/// This file is generated and should not be edited directly.
use super::LanguageData;

#[allow(clippy::type_complexity)]
pub(crate) const OVERVIEW: [LanguageData; 7910] = [
    LanguageData {
        code_3: [97, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghotuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alumu-Tesu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arbëreshë Albanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aranadan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambrak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abu' Arapesh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arifama-Miniafia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ankave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Afade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anambé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Algerian Saharan Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pará Arára",
        #[cfg(feature = "local_names")]
        autonym: Some("Ukarãngmã"),
    },
    LanguageData {
        code_3: [97, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Abnaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 114],
        code_1: Some([97, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Afar",
        #[cfg(feature = "local_names")]
        autonym: Some("Qafar"),
    },
    LanguageData {
        code_3: [97, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aasáx",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arvanitika Albanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abau",
        #[cfg(feature = "local_names")]
        autonym: Some("Abau"),
    },
    LanguageData {
        code_3: [97, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Solong",
        #[cfg(feature = "local_names")]
        autonym: Some("Solong"),
    },
    LanguageData {
        code_3: [97, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandobo Atas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amarasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bankon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambala Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manide",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Abnaki",
        #[cfg(feature = "local_names")]
        autonym: Some("Alnôbak"),
    },
    LanguageData {
        code_3: [97, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abai Sungai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tajiki Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abidji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Bea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 107],
        code_1: Some([97, 98]),
        #[cfg(feature = "english_names")]
        name_en: "Abkhazian",
        #[cfg(feature = "local_names")]
        autonym: Some("аҧсуа бызшәа\u{200e} (Aṗsua byzšwa), аҧсшәа\u{200e} (Aṗsšwa)"),
    },
    LanguageData {
        code_3: [97, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lampung Nyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abanyom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abua",
        #[cfg(feature = "local_names")]
        autonym: Some("Abuan"),
    },
    LanguageData {
        code_3: [97, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abellen Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abaza",
        #[cfg(feature = "local_names")]
        autonym: Some("Abaza, абаза\u{200e} (Abaza), абаза бызшва\u{200e} (Abaza byzšva)"),
    },
    LanguageData {
        code_3: [97, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abron",
        #[cfg(feature = "local_names")]
        autonym: Some("Brong"),
    },
    LanguageData {
        code_3: [97, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambonese Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Ambon, Bahasa Melaju Ambon"),
    },
    LanguageData {
        code_3: [97, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambulas",
        #[cfg(feature = "local_names")]
        autonym: Some("Ambulas"),
    },
    LanguageData {
        code_3: [97, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abure",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baharna Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("العربية البحرانية\u{200e} (Arab Bahraini)"),
    },
    LanguageData {
        code_3: [97, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inabaknon",
        #[cfg(feature = "local_names")]
        autonym: Some("Inabaknon"),
    },
    LanguageData {
        code_3: [97, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aneme Wake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achagua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Áncá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gikyode",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achinese",
        #[cfg(feature = "local_names")]
        autonym: Some("Aceh, Basa Acèh"),
    },
    LanguageData {
        code_3: [97, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saint Lucian Creole French",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwéyòl, Patwa"),
    },
    LanguageData {
        code_3: [97, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Acoli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Cari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Kora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akar-Bale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mesopotamian Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("عراقي\u{200e} (ʕirāgi)"),
    },
    LanguageData {
        code_3: [97, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achang",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngochang"),
    },
    LanguageData {
        code_3: [97, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Acipa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ta'izzi-Adeni Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achi",
        #[cfg(feature = "local_names")]
        autonym: Some("Qach’a’teem"),
    },
    LanguageData {
        code_3: [97, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Acroá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achterhoeks",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achuar-Shiwiar",
        #[cfg(feature = "local_names")]
        autonym: Some("Achuar Chicham, Shiwiar Chicham"),
    },
    LanguageData {
        code_3: [97, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Achumawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hijazi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omani Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cypriot Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("Sanna"),
    },
    LanguageData {
        code_3: [97, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Acheron",
        #[cfg(feature = "local_names")]
        autonym: Some("Garme"),
    },
    LanguageData {
        code_3: [97, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adangme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atauran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lidzonka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adele",
        #[cfg(feature = "local_names")]
        autonym: Some("Gidire"),
    },
    LanguageData {
        code_3: [97, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhofari Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andegerebinha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adhola",
        #[cfg(feature = "local_names")]
        autonym: Some("Dhopadhola"),
    },
    LanguageData {
        code_3: [97, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adi",
        #[cfg(feature = "local_names")]
        autonym: Some("Adi"),
    },
    LanguageData {
        code_3: [97, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adioukrou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adangbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adonara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adamorobe Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adnyamathanha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aduge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amundava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amdo Tibetan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adyghe",
        #[cfg(feature = "local_names")]
        autonym: Some("Адыгабзэ\u{200e} (Adəgăbză)"),
    },
    LanguageData {
        code_3: [97, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adzera",
        #[cfg(feature = "local_names")]
        autonym: Some("Adzera"),
    },
    LanguageData {
        code_3: [97, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Areba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunisian Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("Derja, تونسي\u{200e} (Tounsi)"),
    },
    LanguageData {
        code_3: [97, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saidi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Argentine Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeast Pashai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haeke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Armenian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Arrernte",
        #[cfg(feature = "local_names")]
        autonym: Some("Arrernte"),
    },
    LanguageData {
        code_3: [97, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alsea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akeu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambakich",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amele",
        #[cfg(feature = "local_names")]
        autonym: Some("Amele"),
    },
    LanguageData {
        code_3: [97, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aeka",
        #[cfg(feature = "local_names")]
        autonym: Some("Aeka"),
    },
    LanguageData {
        code_3: [97, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gulf Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("خليجي\u{200e} (Khaliji)"),
    },
    LanguageData {
        code_3: [97, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Putukwam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Afghan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Afrihili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akrukay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanubae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Defaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eloyi",
        #[cfg(feature = "local_names")]
        autonym: Some("Eloyi"),
    },
    LanguageData {
        code_3: [97, 102, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tapei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 114],
        code_1: Some([97, 102]),
        #[cfg(feature = "english_names")]
        name_en: "Afrikaans",
        #[cfg(feature = "local_names")]
        autonym: Some("Afrikaans"),
    },
    LanguageData {
        code_3: [97, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Afro-Seminole Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Afitti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awutu",
        #[cfg(feature = "local_names")]
        autonym: Some("Awutu"),
    },
    LanguageData {
        code_3: [97, 102, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obokuitai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aguano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Legbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agatu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agarabi",
        #[cfg(feature = "local_names")]
        autonym: Some("Agarabi"),
    },
    LanguageData {
        code_3: [97, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angal",
        #[cfg(feature = "local_names")]
        autonym: Some("Angal Heneng"),
    },
    LanguageData {
        code_3: [97, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arguni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngelima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agariya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Argobba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isarog Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fembe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angaataha",
        #[cfg(feature = "local_names")]
        autonym: Some("Angaataha"),
    },
    LanguageData {
        code_3: [97, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agutaynen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tainae",
        #[cfg(feature = "local_names")]
        autonym: Some("Tainae"),
    },
    LanguageData {
        code_3: [97, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aghem",
        #[cfg(feature = "local_names")]
        autonym: Some("Aghem"),
    },
    LanguageData {
        code_3: [97, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aguaruna",
        #[cfg(feature = "local_names")]
        autonym: Some("Awajún"),
    },
    LanguageData {
        code_3: [97, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Esimbi",
        #[cfg(feature = "local_names")]
        autonym: Some("Esimbi"),
    },
    LanguageData {
        code_3: [97, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Cagayan Agta",
        #[cfg(feature = "local_names")]
        autonym: Some("Agta"),
    },
    LanguageData {
        code_3: [97, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aguacateco",
        #[cfg(feature = "local_names")]
        autonym: Some("Qa’yol"),
    },
    LanguageData {
        code_3: [97, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Remontado Dumagat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kahua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aghul",
        #[cfg(feature = "local_names")]
        autonym: Some("агъул\u{200e} (Aġul), агъул чӀал\u{200e} (Ağul ҫ\u{307}al)"),
    },
    LanguageData {
        code_3: [97, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Alta",
        #[cfg(feature = "local_names")]
        autonym: Some("Alta, Kaboloan"),
    },
    LanguageData {
        code_3: [97, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mt. Iriga Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ahanta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Axamb",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qimant",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aghu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiagbamrin Aizi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akha",
        #[cfg(feature = "local_names")]
        autonym: Some("A˯ka˯daw˯\u{200e} (Avkavdawv)"),
    },
    LanguageData {
        code_3: [97, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Igo",
        #[cfg(feature = "local_names")]
        autonym: Some("Igo"),
    },
    LanguageData {
        code_3: [97, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mobumrin Aizi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Àhàn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ahom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aproumu Aizi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ahirani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ashe",
        #[cfg(feature = "local_names")]
        autonym: Some("Ishe"),
    },
    LanguageData {
        code_3: [97, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ahtena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arosi",
        #[cfg(feature = "local_names")]
        autonym: Some("Arosi"),
    },
    LanguageData {
        code_3: [97, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ainu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ainbai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alngith",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Antigua and Barbuda Creole English",
        #[cfg(feature = "local_names")]
        autonym: Some("Creole"),
    },
    LanguageData {
        code_3: [97, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ai-Cham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Assyrian Neo-Aramaic",
        #[cfg(feature = "local_names")]
        autonym: Some("ܐܬܘܪܝܐ\u{200e} (Ātūrāyā), ܣܘܪܬ\u{200e} (Sūrët)"),
    },
    LanguageData {
        code_3: [97, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lishanid Noshan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aimele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aimol",
        #[cfg(feature = "local_names")]
        autonym: Some("Aimol"),
    },
    LanguageData {
        code_3: [97, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ainu",
        #[cfg(feature = "local_names")]
        autonym: Some("アイヌ\u{200e} (Aynu), アイヌ イタㇰ\u{200e} (Aynu itak)"),
    },
    LanguageData {
        code_3: [97, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aiton",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burumakok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aimaq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Airoran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arikem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aari",
        #[cfg(feature = "local_names")]
        autonym: Some("Aari"),
    },
    LanguageData {
        code_3: [97, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aighon",
        #[cfg(feature = "local_names")]
        autonym: Some("Aighon, Apsokok"),
    },
    LanguageData {
        code_3: [97, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aja",
        #[cfg(feature = "local_names")]
        autonym: Some("Ajagbe"),
    },
    LanguageData {
        code_3: [97, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ajië",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andajin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Levantine Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Algerian Jewish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Moroccan Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ajawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 106, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amri Karbi",
        #[cfg(feature = "local_names")]
        autonym: Some("Amri Karbi"),
    },
    LanguageData {
        code_3: [97, 107, 97],
        code_1: Some([97, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Akan",
        #[cfg(feature = "local_names")]
        autonym: Some("Akan"),
    },
    LanguageData {
        code_3: [97, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Angkola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukpet-Ehom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akawaio",
        #[cfg(feature = "local_names")]
        autonym: Some("Kapon"),
    },
    LanguageData {
        code_3: [97, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anakalangu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angal Heneng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aiome",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Jeru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akkadian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aklanon",
        #[cfg(feature = "local_names")]
        autonym: Some("Inakeanon"),
    },
    LanguageData {
        code_3: [97, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Bo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akurio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siwu",
        #[cfg(feature = "local_names")]
        autonym: Some("Siwu"),
    },
    LanguageData {
        code_3: [97, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Araki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akaselem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akolet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akhvakh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Kede",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka-Kol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alabama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alago",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qawasqar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alladian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aleut",
        #[cfg(feature = "local_names")]
        autonym: Some("Unangam tunnu"),
    },
    LanguageData {
        code_3: [97, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alege",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amaimon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Allar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amblong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gheg Albanian",
        #[cfg(feature = "local_names")]
        autonym: Some("Gegnisht-Shqyp, Shqyp"),
    },
    LanguageData {
        code_3: [97, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Larike-Wakasihu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alune",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Algonquin",
        #[cfg(feature = "local_names")]
        autonym: Some("Anishinaabemowin"),
    },
    LanguageData {
        code_3: [97, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alutor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tosk Albanian",
        #[cfg(feature = "local_names")]
        autonym: Some("Shqip"),
    },
    LanguageData {
        code_3: [97, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Altai",
        #[cfg(feature = "local_names")]
        autonym: Some("алтай тили\u{200e} (Altay tili), алтайча\u{200e} (Altajča)"),
    },
    LanguageData {
        code_3: [97, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "'Are'are",
        #[cfg(feature = "local_names")]
        autonym: Some("’Are’are"),
    },
    LanguageData {
        code_3: [97, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alaba-K’abeena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alyawarr",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alur",
        #[cfg(feature = "local_names")]
        autonym: Some("DhuAlur"),
    },
    LanguageData {
        code_3: [97, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amanayé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amahuaca",
        #[cfg(feature = "local_names")]
        autonym: Some("Yora"),
    },
    LanguageData {
        code_3: [97, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanesha'",
        #[cfg(feature = "local_names")]
        autonym: Some("Yanesha’"),
    },
    LanguageData {
        code_3: [97, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hamer-Banna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amurdak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 104],
        code_1: Some([97, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Amharic",
        #[cfg(feature = "local_names")]
        autonym: Some("አማርኛ"),
    },
    LanguageData {
        code_3: [97, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amis",
        #[cfg(feature = "local_names")]
        autonym: Some("Amis"),
    },
    LanguageData {
        code_3: [97, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amdang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "War-Jaintia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amanab",
        #[cfg(feature = "local_names")]
        autonym: Some("Awai"),
    },
    LanguageData {
        code_3: [97, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amo",
        #[cfg(feature = "local_names")]
        autonym: Some("Timap"),
    },
    LanguageData {
        code_3: [97, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alamblak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amahai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amarakaeri",
        #[cfg(feature = "local_names")]
        autonym: Some("Amarakaeri, Harakmbut"),
    },
    LanguageData {
        code_3: [97, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Amami-Oshima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guerrero Amuzgo",
        #[cfg(feature = "local_names")]
        autonym: Some("Ñomndaa"),
    },
    LanguageData {
        code_3: [97, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambelau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Neo-Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anmatyerre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atampaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andaqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andoa",
        #[cfg(feature = "local_names")]
        autonym: Some("Simikai"),
    },
    LanguageData {
        code_3: [97, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ansus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xârâcùù",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Animere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nend",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andi",
        #[cfg(feature = "local_names")]
        autonym: Some("къIaваннаб мицци\u{200e} (Qwavannab Micci), мицци\u{200e} (Micci)"),
    },
    LanguageData {
        code_3: [97, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goemai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anu-Hkongso Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obolo",
        #[cfg(feature = "local_names")]
        autonym: Some("Obolo"),
    },
    LanguageData {
        code_3: [97, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angika",
        #[cfg(feature = "local_names")]
        autonym: Some("अ\u{902}गिका\u{200e} (Angika)"),
    },
    LanguageData {
        code_3: [97, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jarawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anserma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Antakarinya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anuak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Denya",
        #[cfg(feature = "local_names")]
        autonym: Some("Denya"),
    },
    LanguageData {
        code_3: [97, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anaang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andra-Hus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anyin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angolar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pemon",
        #[cfg(feature = "local_names")]
        autonym: Some("Kamarakotos"),
    },
    LanguageData {
        code_3: [97, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andarum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angal Enen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bragat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angoram",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anindilyakwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mufian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arhö",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ömie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bumbita Arapesh",
        #[cfg(feature = "local_names")]
        autonym: Some("Weri"),
    },
    LanguageData {
        code_3: [97, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taikat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "A'ou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atorada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uab Meto",
        #[cfg(feature = "local_names")]
        autonym: Some("Uab Meto"),
    },
    LanguageData {
        code_3: [97, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sa'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Levantine Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("اللهجة العربي\u{651}ة السورى\u{651}ة\u{200e} (El-lahjeẗ el-‛arabīyeẗ es-sūrīyé)"),
    },
    LanguageData {
        code_3: [97, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sudanese Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("Arabi"),
    },
    LanguageData {
        code_3: [97, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukiyip",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahanan Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ampanang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Athpariya",
        #[cfg(feature = "local_names")]
        autonym: Some("आठपहरिया\u{200e} (Āṭhapahariyā)"),
    },
    LanguageData {
        code_3: [97, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apiaká",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jicarilla Apache",
        #[cfg(feature = "local_names")]
        autonym: Some("Abáachi mizaa, Jicarilla Abáachi"),
    },
    LanguageData {
        code_3: [97, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiowa Apache",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lipan Apache",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mescalero-Chiricahua Apache",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndee bizaa"),
    },
    LanguageData {
        code_3: [97, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apinayé",
        #[cfg(feature = "local_names")]
        autonym: Some("Apinayé"),
    },
    LanguageData {
        code_3: [97, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apma",
        #[cfg(feature = "local_names")]
        autonym: Some("Apma, Dalekte"),
    },
    LanguageData {
        code_3: [97, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "A-Pucikwar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arop-Lokep",
        #[cfg(feature = "local_names")]
        autonym: Some("Arop-Lokep"),
    },
    LanguageData {
        code_3: [97, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arop-Sissano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apatani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apurinã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alapmunte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Apache",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndee biyáti’"),
    },
    LanguageData {
        code_3: [97, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aputai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apalaí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Safeyoka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Archi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ampari Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arigidi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atohwaim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Alta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atakapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arhâ",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angaité",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 113, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akuntsu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 97],
        code_1: Some([97, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("العربية"),
    },
    LanguageData {
        code_3: [97, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Standard Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("العربية\u{200e} (al-ʻArabīyah)"),
    },
    LanguageData {
        code_3: [97, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Official Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arabana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Arrarnta",
        #[cfg(feature = "local_names")]
        autonym: Some("Arrente"),
    },
    LanguageData {
        code_3: [97, 114, 103],
        code_1: Some([97, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Aragonese",
        #[cfg(feature = "local_names")]
        autonym: Some("Aragonés"),
    },
    LanguageData {
        code_3: [97, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arhuaco",
        #[cfg(feature = "local_names")]
        autonym: Some("Iku"),
    },
    LanguageData {
        code_3: [97, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arikara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arapaso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arikapú",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arabela",
        #[cfg(feature = "local_names")]
        autonym: Some("Tapueyocaca"),
    },
    LanguageData {
        code_3: [97, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapudungun",
        #[cfg(feature = "local_names")]
        autonym: Some("Mapudungun"),
    },
    LanguageData {
        code_3: [97, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Araona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arapaho",
        #[cfg(feature = "local_names")]
        autonym: Some("Hinónoʼeitíít"),
    },
    LanguageData {
        code_3: [97, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Algerian Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("دزيرية\u{200e} (Dziria)"),
    },
    LanguageData {
        code_3: [97, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Najdi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aruá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arbore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arawak",
        #[cfg(feature = "local_names")]
        autonym: Some("Arawak, Lokono"),
    },
    LanguageData {
        code_3: [97, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aruá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moroccan Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("الدارجة\u{200e} (Darija)"),
    },
    LanguageData {
        code_3: [97, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Egyptian Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("مصري\u{200e} (Masri)"),
    },
    LanguageData {
        code_3: [97, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asu",
        #[cfg(feature = "local_names")]
        autonym: Some("Kipare"),
    },
    LanguageData {
        code_3: [97, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Assiniboine",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Casuarina Coast Asmat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "American Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Auslan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cishingini",
        #[cfg(feature = "local_names")]
        autonym: Some("Cishingini"),
    },
    LanguageData {
        code_3: [97, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abishira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buruwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sari",
        #[cfg(feature = "local_names")]
        autonym: Some("Sari"),
    },
    LanguageData {
        code_3: [97, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ashkun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asilulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 109],
        code_1: Some([97, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Assamese",
        #[cfg(feature = "local_names")]
        autonym: Some("অসমীয\u{9bc}\u{9be}"),
    },
    LanguageData {
        code_3: [97, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xingú Asuriní",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Algerian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Austrian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ipulo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asturian",
        #[cfg(feature = "local_names")]
        autonym: Some("asturianu"),
    },
    LanguageData {
        code_3: [97, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tocantins Asurini",
        #[cfg(feature = "local_names")]
        autonym: Some("Asuriní"),
    },
    LanguageData {
        code_3: [97, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Australian Aborigines Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muratayak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaosakor Asmat",
        #[cfg(feature = "local_names")]
        autonym: Some("Asmat"),
    },
    LanguageData {
        code_3: [97, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "As",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pele-Ata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atsahuaca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ata Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atemble",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ivbie North-Okpela-Arhe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Attié",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atikamekw",
        #[cfg(feature = "local_names")]
        autonym: Some("Atikamekw"),
    },
    LanguageData {
        code_3: [97, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mt. Iraya Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ashtiani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pudtol Atta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aralle-Tabulahan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waimiri-Atroari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gros Ventre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pamplona Atta",
        #[cfg(feature = "local_names")]
        autonym: Some("Atta"),
    },
    LanguageData {
        code_3: [97, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Reel",
        #[cfg(feature = "local_names")]
        autonym: Some("Thok Reel"),
    },
    LanguageData {
        code_3: [97, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Altai",
        #[cfg(feature = "local_names")]
        autonym: Some("Алтай тили\u{200e} (Altay tili)"),
    },
    LanguageData {
        code_3: [97, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atsugewi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arutani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aneityum",
        #[cfg(feature = "local_names")]
        autonym: Some("Anejom, Intas Anejom"),
    },
    LanguageData {
        code_3: [97, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asumboa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alugu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waorani",
        #[cfg(feature = "local_names")]
        autonym: Some("Wao Tededö"),
    },
    LanguageData {
        code_3: [97, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anuta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aguna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aushi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anuki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awjilah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Heyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aulua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molmo One",
        #[cfg(feature = "local_names")]
        autonym: Some("One"),
    },
    LanguageData {
        code_3: [97, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Auyokawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makayam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aruek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Austral",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Auye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aurá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awiyaana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uzbeki Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 97],
        code_1: Some([97, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Avaric",
        #[cfg(feature = "local_names")]
        autonym: Some("авар мацӏ\u{200e} (Awar mac\u{323}), магӏарул мацӏ\u{200e} (Maʿarul mac\u{323})"),
    },
    LanguageData {
        code_3: [97, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Avau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alviri-Vidari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 101],
        code_1: Some([97, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Avestan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Avikam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kotava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Egyptian Bedawi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angkamuthi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Avatime",
        #[cfg(feature = "local_names")]
        autonym: Some("Si-yà"),
    },
    LanguageData {
        code_3: [97, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agavotaguerra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aushiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Au",
        #[cfg(feature = "local_names")]
        autonym: Some("Au"),
    },
    LanguageData {
        code_3: [97, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Avokaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Avá-Canoeiro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awadhi",
        #[cfg(feature = "local_names")]
        autonym: Some("अवधी\u{200e} (Awadhi)"),
    },
    LanguageData {
        code_3: [97, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cicipu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awetí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anguthimri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awbono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aekyom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awabakal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arawum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awngi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Araweté",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jair Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Edera Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abipon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayerrerenge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mato Grosso Arára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaka",
        #[cfg(feature = "local_names")]
        autonym: Some("Aka"),
    },
    LanguageData {
        code_3: [97, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Southern Aranda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Armenian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xârâgurè",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayizo Gbe",
        #[cfg(feature = "local_names")]
        autonym: Some("Ayizo Gbe"),
    },
    LanguageData {
        code_3: [97, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Aymara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayabadhu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ginyanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hadrami Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leyigha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akuku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Libyan Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("ليبي\u{200e} (Lībi)"),
    },
    LanguageData {
        code_3: [97, 121, 109],
        code_1: Some([97, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Aymara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanaani Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayoreo",
        #[cfg(feature = "local_names")]
        autonym: Some("Morotoco"),
    },
    LanguageData {
        code_3: [97, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Mesopotamian Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Aymara",
        #[cfg(feature = "local_names")]
        autonym: Some("Aymar"),
    },
    LanguageData {
        code_3: [97, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sorsogon Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magbukun Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mai Brat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Azha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Azerbaijani",
        #[cfg(feature = "local_names")]
        autonym: Some("آذربایجان دیلی\u{200e} (Azərbaycan dili), آذربایجانجا\u{200e} (Azərbaycanca)"),
    },
    LanguageData {
        code_3: [97, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Durango Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: Some("Meshikan de San Pedro Shikora"),
    },
    LanguageData {
        code_3: [97, 122, 101],
        code_1: Some([97, 122]),
        #[cfg(feature = "english_names")]
        name_en: "Azerbaijani",
        #[cfg(feature = "local_names")]
        autonym: Some("azərbaycan"),
    },
    LanguageData {
        code_3: [97, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Pedro Amuzgos Amuzgo",
        #[cfg(feature = "local_names")]
        autonym: Some("Jnꞌoon"),
    },
    LanguageData {
        code_3: [97, 122, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Azerbaijani",
        #[cfg(feature = "local_names")]
        autonym: Some("азәрбајҹан дили\u{200e} (Azərbaycan dili), азәрбајҹанҹа\u{200e} (azərbaycanca)"),
    },
    LanguageData {
        code_3: [97, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ipalapa Amuzgo",
        #[cfg(feature = "local_names")]
        autonym: Some("Jñunda"),
    },
    LanguageData {
        code_3: [97, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Durango Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: Some("Meshikan de San Agustin Buenaventura y de Santa Cruz"),
    },
    LanguageData {
        code_3: [97, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awing",
        #[cfg(feature = "local_names")]
        autonym: Some("Atembuluwe"),
    },
    LanguageData {
        code_3: [97, 122, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Faire Atta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [97, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Highland Puebla Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: Some("Mexicano, Nahuatlahtolli"),
    },
    LanguageData {
        code_3: [98, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Babatana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bainouk-Gunyuño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baré",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nubaca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahamas Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barakai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 107],
        code_1: Some([98, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Bashkir",
        #[cfg(feature = "local_names")]
        autonym: Some("башҡорт теле\u{200e} (Başķort tele), башҡортса\u{200e} (Başķortsa)"),
    },
    LanguageData {
        code_3: [98, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baluchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 109],
        code_1: Some([98, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Bambara",
        #[cfg(feature = "local_names")]
        autonym: Some("bamanakan"),
    },
    LanguageData {
        code_3: [98, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balinese",
        #[cfg(feature = "local_names")]
        autonym: Some("Basa Bali, Basa Bali"),
    },
    LanguageData {
        code_3: [98, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waimaha",
        #[cfg(feature = "local_names")]
        autonym: Some("Waimaja"),
    },
    LanguageData {
        code_3: [98, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bantawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bavarian",
        #[cfg(feature = "local_names")]
        autonym: Some("Boarisch"),
    },
    LanguageData {
        code_3: [98, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basa",
        #[cfg(feature = "local_names")]
        autonym: Some("Ɓàsàa"),
    },
    LanguageData {
        code_3: [98, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bambili-Bambui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamun",
        #[cfg(feature = "local_names")]
        autonym: Some("Shüpamom"),
    },
    LanguageData {
        code_3: [98, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batuley",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baatonum",
        #[cfg(feature = "local_names")]
        autonym: Some("Baatonum"),
    },
    LanguageData {
        code_3: [98, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Toba",
        #[cfg(feature = "local_names")]
        autonym: Some("Batta"),
    },
    LanguageData {
        code_3: [98, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baibai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bugan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barombi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghomálá'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Babanki",
        #[cfg(feature = "local_names")]
        autonym: Some("Nga-Kejom"),
    },
    LanguageData {
        code_3: [98, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bats",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Babango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uneapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Bobo Madaré",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Central Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamali",
        #[cfg(feature = "local_names")]
        autonym: Some("Chopechop"),
    },
    LanguageData {
        code_3: [98, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Girawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakpinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mburku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karnai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baba",
        #[cfg(feature = "local_names")]
        autonym: Some("Supapyak’"),
    },
    LanguageData {
        code_3: [98, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bubia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Befang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Bai",
        #[cfg(feature = "local_names")]
        autonym: Some("白语\u{200e} (Baip‧ngvp‧zix)"),
    },
    LanguageData {
        code_3: [98, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bainouk-Samik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Balochi",
        #[cfg(feature = "local_names")]
        autonym: Some("بلوچی\u{200e} (Balōčī), بلوچی زبان\u{200e} (Balōčī Zubān)"),
    },
    LanguageData {
        code_3: [98, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Babar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamenyam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Pokur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bariai",
        #[cfg(feature = "local_names")]
        autonym: Some("Bariai"),
    },
    LanguageData {
        code_3: [98, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baoulé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bardi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bannoni",
        #[cfg(feature = "local_names")]
        autonym: Some("Tsunari"),
    },
    LanguageData {
        code_3: [98, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bali",
        #[cfg(feature = "local_names")]
        autonym: Some("Abaali, Maya"),
    },
    LanguageData {
        code_3: [98, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaluli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bench",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Babine",
        #[cfg(feature = "local_names")]
        autonym: Some("Nedut’en"),
    },
    LanguageData {
        code_3: [98, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kohumono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bendi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awad Bing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shoo-Minda-Nye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bacama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bainouk-Gunyaamolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bayot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emberá-Baudó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bade",
        #[cfg(feature = "local_names")]
        autonym: Some("Gabaden"),
    },
    LanguageData {
        code_3: [98, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biage",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonggi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budukh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indonesian Bajau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buduma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baldemu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahnar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Coast Bajau",
        #[cfg(feature = "local_names")]
        autonym: Some("Sama"),
    },
    LanguageData {
        code_3: [98, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burunge",
        #[cfg(feature = "local_names")]
        autonym: Some("Burungaisoo"),
    },
    LanguageData {
        code_3: [98, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bokoto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oroko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bodo Parja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budong-Budong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bandjalang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badeshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beaver",
        #[cfg(feature = "local_names")]
        autonym: Some("Danezaa ZaageɁ\u{200e} (Danezaa Zaage)"),
    },
    LanguageData {
        code_3: [98, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bebele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iceve-Maci",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bedoanas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Byangsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Benabena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belait",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biali",
        #[cfg(feature = "local_names")]
        autonym: Some("Biali"),
    },
    LanguageData {
        code_3: [98, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bekati'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beja",
        #[cfg(feature = "local_names")]
        autonym: Some("Bedàwie, Tu Bdhaawi"),
    },
    LanguageData {
        code_3: [98, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bebeli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 108],
        code_1: Some([98, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Belarusian",
        #[cfg(feature = "local_names")]
        autonym: Some("беларуская"),
    },
    LanguageData {
        code_3: [98, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bemba",
        #[cfg(feature = "local_names")]
        autonym: Some("Ichibemba"),
    },
    LanguageData {
        code_3: [98, 101, 110],
        code_1: Some([98, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Bengali",
        #[cfg(feature = "local_names")]
        autonym: Some("ব\u{9be}ংল\u{9be}"),
    },
    LanguageData {
        code_3: [98, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Besoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beembe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Besme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guiberoua Béte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Blagar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daloa Bété",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Betawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jur Modo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bena",
        #[cfg(feature = "local_names")]
        autonym: Some("Hibena"),
    },
    LanguageData {
        code_3: [98, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bari",
        #[cfg(feature = "local_names")]
        autonym: Some("Kuku"),
    },
    LanguageData {
        code_3: [98, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pauri Bareli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panyi Bai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bafut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Betaf",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bofi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busang Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Blafe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "British Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bafanji",
        #[cfg(feature = "local_names")]
        autonym: Some("Chufie’"),
    },
    LanguageData {
        code_3: [98, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ban Khor Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda-Ndélé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mmen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malba Birifor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badaga",
        #[cfg(feature = "local_names")]
        autonym: Some("Badaga, ಬಡಗ\u{200e} (Badaga)"),
    },
    LanguageData {
        code_3: [98, 102, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bazigar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Bai",
        #[cfg(feature = "local_names")]
        autonym: Some("白语\u{200e} (Baip‧ngvp‧zix)"),
    },
    LanguageData {
        code_3: [98, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balti",
        #[cfg(feature = "local_names")]
        autonym: Some("بلتی\u{200e} (Baltī), ས\u{fa6}ལ་འཐ\u{f74}ས་\u{200e} (Balti)"),
    },
    LanguageData {
        code_3: [98, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gahri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bondo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bantayanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 102, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagheli",
        #[cfg(feature = "local_names")]
        autonym: Some("बघ\u{947}ली\u{200e} (Bagheli)"),
    },
    LanguageData {
        code_3: [98, 102, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mahasu Pahari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwamhi-Wuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bobongko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haryanvi",
        #[cfg(feature = "local_names")]
        autonym: Some("हरियाणवी\u{200e} (Hariyāṇvī), हि\u{902}दी\u{200e} (Hindi)"),
    },
    LanguageData {
        code_3: [98, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rathwi Bareli",
        #[cfg(feature = "local_names")]
        autonym: Some("बर\u{947}ली\u{200e} (Bareli)"),
    },
    LanguageData {
        code_3: [98, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bauria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangandu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bugun",
        #[cfg(feature = "local_names")]
        autonym: Some("Bugun"),
    },
    LanguageData {
        code_3: [98, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Giangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangolan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Balochi",
        #[cfg(feature = "local_names")]
        autonym: Some("بلوچی (رخشانی)"),
    },
    LanguageData {
        code_3: [98, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Koga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Balochi",
        #[cfg(feature = "local_names")]
        autonym: Some("بلوچی\u{200e} (Balōčī), بلوچی زبان\u{200e} (Balōčī Zubān)"),
    },
    LanguageData {
        code_3: [98, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagri",
        #[cfg(feature = "local_names")]
        autonym: Some("बागड\u{93c}ी\u{200e} (Bagri)"),
    },
    LanguageData {
        code_3: [98, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bawm Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagabawa",
        #[cfg(feature = "local_names")]
        autonym: Some("Bagobo"),
    },
    LanguageData {
        code_3: [98, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bughotu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbongno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warkay-Bipim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhatri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balkan Gagauz Turkish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Benggoi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banggai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bharia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhili",
        #[cfg(feature = "local_names")]
        autonym: Some("भीली\u{200e} (Bhīlī)"),
    },
    LanguageData {
        code_3: [98, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhadrawahi",
        #[cfg(feature = "local_names")]
        autonym: Some("भद\u{94d}रवाही\u{200e} (Bhadrawahi)"),
    },
    LanguageData {
        code_3: [98, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Odiai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binandere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukharic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhilali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bimin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bathari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bohtan Neo-Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhojpuri",
        #[cfg(feature = "local_names")]
        autonym: Some("भोजप\u{941}री\u{200e} (Bhōjpurī)"),
    },
    LanguageData {
        code_3: [98, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tukang Besi South",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bara Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buwal",
        #[cfg(feature = "local_names")]
        autonym: Some("Buwal"),
    },
    LanguageData {
        code_3: [98, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhattiyali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhunjia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biak",
        #[cfg(feature = "local_names")]
        autonym: Some("Wós Kovedi, Wós Vyak"),
    },
    LanguageData {
        code_3: [98, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhalay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badimaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bissa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bidiyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bepour",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biafada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biangai",
        #[cfg(feature = "local_names")]
        autonym: Some("Biangai"),
    },
    LanguageData {
        code_3: [98, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bile",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bimoba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bini",
        #[cfg(feature = "local_names")]
        autonym: Some("Edo"),
    },
    LanguageData {
        code_3: [98, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bipi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bisorio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 115],
        code_1: Some([98, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Bislama",
        #[cfg(feature = "local_names")]
        autonym: Some("Bislama"),
    },
    LanguageData {
        code_3: [98, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berinomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Birifor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kol",
        #[cfg(feature = "local_names")]
        autonym: Some("Bòkól"),
    },
    LanguageData {
        code_3: [98, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bijori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birhor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baloi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banggarla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bariji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biao-Jiao Mien",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barzani Jewish Neo-Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bidyogo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahinemo",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahinemo"),
    },
    LanguageData {
        code_3: [98, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanauji",
        #[cfg(feature = "local_names")]
        autonym: Some("द\u{947}हाती\u{200e} (Dehati), हिन\u{94d}दी\u{200e} (Hindi)"),
    },
    LanguageData {
        code_3: [98, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bajelani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banjar",
        #[cfg(feature = "local_names")]
        autonym: Some("بنجر\u{200e} (Banjar)"),
    },
    LanguageData {
        code_3: [98, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mid-Southern Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fanamaket",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binumarien",
        #[cfg(feature = "local_names")]
        autonym: Some("Binumarien"),
    },
    LanguageData {
        code_3: [98, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bajan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balanta-Ganja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busuu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bedjond",
        #[cfg(feature = "local_names")]
        autonym: Some("Bedjond, Nangnda"),
    },
    LanguageData {
        code_3: [98, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakwé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banao Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bayali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 106, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baruga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binukid",
        #[cfg(feature = "local_names")]
        autonym: Some("Binukid"),
    },
    LanguageData {
        code_3: [98, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beeke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buraka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakoko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baki",
        #[cfg(feature = "local_names")]
        autonym: Some("Baki"),
    },
    LanguageData {
        code_3: [98, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pande",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brokskat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kom",
        #[cfg(feature = "local_names")]
        autonym: Some("Itangikom"),
    },
    LanguageData {
        code_3: [98, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukitan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwa'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakairí",
        #[cfg(feature = "local_names")]
        autonym: Some("Kurâ"),
    },
    LanguageData {
        code_3: [98, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakumpai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Sorsoganon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boloki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buhid",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bekwarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bekwel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baikeno",
        #[cfg(feature = "local_names")]
        autonym: Some("Baikeno, Uab Metô"),
    },
    LanguageData {
        code_3: [98, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bokyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bungku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siksika",
        #[cfg(feature = "local_names")]
        autonym: Some("ᓱᖽᐧᖿ\u{200e} (Siksiká)"),
    },
    LanguageData {
        code_3: [98, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bella Coola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balanta-Kentohe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buol",
        #[cfg(feature = "local_names")]
        autonym: Some("Apadu Vuolo"),
    },
    LanguageData {
        code_3: [98, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuwaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolongan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pa'o Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biloxi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beli",
        #[cfg(feature = "local_names")]
        autonym: Some("Beli"),
    },
    LanguageData {
        code_3: [98, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Catanduanes Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anii",
        #[cfg(feature = "local_names")]
        autonym: Some("Anii"),
    },
    LanguageData {
        code_3: [98, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Blablanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baluan-Pam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Blang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balaesang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Dam",
        #[cfg(feature = "local_names")]
        autonym: Some("ꪼꪕꪒ\u{aabe}"),
    },
    LanguageData {
        code_3: [98, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kibala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balangao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mag-Indi Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Notre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balantak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lame",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bembe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Manduri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limassa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bom-Kim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kein",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagirmi",
        #[cfg(feature = "local_names")]
        autonym: Some("Barma"),
    },
    LanguageData {
        code_3: [98, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bote-Majhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghayavi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bomboli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Betsimisaraka Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bambalang",
        #[cfg(feature = "local_names")]
        autonym: Some("Chrambo, Mboyakum"),
    },
    LanguageData {
        code_3: [98, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bulgebi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bomu",
        #[cfg(feature = "local_names")]
        autonym: Some("Bomu, Boré"),
    },
    LanguageData {
        code_3: [98, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muinane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilma Kanuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biao Mon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Somba-Siawari",
        #[cfg(feature = "local_names")]
        autonym: Some("Mindik, Somba-Siawari"),
    },
    LanguageData {
        code_3: [98, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bum",
        #[cfg(feature = "local_names")]
        autonym: Some("Bum"),
    },
    LanguageData {
        code_3: [98, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bomwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baimak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baramu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonerate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bookan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bintauna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masiwang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Benga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Tawbuid",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bierebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunun",
        #[cfg(feature = "local_names")]
        autonym: Some("Bunun"),
    },
    LanguageData {
        code_3: [98, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bantoanon",
        #[cfg(feature = "local_names")]
        autonym: Some("Binisaya, Bisaya"),
    },
    LanguageData {
        code_3: [98, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bantik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Butmas-Tur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bundeli",
        #[cfg(feature = "local_names")]
        autonym: Some("ब\u{941}न\u{94d}द\u{947}ली\u{200e} (Bundeli)"),
    },
    LanguageData {
        code_3: [98, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bentong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonerif",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bisis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangubangu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bintulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beezen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bora",
        #[cfg(feature = "local_names")]
        autonym: Some("Booraa, Miamunaa"),
    },
    LanguageData {
        code_3: [98, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aweer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 100],
        code_1: Some([98, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Tibetan",
        #[cfg(feature = "local_names")]
        autonym: Some("བ\u{f7c}ད་ས\u{f90}ད་"),
    },
    LanguageData {
        code_3: [98, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundabli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamako Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barbareño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anjam",
        #[cfg(feature = "local_names")]
        autonym: Some("Anjam"),
    },
    LanguageData {
        code_3: [98, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonjo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bine",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiemacèwè Bozo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonkiman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bogaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Borôro",
        #[cfg(feature = "local_names")]
        autonym: Some("Boe Wadáru"),
    },
    LanguageData {
        code_3: [98, 111, 115],
        code_1: Some([98, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Bosnian",
        #[cfg(feature = "local_names")]
        autonym: Some("bosanski"),
    },
    LanguageData {
        code_3: [98, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bondei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuwuli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buamu",
        #[cfg(feature = "local_names")]
        autonym: Some("Buamu"),
    },
    LanguageData {
        code_3: [98, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bodo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiéyaxo Bozo",
        #[cfg(feature = "local_names")]
        autonym: Some("Tieyaxo"),
    },
    LanguageData {
        code_3: [98, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daakaka",
        #[cfg(feature = "local_names")]
        autonym: Some("Daakaka"),
    },
    LanguageData {
        code_3: [98, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda-Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bauni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonggo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Botlikh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagupi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orowe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Broome Pearling Lugger Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biyom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dzao Min",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaure",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koronadal Blaan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarangani Blaan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barrow Point",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bongu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bian Marind",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palya Bareli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bishnupriya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tchumbuli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagusa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Kaloum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bago-Kusuntu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakhtiari",
        #[cfg(feature = "local_names")]
        autonym: Some("بختیاری\u{200e} (Bakhtiâri)"),
    },
    LanguageData {
        code_3: [98, 113, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bandial",
        #[cfg(feature = "local_names")]
        autonym: Some("Gubanjalay"),
    },
    LanguageData {
        code_3: [98, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda-Mbrès",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilakura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wumboko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bulgarian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busa",
        #[cfg(feature = "local_names")]
        autonym: Some("Bisã"),
    },
    LanguageData {
        code_3: [98, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biritai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burusu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bosngun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamukumbit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boguru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro Wachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baangi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bengkala Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 113, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Braj",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berbice Creole Dutch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baraamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 101],
        code_1: Some([98, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Breton",
        #[cfg(feature = "local_names")]
        autonym: Some("brezhoneg"),
    },
    LanguageData {
        code_3: [98, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bira",
        #[cfg(feature = "local_names")]
        autonym: Some("Kibira"),
    },
    LanguageData {
        code_3: [98, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baure",
        #[cfg(feature = "local_names")]
        autonym: Some("Baure"),
    },
    LanguageData {
        code_3: [98, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brahui",
        #[cfg(feature = "local_names")]
        autonym: Some("براہوئی\u{200e} (brāhūʾī)"),
    },
    LanguageData {
        code_3: [98, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mokpwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bieria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birked",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barambu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boruca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brokkat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barapasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Breri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baras",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bitare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Bru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Bru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bellari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bodo",
        #[cfg(feature = "local_names")]
        autonym: Some("बड\u{93c}ो"),
    },
    LanguageData {
        code_3: [98, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilbil",
        #[cfg(feature = "local_names")]
        autonym: Some("Bel"),
    },
    LanguageData {
        code_3: [98, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abinomn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brunei Bisaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bassari",
        #[cfg(feature = "local_names")]
        autonym: Some("Oniyan"),
    },
    LanguageData {
        code_3: [98, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wushi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bauchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bashkardi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bassossi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangwinji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burushaski",
        #[cfg(feature = "local_names")]
        autonym: Some("Burushaski, Mishaski, بروشسکی\u{200e} (burū\u{301}šaskī)"),
    },
    LanguageData {
        code_3: [98, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basa-Gumna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barasana-Eduria",
        #[cfg(feature = "local_names")]
        autonym: Some("Barasana-Eduria"),
    },
    LanguageData {
        code_3: [98, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Sitemu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bassa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bassa-Kontagora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akoose",
        #[cfg(feature = "local_names")]
        autonym: Some("Akoose"),
    },
    LanguageData {
        code_3: [98, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basketo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bahonsuai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baga Sobané",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baiso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangkam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabah Bisaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Dairi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamo-Ningi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birgit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gagnoa Bété",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biatah Bidayuh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bacanese Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Mandailing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ratagnon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rinconada Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budibud",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baetora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Simalungun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bete-Bendi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bateri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Butuanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Karo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bobot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak Alas-Kluet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buriat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bushi",
        #[cfg(feature = "local_names")]
        autonym: Some("Kibushi, Shibushi Shimaore"),
    },
    LanguageData {
        code_3: [98, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ntcham",
        #[cfg(feature = "local_names")]
        autonym: Some("Ncham"),
    },
    LanguageData {
        code_3: [98, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beothuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bushoong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buginese",
        #[cfg(feature = "local_names")]
        autonym: Some("Basa Ugi"),
    },
    LanguageData {
        code_3: [98, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Younuo Bunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bongili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basa-Gurmana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bugawac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 108],
        code_1: Some([98, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Bulgarian",
        #[cfg(feature = "local_names")]
        autonym: Some("български"),
    },
    LanguageData {
        code_3: [98, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sherbro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Terei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bokobaru",
        #[cfg(feature = "local_names")]
        autonym: Some("Zogben"),
    },
    LanguageData {
        code_3: [98, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bungain",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boghom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bullom So",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukwen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barein",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bube",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baelelea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baeggu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berau Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonkeng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bure",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belanda Viri",
        #[cfg(feature = "local_names")]
        autonym: Some("Viri"),
    },
    LanguageData {
        code_3: [98, 118, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolivian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bamunka",
        #[cfg(feature = "local_names")]
        autonym: Some("Niemeng"),
    },
    LanguageData {
        code_3: [98, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolgo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bumang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukit Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baniva",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dibole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baybayanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 118, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bauzi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwatoo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namosi-Naitasiri-Serua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwile",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwaidoka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwe Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boselewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bishuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baniwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Láá Láá Bwamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bauwaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biwat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wunai Bunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boro",
        #[cfg(feature = "local_names")]
        autonym: Some("Borna"),
    },
    LanguageData {
        code_3: [98, 119, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandobo Bawah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Bobo Madaré",
        #[cfg(feature = "local_names")]
        autonym: Some("Boboda"),
    },
    LanguageData {
        code_3: [98, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bura-Pabir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bomboma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bafaw-Balong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buli",
        #[cfg(feature = "local_names")]
        autonym: Some("Builsa"),
    },
    LanguageData {
        code_3: [98, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bu-Nao Bunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cwi Bwamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 119, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tairaha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belanda Bor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molengue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangala",
        #[cfg(feature = "local_names")]
        autonym: Some("Bangala"),
    },
    LanguageData {
        code_3: [98, 120, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buhutu",
        #[cfg(feature = "local_names")]
        autonym: Some("Buhutu"),
    },
    LanguageData {
        code_3: [98, 120, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pirlatapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bayungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukusu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jalkunan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mongolia Buriat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burduna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barikanchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bebil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Russia Buriat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Busam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "China Buriat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berakou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bankagooma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 120, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binahari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bikya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ubaghara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Benyadu'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pouye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baygo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhujel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bayono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bidjara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilin",
        #[cfg(feature = "local_names")]
        autonym: Some("ብሊን"),
    },
    LanguageData {
        code_3: [98, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bumaji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baruya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Medumba",
        #[cfg(feature = "local_names")]
        autonym: Some("Medumba"),
    },
    LanguageData {
        code_3: [98, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belhariya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qaqet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banaro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Betsimisaraka Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bribri",
        #[cfg(feature = "local_names")]
        autonym: Some("Bribri"),
    },
    LanguageData {
        code_3: [98, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jenaama Bozo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boikin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Babuza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapos Buang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bisu",
        #[cfg(feature = "local_names")]
        autonym: Some("Mbisu"),
    },
    LanguageData {
        code_3: [98, 122, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belize Kriol English",
        #[cfg(feature = "local_names")]
        autonym: Some("Kriol"),
    },
    LanguageData {
        code_3: [98, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nicaragua Creole English",
        #[cfg(feature = "local_names")]
        autonym: Some("Creole"),
    },
    LanguageData {
        code_3: [98, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolondo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bozaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemberano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brazilian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brithenig",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burmeso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naami",
        #[cfg(feature = "local_names")]
        autonym: Some("Naami"),
    },
    LanguageData {
        code_3: [98, 122, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Basa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kɛlɛngaxo Bozo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obanliku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [98, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Evant",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chortí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garifuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuj",
        #[cfg(feature = "local_names")]
        autonym: Some("Koti’"),
    },
    LanguageData {
        code_3: [99, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caddo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lehar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Carrier",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nivaclé",
        #[cfg(feature = "local_names")]
        autonym: Some("Guisnai"),
    },
    LanguageData {
        code_3: [99, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cahuarano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chané",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaqchikel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carolinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cemuhî",
        #[cfg(feature = "local_names")]
        autonym: Some("Cemuhî"),
    },
    LanguageData {
        code_3: [99, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chambri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chácobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chipaya",
        #[cfg(feature = "local_names")]
        autonym: Some("Chipaya"),
    },
    LanguageData {
        code_3: [99, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Car Nicobarese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galibi Carib",
        #[cfg(feature = "local_names")]
        autonym: Some("Kari’na auran, Kari’ña"),
    },
    LanguageData {
        code_3: [99, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsimané",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 116],
        code_1: Some([99, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Catalan",
        #[cfg(feature = "local_names")]
        autonym: Some("català"),
    },
    LanguageData {
        code_3: [99, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cavineña",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Callawalla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiquitano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cayuga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Canichana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cabiyarí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carapana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carijona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chimila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chachi",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’Palaa"),
    },
    LanguageData {
        code_3: [99, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ede Cabe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chavacano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bualkhaw Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyahkur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Izora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsucuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cashibo-Cacataibo",
        #[cfg(feature = "local_names")]
        autonym: Some("Kakataibo-Kashibo"),
    },
    LanguageData {
        code_3: [99, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cashinahua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chayahuita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Candoshi-Shapra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cacua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinabalian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carabayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chamicuro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cafundo Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chopi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samba Daka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atsam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kasanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cutchi-Swahili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malaccan Creole Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Comaltepec Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jmii’"),
    },
    LanguageData {
        code_3: [99, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chakma",
        #[cfg(feature = "local_names")]
        autonym: Some("𑄌𑄋\u{11134}𑄟\u{11133}𑄦"),
    },
    LanguageData {
        code_3: [99, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cacaopera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Choni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chenchu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chambeali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chodri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Churahi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chepang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chaudangsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Min Dong Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("平話\u{200e} (Bangua), 闽东话\u{200e} (Mindongyu)"),
    },
    LanguageData {
        code_3: [99, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cinda-Regi-Tiyal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chadian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chadong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Chehalis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cebuano",
        #[cfg(feature = "local_names")]
        autonym: Some("Binisaya"),
    },
    LanguageData {
        code_3: [99, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chamacoco",
        #[cfg(feature = "local_names")]
        autonym: Some("Yshyr Ybytoso"),
    },
    LanguageData {
        code_3: [99, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Khumi Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 101, 115],
        code_1: Some([99, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Czech",
        #[cfg(feature = "local_names")]
        autonym: Some("čeština"),
    },
    LanguageData {
        code_3: [99, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Centúúm",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ekai Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dijim-Bwilim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Como Karim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Falam Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Falam Chin"),
    },
    LanguageData {
        code_3: [99, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Changriwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagayanen",
        #[cfg(feature = "local_names")]
        autonym: Some("Kagayanen"),
    },
    LanguageData {
        code_3: [99, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiga",
        #[cfg(feature = "local_names")]
        autonym: Some("Rukiga"),
    },
    LanguageData {
        code_3: [99, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chocangacakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 97],
        code_1: Some([99, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Chamorro",
        #[cfg(feature = "local_names")]
        autonym: Some("Chamorru"),
    },
    LanguageData {
        code_3: [99, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chibcha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Catawba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Highland Oaxaca Chontal",
        #[cfg(feature = "local_names")]
        autonym: Some("Tsome"),
    },
    LanguageData {
        code_3: [99, 104, 101],
        code_1: Some([99, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Chechen",
        #[cfg(feature = "local_names")]
        autonym: Some("нохчийн"),
    },
    LanguageData {
        code_3: [99, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabasco Chontal",
        #[cfg(feature = "local_names")]
        autonym: Some("Yocot’an"),
    },
    LanguageData {
        code_3: [99, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chagatai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinook",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ojitlán Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jmiih kia’ dzä ‘vï ï, Jujmi"),
    },
    LanguageData {
        code_3: [99, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuukese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cahuilla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinook jargon",
        #[cfg(feature = "local_names")]
        autonym: Some("Chinuk Wawa"),
    },
    LanguageData {
        code_3: [99, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Choctaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chipewyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quiotepec Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("juu jmiih"),
    },
    LanguageData {
        code_3: [99, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cherokee",
        #[cfg(feature = "local_names")]
        autonym: Some("ᏣᎳᎩ"),
    },
    LanguageData {
        code_3: [99, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cholón",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 117],
        code_1: Some([99, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Church Slavic",
        #[cfg(feature = "local_names")]
        autonym: Some("церковнослове\u{301}нскїй"),
    },
    LanguageData {
        code_3: [99, 104, 118],
        code_1: Some([99, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Chuvash",
        #[cfg(feature = "local_names")]
        autonym: Some("Чӑвашла"),
    },
    LanguageData {
        code_3: [99, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuwabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chantyal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cheyenne",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ozumacín Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jumi dsa mojai, Juujmii"),
    },
    LanguageData {
        code_3: [99, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cia-Cia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ci Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chickasaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chimariko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cineni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chitkuli Kinnauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cimbrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cinta Larga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiapanec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chippewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chaima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Cham",
        #[cfg(feature = "local_names")]
        autonym: Some("Cham"),
    },
    LanguageData {
        code_3: [99, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Chehalis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chamalal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chokwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Cham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chenapian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ashéninka Pajonal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cabécar",
        #[cfg(feature = "local_names")]
        autonym: Some("Cabécar"),
    },
    LanguageData {
        code_3: [99, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 106, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jinyu Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("晋语\u{200e} (Jin)"),
    },
    LanguageData {
        code_3: [99, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Kurdish",
        #[cfg(feature = "local_names")]
        autonym: Some("کوردیی ناوەندی"),
    },
    LanguageData {
        code_3: [99, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cibak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chakavian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaang Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Kaang Chin"),
    },
    LanguageData {
        code_3: [99, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anufo",
        #[cfg(feature = "local_names")]
        autonym: Some("Anufo"),
    },
    LanguageData {
        code_3: [99, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kajakse",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kairak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chukot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koasati",
        #[cfg(feature = "local_names")]
        autonym: Some("Koasati"),
    },
    LanguageData {
        code_3: [99, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kavalan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cakfem-Mushere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cakchiquel-Quiché Mixed Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ron",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chilcotin",
        #[cfg(feature = "local_names")]
        autonym: Some("Tŝinlhqot’in"),
    },
    LanguageData {
        code_3: [99, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chaldean Neo-Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lealao Chinantec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chilisso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chakali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laitu Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Laitu Chin"),
    },
    LanguageData {
        code_3: [99, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idu-Mishmi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Clallam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lowland Oaxaca Chontal",
        #[cfg(feature = "local_names")]
        autonym: Some("Lajltyaygi"),
    },
    LanguageData {
        code_3: [99, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lautu Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Lautu Chin"),
    },
    LanguageData {
        code_3: [99, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caluyanun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chulym",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Highland Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cerma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Mongolian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emberá-Chamí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Campalagian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Michigamea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandarin Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("普通话\u{200e} (Putonghua)"),
    },
    LanguageData {
        code_3: [99, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Mnong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mro-Khimi Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Khimi, Mro-Khimi"),
    },
    LanguageData {
        code_3: [99, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Messapic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Camtho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Changthang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinbon Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Uppu, Uppu Chin"),
    },
    LanguageData {
        code_3: [99, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Côông",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Qiang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hakha Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Lai, Lai Chin"),
    },
    LanguageData {
        code_3: [99, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asháninka",
        #[cfg(feature = "local_names")]
        autonym: Some("Asháninka"),
    },
    LanguageData {
        code_3: [99, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khumi Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Khumi, Khumi Chin"),
    },
    LanguageData {
        code_3: [99, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lalana Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jujmi"),
    },
    LanguageData {
        code_3: [99, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Con",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Ping Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Montenegrin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Asmat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tepetotutla Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jajmi dzä kï ï’, Jejmi"),
    },
    LanguageData {
        code_3: [99, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chenoua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngawn Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Cornish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cocos Islands Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chicomuceltec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cocopa",
        #[cfg(feature = "local_names")]
        autonym: Some("Kuapá"),
    },
    LanguageData {
        code_3: [99, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cocama-Cocamilla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koreguaje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Colorado",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chonyi-Dzihana-Kauma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cochimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Teresa Cora",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwéimarusa’na"),
    },
    LanguageData {
        code_3: [99, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Columbia-Wenatchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Comanche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cofán",
        #[cfg(feature = "local_names")]
        autonym: Some("A’ingae"),
    },
    LanguageData {
        code_3: [99, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Comox",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coptic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coquille",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 114],
        code_1: Some([107, 119]),
        #[cfg(feature = "english_names")]
        name_en: "Cornish",
        #[cfg(feature = "local_names")]
        autonym: Some("kernewek"),
    },
    LanguageData {
        code_3: [99, 111, 115],
        code_1: Some([99, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Corsican",
        #[cfg(feature = "local_names")]
        autonym: Some("Corsu"),
    },
    LanguageData {
        code_3: [99, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caquinte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wamey",
        #[cfg(feature = "local_names")]
        autonym: Some("Wamay"),
    },
    LanguageData {
        code_3: [99, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cao Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cowlitz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chochotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palantla Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jajme dzä mii, Jmiih kia’ dzä mii"),
    },
    LanguageData {
        code_3: [99, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ucayali-Yurúa Ashéninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ajyíninka Apurucayali",
        #[cfg(feature = "local_names")]
        autonym: Some("Ajyéninka"),
    },
    LanguageData {
        code_3: [99, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cappadocian Greek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinese Pidgin English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cherepon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpeego",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Capiznon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pichis Ashéninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pu-Xian Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Ucayali Ashéninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 113, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuanqiandian Cluster Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Island Carib",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lonwolwol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coeur d'Alene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 101],
        code_1: Some([99, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Cree",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caramanta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Michif",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Crimean Tatar",
        #[cfg(feature = "local_names")]
        autonym: Some("Qirim, Qirimtatar"),
    },
    LanguageData {
        code_3: [99, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sãotomense",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern East Cree",
        #[cfg(feature = "local_names")]
        autonym: Some("ᐄᓅ ᐊᔨᒨᓐ, Īnū Ayimūn "),
    },
    LanguageData {
        code_3: [99, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plains Cree",
        #[cfg(feature = "local_names")]
        autonym: Some("ᓀᐦᐃᔭᐍᐏᐣ, Nēhiyawēwin"),
    },
    LanguageData {
        code_3: [99, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern East Cree",
        #[cfg(feature = "local_names")]
        autonym: Some("ᐄᔨᔫ ᐊᔨᒨᓐ, Īyiyū Ayimūn "),
    },
    LanguageData {
        code_3: [99, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moose Cree",
        #[cfg(feature = "local_names")]
        autonym: Some("Mōsonī, Ililiw"),
    },
    LanguageData {
        code_3: [99, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "El Nayar Cora",
        #[cfg(feature = "local_names")]
        autonym: Some("Naáyeri"),
    },
    LanguageData {
        code_3: [99, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Crow",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iyo'wujwa Chorote",
        #[cfg(feature = "local_names")]
        autonym: Some("Yojwaja"),
    },
    LanguageData {
        code_3: [99, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carolina Algonquian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seselwa Creole French",
        #[cfg(feature = "local_names")]
        autonym: Some("Seselwa"),
    },
    LanguageData {
        code_3: [99, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iyojwa'ja Chorote",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chaura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chrau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carrier",
        #[cfg(feature = "local_names")]
        autonym: Some("Dakelh"),
    },
    LanguageData {
        code_3: [99, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cruzeño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiltepec Chinantec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kashubian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Catalan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiangmai Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Czech Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuba Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chilean Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asho Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Asho, Asho Chin"),
    },
    LanguageData {
        code_3: [99, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coast Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Songlai Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jola-Kasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chinese Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Sierra Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Colombian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sochiapam Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jaú jm_, Jmiih kia’ dzä jii’"),
    },
    LanguageData {
        code_3: [99, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Ping Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Croatia Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Costa Rican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Ohlone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Ohlone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumtu Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Cumtu, Sumtu Chin"),
    },
    LanguageData {
        code_3: [99, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swampy Cree",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cambodian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siyin Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coos",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tataltepec Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chetco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tedim Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Tidim"),
    },
    LanguageData {
        code_3: [99, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tepinapa Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jujmi"),
    },
    LanguageData {
        code_3: [99, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chittagonian",
        #[cfg(feature = "local_names")]
        autonym: Some("চ\u{9be}\u{981}টগ\u{9be}\u{981}ইয\u{9bc}\u{9be} ব\u{9c1}লি\u{200e} (Caṭgãia Buli)"),
    },
    LanguageData {
        code_3: [99, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thaiphum Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Thaiphum"),
    },
    LanguageData {
        code_3: [99, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlacoatzintepec Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jau jmai"),
    },
    LanguageData {
        code_3: [99, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chitimacha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chhintange",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emberá-Catío",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Highland Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Catanduanes Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayanad Chetti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chol",
        #[cfg(feature = "local_names")]
        autonym: Some("Lakty’añ"),
    },
    LanguageData {
        code_3: [99, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moundadan Chetty",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zacatepec Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cubeo",
        #[cfg(feature = "local_names")]
        autonym: Some("Pamié"),
    },
    LanguageData {
        code_3: [99, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usila Chinantec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jaú jm_, Jmiih kia’ dzä jii’"),
    },
    LanguageData {
        code_3: [99, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuiba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mashco Piro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Blas Kuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Culina",
        #[cfg(feature = "local_names")]
        autonym: Some("Madiha"),
    },
    LanguageData {
        code_3: [99, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cumanagoto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cupeño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chhulung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teutila Cuicatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dubaku"),
    },
    LanguageData {
        code_3: [99, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Ya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuvok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chukwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tepeuxila Cuicatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dibaku"),
    },
    LanguageData {
        code_3: [99, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuitlatec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 118, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chug",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Valle Nacional Chinantec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maindo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woods Cree",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwere",
        #[cfg(feature = "local_names")]
        autonym: Some("Nghwele"),
    },
    LanguageData {
        code_3: [99, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chewong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuwaataay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nopala Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cayubaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 121, 109],
        code_1: Some([99, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Welsh",
        #[cfg(feature = "local_names")]
        autonym: Some("Cymraeg"),
    },
    LanguageData {
        code_3: [99, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuyonon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huizhou Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Knaanic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zenzontepec Chatino",
        #[cfg(feature = "local_names")]
        autonym: Some("Cha’ jna’a"),
    },
    LanguageData {
        code_3: [99, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Min Zhong Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [99, 122, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zotung Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Zotung, Zotung Chin"),
    },
    LanguageData {
        code_3: [100, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dangaléat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dambi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duupa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dagbani",
        #[cfg(feature = "local_names")]
        autonym: Some("Dagbanli"),
    },
    LanguageData {
        code_3: [100, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwahatike",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Day",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dar Fur Daju",
        #[cfg(feature = "local_names")]
        autonym: Some("Bekke"),
    },
    LanguageData {
        code_3: [100, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dakota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dahalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Damakawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 110],
        code_1: Some([100, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Danish",
        #[cfg(feature = "local_names")]
        autonym: Some("dansk"),
    },
    LanguageData {
        code_3: [100, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daai Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Daai Chin, Dai"),
    },
    LanguageData {
        code_3: [100, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dandami Maria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dargwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daho-Doo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dar Sila Daju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taita",
        #[cfg(feature = "local_names")]
        autonym: Some("Kitaita"),
    },
    LanguageData {
        code_3: [100, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Davawenyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangime",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Deno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dadiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dabe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Edopi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogul Dom Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ida'an",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyirbal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duguri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duriankere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dulbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dabarre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ben Tey Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bondum Dom Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bankan Tey Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dibiyaso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Deccan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Negerhollands",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dadi Dadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dongotono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doondo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fataluku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Goodenough",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dendi",
        #[cfg(feature = "local_names")]
        autonym: Some("Songhay"),
    },
    LanguageData {
        code_3: [100, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dido",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhudhuroa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Donno So Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dawera-Daweloor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dagik",
        #[cfg(feature = "local_names")]
        autonym: Some("Tharuwa"),
    },
    LanguageData {
        code_3: [100, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dedua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dewoin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dezfuli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Degema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dehwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Demisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Delaware",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Slave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pidgin Delaware",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dendi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Deori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Desano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 117],
        code_1: Some([100, 101]),
        #[cfg(feature = "english_names")]
        name_en: "German",
        #[cfg(feature = "local_names")]
        autonym: Some("Deutsch"),
    },
    LanguageData {
        code_3: [100, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Domung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dengese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Dagaare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunoge Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Casiguran Dumagat Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dagaari Dioula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Degenan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dghwede",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Dagara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dagba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andaandi",
        #[cfg(feature = "local_names")]
        autonym: Some("Andaandi"),
    },
    LanguageData {
        code_3: [100, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dagoman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogrib",
        #[cfg(feature = "local_names")]
        autonym: Some("Tlicho"),
    },
    LanguageData {
        code_3: [100, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogoso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndra'ngith",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daungwurrung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doghoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhundari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhangu-Djangu",
        #[cfg(feature = "local_names")]
        autonym: Some("Dhangu-Djangu, Yolngu-Matha"),
    },
    LanguageData {
        code_3: [100, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhimal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhalandji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhanki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhodia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhargari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhaiso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhurga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dehu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhanwar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhungaloo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Central Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakota Dida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Didinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dieri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Digo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumiai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dimbong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dilling",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dime",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dibo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeastern Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dimli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dirim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dimasa",
        #[cfg(feature = "local_names")]
        autonym: Some("Grau Dima, Magrau"),
    },
    LanguageData {
        code_3: [100, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Diriku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 118],
        code_1: Some([100, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Dhivehi",
        #[cfg(feature = "local_names")]
        autonym: Some("ދ\u{7a8}ވ\u{7ac}ހ\u{7a8}ބ\u{7a6}ސ\u{7b0}"),
    },
    LanguageData {
        code_3: [100, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dixon Reef",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Diuwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ding",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djadjawurrung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djinba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dar Daju Daju",
        #[cfg(feature = "local_names")]
        autonym: Some("Saaronge"),
    },
    LanguageData {
        code_3: [100, 106, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djamindjung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zarma",
        #[cfg(feature = "local_names")]
        autonym: Some("Zarmaciine"),
    },
    LanguageData {
        code_3: [100, 106, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djangun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djinang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djeebbana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Maroon Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamsay Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jawoyn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jangkang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djambarrpuyngu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapriman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dakpakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dakka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuijau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Dinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mazagway",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dolgan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dahalik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dalmatian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Darlong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mombo Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gavak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Madhi Madhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dugwor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Medefaidrin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Kinabatangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Domaaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dameli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemedzung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Damar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dampelas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dubu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dumpas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mudburra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Demta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Grand Valley Dani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daonda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndendeule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dungan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Grand Valley Dani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dan",
        #[cfg(feature = "local_names")]
        autonym: Some("Dan"),
    },
    LanguageData {
        code_3: [100, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dengka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dzùùngoo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndrulo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Danaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mid Grand Valley Dani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Danau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Danu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Dani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dení",
        #[cfg(feature = "local_names")]
        autonym: Some("Madiha"),
    },
    LanguageData {
        code_3: [100, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dobu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Dong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Domu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dondo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lukpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dominican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dori'o",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dogosé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dass",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dombe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doyayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bussa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dompo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dorze",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dair",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minderico",
        #[cfg(feature = "local_names")]
        autonym: Some("Minderico, Piação dos Charales do Ninhou"),
    },
    LanguageData {
        code_3: [100, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Darmiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dolpo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rungus",
        #[cfg(feature = "local_names")]
        autonym: Some("Momogun"),
    },
    LanguageData {
        code_3: [100, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "C'Lela",
        #[cfg(feature = "local_names")]
        autonym: Some("Lela"),
    },
    LanguageData {
        code_3: [100, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paakantyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Damar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daro-Matu Melanau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gedeo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Drents",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rukai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Darai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Sorbian",
        #[cfg(feature = "local_names")]
        autonym: Some("dolnoserbšćina"),
    },
    LanguageData {
        code_3: [100, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dutch Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daasanach",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Disa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Danish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dusner",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Desiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tadaksahak",
        #[cfg(feature = "local_names")]
        autonym: Some("Idaksahak"),
    },
    LanguageData {
        code_3: [100, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mardin Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Labuk-Kinabatangan Kadazan",
        #[cfg(feature = "local_names")]
        autonym: Some("Sungai, Sungai Kinabatangan"),
    },
    LanguageData {
        code_3: [100, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ditidaht",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adithinngithigh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ana Tinga Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tene Kan Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tomo Kan Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daatsʼi\u{301}in",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tommo So Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadazan Dusun",
        #[cfg(feature = "local_names")]
        autonym: Some("Kadazan Dusun"),
    },
    LanguageData {
        code_3: [100, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lotud",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toro So Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toro Tegu Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tebul Ure Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dotyali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duala",
        #[cfg(feature = "local_names")]
        autonym: Some("duálá"),
    },
    LanguageData {
        code_3: [100, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dubli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umiray Dumaget Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dumbea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duruma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dungra Bhil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dumun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uyajitaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alabat Island Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Dutch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dusun Deyah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dupaninan Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dusun Malang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Drung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duvle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dusun Witu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duungooma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dicamay Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duli-Gey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Diri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dawik Kui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dawro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dutton World Speedwords",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhuwal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dawawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhuwaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 119, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dewas Rai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyaberdyaber",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyugun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Villa Viciosa Agta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djimini Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanda Dom Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyangadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jola-Fonyi",
        #[cfg(feature = "local_names")]
        autonym: Some("joola"),
    },
    LanguageData {
        code_3: [100, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dyula",
        #[cfg(feature = "local_names")]
        autonym: Some("Julakan"),
    },
    LanguageData {
        code_3: [100, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djabugay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunzu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djiwarli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dazaga",
        #[cfg(feature = "local_names")]
        autonym: Some("Dazaga"),
    },
    LanguageData {
        code_3: [100, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dzalakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dzando",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [100, 122, 111],
        code_1: Some([100, 122]),
        #[cfg(feature = "english_names")]
        name_en: "Dzongkha",
        #[cfg(feature = "local_names")]
        autonym: Some("ར\u{fab}\u{f7c}ང་ཁ"),
    },
    LanguageData {
        code_3: [101, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karenggapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beginci",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ebughu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Ebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ebrié",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Embu",
        #[cfg(feature = "local_names")]
        autonym: Some("Kĩembu"),
    },
    LanguageData {
        code_3: [101, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eteocretan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ecuadorian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eteocypriot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "E",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Efai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Efe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Efik",
        #[cfg(feature = "local_names")]
        autonym: Some("Efik"),
    },
    LanguageData {
        code_3: [101, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ega",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emilian",
        #[cfg(feature = "local_names")]
        autonym: Some("Emigliân"),
    },
    LanguageData {
        code_3: [101, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Benamanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eggon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Egyptian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miyakubo Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ehueun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eipomek",
        #[cfg(feature = "local_names")]
        autonym: Some("Lik"),
    },
    LanguageData {
        code_3: [101, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eitiep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Askopan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ejamat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ekajuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ekit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ekari",
        #[cfg(feature = "local_names")]
        autonym: Some("Me"),
    },
    LanguageData {
        code_3: [101, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Standard Estonian",
        #[cfg(feature = "local_names")]
        autonym: Some("Eesti Kirjakeel"),
    },
    LanguageData {
        code_3: [101, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elip",
        #[cfg(feature = "local_names")]
        autonym: Some("Nulibié"),
    },
    LanguageData {
        code_3: [101, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ekpeye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yace",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Kayah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elepi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "El Hugeirat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nding",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elkei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 108],
        code_1: Some([101, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Modern Greek",
        #[cfg(feature = "local_names")]
        autonym: Some("Ελληνικά"),
    },
    LanguageData {
        code_3: [101, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eleme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "El Molo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elamite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emai-Iuleha-Ora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Embaloh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emerillon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Meohang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mussau-Emira",
        #[cfg(feature = "local_names")]
        autonym: Some("Mussau"),
    },
    LanguageData {
        code_3: [101, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Maninkakan",
        #[cfg(feature = "local_names")]
        autonym: Some("Maninka"),
    },
    LanguageData {
        code_3: [101, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamulique",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Emberá",
        #[cfg(feature = "local_names")]
        autonym: Some("Embera"),
    },
    LanguageData {
        code_3: [101, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Minyag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pacific Gulf Yupik",
        #[cfg(feature = "local_names")]
        autonym: Some("Aleut"),
    },
    LanguageData {
        code_3: [101, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Muria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emplawas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Erromintxela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Epigraphic Mayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbessa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Markweeta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "En",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Forest Enets",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 103],
        code_1: Some([101, 110]),
        #[cfg(feature = "english_names")]
        name_en: "English",
        #[cfg(feature = "local_names")]
        autonym: Some("English"),
    },
    LanguageData {
        code_3: [101, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tundra Enets",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enlhet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Engenni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enggano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emumu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enxet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Epie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 112, 111],
        code_1: Some([101, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Esperanto",
        #[cfg(feature = "local_names")]
        autonym: Some("esperanto"),
    },
    LanguageData {
        code_3: [101, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eravallan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eruwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ogea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Efate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Horpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Erre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ersu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eritai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Erokwanas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ese Ejja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aheri Gondi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eshtehardi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Alaskan Inupiatun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwest Alaska Inupiatun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Egypt Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Esuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salvadoran Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Estonian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Esselen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Siberian Yupik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 116],
        code_1: Some([101, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Estonian",
        #[cfg(feature = "local_names")]
        autonym: Some("eesti"),
    },
    LanguageData {
        code_3: [101, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Yupik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eskayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Etebi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Etchemin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ethiopian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eton",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eton",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Edolo",
        #[cfg(feature = "local_names")]
        autonym: Some("Edolo"),
    },
    LanguageData {
        code_3: [101, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yekhee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Etruscan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ejagham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eten",
        #[cfg(feature = "local_names")]
        autonym: Some("Iten"),
    },
    LanguageData {
        code_3: [101, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 117, 115],
        code_1: Some([101, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Basque",
        #[cfg(feature = "local_names")]
        autonym: Some("euskara"),
    },
    LanguageData {
        code_3: [101, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Even",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uvbie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Evenki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 119, 101],
        code_1: Some([101, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Ewe",
        #[cfg(feature = "local_names")]
        autonym: Some("Eʋegbe"),
    },
    LanguageData {
        code_3: [101, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ewondo",
        #[cfg(feature = "local_names")]
        autonym: Some("ewondo"),
    },
    LanguageData {
        code_3: [101, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Extremaduran",
        #[cfg(feature = "local_names")]
        autonym: Some("Barranquênhu"),
    },
    LanguageData {
        code_3: [101, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eyak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keiyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [101, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ezaa",
        #[cfg(feature = "local_names")]
        autonym: Some("Ezaa"),
    },
    LanguageData {
        code_3: [101, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uzekwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fasu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fa d'Ambu",
        #[cfg(feature = "local_names")]
        autonym: Some("Fa d’Ambu"),
    },
    LanguageData {
        code_3: [102, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wagi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fagani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Finongan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baissa Fali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Faiwol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Faita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fang",
        #[cfg(feature = "local_names")]
        autonym: Some("Benjube Fenga"),
    },
    LanguageData {
        code_3: [102, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Fali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fang",
        #[cfg(feature = "local_names")]
        autonym: Some("Fang"),
    },
    LanguageData {
        code_3: [102, 97, 111],
        code_1: Some([102, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Faroese",
        #[cfg(feature = "local_names")]
        autonym: Some("føroyskt"),
    },
    LanguageData {
        code_3: [102, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paloor",
        #[cfg(feature = "local_names")]
        autonym: Some("Paloor"),
    },
    LanguageData {
        code_3: [102, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fataleka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 115],
        code_1: Some([102, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Persian",
        #[cfg(feature = "local_names")]
        autonym: Some("فارسی"),
    },
    LanguageData {
        code_3: [102, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fanti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fala",
        #[cfg(feature = "local_names")]
        autonym: Some("Fala, Nosa Fala"),
    },
    LanguageData {
        code_3: [102, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Fars",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Fars",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Albay Bikol",
        #[cfg(feature = "local_names")]
        autonym: Some("Bikol"),
    },
    LanguageData {
        code_3: [102, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quebec Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Feroge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Foia Foia",
        #[cfg(feature = "local_names")]
        autonym: Some("Foia Foia"),
    },
    LanguageData {
        code_3: [102, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maasina Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fongoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nobiin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fyer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Faifi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 106],
        code_1: Some([102, 106]),
        #[cfg(feature = "english_names")]
        name_en: "Fijian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Filipino",
        #[cfg(feature = "local_names")]
        autonym: Some("Filipino"),
    },
    LanguageData {
        code_3: [102, 105, 110],
        code_1: Some([102, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Finnish",
        #[cfg(feature = "local_names")]
        autonym: Some("suomi"),
    },
    LanguageData {
        code_3: [102, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fipa",
        #[cfg(feature = "local_names")]
        autonym: Some("Ichifipa"),
    },
    LanguageData {
        code_3: [102, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Firan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tornedalen Finnish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fiwaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kirya-Konzəl",
        #[cfg(feature = "local_names")]
        autonym: Some("Karya, Konzal"),
    },
    LanguageData {
        code_3: [102, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kven Finnish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalispel-Pend d'Oreille",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Foau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Fali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Flinders Island",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fuliiru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Flaaitaal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fe'fe'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Far Western Muria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fanbak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fanagalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fania",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Foodo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Foi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Foma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fon",
        #[cfg(feature = "local_names")]
        autonym: Some("Fongbe"),
    },
    LanguageData {
        code_3: [102, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siraya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fernando Po Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 113, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 97],
        code_1: Some([102, 114]),
        #[cfg(feature = "english_names")]
        name_en: "French",
        #[cfg(feature = "local_names")]
        autonym: Some("français"),
    },
    LanguageData {
        code_3: [102, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cajun French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fordata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Frankish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arpitan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Forak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Frisian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Frisian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fortsenal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 114, 121],
        code_1: Some([102, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Western Frisian",
        #[cfg(feature = "local_names")]
        autonym: Some("Frysk"),
    },
    LanguageData {
        code_3: [102, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Finnish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "French Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Finland-Swedish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adamawa Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: Some("Fulfulde"),
    },
    LanguageData {
        code_3: [102, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pulaar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Futuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Borgu Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pular",
        #[cfg(feature = "local_names")]
        autonym: Some("Pular"),
    },
    LanguageData {
        code_3: [102, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Niger Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: Some("Fulfulde"),
    },
    LanguageData {
        code_3: [102, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagirmi Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 108],
        code_1: Some([102, 102]),
        #[cfg(feature = "english_names")]
        name_en: "Fulah",
        #[cfg(feature = "local_names")]
        autonym: Some("Pulaar"),
    },
    LanguageData {
        code_3: [102, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fulniô",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central-Eastern Niger Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Friulian",
        #[cfg(feature = "local_names")]
        autonym: Some("furlan"),
    },
    LanguageData {
        code_3: [102, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Futuna-Aniwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Furu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nigerian Fulfulde",
        #[cfg(feature = "local_names")]
        autonym: Some("Fulfulde"),
    },
    LanguageData {
        code_3: [102, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fuyug",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 118, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fwâi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [102, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ga",
        #[cfg(feature = "local_names")]
        autonym: Some("Ga"),
    },
    LanguageData {
        code_3: [103, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gabri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mixed Great Andamanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaddang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guarequena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gagauz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alekano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Borei",
        #[cfg(feature = "local_names")]
        autonym: Some("Mborena Kam"),
    },
    LanguageData {
        code_3: [103, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gadsup",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamkonora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galolen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kandawo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gan Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gants",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gata'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galeya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adiwasi Garasia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mudhili Gadaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nobonob",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Borana-Arsi-Guji Oromo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Central Oromo",
        #[cfg(feature = "local_names")]
        autonym: Some("Afaan Oromoo, Oromiffa"),
    },
    LanguageData {
        code_3: [103, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaytetye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karajarri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niksek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaikundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbanziri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Defi Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bodo Gadaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaddi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garhwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mo'da",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbaya-Bossangoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbaya-Bozoum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbagyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbesi Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gagadu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbanu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gabi-Gabi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Xwla Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zoroastrian Dari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ganggalida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galice",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guadeloupean Creole French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Grenadian Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guianese Creole French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Colonia Tovar German",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gade Lohar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pottangi Ollar Gadaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gugu Badhun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gedaged",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gude",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guduf-Gava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ga'dang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gadjerawang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurdjar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gadang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dirasha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umanakaina",
        #[cfg(feature = "local_names")]
        autonym: Some("Umanakaina"),
    },
    LanguageData {
        code_3: [103, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghodoberi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mehri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wipi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghandruk Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kungardutyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gudu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Godwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geruma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kire",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gboloo Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gerai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gengle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hutterite German",
        #[cfg(feature = "local_names")]
        autonym: Some("Hutterisch"),
    },
    LanguageData {
        code_3: [103, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gebe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gen",
        #[cfg(feature = "local_names")]
        autonym: Some("Gen-Gbe"),
    },
    LanguageData {
        code_3: [103, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ywom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ut-Ma'in",
        #[cfg(feature = "local_names")]
        autonym: Some("ut-Ma’in"),
    },
    LanguageData {
        code_3: [103, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geser-Gorom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eviya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geez",
        #[cfg(feature = "local_names")]
        autonym: Some("ግዕዝኛ"),
    },
    LanguageData {
        code_3: [103, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Patpatar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gafat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gugadj",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurr-goni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurgula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kungarakany",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ganglau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gitua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gagu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gogodala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghadamès",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hiberno-Scottish Gaelic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Ghale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Ghale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geko Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghulfan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghanongga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghomara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guhu-Samane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kija",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gibanawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gail",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gidar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaɓogbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goaria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Githabul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Girirra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gilbertese",
        #[cfg(feature = "local_names")]
        autonym: Some("I-Kiribati"),
    },
    LanguageData {
        code_3: [103, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hinukh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Green Gelao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Red Gelao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Giziga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gitxsan",
        #[cfg(feature = "local_names")]
        autonym: Some("Gitsenimx"),
    },
    LanguageData {
        code_3: [103, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mulao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "White Gelao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gilima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Giyug",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Giziga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachi Koli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gunditjmara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gonja",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngbanyito"),
    },
    LanguageData {
        code_3: [103, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurindji Kriol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gujari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magɨ",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gokana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kok-Nar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guinea Kpelle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǂUngkue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 97],
        code_1: Some([103, 100]),
        #[cfg(feature = "english_names")]
        name_en: "Scottish Gaelic",
        #[cfg(feature = "local_names")]
        autonym: Some("Gàidhlig"),
    },
    LanguageData {
        code_3: [103, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Belning",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bon Gula",
        #[cfg(feature = "local_names")]
        autonym: Some("Êeni"),
    },
    LanguageData {
        code_3: [103, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 101],
        code_1: Some([103, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Irish",
        #[cfg(feature = "local_names")]
        autonym: Some("Gaeilge"),
    },
    LanguageData {
        code_3: [103, 108, 103],
        code_1: Some([103, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Galician",
        #[cfg(feature = "local_names")]
        autonym: Some("galego"),
    },
    LanguageData {
        code_3: [103, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwest Pashai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gula Iro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gilaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garlali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galambu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Glaro-Twabo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 118],
        code_1: Some([103, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Manx",
        #[cfg(feature = "local_names")]
        autonym: Some("Gaelg"),
    },
    LanguageData {
        code_3: [103, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Glavda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gambera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gula'alaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mághdì",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magɨyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle High German",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Low German",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbaya-Mbodomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gimnime",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mirning",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gumalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamo",
        #[cfg(feature = "local_names")]
        autonym: Some("Gamotso"),
    },
    LanguageData {
        code_3: [103, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mycenaean Greek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mgbolizhia",
        #[cfg(feature = "local_names")]
        autonym: Some("Mgbolizhia"),
    },
    LanguageData {
        code_3: [103, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaansa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gangte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guanche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zulgo-Gemzek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ganang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngangam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gooniyandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǁGana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gangulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ginuman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gumatj",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Gondi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gureng Gureng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guntai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gnau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Bolivian Guaraní",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ganzi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guro",
        #[cfg(feature = "local_names")]
        autonym: Some("Golo"),
    },
    LanguageData {
        code_3: [103, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Playero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gorakor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Godié",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gongduk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gofa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gogo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old High German",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gobasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gowlan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gowli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goan Konkani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gondi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gone Dau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yeretuar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gorap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gorontalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gronings",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gothic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gavar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gorowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gobu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goundo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gozarkhani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gupa-Abawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghanaian Pidgin English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taiap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ga'anda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 113, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guiqiong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rajput Garasia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ancient Greek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guruntum-Mbaaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Madi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbiri-Niragu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kota Marudu Talantang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 110],
        code_1: Some([103, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Guarani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Groma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gorovu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taznatit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gresi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garo",
        #[cfg(feature = "local_names")]
        autonym: Some("Mande"),
    },
    LanguageData {
        code_3: [103, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kistane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gweda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guriaso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barclayville Grebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guramalum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ghanaian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "German Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gusilay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guatemalan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwest Gbaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wasembo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Greek Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swiss German",
        #[cfg(feature = "local_names")]
        autonym: Some("Schwiizertüütsch"),
    },
    LanguageData {
        code_3: [103, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guató",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aghu-Tharnggala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shiki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guajajára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayuu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yocoboué Dida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurindji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gupapuyngu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paraguayan Guaraní",
        #[cfg(feature = "local_names")]
        autonym: Some("Guaraní"),
    },
    LanguageData {
        code_3: [103, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guahibo",
        #[cfg(feature = "local_names")]
        autonym: Some("Hivi"),
    },
    LanguageData {
        code_3: [103, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Bolivian Guaraní",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 106],
        code_1: Some([103, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Gujarati",
        #[cfg(feature = "local_names")]
        autonym: Some("ગ\u{ac1}જરાતી"),
    },
    LanguageData {
        code_3: [103, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gumuz",
        #[cfg(feature = "local_names")]
        autonym: Some("Bega-Tse, Sigumza"),
    },
    LanguageData {
        code_3: [103, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sea Island Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guambiano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbyá Guaraní",
        #[cfg(feature = "local_names")]
        autonym: Some("Mbya"),
    },
    LanguageData {
        code_3: [103, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guayabero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gunwinggu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aché",
        #[cfg(feature = "local_names")]
        autonym: Some("Aché"),
    },
    LanguageData {
        code_3: [103, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Farefare",
        #[cfg(feature = "local_names")]
        autonym: Some("Farefare, Ninkarsé"),
    },
    LanguageData {
        code_3: [103, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guinean Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maléku Jaíka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanomamö",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gourmanchéma",
        #[cfg(feature = "local_names")]
        autonym: Some("Gourma"),
    },
    LanguageData {
        code_3: [103, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gusii",
        #[cfg(feature = "local_names")]
        autonym: Some("Ekegusii"),
    },
    LanguageData {
        code_3: [103, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guanano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duwet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Golin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guajá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gulay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurmana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku-Yalanji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gavião Do Jiparaná",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pará Gavião",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gumawana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guyani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbato",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gawri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gawwada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gweno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gowro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwichʼin",
        #[cfg(feature = "local_names")]
        autonym: Some("Dinju Zhuh K’yuu"),
    },
    LanguageData {
        code_3: [103, 119, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǀGwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awngthim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwandara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gawar-Bati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guwamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wè Southern",
        #[cfg(feature = "local_names")]
        autonym: Some("Wèè"),
    },
    LanguageData {
        code_3: [103, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwest Gbaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayardild",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gyem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gungabula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gyele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gayil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngäbere",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngäbere"),
    },
    LanguageData {
        code_3: [103, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guyanese Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gyalsumdo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guarayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gunya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ganza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 122, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gazi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [103, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Han",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hanoi Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gurani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hatam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Oromo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haiphong Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hahon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hajong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hakka Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("客家話\u{200e} (Hakkafa)"),
    },
    LanguageData {
        code_3: [104, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hangaza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hakö",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hupla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Harari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haisla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 116],
        code_1: Some([104, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Haitian",
        #[cfg(feature = "local_names")]
        autonym: Some("Ayisyen, Kreyòl"),
    },
    LanguageData {
        code_3: [104, 97, 117],
        code_1: Some([104, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Hausa",
        #[cfg(feature = "local_names")]
        autonym: Some("Hausa"),
    },
    LanguageData {
        code_3: [104, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Havu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hawaiian",
        #[cfg(feature = "local_names")]
        autonym: Some("ʻŌlelo Hawaiʻi"),
    },
    LanguageData {
        code_3: [104, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Haida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hazaragi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Heiban",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ancient Hebrew",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 115],
        code_1: Some([115, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Serbo-Croatian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Habu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andaman Creole Hindi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huichol",
        #[cfg(feature = "local_names")]
        autonym: Some("Wixárika"),
    },
    LanguageData {
        code_3: [104, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Haida",
        #[cfg(feature = "local_names")]
        autonym: Some("Xaad Kil"),
    },
    LanguageData {
        code_3: [104, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Honduras Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hadiyya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Qiandong Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 98],
        code_1: Some([104, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Hebrew",
        #[cfg(feature = "local_names")]
        autonym: Some("עברית"),
    },
    LanguageData {
        code_3: [104, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Herdé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Helong",
        #[cfg(feature = "local_names")]
        autonym: Some("Helong"),
    },
    LanguageData {
        code_3: [104, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hehe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Heiltsuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 101, 114],
        code_1: Some([104, 122]),
        #[cfg(feature = "english_names")]
        name_en: "Herero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haiǁom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haigwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hoia Hoia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kerak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hoyahoya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hibito",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hidatsa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fiji Hindi",
        #[cfg(feature = "local_names")]
        autonym: Some("Fiji Baat"),
    },
    LanguageData {
        code_3: [104, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pamosu",
        #[cfg(feature = "local_names")]
        autonym: Some("Pamosu"),
    },
    LanguageData {
        code_3: [104, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hinduri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hijuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seit-Kaitetu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hiligaynon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 110],
        code_1: Some([104, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Hindi",
        #[cfg(feature = "local_names")]
        autonym: Some("हिन\u{94d}दी"),
    },
    LanguageData {
        code_3: [104, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Himarimã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hittite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hiw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hixkaryána",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kahe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hunde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hunjara-Kaina Ke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mel-Khaonh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hong Kong Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halang Doan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hlersu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matu Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Matu Chin"),
    },
    LanguageData {
        code_3: [104, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hieroglyphic Luwian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Mashan Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Humburi Senni Songhay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Huishui Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Large Flowery Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Huishui Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong Don",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Guiyang Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Huishui Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Huishui Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luopohe Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Mashan Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 111],
        code_1: Some([104, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Hiri Motu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Mashan Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Qiandong Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Qiandong Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hamtai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hamap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong Dô",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Mashan Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Guiyang Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong Shua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Hindko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chhattisgarhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǁAni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hani",
        #[cfg(feature = "local_names")]
        autonym: Some("Ha Nhi"),
    },
    LanguageData {
        code_3: [104, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong Njua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hanunoo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Hindko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caribbean Hindustani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hoava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Holma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Horom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hobyót",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Holikachuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hadothi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Holu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Homa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Holoholo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hopi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Horo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ho Chi Minh City Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hote",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hovongan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Honi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Holiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hozo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hpon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hawai'i Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hrangkhol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niwer Mil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haruku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Horned Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haroi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nhirrpi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hértevin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hruso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 118],
        code_1: Some([104, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Croatian",
        #[cfg(feature = "local_names")]
        autonym: Some("hrvatski"),
    },
    LanguageData {
        code_3: [104, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warwar Feni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hunsrik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Harzani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Sorbian",
        #[cfg(feature = "local_names")]
        autonym: Some("hornjoserbšćina"),
    },
    LanguageData {
        code_3: [104, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hungarian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hausa Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xiang Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("湘语\u{200e} (Xiang)"),
    },
    LanguageData {
        code_3: [104, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Harsusi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hoti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minica Huitoto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hadza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hitu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Hittite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huambisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǂHua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huaulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Francisco Del Mar Huave",
        #[cfg(feature = "local_names")]
        autonym: Some("Ombeyajts"),
    },
    LanguageData {
        code_3: [104, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Humene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huachipaeri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huilliche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Guiyang Hmong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hulung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hungana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 110],
        code_1: Some([104, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Hungarian",
        #[cfg(feature = "local_names")]
        autonym: Some("magyar"),
    },
    LanguageData {
        code_3: [104, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hupa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halkomelem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huastec",
        #[cfg(feature = "local_names")]
        autonym: Some("Teenek"),
    },
    LanguageData {
        code_3: [104, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Humla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murui Huitoto",
        #[cfg(feature = "local_names")]
        autonym: Some("Murui"),
    },
    LanguageData {
        code_3: [104, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Mateo Del Mar Huave",
        #[cfg(feature = "local_names")]
        autonym: Some("Ombeayiüts"),
    },
    LanguageData {
        code_3: [104, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hukumina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nüpode Huitoto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hulaulá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hunzib",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haitian Vodoun Culture Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Dionisio Del Mar Huave",
        #[cfg(feature = "local_names")]
        autonym: Some("Ombeyajts"),
    },
    LanguageData {
        code_3: [104, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haveke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa María Del Mar Huave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wané",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hawai'i Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hwana",
        #[cfg(feature = "local_names")]
        autonym: Some("Tuftera"),
    },
    LanguageData {
        code_3: [104, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [104, 121, 101],
        code_1: Some([104, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Armenian",
        #[cfg(feature = "local_names")]
        autonym: Some("հայերեն"),
    },
    LanguageData {
        code_3: [104, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Armenian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iaai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iatmul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Purari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iban",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibibio",
        #[cfg(feature = "local_names")]
        autonym: Some("Ibibio"),
    },
    LanguageData {
        code_3: [105, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iwaidja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akpes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibanag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bih",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibaloi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agoi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 111],
        code_1: Some([105, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Igbo",
        #[cfg(feature = "local_names")]
        autonym: Some("Igbo"),
    },
    LanguageData {
        code_3: [105, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibuoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ede Ica",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Etkywan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Icelandic Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Islander Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idakho-Isukha-Tiriki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indo-Portuguese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ede Idaca",
        #[cfg(feature = "local_names")]
        autonym: Some("Ìdàáshà"),
    },
    LanguageData {
        code_3: [105, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 111],
        code_1: Some([105, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Ido",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idesa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idaté",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Idoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amganad Ifugao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batad Ifugao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ifè",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ifo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuwali Ifugao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Fuumu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayoyao Ifugao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 102, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keley-I Kallahan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ebira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Igede",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Igana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Igala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanggape",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ignaciano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isebe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Interglossa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Igwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iha Based Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ihievbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bidhawal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 105, 105],
        code_1: Some([105, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Sichuan Yi",
        #[cfg(feature = "local_names")]
        autonym: Some("ꆈꌠꉙ"),
    },
    LanguageData {
        code_3: [105, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thiin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Izon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Biseni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ede Ije",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalabari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeast Ijo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Canadian Inuktitut",
        #[cfg(feature = "local_names")]
        autonym: Some("ᐃᓄᒃᑎᑐᑦ\u{200e} (Inuktitut)"),
    },
    LanguageData {
        code_3: [105, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olulumo-Ikom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikpeshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikaranggal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inuit Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inuinnaqtun",
        #[cfg(feature = "local_names")]
        autonym: Some("Inuinnaqtun, ᐃᓄᐃᓐᓇᖅᑐᓐ\u{200e} (Inuvialuktun)"),
    },
    LanguageData {
        code_3: [105, 107, 117],
        code_1: Some([105, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Inuktitut",
        #[cfg(feature = "local_names")]
        autonym: Some("ᐃᓄᒃᑎᑐᑦ"),
    },
    LanguageData {
        code_3: [105, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iku-Gora-Ankwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikwere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ik",
        #[cfg(feature = "local_names")]
        autonym: Some("Icetod"),
    },
    LanguageData {
        code_3: [105, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikizu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ile Ape",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 101],
        code_1: Some([105, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Interlingue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garig-Ilgar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ili Turki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ilongot",
        #[cfg(feature = "local_names")]
        autonym: Some("Bogkalot"),
    },
    LanguageData {
        code_3: [105, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iranun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iloko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iranun",
        #[cfg(feature = "local_names")]
        autonym: Some("Iranun"),
    },
    LanguageData {
        code_3: [105, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "International Sign",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ili'uun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ilue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mala Malasar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anamgura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miluk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Imonda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Imbongu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Imroing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marsian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Imotong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Milyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 97],
        code_1: Some([105, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Interlingua",
        #[cfg(feature = "local_names")]
        autonym: Some("interlingua"),
    },
    LanguageData {
        code_3: [105, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 100],
        code_1: Some([105, 100]),
        #[cfg(feature = "english_names")]
        name_en: "Indonesian",
        #[cfg(feature = "local_names")]
        autonym: Some("Indonesia"),
    },
    LanguageData {
        code_3: [105, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Degexit'an",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ingush",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jungle Inga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indonesian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minaean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isinai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inoke-Yate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iñapari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Intha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ineseño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuma-Irumu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iowa-Oto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ipili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 112, 107],
        code_1: Some([105, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Inupiaq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ipiko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iquito",
        #[cfg(feature = "local_names")]
        autonym: Some("Akenóiri"),
    },
    LanguageData {
        code_3: [105, 113, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikwo",
        #[cfg(feature = "local_names")]
        autonym: Some("Ikwo"),
    },
    LanguageData {
        code_3: [105, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iresim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Irarutu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rigwe",
        #[cfg(feature = "local_names")]
        autonym: Some("Nkarigwe, Rigwe"),
    },
    LanguageData {
        code_3: [105, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iraqw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Irántxe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Irula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamberau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iraya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isabi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isconahua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isnag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Italian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Irish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Esan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkem-Nkum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ishkashimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 108],
        code_1: Some([105, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Icelandic",
        #[cfg(feature = "local_names")]
        autonym: Some("íslenska"),
    },
    LanguageData {
        code_3: [105, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masimasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isanzu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isoko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Israeli Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Istriot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 97],
        code_1: Some([105, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Italian",
        #[cfg(feature = "local_names")]
        autonym: Some("italiano"),
    },
    LanguageData {
        code_3: [105, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Binongan Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Tidung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inlaod Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Italian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itelmen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itu Mbon Uzo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itonama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iteri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isekiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maeng Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itawit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ito",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moyadan Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itzá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iu Mien",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibatan",
        #[cfg(feature = "local_names")]
        autonym: Some("Ibatan"),
    },
    LanguageData {
        code_3: [105, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ivatan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "I-Wak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iwam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iwur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sepik Iwam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ixcatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Xwja"),
    },
    LanguageData {
        code_3: [105, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ixil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iyayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mesaka",
        #[cfg(feature = "local_names")]
        autonym: Some("Ugare"),
    },
    LanguageData {
        code_3: [105, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ingrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [105, 122, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Izere",
        #[cfg(feature = "local_names")]
        autonym: Some("Izere"),
    },
    LanguageData {
        code_3: [105, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Izii",
        #[cfg(feature = "local_names")]
        autonym: Some("Izii"),
    },
    LanguageData {
        code_3: [106, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamamadí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hyam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Popti'",
        #[cfg(feature = "local_names")]
        autonym: Some("Jakalteko-Popti’"),
    },
    LanguageData {
        code_3: [106, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jahanka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yabem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jah Hut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zazao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jakun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yalahatan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamaican Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jandai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanyuwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaqay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "New Caledonian Javanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jakati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 118],
        code_1: Some([106, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Javanese",
        #[cfg(feature = "local_names")]
        autonym: Some("Basa Jawa"),
    },
    LanguageData {
        code_3: [106, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jambi Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahaso Daerah, Bahaso Dusun"),
    },
    LanguageData {
        code_3: [106, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yan-nhangu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jawe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Berber",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badjiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arandai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barikewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bijim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nafusi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lojban",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jofotek-Bromnya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jabutí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jukun Takum",
        #[cfg(feature = "local_names")]
        autonym: Some("Njikum"),
    },
    LanguageData {
        code_3: [106, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawijibaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamaican Country Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krymchak",
        #[cfg(feature = "local_names")]
        autonym: Some("Кърымчах тыльы\u{200e} (Krymchakh Tilyi)"),
    },
    LanguageData {
        code_3: [106, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jad",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jadgali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Tat",
        #[cfg(feature = "local_names")]
        autonym: Some("Juwri"),
    },
    LanguageData {
        code_3: [106, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jebero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jerung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jeh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jeri Kuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yelmek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dza",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngwa Idza"),
    },
    LanguageData {
        code_3: [106, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jonkor Bourmataguil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Georgian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gwak",
        #[cfg(feature = "local_names")]
        autonym: Some("Gwak"),
    },
    LanguageData {
        code_3: [106, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngomba",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndaꞌa"),
    },
    LanguageData {
        code_3: [106, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jehai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jhankot Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jibu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jilbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jingulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "sTodsde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiiddu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jilim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiamao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guanyinqiao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Youle Jinuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shuar",
        #[cfg(feature = "local_names")]
        autonym: Some("Shuar Chicham"),
    },
    LanguageData {
        code_3: [106, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buyuan Jinuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jejueo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bankal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mobwa Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kubo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paku Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amami Koniya Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Labir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngile",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamaican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zumbun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Machame",
        #[cfg(feature = "local_names")]
        autonym: Some("Kimachame"),
    },
    LanguageData {
        code_3: [106, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamdena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jumli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makuri Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Makuri"),
    },
    LanguageData {
        code_3: [106, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mashi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mouwase",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Juxtlahuaca Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("To’on Savi"),
    },
    LanguageData {
        code_3: [106, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jangshung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jandavra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Janji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yemsa",
        #[cfg(feature = "local_names")]
        autonym: Some("Yemsa"),
    },
    LanguageData {
        code_3: [106, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rawat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jaunsari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Joba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wojenaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jogi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jorá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jordanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jowulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jewish Palestinian Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 112, 110],
        code_1: Some([106, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Japanese",
        #[cfg(feature = "local_names")]
        autonym: Some("日本語"),
    },
    LanguageData {
        code_3: [106, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Persian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jaqaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jarai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jakattoe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Japrería",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Japanese Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Júma",
        #[cfg(feature = "local_names")]
        autonym: Some("Kagwahiva"),
    },
    LanguageData {
        code_3: [106, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wannu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jurchen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Worodougou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hõne",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngadjuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wapan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jirel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jumjum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hupdë",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jurúna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jumla Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jutish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wãpha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juray",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Javindo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caribbean Javanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jwira-Pepesa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiarong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Yemeni Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [106, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kara-Kalpak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabyle",
        #[cfg(feature = "local_names")]
        autonym: Some("Taqbaylit"),
    },
    LanguageData {
        code_3: [107, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ketangalan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kajaman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karekare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalanguya",
        #[cfg(feature = "local_names")]
        autonym: Some("Kalanguya"),
    },
    LanguageData {
        code_3: [107, 97, 108],
        code_1: Some([107, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Kalaallisut",
        #[cfg(feature = "local_names")]
        autonym: Some("kalaallisut"),
    },
    LanguageData {
        code_3: [107, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamba",
        #[cfg(feature = "local_names")]
        autonym: Some("Kikamba"),
    },
    LanguageData {
        code_3: [107, 97, 110],
        code_1: Some([107, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Kannada",
        #[cfg(feature = "local_names")]
        autonym: Some("ಕನ\u{ccd}ನಡ"),
    },
    LanguageData {
        code_3: [107, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xaasongaxango",
        #[cfg(feature = "local_names")]
        autonym: Some("Xasongo"),
    },
    LanguageData {
        code_3: [107, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bezhta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Capanahua",
        #[cfg(feature = "local_names")]
        autonym: Some("Capa Baquebo"),
    },
    LanguageData {
        code_3: [107, 97, 115],
        code_1: Some([107, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Kashmiri",
        #[cfg(feature = "local_names")]
        autonym: Some("کٲش\u{64f}ر"),
    },
    LanguageData {
        code_3: [107, 97, 116],
        code_1: Some([107, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Georgian",
        #[cfg(feature = "local_names")]
        autonym: Some("ქართული"),
    },
    LanguageData {
        code_3: [107, 97, 117],
        code_1: Some([107, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Kanuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katukína",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamayurá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 97, 122],
        code_1: Some([107, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Kazakh",
        #[cfg(feature = "local_names")]
        autonym: Some("қазақ тілі"),
    },
    LanguageData {
        code_3: [107, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalarko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaxuiâna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadiwéu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabardian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Camsá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaptiau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Grass Koiari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanembu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iwal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keliko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabiyè",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kafa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kande",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabutra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaiep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ap Ma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manga Kanuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duhwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khanty",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kawacha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lubila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngkâlmpw Kanum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaivi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukaan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tyap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamantan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kobiana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinalakna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koenoem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachama-Ganjule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korandje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Worimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kutu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yankunytjatjara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makonde",
        #[cfg(feature = "local_names")]
        autonym: Some("Chimakonde"),
    },
    LanguageData {
        code_3: [107, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamusi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tem",
        #[cfg(feature = "local_names")]
        autonym: Some("Tem"),
    },
    LanguageData {
        code_3: [107, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karamojong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Numèè",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsikimba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaningdon-Nindem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karaim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koneraw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keder",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwaja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabuverdianu",
        #[cfg(feature = "local_names")]
        autonym: Some("kabuverdianu"),
    },
    LanguageData {
        code_3: [107, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kélé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keiga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kerewe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Keres",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpessi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kekchí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemak",
        #[cfg(feature = "local_names")]
        autonym: Some("Ema"),
    },
    LanguageData {
        code_3: [107, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kakwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaikadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kugbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ket",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akebu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanikkaran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Kewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kukna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kupia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kukele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kodava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Kolami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konda-Dora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korra Koraga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kudiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurichiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kannada Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemiehua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinnauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khunsari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bilaspuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanjari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katkari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurmukar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kharam Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kullu Pahari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumaoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 102, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koromfé",
        #[cfg(feature = "local_names")]
        autonym: Some("Koromba"),
    },
    LanguageData {
        code_3: [107, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koyaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kawe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komering",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kube",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kusunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selangor Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamale Kham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaiwá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunggari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karipúna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karingani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krongo",
        #[cfg(feature = "local_names")]
        autonym: Some("Niino mo-di"),
    },
    LanguageData {
        code_3: [107, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaingang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumbainggar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Somyev",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kobol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karon Dori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyerung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khasi",
        #[cfg(feature = "local_names")]
        autonym: Some("Khasi"),
    },
    LanguageData {
        code_3: [107, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lü",
        #[cfg(feature = "local_names")]
        autonym: Some("ᦅᧄᦺᦑᦟᦹᧉ\u{200e} (Kam Tai Lue)"),
    },
    LanguageData {
        code_3: [107, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tukang Besi North",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bädi Kanum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korowai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khuen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khams Tibetan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kehu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuturmi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Halh Mongolian",
        #[cfg(feature = "local_names")]
        autonym: Some("Монгол хэл\u{200e} (Mongol khel)"),
    },
    LanguageData {
        code_3: [107, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lusi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 109],
        code_1: Some([107, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Khmer",
        #[cfg(feature = "local_names")]
        autonym: Some("ខ\u{17d2}មែរ"),
    },
    LanguageData {
        code_3: [107, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khandesi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khotanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koyra Chiini Songhay",
        #[cfg(feature = "local_names")]
        autonym: Some("Koyra ciini"),
    },
    LanguageData {
        code_3: [107, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kharia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kasua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khamti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkhumbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khvarshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khowar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keapara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koalib",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kickapoo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koshin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kibet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Parbate Kham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimaama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kilmeri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kitsai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kilivila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 107],
        code_1: Some([107, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Kikuyu",
        #[cfg(feature = "local_names")]
        autonym: Some("Gikuyu"),
    },
    LanguageData {
        code_3: [107, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kariya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karagas",
        #[cfg(feature = "local_names")]
        autonym: Some("тоъфа дыл\u{200e} (tofa dyl)"),
    },
    LanguageData {
        code_3: [107, 105, 110],
        code_1: Some([114, 119]),
        #[cfg(feature = "english_names")]
        name_en: "Kinyarwanda",
        #[cfg(feature = "local_names")]
        autonym: Some("Kinyarwanda"),
    },
    LanguageData {
        code_3: [107, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiowa",
        #[cfg(feature = "local_names")]
        autonym: Some("Cáuijògà"),
    },
    LanguageData {
        code_3: [107, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sheshi Kham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kosadle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 114],
        code_1: Some([107, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Kirghiz",
        #[cfg(feature = "local_names")]
        autonym: Some("кыргызча"),
    },
    LanguageData {
        code_3: [107, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agob",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kirmanjki",
        #[cfg(feature = "local_names")]
        autonym: Some("Shar Ma"),
    },
    LanguageData {
        code_3: [107, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeast Kiwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khiamniungan Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kirikiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mlap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Q'anjob'al",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coastal Konjo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Kiwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kisar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khmu",
        #[cfg(feature = "local_names")]
        autonym: Some("ກ\u{eb6}ມຫມ\u{eb8}\u{200e} (Kmhmu’)"),
    },
    LanguageData {
        code_3: [107, 106, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khakas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zabana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khinalugh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Highland Konjo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Parbate Kham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kháng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunjen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Harijan Kinnauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pwo Eastern Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Keres",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurudu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Kewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phrae Pwo Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kashaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaikavian Literary Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ramopa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Erave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 106, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bumthangkha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kakanda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwerisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Odoodee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinuku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kakabe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalaktang Monpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mabaka Valley Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khün",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kako",
        #[cfg(feature = "local_names")]
        autonym: Some("kakɔ"),
    },
    LanguageData {
        code_3: [107, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kokota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kosarek Yale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kon Keu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karko",
        #[cfg(feature = "local_names")]
        autonym: Some("Kakenbi"),
    },
    LanguageData {
        code_3: [107, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gugubera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaeku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kir-Balar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Giiwo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kangean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Kukuya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kohin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guugu Yimidhirr",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaska",
        #[cfg(feature = "local_names")]
        autonym: Some("Danezāgé’"),
    },
    LanguageData {
        code_3: [107, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Klamath-Modoc",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiliwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kolbila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gamilaraay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kendeje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagakaulo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Weliki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalumpang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khalaj",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagan Kalagan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Migum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalenjin",
        #[cfg(feature = "local_names")]
        autonym: Some("Kalenjin"),
    },
    LanguageData {
        code_3: [107, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rumu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khaling",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalasha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Klao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maskelynes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tado",
        #[cfg(feature = "local_names")]
        autonym: Some("Tado"),
    },
    LanguageData {
        code_3: [107, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koluwawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimbundu",
        #[cfg(feature = "local_names")]
        autonym: Some("Kimbundu"),
    },
    LanguageData {
        code_3: [107, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Dong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Majukayang Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bakole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kâte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumarbhag Paharia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limos Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanudan Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awtuw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gimme",
        #[cfg(feature = "local_names")]
        autonym: Some("Gimma"),
    },
    LanguageData {
        code_3: [107, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwama",
        #[cfg(feature = "local_names")]
        autonym: Some("T’wa Kwama"),
    },
    LanguageData {
        code_3: [107, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Kurdish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamasau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemtuik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karipúna Creole French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waboda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khorasani Turkish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lubuagan Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Kanuri",
        #[cfg(feature = "local_names")]
        autonym: Some("Kanuri"),
    },
    LanguageData {
        code_3: [107, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kankanaey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mankanya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koongo",
        #[cfg(feature = "local_names")]
        autonym: Some("Kikoongo, Koongo"),
    },
    LanguageData {
        code_3: [107, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanufi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Kanjobal",
        #[cfg(feature = "local_names")]
        autonym: Some("K’anjob’al"),
    },
    LanguageData {
        code_3: [107, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuranko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keninjal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanamarí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konkani",
        #[cfg(feature = "local_names")]
        autonym: Some("को\u{902}कणी\u{200e} (Konknni)"),
    },
    LanguageData {
        code_3: [107, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwanja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kintaq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaningra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kensiu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panoan Katukína",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kung-Ekoka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kendayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanyok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalamsé",
        #[cfg(feature = "local_names")]
        autonym: Some("Sàmó"),
    },
    LanguageData {
        code_3: [107, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konomala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kodi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kacipo-Bale Suri",
        #[cfg(feature = "local_names")]
        autonym: Some("Kacipo"),
    },
    LanguageData {
        code_3: [107, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cogui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komi-Permyak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konkani",
        #[cfg(feature = "local_names")]
        autonym: Some("को\u{902}कणी"),
    },
    LanguageData {
        code_3: [107, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 109],
        code_1: Some([107, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Komi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 110],
        code_1: Some([107, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Kongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konzo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waube",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 114],
        code_1: Some([107, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Korean",
        #[cfg(feature = "local_names")]
        autonym: Some("한국어"),
    },
    LanguageData {
        code_3: [107, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kosraean",
        #[cfg(feature = "local_names")]
        autonym: Some("Kosrae"),
    },
    LanguageData {
        code_3: [107, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lagwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kudu-Camo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kugama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koyukon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kutto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mullu Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Curripaco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpelle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapingamarangi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kplang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kofei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karajá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kepkiriwát",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikposo",
        #[cfg(feature = "local_names")]
        autonym: Some("Ikposo"),
    },
    LanguageData {
        code_3: [107, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korupun-Sela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korafe-Yegha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tehit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kafoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komi-Zyrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kobon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mountain Koiali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koryak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kupsabiny",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kovai",
        #[cfg(feature = "local_names")]
        autonym: Some("Alngubin"),
    },
    LanguageData {
        code_3: [107, 113, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doromu-Koki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koy Sanjaq Surat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalagan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kakabai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kisankasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koitabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koromira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kotafon Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyenele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaonde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Krahn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimré",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krenak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimaragang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Kissi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Klias River Kadazan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seroa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okolod",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kandas",
        #[cfg(feature = "local_names")]
        autonym: Some("Kandas"),
    },
    LanguageData {
        code_3: [107, 113, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mser",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koorete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 113, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumhali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karkin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karachay-Balkar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kairui-Midiki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panará",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krio",
        #[cfg(feature = "local_names")]
        autonym: Some("Krio"),
    },
    LanguageData {
        code_3: [107, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinaray-A",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kerek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karelian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sapo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumari Kanuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurukh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kavet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Krahn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karon",
        #[cfg(feature = "local_names")]
        autonym: Some("Kuloonaay"),
    },
    LanguageData {
        code_3: [107, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kryts",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sota Kanum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shuwa-Zamani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shambala",
        #[cfg(feature = "local_names")]
        autonym: Some("Kishambaa"),
    },
    LanguageData {
        code_3: [107, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuanua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bafia",
        #[cfg(feature = "local_names")]
        autonym: Some("rikpa"),
    },
    LanguageData {
        code_3: [107, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kusaghe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kölsch",
        #[cfg(feature = "local_names")]
        autonym: Some("Kölsch"),
    },
    LanguageData {
        code_3: [107, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kansa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kasiguranin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kofa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwaami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Borong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Kisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Winyé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khamyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kusu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "S'gaw Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kedang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kharia Thar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kodaku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kambaata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kholok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kokata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nubri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalkutung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Muyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plapo Krumen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaniet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koroshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karitiâna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaduo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katabaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Muyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ketum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kituba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Katu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kato",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaxararí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juǀʼhoan",
        #[cfg(feature = "local_names")]
        autonym: Some("Ju|’hoansi"),
    },
    LanguageData {
        code_3: [107, 117, 97],
        code_1: Some([107, 106]),
        #[cfg(feature = "english_names")]
        name_en: "Kuanyama",
        #[cfg(feature = "local_names")]
        autonym: Some("Oshikwanyama"),
    },
    LanguageData {
        code_3: [107, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kutep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwinsu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "'Auhelawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Katu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kupa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kushi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuikúro-Kalapálo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kepo'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumyk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumukio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunimaipa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karipuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 114],
        code_1: Some([107, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Kurdish",
        #[cfg(feature = "local_names")]
        autonym: Some("kurdî"),
    },
    LanguageData {
        code_3: [107, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kusaal",
        #[cfg(feature = "local_names")]
        autonym: Some("Kusaas"),
    },
    LanguageData {
        code_3: [107, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kutenai",
        #[cfg(feature = "local_names")]
        autonym: Some("Ktunaxa"),
    },
    LanguageData {
        code_3: [107, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Kuskokwim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpagua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kukatja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuuku-Ya'u",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bagvalal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kubu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kove",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalabakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabalai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuni-Boazi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komodo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Psikye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korean Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayaw",
        #[cfg(feature = "local_names")]
        autonym: Some("Kayaw"),
    },
    LanguageData {
        code_3: [107, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kendem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Border Kuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dobel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kompane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Geba Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kerinci",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahta Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yinbaw Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wersing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parkari Koli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yintale Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 118, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsakwambo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dâw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Likwala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwaio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwerba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwara'ae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sara Kaba Deme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kowiai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awa-Cuaiquer",
        #[cfg(feature = "local_names")]
        autonym: Some("Awa Pit, Înkal Awa"),
    },
    LanguageData {
        code_3: [107, 119, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwakiutl",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwak’wala"),
    },
    LanguageData {
        code_3: [107, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kofyar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwambi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwangali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwomtari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kodia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwesten",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwakum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sara Kaba Náà",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwinti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khirwar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Salvador Kongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 119, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kairiru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krobu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konso",
        #[cfg(feature = "local_names")]
        autonym: Some("Khonso"),
    },
    LanguageData {
        code_3: [107, 120, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brunei",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Melayu Brunei, بهاس ملايو بروني\u{200e} (Bahasa Melayu Brunei)"),
    },
    LanguageData {
        code_3: [107, 120, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manumanaw Karen",
        #[cfg(feature = "local_names")]
        autonym: Some("Kawyaw"),
    },
    LanguageData {
        code_3: [107, 120, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keningau Murut",
        #[cfg(feature = "local_names")]
        autonym: Some("Keningau Murut"),
    },
    LanguageData {
        code_3: [107, 120, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulfa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zayein Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Khmer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanowit-Tanjong Melanau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanoé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadiyara Koli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Smärky Kanum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kangjia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koiwat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuvi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Likuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 120, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kerewo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Butbut Kalinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krache",
        #[cfg(feature = "local_names")]
        autonym: Some("Kaakyi"),
    },
    LanguageData {
        code_3: [107, 121, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kouya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keyagana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiput",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalapuya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpatili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Binukidnon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kelon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuruáya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baram Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayagar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Kayah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayort",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kudmali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rapoisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kambaira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayabí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Karaboro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaibobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bondoukou Kulango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kosena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Da'a Kaili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kikai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kelabit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kazukuru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayeli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kais",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kokola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaningi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaidipang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaike",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sugut Dusun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayupulau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komyandaret",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karirí-Xocó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamarian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [107, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalabra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Subanen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Linear A",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lacandon",
        #[cfg(feature = "local_names")]
        autonym: Some("Jach-t’aan"),
    },
    LanguageData {
        code_3: [108, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ladino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pattani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lafofa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langi",
        #[cfg(feature = "local_names")]
        autonym: Some("Kɨlaangi"),
    },
    LanguageData {
        code_3: [108, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahnda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lambya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lalia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 111],
        code_1: Some([108, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Lao",
        #[cfg(feature = "local_names")]
        autonym: Some("ລາວ"),
    },
    LanguageData {
        code_3: [108, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qabiao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Larteh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lama",
        #[cfg(feature = "local_names")]
        autonym: Some("Lama"),
    },
    LanguageData {
        code_3: [108, 97, 116],
        code_1: Some([108, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Latin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 118],
        code_1: Some([108, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Latvian",
        #[cfg(feature = "local_names")]
        autonym: Some("latviešu"),
    },
    LanguageData {
        code_3: [108, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lauje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lama Bai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aribwatsa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Label",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakkia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tinani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laopang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "La'bi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ladakhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Libon Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lodhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rmeet",
        #[cfg(feature = "local_names")]
        autonym: Some("Rmeet"),
    },
    LanguageData {
        code_3: [108, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laven",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wampar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lohorung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Libyan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Labu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lavatbura-Lamusong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tolaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lawangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamalama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lardil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Legenyem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loncong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lubu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luchazi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lisela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tungag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Lawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luhu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lisabata-Nuniali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kla-Dan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Du\u{303}ya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lenyima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamja-Dengsa-Tola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lemoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leelau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Landoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Láadan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lufu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lega-Shabunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lala-Bisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lendu",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndrulo"),
    },
    LanguageData {
        code_3: [108, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lyélé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lelemi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lenje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lemio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lengola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leipon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nomaande",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lenca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lepcha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lembena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lenkau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lesing-Gelimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ledo Kaili",
        #[cfg(feature = "local_names")]
        autonym: Some("Kaili-Ledo"),
    },
    LanguageData {
        code_3: [108, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lemolang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lezghian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lefa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lingua Franca Nova",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lungga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laghu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lugbara",
        #[cfg(feature = "local_names")]
        autonym: Some("Lugbarati"),
    },
    LanguageData {
        code_3: [108, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laghuu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lengilu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lingarak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lega-Mwenga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "T'apo",
        #[cfg(feature = "local_names")]
        autonym: Some("T’apo"),
    },
    LanguageData {
        code_3: [108, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Longgu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ligenza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahu Shi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahul Lohar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lhomi",
        #[cfg(feature = "local_names")]
        autonym: Some("ल\u{94d}होमी\u{200e} (Lhomi)"),
    },
    LanguageData {
        code_3: [108, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahanan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lhokpu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mlahsö",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lo-Toga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lahu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West-Central Limba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Likum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hlai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyindrou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Likila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ligbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lihir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ligurian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lillooet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 109],
        code_1: Some([108, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Limburgan",
        #[cfg(feature = "local_names")]
        autonym: Some("Limbourgeois, Limburgs"),
    },
    LanguageData {
        code_3: [108, 105, 110],
        code_1: Some([108, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Lingala",
        #[cfg(feature = "local_names")]
        autonym: Some("lingála"),
    },
    LanguageData {
        code_3: [108, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sekpele",
        #[cfg(feature = "local_names")]
        autonym: Some("Sekpele"),
    },
    LanguageData {
        code_3: [108, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Libido",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liberian English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lisu",
        #[cfg(feature = "local_names")]
        autonym: Some("LI-SU\u{200e} (Li-su), Lisu"),
    },
    LanguageData {
        code_3: [108, 105, 116],
        code_1: Some([108, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Lithuanian",
        #[cfg(feature = "local_names")]
        autonym: Some("lietuvių"),
    },
    LanguageData {
        code_3: [108, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logorik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liv",
        #[cfg(feature = "local_names")]
        autonym: Some("Livõ kel"),
    },
    LanguageData {
        code_3: [108, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Col",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liabuku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda-Bambari",
        #[cfg(feature = "local_names")]
        autonym: Some("Linda"),
    },
    LanguageData {
        code_3: [108, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Libinza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Golpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rampi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laiyolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Li'o",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lampung Api",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yirandali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 106, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakalei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabras",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kucong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakondê",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Remun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laeko-Libuat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalaamaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Päri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakota",
        #[cfg(feature = "local_names")]
        autonym: Some("Lakȟólʼiyapi"),
    },
    LanguageData {
        code_3: [108, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kungkari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lokoya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lala-Roba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ladin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hermit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Laali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ladji Ladji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lelak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lilau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lasalimu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Efate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lolak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lithuanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lauan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Limba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Merei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limilngan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lumun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pévé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Lembata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamogai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lambichhong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lombi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Lembata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamkang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lambadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lombard",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limbum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamatuka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamalera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamenu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lomaiviti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lake Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laimbue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamboya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langbashe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbalanhu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lundayeh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langobardic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lanoh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Daantanai'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leningitij",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Central Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lorediakarkar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamnso'",
        #[cfg(feature = "local_names")]
        autonym: Some("Lamnso’"),
    },
    LanguageData {
        code_3: [108, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Longuda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lanima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lonzo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loloda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lobi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inonhan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saluan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narim",
        #[cfg(feature = "local_names")]
        autonym: Some("Boya"),
    },
    LanguageData {
        code_3: [108, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malawi Lomwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lombo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lopa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lobala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Téén",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loniu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Otuho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Louisiana Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lopi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tampias Lobu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loke",
        #[cfg(feature = "local_names")]
        autonym: Some("ल\u{94d}होवा\u{200e} (Lhowa)"),
    },
    LanguageData {
        code_3: [108, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lozi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lelepa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lepki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Long Phuri Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lipo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lopit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rara Bakati'",
        #[cfg(feature = "local_names")]
        autonym: Some("Rara Bakati’"),
    },
    LanguageData {
        code_3: [108, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Luri",
        #[cfg(feature = "local_names")]
        autonym: Some("لۊری شومالی"),
    },
    LanguageData {
        code_3: [108, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laurentian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laragia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loarki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lorang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Yamphu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Larantuka Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Larevat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lemerig",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lasgerdi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burundian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Albarradas Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lishana Deni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lusengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lashi",
        #[cfg(feature = "local_names")]
        autonym: Some("Lacid"),
    },
    LanguageData {
        code_3: [108, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Latvian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saamia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tibetan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laos Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panamanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aruop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trinidad and Tobago Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sivia Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seychelles Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mauritian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Late Middle Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Latgalian",
        #[cfg(feature = "local_names")]
        autonym: Some("Latgališu"),
    },
    LanguageData {
        code_3: [108, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thur",
        #[cfg(feature = "local_names")]
        autonym: Some("Leb-Thur"),
    },
    LanguageData {
        code_3: [108, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Latundê",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsotso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tachoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Latu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 116, 122],
        code_1: Some([108, 98]),
        #[cfg(feature = "english_names")]
        name_en: "Luxembourgish",
        #[cfg(feature = "local_names")]
        autonym: Some("Lëtzebuergesch"),
    },
    LanguageData {
        code_3: [108, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luba-Lulua",
        #[cfg(feature = "local_names")]
        autonym: Some("Tshiluba"),
    },
    LanguageData {
        code_3: [108, 117, 98],
        code_1: Some([108, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Luba-Katanga",
        #[cfg(feature = "local_names")]
        autonym: Some("Tshiluba"),
    },
    LanguageData {
        code_3: [108, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aringa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ludian",
        #[cfg(feature = "local_names")]
        autonym: Some("Lüüdikiel"),
    },
    LanguageData {
        code_3: [108, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luvale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 103],
        code_1: Some([108, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Ganda",
        #[cfg(feature = "local_names")]
        autonym: Some("Luganda"),
    },
    LanguageData {
        code_3: [108, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luiseno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lunanakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olu'bo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luimbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luo",
        #[cfg(feature = "local_names")]
        autonym: Some("Dholuo"),
    },
    LanguageData {
        code_3: [108, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lumbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lucumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lushai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lushootseed",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lumba-Yakkha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luwati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luyia",
        #[cfg(feature = "local_names")]
        autonym: Some("Luluhia"),
    },
    LanguageData {
        code_3: [108, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Luri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maku'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lavi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lavukaleve",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Standard Latvian",
        #[cfg(feature = "local_names")]
        autonym: Some("Latviešu valoda, Latviski"),
    },
    LanguageData {
        code_3: [108, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Levuka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lwalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lewo Eleng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "White Lachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Lawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laomian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luwo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malawian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lewotobi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lawu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lewo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lakurumau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Layakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lyngngam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luyana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Literary Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Litzlitz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Leinong Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [108, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Laz",
        #[cfg(feature = "local_names")]
        autonym: Some("Lazuri"),
    },
    LanguageData {
        code_3: [109, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Jerónimo Tecóatl Mazatec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yutanduchi Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Madurese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bo-Rukul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mafa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magahi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 104],
        code_1: Some([109, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Marshallese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maithili",
        #[cfg(feature = "local_names")]
        autonym: Some("म\u{948}थिली\u{200e} (Maithilī)"),
    },
    LanguageData {
        code_3: [109, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jalapa De Díaz Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Ntaxjo"),
    },
    LanguageData {
        code_3: [109, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makasar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 108],
        code_1: Some([109, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Malayalam",
        #[cfg(feature = "local_names")]
        autonym: Some("മലയ\u{d3e}ളം"),
    },
    LanguageData {
        code_3: [109, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mam",
        #[cfg(feature = "local_names")]
        autonym: Some("B’anax Mam, Qyool"),
    },
    LanguageData {
        code_3: [109, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandingo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiquihuitlán Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Nne nangui ngaxni"),
    },
    LanguageData {
        code_3: [109, 97, 114],
        code_1: Some([109, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Marathi",
        #[cfg(feature = "local_names")]
        autonym: Some("मराठी"),
    },
    LanguageData {
        code_3: [109, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masai",
        #[cfg(feature = "local_names")]
        autonym: Some("Maa"),
    },
    LanguageData {
        code_3: [109, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Francisco Matlatzinca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huautla Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Enna"),
    },
    LanguageData {
        code_3: [109, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sateré-Mawé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mampruli",
        #[cfg(feature = "local_names")]
        autonym: Some("Nmampurli"),
    },
    LanguageData {
        code_3: [109, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Moluccan Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Pasar"),
    },
    LanguageData {
        code_3: [109, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Mazahua",
        #[cfg(feature = "local_names")]
        autonym: Some("Jnatrjo"),
    },
    LanguageData {
        code_3: [109, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Higaonon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Bukidnon Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macushi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dibabawon Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baba Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangseng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ilianen Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nadëb",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maxakalí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ombamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macaguán",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maisin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukak Makú",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarangani Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matigsalug Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbula-Bwazza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbulungish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maring",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Memoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amoltepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tnu’u Ñuu Savi"),
    },
    LanguageData {
        code_3: [109, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maca",
        #[cfg(feature = "local_names")]
        autonym: Some("Mak’á"),
    },
    LanguageData {
        code_3: [109, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Machiguenga",
        #[cfg(feature = "local_names")]
        autonym: Some("Matsigenka"),
    },
    LanguageData {
        code_3: [109, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bitur",
        #[cfg(feature = "local_names")]
        autonym: Some("Bitur"),
    },
    LanguageData {
        code_3: [109, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sharanahua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Itundujia Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matsés",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapoyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maquiritari",
        #[cfg(feature = "local_names")]
        autonym: Some("Ye’kuana"),
    },
    LanguageData {
        code_3: [109, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mvanip",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macaguaje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malaccan Creole Portuguese",
        #[cfg(feature = "local_names")]
        autonym: Some("Kristang"),
    },
    LanguageData {
        code_3: [109, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coatlán Mixe",
        #[cfg(feature = "local_names")]
        autonym: Some("Ayuk"),
    },
    LanguageData {
        code_3: [109, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Menya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mambai",
        #[cfg(feature = "local_names")]
        autonym: Some("dâg tì mà\u{327}mbày"),
    },
    LanguageData {
        code_3: [109, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mengisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cameroon Mambila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minanibai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpiemo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Watut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morigi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Male",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maba",
        #[cfg(feature = "local_names")]
        autonym: Some("Bura Mabang"),
    },
    LanguageData {
        code_3: [109, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moksha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Massalat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maguindanaon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamvu",
        #[cfg(feature = "local_names")]
        autonym: Some("Tengo"),
    },
    LanguageData {
        code_3: [109, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangbetu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangbutu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maltese Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayogo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mboko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Lucía Monteverde Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbosi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dizin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Male",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suruí Do Pará",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Menka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikobi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Melpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mengen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Megam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Tlaxiaco Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Midob",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meyah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mekeo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Melanau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mende",
        #[cfg(feature = "local_names")]
        autonym: Some("Mende, Mɛnde\u{200e} (Mende), Mɛnde yia\u{200e} (Mende yia)"),
    },
    LanguageData {
        code_3: [109, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kedah Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miriwoong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Merey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meru",
        #[cfg(feature = "local_names")]
        autonym: Some("Kĩmĩrũ"),
    },
    LanguageData {
        code_3: [109, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masmaje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mato",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Motu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mano",
        #[cfg(feature = "local_names")]
        autonym: Some("Maa"),
    },
    LanguageData {
        code_3: [109, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hassaniyya",
        #[cfg(feature = "local_names")]
        autonym: Some("Hassaniyya, حسانية\u{200e}\u{200e}\u{200e} (Ḥassānīya)"),
    },
    LanguageData {
        code_3: [109, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Menominee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pattani Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("ภาษายาว\u{e35}\u{200e} (Baso Jawi)"),
    },
    LanguageData {
        code_3: [109, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bangka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mendankwe-Nkwen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morisyen",
        #[cfg(feature = "local_names")]
        autonym: Some("kreol morisien"),
    },
    LanguageData {
        code_3: [109, 102, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mogofin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wandala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mefele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Mofu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Putai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marghi South",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cross River Mbembe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makassar Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marrithiyel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mexican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mokerang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbwela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandjak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mulaha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Melo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 102, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayo",
        #[cfg(feature = "local_names")]
        autonym: Some("Yoreme Nokki"),
    },
    LanguageData {
        code_3: [109, 102, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mabaan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Irish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mararit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morokodo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maklew",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpumpong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa-Meetto",
        #[cfg(feature = "local_names")]
        autonym: Some("Makua"),
    },
    LanguageData {
        code_3: [109, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lijili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abureni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maleu-Kilenge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mambae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbangi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meta'",
        #[cfg(feature = "local_names")]
        autonym: Some("metaʼ"),
    },
    LanguageData {
        code_3: [109, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Magar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malila",
        #[cfg(feature = "local_names")]
        autonym: Some("Bamalila"),
    },
    LanguageData {
        code_3: [109, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mambwe-Lungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mongol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mailu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matumbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbunga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbugwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mahongwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mocho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbugu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Besisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Margu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma'di",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mogholi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mungaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mauwake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa-Moniga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mócheno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mashi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balinese Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Mari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buru",
        #[cfg(feature = "local_names")]
        autonym: Some("Buru"),
    },
    LanguageData {
        code_3: [109, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandahuaca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Digaro-Mishmi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbukushu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maru",
        #[cfg(feature = "local_names")]
        autonym: Some("Lhao Vo"),
    },
    LanguageData {
        code_3: [109, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma'anyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atatláhuca Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mi'kmaq",
        #[cfg(feature = "local_names")]
        autonym: Some("Míkmawísimk"),
    },
    LanguageData {
        code_3: [109, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ocotepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [109, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mofu-Gudur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Miguel El Grande Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chayuco Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [109, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chigmecatitlán Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Da’an davi"),
    },
    LanguageData {
        code_3: [109, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Abar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mikasuki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Peñoles Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [109, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alacatlatzala Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("To’on Savi"),
    },
    LanguageData {
        code_3: [109, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minangkabau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinotepa Nacional Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apasco-Apoala Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mískito",
        #[cfg(feature = "local_names")]
        autonym: Some("Mískitu"),
    },
    LanguageData {
        code_3: [109, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isthmus Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uncoded languages",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Puebla Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Da’an davi"),
    },
    LanguageData {
        code_3: [109, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cacaloxtepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akoye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mixtepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayutla Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coatzospan Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un davi"),
    },
    LanguageData {
        code_3: [109, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makalero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Juan Colorado Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un sav"),
    },
    LanguageData {
        code_3: [109, 106, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwest Maidu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muskum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kim Mun",
        #[cfg(feature = "local_names")]
        autonym: Some("Kem di mun"),
    },
    LanguageData {
        code_3: [109, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matukar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandeali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Medebur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malankuravan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malapandaram",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malaryan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malavedan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miship",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sauria Paharia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manna-Dora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mannan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mahali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mahican",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 106, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Majhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mal Paharia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siliput",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 100],
        code_1: Some([109, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Macedonian",
        #[cfg(feature = "local_names")]
        autonym: Some("македонски"),
    },
    LanguageData {
        code_3: [109, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhatki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mokilese",
        #[cfg(feature = "local_names")]
        autonym: Some("Mokil, Mwoakilloa"),
    },
    LanguageData {
        code_3: [109, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Byep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mokole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moklen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kupang Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Kupang"),
    },
    LanguageData {
        code_3: [109, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mingang Doso",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moikodi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bay Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Silacayoapan Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vamale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konyanka Maninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mafea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kituba",
        #[cfg(feature = "local_names")]
        autonym: Some("Kituba"),
    },
    LanguageData {
        code_3: [109, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinamiging Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Makian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makasae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cao Lan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manambu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 103],
        code_1: Some([109, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Malagasy",
        #[cfg(feature = "local_names")]
        autonym: Some("Malagasy"),
    },
    LanguageData {
        code_3: [109, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mape",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malimpung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miltu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ilwana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malua Bay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mulam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mlomp",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bargam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Maninkakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vame",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masalit",
        #[cfg(feature = "local_names")]
        autonym: Some("Kana Masaraka"),
    },
    LanguageData {
        code_3: [109, 108, 116],
        code_1: Some([109, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Maltese",
        #[cfg(feature = "local_names")]
        autonym: Some("Malti"),
    },
    LanguageData {
        code_3: [109, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "To'abaita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Motlav",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moloko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malfaxal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malaynon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Momina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Michoacán Mazahua",
        #[cfg(feature = "local_names")]
        autonym: Some("Jnatjo"),
    },
    LanguageData {
        code_3: [109, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maonan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Ambrym",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mehináku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Majhwar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mukha-Dora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Man Met",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamanwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangga Buang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Xiangxi Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malalamai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mmaala",
        #[cfg(feature = "local_names")]
        autonym: Some("Numaala"),
    },
    LanguageData {
        code_3: [109, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miriti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Madak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Migaama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mabaale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muna",
        #[cfg(feature = "local_names")]
        autonym: Some("Wuna"),
    },
    LanguageData {
        code_3: [109, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manchu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mondé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Mnong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manipuri",
        #[cfg(feature = "local_names")]
        autonym: Some("মৈতৈলোন\u{9cd}"),
    },
    LanguageData {
        code_3: [109, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Munji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandinka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Mnong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Min Bei Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minriq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mansi",
        #[cfg(feature = "local_names")]
        autonym: Some("Ма\u{304}ньщи"),
    },
    LanguageData {
        code_3: [109, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rennell-Bellona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manikion",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manyawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mocoví",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mobilian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Innu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mongondow",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mohawk",
        #[cfg(feature = "local_names")]
        autonym: Some("Kanienʼkéha"),
    },
    LanguageData {
        code_3: [109, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mboi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Monzombo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 110],
        code_1: Some([109, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Mongolian",
        #[cfg(feature = "local_names")]
        autonym: Some("монгол"),
    },
    LanguageData {
        code_3: [109, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Monom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mopán Maya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moro",
        #[cfg(feature = "local_names")]
        autonym: Some("Dhimorong"),
    },
    LanguageData {
        code_3: [109, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mossi",
        #[cfg(feature = "local_names")]
        autonym: Some("Mòoré"),
    },
    LanguageData {
        code_3: [109, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mogum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mohave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shekkacho",
        #[cfg(feature = "local_names")]
        autonym: Some("Shekkacho"),
    },
    LanguageData {
        code_3: [109, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mukulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpoto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malak Malak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangarrayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Machinere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Majang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Martu Wangka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Watut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yosondúa Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mindiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Migabac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matís",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vangunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dadibi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makuráp",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mungkip",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapidian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Misima-Panaeati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbuko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matepi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Momuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kota Bangun Kutai Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlazoyaltepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mariri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rajah Kabunsuwan Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbelime",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Marquesan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moronene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Modole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manipa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minokok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mander",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Makian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandari",
        #[cfg(feature = "local_names")]
        autonym: Some("Mundari"),
    },
    LanguageData {
        code_3: [109, 113, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mosimo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murupi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamuju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manggarai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 113, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mlabri",
        #[cfg(feature = "local_names")]
        autonym: Some("มละบร\u{e34}\u{200e} (Mlabri), มาลาบร\u{e35}\u{200e} (Malabri)"),
    },
    LanguageData {
        code_3: [109, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maricopa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Magar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Martha's Vineyard Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elseng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mising",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mara Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Mara"),
    },
    LanguageData {
        code_3: [109, 114, 105],
        code_1: Some([109, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Maori",
        #[cfg(feature = "local_names")]
        autonym: Some("te reo Maori"),
    },
    LanguageData {
        code_3: [109, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Mari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmwaveke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mortlockese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Merlav",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cheke Holo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morouas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Marquesan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maragus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marghi Central",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangareva",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maranao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maremgi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marind",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 97],
        code_1: Some([109, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Melayu"),
    },
    LanguageData {
        code_3: [109, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masbatenyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sankaran Maninka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yucatec Maya Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mekwei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moraid",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masikoro Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabah Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Sabah"),
    },
    LanguageData {
        code_3: [109, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mansaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molof",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agusan Manobo",
        #[cfg(feature = "local_names")]
        autonym: Some("Minanubu"),
    },
    LanguageData {
        code_3: [109, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vurës",
        #[cfg(feature = "local_names")]
        autonym: Some("Vurës"),
    },
    LanguageData {
        code_3: [109, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mombum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maritsauá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mongolian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Masela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maslam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mansoanka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moresada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aruamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Momare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cotabato Manobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anyin Morofo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Munit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mualang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Una",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Munggui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moskona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbe'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Montol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mator",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matagalpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Totontepec Mixe",
        #[cfg(feature = "local_names")]
        autonym: Some("Ayuk"),
    },
    LanguageData {
        code_3: [109, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wichí Lhamtés Nocten",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mewari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tututepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [109, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asaro'o",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Binukidnon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tidaá Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tnu’un dawi"),
    },
    LanguageData {
        code_3: [109, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nabi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundang",
        #[cfg(feature = "local_names")]
        autonym: Some("MUNDAŊ"),
    },
    LanguageData {
        code_3: [109, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ajumbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mednyj Aleut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Media Lengua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musgu",
        #[cfg(feature = "local_names")]
        autonym: Some("Mulwi"),
    },
    LanguageData {
        code_3: [109, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mündü",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mabire",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mugom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Multiple languages",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maiwala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malvi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Xiangxi Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Creek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Muria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaaku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muthuvan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bo-Ung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mursi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mattole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamboru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Peripheral Mongolian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yucuañe Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tnu’u savi"),
    },
    LanguageData {
        code_3: [109, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mulgi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miyako",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mekmek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minaveha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marovo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Massep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpotovoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marfa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagal Murut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Machinga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meoswar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indus Kohistani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 118, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mesqan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwatebu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juwal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Are",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murrinh-Patha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aiklep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mouk-Aria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Labo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kita Maninkakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mirandese",
        #[cfg(feature = "local_names")]
        autonym: Some("Mirandés"),
    },
    LanguageData {
        code_3: [109, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyamwanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Maewo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kala Lagaw Ya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mün Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("K’cho"),
    },
    LanguageData {
        code_3: [109, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwimbi-Muthambi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moken",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mittu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mentawai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hmong Daw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 119, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moingi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwest Oaxaca Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un sav"),
    },
    LanguageData {
        code_3: [109, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tezoatlán Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un nda’i"),
    },
    LanguageData {
        code_3: [109, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manyika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Modang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mele-Fila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malgbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbangala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mvuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mozarabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miju-Mishmi",
        #[cfg(feature = "local_names")]
        autonym: Some("Kman"),
    },
    LanguageData {
        code_3: [109, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Monumbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maxi Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meramera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbowe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlahuitoltepec Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juquila Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huitepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un sav"),
    },
    LanguageData {
        code_3: [109, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jamiltepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mada",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Metlatónoc Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mahou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Nochixtlán Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 120, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Masela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 97],
        code_1: Some([109, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Burmese",
        #[cfg(feature = "local_names")]
        autonym: Some("မြန\u{103a}မာ"),
    },
    LanguageData {
        code_3: [109, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayeka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Myene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bambassi",
        #[cfg(feature = "local_names")]
        autonym: Some("Màwés Aasʼè"),
    },
    LanguageData {
        code_3: [109, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangayat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamara Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Me'en",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anfillo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pirahã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muniche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mesmes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundurukú",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Erzya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muyuw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masaaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Mandaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa María Zacatepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un Va’a"),
    },
    LanguageData {
        code_3: [109, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumzabt",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Madagascar Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malimba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Monastic Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wichí Lhamtés Güisnay",
        #[cfg(feature = "local_names")]
        autonym: Some("Wichí Lhamtés"),
    },
    LanguageData {
        code_3: [109, 122, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ixcatlán Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("En ningotsie"),
    },
    LanguageData {
        code_3: [109, 122, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nigeria Mambila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mazatlán Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mumuye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mazanderani",
        #[cfg(feature = "local_names")]
        autonym: Some("مازرونی"),
    },
    LanguageData {
        code_3: [109, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matipuhy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Movima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mori Atas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marúbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mintil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inapang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Deg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawayana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mozambican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [109, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maiadomu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Nambikuára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naka'ela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nabak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naga Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nakanai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nalik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngan'gityemerri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Min Nan Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("闽南语\u{200e} (Minnanyu)"),
    },
    LanguageData {
        code_3: [110, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neapolitan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khoekhoe",
        #[cfg(feature = "local_names")]
        autonym: Some("Khoekhoegowab"),
    },
    LanguageData {
        code_3: [110, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iguta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naasioi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ca\u{331}hungwa\u{331}rya\u{331}",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 117],
        code_1: Some([110, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Nauru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 118],
        code_1: Some([110, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Navajo",
        #[cfg(feature = "local_names")]
        autonym: Some("Diné"),
    },
    LanguageData {
        code_3: [110, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nawuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nakwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarrindjeri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coatepec Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndoe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chang Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbinda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konyak Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nagarchal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngamo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mao Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarinyman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 108],
        code_1: Some([110, 114]),
        #[cfg(feature = "english_names")]
        name_en: "South Ndebele",
        #[cfg(feature = "local_names")]
        autonym: Some("isiNdebele"),
    },
    LanguageData {
        code_3: [110, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbaka Ma'bo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkukoli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nnam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nggem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Numana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namibian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Na",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rongmei Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngamambo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Ngbandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ningera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Nicobarese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ponam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nachering",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Notsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nisga'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Huasteca Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Puebla Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Na-kara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Michoacán Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nambo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nauna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sibe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Katang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ncane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nicaraguan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chothe Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chumburung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Puebla Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Natchez",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenswei Nsei",
        #[cfg(feature = "local_names")]
        autonym: Some("Nsei"),
    },
    LanguageData {
        code_3: [110, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nde-Nsele-Nta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 101],
        code_1: Some([110, 100]),
        #[cfg(feature = "english_names")]
        name_en: "North Ndebele",
        #[cfg(feature = "local_names")]
        autonym: Some("isiNdebele"),
    },
    LanguageData {
        code_3: [110, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nadruvian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndengereko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samba Leko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 111],
        code_1: Some([110, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Ndonga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndombe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndoola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Low German",
        #[cfg(feature = "local_names")]
        autonym: Some("Neddersass’sch"),
    },
    LanguageData {
        code_3: [110, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndunga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dugun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndut",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndut"),
    },
    LanguageData {
        code_3: [110, 100, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nduga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lutos",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndogo",
        #[cfg(feature = "local_names")]
        autonym: Some("Co Ndogo"),
    },
    LanguageData {
        code_3: [110, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Ngad'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nedebang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nde-Gbite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nêlêmwa-Nixumwak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nefamese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Negidal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyenkha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neo-Hittite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nemi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nengone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ná-Meo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 112],
        code_1: Some([110, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Nepali",
        #[cfg(feature = "local_names")]
        autonym: Some("न\u{947}पाली"),
    },
    LanguageData {
        code_3: [110, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Central Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yahadian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bhoti Kinnauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyaheun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Newari",
        #[cfg(feature = "local_names")]
        autonym: Some("न\u{947}वाः भाय\u{94d}\u{200e} (Newah Bhaaye)"),
    },
    LanguageData {
        code_3: [110, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Neyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nez Perce",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dhao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 102, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ahwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 102, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayiwo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 102, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nafaanra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mfumte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Ngbandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngombe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngando",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbaka Manza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nǁng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngizim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dalabon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lomwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngatik Men's Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngwo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngurimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Engdewu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gvoko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kriang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guerrero Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nagumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngwaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nggwahyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tibea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngungwel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nhanda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Beng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabasco Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiripá",
        #[cfg(feature = "local_names")]
        autonym: Some("Ava Guaraní"),
    },
    LanguageData {
        code_3: [110, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Huasteca Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nhuwala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetelcingo Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nahari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zacatlán-Ahuacatlán-Tepetzintla Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isthmus-Cosoleacaque Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morelos Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takuu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isthmus-Pajapan Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huaxcaleca Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naro",
        #[cfg(feature = "local_names")]
        autonym: Some("Naro"),
    },
    LanguageData {
        code_3: [110, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ometepec Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temascaltepec Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Huasteca Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isthmus-Mecayapan Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Oaxaca Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa María La Alta Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nias",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nakame",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niellim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngalakgan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyiha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngaju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Nicobarese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nilamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ninzo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nganasan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimboran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Kolami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niuean",
        #[cfg(feature = "local_names")]
        autonym: Some("Vagahau Niuē"),
    },
    LanguageData {
        code_3: [110, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gilyak",
        #[cfg(feature = "local_names")]
        autonym: Some("Nivkh"),
    },
    LanguageData {
        code_3: [110, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngiti",
        #[cfg(feature = "local_names")]
        autonym: Some("Ndruna"),
    },
    LanguageData {
        code_3: [110, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ningil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nzanyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nocte Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndonde Hamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lotha Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gudanji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Njen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Njalgulgule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angami Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liangmai Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ao Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Njerep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nisa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndyuka-Trio Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngadjunmaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Njyem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 106, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyishi",
        #[cfg(feature = "local_names")]
        autonym: Some("Nil"),
    },
    LanguageData {
        code_3: [110, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkoya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khoibu Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkongho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koireng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inpui Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Inpui"),
    },
    LanguageData {
        code_3: [110, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nekgini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khezha Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thangal Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nakai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nokuku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkangala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkonya",
        #[cfg(feature = "local_names")]
        autonym: Some("Nkunya"),
    },
    LanguageData {
        code_3: [110, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niuatoputapu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukuoro",
        #[cfg(feature = "local_names")]
        autonym: Some("Nukuoro"),
    },
    LanguageData {
        code_3: [110, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Asmat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bouna Kulango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkutu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkoroo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nkari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngombale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nalca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 100],
        code_1: Some([110, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Dutch",
        #[cfg(feature = "local_names")]
        autonym: Some("Nederlands"),
    },
    LanguageData {
        code_3: [110, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Nyala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Grangali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ninia Yali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nihali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mankiyali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lao Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nchumbulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orizaba Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walangama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nahali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyamal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nalögo",
        #[cfg(feature = "local_names")]
        autonym: Some("Nalögo"),
    },
    LanguageData {
        code_3: [110, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maram Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Big Nambas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndumu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mzieme Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangkhul Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwasio",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwasio"),
    },
    LanguageData {
        code_3: [110, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Monsang Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngombe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namakura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndemli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manangba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǃXóõ",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moyon Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimanbur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nambya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimbari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Letemboi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namonuito",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeast Maidu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngamini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namuyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nawdm",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyangumarta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nande",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nancere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Ambae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngandyera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngaing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maring Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngiemboon",
        #[cfg(feature = "local_names")]
        autonym: Some("Shwóŋò ngiembɔɔn"),
    },
    LanguageData {
        code_3: [110, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Nuaulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyangatom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nankina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Rengma Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 111],
        code_1: Some([110, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Norwegian Nynorsk",
        #[cfg(feature = "local_names")]
        autonym: Some("nynorsk"),
    },
    LanguageData {
        code_3: [110, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wancho Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngindo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narungga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanticoke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dwang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nugunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Nuni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyangga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nda'nda'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woun Meu",
        #[cfg(feature = "local_names")]
        autonym: Some("Woun Meu"),
    },
    LanguageData {
        code_3: [110, 111, 98],
        code_1: Some([110, 98]),
        #[cfg(feature = "english_names")]
        name_en: "Norwegian Bokmål",
        #[cfg(feature = "local_names")]
        autonym: Some("norsk bokmål"),
    },
    LanguageData {
        code_3: [110, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Thai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nimadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nomane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nogai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nomu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nonuya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nooksack",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nomlaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nocamán",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Norse",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Numanggang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 114],
        code_1: Some([110, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Norwegian",
        #[cfg(feature = "local_names")]
        autonym: Some("Norsk"),
    },
    LanguageData {
        code_3: [110, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Nisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nomatsiguenga",
        #[cfg(feature = "local_names")]
        autonym: Some("Nomatsigenga"),
    },
    LanguageData {
        code_3: [110, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ewage-Notu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Novial",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyambo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nar Phu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nupbikha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ponyo-Gongwang Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Gongwang Naga, Ponyo Naga"),
    },
    LanguageData {
        code_3: [110, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phom Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nepali",
        #[cfg(feature = "local_names")]
        autonym: Some("न\u{947}पाली\u{200e} (Nepālī)"),
    },
    LanguageData {
        code_3: [110, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Puebla Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mondropolon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pochuri Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nipsan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puimei Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noipx",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Napu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Nago",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kura Ede Nago",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngendelengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "N'Ko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyan-Karyaw Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nteng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 113, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akyaung Ari Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noric",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Rengma Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jèrriais",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chokri Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarluma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Norn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Picene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Norra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Kalapuya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngurmbur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangtam Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Nossob",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Nisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nsenga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Nisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngoshie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nigerian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naskapi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Norwegian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumi Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nehan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pedi",
        #[cfg(feature = "local_names")]
        autonym: Some("Sesotho sa Leboa"),
    },
    LanguageData {
        code_3: [110, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nepalese Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Sierra Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maritime Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tase Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Tangshang Naga"),
    },
    LanguageData {
        code_3: [110, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sierra Negra Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Nisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Navut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nsongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nasal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nisenan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tidung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nathembo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngantangarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Natioro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngaanyatjarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikoma-Nata-Isenye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nateni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ntomba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tepehuan",
        #[cfg(feature = "local_names")]
        autonym: Some("O’otham"),
    },
    LanguageData {
        code_3: [110, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Delo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Natügu",
        #[cfg(feature = "local_names")]
        autonym: Some("Natügu"),
    },
    LanguageData {
        code_3: [110, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nottoway",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangkhul Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Tangkhul Naga"),
    },
    LanguageData {
        code_3: [110, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mantsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Natanzi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukuini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngundu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nusu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nungali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngumbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nuu-chah-nulth",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nusa Laut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Niuafo'ou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anong",
        #[cfg(feature = "local_names")]
        autonym: Some("Anong"),
    },
    LanguageData {
        code_3: [110, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nguôn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nupe-Nupe-Tako",
        #[cfg(feature = "local_names")]
        autonym: Some("Nupe"),
    },
    LanguageData {
        code_3: [110, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukumanu",
        #[cfg(feature = "local_names")]
        autonym: Some("Nukumanu"),
    },
    LanguageData {
        code_3: [110, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukuria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nuer",
        #[cfg(feature = "local_names")]
        autonym: Some("Thok Nath"),
    },
    LanguageData {
        code_3: [110, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngbundu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Nuni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nguluwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mehek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nunggubuyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlamacazapa Nahuatl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nasarian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Namiae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyokon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nawathinehena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyabwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Newari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngayawung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwest Tanna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyamusa-Molo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nauo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nawaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndwewe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Newar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nottoway-Meherrin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nauete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngando",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nage",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngad'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nindi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koki Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Nuaulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Numidian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngawun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndambomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naxi",
        #[cfg(feature = "local_names")]
        autonym: Some("Naqxi"),
    },
    LanguageData {
        code_3: [110, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ninggerum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nafri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 97],
        code_1: Some([110, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Nyanja",
        #[cfg(feature = "local_names")]
        autonym: Some("Chichewa, chiCheŵa\u{200e} (Chichewa), chiNyanja"),
    },
    LanguageData {
        code_3: [110, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyangbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyanga-li",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Giryama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyindu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyikina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyaneka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyeu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyamwezi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyankole",
        #[cfg(feature = "local_names")]
        autonym: Some("Runyankore"),
    },
    LanguageData {
        code_3: [110, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyang'i",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nayini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyiha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyungar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyawaygi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyungwe",
        #[cfg(feature = "local_names")]
        autonym: Some("Cinyungwe"),
    },
    LanguageData {
        code_3: [110, 121, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyulnyul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nganyaywana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyakyusa-Ngonde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tigon Mbembe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Njebi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nzadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nzima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nzakara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zeme Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "New Zealand Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Nzikou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nzakambay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [110, 122, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanga Dama Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oroch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Avar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obispeño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oblo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moabite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obo Manobo",
        #[cfg(feature = "local_names")]
        autonym: Some("Manobo"),
    },
    LanguageData {
        code_3: [111, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Burmese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Breton",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Obulom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ocaina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 99, 105],
        code_1: Some([111, 99]),
        #[cfg(feature = "english_names")]
        name_en: "Occitan",
        #[cfg(feature = "local_names")]
        autonym: Some("occitan"),
    },
    LanguageData {
        code_3: [111, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Cham",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Cornish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atzingo Matlatzinca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Odut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Od",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Dutch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Odual",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ofo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Frisian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 102, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Efutop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ogbia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ogbah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Georgian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ogbogolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ogbronuagum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Hittite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Hungarian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oirata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okolie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inebu One",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 105],
        code_1: Some([111, 106]),
        #[cfg(feature = "english_names")]
        name_en: "Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Japanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Severn Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ontong Java",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Ojibwa",
        #[cfg(feature = "local_names")]
        autonym: Some("Anishnaubemowin"),
    },
    LanguageData {
        code_3: [111, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okanagan",
        #[cfg(feature = "local_names")]
        autonym: Some("Nsyilxcen"),
    },
    LanguageData {
        code_3: [111, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okodia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okpe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koko Babangk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koresh-e Rostam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okiek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oko-Juwoi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwamtim One",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Kentish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Korean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oki-No-Erabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Korean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kirike",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oko-Eni-Osayen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orokaiva",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okpe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Khmer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walungge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mochi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olekha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olkol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oloma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Livvi",
        #[cfg(feature = "local_names")]
        autonym: Some("Livvin kieli"),
    },
    LanguageData {
        code_3: [111, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olrat",
        #[cfg(feature = "local_names")]
        autonym: Some("Olrat"),
    },
    LanguageData {
        code_3: [111, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Lithuanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuvale",
        #[cfg(feature = "local_names")]
        autonym: Some("Kuvale, Olukuvale"),
    },
    LanguageData {
        code_3: [111, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omaha-Ponca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Ambae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mochica",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omagua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ombo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minoan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Utarmbung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Manipuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Marathi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omotik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Omurano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Tairora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Mon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lingao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oneida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Olo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Onin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Onjob",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabore One",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Onobasulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Onondaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sartang",
        #[cfg(feature = "local_names")]
        autonym: Some("Sartang"),
    },
    LanguageData {
        code_3: [111, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern One",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ontenu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Nubian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Onin Based Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tohono O'odham",
        #[cfg(feature = "local_names")]
        autonym: Some("Tohono O’otham"),
    },
    LanguageData {
        code_3: [111, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Önge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oorlams",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Ossetic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Okpamheri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kopkaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oksapmin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Opao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Opata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ofayé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oroha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orejón",
        #[cfg(feature = "local_names")]
        autonym: Some("Maijuna"),
    },
    LanguageData {
        code_3: [111, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oring",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oroqen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 105],
        code_1: Some([111, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Oriya",
        #[cfg(feature = "local_names")]
        autonym: Some("ଓଡ\u{b3c}\u{b3f}ଆ"),
    },
    LanguageData {
        code_3: [111, 114, 109],
        code_1: Some([111, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Oromo",
        #[cfg(feature = "local_names")]
        autonym: Some("Oromoo"),
    },
    LanguageData {
        code_3: [111, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orang Kanaq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orokolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oruma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orang Seletar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adivasi Oriya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ormuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Russian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oro Win",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Odia",
        #[cfg(feature = "local_names")]
        autonym: Some("ଓଡ\u{b3c}\u{b3f}ଆ\u{200e} (Oḍiā)"),
    },
    LanguageData {
        code_3: [111, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ormu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Osage",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oscan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Osing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Sundanese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ososo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Spanish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 115],
        code_1: Some([111, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Ossetian",
        #[cfg(feature = "local_names")]
        autonym: Some("ирон"),
    },
    LanguageData {
        code_3: [111, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Osatu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern One",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Saxon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ottoman Turkish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Tibetan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ot Danum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mezquital Otomi",
        #[cfg(feature = "local_names")]
        autonym: Some("Hñahñu"),
    },
    LanguageData {
        code_3: [111, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Turkish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tilapa Otomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Highland Otomi",
        #[cfg(feature = "local_names")]
        autonym: Some("Yuhu"),
    },
    LanguageData {
        code_3: [111, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tenango Otomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Querétaro Otomi",
        #[cfg(feature = "local_names")]
        autonym: Some("Hñohño"),
    },
    LanguageData {
        code_3: [111, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Otoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Estado de México Otomi",
        #[cfg(feature = "local_names")]
        autonym: Some("Hñatho"),
    },
    LanguageData {
        code_3: [111, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temoaya Otomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Otuke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ottawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Texcatepec Otomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Tamil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ixtenco Otomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagargrent",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Glio-Oubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oune",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Uighur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ouma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elfdalian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Owiniga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Welsh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 121, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oyda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayampi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oya'oya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [111, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koonzime",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parecís",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pacoh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paumarí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pagibete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paranawát",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangasinan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tenharim",
        #[cfg(feature = "local_names")]
        autonym: Some("Kagwahiva"),
    },
    LanguageData {
        code_3: [112, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parakanã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahlavi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pampanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 110],
        code_1: Some([112, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Panjabi",
        #[cfg(feature = "local_names")]
        autonym: Some("ਪ\u{a70}ਜਾਬੀ"),
    },
    LanguageData {
        code_3: [112, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Paiute",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papiamento",
        #[cfg(feature = "local_names")]
        autonym: Some("Papiamentu"),
    },
    LanguageData {
        code_3: [112, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panamint",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papasena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palauan",
        #[cfg(feature = "local_names")]
        autonym: Some("Tekoi ra Belau"),
    },
    LanguageData {
        code_3: [112, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pakaásnovos",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pawnee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pankararé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pech",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pankararú",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Páez",
        #[cfg(feature = "local_names")]
        autonym: Some("Nasa Yuwe"),
    },
    LanguageData {
        code_3: [112, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Patamona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mezontla Popoloca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coyotepec Popoloca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paraujano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "E'ñapa Woromaipu",
        #[cfg(feature = "local_names")]
        autonym: Some("E’ñepa"),
    },
    LanguageData {
        code_3: [112, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parkwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puebla Mazatec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kpasam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Badyara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Pame",
        #[cfg(feature = "local_names")]
        autonym: Some("Xi’oi"),
    },
    LanguageData {
        code_3: [112, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Pashto",
        #[cfg(feature = "local_names")]
        autonym: Some("پښتو\u{200e}\u{200e} (Pax\u{30c}tō)"),
    },
    LanguageData {
        code_3: [112, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Pashto",
        #[cfg(feature = "local_names")]
        autonym: Some("پښتو\u{200e} (Pashto)"),
    },
    LanguageData {
        code_3: [112, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pnar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Inés Ahuatempan Popoloca",
        #[cfg(feature = "local_names")]
        autonym: Some("Nquivā"),
    },
    LanguageData {
        code_3: [112, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pear",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bouyei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Picard",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruching Palaung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paliyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paniya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pardhan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Duruwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parenga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paite Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Paite Chin"),
    },
    LanguageData {
        code_3: [112, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pardhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nigerian Pidgin",
        #[cfg(feature = "local_names")]
        autonym: Some("Naijá, Pidgin"),
    },
    LanguageData {
        code_3: [112, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pacahuara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pyapun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pennsylvania German",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pa Di",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Podena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Padoe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plautdietsch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayan",
        #[cfg(feature = "local_names")]
        autonym: Some("Kayan, Kayan"),
    },
    LanguageData {
        code_3: [112, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Peranakan Indonesian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeastern Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bonan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chichimeca-Jonaz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Penchal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pekal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Persian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iranian Persian",
        #[cfg(feature = "local_names")]
        autonym: Some("فارسی\u{200e} (Fārsi)"),
    },
    LanguageData {
        code_3: [112, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pémono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Petats",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Petjo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Penan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pááfang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 102, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pfaelzisch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sudanese Creole Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("Arabi Juba"),
    },
    LanguageData {
        code_3: [112, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gāndhārī",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pagi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rerep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Primitive Irish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paelignian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangseng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pagu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papua New Guinean Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pa-Hng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phudagi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phuong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phukha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phalura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phimbi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phoenician",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phunoi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phana'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahari-Potwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phu Thai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pahlavani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phangduwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pima Bajo",
        #[cfg(feature = "local_names")]
        autonym: Some("Oob No’ok"),
    },
    LanguageData {
        code_3: [112, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yine",
        #[cfg(feature = "local_names")]
        autonym: Some("Yine"),
    },
    LanguageData {
        code_3: [112, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piaroa",
        #[cfg(feature = "local_names")]
        autonym: Some("De’aruwã thiwene"),
    },
    LanguageData {
        code_3: [112, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pingelapese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pisabo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pitcairn-Norfolk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pijao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Powhatan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piame",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piapoco",
        #[cfg(feature = "local_names")]
        autonym: Some("Wenewika"),
    },
    LanguageData {
        code_3: [112, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piratapuyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pijin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pitta Pitta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pintupi-Luritja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pileni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pimbwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piya-Kwonci",
        #[cfg(feature = "local_names")]
        autonym: Some("Ambandi"),
    },
    LanguageData {
        code_3: [112, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pije",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pitjantjatjara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ardhamāgadhī Prākrit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pokomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paekche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pak-Tong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pankhu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pakanha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pökoot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pukapuka",
        #[cfg(feature = "local_names")]
        autonym: Some("Te Leo Wale"),
    },
    LanguageData {
        code_3: [112, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Attapady Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pakistan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maleng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Polonombauk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Palawano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Polari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palu'e",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pilagá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paulohi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 105],
        code_1: Some([112, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Pali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Polci",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kohistani Shina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shwe Palaung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palenquero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oluta Popoluca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palaka Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Marcos Tlacoyalco Popoloca",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngigua"),
    },
    LanguageData {
        code_3: [112, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plateau Malagasy",
        #[cfg(feature = "local_names")]
        autonym: Some("Malagasy, Malagasy fiteny"),
    },
    LanguageData {
        code_3: [112, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Palikúr",
        #[cfg(feature = "local_names")]
        autonym: Some("Pa’ikwaki"),
    },
    LanguageData {
        code_3: [112, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwest Palawano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brooke's Point Palawano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paluan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pambia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pallanganmiddang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pwaamei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pamona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Māhārāṣṭri Prākrit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Pumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Pumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pamlico",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lingua Franca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Pame",
        #[cfg(feature = "local_names")]
        autonym: Some("Xi’iuy"),
    },
    LanguageData {
        code_3: [112, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paynamar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piemontese",
        #[cfg(feature = "local_names")]
        autonym: Some("Piemontèis"),
    },
    LanguageData {
        code_3: [112, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuamotuan",
        #[cfg(feature = "local_names")]
        autonym: Some("Pa’umotu"),
    },
    LanguageData {
        code_3: [112, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plains Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Poumei Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papuan Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Melayu Papua"),
    },
    LanguageData {
        code_3: [112, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Pame",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Bah-Biau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Panjabi",
        #[cfg(feature = "local_names")]
        autonym: Some("پنجابی\u{200e} (Panjābī)"),
    },
    LanguageData {
        code_3: [112, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pannei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpinda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Penan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangu",
        #[cfg(feature = "local_names")]
        autonym: Some("Tarin"),
    },
    LanguageData {
        code_3: [112, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Penrhyn",
        #[cfg(feature = "local_names")]
        autonym: Some("Mangarongaro, Tongareva"),
    },
    LanguageData {
        code_3: [112, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aoheng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinjarup",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paunaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paleni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Batu 1",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinai-Hagahai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pancana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ponosakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pontic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jiongnai Bunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinigura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banyjima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phong-Kniang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pinyin",
        #[cfg(feature = "local_names")]
        autonym: Some("Menka, Pinyin"),
    },
    LanguageData {
        code_3: [112, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Poqomam",
        #[cfg(feature = "local_names")]
        autonym: Some("Qaq’oral"),
    },
    LanguageData {
        code_3: [112, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Juan Atzingo Popoloca",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngigua"),
    },
    LanguageData {
        code_3: [112, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Poke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Potiguára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Poqomchi'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Highland Popoluca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pokangá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 108],
        code_1: Some([112, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Polish",
        #[cfg(feature = "local_names")]
        autonym: Some("polski"),
    },
    LanguageData {
        code_3: [112, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pohnpeian",
        #[cfg(feature = "local_names")]
        autonym: Some("Pohnpei"),
    },
    LanguageData {
        code_3: [112, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Pomo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pwapwâ",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Texistepec Popoluca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 114],
        code_1: Some([112, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Portuguese",
        #[cfg(feature = "local_names")]
        autonym: Some("português"),
    },
    LanguageData {
        code_3: [112, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sayula Popoluca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Potawatomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Guinea Crioulo",
        #[cfg(feature = "local_names")]
        autonym: Some("Kriyol"),
    },
    LanguageData {
        code_3: [112, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Felipe Otlaltepec Popoloca",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngiba"),
    },
    LanguageData {
        code_3: [112, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Polabian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pogolo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paipai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uma",
        #[cfg(feature = "local_names")]
        autonym: Some("Uma"),
    },
    LanguageData {
        code_3: [112, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pipil",
        #[cfg(feature = "local_names")]
        autonym: Some("Nahuat"),
    },
    LanguageData {
        code_3: [112, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papapana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Folopa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pelende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Luís Temalacayuca Popoloca",
        #[cfg(feature = "local_names")]
        autonym: Some("Ngigua"),
    },
    LanguageData {
        code_3: [112, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pa'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malecite-Passamaquoddy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parsi-Dari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Principense",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paranan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Prussian",
        #[cfg(feature = "local_names")]
        autonym: Some("prūsiskan"),
    },
    LanguageData {
        code_3: [112, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Porohanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paicî",
        #[cfg(feature = "local_names")]
        autonym: Some("Paicî"),
    },
    LanguageData {
        code_3: [112, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parauk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Peruvian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kibiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Prasuni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Provençal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ashéninka Perené",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dari",
        #[cfg(feature = "local_names")]
        autonym: Some("فارسی\u{200e} (Farsi)"),
    },
    LanguageData {
        code_3: [112, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puragi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parawen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Purik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Providencia Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asue Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iranian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Plains Indian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Penang Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwest Pashai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeast Pashai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puerto Rican Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pauserna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panasuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Polish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Philippine Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Portuguese Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaulong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Pashto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sauraseni Prākrit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Port Sandwich",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piscataway",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pai Tavytera",
        #[cfg(feature = "local_names")]
        autonym: Some("Pai Tavyterã"),
    },
    LanguageData {
        code_3: [112, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pataxó Hã-Ha-Hãe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pindiini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Patani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zo'é",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Patep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pattapu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Piamatsina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enrekang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bambam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Port Vato",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pentlatch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pathiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Highland Purepecha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Purum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Merap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Aput",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puelche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Merah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phuie",
        #[cfg(feature = "local_names")]
        autonym: Some("Phuie"),
    },
    LanguageData {
        code_3: [112, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puinave",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punan Tubu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puoc",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pulabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puquina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puruborá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 115],
        code_1: Some([112, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Pushto",
        #[cfg(feature = "local_names")]
        autonym: Some("پښتو"),
    },
    LanguageData {
        code_3: [112, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Putoh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puluwatese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Purisimeño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pawaia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gapapaiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Patwin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molbog",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paiwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pwo Western Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Powari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pwo Northern Karen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quetzaltepec Mixe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pye Krumen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fyam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Poyanáwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paraguayan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puyuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pyen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pazeh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [112, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jejara Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Naga"),
    },
    LanguageData {
        code_3: [113, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quapaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huallaga Huánuco Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "K'iche'",
        #[cfg(feature = "local_names")]
        autonym: Some("Kʼicheʼ"),
    },
    LanguageData {
        code_3: [113, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Calderón Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 101],
        code_1: Some([113, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Runasimi"),
    },
    LanguageData {
        code_3: [113, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lambayeque Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chimborazo Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Bolivian Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quileute",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chachapoyas Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Llakwash"),
    },
    LanguageData {
        code_3: [113, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Bolivian Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Quechua"),
    },
    LanguageData {
        code_3: [113, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sipacapense",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quinault",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Pastaza Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Inka"),
    },
    LanguageData {
        code_3: [113, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quinqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanahuanca Pasco Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santiago del Estero Quichua",
        #[cfg(feature = "local_names")]
        autonym: Some("Quichua"),
    },
    LanguageData {
        code_3: [113, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sacapulteco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tena Lowland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yauyos Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayacucho Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Runasimi"),
    },
    LanguageData {
        code_3: [113, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cusco Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Qheswasimi, Runasimi"),
    },
    LanguageData {
        code_3: [113, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ambo-Pasco Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cajamarca Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Kichwa"),
    },
    LanguageData {
        code_3: [113, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Apurímac Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huamalíes-Dos de Mayo Huánuco Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Imbabura Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loja Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cajatambo North Lima Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Margos-Yarowilca-Lauricocha Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Junín Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Napo Lowland Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Runa Shimi"),
    },
    LanguageData {
        code_3: [113, 118, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pacaraos Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Martín Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Llakwash Quechua"),
    },
    LanguageData {
        code_3: [113, 118, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huaylla Wanca Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Queyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 118, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Pastaza Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Corongo Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huaylas Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Quechua"),
    },
    LanguageData {
        code_3: [113, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sihuas Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwalhioqua-Tlatskanai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chiquián Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chincha Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panao Huánuco Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Panao runacuna"),
    },
    LanguageData {
        code_3: [113, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salasaca Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Conchucos Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: Some("Quechua"),
    },
    LanguageData {
        code_3: [113, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Conchucos Ancash Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puno Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qashqa'i",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cañar Highland Quichua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Qiang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Ana de Tusi Pasco Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arequipa-La Unión Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jauja Wanca Quechua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quenya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [113, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quiripi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dungmali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Camling",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rasawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Meohang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logooli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rabha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ramoaaina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rajasthani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tulu-Bohuai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ralte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Canela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riantana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rapanui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rarotongan",
        #[cfg(feature = "local_names")]
        autonym: Some("Te Reo Maori"),
    },
    LanguageData {
        code_3: [114, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tegali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Razajerdi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Raute",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sampang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rawang",
        #[cfg(feature = "local_names")]
        autonym: Some("Nung, Rawang"),
    },
    LanguageData {
        code_3: [114, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rapa",
        #[cfg(feature = "local_names")]
        autonym: Some("Reo Rapa"),
    },
    LanguageData {
        code_3: [114, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rahambuu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rumai Palaung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miraya Bikol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barababaraba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Réunion Creole French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rudbari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rerau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rembong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rejang Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Reli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rejang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rendille",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Remo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rengao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rer Bare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Reshe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Retta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Reyesano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Roria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romano-Greek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rangkas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romagnol",
        #[cfg(feature = "local_names")]
        autonym: Some("Rumagnol"),
    },
    LanguageData {
        code_3: [114, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Resígaro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Roglai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ringgou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rohingya",
        #[cfg(feature = "local_names")]
        autonym: Some("Ruwainggya"),
    },
    LanguageData {
        code_3: [114, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yahang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bribri Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tarifit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riang Lang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyaturu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ribun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ritharrngu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rajong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Raji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rajbanshi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kraol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rikbaktsa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rakahanga-Manihiki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rakhine",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rangpuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arakwal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rembarrnga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carpathian Romani",
        #[cfg(feature = "local_names")]
        autonym: Some("Romanes"),
    },
    LanguageData {
        code_3: [114, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Traveller Danish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angloromani",
        #[cfg(feature = "local_names")]
        autonym: Some("Romanes, Romani"),
    },
    LanguageData {
        code_3: [114, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalo Finnish Romani",
        #[cfg(feature = "local_names")]
        autonym: Some("Kaalengo tšibbaha"),
    },
    LanguageData {
        code_3: [114, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Traveller Norwegian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Murkim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lomavren",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romkun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baltic Romani",
        #[cfg(feature = "local_names")]
        autonym: Some("Romani"),
    },
    LanguageData {
        code_3: [114, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Roma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balkan Romani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinte Romani",
        #[cfg(feature = "local_names")]
        autonym: Some("Romanes"),
    },
    LanguageData {
        code_3: [114, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rempi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Caló",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Domari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tavringer Romani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romanova",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Welsh Romani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vlax Romani",
        #[cfg(feature = "local_names")]
        autonym: Some("Romani"),
    },
    LanguageData {
        code_3: [114, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brunca Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruund",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ronga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ranglong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Roon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rongpo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nari Nari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rungwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tae'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cacgia Roglai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rogo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ronji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rombo",
        #[cfg(feature = "local_names")]
        autonym: Some("Kihorombo"),
    },
    LanguageData {
        code_3: [114, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Roglai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 104],
        code_1: Some([114, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Romansh",
        #[cfg(feature = "local_names")]
        autonym: Some("rumantsch"),
    },
    LanguageData {
        code_3: [114, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romblomanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romany",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 110],
        code_1: Some([114, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Romanian",
        #[cfg(feature = "local_names")]
        autonym: Some("română"),
    },
    LanguageData {
        code_3: [114, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rotokas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kriol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rongga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Runga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dela-Oenale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Repanbitip",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rapting",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ririo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arritinngithigh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Romano-Serbian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruthenian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Russian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miriwoong Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rwandan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rungtu Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ratahan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rotuman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yurats",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rathawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruuli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rusyn",
        #[cfg(feature = "local_names")]
        autonym: Some("русиньскый язык\u{200e} (Rusynska mova)"),
    },
    LanguageData {
        code_3: [114, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Luguru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Roviana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rufiji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Che",
        #[cfg(feature = "local_names")]
        autonym: Some("Kuche"),
    },
    LanguageData {
        code_3: [114, 117, 110],
        code_1: Some([114, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Rundi",
        #[cfg(feature = "local_names")]
        autonym: Some("Ikirundi"),
    },
    LanguageData {
        code_3: [114, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Istro Romanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Macedo-Romanian",
        #[cfg(feature = "local_names")]
        autonym: Some("Armani"),
    },
    LanguageData {
        code_3: [114, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Megleno Romanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 115],
        code_1: Some([114, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Russian",
        #[cfg(feature = "local_names")]
        autonym: Some("русский"),
    },
    LanguageData {
        code_3: [114, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rutul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lanas Lobu",
        #[cfg(feature = "local_names")]
        autonym: Some("Lobu"),
    },
    LanguageData {
        code_3: [114, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rawo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rwa",
        #[cfg(feature = "local_names")]
        autonym: Some("Kiruwa"),
    },
    LanguageData {
        code_3: [114, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ruwila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amba",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwamba"),
    },
    LanguageData {
        code_3: [114, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marwari",
        #[cfg(feature = "local_names")]
        autonym: Some("मारवाड\u{93c}ी\u{200e} (Mārwāṛī)"),
    },
    LanguageData {
        code_3: [114, 120, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngardi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karuwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Amami-Oshima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaeyama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Okinawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [114, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rāziḥī",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buglere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meskwaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sandawe",
        #[cfg(feature = "local_names")]
        autonym: Some("Sandawso"),
    },
    LanguageData {
        code_3: [115, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabanê",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Safaliba",
        #[cfg(feature = "local_names")]
        autonym: Some("Safaliba"),
    },
    LanguageData {
        code_3: [115, 97, 103],
        code_1: Some([115, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Sango",
        #[cfg(feature = "local_names")]
        autonym: Some("Sängö"),
    },
    LanguageData {
        code_3: [115, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakut",
        #[cfg(feature = "local_names")]
        autonym: Some("саха тыла"),
    },
    LanguageData {
        code_3: [115, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sahu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samaritan Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 110],
        code_1: Some([115, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Sanskrit",
        #[cfg(feature = "local_names")]
        autonym: Some("स\u{902}स\u{94d}क\u{943}त भाषा"),
    },
    LanguageData {
        code_3: [115, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sause",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samburu",
        #[cfg(feature = "local_names")]
        autonym: Some("Kisampur"),
    },
    LanguageData {
        code_3: [115, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saraveca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sasak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santali",
        #[cfg(feature = "local_names")]
        autonym: Some("Har Rar"),
    },
    LanguageData {
        code_3: [115, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saleman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saafi-Saafi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sawi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saurashtra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngambay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Samo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saliba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seget",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sori-Harengan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Surbakhal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Safwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Botolan Sambal",
        #[cfg(feature = "local_names")]
        autonym: Some("Hambali"),
    },
    LanguageData {
        code_3: [115, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sagala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sindhi Bhil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabüm",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangu",
        #[cfg(feature = "local_names")]
        autonym: Some("Ishisangu"),
    },
    LanguageData {
        code_3: [115, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sileibi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sembakung Murut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Subiya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kimki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Stod Bhoti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabine",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seberuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sara Kaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dongxiang",
        #[cfg(feature = "local_names")]
        autonym: Some("Santa"),
    },
    LanguageData {
        code_3: [115, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Miguel Creole French",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanggau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakachep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sri Lankan Creole Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sadri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shina",
        #[cfg(feature = "local_names")]
        autonym: Some("شینا\u{200e} (Šīnā)"),
    },
    LanguageData {
        code_3: [115, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sicilian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Scots",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hyolmo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sa'och",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Slavey",
        #[cfg(feature = "local_names")]
        autonym: Some("Satúotine Yatí"),
    },
    LanguageData {
        code_3: [115, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Katang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shumcho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sheni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sicel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toraja-Sa'dan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shabak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sassarese Sardinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Surubu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Savi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Kurdish",
        #[cfg(feature = "local_names")]
        autonym: Some("کوردی خوارگ"),
    },
    LanguageData {
        code_3: [115, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sos Kundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saudi Arabian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gallurese Sardinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bukar-Sadung Bidayuh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sherdukpen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semandang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oraon Sadri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sened",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shuadit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarudu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sibu Melanau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 100, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sallands",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shempire Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sechelt",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sedang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seneca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cebaara Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Segeju",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sena",
        #[cfg(feature = "local_names")]
        autonym: Some("sena"),
    },
    LanguageData {
        code_3: [115, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seri",
        #[cfg(feature = "local_names")]
        autonym: Some("Cmiique Iitom"),
    },
    LanguageData {
        code_3: [115, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sekani",
        #[cfg(feature = "local_names")]
        autonym: Some("Tse’khene"),
    },
    LanguageData {
        code_3: [115, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selkup",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nanerigé Sénoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suarmin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sìcìté Sénoufo",
        #[cfg(feature = "local_names")]
        autonym: Some("Sìcijuungé"),
    },
    LanguageData {
        code_3: [115, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Senara Sénoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serrano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koyraboro Senni Songhai",
        #[cfg(feature = "local_names")]
        autonym: Some("Koyraboro senni"),
    },
    LanguageData {
        code_3: [115, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sentani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serui-Laut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyarafolo Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sewa Bay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Secoya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Senthang Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Senthang Chin"),
    },
    LanguageData {
        code_3: [115, 102, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langue des signes de Belgique Francophone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 102, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Subanen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 102, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Small Flowery Miao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 102, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South African Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 102, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sehwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Old Irish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mag-antsi Ayta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kipsigis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Surigaonon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Segai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swiss-German Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shughni",
        #[cfg(feature = "local_names")]
        autonym: Some("Khugnone"),
    },
    LanguageData {
        code_3: [115, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Surgujia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangkong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Singa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Singpho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangisari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samogitian",
        #[cfg(feature = "local_names")]
        autonym: Some("Žemaitiu"),
    },
    LanguageData {
        code_3: [115, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Brokpake",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sebat Bet Gurage",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sierra Leone Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanglechi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sursurunga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shall-Zwall",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ninam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sonde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kundal Shahi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sheko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shoshoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tachelhit",
        #[cfg(feature = "local_names")]
        autonym: Some("ⵜⴰⵛⵍⵃⵉⵜ"),
    },
    LanguageData {
        code_3: [115, 104, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shatt",
        #[cfg(feature = "local_names")]
        autonym: Some("Caning"),
    },
    LanguageData {
        code_3: [115, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shilluk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shendu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shahrudi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shipibo-Conibo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shuswap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shasta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chadian Arabic",
        #[cfg(feature = "local_names")]
        autonym: Some("العربية\u{200e} (Alearabia)"),
    },
    LanguageData {
        code_3: [115, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shehri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shwai",
        #[cfg(feature = "local_names")]
        autonym: Some("Cwaya"),
    },
    LanguageData {
        code_3: [115, 104, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "She",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tachawit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Syenara Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akkala Sami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sebop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sidamo",
        #[cfg(feature = "local_names")]
        autonym: Some("Sidaamu Afo"),
    },
    LanguageData {
        code_3: [115, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siamou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Paasaal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zire",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shom Peng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Numbami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sikiana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumulung Sisaala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mende",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 110],
        code_1: Some([115, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Sinhala",
        #[cfg(feature = "local_names")]
        autonym: Some("ස\u{dd2}ංහල"),
    },
    LanguageData {
        code_3: [115, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sikkimese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sonia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siuslaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinagen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumariup",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sivandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Epena",
        #[cfg(feature = "local_names")]
        autonym: Some("Epéna Pedée"),
    },
    LanguageData {
        code_3: [115, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sajau Basap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kildin Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("Кӣллт са\u{304}мь кӣлл"),
    },
    LanguageData {
        code_3: [115, 106, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pite Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("Bidumsámegiella"),
    },
    LanguageData {
        code_3: [115, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Assangori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kemi Sami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sajalong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mapun",
        #[cfg(feature = "local_names")]
        autonym: Some("Jama Mapun"),
    },
    LanguageData {
        code_3: [115, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sindarin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xibe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Surjapuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siar-Lak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Senhaja De Srair",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ter Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("saa´mekiill"),
    },
    LanguageData {
        code_3: [115, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ume Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("Ubmejensámien giella"),
    },
    LanguageData {
        code_3: [115, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shawnee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Skagit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma Manda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Sierra Miwok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakirabiá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakalava Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sikule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sika",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kutong",
        #[cfg(feature = "local_names")]
        autonym: Some("Kutong gin"),
    },
    LanguageData {
        code_3: [115, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kolibugan Subanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seko Tengah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sekapan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sininkere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saraiki",
        #[cfg(feature = "local_names")]
        autonym: Some("سرائیکی\u{200e} (Saraiki)"),
    },
    LanguageData {
        code_3: [115, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Skou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Skepi Creole Dutch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seko Padang",
        #[cfg(feature = "local_names")]
        autonym: Some("Sua Tu Padang"),
    },
    LanguageData {
        code_3: [115, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sikaiana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sekar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sáliba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sissala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sholaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swiss-Italian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selungai Murut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Puget Sound Salish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Silesian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salumá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 107],
        code_1: Some([115, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Slovak",
        #[cfg(feature = "local_names")]
        autonym: Some("slovenčina"),
    },
    LanguageData {
        code_3: [115, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salt-Yui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pangutaran Sama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salinan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lamaholot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salchuq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Singapore Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 118],
        code_1: Some([115, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Slovenian",
        #[cfg(feature = "local_names")]
        autonym: Some("slovenščina"),
    },
    LanguageData {
        code_3: [115, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sialum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salampasu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selayar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 108, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ma'ya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("åarjelsaemien gïele"),
    },
    LanguageData {
        code_3: [115, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simbari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Som",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 101],
        code_1: Some([115, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Northern Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("davvisámegiella"),
    },
    LanguageData {
        code_3: [115, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Auwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simbali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lule Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("julevsámegiella"),
    },
    LanguageData {
        code_3: [115, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolinao",
        #[cfg(feature = "local_names")]
        autonym: Some("Binobolinao"),
    },
    LanguageData {
        code_3: [115, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Sama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Musasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Inari Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("anarâškielâ"),
    },
    LanguageData {
        code_3: [115, 109, 111],
        code_1: Some([115, 109]),
        #[cfg(feature = "english_names")]
        name_en: "Samoan",
        #[cfg(feature = "local_names")]
        autonym: Some("Gagana Samoa"),
    },
    LanguageData {
        code_3: [115, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samaritan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simeulue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Skolt Sami",
        #[cfg(feature = "local_names")]
        autonym: Some("nuõrttsääʹmǩiõll"),
    },
    LanguageData {
        code_3: [115, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Somray",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samvedi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumbawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semnani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Simeku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 97],
        code_1: Some([115, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Shona",
        #[cfg(feature = "local_names")]
        autonym: Some("chiShona"),
    },
    LanguageData {
        code_3: [115, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinaugoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 100],
        code_1: Some([115, 100]),
        #[cfg(feature = "english_names")]
        name_en: "Sindhi",
        #[cfg(feature = "local_names")]
        autonym: Some("سنڌي"),
    },
    LanguageData {
        code_3: [115, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bau Bidayuh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Noon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sensi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riverain Sango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soninke",
        #[cfg(feature = "local_names")]
        autonym: Some("Sooninkanxanne"),
    },
    LanguageData {
        code_3: [115, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Ma'di",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Snohomish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sihan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South West Bay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Senggi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sa'ban",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selee",
        #[cfg(feature = "local_names")]
        autonym: Some("Bale"),
    },
    LanguageData {
        code_3: [115, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saniyo-Hiyewe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thai Song",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sobei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "So",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Songoora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Songomeno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sogdian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sonha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sokoro",
        #[cfg(feature = "local_names")]
        autonym: Some("soŋoroŋ"),
    },
    LanguageData {
        code_3: [115, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Solos",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 109],
        code_1: Some([115, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Somali",
        #[cfg(feature = "local_names")]
        autonym: Some("Soomaali"),
    },
    LanguageData {
        code_3: [115, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Songo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Songe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Somrai",
        #[cfg(feature = "local_names")]
        autonym: Some("Shibne"),
    },
    LanguageData {
        code_3: [115, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seeku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 116],
        code_1: Some([115, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Southern Sotho",
        #[cfg(feature = "local_names")]
        autonym: Some("Sesotho"),
    },
    LanguageData {
        code_3: [115, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Thai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sonsorol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sowanda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miyobe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temi",
        #[cfg(feature = "local_names")]
        autonym: Some("GiTemi"),
    },
    LanguageData {
        code_3: [115, 112, 97],
        code_1: Some([101, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Spanish",
        #[cfg(feature = "local_names")]
        autonym: Some("español"),
    },
    LanguageData {
        code_3: [115, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sepa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sapé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saep",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sepa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saponi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sengo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selepet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akukem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanapaná",
        #[cfg(feature = "local_names")]
        autonym: Some("Nenlhet"),
    },
    LanguageData {
        code_3: [115, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Spokane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Supyire Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loreto-Ucayali Spanish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saparua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saposa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Spiti Bhoti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sapuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sambalpuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Picene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabaot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shama-Sambuga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 105],
        code_1: Some([115, 113]),
        #[cfg(feature = "english_names")]
        name_en: "Albanian",
        #[cfg(feature = "local_names")]
        autonym: Some("shqip"),
    },
    LanguageData {
        code_3: [115, 113, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Albanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Susquehannock",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sorkhei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siculo Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sri Lankan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soqotri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Squamish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 113, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kufr Qassem Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saruga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Logudorese Sardinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 100],
        code_1: Some([115, 99]),
        #[cfg(feature = "english_names")]
        name_en: "Sardinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nafi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sulod",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarikoli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siriano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serudung Murut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isirawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saramaccan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sranan Tongo",
        #[cfg(feature = "local_names")]
        autonym: Some("Sranan, Sranan Tongo"),
    },
    LanguageData {
        code_3: [115, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Campidanese Sardinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 112],
        code_1: Some([115, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Serbian",
        #[cfg(feature = "local_names")]
        autonym: Some("српски"),
    },
    LanguageData {
        code_3: [115, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sirionó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suruí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Sorsoganon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sirmauri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shahmirzadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Sama",
        #[cfg(feature = "local_names")]
        autonym: Some("Sinama, Tawi-Tawi Sinama"),
    },
    LanguageData {
        code_3: [115, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suba-Simbiti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siroi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balangingi",
        #[cfg(feature = "local_names")]
        autonym: Some("Bangingih"),
    },
    LanguageData {
        code_3: [115, 115, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seimat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shihhi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sansi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sausi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sunam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Sisaala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semnam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sissano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Spanish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "So'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swiss-French Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sô",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinasina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Susuami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shark Bay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 119],
        code_1: Some([115, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Swati",
        #[cfg(feature = "local_names")]
        autonym: Some("siSwati"),
    },
    LanguageData {
        code_3: [115, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samberigi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sengseng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Settla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Subanen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sentinel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liana-Seti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trieng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shelta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bulo Stieng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matya Samo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arammba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Stellingwerfs",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Setaman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Owa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Stoney",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Tepehuan",
        #[cfg(feature = "local_names")]
        autonym: Some("O’dam"),
    },
    LanguageData {
        code_3: [115, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saterfriesisch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Straits Salish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shumashti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Budeh Stieng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samtao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Silt'e",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Satawalese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siberian Tatar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sulka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Subanon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suganga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sukuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 110],
        code_1: Some([115, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Sundanese",
        #[cfg(feature = "local_names")]
        autonym: Some("Sunda"),
    },
    LanguageData {
        code_3: [115, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bouni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tirmaga-Chai Suri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwaghavul",
        #[cfg(feature = "local_names")]
        autonym: Some("Mwaghavul"),
    },
    LanguageData {
        code_3: [115, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Susu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Subtiaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puroik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumbwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sumerian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suyá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sunwar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Svan",
        #[cfg(feature = "local_names")]
        autonym: Some("Lushnu"),
    },
    LanguageData {
        code_3: [115, 118, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulau-Suain",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 118, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vincentian Creole English",
        #[cfg(feature = "local_names")]
        autonym: Some("Vincy Twang"),
    },
    LanguageData {
        code_3: [115, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Serili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Slovakian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Slavomolisano",
        #[cfg(feature = "local_names")]
        autonym: Some("Na-našu"),
    },
    LanguageData {
        code_3: [115, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Savosavo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 118, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Skalvian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 97],
        code_1: Some([115, 119]),
        #[cfg(feature = "english_names")]
        name_en: "Swahili",
        #[cfg(feature = "local_names")]
        autonym: Some("Kiswahili"),
    },
    LanguageData {
        code_3: [115, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maore Comorian",
        #[cfg(feature = "local_names")]
        autonym: Some("Maore Comorian, Shimaore"),
    },
    LanguageData {
        code_3: [115, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Congo Swahili",
        #[cfg(feature = "local_names")]
        autonym: Some("Kiswahili"),
    },
    LanguageData {
        code_3: [115, 119, 101],
        code_1: Some([115, 118]),
        #[cfg(feature = "english_names")]
        name_en: "Swedish",
        #[cfg(feature = "local_names")]
        autonym: Some("svenska"),
    },
    LanguageData {
        code_3: [115, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swabian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swahili",
        #[cfg(feature = "local_names")]
        autonym: Some("Kiswahili, Swahili"),
    },
    LanguageData {
        code_3: [115, 119, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malawi Sena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Swedish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samosa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sawknah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shanenawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sharwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saweru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seluwasan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sawila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suwawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shekhawati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suruahá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sicanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sighu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shuhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Kalapuya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Selian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sangir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sorothaptic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saaroa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sasaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Saxon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saxwe Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Siang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Subanen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Syriac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sukur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sylheti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maya Samo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Senaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suoy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Syriac",
        #[cfg(feature = "local_names")]
        autonym: Some("ܣܘܪܝܝܐ"),
    },
    LanguageData {
        code_3: [115, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinyar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagate",
        #[cfg(feature = "local_names")]
        autonym: Some("स\u{94d}य\u{941}बा\u{200e} (Syuba)"),
    },
    LanguageData {
        code_3: [115, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Al-Sayyid Bedouin Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semelai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngalum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Semaq Beri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Seze",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sengele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Silesian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Suabo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Solomon Islands Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sawai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [115, 122, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sakizaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Tanana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabassaran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lowland Tarahumara",
        #[cfg(feature = "local_names")]
        autonym: Some("Rarómari raicha"),
    },
    LanguageData {
        code_3: [116, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tause",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tariana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tapirapé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagoi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 104],
        code_1: Some([116, 121]),
        #[cfg(feature = "english_names")]
        name_en: "Tahitian",
        #[cfg(feature = "local_names")]
        autonym: Some("Reo Tahiti"),
    },
    LanguageData {
        code_3: [116, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Tamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 109],
        code_1: Some([116, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Tamil",
        #[cfg(feature = "local_names")]
        autonym: Some("தமிழ\u{bcd}"),
    },
    LanguageData {
        code_3: [116, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yami",
        #[cfg(feature = "local_names")]
        autonym: Some("Pongso no Tao"),
    },
    LanguageData {
        code_3: [116, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taabwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamasheq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Tarahumara",
        #[cfg(feature = "local_names")]
        autonym: Some("Ralámuli raicha"),
    },
    LanguageData {
        code_3: [116, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tay Boi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 116],
        code_1: Some([116, 116]),
        #[cfg(feature = "english_names")]
        name_en: "Tatar",
        #[cfg(feature = "local_names")]
        autonym: Some("татар"),
    },
    LanguageData {
        code_3: [116, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Tanana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tatuyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Atayal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tocho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aikanã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaki Ae",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanimbili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Tairora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dharawal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaam",
        #[cfg(feature = "local_names")]
        autonym: Some("Gaahmg"),
    },
    LanguageData {
        code_3: [116, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Calamian Tagbanwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tboli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barro Negro Tunebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taworta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumtum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanguat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tembo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tubar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagbanwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabaru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 98, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ditammari",
        #[cfg(feature = "local_names")]
        autonym: Some("Ditammari"),
    },
    LanguageData {
        code_3: [116, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ticuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanacross",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Datooga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tafi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Tutchone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malinaltepec Me'phaa",
        #[cfg(feature = "local_names")]
        autonym: Some("Mè’phàà Mañuwìín"),
    },
    LanguageData {
        code_3: [116, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamagario",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turks And Caicos Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tchitchege",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanahmerah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tichurong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taungyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawr Chin",
        #[cfg(feature = "local_names")]
        autonym: Some("Thawr"),
    },
    LanguageData {
        code_3: [116, 99, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaiy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torres Strait Creole",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "T'en",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Tarahumara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tecpatlán Totonac",
        #[cfg(feature = "local_names")]
        autonym: Some("Totonaca"),
    },
    LanguageData {
        code_3: [116, 99, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 99, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thado Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagdal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Panchpargania",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Emberá-Tadó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Nüa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiranige Diga Dogon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talieng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Tamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thulung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tomadino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tajio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tambas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taruma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tondano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Todrah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doutai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetun Dili",
        #[cfg(feature = "local_names")]
        autonym: Some("Lia-Tetun, Tetun Dili"),
    },
    LanguageData {
        code_3: [116, 100, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tandroy-Mahafaly Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tadyawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temiar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetete",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Terik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tepo Krumen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huehuetla Tepehua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teressa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Tege",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tehuelche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torricelli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ibali Teke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 108],
        code_1: Some([116, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Telugu",
        #[cfg(feature = "local_names")]
        autonym: Some("త\u{c46}లుగు"),
    },
    LanguageData {
        code_3: [116, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Timne",
        #[cfg(feature = "local_names")]
        autonym: Some("Themne"),
    },
    LanguageData {
        code_3: [116, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teso",
        #[cfg(feature = "local_names")]
        autonym: Some("Kiteso"),
    },
    LanguageData {
        code_3: [116, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tepecano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temein",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tereno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tengger",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetum",
        #[cfg(feature = "local_names")]
        autonym: Some("Tetun"),
    },
    LanguageData {
        code_3: [116, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tennet",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tulishi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 101, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetserret",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tofin Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 102, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanaina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 102, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tefaro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 102, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teribe",
        #[cfg(feature = "local_names")]
        autonym: Some("Naso"),
    },
    LanguageData {
        code_3: [116, 102, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ternate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sagalla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobilung",
        #[cfg(feature = "local_names")]
        autonym: Some("Momogun, Tobilung"),
    },
    LanguageData {
        code_3: [116, 103, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tigak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ciwogai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Gorkha Tamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chalikha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobagonian Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lawunuia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 107],
        code_1: Some([116, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Tajik",
        #[cfg(feature = "local_names")]
        autonym: Some("тоҷикӣ"),
    },
    LanguageData {
        code_3: [116, 103, 108],
        code_1: Some([116, 108]),
        #[cfg(feature = "english_names")]
        name_en: "Tagalog",
        #[cfg(feature = "local_names")]
        autonym: Some("Tagalog"),
    },
    LanguageData {
        code_3: [116, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tandaganon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sudest",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tring",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tareng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nume",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Tagbanwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanggu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tingui-Boto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagwana Senoufo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Togoyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 103, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tagalaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 97],
        code_1: Some([116, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Thai",
        #[cfg(feature = "local_names")]
        autonym: Some("ไทย"),
    },
    LanguageData {
        code_3: [116, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuuk Thaayorre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chitwania Tharu",
        #[cfg(feature = "local_names")]
        autonym: Some("मध\u{94d}य क\u{94d}ष\u{947}त\u{94d}रीय थार\u{942}\u{200e} (Madhya Ksetriya Tharu)"),
    },
    LanguageData {
        code_3: [116, 104, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thangmi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tarahumara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Long",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tharaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dangaura Tharu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aheu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thachanadan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thompson",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kochila Tharu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rana Tharu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thakali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tahltan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tahaggart Tamahaq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 104, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tayart Tamajeq",
        #[cfg(feature = "local_names")]
        autonym: Some("Tamajeq"),
    },
    LanguageData {
        code_3: [116, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tidikelt Tamazight",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tifal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tigre",
        #[cfg(feature = "local_names")]
        autonym: Some("ትግረ"),
    },
    LanguageData {
        code_3: [116, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Timugon Murut",
        #[cfg(feature = "local_names")]
        autonym: Some("Timugon"),
    },
    LanguageData {
        code_3: [116, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiene",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tilung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tikar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tillamook",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Timbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tindi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trimuris",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiéfo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 114],
        code_1: Some([116, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Tigrinya",
        #[cfg(feature = "local_names")]
        autonym: Some("ትግርኛ"),
    },
    LanguageData {
        code_3: [116, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Masadiit Itneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tinigua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adasen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiv",
        #[cfg(feature = "local_names")]
        autonym: Some("Tiv"),
    },
    LanguageData {
        code_3: [116, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Tiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiruray",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Hongjin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tajuasohn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunjung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tujia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tjungundji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Laing",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Timucua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonjon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temacine Tamazight",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tjupany",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Tujia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tjurruru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 106, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Djabwurrung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Truká",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buksa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tukudede",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takwane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tukumanféd",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tesaka Malagasy",
        #[cfg(feature = "local_names")]
        autonym: Some("Tesaka Malagasy"),
    },
    LanguageData {
        code_3: [116, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tokelau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takelma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toku-No-Shima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tikopia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsakhur",
        #[cfg(feature = "local_names")]
        autonym: Some("Cʼäxna miz"),
    },
    LanguageData {
        code_3: [116, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takestani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kathoriya Tharu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Necaxa Totonac",
        #[cfg(feature = "local_names")]
        autonym: Some("Totonaca"),
    },
    LanguageData {
        code_3: [116, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mur Pano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teanu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Takua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Tepehuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobelo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yecuatla Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talaud",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Telefol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tofanma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Klingon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlingit",
        #[cfg(feature = "local_names")]
        autonym: Some("Łingít"),
    },
    LanguageData {
        code_3: [116, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talinga-Bwisi",
        #[cfg(feature = "local_names")]
        autonym: Some("Bwisi"),
    },
    LanguageData {
        code_3: [116, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taloki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tetela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tolomako",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talondo'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talodi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Filomena Mata-Coahuitlán Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Loi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talise",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tambotalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sou Nama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tulehu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taliabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khehek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talysh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katbol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Haruai",
        #[cfg(feature = "local_names")]
        autonym: Some("Haruai"),
    },
    LanguageData {
        code_3: [116, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tremembé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toba-Maskoy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ternateño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamashek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tutuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samarokena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Tamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamnim Citak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Thanh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temoq",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumleo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jewish Babylonian Aramaic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tasmate",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tembo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Temuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamanaku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tacana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Tunebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanimuca-Retuarã",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angosturas Tunebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maiani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tandia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwamera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lenakel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Tanna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toromono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Whitesands",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ménik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tenis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tontemboan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tay Khang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangchangya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonsawang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tongwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ten'edn",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toba",
        #[cfg(feature = "local_names")]
        autonym: Some("Qom"),
    },
    LanguageData {
        code_3: [116, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coyutla Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gizrra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gitonga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonga",
        #[cfg(feature = "local_names")]
        autonym: Some("Chitonga, Tonga"),
    },
    LanguageData {
        code_3: [116, 111, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tojolabal",
        #[cfg(feature = "local_names")]
        autonym: Some("Tojolabal"),
    },
    LanguageData {
        code_3: [116, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toki Pona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tolowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tombulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 110],
        code_1: Some([116, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Tonga",
        #[cfg(feature = "local_names")]
        autonym: Some("lea fakatonga"),
    },
    LanguageData {
        code_3: [116, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xicotepec De Juárez Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Papantla Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toposa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Togbo-Vara Banda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Highland Totonac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Taromi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jemez",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Topoiyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 111, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "To",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taupota",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Azoyú Me'phaa",
        #[cfg(feature = "local_names")]
        autonym: Some("Mè’pháà Tsìndíì"),
    },
    LanguageData {
        code_3: [116, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tippera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tarpia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tok Pisin",
        #[cfg(feature = "local_names")]
        autonym: Some("Tok Pisin"),
    },
    LanguageData {
        code_3: [116, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tapieté",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tupinikin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlacoapa Me'phaa",
        #[cfg(feature = "local_names")]
        autonym: Some("Mi’phàà Mí’uíí"),
    },
    LanguageData {
        code_3: [116, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tampulma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tupinambá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Pao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pisaflores Tepehua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tukpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuparí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlachichilco Tepehua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tampuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanapag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tupí",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Acatepec Me'phaa",
        #[cfg(feature = "local_names")]
        autonym: Some("Me’pàà Wí’ììn"),
    },
    LanguageData {
        code_3: [116, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trumai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tinputz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tembé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lehali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turumsa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tenino",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toaripi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tomoip",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torona",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Totonac",
        #[cfg(feature = "local_names")]
        autonym: Some("Xinulajgsípij Totonaco"),
    },
    LanguageData {
        code_3: [116, 113, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Touo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 113, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonkawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tirahi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Terebu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Copala Triqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Tarangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trinidadian Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lishán Didán",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trió",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toram",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Traveller Scottish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tregami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Trinitario",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tarao Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kok Borok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Martín Itunyoso Triqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taushiro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chicahuaxtla Triqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunggare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turoyo",
        #[cfg(feature = "local_names")]
        autonym: Some("Surayt"),
    },
    LanguageData {
        code_3: [116, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sediq",
        #[cfg(feature = "local_names")]
        autonym: Some("Sediq Taroko"),
    },
    LanguageData {
        code_3: [116, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torwali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tringgus-Sembaan Bidayuh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsaangi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsamai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tswa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsakonian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunisian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tausug",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsuvan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsimshian",
        #[cfg(feature = "local_names")]
        autonym: Some("Sm’algyax"),
    },
    LanguageData {
        code_3: [116, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tshangla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tseku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ts'ün-Lao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turkish Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 110],
        code_1: Some([116, 110]),
        #[cfg(feature = "english_names")]
        name_en: "Tswana",
        #[cfg(feature = "local_names")]
        autonym: Some("Setswana"),
    },
    LanguageData {
        code_3: [116, 115, 111],
        code_1: Some([116, 115]),
        #[cfg(feature = "english_names")]
        name_en: "Tsonga",
        #[cfg(feature = "local_names")]
        autonym: Some("Xitsonga"),
    },
    LanguageData {
        code_3: [116, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Toussian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thai Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Akei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taiwan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tondi Songway Kiini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsogo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsishingini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mubami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tebul Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 115, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Purepecha",
        #[cfg(feature = "local_names")]
        autonym: Some("P’urhepecha"),
    },
    LanguageData {
        code_3: [116, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tutelo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tektiteko",
        #[cfg(feature = "local_names")]
        autonym: Some("B’a’aj"),
    },
    LanguageData {
        code_3: [116, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tauade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bwanabwana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuotomb",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tutong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Ta'oih",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tobati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tooro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Totoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Totela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tutchone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Towei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Ta'oih",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tombelala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawallammat Tamajaq",
        #[cfg(feature = "local_names")]
        autonym: Some("Tamajaq"),
    },
    LanguageData {
        code_3: [116, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeastern Thai",
        #[cfg(feature = "local_names")]
        autonym: Some("อ\u{e35}สาน\u{200e} (Isan)"),
    },
    LanguageData {
        code_3: [116, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muslim Tat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Torau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Titan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Long Wat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sikaritai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 116, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wiarumus",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tübatulabal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mutu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuxá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuyuca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Tunebo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taulil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tupuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tugutil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 107],
        code_1: Some([116, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Turkmen",
        #[cfg(feature = "local_names")]
        autonym: Some("Türkmen dili"),
    },
    LanguageData {
        code_3: [116, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumbuka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunica",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tucano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tedaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 114],
        code_1: Some([116, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Turkish",
        #[cfg(feature = "local_names")]
        autonym: Some("Türkçe"),
    },
    LanguageData {
        code_3: [116, 117, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuscarora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tututni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turkana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuxináwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tugen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vaghua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsuvadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Te'un",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeast Ambrym",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuvalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tela-Masbuar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tavoyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tidore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taveta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tutsa Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tunen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sedoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Taivoan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 118, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Timor Pidgin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Twana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Tawbuid",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teshenawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Twents",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tereweng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Dón",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 105],
        code_1: Some([116, 119]),
        #[cfg(feature = "english_names")]
        name_en: "Twi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawang Monpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Twendi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tswapong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tasawaq",
        #[cfg(feature = "local_names")]
        autonym: Some("Tasawaq senni"),
    },
    LanguageData {
        code_3: [116, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Tarahumara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Turiwára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Termanu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tewe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 119, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawoyan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tombonuo",
        #[cfg(feature = "local_names")]
        autonym: Some("Tangar nu Tombonuo"),
    },
    LanguageData {
        code_3: [116, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tokharian B",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsetsaut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Totoli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tangut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thracian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ikpeng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tarjumo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tomini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Tarangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Toto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tartessian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tonsea",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Citak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayapó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tatana",
        #[cfg(feature = "local_names")]
        autonym: Some("Tatana"),
    },
    LanguageData {
        code_3: [116, 120, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanosy Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tauya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kyanga",
        #[cfg(feature = "local_names")]
        autonym: Some("Kyanga"),
    },
    LanguageData {
        code_3: [116, 121, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "O'du",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Tsaayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Do",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thu Lao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kombai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thaypan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tai Daeng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tày Sa Pa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tày Tac",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tuvinian",
        #[cfg(feature = "local_names")]
        autonym: Some("тыва\u{200e} (tyva), тыва дыл\u{200e} (tyva dyl)"),
    },
    LanguageData {
        code_3: [116, 121, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Teke-Tyee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tiyaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tày",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanzanian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tzeltal",
        #[cfg(feature = "local_names")]
        autonym: Some("Bats’il k’op"),
    },
    LanguageData {
        code_3: [116, 122, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tz'utujil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 122, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talossan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [116, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Atlas Tamazight",
        #[cfg(feature = "local_names")]
        autonym: Some("Tamaziɣt n laṭlaṣ"),
    },
    LanguageData {
        code_3: [116, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tugun",
        #[cfg(feature = "local_names")]
        autonym: Some("Tugun"),
    },
    LanguageData {
        code_3: [116, 122, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tzotzil",
        #[cfg(feature = "local_names")]
        autonym: Some("Bats’i k’op"),
    },
    LanguageData {
        code_3: [116, 122, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabriak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uamué",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tairuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ubang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ubi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Buhi'non Bikol",
        #[cfg(feature = "local_names")]
        autonym: Some("Boie’nen, Buhi"),
    },
    LanguageData {
        code_3: [117, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ubir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umbu-Ungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ubykh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Udihe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muduga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Udi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ujir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wuzlam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Udmurt",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uduk",
        #[cfg(feature = "local_names")]
        autonym: Some("Twampa"),
    },
    LanguageData {
        code_3: [117, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kioko",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 102, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ufim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ugaritic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku-Ugbanh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ughele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kubachi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ugandan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ugong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uruguayan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uhami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Damal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 105, 103],
        code_1: Some([117, 103]),
        #[cfg(feature = "english_names")]
        name_en: "Uighur",
        #[cfg(feature = "local_names")]
        autonym: Some("ئۇيغۇرچە"),
    },
    LanguageData {
        code_3: [117, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uisai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iyive",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanjijili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaburi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukuriguma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukhwejo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muak Sa-aak",
        #[cfg(feature = "local_names")]
        autonym: Some("Muak Sa-aak"),
    },
    LanguageData {
        code_3: [117, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukrainian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukpe-Bayobiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 114],
        code_1: Some([117, 107]),
        #[cfg(feature = "english_names")]
        name_en: "Ukrainian",
        #[cfg(feature = "local_names")]
        autonym: Some("українська"),
    },
    LanguageData {
        code_3: [117, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urubú-Kaapor Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukue",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukwuani-Aboh-Ndoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuuk-Yak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Fungwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulukwumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulithian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meriam Mir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ullatan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulumanda'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unserdeutsch",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uma' Lung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ulwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umatilla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umbundu",
        #[cfg(feature = "local_names")]
        autonym: Some("Umbundu"),
    },
    LanguageData {
        code_3: [117, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marrucinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umbindhamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morrobalama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ukit",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makyan Naga",
        #[cfg(feature = "local_names")]
        autonym: Some("Paungnyuan Naga"),
    },
    LanguageData {
        code_3: [117, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umotína",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umpila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umbugarla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pendau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Munsee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Watut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Undetermined",
        #[cfg(feature = "local_names")]
        autonym: Some("und"),
    },
    LanguageData {
        code_3: [117, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uneme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarinyin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Enawené-Nawé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurnai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mundari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unubahe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Munda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unde Kaili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umeda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uripiv-Wala-Rano-Atchin",
        #[cfg(feature = "local_names")]
        autonym: Some("Tirax"),
    },
    LanguageData {
        code_3: [117, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urarina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urubú-Kaapor",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urningangg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 100],
        code_1: Some([117, 114]),
        #[cfg(feature = "english_names")]
        name_en: "Urdu",
        #[cfg(feature = "local_names")]
        autonym: Some("اردو"),
    },
    LanguageData {
        code_3: [117, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uradhi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urigina",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urhobo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urim",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urak Lawoi'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urapmin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uruangnirin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uru-Pa-In",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lehalurup",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uruava",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sop",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urimo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Orya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uru-Eu-Wau-Wau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usarufa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ushojo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usaghade",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uspanteco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "us-Saare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Otank",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ute-Southern Paiute",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ut-Hun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Etulo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Utu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ura",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "U",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Uvean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 118, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 118, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lote",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku-Uwanh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Doko-Uyanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [117, 122, 98],
        code_1: Some([117, 122]),
        #[cfg(feature = "english_names")]
        name_en: "Uzbek",
        #[cfg(feature = "local_names")]
        autonym: Some("o‘zbek"),
    },
    LanguageData {
        code_3: [117, 122, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Uzbek",
        #[cfg(feature = "local_names")]
        autonym: Some("O’zbek"),
    },
    LanguageData {
        code_3: [117, 122, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Uzbek",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vaagri Booli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vafsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vagla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Varhadi-Nagpuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vai",
        #[cfg(feature = "local_names")]
        autonym: Some("ꕙꔤ"),
    },
    LanguageData {
        code_3: [118, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sekele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vehes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vanimo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Valman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vaiphei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Huarijio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vasavi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vanuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Varli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeast Babar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Bontok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Venetian",
        #[cfg(feature = "local_names")]
        autonym: Some("Veneto"),
    },
    LanguageData {
        code_3: [118, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Veddah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Veluws",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vemgo-Mabas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 110],
        code_1: Some([118, 101]),
        #[cfg(feature = "english_names")]
        name_en: "Venda",
        #[cfg(feature = "local_names")]
        autonym: Some("Tshivenḓa"),
    },
    LanguageData {
        code_3: [118, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ventureño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Veps",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mom Jango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vaghri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 103, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vlaamse Gebarentaal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Virgin Islands Creole English",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vidunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 101],
        code_1: Some([118, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Vietnamese",
        #[cfg(feature = "local_names")]
        autonym: Some("Tiếng Việt"),
    },
    LanguageData {
        code_3: [118, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Viemo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vilela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vinza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vishavan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Viti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iduna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kariyarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kujarge",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulisusu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro Nulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kodeoha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korlai Creole Portuguese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tenggarong Kutai Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurrama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koro Zuba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Valpei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vlaams",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Martuyhunira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Barbaram",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Juxtlahuaca Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mudu Koraga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Masela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mainfränkisch",
        #[cfg(feature = "local_names")]
        autonym: Some("Mainfränkisch, Ostfränkisch"),
    },
    LanguageData {
        code_3: [118, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lungalunga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maraghei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ixtayutla Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [118, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa-Shirima",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malgana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mitlatongo Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Jnu’u lavi"),
    },
    LanguageData {
        code_3: [118, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soyaltepec Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("En naxijen"),
    },
    LanguageData {
        code_3: [118, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soyaltepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marenje",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moksela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muluridyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Valley Maidu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tamazola Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayautla Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Enre naxinanda nguifi"),
    },
    LanguageData {
        code_3: [118, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mazatlán Mazatec",
        #[cfg(feature = "local_names")]
        autonym: Some("Ienra naxinandana nnandia"),
    },
    LanguageData {
        code_3: [118, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vinmavis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vunapu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 111, 108],
        code_1: Some([118, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Volapük",
        #[cfg(feature = "local_names")]
        autonym: Some("Volapük"),
    },
    LanguageData {
        code_3: [118, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Voro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Votic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vera'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Võro",
        #[cfg(feature = "local_names")]
        autonym: Some("Võro kiil"),
    },
    LanguageData {
        code_3: [118, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Varisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Burmbar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moldova Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Venezuelan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Valencian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vitou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vumbu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vunjo",
        #[cfg(feature = "local_names")]
        autonym: Some("Kyivunjo"),
    },
    LanguageData {
        code_3: [118, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vute",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [118, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walla Walla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wab",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wasco-Wishram",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wamesa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walser",
        #[cfg(feature = "local_names")]
        autonym: Some("Walser"),
    },
    LanguageData {
        code_3: [119, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wakoná",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wa'ema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Watubela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wares",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waffa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wolaytta",
        #[cfg(feature = "local_names")]
        autonym: Some("ወላይታቱ"),
    },
    LanguageData {
        code_3: [119, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wampanoag",
        #[cfg(feature = "local_names")]
        autonym: Some("Wôpanâak"),
    },
    LanguageData {
        code_3: [119, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wappo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wapishana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wagiman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waray",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Washo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaninuwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waurá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waiwai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Watam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wampur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wabo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waritai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vwanji",
        #[cfg(feature = "local_names")]
        autonym: Some("Bavwanji"),
    },
    LanguageData {
        code_3: [119, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alagwa",
        #[cfg(feature = "local_names")]
        autonym: Some("Alagwa, Alagwaisa"),
    },
    LanguageData {
        code_3: [119, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waigali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wakhi",
        #[cfg(feature = "local_names")]
        autonym: Some("Khik, Khikwar"),
    },
    LanguageData {
        code_3: [119, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warlpiri",
        #[cfg(feature = "local_names")]
        autonym: Some("Warlpiri"),
    },
    LanguageData {
        code_3: [119, 98, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waddar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wagdi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Bengal Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warnman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wajarri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanomámi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 99, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waci Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wandji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadaginam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadjiginy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadikali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wendat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadjigu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadjabangayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wewaw",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wè Western",
        #[cfg(feature = "local_names")]
        autonym: Some("Wèè"),
    },
    LanguageData {
        code_3: [119, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wedau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wergaia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Weh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiunum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Weme Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wemale",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Westphalien",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Weri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cameroon Pidgin",
        #[cfg(feature = "local_names")]
        autonym: Some("Pidgin"),
    },
    LanguageData {
        code_3: [119, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Perai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rawngtu Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 101, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wejewa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 102, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yafi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wagaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wagawaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wangkangurru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wahgi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waigeo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wirangu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 103, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warrgamay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sou Upaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 104, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Wahgi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 104, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wahau Kenyah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wahau Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Toussian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wichita",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik-Epa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik-Keyangan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik Ngathan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik-Me'anha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minidien",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik-Iiyanh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wikalkan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wilawila",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wik-Mungkan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ho-Chunk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wiraféd",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wiru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vitu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wiyot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 106, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 106, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kw'adza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumbaran",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wakde",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalanadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keerray-Woorroong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunduvadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wakawaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wangkayutyuru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwali Comorian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wolane",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunbarlang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Welaun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waioli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wailaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Welsh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 110],
        code_1: Some([119, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Walloon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wolio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wailapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wallisian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wuliwuli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wichí Lhamtés Vejoz",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waling",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wambaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wamas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mamaindé",
        #[cfg(feature = "local_names")]
        autonym: Some("Mamainsahai’gidu"),
    },
    LanguageData {
        code_3: [119, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wambule",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Minyag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waima'a",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wamin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waamwang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wambon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Walmajarri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mwani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Womo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanambre",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wantoat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wandarang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waneci",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanggom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ndzwani Comorian",
        #[cfg(feature = "local_names")]
        autonym: Some("Shindzwani"),
    },
    LanguageData {
        code_3: [119, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanukaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanggamala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wunumara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wano",
        #[cfg(feature = "local_names")]
        autonym: Some("Wano"),
    },
    LanguageData {
        code_3: [119, 110, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Usan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wintu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wanyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuwema",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wè Northern",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wogeo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wolani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woleaian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gambian Wolof",
        #[cfg(feature = "local_names")]
        autonym: Some("Wolof"),
    },
    LanguageData {
        code_3: [119, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wogamusin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Longto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 108],
        code_1: Some([119, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Wolof",
        #[cfg(feature = "local_names")]
        autonym: Some("Wolof"),
    },
    LanguageData {
        code_3: [119, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wongo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manombai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hanga Hundi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wawonii",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Weyto",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waluwarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wiradjuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wariyangga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garrwa",
        #[cfg(feature = "local_names")]
        autonym: Some("Karrwa"),
    },
    LanguageData {
        code_3: [119, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warlmanpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warumungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warnang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Worrorra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waropen",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wardaman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waris",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waruna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gugu Warra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wae Rana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Merwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 114, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waray",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warembori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adilabad Gondi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wusi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waskia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Owenia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wasu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wotapuri-Katarqalai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 116, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Watiwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wathawurrung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Berta",
        #[cfg(feature = "local_names")]
        autonym: Some("Rotana"),
    },
    LanguageData {
        code_3: [119, 116, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Watakataui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mewati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wotu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wikngenchera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wunambal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wudu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wutunhua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Silimo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wumbvu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bungu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wurrugu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wutung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wu Chinese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wuvulu-Aua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wulna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wauyai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 119, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wakabunga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wetamut",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Warrwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Waxianghua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 120, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wardandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wangaaybuwan-Ngiyambaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woiwurrung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wymysorys",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wyandot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wayoró",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [119, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Fijian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Andalusian Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sambe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Adai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aequian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aghwan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaimbé",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ararandewára",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Máku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalmyk",
        #[cfg(feature = "local_names")]
        autonym: Some("хальмг\u{200e} (Xaľmg), хальмг келн\u{200e} (Xaľmg keln)"),
    },
    LanguageData {
        code_3: [120, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǀXam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xamtanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Apalachee",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aquitanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katawixi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kauwera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xavánte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kawaiisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayan Mahakam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lower Burdekin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bactrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bindal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bigambal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bunganditj",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kombio",
        #[cfg(feature = "local_names")]
        autonym: Some("Akwun"),
    },
    LanguageData {
        code_3: [120, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Birrpayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Breton",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenaboi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bolgarian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bibbulman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kambera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kambiwá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batjala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cumbric",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Camunic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Celtiberian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cisalpine Gaulish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chemakum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Armenian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Comecrudo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cotoname",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chorasmian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Carian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Classical Tibetan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Curonian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chuvantsy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coahuilteco",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 99, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cayuse",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Darkinyung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dacian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dharuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Edomite",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwandu",
        #[cfg(feature = "local_names")]
        autonym: Some("Kwandu, Olukwandu"),
    },
    LanguageData {
        code_3: [120, 100, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaitag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 100, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malayic Dayak",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eblan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hdi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "ǁXegwi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kelo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kembayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Epi-Olmec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xerénte",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kesawai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xetá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Keoru-Ahia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 102, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Faliscan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galatian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gbin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gudang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gabrielino-Fernandeño",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Goreng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garingbal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Galindan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dharumbal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Garza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unggumi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guwa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Harami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hunnic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hadrami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khetrani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Khmer",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 111],
        code_1: Some([120, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Xhosa",
        #[cfg(feature = "local_names")]
        autonym: Some("isiXhosa"),
    },
    LanguageData {
        code_3: [120, 104, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hernican",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hattic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hurrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 104, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iberian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xiri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Illyrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xinca",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xiriâna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kisan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Indus Valley Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 105, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xipaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 106, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minjungbal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 106, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jaitmatang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalkoti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Nago",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kho'ini",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mendalam Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kereho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khengkha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kagoro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenyan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kajali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kachok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mainstream Kenyah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kayan River Kayan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kiorr",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kabatei",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koroni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xakriabá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumbewaha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kantosi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kgalagadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kembra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Uma' Lasan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kurtokha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loup B",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lycian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lydian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lemnian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ligurian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liburnian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alanic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loup A",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lepontic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lusitanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuneiform Luwian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elymian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mushungulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbonga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa-Marrevone",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbudum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Median",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mingrelian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mengaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kugu-Muminh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Majera",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ancient Macedonian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malaysian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manado Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Manado"),
    },
    LanguageData {
        code_3: [120, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manichaean Middle Persian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Morerebi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku-Mu'inh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuku-Mangk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Meroitic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moroccan Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matbat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Antankarana Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tsimihety Malagasy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Salawati",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayaguduna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mori Bawah",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ancient North Arabian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanakanabu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Middle Mongolian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuanhua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngarigu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nganakarti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngumbarl",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Kankanay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Anglo-Norman",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngoni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kangri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanashi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Narragansett",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nukunul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyiyaparli",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 110, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kenzi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "O'chi'chi'",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kokoda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Soga",
        #[cfg(feature = "local_names")]
        autonym: Some("Olusoga"),
    },
    LanguageData {
        code_3: [120, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kominimung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xokleng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Komo",
        #[cfg(feature = "local_names")]
        autonym: Some("Tta Komo"),
    },
    LanguageData {
        code_3: [120, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Konkomba",
        #[cfg(feature = "local_names")]
        autonym: Some("Likpakpaanl"),
    },
    LanguageData {
        code_3: [120, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xukurú",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kopar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Korubo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 111, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kowaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pirriya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northeastern Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pecheneg",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Oyster Bay Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liberia Kpelle",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeast Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phrygian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Midlands Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pictish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpalitjanh",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kulina Pano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Port Sorell Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pumpokol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kapinawá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pochutec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puyo-Paekche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mohegan-Pequot",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Parthian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pisidian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punthamara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Punic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northwestern Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southwestern Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Puyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bruny Island Tasmanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 113, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karakhanid",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 113, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qatabanian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krahô",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Karaboro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gundungurra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kreye",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krikati-Timbira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Armazic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Arin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Raetic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aranama-Tamique",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marriammu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sabaean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sambal",
        #[cfg(feature = "local_names")]
        autonym: Some("Sambalì"),
    },
    LanguageData {
        code_3: [120, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Scythian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sidetic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sempan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Shamang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sio",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Subi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "South Slavey",
        #[cfg(feature = "local_names")]
        autonym: Some("Deh Gáh Ghotie Zhatie"),
    },
    LanguageData {
        code_3: [120, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kasem",
        #[cfg(feature = "local_names")]
        autonym: Some("Kasem"),
    },
    LanguageData {
        code_3: [120, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Solano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Silopi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makhuwa-Saka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sherpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Assan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanumá",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sudovian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Saisiyat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alcozauca Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chazumba Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Katcha-Kadugli-Miri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Diuxi-Tilantongo Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tnu’un dau"),
    },
    LanguageData {
        code_3: [120, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ketengban",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Transalpine Gaulish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yitha Yitha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sinicahua Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [120, 116, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Juan Teita Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dañudavi"),
    },
    LanguageData {
        code_3: [120, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tijaltepec Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [120, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magdalena Peñasco Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Tlaxiaco Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Sa’an nda’u, Sa’an savi"),
    },
    LanguageData {
        code_3: [120, 116, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tokharian A",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Miguel Piedras Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tumshuqese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Early Tripuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sindihui Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tacahua Mixtec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cuyamecalco Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [120, 116, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tawandê",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoloxochitl Mixtec",
        #[cfg(feature = "local_names")]
        autonym: Some("Tu’un savi"),
    },
    LanguageData {
        code_3: [120, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alu Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Betta Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umiida",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kunigami",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Jennu Kurumba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngunawal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Umbrian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Unggaranggu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Upper Umpqua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Urartian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuthant",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 117, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kxoe",
        #[cfg(feature = "local_names")]
        autonym: Some("Khwe, Khwedam"),
    },
    LanguageData {
        code_3: [120, 118, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Venetic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 118, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kamviri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 118, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vandalic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 118, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Volscian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 118, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Vestinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwaza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Woccon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wadi Wadi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xwela Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwegu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wajuk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wangkumara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Xwla Gbe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Written Oirat",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kwerba Mamberamo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wotjobaluk",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wemba Wemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 120, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Boro",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 120, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ke'o",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minkin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 120, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koropó",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 120, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tambora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaygir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yandjibara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayi-Yapi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayi-Kulan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yalakalore",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayi-Thakurti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 121, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yorta Yorta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 122, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zhang-Zhung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 122, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zemgalian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [120, 122, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ancient Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaminahua",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuhup",
        #[cfg(feature = "local_names")]
        autonym: Some("Yuhup"),
    },
    LanguageData {
        code_3: [121, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pass Valley Yali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagua",
        #[cfg(feature = "local_names")]
        autonym: Some("Nijyamii, Nijyamïï Nikyejaada"),
    },
    LanguageData {
        code_3: [121, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pumé",
        #[cfg(feature = "local_names")]
        autonym: Some("Pumé"),
    },
    LanguageData {
        code_3: [121, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yámana",
        #[cfg(feature = "local_names")]
        autonym: Some("Háusi Kúta"),
    },
    LanguageData {
        code_3: [121, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yazgulyam",
        #[cfg(feature = "local_names")]
        autonym: Some("Yuzdomi zavég"),
    },
    LanguageData {
        code_3: [121, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagnobi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Banda-Yangere",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakama",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yalunka",
        #[cfg(feature = "local_names")]
        autonym: Some("Jalunga xuwiina’"),
    },
    LanguageData {
        code_3: [121, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayangna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yapese",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaqui",
        #[cfg(feature = "local_names")]
        autonym: Some("Hiak-nooki, Yoeme"),
    },
    LanguageData {
        code_3: [121, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yabarana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nugunu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yambeta",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuwana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangben",
        #[cfg(feature = "local_names")]
        autonym: Some("nuasue"),
    },
    LanguageData {
        code_3: [121, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawalapití",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yauma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Agwagwune",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lokaa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yala",
        #[cfg(feature = "local_names")]
        autonym: Some("Yala"),
    },
    LanguageData {
        code_3: [121, 98, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yemba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Yugur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamphu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hasha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bokha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yukuben",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaben",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yabaâna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yabong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawiyo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 98, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaweyuha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chesu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 99, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lolopo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 99, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yucuna",
        #[cfg(feature = "local_names")]
        autonym: Some("Yucuna"),
    },
    LanguageData {
        code_3: [121, 99, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chepya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 100, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yanda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 100, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Yiddish",
        #[cfg(feature = "local_names")]
        autonym: Some("יידיש\u{200e} (Yiddish)"),
    },
    LanguageData {
        code_3: [121, 100, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangum Dey",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 100, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yidgha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 100, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoidik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ravula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yeniche",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yimas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yeni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yevanic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yela",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tarok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyankpa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yetfa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yerukula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yapunda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 101, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yeyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malyangapa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yiningayi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangum Gel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagomi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Gepo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagaria",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yolŋu Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yugul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 103, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagwoia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 104, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Baha Buyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Iraqi Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 104, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Hlepho Phowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 104, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yan-nhaŋu Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yinggarda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 100],
        code_1: Some([121, 105]),
        #[cfg(feature = "english_names")]
        name_en: "Yiddish",
        #[cfg(feature = "local_names")]
        autonym: Some("יי\u{5b4}דיש"),
    },
    LanguageData {
        code_3: [121, 105, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ache",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wusa Nasu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Yiddish",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yidiny",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yindjibarndi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dongshanba Lalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yindjilandji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yimchungru Naga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Riang Lai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Pholo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miqie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "North Awyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yis",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Lalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Awu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Nisu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Axi Yi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Azhe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Yukaghir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakaikeke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khlula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kua-nsi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yasa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yekora",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kathu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kuamasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 107, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yakoma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaul",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaleba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yelogu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Angguruk Yali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yil",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Limi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Langnian Buyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naluo Yi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yalarnnga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aribwaung",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 108, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nyâlayu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yambes",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Muji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yameo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamongeri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mili",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Moji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Iamalele",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamna",
        #[cfg(feature = "local_names")]
        autonym: Some("Sunum"),
    },
    LanguageData {
        code_3: [121, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangum Mon",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yamap",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qila Muji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malasar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mysian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Northern Muji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muzi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aluo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yandruwandha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lang'e",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yango",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Naukan Yupik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yangulam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yana",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yendang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yansi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 110, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yahuna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yogad",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yonaguni",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yokuts",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yombe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yongkom",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 114],
        code_1: Some([121, 111]),
        #[cfg(feature = "english_names")]
        name_en: "Yoruba",
        #[cfg(feature = "local_names")]
        autonym: Some("Èdè Yorùbá"),
    },
    LanguageData {
        code_3: [121, 111, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yotti",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoron",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 111, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yoy",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Labo Phowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phupha",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phuma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ani Phowa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Alo Phola",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phupa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Phuza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yerakai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yareba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaouré",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nenets",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nhengatu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yirrk-Mel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yerong",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yaroamë",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yarsun",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yarawata",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 114, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yarluyandi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yassic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Samatao",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sonaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yugoslavian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Myanmar Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nisi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Lolopo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sirenik Yupik",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yessan-Mayo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 115, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sanie",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 116, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Talu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tanglang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Thopho",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 116, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yout Wam",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yatay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yucateco",
        #[cfg(feature = "local_names")]
        autonym: Some("Maaya t’aan"),
    },
    LanguageData {
        code_3: [121, 117, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yugambal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuchi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Judeo-Tripolitanian Arabic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yue Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("粵語"),
    },
    LanguageData {
        code_3: [121, 117, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Havasupai-Walapai-Yavapai",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yug",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yurutí",
        #[cfg(feature = "local_names")]
        autonym: Some("Wajiaraye"),
    },
    LanguageData {
        code_3: [121, 117, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karkar-Yuri",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yulu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quechan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bena",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yukpa",
        #[cfg(feature = "local_names")]
        autonym: Some("Yukpa"),
    },
    LanguageData {
        code_3: [121, 117, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuqui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yurok",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yopno",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Yukaghir",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Yugur",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 117, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuracare",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 118, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawa",
        #[cfg(feature = "local_names")]
        autonym: Some("Yawa Unat"),
    },
    LanguageData {
        code_3: [121, 118, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yavitero",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kalou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yinhawangka",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Lalu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawanawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wuding-Luquan Yi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawuru",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xishanba Lalo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Wumeng Nasu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 119, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yawarawarga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mayawali",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yagara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yardliyawarra",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yinwum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yuyu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 120, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yabula Yabula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 121, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yir Yoront",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 121, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yau",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 121, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayizi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 122, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "E'ma Buyang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [121, 122, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zokhuo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sierra de Juárez Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Western Tlacolula Valley Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ocotlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Cajonos Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yareni Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ayoquesco Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaghawa",
        #[cfg(feature = "local_names")]
        autonym: Some("Beria"),
    },
    LanguageData {
        code_3: [122, 97, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zangwal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Isthmus Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("diidxazá"),
    },
    LanguageData {
        code_3: [122, 97, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaramo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zanaki",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zauzou",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Miahuatlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ozolotepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Aloápam Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rincón Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santo Domingo Albarradas Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dihidx Bilyáhab"),
    },
    LanguageData {
        code_3: [122, 97, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabaa Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zangskari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yatzachi Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mitla Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Didxsaj"),
    },
    LanguageData {
        code_3: [122, 97, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xadani Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zayse-Zergulla",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 97, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Balaibalan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Berawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "East Berawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Blissymbols",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Batui",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 98, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "West Berawan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 99, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coatecas Altas Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 99, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Las Delicias Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 99, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Central Hongshuihe Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 100, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Ngazidja Comorian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 101, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zeeuws",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 101, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zenag",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 101, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Eastern Hongshuihe Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 101, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zenaga",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 103, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kinga",
        #[cfg(feature = "local_names")]
        autonym: Some("Bakinga"),
    },
    LanguageData {
        code_3: [122, 103, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guibei Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 103, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Standard Moroccan Tamazight",
        #[cfg(feature = "local_names")]
        autonym: Some("ⵜⴰⵎⴰⵣⵉⵖⵜ"),
    },
    LanguageData {
        code_3: [122, 103, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Minz Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 103, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guibian Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 103, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Magori",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 104, 97],
        code_1: Some([122, 97]),
        #[cfg(feature = "english_names")]
        name_en: "Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 104, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zhaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 104, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Dai Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 104, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zhire",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 104, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Nong Zhuang",
        #[cfg(feature = "local_names")]
        autonym: Some("Tei Nong"),
    },
    LanguageData {
        code_3: [122, 104, 111],
        code_1: Some([122, 104]),
        #[cfg(feature = "english_names")]
        name_en: "Chinese",
        #[cfg(feature = "local_names")]
        autonym: Some("中文"),
    },
    LanguageData {
        code_3: [122, 104, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zhoa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zia",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zimbabwe Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zimakani",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zialo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mesme",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zinza",
        #[cfg(feature = "local_names")]
        autonym: Some("Abazinza"),
    },
    LanguageData {
        code_3: [122, 105, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zigula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 105, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zizilivakan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaimbulawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koibal",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kadu",
        #[cfg(feature = "local_names")]
        autonym: Some("Asak"),
    },
    LanguageData {
        code_3: [122, 107, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Koguryo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khorezmian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Karankawa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kanan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kott",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "São Paulo Kaingáng",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zakhring",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kitan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaurna",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Krevinian",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 107, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Khazar",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 108, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zula",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 108, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liujiang Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 108, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Melayu, ملايو\u{200e} (Melayu)"),
    },
    LanguageData {
        code_3: [122, 108, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lianshan Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 108, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Liuqian Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manda",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zimba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Margany",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maridan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangerr",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mfinu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marti Ke",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Makolkol",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Negeri Sembilan Malay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maridjabin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mandandanyi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Matngala",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Marimanindji",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbangwe",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Molo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mpuono",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mituku",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maranunggu",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbesa",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Maringarr",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Muruwari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbariman-Gudhinma",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Bomitaba",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mariyedi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 109, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mbandja",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 110, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zan Gula",
        #[cfg(feature = "local_names")]
        autonym: Some("More"),
    },
    LanguageData {
        code_3: [122, 110, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zande",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 110, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 110, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Manangkari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 110, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mangas",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Copainalá Zoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chimalapa Zoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zou",
        #[cfg(feature = "local_names")]
        autonym: Some("Zokam"),
    },
    LanguageData {
        code_3: [122, 111, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Asunción Mixtepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tabasco Zoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Rayón Zoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 111, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Francisco León Zoque",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lachiguiri Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yautepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 99],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Choapan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 100],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southeastern Ixtlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Petapa Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 102],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Pedro Quiatoni Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Guevea De Humboldt Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Totomachapan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 105],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa María Quiegolani Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quiavicuzas Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tlacolulita Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lachixío Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dialu"),
    },
    LanguageData {
        code_3: [122, 112, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mixtepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Inés Yatzechi Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Amatlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Dizhze"),
    },
    LanguageData {
        code_3: [122, 112, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "El Alto Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zoogocho Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santiago Xanica Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Coatlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Vicente Coatlán Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yalálag Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 118],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Chichicapan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 119],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaniza Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Baltazar Loxicha Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mazaltepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 112, 122],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Texmelucan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 113, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Qiubei Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kara",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mirgan",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zerenkel",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 111],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Záparo",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zarphatic",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 114, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Mairasi",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 115, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sarasira",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 115, 107],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kaskean",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 115, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zambian Sign Language",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 115, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Standard Malay",
        #[cfg(feature = "local_names")]
        autonym: Some("Bahasa Malaysia"),
    },
    LanguageData {
        code_3: [122, 115, 114],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Southern Rincon Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 115, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Sukurum",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 101],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Elotepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Xanaguía Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Diidz Zë"),
    },
    LanguageData {
        code_3: [122, 116, 108],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Lapaguía-Guivini Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "San Agustín Mixtepec Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Santa Catarina Albarradas Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Loxicha Zapotec",
        #[cfg(feature = "local_names")]
        autonym: Some("Diste"),
    },
    LanguageData {
        code_3: [122, 116, 113],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Quioquitani-Quierí Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 115],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tilquiapan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 116],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tejalapan Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 117],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Güilá Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaachila Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 116, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yatee Zapotec",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 117, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zeem",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 117, 104],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Tokano",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 117, 108],
        code_1: Some([122, 117]),
        #[cfg(feature = "english_names")]
        name_en: "Zulu",
        #[cfg(feature = "local_names")]
        autonym: Some("isiZulu"),
    },
    LanguageData {
        code_3: [122, 117, 109],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Kumzari",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 117, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zuni",
        #[cfg(feature = "local_names")]
        autonym: Some("Shiwi’ma"),
    },
    LanguageData {
        code_3: [122, 117, 121],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zumaya",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 119, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zay",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 120, 120],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "No linguistic content",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 121, 98],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yongbei Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 121, 103],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yang Zhuang",
        #[cfg(feature = "local_names")]
        autonym: Some("壮语徳靖\u{200e} (Deijing Zhuang)"),
    },
    LanguageData {
        code_3: [122, 121, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Youjiang Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 121, 110],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Yongnan Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 121, 112],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zyphe Chin",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 122, 97],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zaza",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
    LanguageData {
        code_3: [122, 122, 106],
        code_1: None,
        #[cfg(feature = "english_names")]
        name_en: "Zuojiang Zhuang",
        #[cfg(feature = "local_names")]
        autonym: None,
    },
];
#[derive(Clone, Copy, Hash, Eq, PartialEq, PartialOrd, Ord)]
pub enum Language {
    #[doc(hidden)]
    Aaa = 0,
    #[doc(hidden)]
    Aab = 1,
    #[doc(hidden)]
    Aac = 2,
    #[doc(hidden)]
    Aad = 3,
    #[doc(hidden)]
    Aae = 4,
    #[doc(hidden)]
    Aaf = 5,
    #[doc(hidden)]
    Aag = 6,
    #[doc(hidden)]
    Aah = 7,
    #[doc(hidden)]
    Aai = 8,
    #[doc(hidden)]
    Aak = 9,
    #[doc(hidden)]
    Aal = 10,
    #[doc(hidden)]
    Aan = 11,
    #[doc(hidden)]
    Aao = 12,
    #[doc(hidden)]
    Aap = 13,
    #[doc(hidden)]
    Aaq = 14,
    #[doc(hidden)]
    Aar = 15,
    #[doc(hidden)]
    Aas = 16,
    #[doc(hidden)]
    Aat = 17,
    #[doc(hidden)]
    Aau = 18,
    #[doc(hidden)]
    Aaw = 19,
    #[doc(hidden)]
    Aax = 20,
    #[doc(hidden)]
    Aaz = 21,
    #[doc(hidden)]
    Aba = 22,
    #[doc(hidden)]
    Abb = 23,
    #[doc(hidden)]
    Abc = 24,
    #[doc(hidden)]
    Abd = 25,
    #[doc(hidden)]
    Abe = 26,
    #[doc(hidden)]
    Abf = 27,
    #[doc(hidden)]
    Abg = 28,
    #[doc(hidden)]
    Abh = 29,
    #[doc(hidden)]
    Abi = 30,
    #[doc(hidden)]
    Abj = 31,
    #[doc(hidden)]
    Abk = 32,
    #[doc(hidden)]
    Abl = 33,
    #[doc(hidden)]
    Abm = 34,
    #[doc(hidden)]
    Abn = 35,
    #[doc(hidden)]
    Abo = 36,
    #[doc(hidden)]
    Abp = 37,
    #[doc(hidden)]
    Abq = 38,
    #[doc(hidden)]
    Abr = 39,
    #[doc(hidden)]
    Abs = 40,
    #[doc(hidden)]
    Abt = 41,
    #[doc(hidden)]
    Abu = 42,
    #[doc(hidden)]
    Abv = 43,
    #[doc(hidden)]
    Abw = 44,
    #[doc(hidden)]
    Abx = 45,
    #[doc(hidden)]
    Aby = 46,
    #[doc(hidden)]
    Abz = 47,
    #[doc(hidden)]
    Aca = 48,
    #[doc(hidden)]
    Acb = 49,
    #[doc(hidden)]
    Acd = 50,
    #[doc(hidden)]
    Ace = 51,
    #[doc(hidden)]
    Acf = 52,
    #[doc(hidden)]
    Ach = 53,
    #[doc(hidden)]
    Aci = 54,
    #[doc(hidden)]
    Ack = 55,
    #[doc(hidden)]
    Acl = 56,
    #[doc(hidden)]
    Acm = 57,
    #[doc(hidden)]
    Acn = 58,
    #[doc(hidden)]
    Acp = 59,
    #[doc(hidden)]
    Acq = 60,
    #[doc(hidden)]
    Acr = 61,
    #[doc(hidden)]
    Acs = 62,
    #[doc(hidden)]
    Act = 63,
    #[doc(hidden)]
    Acu = 64,
    #[doc(hidden)]
    Acv = 65,
    #[doc(hidden)]
    Acw = 66,
    #[doc(hidden)]
    Acx = 67,
    #[doc(hidden)]
    Acy = 68,
    #[doc(hidden)]
    Acz = 69,
    #[doc(hidden)]
    Ada = 70,
    #[doc(hidden)]
    Adb = 71,
    #[doc(hidden)]
    Add = 72,
    #[doc(hidden)]
    Ade = 73,
    #[doc(hidden)]
    Adf = 74,
    #[doc(hidden)]
    Adg = 75,
    #[doc(hidden)]
    Adh = 76,
    #[doc(hidden)]
    Adi = 77,
    #[doc(hidden)]
    Adj = 78,
    #[doc(hidden)]
    Adl = 79,
    #[doc(hidden)]
    Adn = 80,
    #[doc(hidden)]
    Ado = 81,
    #[doc(hidden)]
    Adq = 82,
    #[doc(hidden)]
    Adr = 83,
    #[doc(hidden)]
    Ads = 84,
    #[doc(hidden)]
    Adt = 85,
    #[doc(hidden)]
    Adu = 86,
    #[doc(hidden)]
    Adw = 87,
    #[doc(hidden)]
    Adx = 88,
    #[doc(hidden)]
    Ady = 89,
    #[doc(hidden)]
    Adz = 90,
    #[doc(hidden)]
    Aea = 91,
    #[doc(hidden)]
    Aeb = 92,
    #[doc(hidden)]
    Aec = 93,
    #[doc(hidden)]
    Aed = 94,
    #[doc(hidden)]
    Aee = 95,
    #[doc(hidden)]
    Aek = 96,
    #[doc(hidden)]
    Ael = 97,
    #[doc(hidden)]
    Aem = 98,
    #[doc(hidden)]
    Aen = 99,
    #[doc(hidden)]
    Aeq = 100,
    #[doc(hidden)]
    Aer = 101,
    #[doc(hidden)]
    Aes = 102,
    #[doc(hidden)]
    Aeu = 103,
    #[doc(hidden)]
    Aew = 104,
    #[doc(hidden)]
    Aey = 105,
    #[doc(hidden)]
    Aez = 106,
    #[doc(hidden)]
    Afb = 107,
    #[doc(hidden)]
    Afd = 108,
    #[doc(hidden)]
    Afe = 109,
    #[doc(hidden)]
    Afg = 110,
    #[doc(hidden)]
    Afh = 111,
    #[doc(hidden)]
    Afi = 112,
    #[doc(hidden)]
    Afk = 113,
    #[doc(hidden)]
    Afn = 114,
    #[doc(hidden)]
    Afo = 115,
    #[doc(hidden)]
    Afp = 116,
    #[doc(hidden)]
    Afr = 117,
    #[doc(hidden)]
    Afs = 118,
    #[doc(hidden)]
    Aft = 119,
    #[doc(hidden)]
    Afu = 120,
    #[doc(hidden)]
    Afz = 121,
    #[doc(hidden)]
    Aga = 122,
    #[doc(hidden)]
    Agb = 123,
    #[doc(hidden)]
    Agc = 124,
    #[doc(hidden)]
    Agd = 125,
    #[doc(hidden)]
    Age = 126,
    #[doc(hidden)]
    Agf = 127,
    #[doc(hidden)]
    Agg = 128,
    #[doc(hidden)]
    Agh = 129,
    #[doc(hidden)]
    Agi = 130,
    #[doc(hidden)]
    Agj = 131,
    #[doc(hidden)]
    Agk = 132,
    #[doc(hidden)]
    Agl = 133,
    #[doc(hidden)]
    Agm = 134,
    #[doc(hidden)]
    Agn = 135,
    #[doc(hidden)]
    Ago = 136,
    #[doc(hidden)]
    Agq = 137,
    #[doc(hidden)]
    Agr = 138,
    #[doc(hidden)]
    Ags = 139,
    #[doc(hidden)]
    Agt = 140,
    #[doc(hidden)]
    Agu = 141,
    #[doc(hidden)]
    Agv = 142,
    #[doc(hidden)]
    Agw = 143,
    #[doc(hidden)]
    Agx = 144,
    #[doc(hidden)]
    Agy = 145,
    #[doc(hidden)]
    Agz = 146,
    #[doc(hidden)]
    Aha = 147,
    #[doc(hidden)]
    Ahb = 148,
    #[doc(hidden)]
    Ahg = 149,
    #[doc(hidden)]
    Ahh = 150,
    #[doc(hidden)]
    Ahi = 151,
    #[doc(hidden)]
    Ahk = 152,
    #[doc(hidden)]
    Ahl = 153,
    #[doc(hidden)]
    Ahm = 154,
    #[doc(hidden)]
    Ahn = 155,
    #[doc(hidden)]
    Aho = 156,
    #[doc(hidden)]
    Ahp = 157,
    #[doc(hidden)]
    Ahr = 158,
    #[doc(hidden)]
    Ahs = 159,
    #[doc(hidden)]
    Aht = 160,
    #[doc(hidden)]
    Aia = 161,
    #[doc(hidden)]
    Aib = 162,
    #[doc(hidden)]
    Aic = 163,
    #[doc(hidden)]
    Aid = 164,
    #[doc(hidden)]
    Aie = 165,
    #[doc(hidden)]
    Aif = 166,
    #[doc(hidden)]
    Aig = 167,
    #[doc(hidden)]
    Aih = 168,
    #[doc(hidden)]
    Aii = 169,
    #[doc(hidden)]
    Aij = 170,
    #[doc(hidden)]
    Aik = 171,
    #[doc(hidden)]
    Ail = 172,
    #[doc(hidden)]
    Aim = 173,
    #[doc(hidden)]
    Ain = 174,
    #[doc(hidden)]
    Aio = 175,
    #[doc(hidden)]
    Aip = 176,
    #[doc(hidden)]
    Aiq = 177,
    #[doc(hidden)]
    Air = 178,
    #[doc(hidden)]
    Ait = 179,
    #[doc(hidden)]
    Aiw = 180,
    #[doc(hidden)]
    Aix = 181,
    #[doc(hidden)]
    Aiy = 182,
    #[doc(hidden)]
    Aja = 183,
    #[doc(hidden)]
    Ajg = 184,
    #[doc(hidden)]
    Aji = 185,
    #[doc(hidden)]
    Ajn = 186,
    #[doc(hidden)]
    Ajp = 187,
    #[doc(hidden)]
    Ajs = 188,
    #[doc(hidden)]
    Aju = 189,
    #[doc(hidden)]
    Ajw = 190,
    #[doc(hidden)]
    Ajz = 191,
    #[doc(hidden)]
    Aka = 192,
    #[doc(hidden)]
    Akb = 193,
    #[doc(hidden)]
    Akc = 194,
    #[doc(hidden)]
    Akd = 195,
    #[doc(hidden)]
    Ake = 196,
    #[doc(hidden)]
    Akf = 197,
    #[doc(hidden)]
    Akg = 198,
    #[doc(hidden)]
    Akh = 199,
    #[doc(hidden)]
    Aki = 200,
    #[doc(hidden)]
    Akj = 201,
    #[doc(hidden)]
    Akk = 202,
    #[doc(hidden)]
    Akl = 203,
    #[doc(hidden)]
    Akm = 204,
    #[doc(hidden)]
    Ako = 205,
    #[doc(hidden)]
    Akp = 206,
    #[doc(hidden)]
    Akq = 207,
    #[doc(hidden)]
    Akr = 208,
    #[doc(hidden)]
    Aks = 209,
    #[doc(hidden)]
    Akt = 210,
    #[doc(hidden)]
    Aku = 211,
    #[doc(hidden)]
    Akv = 212,
    #[doc(hidden)]
    Akw = 213,
    #[doc(hidden)]
    Akx = 214,
    #[doc(hidden)]
    Aky = 215,
    #[doc(hidden)]
    Akz = 216,
    #[doc(hidden)]
    Ala = 217,
    #[doc(hidden)]
    Alc = 218,
    #[doc(hidden)]
    Ald = 219,
    #[doc(hidden)]
    Ale = 220,
    #[doc(hidden)]
    Alf = 221,
    #[doc(hidden)]
    Alh = 222,
    #[doc(hidden)]
    Ali = 223,
    #[doc(hidden)]
    Alj = 224,
    #[doc(hidden)]
    Alk = 225,
    #[doc(hidden)]
    All = 226,
    #[doc(hidden)]
    Alm = 227,
    #[doc(hidden)]
    Aln = 228,
    #[doc(hidden)]
    Alo = 229,
    #[doc(hidden)]
    Alp = 230,
    #[doc(hidden)]
    Alq = 231,
    #[doc(hidden)]
    Alr = 232,
    #[doc(hidden)]
    Als = 233,
    #[doc(hidden)]
    Alt = 234,
    #[doc(hidden)]
    Alu = 235,
    #[doc(hidden)]
    Alw = 236,
    #[doc(hidden)]
    Alx = 237,
    #[doc(hidden)]
    Aly = 238,
    #[doc(hidden)]
    Alz = 239,
    #[doc(hidden)]
    Ama = 240,
    #[doc(hidden)]
    Amb = 241,
    #[doc(hidden)]
    Amc = 242,
    #[doc(hidden)]
    Ame = 243,
    #[doc(hidden)]
    Amf = 244,
    #[doc(hidden)]
    Amg = 245,
    #[doc(hidden)]
    Amh = 246,
    #[doc(hidden)]
    Ami = 247,
    #[doc(hidden)]
    Amj = 248,
    #[doc(hidden)]
    Amk = 249,
    #[doc(hidden)]
    Aml = 250,
    #[doc(hidden)]
    Amm = 251,
    #[doc(hidden)]
    Amn = 252,
    #[doc(hidden)]
    Amo = 253,
    #[doc(hidden)]
    Amp = 254,
    #[doc(hidden)]
    Amq = 255,
    #[doc(hidden)]
    Amr = 256,
    #[doc(hidden)]
    Ams = 257,
    #[doc(hidden)]
    Amt = 258,
    #[doc(hidden)]
    Amu = 259,
    #[doc(hidden)]
    Amv = 260,
    #[doc(hidden)]
    Amw = 261,
    #[doc(hidden)]
    Amx = 262,
    #[doc(hidden)]
    Amy = 263,
    #[doc(hidden)]
    Amz = 264,
    #[doc(hidden)]
    Ana = 265,
    #[doc(hidden)]
    Anb = 266,
    #[doc(hidden)]
    Anc = 267,
    #[doc(hidden)]
    And = 268,
    #[doc(hidden)]
    Ane = 269,
    #[doc(hidden)]
    Anf = 270,
    #[doc(hidden)]
    Ang = 271,
    #[doc(hidden)]
    Anh = 272,
    #[doc(hidden)]
    Ani = 273,
    #[doc(hidden)]
    Anj = 274,
    #[doc(hidden)]
    Ank = 275,
    #[doc(hidden)]
    Anl = 276,
    #[doc(hidden)]
    Anm = 277,
    #[doc(hidden)]
    Ann = 278,
    #[doc(hidden)]
    Ano = 279,
    #[doc(hidden)]
    Anp = 280,
    #[doc(hidden)]
    Anq = 281,
    #[doc(hidden)]
    Anr = 282,
    #[doc(hidden)]
    Ans = 283,
    #[doc(hidden)]
    Ant = 284,
    #[doc(hidden)]
    Anu = 285,
    #[doc(hidden)]
    Anv = 286,
    #[doc(hidden)]
    Anw = 287,
    #[doc(hidden)]
    Anx = 288,
    #[doc(hidden)]
    Any = 289,
    #[doc(hidden)]
    Anz = 290,
    #[doc(hidden)]
    Aoa = 291,
    #[doc(hidden)]
    Aob = 292,
    #[doc(hidden)]
    Aoc = 293,
    #[doc(hidden)]
    Aod = 294,
    #[doc(hidden)]
    Aoe = 295,
    #[doc(hidden)]
    Aof = 296,
    #[doc(hidden)]
    Aog = 297,
    #[doc(hidden)]
    Aoi = 298,
    #[doc(hidden)]
    Aoj = 299,
    #[doc(hidden)]
    Aok = 300,
    #[doc(hidden)]
    Aol = 301,
    #[doc(hidden)]
    Aom = 302,
    #[doc(hidden)]
    Aon = 303,
    #[doc(hidden)]
    Aor = 304,
    #[doc(hidden)]
    Aos = 305,
    #[doc(hidden)]
    Aot = 306,
    #[doc(hidden)]
    Aou = 307,
    #[doc(hidden)]
    Aox = 308,
    #[doc(hidden)]
    Aoz = 309,
    #[doc(hidden)]
    Apb = 310,
    #[doc(hidden)]
    Apc = 311,
    #[doc(hidden)]
    Apd = 312,
    #[doc(hidden)]
    Ape = 313,
    #[doc(hidden)]
    Apf = 314,
    #[doc(hidden)]
    Apg = 315,
    #[doc(hidden)]
    Aph = 316,
    #[doc(hidden)]
    Api = 317,
    #[doc(hidden)]
    Apj = 318,
    #[doc(hidden)]
    Apk = 319,
    #[doc(hidden)]
    Apl = 320,
    #[doc(hidden)]
    Apm = 321,
    #[doc(hidden)]
    Apn = 322,
    #[doc(hidden)]
    Apo = 323,
    #[doc(hidden)]
    App = 324,
    #[doc(hidden)]
    Apq = 325,
    #[doc(hidden)]
    Apr = 326,
    #[doc(hidden)]
    Aps = 327,
    #[doc(hidden)]
    Apt = 328,
    #[doc(hidden)]
    Apu = 329,
    #[doc(hidden)]
    Apv = 330,
    #[doc(hidden)]
    Apw = 331,
    #[doc(hidden)]
    Apx = 332,
    #[doc(hidden)]
    Apy = 333,
    #[doc(hidden)]
    Apz = 334,
    #[doc(hidden)]
    Aqc = 335,
    #[doc(hidden)]
    Aqd = 336,
    #[doc(hidden)]
    Aqg = 337,
    #[doc(hidden)]
    Aqk = 338,
    #[doc(hidden)]
    Aqm = 339,
    #[doc(hidden)]
    Aqn = 340,
    #[doc(hidden)]
    Aqp = 341,
    #[doc(hidden)]
    Aqr = 342,
    #[doc(hidden)]
    Aqt = 343,
    #[doc(hidden)]
    Aqz = 344,
    #[doc(hidden)]
    Ara = 345,
    #[doc(hidden)]
    Arb = 346,
    #[doc(hidden)]
    Arc = 347,
    #[doc(hidden)]
    Ard = 348,
    #[doc(hidden)]
    Are = 349,
    #[doc(hidden)]
    Arg = 350,
    #[doc(hidden)]
    Arh = 351,
    #[doc(hidden)]
    Ari = 352,
    #[doc(hidden)]
    Arj = 353,
    #[doc(hidden)]
    Ark = 354,
    #[doc(hidden)]
    Arl = 355,
    #[doc(hidden)]
    Arn = 356,
    #[doc(hidden)]
    Aro = 357,
    #[doc(hidden)]
    Arp = 358,
    #[doc(hidden)]
    Arq = 359,
    #[doc(hidden)]
    Arr = 360,
    #[doc(hidden)]
    Ars = 361,
    #[doc(hidden)]
    Aru = 362,
    #[doc(hidden)]
    Arv = 363,
    #[doc(hidden)]
    Arw = 364,
    #[doc(hidden)]
    Arx = 365,
    #[doc(hidden)]
    Ary = 366,
    #[doc(hidden)]
    Arz = 367,
    #[doc(hidden)]
    Asa = 368,
    #[doc(hidden)]
    Asb = 369,
    #[doc(hidden)]
    Asc = 370,
    #[doc(hidden)]
    Ase = 371,
    #[doc(hidden)]
    Asf = 372,
    #[doc(hidden)]
    Asg = 373,
    #[doc(hidden)]
    Ash = 374,
    #[doc(hidden)]
    Asi = 375,
    #[doc(hidden)]
    Asj = 376,
    #[doc(hidden)]
    Ask = 377,
    #[doc(hidden)]
    Asl = 378,
    #[doc(hidden)]
    Asm = 379,
    #[doc(hidden)]
    Asn = 380,
    #[doc(hidden)]
    Aso = 381,
    #[doc(hidden)]
    Asp = 382,
    #[doc(hidden)]
    Asq = 383,
    #[doc(hidden)]
    Asr = 384,
    #[doc(hidden)]
    Ass = 385,
    #[doc(hidden)]
    Ast = 386,
    #[doc(hidden)]
    Asu = 387,
    #[doc(hidden)]
    Asv = 388,
    #[doc(hidden)]
    Asw = 389,
    #[doc(hidden)]
    Asx = 390,
    #[doc(hidden)]
    Asy = 391,
    #[doc(hidden)]
    Asz = 392,
    #[doc(hidden)]
    Ata = 393,
    #[doc(hidden)]
    Atb = 394,
    #[doc(hidden)]
    Atc = 395,
    #[doc(hidden)]
    Atd = 396,
    #[doc(hidden)]
    Ate = 397,
    #[doc(hidden)]
    Atg = 398,
    #[doc(hidden)]
    Ati = 399,
    #[doc(hidden)]
    Atj = 400,
    #[doc(hidden)]
    Atk = 401,
    #[doc(hidden)]
    Atl = 402,
    #[doc(hidden)]
    Atm = 403,
    #[doc(hidden)]
    Atn = 404,
    #[doc(hidden)]
    Ato = 405,
    #[doc(hidden)]
    Atp = 406,
    #[doc(hidden)]
    Atq = 407,
    #[doc(hidden)]
    Atr = 408,
    #[doc(hidden)]
    Ats = 409,
    #[doc(hidden)]
    Att = 410,
    #[doc(hidden)]
    Atu = 411,
    #[doc(hidden)]
    Atv = 412,
    #[doc(hidden)]
    Atw = 413,
    #[doc(hidden)]
    Atx = 414,
    #[doc(hidden)]
    Aty = 415,
    #[doc(hidden)]
    Atz = 416,
    #[doc(hidden)]
    Aua = 417,
    #[doc(hidden)]
    Aub = 418,
    #[doc(hidden)]
    Auc = 419,
    #[doc(hidden)]
    Aud = 420,
    #[doc(hidden)]
    Aug = 421,
    #[doc(hidden)]
    Auh = 422,
    #[doc(hidden)]
    Aui = 423,
    #[doc(hidden)]
    Auj = 424,
    #[doc(hidden)]
    Auk = 425,
    #[doc(hidden)]
    Aul = 426,
    #[doc(hidden)]
    Aum = 427,
    #[doc(hidden)]
    Aun = 428,
    #[doc(hidden)]
    Auo = 429,
    #[doc(hidden)]
    Aup = 430,
    #[doc(hidden)]
    Auq = 431,
    #[doc(hidden)]
    Aur = 432,
    #[doc(hidden)]
    Aut = 433,
    #[doc(hidden)]
    Auu = 434,
    #[doc(hidden)]
    Auw = 435,
    #[doc(hidden)]
    Aux = 436,
    #[doc(hidden)]
    Auy = 437,
    #[doc(hidden)]
    Auz = 438,
    #[doc(hidden)]
    Ava = 439,
    #[doc(hidden)]
    Avb = 440,
    #[doc(hidden)]
    Avd = 441,
    #[doc(hidden)]
    Ave = 442,
    #[doc(hidden)]
    Avi = 443,
    #[doc(hidden)]
    Avk = 444,
    #[doc(hidden)]
    Avl = 445,
    #[doc(hidden)]
    Avm = 446,
    #[doc(hidden)]
    Avn = 447,
    #[doc(hidden)]
    Avo = 448,
    #[doc(hidden)]
    Avs = 449,
    #[doc(hidden)]
    Avt = 450,
    #[doc(hidden)]
    Avu = 451,
    #[doc(hidden)]
    Avv = 452,
    #[doc(hidden)]
    Awa = 453,
    #[doc(hidden)]
    Awb = 454,
    #[doc(hidden)]
    Awc = 455,
    #[doc(hidden)]
    Awe = 456,
    #[doc(hidden)]
    Awg = 457,
    #[doc(hidden)]
    Awh = 458,
    #[doc(hidden)]
    Awi = 459,
    #[doc(hidden)]
    Awk = 460,
    #[doc(hidden)]
    Awm = 461,
    #[doc(hidden)]
    Awn = 462,
    #[doc(hidden)]
    Awo = 463,
    #[doc(hidden)]
    Awr = 464,
    #[doc(hidden)]
    Aws = 465,
    #[doc(hidden)]
    Awt = 466,
    #[doc(hidden)]
    Awu = 467,
    #[doc(hidden)]
    Awv = 468,
    #[doc(hidden)]
    Aww = 469,
    #[doc(hidden)]
    Awx = 470,
    #[doc(hidden)]
    Awy = 471,
    #[doc(hidden)]
    Axb = 472,
    #[doc(hidden)]
    Axe = 473,
    #[doc(hidden)]
    Axg = 474,
    #[doc(hidden)]
    Axk = 475,
    #[doc(hidden)]
    Axl = 476,
    #[doc(hidden)]
    Axm = 477,
    #[doc(hidden)]
    Axx = 478,
    #[doc(hidden)]
    Aya = 479,
    #[doc(hidden)]
    Ayb = 480,
    #[doc(hidden)]
    Ayc = 481,
    #[doc(hidden)]
    Ayd = 482,
    #[doc(hidden)]
    Aye = 483,
    #[doc(hidden)]
    Ayg = 484,
    #[doc(hidden)]
    Ayh = 485,
    #[doc(hidden)]
    Ayi = 486,
    #[doc(hidden)]
    Ayk = 487,
    #[doc(hidden)]
    Ayl = 488,
    #[doc(hidden)]
    Aym = 489,
    #[doc(hidden)]
    Ayn = 490,
    #[doc(hidden)]
    Ayo = 491,
    #[doc(hidden)]
    Ayp = 492,
    #[doc(hidden)]
    Ayq = 493,
    #[doc(hidden)]
    Ayr = 494,
    #[doc(hidden)]
    Ays = 495,
    #[doc(hidden)]
    Ayt = 496,
    #[doc(hidden)]
    Ayu = 497,
    #[doc(hidden)]
    Ayz = 498,
    #[doc(hidden)]
    Aza = 499,
    #[doc(hidden)]
    Azb = 500,
    #[doc(hidden)]
    Azd = 501,
    #[doc(hidden)]
    Aze = 502,
    #[doc(hidden)]
    Azg = 503,
    #[doc(hidden)]
    Azj = 504,
    #[doc(hidden)]
    Azm = 505,
    #[doc(hidden)]
    Azn = 506,
    #[doc(hidden)]
    Azo = 507,
    #[doc(hidden)]
    Azt = 508,
    #[doc(hidden)]
    Azz = 509,
    #[doc(hidden)]
    Baa = 510,
    #[doc(hidden)]
    Bab = 511,
    #[doc(hidden)]
    Bac = 512,
    #[doc(hidden)]
    Bae = 513,
    #[doc(hidden)]
    Baf = 514,
    #[doc(hidden)]
    Bag = 515,
    #[doc(hidden)]
    Bah = 516,
    #[doc(hidden)]
    Baj = 517,
    #[doc(hidden)]
    Bak = 518,
    #[doc(hidden)]
    Bal = 519,
    #[doc(hidden)]
    Bam = 520,
    #[doc(hidden)]
    Ban = 521,
    #[doc(hidden)]
    Bao = 522,
    #[doc(hidden)]
    Bap = 523,
    #[doc(hidden)]
    Bar = 524,
    #[doc(hidden)]
    Bas = 525,
    #[doc(hidden)]
    Bau = 526,
    #[doc(hidden)]
    Bav = 527,
    #[doc(hidden)]
    Baw = 528,
    #[doc(hidden)]
    Bax = 529,
    #[doc(hidden)]
    Bay = 530,
    #[doc(hidden)]
    Bba = 531,
    #[doc(hidden)]
    Bbb = 532,
    #[doc(hidden)]
    Bbc = 533,
    #[doc(hidden)]
    Bbd = 534,
    #[doc(hidden)]
    Bbe = 535,
    #[doc(hidden)]
    Bbf = 536,
    #[doc(hidden)]
    Bbg = 537,
    #[doc(hidden)]
    Bbh = 538,
    #[doc(hidden)]
    Bbi = 539,
    #[doc(hidden)]
    Bbj = 540,
    #[doc(hidden)]
    Bbk = 541,
    #[doc(hidden)]
    Bbl = 542,
    #[doc(hidden)]
    Bbm = 543,
    #[doc(hidden)]
    Bbn = 544,
    #[doc(hidden)]
    Bbo = 545,
    #[doc(hidden)]
    Bbp = 546,
    #[doc(hidden)]
    Bbq = 547,
    #[doc(hidden)]
    Bbr = 548,
    #[doc(hidden)]
    Bbs = 549,
    #[doc(hidden)]
    Bbt = 550,
    #[doc(hidden)]
    Bbu = 551,
    #[doc(hidden)]
    Bbv = 552,
    #[doc(hidden)]
    Bbw = 553,
    #[doc(hidden)]
    Bbx = 554,
    #[doc(hidden)]
    Bby = 555,
    #[doc(hidden)]
    Bca = 556,
    #[doc(hidden)]
    Bcb = 557,
    #[doc(hidden)]
    Bcc = 558,
    #[doc(hidden)]
    Bcd = 559,
    #[doc(hidden)]
    Bce = 560,
    #[doc(hidden)]
    Bcf = 561,
    #[doc(hidden)]
    Bcg = 562,
    #[doc(hidden)]
    Bch = 563,
    #[doc(hidden)]
    Bci = 564,
    #[doc(hidden)]
    Bcj = 565,
    #[doc(hidden)]
    Bck = 566,
    #[doc(hidden)]
    Bcl = 567,
    #[doc(hidden)]
    Bcm = 568,
    #[doc(hidden)]
    Bcn = 569,
    #[doc(hidden)]
    Bco = 570,
    #[doc(hidden)]
    Bcp = 571,
    #[doc(hidden)]
    Bcq = 572,
    #[doc(hidden)]
    Bcr = 573,
    #[doc(hidden)]
    Bcs = 574,
    #[doc(hidden)]
    Bct = 575,
    #[doc(hidden)]
    Bcu = 576,
    #[doc(hidden)]
    Bcv = 577,
    #[doc(hidden)]
    Bcw = 578,
    #[doc(hidden)]
    Bcy = 579,
    #[doc(hidden)]
    Bcz = 580,
    #[doc(hidden)]
    Bda = 581,
    #[doc(hidden)]
    Bdb = 582,
    #[doc(hidden)]
    Bdc = 583,
    #[doc(hidden)]
    Bdd = 584,
    #[doc(hidden)]
    Bde = 585,
    #[doc(hidden)]
    Bdf = 586,
    #[doc(hidden)]
    Bdg = 587,
    #[doc(hidden)]
    Bdh = 588,
    #[doc(hidden)]
    Bdi = 589,
    #[doc(hidden)]
    Bdj = 590,
    #[doc(hidden)]
    Bdk = 591,
    #[doc(hidden)]
    Bdl = 592,
    #[doc(hidden)]
    Bdm = 593,
    #[doc(hidden)]
    Bdn = 594,
    #[doc(hidden)]
    Bdo = 595,
    #[doc(hidden)]
    Bdp = 596,
    #[doc(hidden)]
    Bdq = 597,
    #[doc(hidden)]
    Bdr = 598,
    #[doc(hidden)]
    Bds = 599,
    #[doc(hidden)]
    Bdt = 600,
    #[doc(hidden)]
    Bdu = 601,
    #[doc(hidden)]
    Bdv = 602,
    #[doc(hidden)]
    Bdw = 603,
    #[doc(hidden)]
    Bdx = 604,
    #[doc(hidden)]
    Bdy = 605,
    #[doc(hidden)]
    Bdz = 606,
    #[doc(hidden)]
    Bea = 607,
    #[doc(hidden)]
    Beb = 608,
    #[doc(hidden)]
    Bec = 609,
    #[doc(hidden)]
    Bed = 610,
    #[doc(hidden)]
    Bee = 611,
    #[doc(hidden)]
    Bef = 612,
    #[doc(hidden)]
    Beg = 613,
    #[doc(hidden)]
    Beh = 614,
    #[doc(hidden)]
    Bei = 615,
    #[doc(hidden)]
    Bej = 616,
    #[doc(hidden)]
    Bek = 617,
    #[doc(hidden)]
    Bel = 618,
    #[doc(hidden)]
    Bem = 619,
    #[doc(hidden)]
    Ben = 620,
    #[doc(hidden)]
    Beo = 621,
    #[doc(hidden)]
    Bep = 622,
    #[doc(hidden)]
    Beq = 623,
    #[doc(hidden)]
    Bes = 624,
    #[doc(hidden)]
    Bet = 625,
    #[doc(hidden)]
    Beu = 626,
    #[doc(hidden)]
    Bev = 627,
    #[doc(hidden)]
    Bew = 628,
    #[doc(hidden)]
    Bex = 629,
    #[doc(hidden)]
    Bey = 630,
    #[doc(hidden)]
    Bez = 631,
    #[doc(hidden)]
    Bfa = 632,
    #[doc(hidden)]
    Bfb = 633,
    #[doc(hidden)]
    Bfc = 634,
    #[doc(hidden)]
    Bfd = 635,
    #[doc(hidden)]
    Bfe = 636,
    #[doc(hidden)]
    Bff = 637,
    #[doc(hidden)]
    Bfg = 638,
    #[doc(hidden)]
    Bfh = 639,
    #[doc(hidden)]
    Bfi = 640,
    #[doc(hidden)]
    Bfj = 641,
    #[doc(hidden)]
    Bfk = 642,
    #[doc(hidden)]
    Bfl = 643,
    #[doc(hidden)]
    Bfm = 644,
    #[doc(hidden)]
    Bfn = 645,
    #[doc(hidden)]
    Bfo = 646,
    #[doc(hidden)]
    Bfp = 647,
    #[doc(hidden)]
    Bfq = 648,
    #[doc(hidden)]
    Bfr = 649,
    #[doc(hidden)]
    Bfs = 650,
    #[doc(hidden)]
    Bft = 651,
    #[doc(hidden)]
    Bfu = 652,
    #[doc(hidden)]
    Bfw = 653,
    #[doc(hidden)]
    Bfx = 654,
    #[doc(hidden)]
    Bfy = 655,
    #[doc(hidden)]
    Bfz = 656,
    #[doc(hidden)]
    Bga = 657,
    #[doc(hidden)]
    Bgb = 658,
    #[doc(hidden)]
    Bgc = 659,
    #[doc(hidden)]
    Bgd = 660,
    #[doc(hidden)]
    Bge = 661,
    #[doc(hidden)]
    Bgf = 662,
    #[doc(hidden)]
    Bgg = 663,
    #[doc(hidden)]
    Bgi = 664,
    #[doc(hidden)]
    Bgj = 665,
    #[doc(hidden)]
    Bgk = 666,
    #[doc(hidden)]
    Bgl = 667,
    #[doc(hidden)]
    Bgn = 668,
    #[doc(hidden)]
    Bgo = 669,
    #[doc(hidden)]
    Bgp = 670,
    #[doc(hidden)]
    Bgq = 671,
    #[doc(hidden)]
    Bgr = 672,
    #[doc(hidden)]
    Bgs = 673,
    #[doc(hidden)]
    Bgt = 674,
    #[doc(hidden)]
    Bgu = 675,
    #[doc(hidden)]
    Bgv = 676,
    #[doc(hidden)]
    Bgw = 677,
    #[doc(hidden)]
    Bgx = 678,
    #[doc(hidden)]
    Bgy = 679,
    #[doc(hidden)]
    Bgz = 680,
    #[doc(hidden)]
    Bha = 681,
    #[doc(hidden)]
    Bhb = 682,
    #[doc(hidden)]
    Bhc = 683,
    #[doc(hidden)]
    Bhd = 684,
    #[doc(hidden)]
    Bhe = 685,
    #[doc(hidden)]
    Bhf = 686,
    #[doc(hidden)]
    Bhg = 687,
    #[doc(hidden)]
    Bhh = 688,
    #[doc(hidden)]
    Bhi = 689,
    #[doc(hidden)]
    Bhj = 690,
    #[doc(hidden)]
    Bhl = 691,
    #[doc(hidden)]
    Bhm = 692,
    #[doc(hidden)]
    Bhn = 693,
    #[doc(hidden)]
    Bho = 694,
    #[doc(hidden)]
    Bhp = 695,
    #[doc(hidden)]
    Bhq = 696,
    #[doc(hidden)]
    Bhr = 697,
    #[doc(hidden)]
    Bhs = 698,
    #[doc(hidden)]
    Bht = 699,
    #[doc(hidden)]
    Bhu = 700,
    #[doc(hidden)]
    Bhv = 701,
    #[doc(hidden)]
    Bhw = 702,
    #[doc(hidden)]
    Bhx = 703,
    #[doc(hidden)]
    Bhy = 704,
    #[doc(hidden)]
    Bhz = 705,
    #[doc(hidden)]
    Bia = 706,
    #[doc(hidden)]
    Bib = 707,
    #[doc(hidden)]
    Bid = 708,
    #[doc(hidden)]
    Bie = 709,
    #[doc(hidden)]
    Bif = 710,
    #[doc(hidden)]
    Big = 711,
    #[doc(hidden)]
    Bik = 712,
    #[doc(hidden)]
    Bil = 713,
    #[doc(hidden)]
    Bim = 714,
    #[doc(hidden)]
    Bin = 715,
    #[doc(hidden)]
    Bio = 716,
    #[doc(hidden)]
    Bip = 717,
    #[doc(hidden)]
    Biq = 718,
    #[doc(hidden)]
    Bir = 719,
    #[doc(hidden)]
    Bis = 720,
    #[doc(hidden)]
    Bit = 721,
    #[doc(hidden)]
    Biu = 722,
    #[doc(hidden)]
    Biv = 723,
    #[doc(hidden)]
    Biw = 724,
    #[doc(hidden)]
    Bix = 725,
    #[doc(hidden)]
    Biy = 726,
    #[doc(hidden)]
    Biz = 727,
    #[doc(hidden)]
    Bja = 728,
    #[doc(hidden)]
    Bjb = 729,
    #[doc(hidden)]
    Bjc = 730,
    #[doc(hidden)]
    Bje = 731,
    #[doc(hidden)]
    Bjf = 732,
    #[doc(hidden)]
    Bjg = 733,
    #[doc(hidden)]
    Bjh = 734,
    #[doc(hidden)]
    Bji = 735,
    #[doc(hidden)]
    Bjj = 736,
    #[doc(hidden)]
    Bjk = 737,
    #[doc(hidden)]
    Bjl = 738,
    #[doc(hidden)]
    Bjm = 739,
    #[doc(hidden)]
    Bjn = 740,
    #[doc(hidden)]
    Bjo = 741,
    #[doc(hidden)]
    Bjp = 742,
    #[doc(hidden)]
    Bjr = 743,
    #[doc(hidden)]
    Bjs = 744,
    #[doc(hidden)]
    Bjt = 745,
    #[doc(hidden)]
    Bju = 746,
    #[doc(hidden)]
    Bjv = 747,
    #[doc(hidden)]
    Bjw = 748,
    #[doc(hidden)]
    Bjx = 749,
    #[doc(hidden)]
    Bjy = 750,
    #[doc(hidden)]
    Bjz = 751,
    #[doc(hidden)]
    Bka = 752,
    #[doc(hidden)]
    Bkc = 753,
    #[doc(hidden)]
    Bkd = 754,
    #[doc(hidden)]
    Bkf = 755,
    #[doc(hidden)]
    Bkg = 756,
    #[doc(hidden)]
    Bkh = 757,
    #[doc(hidden)]
    Bki = 758,
    #[doc(hidden)]
    Bkj = 759,
    #[doc(hidden)]
    Bkk = 760,
    #[doc(hidden)]
    Bkl = 761,
    #[doc(hidden)]
    Bkm = 762,
    #[doc(hidden)]
    Bkn = 763,
    #[doc(hidden)]
    Bko = 764,
    #[doc(hidden)]
    Bkp = 765,
    #[doc(hidden)]
    Bkq = 766,
    #[doc(hidden)]
    Bkr = 767,
    #[doc(hidden)]
    Bks = 768,
    #[doc(hidden)]
    Bkt = 769,
    #[doc(hidden)]
    Bku = 770,
    #[doc(hidden)]
    Bkv = 771,
    #[doc(hidden)]
    Bkw = 772,
    #[doc(hidden)]
    Bkx = 773,
    #[doc(hidden)]
    Bky = 774,
    #[doc(hidden)]
    Bkz = 775,
    #[doc(hidden)]
    Bla = 776,
    #[doc(hidden)]
    Blb = 777,
    #[doc(hidden)]
    Blc = 778,
    #[doc(hidden)]
    Bld = 779,
    #[doc(hidden)]
    Ble = 780,
    #[doc(hidden)]
    Blf = 781,
    #[doc(hidden)]
    Blh = 782,
    #[doc(hidden)]
    Bli = 783,
    #[doc(hidden)]
    Blj = 784,
    #[doc(hidden)]
    Blk = 785,
    #[doc(hidden)]
    Bll = 786,
    #[doc(hidden)]
    Blm = 787,
    #[doc(hidden)]
    Bln = 788,
    #[doc(hidden)]
    Blo = 789,
    #[doc(hidden)]
    Blp = 790,
    #[doc(hidden)]
    Blq = 791,
    #[doc(hidden)]
    Blr = 792,
    #[doc(hidden)]
    Bls = 793,
    #[doc(hidden)]
    Blt = 794,
    #[doc(hidden)]
    Blv = 795,
    #[doc(hidden)]
    Blw = 796,
    #[doc(hidden)]
    Blx = 797,
    #[doc(hidden)]
    Bly = 798,
    #[doc(hidden)]
    Blz = 799,
    #[doc(hidden)]
    Bma = 800,
    #[doc(hidden)]
    Bmb = 801,
    #[doc(hidden)]
    Bmc = 802,
    #[doc(hidden)]
    Bmd = 803,
    #[doc(hidden)]
    Bme = 804,
    #[doc(hidden)]
    Bmf = 805,
    #[doc(hidden)]
    Bmg = 806,
    #[doc(hidden)]
    Bmh = 807,
    #[doc(hidden)]
    Bmi = 808,
    #[doc(hidden)]
    Bmj = 809,
    #[doc(hidden)]
    Bmk = 810,
    #[doc(hidden)]
    Bml = 811,
    #[doc(hidden)]
    Bmm = 812,
    #[doc(hidden)]
    Bmn = 813,
    #[doc(hidden)]
    Bmo = 814,
    #[doc(hidden)]
    Bmp = 815,
    #[doc(hidden)]
    Bmq = 816,
    #[doc(hidden)]
    Bmr = 817,
    #[doc(hidden)]
    Bms = 818,
    #[doc(hidden)]
    Bmt = 819,
    #[doc(hidden)]
    Bmu = 820,
    #[doc(hidden)]
    Bmv = 821,
    #[doc(hidden)]
    Bmw = 822,
    #[doc(hidden)]
    Bmx = 823,
    #[doc(hidden)]
    Bmz = 824,
    #[doc(hidden)]
    Bna = 825,
    #[doc(hidden)]
    Bnb = 826,
    #[doc(hidden)]
    Bnc = 827,
    #[doc(hidden)]
    Bnd = 828,
    #[doc(hidden)]
    Bne = 829,
    #[doc(hidden)]
    Bnf = 830,
    #[doc(hidden)]
    Bng = 831,
    #[doc(hidden)]
    Bni = 832,
    #[doc(hidden)]
    Bnj = 833,
    #[doc(hidden)]
    Bnk = 834,
    #[doc(hidden)]
    Bnl = 835,
    #[doc(hidden)]
    Bnm = 836,
    #[doc(hidden)]
    Bnn = 837,
    #[doc(hidden)]
    Bno = 838,
    #[doc(hidden)]
    Bnp = 839,
    #[doc(hidden)]
    Bnq = 840,
    #[doc(hidden)]
    Bnr = 841,
    #[doc(hidden)]
    Bns = 842,
    #[doc(hidden)]
    Bnu = 843,
    #[doc(hidden)]
    Bnv = 844,
    #[doc(hidden)]
    Bnw = 845,
    #[doc(hidden)]
    Bnx = 846,
    #[doc(hidden)]
    Bny = 847,
    #[doc(hidden)]
    Bnz = 848,
    #[doc(hidden)]
    Boa = 849,
    #[doc(hidden)]
    Bob = 850,
    #[doc(hidden)]
    Bod = 851,
    #[doc(hidden)]
    Boe = 852,
    #[doc(hidden)]
    Bof = 853,
    #[doc(hidden)]
    Bog = 854,
    #[doc(hidden)]
    Boh = 855,
    #[doc(hidden)]
    Boi = 856,
    #[doc(hidden)]
    Boj = 857,
    #[doc(hidden)]
    Bok = 858,
    #[doc(hidden)]
    Bol = 859,
    #[doc(hidden)]
    Bom = 860,
    #[doc(hidden)]
    Bon = 861,
    #[doc(hidden)]
    Boo = 862,
    #[doc(hidden)]
    Bop = 863,
    #[doc(hidden)]
    Boq = 864,
    #[doc(hidden)]
    Bor = 865,
    #[doc(hidden)]
    Bos = 866,
    #[doc(hidden)]
    Bot = 867,
    #[doc(hidden)]
    Bou = 868,
    #[doc(hidden)]
    Bov = 869,
    #[doc(hidden)]
    Bow = 870,
    #[doc(hidden)]
    Box = 871,
    #[doc(hidden)]
    Boy = 872,
    #[doc(hidden)]
    Boz = 873,
    #[doc(hidden)]
    Bpa = 874,
    #[doc(hidden)]
    Bpc = 875,
    #[doc(hidden)]
    Bpd = 876,
    #[doc(hidden)]
    Bpe = 877,
    #[doc(hidden)]
    Bpg = 878,
    #[doc(hidden)]
    Bph = 879,
    #[doc(hidden)]
    Bpi = 880,
    #[doc(hidden)]
    Bpj = 881,
    #[doc(hidden)]
    Bpk = 882,
    #[doc(hidden)]
    Bpl = 883,
    #[doc(hidden)]
    Bpm = 884,
    #[doc(hidden)]
    Bpn = 885,
    #[doc(hidden)]
    Bpo = 886,
    #[doc(hidden)]
    Bpp = 887,
    #[doc(hidden)]
    Bpq = 888,
    #[doc(hidden)]
    Bpr = 889,
    #[doc(hidden)]
    Bps = 890,
    #[doc(hidden)]
    Bpt = 891,
    #[doc(hidden)]
    Bpu = 892,
    #[doc(hidden)]
    Bpv = 893,
    #[doc(hidden)]
    Bpw = 894,
    #[doc(hidden)]
    Bpx = 895,
    #[doc(hidden)]
    Bpy = 896,
    #[doc(hidden)]
    Bpz = 897,
    #[doc(hidden)]
    Bqa = 898,
    #[doc(hidden)]
    Bqb = 899,
    #[doc(hidden)]
    Bqc = 900,
    #[doc(hidden)]
    Bqd = 901,
    #[doc(hidden)]
    Bqf = 902,
    #[doc(hidden)]
    Bqg = 903,
    #[doc(hidden)]
    Bqh = 904,
    #[doc(hidden)]
    Bqi = 905,
    #[doc(hidden)]
    Bqj = 906,
    #[doc(hidden)]
    Bqk = 907,
    #[doc(hidden)]
    Bql = 908,
    #[doc(hidden)]
    Bqm = 909,
    #[doc(hidden)]
    Bqn = 910,
    #[doc(hidden)]
    Bqo = 911,
    #[doc(hidden)]
    Bqp = 912,
    #[doc(hidden)]
    Bqq = 913,
    #[doc(hidden)]
    Bqr = 914,
    #[doc(hidden)]
    Bqs = 915,
    #[doc(hidden)]
    Bqt = 916,
    #[doc(hidden)]
    Bqu = 917,
    #[doc(hidden)]
    Bqv = 918,
    #[doc(hidden)]
    Bqw = 919,
    #[doc(hidden)]
    Bqx = 920,
    #[doc(hidden)]
    Bqy = 921,
    #[doc(hidden)]
    Bqz = 922,
    #[doc(hidden)]
    Bra = 923,
    #[doc(hidden)]
    Brb = 924,
    #[doc(hidden)]
    Brc = 925,
    #[doc(hidden)]
    Brd = 926,
    #[doc(hidden)]
    Bre = 927,
    #[doc(hidden)]
    Brf = 928,
    #[doc(hidden)]
    Brg = 929,
    #[doc(hidden)]
    Brh = 930,
    #[doc(hidden)]
    Bri = 931,
    #[doc(hidden)]
    Brj = 932,
    #[doc(hidden)]
    Brk = 933,
    #[doc(hidden)]
    Brl = 934,
    #[doc(hidden)]
    Brm = 935,
    #[doc(hidden)]
    Brn = 936,
    #[doc(hidden)]
    Bro = 937,
    #[doc(hidden)]
    Brp = 938,
    #[doc(hidden)]
    Brq = 939,
    #[doc(hidden)]
    Brr = 940,
    #[doc(hidden)]
    Brs = 941,
    #[doc(hidden)]
    Brt = 942,
    #[doc(hidden)]
    Bru = 943,
    #[doc(hidden)]
    Brv = 944,
    #[doc(hidden)]
    Brw = 945,
    #[doc(hidden)]
    Brx = 946,
    #[doc(hidden)]
    Bry = 947,
    #[doc(hidden)]
    Brz = 948,
    #[doc(hidden)]
    Bsa = 949,
    #[doc(hidden)]
    Bsb = 950,
    #[doc(hidden)]
    Bsc = 951,
    #[doc(hidden)]
    Bse = 952,
    #[doc(hidden)]
    Bsf = 953,
    #[doc(hidden)]
    Bsg = 954,
    #[doc(hidden)]
    Bsh = 955,
    #[doc(hidden)]
    Bsi = 956,
    #[doc(hidden)]
    Bsj = 957,
    #[doc(hidden)]
    Bsk = 958,
    #[doc(hidden)]
    Bsl = 959,
    #[doc(hidden)]
    Bsm = 960,
    #[doc(hidden)]
    Bsn = 961,
    #[doc(hidden)]
    Bso = 962,
    #[doc(hidden)]
    Bsp = 963,
    #[doc(hidden)]
    Bsq = 964,
    #[doc(hidden)]
    Bsr = 965,
    #[doc(hidden)]
    Bss = 966,
    #[doc(hidden)]
    Bst = 967,
    #[doc(hidden)]
    Bsu = 968,
    #[doc(hidden)]
    Bsv = 969,
    #[doc(hidden)]
    Bsw = 970,
    #[doc(hidden)]
    Bsx = 971,
    #[doc(hidden)]
    Bsy = 972,
    #[doc(hidden)]
    Bta = 973,
    #[doc(hidden)]
    Btc = 974,
    #[doc(hidden)]
    Btd = 975,
    #[doc(hidden)]
    Bte = 976,
    #[doc(hidden)]
    Btf = 977,
    #[doc(hidden)]
    Btg = 978,
    #[doc(hidden)]
    Bth = 979,
    #[doc(hidden)]
    Bti = 980,
    #[doc(hidden)]
    Btj = 981,
    #[doc(hidden)]
    Btm = 982,
    #[doc(hidden)]
    Btn = 983,
    #[doc(hidden)]
    Bto = 984,
    #[doc(hidden)]
    Btp = 985,
    #[doc(hidden)]
    Btq = 986,
    #[doc(hidden)]
    Btr = 987,
    #[doc(hidden)]
    Bts = 988,
    #[doc(hidden)]
    Btt = 989,
    #[doc(hidden)]
    Btu = 990,
    #[doc(hidden)]
    Btv = 991,
    #[doc(hidden)]
    Btw = 992,
    #[doc(hidden)]
    Btx = 993,
    #[doc(hidden)]
    Bty = 994,
    #[doc(hidden)]
    Btz = 995,
    #[doc(hidden)]
    Bua = 996,
    #[doc(hidden)]
    Bub = 997,
    #[doc(hidden)]
    Buc = 998,
    #[doc(hidden)]
    Bud = 999,
    #[doc(hidden)]
    Bue = 1000,
    #[doc(hidden)]
    Buf = 1001,
    #[doc(hidden)]
    Bug = 1002,
    #[doc(hidden)]
    Buh = 1003,
    #[doc(hidden)]
    Bui = 1004,
    #[doc(hidden)]
    Buj = 1005,
    #[doc(hidden)]
    Buk = 1006,
    #[doc(hidden)]
    Bul = 1007,
    #[doc(hidden)]
    Bum = 1008,
    #[doc(hidden)]
    Bun = 1009,
    #[doc(hidden)]
    Buo = 1010,
    #[doc(hidden)]
    Bup = 1011,
    #[doc(hidden)]
    Buq = 1012,
    #[doc(hidden)]
    Bus = 1013,
    #[doc(hidden)]
    But = 1014,
    #[doc(hidden)]
    Buu = 1015,
    #[doc(hidden)]
    Buv = 1016,
    #[doc(hidden)]
    Buw = 1017,
    #[doc(hidden)]
    Bux = 1018,
    #[doc(hidden)]
    Buy = 1019,
    #[doc(hidden)]
    Buz = 1020,
    #[doc(hidden)]
    Bva = 1021,
    #[doc(hidden)]
    Bvb = 1022,
    #[doc(hidden)]
    Bvc = 1023,
    #[doc(hidden)]
    Bvd = 1024,
    #[doc(hidden)]
    Bve = 1025,
    #[doc(hidden)]
    Bvf = 1026,
    #[doc(hidden)]
    Bvg = 1027,
    #[doc(hidden)]
    Bvh = 1028,
    #[doc(hidden)]
    Bvi = 1029,
    #[doc(hidden)]
    Bvj = 1030,
    #[doc(hidden)]
    Bvk = 1031,
    #[doc(hidden)]
    Bvl = 1032,
    #[doc(hidden)]
    Bvm = 1033,
    #[doc(hidden)]
    Bvn = 1034,
    #[doc(hidden)]
    Bvo = 1035,
    #[doc(hidden)]
    Bvp = 1036,
    #[doc(hidden)]
    Bvq = 1037,
    #[doc(hidden)]
    Bvr = 1038,
    #[doc(hidden)]
    Bvt = 1039,
    #[doc(hidden)]
    Bvu = 1040,
    #[doc(hidden)]
    Bvv = 1041,
    #[doc(hidden)]
    Bvw = 1042,
    #[doc(hidden)]
    Bvx = 1043,
    #[doc(hidden)]
    Bvy = 1044,
    #[doc(hidden)]
    Bvz = 1045,
    #[doc(hidden)]
    Bwa = 1046,
    #[doc(hidden)]
    Bwb = 1047,
    #[doc(hidden)]
    Bwc = 1048,
    #[doc(hidden)]
    Bwd = 1049,
    #[doc(hidden)]
    Bwe = 1050,
    #[doc(hidden)]
    Bwf = 1051,
    #[doc(hidden)]
    Bwg = 1052,
    #[doc(hidden)]
    Bwh = 1053,
    #[doc(hidden)]
    Bwi = 1054,
    #[doc(hidden)]
    Bwj = 1055,
    #[doc(hidden)]
    Bwk = 1056,
    #[doc(hidden)]
    Bwl = 1057,
    #[doc(hidden)]
    Bwm = 1058,
    #[doc(hidden)]
    Bwn = 1059,
    #[doc(hidden)]
    Bwo = 1060,
    #[doc(hidden)]
    Bwp = 1061,
    #[doc(hidden)]
    Bwq = 1062,
    #[doc(hidden)]
    Bwr = 1063,
    #[doc(hidden)]
    Bws = 1064,
    #[doc(hidden)]
    Bwt = 1065,
    #[doc(hidden)]
    Bwu = 1066,
    #[doc(hidden)]
    Bww = 1067,
    #[doc(hidden)]
    Bwx = 1068,
    #[doc(hidden)]
    Bwy = 1069,
    #[doc(hidden)]
    Bwz = 1070,
    #[doc(hidden)]
    Bxa = 1071,
    #[doc(hidden)]
    Bxb = 1072,
    #[doc(hidden)]
    Bxc = 1073,
    #[doc(hidden)]
    Bxd = 1074,
    #[doc(hidden)]
    Bxe = 1075,
    #[doc(hidden)]
    Bxf = 1076,
    #[doc(hidden)]
    Bxg = 1077,
    #[doc(hidden)]
    Bxh = 1078,
    #[doc(hidden)]
    Bxi = 1079,
    #[doc(hidden)]
    Bxj = 1080,
    #[doc(hidden)]
    Bxk = 1081,
    #[doc(hidden)]
    Bxl = 1082,
    #[doc(hidden)]
    Bxm = 1083,
    #[doc(hidden)]
    Bxn = 1084,
    #[doc(hidden)]
    Bxo = 1085,
    #[doc(hidden)]
    Bxp = 1086,
    #[doc(hidden)]
    Bxq = 1087,
    #[doc(hidden)]
    Bxr = 1088,
    #[doc(hidden)]
    Bxs = 1089,
    #[doc(hidden)]
    Bxu = 1090,
    #[doc(hidden)]
    Bxv = 1091,
    #[doc(hidden)]
    Bxw = 1092,
    #[doc(hidden)]
    Bxz = 1093,
    #[doc(hidden)]
    Bya = 1094,
    #[doc(hidden)]
    Byb = 1095,
    #[doc(hidden)]
    Byc = 1096,
    #[doc(hidden)]
    Byd = 1097,
    #[doc(hidden)]
    Bye = 1098,
    #[doc(hidden)]
    Byf = 1099,
    #[doc(hidden)]
    Byg = 1100,
    #[doc(hidden)]
    Byh = 1101,
    #[doc(hidden)]
    Byi = 1102,
    #[doc(hidden)]
    Byj = 1103,
    #[doc(hidden)]
    Byk = 1104,
    #[doc(hidden)]
    Byl = 1105,
    #[doc(hidden)]
    Bym = 1106,
    #[doc(hidden)]
    Byn = 1107,
    #[doc(hidden)]
    Byo = 1108,
    #[doc(hidden)]
    Byp = 1109,
    #[doc(hidden)]
    Byq = 1110,
    #[doc(hidden)]
    Byr = 1111,
    #[doc(hidden)]
    Bys = 1112,
    #[doc(hidden)]
    Byt = 1113,
    #[doc(hidden)]
    Byv = 1114,
    #[doc(hidden)]
    Byw = 1115,
    #[doc(hidden)]
    Byx = 1116,
    #[doc(hidden)]
    Byz = 1117,
    #[doc(hidden)]
    Bza = 1118,
    #[doc(hidden)]
    Bzb = 1119,
    #[doc(hidden)]
    Bzc = 1120,
    #[doc(hidden)]
    Bzd = 1121,
    #[doc(hidden)]
    Bze = 1122,
    #[doc(hidden)]
    Bzf = 1123,
    #[doc(hidden)]
    Bzg = 1124,
    #[doc(hidden)]
    Bzh = 1125,
    #[doc(hidden)]
    Bzi = 1126,
    #[doc(hidden)]
    Bzj = 1127,
    #[doc(hidden)]
    Bzk = 1128,
    #[doc(hidden)]
    Bzl = 1129,
    #[doc(hidden)]
    Bzm = 1130,
    #[doc(hidden)]
    Bzn = 1131,
    #[doc(hidden)]
    Bzo = 1132,
    #[doc(hidden)]
    Bzp = 1133,
    #[doc(hidden)]
    Bzq = 1134,
    #[doc(hidden)]
    Bzr = 1135,
    #[doc(hidden)]
    Bzs = 1136,
    #[doc(hidden)]
    Bzt = 1137,
    #[doc(hidden)]
    Bzu = 1138,
    #[doc(hidden)]
    Bzv = 1139,
    #[doc(hidden)]
    Bzw = 1140,
    #[doc(hidden)]
    Bzx = 1141,
    #[doc(hidden)]
    Bzy = 1142,
    #[doc(hidden)]
    Bzz = 1143,
    #[doc(hidden)]
    Caa = 1144,
    #[doc(hidden)]
    Cab = 1145,
    #[doc(hidden)]
    Cac = 1146,
    #[doc(hidden)]
    Cad = 1147,
    #[doc(hidden)]
    Cae = 1148,
    #[doc(hidden)]
    Caf = 1149,
    #[doc(hidden)]
    Cag = 1150,
    #[doc(hidden)]
    Cah = 1151,
    #[doc(hidden)]
    Caj = 1152,
    #[doc(hidden)]
    Cak = 1153,
    #[doc(hidden)]
    Cal = 1154,
    #[doc(hidden)]
    Cam = 1155,
    #[doc(hidden)]
    Can = 1156,
    #[doc(hidden)]
    Cao = 1157,
    #[doc(hidden)]
    Cap = 1158,
    #[doc(hidden)]
    Caq = 1159,
    #[doc(hidden)]
    Car = 1160,
    #[doc(hidden)]
    Cas = 1161,
    #[doc(hidden)]
    Cat = 1162,
    #[doc(hidden)]
    Cav = 1163,
    #[doc(hidden)]
    Caw = 1164,
    #[doc(hidden)]
    Cax = 1165,
    #[doc(hidden)]
    Cay = 1166,
    #[doc(hidden)]
    Caz = 1167,
    #[doc(hidden)]
    Cbb = 1168,
    #[doc(hidden)]
    Cbc = 1169,
    #[doc(hidden)]
    Cbd = 1170,
    #[doc(hidden)]
    Cbg = 1171,
    #[doc(hidden)]
    Cbi = 1172,
    #[doc(hidden)]
    Cbj = 1173,
    #[doc(hidden)]
    Cbk = 1174,
    #[doc(hidden)]
    Cbl = 1175,
    #[doc(hidden)]
    Cbn = 1176,
    #[doc(hidden)]
    Cbo = 1177,
    #[doc(hidden)]
    Cbq = 1178,
    #[doc(hidden)]
    Cbr = 1179,
    #[doc(hidden)]
    Cbs = 1180,
    #[doc(hidden)]
    Cbt = 1181,
    #[doc(hidden)]
    Cbu = 1182,
    #[doc(hidden)]
    Cbv = 1183,
    #[doc(hidden)]
    Cbw = 1184,
    #[doc(hidden)]
    Cby = 1185,
    #[doc(hidden)]
    Ccc = 1186,
    #[doc(hidden)]
    Ccd = 1187,
    #[doc(hidden)]
    Cce = 1188,
    #[doc(hidden)]
    Ccg = 1189,
    #[doc(hidden)]
    Cch = 1190,
    #[doc(hidden)]
    Ccj = 1191,
    #[doc(hidden)]
    Ccl = 1192,
    #[doc(hidden)]
    Ccm = 1193,
    #[doc(hidden)]
    Cco = 1194,
    #[doc(hidden)]
    Ccp = 1195,
    #[doc(hidden)]
    Ccr = 1196,
    #[doc(hidden)]
    Cda = 1197,
    #[doc(hidden)]
    Cde = 1198,
    #[doc(hidden)]
    Cdf = 1199,
    #[doc(hidden)]
    Cdh = 1200,
    #[doc(hidden)]
    Cdi = 1201,
    #[doc(hidden)]
    Cdj = 1202,
    #[doc(hidden)]
    Cdm = 1203,
    #[doc(hidden)]
    Cdn = 1204,
    #[doc(hidden)]
    Cdo = 1205,
    #[doc(hidden)]
    Cdr = 1206,
    #[doc(hidden)]
    Cds = 1207,
    #[doc(hidden)]
    Cdy = 1208,
    #[doc(hidden)]
    Cdz = 1209,
    #[doc(hidden)]
    Cea = 1210,
    #[doc(hidden)]
    Ceb = 1211,
    #[doc(hidden)]
    Ceg = 1212,
    #[doc(hidden)]
    Cek = 1213,
    #[doc(hidden)]
    Cen = 1214,
    #[doc(hidden)]
    Ces = 1215,
    #[doc(hidden)]
    Cet = 1216,
    #[doc(hidden)]
    Cey = 1217,
    #[doc(hidden)]
    Cfa = 1218,
    #[doc(hidden)]
    Cfd = 1219,
    #[doc(hidden)]
    Cfg = 1220,
    #[doc(hidden)]
    Cfm = 1221,
    #[doc(hidden)]
    Cga = 1222,
    #[doc(hidden)]
    Cgc = 1223,
    #[doc(hidden)]
    Cgg = 1224,
    #[doc(hidden)]
    Cgk = 1225,
    #[doc(hidden)]
    Cha = 1226,
    #[doc(hidden)]
    Chb = 1227,
    #[doc(hidden)]
    Chc = 1228,
    #[doc(hidden)]
    Chd = 1229,
    #[doc(hidden)]
    Che = 1230,
    #[doc(hidden)]
    Chf = 1231,
    #[doc(hidden)]
    Chg = 1232,
    #[doc(hidden)]
    Chh = 1233,
    #[doc(hidden)]
    Chj = 1234,
    #[doc(hidden)]
    Chk = 1235,
    #[doc(hidden)]
    Chl = 1236,
    #[doc(hidden)]
    Chm = 1237,
    #[doc(hidden)]
    Chn = 1238,
    #[doc(hidden)]
    Cho = 1239,
    #[doc(hidden)]
    Chp = 1240,
    #[doc(hidden)]
    Chq = 1241,
    #[doc(hidden)]
    Chr = 1242,
    #[doc(hidden)]
    Cht = 1243,
    #[doc(hidden)]
    Chu = 1244,
    #[doc(hidden)]
    Chv = 1245,
    #[doc(hidden)]
    Chw = 1246,
    #[doc(hidden)]
    Chx = 1247,
    #[doc(hidden)]
    Chy = 1248,
    #[doc(hidden)]
    Chz = 1249,
    #[doc(hidden)]
    Cia = 1250,
    #[doc(hidden)]
    Cib = 1251,
    #[doc(hidden)]
    Cic = 1252,
    #[doc(hidden)]
    Cid = 1253,
    #[doc(hidden)]
    Cie = 1254,
    #[doc(hidden)]
    Cih = 1255,
    #[doc(hidden)]
    Cik = 1256,
    #[doc(hidden)]
    Cim = 1257,
    #[doc(hidden)]
    Cin = 1258,
    #[doc(hidden)]
    Cip = 1259,
    #[doc(hidden)]
    Cir = 1260,
    #[doc(hidden)]
    Ciw = 1261,
    #[doc(hidden)]
    Ciy = 1262,
    #[doc(hidden)]
    Cja = 1263,
    #[doc(hidden)]
    Cje = 1264,
    #[doc(hidden)]
    Cjh = 1265,
    #[doc(hidden)]
    Cji = 1266,
    #[doc(hidden)]
    Cjk = 1267,
    #[doc(hidden)]
    Cjm = 1268,
    #[doc(hidden)]
    Cjn = 1269,
    #[doc(hidden)]
    Cjo = 1270,
    #[doc(hidden)]
    Cjp = 1271,
    #[doc(hidden)]
    Cjs = 1272,
    #[doc(hidden)]
    Cjv = 1273,
    #[doc(hidden)]
    Cjy = 1274,
    #[doc(hidden)]
    Ckb = 1275,
    #[doc(hidden)]
    Ckh = 1276,
    #[doc(hidden)]
    Ckl = 1277,
    #[doc(hidden)]
    Ckm = 1278,
    #[doc(hidden)]
    Ckn = 1279,
    #[doc(hidden)]
    Cko = 1280,
    #[doc(hidden)]
    Ckq = 1281,
    #[doc(hidden)]
    Ckr = 1282,
    #[doc(hidden)]
    Cks = 1283,
    #[doc(hidden)]
    Ckt = 1284,
    #[doc(hidden)]
    Cku = 1285,
    #[doc(hidden)]
    Ckv = 1286,
    #[doc(hidden)]
    Ckx = 1287,
    #[doc(hidden)]
    Cky = 1288,
    #[doc(hidden)]
    Ckz = 1289,
    #[doc(hidden)]
    Cla = 1290,
    #[doc(hidden)]
    Clc = 1291,
    #[doc(hidden)]
    Cld = 1292,
    #[doc(hidden)]
    Cle = 1293,
    #[doc(hidden)]
    Clh = 1294,
    #[doc(hidden)]
    Cli = 1295,
    #[doc(hidden)]
    Clj = 1296,
    #[doc(hidden)]
    Clk = 1297,
    #[doc(hidden)]
    Cll = 1298,
    #[doc(hidden)]
    Clm = 1299,
    #[doc(hidden)]
    Clo = 1300,
    #[doc(hidden)]
    Clt = 1301,
    #[doc(hidden)]
    Clu = 1302,
    #[doc(hidden)]
    Clw = 1303,
    #[doc(hidden)]
    Cly = 1304,
    #[doc(hidden)]
    Cma = 1305,
    #[doc(hidden)]
    Cme = 1306,
    #[doc(hidden)]
    Cmg = 1307,
    #[doc(hidden)]
    Cmi = 1308,
    #[doc(hidden)]
    Cml = 1309,
    #[doc(hidden)]
    Cmm = 1310,
    #[doc(hidden)]
    Cmn = 1311,
    #[doc(hidden)]
    Cmo = 1312,
    #[doc(hidden)]
    Cmr = 1313,
    #[doc(hidden)]
    Cms = 1314,
    #[doc(hidden)]
    Cmt = 1315,
    #[doc(hidden)]
    Cna = 1316,
    #[doc(hidden)]
    Cnb = 1317,
    #[doc(hidden)]
    Cnc = 1318,
    #[doc(hidden)]
    Cng = 1319,
    #[doc(hidden)]
    Cnh = 1320,
    #[doc(hidden)]
    Cni = 1321,
    #[doc(hidden)]
    Cnk = 1322,
    #[doc(hidden)]
    Cnl = 1323,
    #[doc(hidden)]
    Cno = 1324,
    #[doc(hidden)]
    Cnp = 1325,
    #[doc(hidden)]
    Cnq = 1326,
    #[doc(hidden)]
    Cnr = 1327,
    #[doc(hidden)]
    Cns = 1328,
    #[doc(hidden)]
    Cnt = 1329,
    #[doc(hidden)]
    Cnu = 1330,
    #[doc(hidden)]
    Cnw = 1331,
    #[doc(hidden)]
    Cnx = 1332,
    #[doc(hidden)]
    Coa = 1333,
    #[doc(hidden)]
    Cob = 1334,
    #[doc(hidden)]
    Coc = 1335,
    #[doc(hidden)]
    Cod = 1336,
    #[doc(hidden)]
    Coe = 1337,
    #[doc(hidden)]
    Cof = 1338,
    #[doc(hidden)]
    Cog = 1339,
    #[doc(hidden)]
    Coh = 1340,
    #[doc(hidden)]
    Coj = 1341,
    #[doc(hidden)]
    Cok = 1342,
    #[doc(hidden)]
    Col = 1343,
    #[doc(hidden)]
    Com = 1344,
    #[doc(hidden)]
    Con = 1345,
    #[doc(hidden)]
    Coo = 1346,
    #[doc(hidden)]
    Cop = 1347,
    #[doc(hidden)]
    Coq = 1348,
    #[doc(hidden)]
    Cor = 1349,
    #[doc(hidden)]
    Cos = 1350,
    #[doc(hidden)]
    Cot = 1351,
    #[doc(hidden)]
    Cou = 1352,
    #[doc(hidden)]
    Cov = 1353,
    #[doc(hidden)]
    Cow = 1354,
    #[doc(hidden)]
    Cox = 1355,
    #[doc(hidden)]
    Coz = 1356,
    #[doc(hidden)]
    Cpa = 1357,
    #[doc(hidden)]
    Cpb = 1358,
    #[doc(hidden)]
    Cpc = 1359,
    #[doc(hidden)]
    Cpg = 1360,
    #[doc(hidden)]
    Cpi = 1361,
    #[doc(hidden)]
    Cpn = 1362,
    #[doc(hidden)]
    Cpo = 1363,
    #[doc(hidden)]
    Cps = 1364,
    #[doc(hidden)]
    Cpu = 1365,
    #[doc(hidden)]
    Cpx = 1366,
    #[doc(hidden)]
    Cpy = 1367,
    #[doc(hidden)]
    Cqd = 1368,
    #[doc(hidden)]
    Cra = 1369,
    #[doc(hidden)]
    Crb = 1370,
    #[doc(hidden)]
    Crc = 1371,
    #[doc(hidden)]
    Crd = 1372,
    #[doc(hidden)]
    Cre = 1373,
    #[doc(hidden)]
    Crf = 1374,
    #[doc(hidden)]
    Crg = 1375,
    #[doc(hidden)]
    Crh = 1376,
    #[doc(hidden)]
    Cri = 1377,
    #[doc(hidden)]
    Crj = 1378,
    #[doc(hidden)]
    Crk = 1379,
    #[doc(hidden)]
    Crl = 1380,
    #[doc(hidden)]
    Crm = 1381,
    #[doc(hidden)]
    Crn = 1382,
    #[doc(hidden)]
    Cro = 1383,
    #[doc(hidden)]
    Crq = 1384,
    #[doc(hidden)]
    Crr = 1385,
    #[doc(hidden)]
    Crs = 1386,
    #[doc(hidden)]
    Crt = 1387,
    #[doc(hidden)]
    Crv = 1388,
    #[doc(hidden)]
    Crw = 1389,
    #[doc(hidden)]
    Crx = 1390,
    #[doc(hidden)]
    Cry = 1391,
    #[doc(hidden)]
    Crz = 1392,
    #[doc(hidden)]
    Csa = 1393,
    #[doc(hidden)]
    Csb = 1394,
    #[doc(hidden)]
    Csc = 1395,
    #[doc(hidden)]
    Csd = 1396,
    #[doc(hidden)]
    Cse = 1397,
    #[doc(hidden)]
    Csf = 1398,
    #[doc(hidden)]
    Csg = 1399,
    #[doc(hidden)]
    Csh = 1400,
    #[doc(hidden)]
    Csi = 1401,
    #[doc(hidden)]
    Csj = 1402,
    #[doc(hidden)]
    Csk = 1403,
    #[doc(hidden)]
    Csl = 1404,
    #[doc(hidden)]
    Csm = 1405,
    #[doc(hidden)]
    Csn = 1406,
    #[doc(hidden)]
    Cso = 1407,
    #[doc(hidden)]
    Csp = 1408,
    #[doc(hidden)]
    Csq = 1409,
    #[doc(hidden)]
    Csr = 1410,
    #[doc(hidden)]
    Css = 1411,
    #[doc(hidden)]
    Cst = 1412,
    #[doc(hidden)]
    Csv = 1413,
    #[doc(hidden)]
    Csw = 1414,
    #[doc(hidden)]
    Csx = 1415,
    #[doc(hidden)]
    Csy = 1416,
    #[doc(hidden)]
    Csz = 1417,
    #[doc(hidden)]
    Cta = 1418,
    #[doc(hidden)]
    Ctc = 1419,
    #[doc(hidden)]
    Ctd = 1420,
    #[doc(hidden)]
    Cte = 1421,
    #[doc(hidden)]
    Ctg = 1422,
    #[doc(hidden)]
    Cth = 1423,
    #[doc(hidden)]
    Ctl = 1424,
    #[doc(hidden)]
    Ctm = 1425,
    #[doc(hidden)]
    Ctn = 1426,
    #[doc(hidden)]
    Cto = 1427,
    #[doc(hidden)]
    Ctp = 1428,
    #[doc(hidden)]
    Cts = 1429,
    #[doc(hidden)]
    Ctt = 1430,
    #[doc(hidden)]
    Ctu = 1431,
    #[doc(hidden)]
    Cty = 1432,
    #[doc(hidden)]
    Ctz = 1433,
    #[doc(hidden)]
    Cua = 1434,
    #[doc(hidden)]
    Cub = 1435,
    #[doc(hidden)]
    Cuc = 1436,
    #[doc(hidden)]
    Cuh = 1437,
    #[doc(hidden)]
    Cui = 1438,
    #[doc(hidden)]
    Cuj = 1439,
    #[doc(hidden)]
    Cuk = 1440,
    #[doc(hidden)]
    Cul = 1441,
    #[doc(hidden)]
    Cuo = 1442,
    #[doc(hidden)]
    Cup = 1443,
    #[doc(hidden)]
    Cuq = 1444,
    #[doc(hidden)]
    Cur = 1445,
    #[doc(hidden)]
    Cut = 1446,
    #[doc(hidden)]
    Cuu = 1447,
    #[doc(hidden)]
    Cuv = 1448,
    #[doc(hidden)]
    Cuw = 1449,
    #[doc(hidden)]
    Cux = 1450,
    #[doc(hidden)]
    Cuy = 1451,
    #[doc(hidden)]
    Cvg = 1452,
    #[doc(hidden)]
    Cvn = 1453,
    #[doc(hidden)]
    Cwa = 1454,
    #[doc(hidden)]
    Cwb = 1455,
    #[doc(hidden)]
    Cwd = 1456,
    #[doc(hidden)]
    Cwe = 1457,
    #[doc(hidden)]
    Cwg = 1458,
    #[doc(hidden)]
    Cwt = 1459,
    #[doc(hidden)]
    Cya = 1460,
    #[doc(hidden)]
    Cyb = 1461,
    #[doc(hidden)]
    Cym = 1462,
    #[doc(hidden)]
    Cyo = 1463,
    #[doc(hidden)]
    Czh = 1464,
    #[doc(hidden)]
    Czk = 1465,
    #[doc(hidden)]
    Czn = 1466,
    #[doc(hidden)]
    Czo = 1467,
    #[doc(hidden)]
    Czt = 1468,
    #[doc(hidden)]
    Daa = 1469,
    #[doc(hidden)]
    Dac = 1470,
    #[doc(hidden)]
    Dad = 1471,
    #[doc(hidden)]
    Dae = 1472,
    #[doc(hidden)]
    Dag = 1473,
    #[doc(hidden)]
    Dah = 1474,
    #[doc(hidden)]
    Dai = 1475,
    #[doc(hidden)]
    Daj = 1476,
    #[doc(hidden)]
    Dak = 1477,
    #[doc(hidden)]
    Dal = 1478,
    #[doc(hidden)]
    Dam = 1479,
    #[doc(hidden)]
    Dan = 1480,
    #[doc(hidden)]
    Dao = 1481,
    #[doc(hidden)]
    Daq = 1482,
    #[doc(hidden)]
    Dar = 1483,
    #[doc(hidden)]
    Das = 1484,
    #[doc(hidden)]
    Dau = 1485,
    #[doc(hidden)]
    Dav = 1486,
    #[doc(hidden)]
    Daw = 1487,
    #[doc(hidden)]
    Dax = 1488,
    #[doc(hidden)]
    Daz = 1489,
    #[doc(hidden)]
    Dba = 1490,
    #[doc(hidden)]
    Dbb = 1491,
    #[doc(hidden)]
    Dbd = 1492,
    #[doc(hidden)]
    Dbe = 1493,
    #[doc(hidden)]
    Dbf = 1494,
    #[doc(hidden)]
    Dbg = 1495,
    #[doc(hidden)]
    Dbi = 1496,
    #[doc(hidden)]
    Dbj = 1497,
    #[doc(hidden)]
    Dbl = 1498,
    #[doc(hidden)]
    Dbm = 1499,
    #[doc(hidden)]
    Dbn = 1500,
    #[doc(hidden)]
    Dbo = 1501,
    #[doc(hidden)]
    Dbp = 1502,
    #[doc(hidden)]
    Dbq = 1503,
    #[doc(hidden)]
    Dbr = 1504,
    #[doc(hidden)]
    Dbt = 1505,
    #[doc(hidden)]
    Dbu = 1506,
    #[doc(hidden)]
    Dbv = 1507,
    #[doc(hidden)]
    Dbw = 1508,
    #[doc(hidden)]
    Dby = 1509,
    #[doc(hidden)]
    Dcc = 1510,
    #[doc(hidden)]
    Dcr = 1511,
    #[doc(hidden)]
    Dda = 1512,
    #[doc(hidden)]
    Ddd = 1513,
    #[doc(hidden)]
    Dde = 1514,
    #[doc(hidden)]
    Ddg = 1515,
    #[doc(hidden)]
    Ddi = 1516,
    #[doc(hidden)]
    Ddj = 1517,
    #[doc(hidden)]
    Ddn = 1518,
    #[doc(hidden)]
    Ddo = 1519,
    #[doc(hidden)]
    Ddr = 1520,
    #[doc(hidden)]
    Dds = 1521,
    #[doc(hidden)]
    Ddw = 1522,
    #[doc(hidden)]
    Dec = 1523,
    #[doc(hidden)]
    Ded = 1524,
    #[doc(hidden)]
    Dee = 1525,
    #[doc(hidden)]
    Def = 1526,
    #[doc(hidden)]
    Deg = 1527,
    #[doc(hidden)]
    Deh = 1528,
    #[doc(hidden)]
    Dei = 1529,
    #[doc(hidden)]
    Dek = 1530,
    #[doc(hidden)]
    Del = 1531,
    #[doc(hidden)]
    Dem = 1532,
    #[doc(hidden)]
    Den = 1533,
    #[doc(hidden)]
    Dep = 1534,
    #[doc(hidden)]
    Deq = 1535,
    #[doc(hidden)]
    Der = 1536,
    #[doc(hidden)]
    Des = 1537,
    #[doc(hidden)]
    Deu = 1538,
    #[doc(hidden)]
    Dev = 1539,
    #[doc(hidden)]
    Dez = 1540,
    #[doc(hidden)]
    Dga = 1541,
    #[doc(hidden)]
    Dgb = 1542,
    #[doc(hidden)]
    Dgc = 1543,
    #[doc(hidden)]
    Dgd = 1544,
    #[doc(hidden)]
    Dge = 1545,
    #[doc(hidden)]
    Dgg = 1546,
    #[doc(hidden)]
    Dgh = 1547,
    #[doc(hidden)]
    Dgi = 1548,
    #[doc(hidden)]
    Dgk = 1549,
    #[doc(hidden)]
    Dgl = 1550,
    #[doc(hidden)]
    Dgn = 1551,
    #[doc(hidden)]
    Dgo = 1552,
    #[doc(hidden)]
    Dgr = 1553,
    #[doc(hidden)]
    Dgs = 1554,
    #[doc(hidden)]
    Dgt = 1555,
    #[doc(hidden)]
    Dgw = 1556,
    #[doc(hidden)]
    Dgx = 1557,
    #[doc(hidden)]
    Dgz = 1558,
    #[doc(hidden)]
    Dhd = 1559,
    #[doc(hidden)]
    Dhg = 1560,
    #[doc(hidden)]
    Dhi = 1561,
    #[doc(hidden)]
    Dhl = 1562,
    #[doc(hidden)]
    Dhm = 1563,
    #[doc(hidden)]
    Dhn = 1564,
    #[doc(hidden)]
    Dho = 1565,
    #[doc(hidden)]
    Dhr = 1566,
    #[doc(hidden)]
    Dhs = 1567,
    #[doc(hidden)]
    Dhu = 1568,
    #[doc(hidden)]
    Dhv = 1569,
    #[doc(hidden)]
    Dhw = 1570,
    #[doc(hidden)]
    Dhx = 1571,
    #[doc(hidden)]
    Dia = 1572,
    #[doc(hidden)]
    Dib = 1573,
    #[doc(hidden)]
    Dic = 1574,
    #[doc(hidden)]
    Did = 1575,
    #[doc(hidden)]
    Dif = 1576,
    #[doc(hidden)]
    Dig = 1577,
    #[doc(hidden)]
    Dih = 1578,
    #[doc(hidden)]
    Dii = 1579,
    #[doc(hidden)]
    Dij = 1580,
    #[doc(hidden)]
    Dik = 1581,
    #[doc(hidden)]
    Dil = 1582,
    #[doc(hidden)]
    Dim = 1583,
    #[doc(hidden)]
    Din = 1584,
    #[doc(hidden)]
    Dio = 1585,
    #[doc(hidden)]
    Dip = 1586,
    #[doc(hidden)]
    Diq = 1587,
    #[doc(hidden)]
    Dir = 1588,
    #[doc(hidden)]
    Dis = 1589,
    #[doc(hidden)]
    Diu = 1590,
    #[doc(hidden)]
    Div = 1591,
    #[doc(hidden)]
    Diw = 1592,
    #[doc(hidden)]
    Dix = 1593,
    #[doc(hidden)]
    Diy = 1594,
    #[doc(hidden)]
    Diz = 1595,
    #[doc(hidden)]
    Dja = 1596,
    #[doc(hidden)]
    Djb = 1597,
    #[doc(hidden)]
    Djc = 1598,
    #[doc(hidden)]
    Djd = 1599,
    #[doc(hidden)]
    Dje = 1600,
    #[doc(hidden)]
    Djf = 1601,
    #[doc(hidden)]
    Dji = 1602,
    #[doc(hidden)]
    Djj = 1603,
    #[doc(hidden)]
    Djk = 1604,
    #[doc(hidden)]
    Djm = 1605,
    #[doc(hidden)]
    Djn = 1606,
    #[doc(hidden)]
    Djo = 1607,
    #[doc(hidden)]
    Djr = 1608,
    #[doc(hidden)]
    Dju = 1609,
    #[doc(hidden)]
    Djw = 1610,
    #[doc(hidden)]
    Dka = 1611,
    #[doc(hidden)]
    Dkg = 1612,
    #[doc(hidden)]
    Dkk = 1613,
    #[doc(hidden)]
    Dkr = 1614,
    #[doc(hidden)]
    Dks = 1615,
    #[doc(hidden)]
    Dkx = 1616,
    #[doc(hidden)]
    Dlg = 1617,
    #[doc(hidden)]
    Dlk = 1618,
    #[doc(hidden)]
    Dlm = 1619,
    #[doc(hidden)]
    Dln = 1620,
    #[doc(hidden)]
    Dma = 1621,
    #[doc(hidden)]
    Dmb = 1622,
    #[doc(hidden)]
    Dmc = 1623,
    #[doc(hidden)]
    Dmd = 1624,
    #[doc(hidden)]
    Dme = 1625,
    #[doc(hidden)]
    Dmf = 1626,
    #[doc(hidden)]
    Dmg = 1627,
    #[doc(hidden)]
    Dmk = 1628,
    #[doc(hidden)]
    Dml = 1629,
    #[doc(hidden)]
    Dmm = 1630,
    #[doc(hidden)]
    Dmo = 1631,
    #[doc(hidden)]
    Dmr = 1632,
    #[doc(hidden)]
    Dms = 1633,
    #[doc(hidden)]
    Dmu = 1634,
    #[doc(hidden)]
    Dmv = 1635,
    #[doc(hidden)]
    Dmw = 1636,
    #[doc(hidden)]
    Dmx = 1637,
    #[doc(hidden)]
    Dmy = 1638,
    #[doc(hidden)]
    Dna = 1639,
    #[doc(hidden)]
    Dnd = 1640,
    #[doc(hidden)]
    Dne = 1641,
    #[doc(hidden)]
    Dng = 1642,
    #[doc(hidden)]
    Dni = 1643,
    #[doc(hidden)]
    Dnj = 1644,
    #[doc(hidden)]
    Dnk = 1645,
    #[doc(hidden)]
    Dnn = 1646,
    #[doc(hidden)]
    Dno = 1647,
    #[doc(hidden)]
    Dnr = 1648,
    #[doc(hidden)]
    Dnt = 1649,
    #[doc(hidden)]
    Dnu = 1650,
    #[doc(hidden)]
    Dnv = 1651,
    #[doc(hidden)]
    Dnw = 1652,
    #[doc(hidden)]
    Dny = 1653,
    #[doc(hidden)]
    Doa = 1654,
    #[doc(hidden)]
    Dob = 1655,
    #[doc(hidden)]
    Doc = 1656,
    #[doc(hidden)]
    Doe = 1657,
    #[doc(hidden)]
    Dof = 1658,
    #[doc(hidden)]
    Doh = 1659,
    #[doc(hidden)]
    Doi = 1660,
    #[doc(hidden)]
    Dok = 1661,
    #[doc(hidden)]
    Dol = 1662,
    #[doc(hidden)]
    Don = 1663,
    #[doc(hidden)]
    Doo = 1664,
    #[doc(hidden)]
    Dop = 1665,
    #[doc(hidden)]
    Doq = 1666,
    #[doc(hidden)]
    Dor = 1667,
    #[doc(hidden)]
    Dos = 1668,
    #[doc(hidden)]
    Dot = 1669,
    #[doc(hidden)]
    Dov = 1670,
    #[doc(hidden)]
    Dow = 1671,
    #[doc(hidden)]
    Dox = 1672,
    #[doc(hidden)]
    Doy = 1673,
    #[doc(hidden)]
    Doz = 1674,
    #[doc(hidden)]
    Dpp = 1675,
    #[doc(hidden)]
    Drb = 1676,
    #[doc(hidden)]
    Drc = 1677,
    #[doc(hidden)]
    Drd = 1678,
    #[doc(hidden)]
    Dre = 1679,
    #[doc(hidden)]
    Drg = 1680,
    #[doc(hidden)]
    Dri = 1681,
    #[doc(hidden)]
    Drl = 1682,
    #[doc(hidden)]
    Drn = 1683,
    #[doc(hidden)]
    Dro = 1684,
    #[doc(hidden)]
    Drq = 1685,
    #[doc(hidden)]
    Drs = 1686,
    #[doc(hidden)]
    Drt = 1687,
    #[doc(hidden)]
    Dru = 1688,
    #[doc(hidden)]
    Dry = 1689,
    #[doc(hidden)]
    Dsb = 1690,
    #[doc(hidden)]
    Dse = 1691,
    #[doc(hidden)]
    Dsh = 1692,
    #[doc(hidden)]
    Dsi = 1693,
    #[doc(hidden)]
    Dsl = 1694,
    #[doc(hidden)]
    Dsn = 1695,
    #[doc(hidden)]
    Dso = 1696,
    #[doc(hidden)]
    Dsq = 1697,
    #[doc(hidden)]
    Dsz = 1698,
    #[doc(hidden)]
    Dta = 1699,
    #[doc(hidden)]
    Dtb = 1700,
    #[doc(hidden)]
    Dtd = 1701,
    #[doc(hidden)]
    Dth = 1702,
    #[doc(hidden)]
    Dti = 1703,
    #[doc(hidden)]
    Dtk = 1704,
    #[doc(hidden)]
    Dtm = 1705,
    #[doc(hidden)]
    Dtn = 1706,
    #[doc(hidden)]
    Dto = 1707,
    #[doc(hidden)]
    Dtp = 1708,
    #[doc(hidden)]
    Dtr = 1709,
    #[doc(hidden)]
    Dts = 1710,
    #[doc(hidden)]
    Dtt = 1711,
    #[doc(hidden)]
    Dtu = 1712,
    #[doc(hidden)]
    Dty = 1713,
    #[doc(hidden)]
    Dua = 1714,
    #[doc(hidden)]
    Dub = 1715,
    #[doc(hidden)]
    Duc = 1716,
    #[doc(hidden)]
    Due = 1717,
    #[doc(hidden)]
    Duf = 1718,
    #[doc(hidden)]
    Dug = 1719,
    #[doc(hidden)]
    Duh = 1720,
    #[doc(hidden)]
    Dui = 1721,
    #[doc(hidden)]
    Duk = 1722,
    #[doc(hidden)]
    Dul = 1723,
    #[doc(hidden)]
    Dum = 1724,
    #[doc(hidden)]
    Dun = 1725,
    #[doc(hidden)]
    Duo = 1726,
    #[doc(hidden)]
    Dup = 1727,
    #[doc(hidden)]
    Duq = 1728,
    #[doc(hidden)]
    Dur = 1729,
    #[doc(hidden)]
    Dus = 1730,
    #[doc(hidden)]
    Duu = 1731,
    #[doc(hidden)]
    Duv = 1732,
    #[doc(hidden)]
    Duw = 1733,
    #[doc(hidden)]
    Dux = 1734,
    #[doc(hidden)]
    Duy = 1735,
    #[doc(hidden)]
    Duz = 1736,
    #[doc(hidden)]
    Dva = 1737,
    #[doc(hidden)]
    Dwa = 1738,
    #[doc(hidden)]
    Dwk = 1739,
    #[doc(hidden)]
    Dwr = 1740,
    #[doc(hidden)]
    Dws = 1741,
    #[doc(hidden)]
    Dwu = 1742,
    #[doc(hidden)]
    Dww = 1743,
    #[doc(hidden)]
    Dwy = 1744,
    #[doc(hidden)]
    Dwz = 1745,
    #[doc(hidden)]
    Dya = 1746,
    #[doc(hidden)]
    Dyb = 1747,
    #[doc(hidden)]
    Dyd = 1748,
    #[doc(hidden)]
    Dyg = 1749,
    #[doc(hidden)]
    Dyi = 1750,
    #[doc(hidden)]
    Dym = 1751,
    #[doc(hidden)]
    Dyn = 1752,
    #[doc(hidden)]
    Dyo = 1753,
    #[doc(hidden)]
    Dyu = 1754,
    #[doc(hidden)]
    Dyy = 1755,
    #[doc(hidden)]
    Dza = 1756,
    #[doc(hidden)]
    Dze = 1757,
    #[doc(hidden)]
    Dzg = 1758,
    #[doc(hidden)]
    Dzl = 1759,
    #[doc(hidden)]
    Dzn = 1760,
    #[doc(hidden)]
    Dzo = 1761,
    #[doc(hidden)]
    Eaa = 1762,
    #[doc(hidden)]
    Ebc = 1763,
    #[doc(hidden)]
    Ebg = 1764,
    #[doc(hidden)]
    Ebk = 1765,
    #[doc(hidden)]
    Ebo = 1766,
    #[doc(hidden)]
    Ebr = 1767,
    #[doc(hidden)]
    Ebu = 1768,
    #[doc(hidden)]
    Ecr = 1769,
    #[doc(hidden)]
    Ecs = 1770,
    #[doc(hidden)]
    Ecy = 1771,
    #[doc(hidden)]
    Eee = 1772,
    #[doc(hidden)]
    Efa = 1773,
    #[doc(hidden)]
    Efe = 1774,
    #[doc(hidden)]
    Efi = 1775,
    #[doc(hidden)]
    Ega = 1776,
    #[doc(hidden)]
    Egl = 1777,
    #[doc(hidden)]
    Egm = 1778,
    #[doc(hidden)]
    Ego = 1779,
    #[doc(hidden)]
    Egy = 1780,
    #[doc(hidden)]
    Ehs = 1781,
    #[doc(hidden)]
    Ehu = 1782,
    #[doc(hidden)]
    Eip = 1783,
    #[doc(hidden)]
    Eit = 1784,
    #[doc(hidden)]
    Eiv = 1785,
    #[doc(hidden)]
    Eja = 1786,
    #[doc(hidden)]
    Eka = 1787,
    #[doc(hidden)]
    Eke = 1788,
    #[doc(hidden)]
    Ekg = 1789,
    #[doc(hidden)]
    Eki = 1790,
    #[doc(hidden)]
    Ekk = 1791,
    #[doc(hidden)]
    Ekl = 1792,
    #[doc(hidden)]
    Ekm = 1793,
    #[doc(hidden)]
    Eko = 1794,
    #[doc(hidden)]
    Ekp = 1795,
    #[doc(hidden)]
    Ekr = 1796,
    #[doc(hidden)]
    Eky = 1797,
    #[doc(hidden)]
    Ele = 1798,
    #[doc(hidden)]
    Elh = 1799,
    #[doc(hidden)]
    Eli = 1800,
    #[doc(hidden)]
    Elk = 1801,
    #[doc(hidden)]
    Ell = 1802,
    #[doc(hidden)]
    Elm = 1803,
    #[doc(hidden)]
    Elo = 1804,
    #[doc(hidden)]
    Elu = 1805,
    #[doc(hidden)]
    Elx = 1806,
    #[doc(hidden)]
    Ema = 1807,
    #[doc(hidden)]
    Emb = 1808,
    #[doc(hidden)]
    Eme = 1809,
    #[doc(hidden)]
    Emg = 1810,
    #[doc(hidden)]
    Emi = 1811,
    #[doc(hidden)]
    Emk = 1812,
    #[doc(hidden)]
    Emm = 1813,
    #[doc(hidden)]
    Emn = 1814,
    #[doc(hidden)]
    Emp = 1815,
    #[doc(hidden)]
    Emq = 1816,
    #[doc(hidden)]
    Ems = 1817,
    #[doc(hidden)]
    Emu = 1818,
    #[doc(hidden)]
    Emw = 1819,
    #[doc(hidden)]
    Emx = 1820,
    #[doc(hidden)]
    Emy = 1821,
    #[doc(hidden)]
    Emz = 1822,
    #[doc(hidden)]
    Ena = 1823,
    #[doc(hidden)]
    Enb = 1824,
    #[doc(hidden)]
    Enc = 1825,
    #[doc(hidden)]
    End = 1826,
    #[doc(hidden)]
    Enf = 1827,
    #[doc(hidden)]
    Eng = 1828,
    #[doc(hidden)]
    Enh = 1829,
    #[doc(hidden)]
    Enl = 1830,
    #[doc(hidden)]
    Enm = 1831,
    #[doc(hidden)]
    Enn = 1832,
    #[doc(hidden)]
    Eno = 1833,
    #[doc(hidden)]
    Enq = 1834,
    #[doc(hidden)]
    Enr = 1835,
    #[doc(hidden)]
    Enu = 1836,
    #[doc(hidden)]
    Env = 1837,
    #[doc(hidden)]
    Enw = 1838,
    #[doc(hidden)]
    Enx = 1839,
    #[doc(hidden)]
    Eot = 1840,
    #[doc(hidden)]
    Epi = 1841,
    #[doc(hidden)]
    Epo = 1842,
    #[doc(hidden)]
    Era = 1843,
    #[doc(hidden)]
    Erg = 1844,
    #[doc(hidden)]
    Erh = 1845,
    #[doc(hidden)]
    Eri = 1846,
    #[doc(hidden)]
    Erk = 1847,
    #[doc(hidden)]
    Ero = 1848,
    #[doc(hidden)]
    Err = 1849,
    #[doc(hidden)]
    Ers = 1850,
    #[doc(hidden)]
    Ert = 1851,
    #[doc(hidden)]
    Erw = 1852,
    #[doc(hidden)]
    Ese = 1853,
    #[doc(hidden)]
    Esg = 1854,
    #[doc(hidden)]
    Esh = 1855,
    #[doc(hidden)]
    Esi = 1856,
    #[doc(hidden)]
    Esk = 1857,
    #[doc(hidden)]
    Esl = 1858,
    #[doc(hidden)]
    Esm = 1859,
    #[doc(hidden)]
    Esn = 1860,
    #[doc(hidden)]
    Eso = 1861,
    #[doc(hidden)]
    Esq = 1862,
    #[doc(hidden)]
    Ess = 1863,
    #[doc(hidden)]
    Est = 1864,
    #[doc(hidden)]
    Esu = 1865,
    #[doc(hidden)]
    Esy = 1866,
    #[doc(hidden)]
    Etb = 1867,
    #[doc(hidden)]
    Etc = 1868,
    #[doc(hidden)]
    Eth = 1869,
    #[doc(hidden)]
    Etn = 1870,
    #[doc(hidden)]
    Eto = 1871,
    #[doc(hidden)]
    Etr = 1872,
    #[doc(hidden)]
    Ets = 1873,
    #[doc(hidden)]
    Ett = 1874,
    #[doc(hidden)]
    Etu = 1875,
    #[doc(hidden)]
    Etx = 1876,
    #[doc(hidden)]
    Etz = 1877,
    #[doc(hidden)]
    Eus = 1878,
    #[doc(hidden)]
    Eve = 1879,
    #[doc(hidden)]
    Evh = 1880,
    #[doc(hidden)]
    Evn = 1881,
    #[doc(hidden)]
    Ewe = 1882,
    #[doc(hidden)]
    Ewo = 1883,
    #[doc(hidden)]
    Ext = 1884,
    #[doc(hidden)]
    Eya = 1885,
    #[doc(hidden)]
    Eyo = 1886,
    #[doc(hidden)]
    Eza = 1887,
    #[doc(hidden)]
    Eze = 1888,
    #[doc(hidden)]
    Faa = 1889,
    #[doc(hidden)]
    Fab = 1890,
    #[doc(hidden)]
    Fad = 1891,
    #[doc(hidden)]
    Faf = 1892,
    #[doc(hidden)]
    Fag = 1893,
    #[doc(hidden)]
    Fah = 1894,
    #[doc(hidden)]
    Fai = 1895,
    #[doc(hidden)]
    Faj = 1896,
    #[doc(hidden)]
    Fak = 1897,
    #[doc(hidden)]
    Fal = 1898,
    #[doc(hidden)]
    Fam = 1899,
    #[doc(hidden)]
    Fan = 1900,
    #[doc(hidden)]
    Fao = 1901,
    #[doc(hidden)]
    Fap = 1902,
    #[doc(hidden)]
    Far = 1903,
    #[doc(hidden)]
    Fas = 1904,
    #[doc(hidden)]
    Fat = 1905,
    #[doc(hidden)]
    Fau = 1906,
    #[doc(hidden)]
    Fax = 1907,
    #[doc(hidden)]
    Fay = 1908,
    #[doc(hidden)]
    Faz = 1909,
    #[doc(hidden)]
    Fbl = 1910,
    #[doc(hidden)]
    Fcs = 1911,
    #[doc(hidden)]
    Fer = 1912,
    #[doc(hidden)]
    Ffi = 1913,
    #[doc(hidden)]
    Ffm = 1914,
    #[doc(hidden)]
    Fgr = 1915,
    #[doc(hidden)]
    Fia = 1916,
    #[doc(hidden)]
    Fie = 1917,
    #[doc(hidden)]
    Fif = 1918,
    #[doc(hidden)]
    Fij = 1919,
    #[doc(hidden)]
    Fil = 1920,
    #[doc(hidden)]
    Fin = 1921,
    #[doc(hidden)]
    Fip = 1922,
    #[doc(hidden)]
    Fir = 1923,
    #[doc(hidden)]
    Fit = 1924,
    #[doc(hidden)]
    Fiw = 1925,
    #[doc(hidden)]
    Fkk = 1926,
    #[doc(hidden)]
    Fkv = 1927,
    #[doc(hidden)]
    Fla = 1928,
    #[doc(hidden)]
    Flh = 1929,
    #[doc(hidden)]
    Fli = 1930,
    #[doc(hidden)]
    Fll = 1931,
    #[doc(hidden)]
    Fln = 1932,
    #[doc(hidden)]
    Flr = 1933,
    #[doc(hidden)]
    Fly = 1934,
    #[doc(hidden)]
    Fmp = 1935,
    #[doc(hidden)]
    Fmu = 1936,
    #[doc(hidden)]
    Fnb = 1937,
    #[doc(hidden)]
    Fng = 1938,
    #[doc(hidden)]
    Fni = 1939,
    #[doc(hidden)]
    Fod = 1940,
    #[doc(hidden)]
    Foi = 1941,
    #[doc(hidden)]
    Fom = 1942,
    #[doc(hidden)]
    Fon = 1943,
    #[doc(hidden)]
    For = 1944,
    #[doc(hidden)]
    Fos = 1945,
    #[doc(hidden)]
    Fpe = 1946,
    #[doc(hidden)]
    Fqs = 1947,
    #[doc(hidden)]
    Fra = 1948,
    #[doc(hidden)]
    Frc = 1949,
    #[doc(hidden)]
    Frd = 1950,
    #[doc(hidden)]
    Frk = 1951,
    #[doc(hidden)]
    Frm = 1952,
    #[doc(hidden)]
    Fro = 1953,
    #[doc(hidden)]
    Frp = 1954,
    #[doc(hidden)]
    Frq = 1955,
    #[doc(hidden)]
    Frr = 1956,
    #[doc(hidden)]
    Frs = 1957,
    #[doc(hidden)]
    Frt = 1958,
    #[doc(hidden)]
    Fry = 1959,
    #[doc(hidden)]
    Fse = 1960,
    #[doc(hidden)]
    Fsl = 1961,
    #[doc(hidden)]
    Fss = 1962,
    #[doc(hidden)]
    Fub = 1963,
    #[doc(hidden)]
    Fuc = 1964,
    #[doc(hidden)]
    Fud = 1965,
    #[doc(hidden)]
    Fue = 1966,
    #[doc(hidden)]
    Fuf = 1967,
    #[doc(hidden)]
    Fuh = 1968,
    #[doc(hidden)]
    Fui = 1969,
    #[doc(hidden)]
    Fuj = 1970,
    #[doc(hidden)]
    Ful = 1971,
    #[doc(hidden)]
    Fum = 1972,
    #[doc(hidden)]
    Fun = 1973,
    #[doc(hidden)]
    Fuq = 1974,
    #[doc(hidden)]
    Fur = 1975,
    #[doc(hidden)]
    Fut = 1976,
    #[doc(hidden)]
    Fuu = 1977,
    #[doc(hidden)]
    Fuv = 1978,
    #[doc(hidden)]
    Fuy = 1979,
    #[doc(hidden)]
    Fvr = 1980,
    #[doc(hidden)]
    Fwa = 1981,
    #[doc(hidden)]
    Fwe = 1982,
    #[doc(hidden)]
    Gaa = 1983,
    #[doc(hidden)]
    Gab = 1984,
    #[doc(hidden)]
    Gac = 1985,
    #[doc(hidden)]
    Gad = 1986,
    #[doc(hidden)]
    Gae = 1987,
    #[doc(hidden)]
    Gaf = 1988,
    #[doc(hidden)]
    Gag = 1989,
    #[doc(hidden)]
    Gah = 1990,
    #[doc(hidden)]
    Gai = 1991,
    #[doc(hidden)]
    Gaj = 1992,
    #[doc(hidden)]
    Gak = 1993,
    #[doc(hidden)]
    Gal = 1994,
    #[doc(hidden)]
    Gam = 1995,
    #[doc(hidden)]
    Gan = 1996,
    #[doc(hidden)]
    Gao = 1997,
    #[doc(hidden)]
    Gap = 1998,
    #[doc(hidden)]
    Gaq = 1999,
    #[doc(hidden)]
    Gar = 2000,
    #[doc(hidden)]
    Gas = 2001,
    #[doc(hidden)]
    Gat = 2002,
    #[doc(hidden)]
    Gau = 2003,
    #[doc(hidden)]
    Gaw = 2004,
    #[doc(hidden)]
    Gax = 2005,
    #[doc(hidden)]
    Gay = 2006,
    #[doc(hidden)]
    Gaz = 2007,
    #[doc(hidden)]
    Gba = 2008,
    #[doc(hidden)]
    Gbb = 2009,
    #[doc(hidden)]
    Gbd = 2010,
    #[doc(hidden)]
    Gbe = 2011,
    #[doc(hidden)]
    Gbf = 2012,
    #[doc(hidden)]
    Gbg = 2013,
    #[doc(hidden)]
    Gbh = 2014,
    #[doc(hidden)]
    Gbi = 2015,
    #[doc(hidden)]
    Gbj = 2016,
    #[doc(hidden)]
    Gbk = 2017,
    #[doc(hidden)]
    Gbl = 2018,
    #[doc(hidden)]
    Gbm = 2019,
    #[doc(hidden)]
    Gbn = 2020,
    #[doc(hidden)]
    Gbo = 2021,
    #[doc(hidden)]
    Gbp = 2022,
    #[doc(hidden)]
    Gbq = 2023,
    #[doc(hidden)]
    Gbr = 2024,
    #[doc(hidden)]
    Gbs = 2025,
    #[doc(hidden)]
    Gbu = 2026,
    #[doc(hidden)]
    Gbv = 2027,
    #[doc(hidden)]
    Gbw = 2028,
    #[doc(hidden)]
    Gbx = 2029,
    #[doc(hidden)]
    Gby = 2030,
    #[doc(hidden)]
    Gbz = 2031,
    #[doc(hidden)]
    Gcc = 2032,
    #[doc(hidden)]
    Gcd = 2033,
    #[doc(hidden)]
    Gce = 2034,
    #[doc(hidden)]
    Gcf = 2035,
    #[doc(hidden)]
    Gcl = 2036,
    #[doc(hidden)]
    Gcn = 2037,
    #[doc(hidden)]
    Gcr = 2038,
    #[doc(hidden)]
    Gct = 2039,
    #[doc(hidden)]
    Gda = 2040,
    #[doc(hidden)]
    Gdb = 2041,
    #[doc(hidden)]
    Gdc = 2042,
    #[doc(hidden)]
    Gdd = 2043,
    #[doc(hidden)]
    Gde = 2044,
    #[doc(hidden)]
    Gdf = 2045,
    #[doc(hidden)]
    Gdg = 2046,
    #[doc(hidden)]
    Gdh = 2047,
    #[doc(hidden)]
    Gdi = 2048,
    #[doc(hidden)]
    Gdj = 2049,
    #[doc(hidden)]
    Gdk = 2050,
    #[doc(hidden)]
    Gdl = 2051,
    #[doc(hidden)]
    Gdm = 2052,
    #[doc(hidden)]
    Gdn = 2053,
    #[doc(hidden)]
    Gdo = 2054,
    #[doc(hidden)]
    Gdq = 2055,
    #[doc(hidden)]
    Gdr = 2056,
    #[doc(hidden)]
    Gds = 2057,
    #[doc(hidden)]
    Gdt = 2058,
    #[doc(hidden)]
    Gdu = 2059,
    #[doc(hidden)]
    Gdx = 2060,
    #[doc(hidden)]
    Gea = 2061,
    #[doc(hidden)]
    Geb = 2062,
    #[doc(hidden)]
    Gec = 2063,
    #[doc(hidden)]
    Ged = 2064,
    #[doc(hidden)]
    Gef = 2065,
    #[doc(hidden)]
    Geg = 2066,
    #[doc(hidden)]
    Geh = 2067,
    #[doc(hidden)]
    Gei = 2068,
    #[doc(hidden)]
    Gej = 2069,
    #[doc(hidden)]
    Gek = 2070,
    #[doc(hidden)]
    Gel = 2071,
    #[doc(hidden)]
    Geq = 2072,
    #[doc(hidden)]
    Ges = 2073,
    #[doc(hidden)]
    Gev = 2074,
    #[doc(hidden)]
    Gew = 2075,
    #[doc(hidden)]
    Gex = 2076,
    #[doc(hidden)]
    Gey = 2077,
    #[doc(hidden)]
    Gez = 2078,
    #[doc(hidden)]
    Gfk = 2079,
    #[doc(hidden)]
    Gft = 2080,
    #[doc(hidden)]
    Gga = 2081,
    #[doc(hidden)]
    Ggb = 2082,
    #[doc(hidden)]
    Ggd = 2083,
    #[doc(hidden)]
    Gge = 2084,
    #[doc(hidden)]
    Ggg = 2085,
    #[doc(hidden)]
    Ggk = 2086,
    #[doc(hidden)]
    Ggl = 2087,
    #[doc(hidden)]
    Ggt = 2088,
    #[doc(hidden)]
    Ggu = 2089,
    #[doc(hidden)]
    Ggw = 2090,
    #[doc(hidden)]
    Gha = 2091,
    #[doc(hidden)]
    Ghc = 2092,
    #[doc(hidden)]
    Ghe = 2093,
    #[doc(hidden)]
    Ghh = 2094,
    #[doc(hidden)]
    Ghk = 2095,
    #[doc(hidden)]
    Ghl = 2096,
    #[doc(hidden)]
    Ghn = 2097,
    #[doc(hidden)]
    Gho = 2098,
    #[doc(hidden)]
    Ghr = 2099,
    #[doc(hidden)]
    Ghs = 2100,
    #[doc(hidden)]
    Ght = 2101,
    #[doc(hidden)]
    Gia = 2102,
    #[doc(hidden)]
    Gib = 2103,
    #[doc(hidden)]
    Gic = 2104,
    #[doc(hidden)]
    Gid = 2105,
    #[doc(hidden)]
    Gie = 2106,
    #[doc(hidden)]
    Gig = 2107,
    #[doc(hidden)]
    Gih = 2108,
    #[doc(hidden)]
    Gii = 2109,
    #[doc(hidden)]
    Gil = 2110,
    #[doc(hidden)]
    Gim = 2111,
    #[doc(hidden)]
    Gin = 2112,
    #[doc(hidden)]
    Gip = 2113,
    #[doc(hidden)]
    Giq = 2114,
    #[doc(hidden)]
    Gir = 2115,
    #[doc(hidden)]
    Gis = 2116,
    #[doc(hidden)]
    Git = 2117,
    #[doc(hidden)]
    Giu = 2118,
    #[doc(hidden)]
    Giw = 2119,
    #[doc(hidden)]
    Gix = 2120,
    #[doc(hidden)]
    Giy = 2121,
    #[doc(hidden)]
    Giz = 2122,
    #[doc(hidden)]
    Gjk = 2123,
    #[doc(hidden)]
    Gjm = 2124,
    #[doc(hidden)]
    Gjn = 2125,
    #[doc(hidden)]
    Gjr = 2126,
    #[doc(hidden)]
    Gju = 2127,
    #[doc(hidden)]
    Gka = 2128,
    #[doc(hidden)]
    Gkd = 2129,
    #[doc(hidden)]
    Gke = 2130,
    #[doc(hidden)]
    Gkn = 2131,
    #[doc(hidden)]
    Gko = 2132,
    #[doc(hidden)]
    Gkp = 2133,
    #[doc(hidden)]
    Gku = 2134,
    #[doc(hidden)]
    Gla = 2135,
    #[doc(hidden)]
    Glb = 2136,
    #[doc(hidden)]
    Glc = 2137,
    #[doc(hidden)]
    Gld = 2138,
    #[doc(hidden)]
    Gle = 2139,
    #[doc(hidden)]
    Glg = 2140,
    #[doc(hidden)]
    Glh = 2141,
    #[doc(hidden)]
    Glj = 2142,
    #[doc(hidden)]
    Glk = 2143,
    #[doc(hidden)]
    Gll = 2144,
    #[doc(hidden)]
    Glo = 2145,
    #[doc(hidden)]
    Glr = 2146,
    #[doc(hidden)]
    Glu = 2147,
    #[doc(hidden)]
    Glv = 2148,
    #[doc(hidden)]
    Glw = 2149,
    #[doc(hidden)]
    Gly = 2150,
    #[doc(hidden)]
    Gma = 2151,
    #[doc(hidden)]
    Gmb = 2152,
    #[doc(hidden)]
    Gmd = 2153,
    #[doc(hidden)]
    Gmg = 2154,
    #[doc(hidden)]
    Gmh = 2155,
    #[doc(hidden)]
    Gml = 2156,
    #[doc(hidden)]
    Gmm = 2157,
    #[doc(hidden)]
    Gmn = 2158,
    #[doc(hidden)]
    Gmr = 2159,
    #[doc(hidden)]
    Gmu = 2160,
    #[doc(hidden)]
    Gmv = 2161,
    #[doc(hidden)]
    Gmx = 2162,
    #[doc(hidden)]
    Gmy = 2163,
    #[doc(hidden)]
    Gmz = 2164,
    #[doc(hidden)]
    Gna = 2165,
    #[doc(hidden)]
    Gnb = 2166,
    #[doc(hidden)]
    Gnc = 2167,
    #[doc(hidden)]
    Gnd = 2168,
    #[doc(hidden)]
    Gne = 2169,
    #[doc(hidden)]
    Gng = 2170,
    #[doc(hidden)]
    Gnh = 2171,
    #[doc(hidden)]
    Gni = 2172,
    #[doc(hidden)]
    Gnj = 2173,
    #[doc(hidden)]
    Gnk = 2174,
    #[doc(hidden)]
    Gnl = 2175,
    #[doc(hidden)]
    Gnm = 2176,
    #[doc(hidden)]
    Gnn = 2177,
    #[doc(hidden)]
    Gno = 2178,
    #[doc(hidden)]
    Gnq = 2179,
    #[doc(hidden)]
    Gnr = 2180,
    #[doc(hidden)]
    Gnt = 2181,
    #[doc(hidden)]
    Gnu = 2182,
    #[doc(hidden)]
    Gnw = 2183,
    #[doc(hidden)]
    Gnz = 2184,
    #[doc(hidden)]
    Goa = 2185,
    #[doc(hidden)]
    Gob = 2186,
    #[doc(hidden)]
    Goc = 2187,
    #[doc(hidden)]
    God = 2188,
    #[doc(hidden)]
    Goe = 2189,
    #[doc(hidden)]
    Gof = 2190,
    #[doc(hidden)]
    Gog = 2191,
    #[doc(hidden)]
    Goh = 2192,
    #[doc(hidden)]
    Goi = 2193,
    #[doc(hidden)]
    Goj = 2194,
    #[doc(hidden)]
    Gok = 2195,
    #[doc(hidden)]
    Gol = 2196,
    #[doc(hidden)]
    Gom = 2197,
    #[doc(hidden)]
    Gon = 2198,
    #[doc(hidden)]
    Goo = 2199,
    #[doc(hidden)]
    Gop = 2200,
    #[doc(hidden)]
    Goq = 2201,
    #[doc(hidden)]
    Gor = 2202,
    #[doc(hidden)]
    Gos = 2203,
    #[doc(hidden)]
    Got = 2204,
    #[doc(hidden)]
    Gou = 2205,
    #[doc(hidden)]
    Gov = 2206,
    #[doc(hidden)]
    Gow = 2207,
    #[doc(hidden)]
    Gox = 2208,
    #[doc(hidden)]
    Goy = 2209,
    #[doc(hidden)]
    Goz = 2210,
    #[doc(hidden)]
    Gpa = 2211,
    #[doc(hidden)]
    Gpe = 2212,
    #[doc(hidden)]
    Gpn = 2213,
    #[doc(hidden)]
    Gqa = 2214,
    #[doc(hidden)]
    Gqi = 2215,
    #[doc(hidden)]
    Gqn = 2216,
    #[doc(hidden)]
    Gqr = 2217,
    #[doc(hidden)]
    Gqu = 2218,
    #[doc(hidden)]
    Gra = 2219,
    #[doc(hidden)]
    Grb = 2220,
    #[doc(hidden)]
    Grc = 2221,
    #[doc(hidden)]
    Grd = 2222,
    #[doc(hidden)]
    Grg = 2223,
    #[doc(hidden)]
    Grh = 2224,
    #[doc(hidden)]
    Gri = 2225,
    #[doc(hidden)]
    Grj = 2226,
    #[doc(hidden)]
    Grm = 2227,
    #[doc(hidden)]
    Grn = 2228,
    #[doc(hidden)]
    Gro = 2229,
    #[doc(hidden)]
    Grq = 2230,
    #[doc(hidden)]
    Grr = 2231,
    #[doc(hidden)]
    Grs = 2232,
    #[doc(hidden)]
    Grt = 2233,
    #[doc(hidden)]
    Gru = 2234,
    #[doc(hidden)]
    Grv = 2235,
    #[doc(hidden)]
    Grw = 2236,
    #[doc(hidden)]
    Grx = 2237,
    #[doc(hidden)]
    Gry = 2238,
    #[doc(hidden)]
    Grz = 2239,
    #[doc(hidden)]
    Gse = 2240,
    #[doc(hidden)]
    Gsg = 2241,
    #[doc(hidden)]
    Gsl = 2242,
    #[doc(hidden)]
    Gsm = 2243,
    #[doc(hidden)]
    Gsn = 2244,
    #[doc(hidden)]
    Gso = 2245,
    #[doc(hidden)]
    Gsp = 2246,
    #[doc(hidden)]
    Gss = 2247,
    #[doc(hidden)]
    Gsw = 2248,
    #[doc(hidden)]
    Gta = 2249,
    #[doc(hidden)]
    Gtu = 2250,
    #[doc(hidden)]
    Gua = 2251,
    #[doc(hidden)]
    Gub = 2252,
    #[doc(hidden)]
    Guc = 2253,
    #[doc(hidden)]
    Gud = 2254,
    #[doc(hidden)]
    Gue = 2255,
    #[doc(hidden)]
    Guf = 2256,
    #[doc(hidden)]
    Gug = 2257,
    #[doc(hidden)]
    Guh = 2258,
    #[doc(hidden)]
    Gui = 2259,
    #[doc(hidden)]
    Guj = 2260,
    #[doc(hidden)]
    Guk = 2261,
    #[doc(hidden)]
    Gul = 2262,
    #[doc(hidden)]
    Gum = 2263,
    #[doc(hidden)]
    Gun = 2264,
    #[doc(hidden)]
    Guo = 2265,
    #[doc(hidden)]
    Gup = 2266,
    #[doc(hidden)]
    Guq = 2267,
    #[doc(hidden)]
    Gur = 2268,
    #[doc(hidden)]
    Gus = 2269,
    #[doc(hidden)]
    Gut = 2270,
    #[doc(hidden)]
    Guu = 2271,
    #[doc(hidden)]
    Guw = 2272,
    #[doc(hidden)]
    Gux = 2273,
    #[doc(hidden)]
    Guz = 2274,
    #[doc(hidden)]
    Gva = 2275,
    #[doc(hidden)]
    Gvc = 2276,
    #[doc(hidden)]
    Gve = 2277,
    #[doc(hidden)]
    Gvf = 2278,
    #[doc(hidden)]
    Gvj = 2279,
    #[doc(hidden)]
    Gvl = 2280,
    #[doc(hidden)]
    Gvm = 2281,
    #[doc(hidden)]
    Gvn = 2282,
    #[doc(hidden)]
    Gvo = 2283,
    #[doc(hidden)]
    Gvp = 2284,
    #[doc(hidden)]
    Gvr = 2285,
    #[doc(hidden)]
    Gvs = 2286,
    #[doc(hidden)]
    Gvy = 2287,
    #[doc(hidden)]
    Gwa = 2288,
    #[doc(hidden)]
    Gwb = 2289,
    #[doc(hidden)]
    Gwc = 2290,
    #[doc(hidden)]
    Gwd = 2291,
    #[doc(hidden)]
    Gwe = 2292,
    #[doc(hidden)]
    Gwf = 2293,
    #[doc(hidden)]
    Gwg = 2294,
    #[doc(hidden)]
    Gwi = 2295,
    #[doc(hidden)]
    Gwj = 2296,
    #[doc(hidden)]
    Gwm = 2297,
    #[doc(hidden)]
    Gwn = 2298,
    #[doc(hidden)]
    Gwr = 2299,
    #[doc(hidden)]
    Gwt = 2300,
    #[doc(hidden)]
    Gwu = 2301,
    #[doc(hidden)]
    Gww = 2302,
    #[doc(hidden)]
    Gwx = 2303,
    #[doc(hidden)]
    Gxx = 2304,
    #[doc(hidden)]
    Gya = 2305,
    #[doc(hidden)]
    Gyb = 2306,
    #[doc(hidden)]
    Gyd = 2307,
    #[doc(hidden)]
    Gye = 2308,
    #[doc(hidden)]
    Gyf = 2309,
    #[doc(hidden)]
    Gyg = 2310,
    #[doc(hidden)]
    Gyi = 2311,
    #[doc(hidden)]
    Gyl = 2312,
    #[doc(hidden)]
    Gym = 2313,
    #[doc(hidden)]
    Gyn = 2314,
    #[doc(hidden)]
    Gyo = 2315,
    #[doc(hidden)]
    Gyr = 2316,
    #[doc(hidden)]
    Gyy = 2317,
    #[doc(hidden)]
    Gyz = 2318,
    #[doc(hidden)]
    Gza = 2319,
    #[doc(hidden)]
    Gzi = 2320,
    #[doc(hidden)]
    Gzn = 2321,
    #[doc(hidden)]
    Haa = 2322,
    #[doc(hidden)]
    Hab = 2323,
    #[doc(hidden)]
    Hac = 2324,
    #[doc(hidden)]
    Had = 2325,
    #[doc(hidden)]
    Hae = 2326,
    #[doc(hidden)]
    Haf = 2327,
    #[doc(hidden)]
    Hag = 2328,
    #[doc(hidden)]
    Hah = 2329,
    #[doc(hidden)]
    Hai = 2330,
    #[doc(hidden)]
    Haj = 2331,
    #[doc(hidden)]
    Hak = 2332,
    #[doc(hidden)]
    Hal = 2333,
    #[doc(hidden)]
    Ham = 2334,
    #[doc(hidden)]
    Han = 2335,
    #[doc(hidden)]
    Hao = 2336,
    #[doc(hidden)]
    Hap = 2337,
    #[doc(hidden)]
    Haq = 2338,
    #[doc(hidden)]
    Har = 2339,
    #[doc(hidden)]
    Has = 2340,
    #[doc(hidden)]
    Hat = 2341,
    #[doc(hidden)]
    Hau = 2342,
    #[doc(hidden)]
    Hav = 2343,
    #[doc(hidden)]
    Haw = 2344,
    #[doc(hidden)]
    Hax = 2345,
    #[doc(hidden)]
    Hay = 2346,
    #[doc(hidden)]
    Haz = 2347,
    #[doc(hidden)]
    Hba = 2348,
    #[doc(hidden)]
    Hbb = 2349,
    #[doc(hidden)]
    Hbn = 2350,
    #[doc(hidden)]
    Hbo = 2351,
    #[doc(hidden)]
    Hbs = 2352,
    #[doc(hidden)]
    Hbu = 2353,
    #[doc(hidden)]
    Hca = 2354,
    #[doc(hidden)]
    Hch = 2355,
    #[doc(hidden)]
    Hdn = 2356,
    #[doc(hidden)]
    Hds = 2357,
    #[doc(hidden)]
    Hdy = 2358,
    #[doc(hidden)]
    Hea = 2359,
    #[doc(hidden)]
    Heb = 2360,
    #[doc(hidden)]
    Hed = 2361,
    #[doc(hidden)]
    Heg = 2362,
    #[doc(hidden)]
    Heh = 2363,
    #[doc(hidden)]
    Hei = 2364,
    #[doc(hidden)]
    Hem = 2365,
    #[doc(hidden)]
    Her = 2366,
    #[doc(hidden)]
    Hgm = 2367,
    #[doc(hidden)]
    Hgw = 2368,
    #[doc(hidden)]
    Hhi = 2369,
    #[doc(hidden)]
    Hhr = 2370,
    #[doc(hidden)]
    Hhy = 2371,
    #[doc(hidden)]
    Hia = 2372,
    #[doc(hidden)]
    Hib = 2373,
    #[doc(hidden)]
    Hid = 2374,
    #[doc(hidden)]
    Hif = 2375,
    #[doc(hidden)]
    Hig = 2376,
    #[doc(hidden)]
    Hih = 2377,
    #[doc(hidden)]
    Hii = 2378,
    #[doc(hidden)]
    Hij = 2379,
    #[doc(hidden)]
    Hik = 2380,
    #[doc(hidden)]
    Hil = 2381,
    #[doc(hidden)]
    Hin = 2382,
    #[doc(hidden)]
    Hio = 2383,
    #[doc(hidden)]
    Hir = 2384,
    #[doc(hidden)]
    Hit = 2385,
    #[doc(hidden)]
    Hiw = 2386,
    #[doc(hidden)]
    Hix = 2387,
    #[doc(hidden)]
    Hji = 2388,
    #[doc(hidden)]
    Hka = 2389,
    #[doc(hidden)]
    Hke = 2390,
    #[doc(hidden)]
    Hkh = 2391,
    #[doc(hidden)]
    Hkk = 2392,
    #[doc(hidden)]
    Hkn = 2393,
    #[doc(hidden)]
    Hks = 2394,
    #[doc(hidden)]
    Hla = 2395,
    #[doc(hidden)]
    Hlb = 2396,
    #[doc(hidden)]
    Hld = 2397,
    #[doc(hidden)]
    Hle = 2398,
    #[doc(hidden)]
    Hlt = 2399,
    #[doc(hidden)]
    Hlu = 2400,
    #[doc(hidden)]
    Hma = 2401,
    #[doc(hidden)]
    Hmb = 2402,
    #[doc(hidden)]
    Hmc = 2403,
    #[doc(hidden)]
    Hmd = 2404,
    #[doc(hidden)]
    Hme = 2405,
    #[doc(hidden)]
    Hmf = 2406,
    #[doc(hidden)]
    Hmg = 2407,
    #[doc(hidden)]
    Hmh = 2408,
    #[doc(hidden)]
    Hmi = 2409,
    #[doc(hidden)]
    Hmj = 2410,
    #[doc(hidden)]
    Hmk = 2411,
    #[doc(hidden)]
    Hml = 2412,
    #[doc(hidden)]
    Hmm = 2413,
    #[doc(hidden)]
    Hmn = 2414,
    #[doc(hidden)]
    Hmo = 2415,
    #[doc(hidden)]
    Hmp = 2416,
    #[doc(hidden)]
    Hmq = 2417,
    #[doc(hidden)]
    Hmr = 2418,
    #[doc(hidden)]
    Hms = 2419,
    #[doc(hidden)]
    Hmt = 2420,
    #[doc(hidden)]
    Hmu = 2421,
    #[doc(hidden)]
    Hmv = 2422,
    #[doc(hidden)]
    Hmw = 2423,
    #[doc(hidden)]
    Hmy = 2424,
    #[doc(hidden)]
    Hmz = 2425,
    #[doc(hidden)]
    Hna = 2426,
    #[doc(hidden)]
    Hnd = 2427,
    #[doc(hidden)]
    Hne = 2428,
    #[doc(hidden)]
    Hng = 2429,
    #[doc(hidden)]
    Hnh = 2430,
    #[doc(hidden)]
    Hni = 2431,
    #[doc(hidden)]
    Hnj = 2432,
    #[doc(hidden)]
    Hnn = 2433,
    #[doc(hidden)]
    Hno = 2434,
    #[doc(hidden)]
    Hns = 2435,
    #[doc(hidden)]
    Hnu = 2436,
    #[doc(hidden)]
    Hoa = 2437,
    #[doc(hidden)]
    Hob = 2438,
    #[doc(hidden)]
    Hoc = 2439,
    #[doc(hidden)]
    Hod = 2440,
    #[doc(hidden)]
    Hoe = 2441,
    #[doc(hidden)]
    Hoh = 2442,
    #[doc(hidden)]
    Hoi = 2443,
    #[doc(hidden)]
    Hoj = 2444,
    #[doc(hidden)]
    Hol = 2445,
    #[doc(hidden)]
    Hom = 2446,
    #[doc(hidden)]
    Hoo = 2447,
    #[doc(hidden)]
    Hop = 2448,
    #[doc(hidden)]
    Hor = 2449,
    #[doc(hidden)]
    Hos = 2450,
    #[doc(hidden)]
    Hot = 2451,
    #[doc(hidden)]
    Hov = 2452,
    #[doc(hidden)]
    How = 2453,
    #[doc(hidden)]
    Hoy = 2454,
    #[doc(hidden)]
    Hoz = 2455,
    #[doc(hidden)]
    Hpo = 2456,
    #[doc(hidden)]
    Hps = 2457,
    #[doc(hidden)]
    Hra = 2458,
    #[doc(hidden)]
    Hrc = 2459,
    #[doc(hidden)]
    Hre = 2460,
    #[doc(hidden)]
    Hrk = 2461,
    #[doc(hidden)]
    Hrm = 2462,
    #[doc(hidden)]
    Hro = 2463,
    #[doc(hidden)]
    Hrp = 2464,
    #[doc(hidden)]
    Hrt = 2465,
    #[doc(hidden)]
    Hru = 2466,
    #[doc(hidden)]
    Hrv = 2467,
    #[doc(hidden)]
    Hrw = 2468,
    #[doc(hidden)]
    Hrx = 2469,
    #[doc(hidden)]
    Hrz = 2470,
    #[doc(hidden)]
    Hsb = 2471,
    #[doc(hidden)]
    Hsh = 2472,
    #[doc(hidden)]
    Hsl = 2473,
    #[doc(hidden)]
    Hsn = 2474,
    #[doc(hidden)]
    Hss = 2475,
    #[doc(hidden)]
    Hti = 2476,
    #[doc(hidden)]
    Hto = 2477,
    #[doc(hidden)]
    Hts = 2478,
    #[doc(hidden)]
    Htu = 2479,
    #[doc(hidden)]
    Htx = 2480,
    #[doc(hidden)]
    Hub = 2481,
    #[doc(hidden)]
    Huc = 2482,
    #[doc(hidden)]
    Hud = 2483,
    #[doc(hidden)]
    Hue = 2484,
    #[doc(hidden)]
    Huf = 2485,
    #[doc(hidden)]
    Hug = 2486,
    #[doc(hidden)]
    Huh = 2487,
    #[doc(hidden)]
    Hui = 2488,
    #[doc(hidden)]
    Huj = 2489,
    #[doc(hidden)]
    Huk = 2490,
    #[doc(hidden)]
    Hul = 2491,
    #[doc(hidden)]
    Hum = 2492,
    #[doc(hidden)]
    Hun = 2493,
    #[doc(hidden)]
    Huo = 2494,
    #[doc(hidden)]
    Hup = 2495,
    #[doc(hidden)]
    Huq = 2496,
    #[doc(hidden)]
    Hur = 2497,
    #[doc(hidden)]
    Hus = 2498,
    #[doc(hidden)]
    Hut = 2499,
    #[doc(hidden)]
    Huu = 2500,
    #[doc(hidden)]
    Huv = 2501,
    #[doc(hidden)]
    Huw = 2502,
    #[doc(hidden)]
    Hux = 2503,
    #[doc(hidden)]
    Huy = 2504,
    #[doc(hidden)]
    Huz = 2505,
    #[doc(hidden)]
    Hvc = 2506,
    #[doc(hidden)]
    Hve = 2507,
    #[doc(hidden)]
    Hvk = 2508,
    #[doc(hidden)]
    Hvn = 2509,
    #[doc(hidden)]
    Hvv = 2510,
    #[doc(hidden)]
    Hwa = 2511,
    #[doc(hidden)]
    Hwc = 2512,
    #[doc(hidden)]
    Hwo = 2513,
    #[doc(hidden)]
    Hya = 2514,
    #[doc(hidden)]
    Hye = 2515,
    #[doc(hidden)]
    Hyw = 2516,
    #[doc(hidden)]
    Iai = 2517,
    #[doc(hidden)]
    Ian = 2518,
    #[doc(hidden)]
    Iar = 2519,
    #[doc(hidden)]
    Iba = 2520,
    #[doc(hidden)]
    Ibb = 2521,
    #[doc(hidden)]
    Ibd = 2522,
    #[doc(hidden)]
    Ibe = 2523,
    #[doc(hidden)]
    Ibg = 2524,
    #[doc(hidden)]
    Ibh = 2525,
    #[doc(hidden)]
    Ibl = 2526,
    #[doc(hidden)]
    Ibm = 2527,
    #[doc(hidden)]
    Ibn = 2528,
    #[doc(hidden)]
    Ibo = 2529,
    #[doc(hidden)]
    Ibr = 2530,
    #[doc(hidden)]
    Ibu = 2531,
    #[doc(hidden)]
    Iby = 2532,
    #[doc(hidden)]
    Ica = 2533,
    #[doc(hidden)]
    Ich = 2534,
    #[doc(hidden)]
    Icl = 2535,
    #[doc(hidden)]
    Icr = 2536,
    #[doc(hidden)]
    Ida = 2537,
    #[doc(hidden)]
    Idb = 2538,
    #[doc(hidden)]
    Idc = 2539,
    #[doc(hidden)]
    Idd = 2540,
    #[doc(hidden)]
    Ide = 2541,
    #[doc(hidden)]
    Idi = 2542,
    #[doc(hidden)]
    Ido = 2543,
    #[doc(hidden)]
    Idr = 2544,
    #[doc(hidden)]
    Ids = 2545,
    #[doc(hidden)]
    Idt = 2546,
    #[doc(hidden)]
    Idu = 2547,
    #[doc(hidden)]
    Ifa = 2548,
    #[doc(hidden)]
    Ifb = 2549,
    #[doc(hidden)]
    Ife = 2550,
    #[doc(hidden)]
    Iff = 2551,
    #[doc(hidden)]
    Ifk = 2552,
    #[doc(hidden)]
    Ifm = 2553,
    #[doc(hidden)]
    Ifu = 2554,
    #[doc(hidden)]
    Ify = 2555,
    #[doc(hidden)]
    Igb = 2556,
    #[doc(hidden)]
    Ige = 2557,
    #[doc(hidden)]
    Igg = 2558,
    #[doc(hidden)]
    Igl = 2559,
    #[doc(hidden)]
    Igm = 2560,
    #[doc(hidden)]
    Ign = 2561,
    #[doc(hidden)]
    Igo = 2562,
    #[doc(hidden)]
    Igs = 2563,
    #[doc(hidden)]
    Igw = 2564,
    #[doc(hidden)]
    Ihb = 2565,
    #[doc(hidden)]
    Ihi = 2566,
    #[doc(hidden)]
    Ihp = 2567,
    #[doc(hidden)]
    Ihw = 2568,
    #[doc(hidden)]
    Iii = 2569,
    #[doc(hidden)]
    Iin = 2570,
    #[doc(hidden)]
    Ijc = 2571,
    #[doc(hidden)]
    Ije = 2572,
    #[doc(hidden)]
    Ijj = 2573,
    #[doc(hidden)]
    Ijn = 2574,
    #[doc(hidden)]
    Ijs = 2575,
    #[doc(hidden)]
    Ike = 2576,
    #[doc(hidden)]
    Iki = 2577,
    #[doc(hidden)]
    Ikk = 2578,
    #[doc(hidden)]
    Ikl = 2579,
    #[doc(hidden)]
    Iko = 2580,
    #[doc(hidden)]
    Ikp = 2581,
    #[doc(hidden)]
    Ikr = 2582,
    #[doc(hidden)]
    Iks = 2583,
    #[doc(hidden)]
    Ikt = 2584,
    #[doc(hidden)]
    Iku = 2585,
    #[doc(hidden)]
    Ikv = 2586,
    #[doc(hidden)]
    Ikw = 2587,
    #[doc(hidden)]
    Ikx = 2588,
    #[doc(hidden)]
    Ikz = 2589,
    #[doc(hidden)]
    Ila = 2590,
    #[doc(hidden)]
    Ilb = 2591,
    #[doc(hidden)]
    Ile = 2592,
    #[doc(hidden)]
    Ilg = 2593,
    #[doc(hidden)]
    Ili = 2594,
    #[doc(hidden)]
    Ilk = 2595,
    #[doc(hidden)]
    Ilm = 2596,
    #[doc(hidden)]
    Ilo = 2597,
    #[doc(hidden)]
    Ilp = 2598,
    #[doc(hidden)]
    Ils = 2599,
    #[doc(hidden)]
    Ilu = 2600,
    #[doc(hidden)]
    Ilv = 2601,
    #[doc(hidden)]
    Ima = 2602,
    #[doc(hidden)]
    Imi = 2603,
    #[doc(hidden)]
    Iml = 2604,
    #[doc(hidden)]
    Imn = 2605,
    #[doc(hidden)]
    Imo = 2606,
    #[doc(hidden)]
    Imr = 2607,
    #[doc(hidden)]
    Ims = 2608,
    #[doc(hidden)]
    Imt = 2609,
    #[doc(hidden)]
    Imy = 2610,
    #[doc(hidden)]
    Ina = 2611,
    #[doc(hidden)]
    Inb = 2612,
    #[doc(hidden)]
    Ind = 2613,
    #[doc(hidden)]
    Ing = 2614,
    #[doc(hidden)]
    Inh = 2615,
    #[doc(hidden)]
    Inj = 2616,
    #[doc(hidden)]
    Inl = 2617,
    #[doc(hidden)]
    Inm = 2618,
    #[doc(hidden)]
    Inn = 2619,
    #[doc(hidden)]
    Ino = 2620,
    #[doc(hidden)]
    Inp = 2621,
    #[doc(hidden)]
    Ins = 2622,
    #[doc(hidden)]
    Int = 2623,
    #[doc(hidden)]
    Inz = 2624,
    #[doc(hidden)]
    Ior = 2625,
    #[doc(hidden)]
    Iou = 2626,
    #[doc(hidden)]
    Iow = 2627,
    #[doc(hidden)]
    Ipi = 2628,
    #[doc(hidden)]
    Ipk = 2629,
    #[doc(hidden)]
    Ipo = 2630,
    #[doc(hidden)]
    Iqu = 2631,
    #[doc(hidden)]
    Iqw = 2632,
    #[doc(hidden)]
    Ire = 2633,
    #[doc(hidden)]
    Irh = 2634,
    #[doc(hidden)]
    Iri = 2635,
    #[doc(hidden)]
    Irk = 2636,
    #[doc(hidden)]
    Irn = 2637,
    #[doc(hidden)]
    Irr = 2638,
    #[doc(hidden)]
    Iru = 2639,
    #[doc(hidden)]
    Irx = 2640,
    #[doc(hidden)]
    Iry = 2641,
    #[doc(hidden)]
    Isa = 2642,
    #[doc(hidden)]
    Isc = 2643,
    #[doc(hidden)]
    Isd = 2644,
    #[doc(hidden)]
    Ise = 2645,
    #[doc(hidden)]
    Isg = 2646,
    #[doc(hidden)]
    Ish = 2647,
    #[doc(hidden)]
    Isi = 2648,
    #[doc(hidden)]
    Isk = 2649,
    #[doc(hidden)]
    Isl = 2650,
    #[doc(hidden)]
    Ism = 2651,
    #[doc(hidden)]
    Isn = 2652,
    #[doc(hidden)]
    Iso = 2653,
    #[doc(hidden)]
    Isr = 2654,
    #[doc(hidden)]
    Ist = 2655,
    #[doc(hidden)]
    Isu = 2656,
    #[doc(hidden)]
    Ita = 2657,
    #[doc(hidden)]
    Itb = 2658,
    #[doc(hidden)]
    Itd = 2659,
    #[doc(hidden)]
    Ite = 2660,
    #[doc(hidden)]
    Iti = 2661,
    #[doc(hidden)]
    Itk = 2662,
    #[doc(hidden)]
    Itl = 2663,
    #[doc(hidden)]
    Itm = 2664,
    #[doc(hidden)]
    Ito = 2665,
    #[doc(hidden)]
    Itr = 2666,
    #[doc(hidden)]
    Its = 2667,
    #[doc(hidden)]
    Itt = 2668,
    #[doc(hidden)]
    Itv = 2669,
    #[doc(hidden)]
    Itw = 2670,
    #[doc(hidden)]
    Itx = 2671,
    #[doc(hidden)]
    Ity = 2672,
    #[doc(hidden)]
    Itz = 2673,
    #[doc(hidden)]
    Ium = 2674,
    #[doc(hidden)]
    Ivb = 2675,
    #[doc(hidden)]
    Ivv = 2676,
    #[doc(hidden)]
    Iwk = 2677,
    #[doc(hidden)]
    Iwm = 2678,
    #[doc(hidden)]
    Iwo = 2679,
    #[doc(hidden)]
    Iws = 2680,
    #[doc(hidden)]
    Ixc = 2681,
    #[doc(hidden)]
    Ixl = 2682,
    #[doc(hidden)]
    Iya = 2683,
    #[doc(hidden)]
    Iyo = 2684,
    #[doc(hidden)]
    Iyx = 2685,
    #[doc(hidden)]
    Izh = 2686,
    #[doc(hidden)]
    Izr = 2687,
    #[doc(hidden)]
    Izz = 2688,
    #[doc(hidden)]
    Jaa = 2689,
    #[doc(hidden)]
    Jab = 2690,
    #[doc(hidden)]
    Jac = 2691,
    #[doc(hidden)]
    Jad = 2692,
    #[doc(hidden)]
    Jae = 2693,
    #[doc(hidden)]
    Jaf = 2694,
    #[doc(hidden)]
    Jah = 2695,
    #[doc(hidden)]
    Jaj = 2696,
    #[doc(hidden)]
    Jak = 2697,
    #[doc(hidden)]
    Jal = 2698,
    #[doc(hidden)]
    Jam = 2699,
    #[doc(hidden)]
    Jan = 2700,
    #[doc(hidden)]
    Jao = 2701,
    #[doc(hidden)]
    Jaq = 2702,
    #[doc(hidden)]
    Jas = 2703,
    #[doc(hidden)]
    Jat = 2704,
    #[doc(hidden)]
    Jau = 2705,
    #[doc(hidden)]
    Jav = 2706,
    #[doc(hidden)]
    Jax = 2707,
    #[doc(hidden)]
    Jay = 2708,
    #[doc(hidden)]
    Jaz = 2709,
    #[doc(hidden)]
    Jbe = 2710,
    #[doc(hidden)]
    Jbi = 2711,
    #[doc(hidden)]
    Jbj = 2712,
    #[doc(hidden)]
    Jbk = 2713,
    #[doc(hidden)]
    Jbm = 2714,
    #[doc(hidden)]
    Jbn = 2715,
    #[doc(hidden)]
    Jbo = 2716,
    #[doc(hidden)]
    Jbr = 2717,
    #[doc(hidden)]
    Jbt = 2718,
    #[doc(hidden)]
    Jbu = 2719,
    #[doc(hidden)]
    Jbw = 2720,
    #[doc(hidden)]
    Jcs = 2721,
    #[doc(hidden)]
    Jct = 2722,
    #[doc(hidden)]
    Jda = 2723,
    #[doc(hidden)]
    Jdg = 2724,
    #[doc(hidden)]
    Jdt = 2725,
    #[doc(hidden)]
    Jeb = 2726,
    #[doc(hidden)]
    Jee = 2727,
    #[doc(hidden)]
    Jeh = 2728,
    #[doc(hidden)]
    Jei = 2729,
    #[doc(hidden)]
    Jek = 2730,
    #[doc(hidden)]
    Jel = 2731,
    #[doc(hidden)]
    Jen = 2732,
    #[doc(hidden)]
    Jer = 2733,
    #[doc(hidden)]
    Jet = 2734,
    #[doc(hidden)]
    Jeu = 2735,
    #[doc(hidden)]
    Jgb = 2736,
    #[doc(hidden)]
    Jge = 2737,
    #[doc(hidden)]
    Jgk = 2738,
    #[doc(hidden)]
    Jgo = 2739,
    #[doc(hidden)]
    Jhi = 2740,
    #[doc(hidden)]
    Jhs = 2741,
    #[doc(hidden)]
    Jia = 2742,
    #[doc(hidden)]
    Jib = 2743,
    #[doc(hidden)]
    Jic = 2744,
    #[doc(hidden)]
    Jid = 2745,
    #[doc(hidden)]
    Jie = 2746,
    #[doc(hidden)]
    Jig = 2747,
    #[doc(hidden)]
    Jih = 2748,
    #[doc(hidden)]
    Jii = 2749,
    #[doc(hidden)]
    Jil = 2750,
    #[doc(hidden)]
    Jim = 2751,
    #[doc(hidden)]
    Jio = 2752,
    #[doc(hidden)]
    Jiq = 2753,
    #[doc(hidden)]
    Jit = 2754,
    #[doc(hidden)]
    Jiu = 2755,
    #[doc(hidden)]
    Jiv = 2756,
    #[doc(hidden)]
    Jiy = 2757,
    #[doc(hidden)]
    Jje = 2758,
    #[doc(hidden)]
    Jjr = 2759,
    #[doc(hidden)]
    Jka = 2760,
    #[doc(hidden)]
    Jkm = 2761,
    #[doc(hidden)]
    Jko = 2762,
    #[doc(hidden)]
    Jkp = 2763,
    #[doc(hidden)]
    Jkr = 2764,
    #[doc(hidden)]
    Jks = 2765,
    #[doc(hidden)]
    Jku = 2766,
    #[doc(hidden)]
    Jle = 2767,
    #[doc(hidden)]
    Jls = 2768,
    #[doc(hidden)]
    Jma = 2769,
    #[doc(hidden)]
    Jmb = 2770,
    #[doc(hidden)]
    Jmc = 2771,
    #[doc(hidden)]
    Jmd = 2772,
    #[doc(hidden)]
    Jmi = 2773,
    #[doc(hidden)]
    Jml = 2774,
    #[doc(hidden)]
    Jmn = 2775,
    #[doc(hidden)]
    Jmr = 2776,
    #[doc(hidden)]
    Jms = 2777,
    #[doc(hidden)]
    Jmw = 2778,
    #[doc(hidden)]
    Jmx = 2779,
    #[doc(hidden)]
    Jna = 2780,
    #[doc(hidden)]
    Jnd = 2781,
    #[doc(hidden)]
    Jng = 2782,
    #[doc(hidden)]
    Jni = 2783,
    #[doc(hidden)]
    Jnj = 2784,
    #[doc(hidden)]
    Jnl = 2785,
    #[doc(hidden)]
    Jns = 2786,
    #[doc(hidden)]
    Job = 2787,
    #[doc(hidden)]
    Jod = 2788,
    #[doc(hidden)]
    Jog = 2789,
    #[doc(hidden)]
    Jor = 2790,
    #[doc(hidden)]
    Jos = 2791,
    #[doc(hidden)]
    Jow = 2792,
    #[doc(hidden)]
    Jpa = 2793,
    #[doc(hidden)]
    Jpn = 2794,
    #[doc(hidden)]
    Jpr = 2795,
    #[doc(hidden)]
    Jqr = 2796,
    #[doc(hidden)]
    Jra = 2797,
    #[doc(hidden)]
    Jrb = 2798,
    #[doc(hidden)]
    Jrr = 2799,
    #[doc(hidden)]
    Jrt = 2800,
    #[doc(hidden)]
    Jru = 2801,
    #[doc(hidden)]
    Jsl = 2802,
    #[doc(hidden)]
    Jua = 2803,
    #[doc(hidden)]
    Jub = 2804,
    #[doc(hidden)]
    Juc = 2805,
    #[doc(hidden)]
    Jud = 2806,
    #[doc(hidden)]
    Juh = 2807,
    #[doc(hidden)]
    Jui = 2808,
    #[doc(hidden)]
    Juk = 2809,
    #[doc(hidden)]
    Jul = 2810,
    #[doc(hidden)]
    Jum = 2811,
    #[doc(hidden)]
    Jun = 2812,
    #[doc(hidden)]
    Juo = 2813,
    #[doc(hidden)]
    Jup = 2814,
    #[doc(hidden)]
    Jur = 2815,
    #[doc(hidden)]
    Jus = 2816,
    #[doc(hidden)]
    Jut = 2817,
    #[doc(hidden)]
    Juu = 2818,
    #[doc(hidden)]
    Juw = 2819,
    #[doc(hidden)]
    Juy = 2820,
    #[doc(hidden)]
    Jvd = 2821,
    #[doc(hidden)]
    Jvn = 2822,
    #[doc(hidden)]
    Jwi = 2823,
    #[doc(hidden)]
    Jya = 2824,
    #[doc(hidden)]
    Jye = 2825,
    #[doc(hidden)]
    Jyy = 2826,
    #[doc(hidden)]
    Kaa = 2827,
    #[doc(hidden)]
    Kab = 2828,
    #[doc(hidden)]
    Kac = 2829,
    #[doc(hidden)]
    Kad = 2830,
    #[doc(hidden)]
    Kae = 2831,
    #[doc(hidden)]
    Kaf = 2832,
    #[doc(hidden)]
    Kag = 2833,
    #[doc(hidden)]
    Kah = 2834,
    #[doc(hidden)]
    Kai = 2835,
    #[doc(hidden)]
    Kaj = 2836,
    #[doc(hidden)]
    Kak = 2837,
    #[doc(hidden)]
    Kal = 2838,
    #[doc(hidden)]
    Kam = 2839,
    #[doc(hidden)]
    Kan = 2840,
    #[doc(hidden)]
    Kao = 2841,
    #[doc(hidden)]
    Kap = 2842,
    #[doc(hidden)]
    Kaq = 2843,
    #[doc(hidden)]
    Kas = 2844,
    #[doc(hidden)]
    Kat = 2845,
    #[doc(hidden)]
    Kau = 2846,
    #[doc(hidden)]
    Kav = 2847,
    #[doc(hidden)]
    Kaw = 2848,
    #[doc(hidden)]
    Kax = 2849,
    #[doc(hidden)]
    Kay = 2850,
    #[doc(hidden)]
    Kaz = 2851,
    #[doc(hidden)]
    Kba = 2852,
    #[doc(hidden)]
    Kbb = 2853,
    #[doc(hidden)]
    Kbc = 2854,
    #[doc(hidden)]
    Kbd = 2855,
    #[doc(hidden)]
    Kbe = 2856,
    #[doc(hidden)]
    Kbg = 2857,
    #[doc(hidden)]
    Kbh = 2858,
    #[doc(hidden)]
    Kbi = 2859,
    #[doc(hidden)]
    Kbj = 2860,
    #[doc(hidden)]
    Kbk = 2861,
    #[doc(hidden)]
    Kbl = 2862,
    #[doc(hidden)]
    Kbm = 2863,
    #[doc(hidden)]
    Kbn = 2864,
    #[doc(hidden)]
    Kbo = 2865,
    #[doc(hidden)]
    Kbp = 2866,
    #[doc(hidden)]
    Kbq = 2867,
    #[doc(hidden)]
    Kbr = 2868,
    #[doc(hidden)]
    Kbs = 2869,
    #[doc(hidden)]
    Kbt = 2870,
    #[doc(hidden)]
    Kbu = 2871,
    #[doc(hidden)]
    Kbv = 2872,
    #[doc(hidden)]
    Kbw = 2873,
    #[doc(hidden)]
    Kbx = 2874,
    #[doc(hidden)]
    Kby = 2875,
    #[doc(hidden)]
    Kbz = 2876,
    #[doc(hidden)]
    Kca = 2877,
    #[doc(hidden)]
    Kcb = 2878,
    #[doc(hidden)]
    Kcc = 2879,
    #[doc(hidden)]
    Kcd = 2880,
    #[doc(hidden)]
    Kce = 2881,
    #[doc(hidden)]
    Kcf = 2882,
    #[doc(hidden)]
    Kcg = 2883,
    #[doc(hidden)]
    Kch = 2884,
    #[doc(hidden)]
    Kci = 2885,
    #[doc(hidden)]
    Kcj = 2886,
    #[doc(hidden)]
    Kck = 2887,
    #[doc(hidden)]
    Kcl = 2888,
    #[doc(hidden)]
    Kcm = 2889,
    #[doc(hidden)]
    Kcn = 2890,
    #[doc(hidden)]
    Kco = 2891,
    #[doc(hidden)]
    Kcp = 2892,
    #[doc(hidden)]
    Kcq = 2893,
    #[doc(hidden)]
    Kcr = 2894,
    #[doc(hidden)]
    Kcs = 2895,
    #[doc(hidden)]
    Kct = 2896,
    #[doc(hidden)]
    Kcu = 2897,
    #[doc(hidden)]
    Kcv = 2898,
    #[doc(hidden)]
    Kcw = 2899,
    #[doc(hidden)]
    Kcx = 2900,
    #[doc(hidden)]
    Kcy = 2901,
    #[doc(hidden)]
    Kcz = 2902,
    #[doc(hidden)]
    Kda = 2903,
    #[doc(hidden)]
    Kdc = 2904,
    #[doc(hidden)]
    Kdd = 2905,
    #[doc(hidden)]
    Kde = 2906,
    #[doc(hidden)]
    Kdf = 2907,
    #[doc(hidden)]
    Kdg = 2908,
    #[doc(hidden)]
    Kdh = 2909,
    #[doc(hidden)]
    Kdi = 2910,
    #[doc(hidden)]
    Kdj = 2911,
    #[doc(hidden)]
    Kdk = 2912,
    #[doc(hidden)]
    Kdl = 2913,
    #[doc(hidden)]
    Kdm = 2914,
    #[doc(hidden)]
    Kdn = 2915,
    #[doc(hidden)]
    Kdp = 2916,
    #[doc(hidden)]
    Kdq = 2917,
    #[doc(hidden)]
    Kdr = 2918,
    #[doc(hidden)]
    Kdt = 2919,
    #[doc(hidden)]
    Kdu = 2920,
    #[doc(hidden)]
    Kdw = 2921,
    #[doc(hidden)]
    Kdx = 2922,
    #[doc(hidden)]
    Kdy = 2923,
    #[doc(hidden)]
    Kdz = 2924,
    #[doc(hidden)]
    Kea = 2925,
    #[doc(hidden)]
    Keb = 2926,
    #[doc(hidden)]
    Kec = 2927,
    #[doc(hidden)]
    Ked = 2928,
    #[doc(hidden)]
    Kee = 2929,
    #[doc(hidden)]
    Kef = 2930,
    #[doc(hidden)]
    Keg = 2931,
    #[doc(hidden)]
    Keh = 2932,
    #[doc(hidden)]
    Kei = 2933,
    #[doc(hidden)]
    Kej = 2934,
    #[doc(hidden)]
    Kek = 2935,
    #[doc(hidden)]
    Kel = 2936,
    #[doc(hidden)]
    Kem = 2937,
    #[doc(hidden)]
    Ken = 2938,
    #[doc(hidden)]
    Keo = 2939,
    #[doc(hidden)]
    Kep = 2940,
    #[doc(hidden)]
    Keq = 2941,
    #[doc(hidden)]
    Ker = 2942,
    #[doc(hidden)]
    Kes = 2943,
    #[doc(hidden)]
    Ket = 2944,
    #[doc(hidden)]
    Keu = 2945,
    #[doc(hidden)]
    Kev = 2946,
    #[doc(hidden)]
    Kew = 2947,
    #[doc(hidden)]
    Kex = 2948,
    #[doc(hidden)]
    Key = 2949,
    #[doc(hidden)]
    Kez = 2950,
    #[doc(hidden)]
    Kfa = 2951,
    #[doc(hidden)]
    Kfb = 2952,
    #[doc(hidden)]
    Kfc = 2953,
    #[doc(hidden)]
    Kfd = 2954,
    #[doc(hidden)]
    Kfe = 2955,
    #[doc(hidden)]
    Kff = 2956,
    #[doc(hidden)]
    Kfg = 2957,
    #[doc(hidden)]
    Kfh = 2958,
    #[doc(hidden)]
    Kfi = 2959,
    #[doc(hidden)]
    Kfj = 2960,
    #[doc(hidden)]
    Kfk = 2961,
    #[doc(hidden)]
    Kfl = 2962,
    #[doc(hidden)]
    Kfm = 2963,
    #[doc(hidden)]
    Kfn = 2964,
    #[doc(hidden)]
    Kfo = 2965,
    #[doc(hidden)]
    Kfp = 2966,
    #[doc(hidden)]
    Kfq = 2967,
    #[doc(hidden)]
    Kfr = 2968,
    #[doc(hidden)]
    Kfs = 2969,
    #[doc(hidden)]
    Kft = 2970,
    #[doc(hidden)]
    Kfu = 2971,
    #[doc(hidden)]
    Kfv = 2972,
    #[doc(hidden)]
    Kfw = 2973,
    #[doc(hidden)]
    Kfx = 2974,
    #[doc(hidden)]
    Kfy = 2975,
    #[doc(hidden)]
    Kfz = 2976,
    #[doc(hidden)]
    Kga = 2977,
    #[doc(hidden)]
    Kgb = 2978,
    #[doc(hidden)]
    Kge = 2979,
    #[doc(hidden)]
    Kgf = 2980,
    #[doc(hidden)]
    Kgg = 2981,
    #[doc(hidden)]
    Kgi = 2982,
    #[doc(hidden)]
    Kgj = 2983,
    #[doc(hidden)]
    Kgk = 2984,
    #[doc(hidden)]
    Kgl = 2985,
    #[doc(hidden)]
    Kgm = 2986,
    #[doc(hidden)]
    Kgn = 2987,
    #[doc(hidden)]
    Kgo = 2988,
    #[doc(hidden)]
    Kgp = 2989,
    #[doc(hidden)]
    Kgq = 2990,
    #[doc(hidden)]
    Kgr = 2991,
    #[doc(hidden)]
    Kgs = 2992,
    #[doc(hidden)]
    Kgt = 2993,
    #[doc(hidden)]
    Kgu = 2994,
    #[doc(hidden)]
    Kgv = 2995,
    #[doc(hidden)]
    Kgw = 2996,
    #[doc(hidden)]
    Kgx = 2997,
    #[doc(hidden)]
    Kgy = 2998,
    #[doc(hidden)]
    Kha = 2999,
    #[doc(hidden)]
    Khb = 3000,
    #[doc(hidden)]
    Khc = 3001,
    #[doc(hidden)]
    Khd = 3002,
    #[doc(hidden)]
    Khe = 3003,
    #[doc(hidden)]
    Khf = 3004,
    #[doc(hidden)]
    Khg = 3005,
    #[doc(hidden)]
    Khh = 3006,
    #[doc(hidden)]
    Khj = 3007,
    #[doc(hidden)]
    Khk = 3008,
    #[doc(hidden)]
    Khl = 3009,
    #[doc(hidden)]
    Khm = 3010,
    #[doc(hidden)]
    Khn = 3011,
    #[doc(hidden)]
    Kho = 3012,
    #[doc(hidden)]
    Khp = 3013,
    #[doc(hidden)]
    Khq = 3014,
    #[doc(hidden)]
    Khr = 3015,
    #[doc(hidden)]
    Khs = 3016,
    #[doc(hidden)]
    Kht = 3017,
    #[doc(hidden)]
    Khu = 3018,
    #[doc(hidden)]
    Khv = 3019,
    #[doc(hidden)]
    Khw = 3020,
    #[doc(hidden)]
    Khx = 3021,
    #[doc(hidden)]
    Khy = 3022,
    #[doc(hidden)]
    Khz = 3023,
    #[doc(hidden)]
    Kia = 3024,
    #[doc(hidden)]
    Kib = 3025,
    #[doc(hidden)]
    Kic = 3026,
    #[doc(hidden)]
    Kid = 3027,
    #[doc(hidden)]
    Kie = 3028,
    #[doc(hidden)]
    Kif = 3029,
    #[doc(hidden)]
    Kig = 3030,
    #[doc(hidden)]
    Kih = 3031,
    #[doc(hidden)]
    Kii = 3032,
    #[doc(hidden)]
    Kij = 3033,
    #[doc(hidden)]
    Kik = 3034,
    #[doc(hidden)]
    Kil = 3035,
    #[doc(hidden)]
    Kim = 3036,
    #[doc(hidden)]
    Kin = 3037,
    #[doc(hidden)]
    Kio = 3038,
    #[doc(hidden)]
    Kip = 3039,
    #[doc(hidden)]
    Kiq = 3040,
    #[doc(hidden)]
    Kir = 3041,
    #[doc(hidden)]
    Kis = 3042,
    #[doc(hidden)]
    Kit = 3043,
    #[doc(hidden)]
    Kiu = 3044,
    #[doc(hidden)]
    Kiv = 3045,
    #[doc(hidden)]
    Kiw = 3046,
    #[doc(hidden)]
    Kix = 3047,
    #[doc(hidden)]
    Kiy = 3048,
    #[doc(hidden)]
    Kiz = 3049,
    #[doc(hidden)]
    Kja = 3050,
    #[doc(hidden)]
    Kjb = 3051,
    #[doc(hidden)]
    Kjc = 3052,
    #[doc(hidden)]
    Kjd = 3053,
    #[doc(hidden)]
    Kje = 3054,
    #[doc(hidden)]
    Kjg = 3055,
    #[doc(hidden)]
    Kjh = 3056,
    #[doc(hidden)]
    Kji = 3057,
    #[doc(hidden)]
    Kjj = 3058,
    #[doc(hidden)]
    Kjk = 3059,
    #[doc(hidden)]
    Kjl = 3060,
    #[doc(hidden)]
    Kjm = 3061,
    #[doc(hidden)]
    Kjn = 3062,
    #[doc(hidden)]
    Kjo = 3063,
    #[doc(hidden)]
    Kjp = 3064,
    #[doc(hidden)]
    Kjq = 3065,
    #[doc(hidden)]
    Kjr = 3066,
    #[doc(hidden)]
    Kjs = 3067,
    #[doc(hidden)]
    Kjt = 3068,
    #[doc(hidden)]
    Kju = 3069,
    #[doc(hidden)]
    Kjv = 3070,
    #[doc(hidden)]
    Kjx = 3071,
    #[doc(hidden)]
    Kjy = 3072,
    #[doc(hidden)]
    Kjz = 3073,
    #[doc(hidden)]
    Kka = 3074,
    #[doc(hidden)]
    Kkb = 3075,
    #[doc(hidden)]
    Kkc = 3076,
    #[doc(hidden)]
    Kkd = 3077,
    #[doc(hidden)]
    Kke = 3078,
    #[doc(hidden)]
    Kkf = 3079,
    #[doc(hidden)]
    Kkg = 3080,
    #[doc(hidden)]
    Kkh = 3081,
    #[doc(hidden)]
    Kki = 3082,
    #[doc(hidden)]
    Kkj = 3083,
    #[doc(hidden)]
    Kkk = 3084,
    #[doc(hidden)]
    Kkl = 3085,
    #[doc(hidden)]
    Kkm = 3086,
    #[doc(hidden)]
    Kkn = 3087,
    #[doc(hidden)]
    Kko = 3088,
    #[doc(hidden)]
    Kkp = 3089,
    #[doc(hidden)]
    Kkq = 3090,
    #[doc(hidden)]
    Kkr = 3091,
    #[doc(hidden)]
    Kks = 3092,
    #[doc(hidden)]
    Kkt = 3093,
    #[doc(hidden)]
    Kku = 3094,
    #[doc(hidden)]
    Kkv = 3095,
    #[doc(hidden)]
    Kkw = 3096,
    #[doc(hidden)]
    Kkx = 3097,
    #[doc(hidden)]
    Kky = 3098,
    #[doc(hidden)]
    Kkz = 3099,
    #[doc(hidden)]
    Kla = 3100,
    #[doc(hidden)]
    Klb = 3101,
    #[doc(hidden)]
    Klc = 3102,
    #[doc(hidden)]
    Kld = 3103,
    #[doc(hidden)]
    Kle = 3104,
    #[doc(hidden)]
    Klf = 3105,
    #[doc(hidden)]
    Klg = 3106,
    #[doc(hidden)]
    Klh = 3107,
    #[doc(hidden)]
    Kli = 3108,
    #[doc(hidden)]
    Klj = 3109,
    #[doc(hidden)]
    Klk = 3110,
    #[doc(hidden)]
    Kll = 3111,
    #[doc(hidden)]
    Klm = 3112,
    #[doc(hidden)]
    Kln = 3113,
    #[doc(hidden)]
    Klo = 3114,
    #[doc(hidden)]
    Klp = 3115,
    #[doc(hidden)]
    Klq = 3116,
    #[doc(hidden)]
    Klr = 3117,
    #[doc(hidden)]
    Kls = 3118,
    #[doc(hidden)]
    Klt = 3119,
    #[doc(hidden)]
    Klu = 3120,
    #[doc(hidden)]
    Klv = 3121,
    #[doc(hidden)]
    Klw = 3122,
    #[doc(hidden)]
    Klx = 3123,
    #[doc(hidden)]
    Kly = 3124,
    #[doc(hidden)]
    Klz = 3125,
    #[doc(hidden)]
    Kma = 3126,
    #[doc(hidden)]
    Kmb = 3127,
    #[doc(hidden)]
    Kmc = 3128,
    #[doc(hidden)]
    Kmd = 3129,
    #[doc(hidden)]
    Kme = 3130,
    #[doc(hidden)]
    Kmf = 3131,
    #[doc(hidden)]
    Kmg = 3132,
    #[doc(hidden)]
    Kmh = 3133,
    #[doc(hidden)]
    Kmi = 3134,
    #[doc(hidden)]
    Kmj = 3135,
    #[doc(hidden)]
    Kmk = 3136,
    #[doc(hidden)]
    Kml = 3137,
    #[doc(hidden)]
    Kmm = 3138,
    #[doc(hidden)]
    Kmn = 3139,
    #[doc(hidden)]
    Kmo = 3140,
    #[doc(hidden)]
    Kmp = 3141,
    #[doc(hidden)]
    Kmq = 3142,
    #[doc(hidden)]
    Kmr = 3143,
    #[doc(hidden)]
    Kms = 3144,
    #[doc(hidden)]
    Kmt = 3145,
    #[doc(hidden)]
    Kmu = 3146,
    #[doc(hidden)]
    Kmv = 3147,
    #[doc(hidden)]
    Kmw = 3148,
    #[doc(hidden)]
    Kmx = 3149,
    #[doc(hidden)]
    Kmy = 3150,
    #[doc(hidden)]
    Kmz = 3151,
    #[doc(hidden)]
    Kna = 3152,
    #[doc(hidden)]
    Knb = 3153,
    #[doc(hidden)]
    Knc = 3154,
    #[doc(hidden)]
    Knd = 3155,
    #[doc(hidden)]
    Kne = 3156,
    #[doc(hidden)]
    Knf = 3157,
    #[doc(hidden)]
    Kng = 3158,
    #[doc(hidden)]
    Kni = 3159,
    #[doc(hidden)]
    Knj = 3160,
    #[doc(hidden)]
    Knk = 3161,
    #[doc(hidden)]
    Knl = 3162,
    #[doc(hidden)]
    Knm = 3163,
    #[doc(hidden)]
    Knn = 3164,
    #[doc(hidden)]
    Kno = 3165,
    #[doc(hidden)]
    Knp = 3166,
    #[doc(hidden)]
    Knq = 3167,
    #[doc(hidden)]
    Knr = 3168,
    #[doc(hidden)]
    Kns = 3169,
    #[doc(hidden)]
    Knt = 3170,
    #[doc(hidden)]
    Knu = 3171,
    #[doc(hidden)]
    Knv = 3172,
    #[doc(hidden)]
    Knw = 3173,
    #[doc(hidden)]
    Knx = 3174,
    #[doc(hidden)]
    Kny = 3175,
    #[doc(hidden)]
    Knz = 3176,
    #[doc(hidden)]
    Koa = 3177,
    #[doc(hidden)]
    Koc = 3178,
    #[doc(hidden)]
    Kod = 3179,
    #[doc(hidden)]
    Koe = 3180,
    #[doc(hidden)]
    Kof = 3181,
    #[doc(hidden)]
    Kog = 3182,
    #[doc(hidden)]
    Koh = 3183,
    #[doc(hidden)]
    Koi = 3184,
    #[doc(hidden)]
    Kok = 3185,
    #[doc(hidden)]
    Kol = 3186,
    #[doc(hidden)]
    Kom = 3187,
    #[doc(hidden)]
    Kon = 3188,
    #[doc(hidden)]
    Koo = 3189,
    #[doc(hidden)]
    Kop = 3190,
    #[doc(hidden)]
    Koq = 3191,
    #[doc(hidden)]
    Kor = 3192,
    #[doc(hidden)]
    Kos = 3193,
    #[doc(hidden)]
    Kot = 3194,
    #[doc(hidden)]
    Kou = 3195,
    #[doc(hidden)]
    Kov = 3196,
    #[doc(hidden)]
    Kow = 3197,
    #[doc(hidden)]
    Koy = 3198,
    #[doc(hidden)]
    Koz = 3199,
    #[doc(hidden)]
    Kpa = 3200,
    #[doc(hidden)]
    Kpb = 3201,
    #[doc(hidden)]
    Kpc = 3202,
    #[doc(hidden)]
    Kpd = 3203,
    #[doc(hidden)]
    Kpe = 3204,
    #[doc(hidden)]
    Kpf = 3205,
    #[doc(hidden)]
    Kpg = 3206,
    #[doc(hidden)]
    Kph = 3207,
    #[doc(hidden)]
    Kpi = 3208,
    #[doc(hidden)]
    Kpj = 3209,
    #[doc(hidden)]
    Kpk = 3210,
    #[doc(hidden)]
    Kpl = 3211,
    #[doc(hidden)]
    Kpm = 3212,
    #[doc(hidden)]
    Kpn = 3213,
    #[doc(hidden)]
    Kpo = 3214,
    #[doc(hidden)]
    Kpq = 3215,
    #[doc(hidden)]
    Kpr = 3216,
    #[doc(hidden)]
    Kps = 3217,
    #[doc(hidden)]
    Kpt = 3218,
    #[doc(hidden)]
    Kpu = 3219,
    #[doc(hidden)]
    Kpv = 3220,
    #[doc(hidden)]
    Kpw = 3221,
    #[doc(hidden)]
    Kpx = 3222,
    #[doc(hidden)]
    Kpy = 3223,
    #[doc(hidden)]
    Kpz = 3224,
    #[doc(hidden)]
    Kqa = 3225,
    #[doc(hidden)]
    Kqb = 3226,
    #[doc(hidden)]
    Kqc = 3227,
    #[doc(hidden)]
    Kqd = 3228,
    #[doc(hidden)]
    Kqe = 3229,
    #[doc(hidden)]
    Kqf = 3230,
    #[doc(hidden)]
    Kqg = 3231,
    #[doc(hidden)]
    Kqh = 3232,
    #[doc(hidden)]
    Kqi = 3233,
    #[doc(hidden)]
    Kqj = 3234,
    #[doc(hidden)]
    Kqk = 3235,
    #[doc(hidden)]
    Kql = 3236,
    #[doc(hidden)]
    Kqm = 3237,
    #[doc(hidden)]
    Kqn = 3238,
    #[doc(hidden)]
    Kqo = 3239,
    #[doc(hidden)]
    Kqp = 3240,
    #[doc(hidden)]
    Kqq = 3241,
    #[doc(hidden)]
    Kqr = 3242,
    #[doc(hidden)]
    Kqs = 3243,
    #[doc(hidden)]
    Kqt = 3244,
    #[doc(hidden)]
    Kqu = 3245,
    #[doc(hidden)]
    Kqv = 3246,
    #[doc(hidden)]
    Kqw = 3247,
    #[doc(hidden)]
    Kqx = 3248,
    #[doc(hidden)]
    Kqy = 3249,
    #[doc(hidden)]
    Kqz = 3250,
    #[doc(hidden)]
    Kra = 3251,
    #[doc(hidden)]
    Krb = 3252,
    #[doc(hidden)]
    Krc = 3253,
    #[doc(hidden)]
    Krd = 3254,
    #[doc(hidden)]
    Kre = 3255,
    #[doc(hidden)]
    Krf = 3256,
    #[doc(hidden)]
    Krh = 3257,
    #[doc(hidden)]
    Kri = 3258,
    #[doc(hidden)]
    Krj = 3259,
    #[doc(hidden)]
    Krk = 3260,
    #[doc(hidden)]
    Krl = 3261,
    #[doc(hidden)]
    Krn = 3262,
    #[doc(hidden)]
    Krp = 3263,
    #[doc(hidden)]
    Krr = 3264,
    #[doc(hidden)]
    Krs = 3265,
    #[doc(hidden)]
    Krt = 3266,
    #[doc(hidden)]
    Kru = 3267,
    #[doc(hidden)]
    Krv = 3268,
    #[doc(hidden)]
    Krw = 3269,
    #[doc(hidden)]
    Krx = 3270,
    #[doc(hidden)]
    Kry = 3271,
    #[doc(hidden)]
    Krz = 3272,
    #[doc(hidden)]
    Ksa = 3273,
    #[doc(hidden)]
    Ksb = 3274,
    #[doc(hidden)]
    Ksc = 3275,
    #[doc(hidden)]
    Ksd = 3276,
    #[doc(hidden)]
    Kse = 3277,
    #[doc(hidden)]
    Ksf = 3278,
    #[doc(hidden)]
    Ksg = 3279,
    #[doc(hidden)]
    Ksh = 3280,
    #[doc(hidden)]
    Ksi = 3281,
    #[doc(hidden)]
    Ksj = 3282,
    #[doc(hidden)]
    Ksk = 3283,
    #[doc(hidden)]
    Ksl = 3284,
    #[doc(hidden)]
    Ksm = 3285,
    #[doc(hidden)]
    Ksn = 3286,
    #[doc(hidden)]
    Kso = 3287,
    #[doc(hidden)]
    Ksp = 3288,
    #[doc(hidden)]
    Ksq = 3289,
    #[doc(hidden)]
    Ksr = 3290,
    #[doc(hidden)]
    Kss = 3291,
    #[doc(hidden)]
    Kst = 3292,
    #[doc(hidden)]
    Ksu = 3293,
    #[doc(hidden)]
    Ksv = 3294,
    #[doc(hidden)]
    Ksw = 3295,
    #[doc(hidden)]
    Ksx = 3296,
    #[doc(hidden)]
    Ksy = 3297,
    #[doc(hidden)]
    Ksz = 3298,
    #[doc(hidden)]
    Kta = 3299,
    #[doc(hidden)]
    Ktb = 3300,
    #[doc(hidden)]
    Ktc = 3301,
    #[doc(hidden)]
    Ktd = 3302,
    #[doc(hidden)]
    Kte = 3303,
    #[doc(hidden)]
    Ktf = 3304,
    #[doc(hidden)]
    Ktg = 3305,
    #[doc(hidden)]
    Kth = 3306,
    #[doc(hidden)]
    Kti = 3307,
    #[doc(hidden)]
    Ktj = 3308,
    #[doc(hidden)]
    Ktk = 3309,
    #[doc(hidden)]
    Ktl = 3310,
    #[doc(hidden)]
    Ktm = 3311,
    #[doc(hidden)]
    Ktn = 3312,
    #[doc(hidden)]
    Kto = 3313,
    #[doc(hidden)]
    Ktp = 3314,
    #[doc(hidden)]
    Ktq = 3315,
    #[doc(hidden)]
    Kts = 3316,
    #[doc(hidden)]
    Ktt = 3317,
    #[doc(hidden)]
    Ktu = 3318,
    #[doc(hidden)]
    Ktv = 3319,
    #[doc(hidden)]
    Ktw = 3320,
    #[doc(hidden)]
    Ktx = 3321,
    #[doc(hidden)]
    Kty = 3322,
    #[doc(hidden)]
    Ktz = 3323,
    #[doc(hidden)]
    Kua = 3324,
    #[doc(hidden)]
    Kub = 3325,
    #[doc(hidden)]
    Kuc = 3326,
    #[doc(hidden)]
    Kud = 3327,
    #[doc(hidden)]
    Kue = 3328,
    #[doc(hidden)]
    Kuf = 3329,
    #[doc(hidden)]
    Kug = 3330,
    #[doc(hidden)]
    Kuh = 3331,
    #[doc(hidden)]
    Kui = 3332,
    #[doc(hidden)]
    Kuj = 3333,
    #[doc(hidden)]
    Kuk = 3334,
    #[doc(hidden)]
    Kul = 3335,
    #[doc(hidden)]
    Kum = 3336,
    #[doc(hidden)]
    Kun = 3337,
    #[doc(hidden)]
    Kuo = 3338,
    #[doc(hidden)]
    Kup = 3339,
    #[doc(hidden)]
    Kuq = 3340,
    #[doc(hidden)]
    Kur = 3341,
    #[doc(hidden)]
    Kus = 3342,
    #[doc(hidden)]
    Kut = 3343,
    #[doc(hidden)]
    Kuu = 3344,
    #[doc(hidden)]
    Kuv = 3345,
    #[doc(hidden)]
    Kuw = 3346,
    #[doc(hidden)]
    Kux = 3347,
    #[doc(hidden)]
    Kuy = 3348,
    #[doc(hidden)]
    Kuz = 3349,
    #[doc(hidden)]
    Kva = 3350,
    #[doc(hidden)]
    Kvb = 3351,
    #[doc(hidden)]
    Kvc = 3352,
    #[doc(hidden)]
    Kvd = 3353,
    #[doc(hidden)]
    Kve = 3354,
    #[doc(hidden)]
    Kvf = 3355,
    #[doc(hidden)]
    Kvg = 3356,
    #[doc(hidden)]
    Kvh = 3357,
    #[doc(hidden)]
    Kvi = 3358,
    #[doc(hidden)]
    Kvj = 3359,
    #[doc(hidden)]
    Kvk = 3360,
    #[doc(hidden)]
    Kvl = 3361,
    #[doc(hidden)]
    Kvm = 3362,
    #[doc(hidden)]
    Kvn = 3363,
    #[doc(hidden)]
    Kvo = 3364,
    #[doc(hidden)]
    Kvp = 3365,
    #[doc(hidden)]
    Kvq = 3366,
    #[doc(hidden)]
    Kvr = 3367,
    #[doc(hidden)]
    Kvt = 3368,
    #[doc(hidden)]
    Kvu = 3369,
    #[doc(hidden)]
    Kvv = 3370,
    #[doc(hidden)]
    Kvw = 3371,
    #[doc(hidden)]
    Kvx = 3372,
    #[doc(hidden)]
    Kvy = 3373,
    #[doc(hidden)]
    Kvz = 3374,
    #[doc(hidden)]
    Kwa = 3375,
    #[doc(hidden)]
    Kwb = 3376,
    #[doc(hidden)]
    Kwc = 3377,
    #[doc(hidden)]
    Kwd = 3378,
    #[doc(hidden)]
    Kwe = 3379,
    #[doc(hidden)]
    Kwf = 3380,
    #[doc(hidden)]
    Kwg = 3381,
    #[doc(hidden)]
    Kwh = 3382,
    #[doc(hidden)]
    Kwi = 3383,
    #[doc(hidden)]
    Kwj = 3384,
    #[doc(hidden)]
    Kwk = 3385,
    #[doc(hidden)]
    Kwl = 3386,
    #[doc(hidden)]
    Kwm = 3387,
    #[doc(hidden)]
    Kwn = 3388,
    #[doc(hidden)]
    Kwo = 3389,
    #[doc(hidden)]
    Kwp = 3390,
    #[doc(hidden)]
    Kwr = 3391,
    #[doc(hidden)]
    Kws = 3392,
    #[doc(hidden)]
    Kwt = 3393,
    #[doc(hidden)]
    Kwu = 3394,
    #[doc(hidden)]
    Kwv = 3395,
    #[doc(hidden)]
    Kww = 3396,
    #[doc(hidden)]
    Kwx = 3397,
    #[doc(hidden)]
    Kwy = 3398,
    #[doc(hidden)]
    Kwz = 3399,
    #[doc(hidden)]
    Kxa = 3400,
    #[doc(hidden)]
    Kxb = 3401,
    #[doc(hidden)]
    Kxc = 3402,
    #[doc(hidden)]
    Kxd = 3403,
    #[doc(hidden)]
    Kxf = 3404,
    #[doc(hidden)]
    Kxh = 3405,
    #[doc(hidden)]
    Kxi = 3406,
    #[doc(hidden)]
    Kxj = 3407,
    #[doc(hidden)]
    Kxk = 3408,
    #[doc(hidden)]
    Kxm = 3409,
    #[doc(hidden)]
    Kxn = 3410,
    #[doc(hidden)]
    Kxo = 3411,
    #[doc(hidden)]
    Kxp = 3412,
    #[doc(hidden)]
    Kxq = 3413,
    #[doc(hidden)]
    Kxr = 3414,
    #[doc(hidden)]
    Kxs = 3415,
    #[doc(hidden)]
    Kxt = 3416,
    #[doc(hidden)]
    Kxv = 3417,
    #[doc(hidden)]
    Kxw = 3418,
    #[doc(hidden)]
    Kxx = 3419,
    #[doc(hidden)]
    Kxy = 3420,
    #[doc(hidden)]
    Kxz = 3421,
    #[doc(hidden)]
    Kya = 3422,
    #[doc(hidden)]
    Kyb = 3423,
    #[doc(hidden)]
    Kyc = 3424,
    #[doc(hidden)]
    Kyd = 3425,
    #[doc(hidden)]
    Kye = 3426,
    #[doc(hidden)]
    Kyf = 3427,
    #[doc(hidden)]
    Kyg = 3428,
    #[doc(hidden)]
    Kyh = 3429,
    #[doc(hidden)]
    Kyi = 3430,
    #[doc(hidden)]
    Kyj = 3431,
    #[doc(hidden)]
    Kyk = 3432,
    #[doc(hidden)]
    Kyl = 3433,
    #[doc(hidden)]
    Kym = 3434,
    #[doc(hidden)]
    Kyn = 3435,
    #[doc(hidden)]
    Kyo = 3436,
    #[doc(hidden)]
    Kyp = 3437,
    #[doc(hidden)]
    Kyq = 3438,
    #[doc(hidden)]
    Kyr = 3439,
    #[doc(hidden)]
    Kys = 3440,
    #[doc(hidden)]
    Kyt = 3441,
    #[doc(hidden)]
    Kyu = 3442,
    #[doc(hidden)]
    Kyv = 3443,
    #[doc(hidden)]
    Kyw = 3444,
    #[doc(hidden)]
    Kyx = 3445,
    #[doc(hidden)]
    Kyy = 3446,
    #[doc(hidden)]
    Kyz = 3447,
    #[doc(hidden)]
    Kza = 3448,
    #[doc(hidden)]
    Kzb = 3449,
    #[doc(hidden)]
    Kzc = 3450,
    #[doc(hidden)]
    Kzd = 3451,
    #[doc(hidden)]
    Kze = 3452,
    #[doc(hidden)]
    Kzf = 3453,
    #[doc(hidden)]
    Kzg = 3454,
    #[doc(hidden)]
    Kzi = 3455,
    #[doc(hidden)]
    Kzk = 3456,
    #[doc(hidden)]
    Kzl = 3457,
    #[doc(hidden)]
    Kzm = 3458,
    #[doc(hidden)]
    Kzn = 3459,
    #[doc(hidden)]
    Kzo = 3460,
    #[doc(hidden)]
    Kzp = 3461,
    #[doc(hidden)]
    Kzq = 3462,
    #[doc(hidden)]
    Kzr = 3463,
    #[doc(hidden)]
    Kzs = 3464,
    #[doc(hidden)]
    Kzu = 3465,
    #[doc(hidden)]
    Kzv = 3466,
    #[doc(hidden)]
    Kzw = 3467,
    #[doc(hidden)]
    Kzx = 3468,
    #[doc(hidden)]
    Kzy = 3469,
    #[doc(hidden)]
    Kzz = 3470,
    #[doc(hidden)]
    Laa = 3471,
    #[doc(hidden)]
    Lab = 3472,
    #[doc(hidden)]
    Lac = 3473,
    #[doc(hidden)]
    Lad = 3474,
    #[doc(hidden)]
    Lae = 3475,
    #[doc(hidden)]
    Laf = 3476,
    #[doc(hidden)]
    Lag = 3477,
    #[doc(hidden)]
    Lah = 3478,
    #[doc(hidden)]
    Lai = 3479,
    #[doc(hidden)]
    Laj = 3480,
    #[doc(hidden)]
    Lal = 3481,
    #[doc(hidden)]
    Lam = 3482,
    #[doc(hidden)]
    Lan = 3483,
    #[doc(hidden)]
    Lao = 3484,
    #[doc(hidden)]
    Lap = 3485,
    #[doc(hidden)]
    Laq = 3486,
    #[doc(hidden)]
    Lar = 3487,
    #[doc(hidden)]
    Las = 3488,
    #[doc(hidden)]
    Lat = 3489,
    #[doc(hidden)]
    Lau = 3490,
    #[doc(hidden)]
    Lav = 3491,
    #[doc(hidden)]
    Law = 3492,
    #[doc(hidden)]
    Lax = 3493,
    #[doc(hidden)]
    Lay = 3494,
    #[doc(hidden)]
    Laz = 3495,
    #[doc(hidden)]
    Lbb = 3496,
    #[doc(hidden)]
    Lbc = 3497,
    #[doc(hidden)]
    Lbe = 3498,
    #[doc(hidden)]
    Lbf = 3499,
    #[doc(hidden)]
    Lbg = 3500,
    #[doc(hidden)]
    Lbi = 3501,
    #[doc(hidden)]
    Lbj = 3502,
    #[doc(hidden)]
    Lbk = 3503,
    #[doc(hidden)]
    Lbl = 3504,
    #[doc(hidden)]
    Lbm = 3505,
    #[doc(hidden)]
    Lbn = 3506,
    #[doc(hidden)]
    Lbo = 3507,
    #[doc(hidden)]
    Lbq = 3508,
    #[doc(hidden)]
    Lbr = 3509,
    #[doc(hidden)]
    Lbs = 3510,
    #[doc(hidden)]
    Lbt = 3511,
    #[doc(hidden)]
    Lbu = 3512,
    #[doc(hidden)]
    Lbv = 3513,
    #[doc(hidden)]
    Lbw = 3514,
    #[doc(hidden)]
    Lbx = 3515,
    #[doc(hidden)]
    Lby = 3516,
    #[doc(hidden)]
    Lbz = 3517,
    #[doc(hidden)]
    Lcc = 3518,
    #[doc(hidden)]
    Lcd = 3519,
    #[doc(hidden)]
    Lce = 3520,
    #[doc(hidden)]
    Lcf = 3521,
    #[doc(hidden)]
    Lch = 3522,
    #[doc(hidden)]
    Lcl = 3523,
    #[doc(hidden)]
    Lcm = 3524,
    #[doc(hidden)]
    Lcp = 3525,
    #[doc(hidden)]
    Lcq = 3526,
    #[doc(hidden)]
    Lcs = 3527,
    #[doc(hidden)]
    Lda = 3528,
    #[doc(hidden)]
    Ldb = 3529,
    #[doc(hidden)]
    Ldd = 3530,
    #[doc(hidden)]
    Ldg = 3531,
    #[doc(hidden)]
    Ldh = 3532,
    #[doc(hidden)]
    Ldi = 3533,
    #[doc(hidden)]
    Ldj = 3534,
    #[doc(hidden)]
    Ldk = 3535,
    #[doc(hidden)]
    Ldl = 3536,
    #[doc(hidden)]
    Ldm = 3537,
    #[doc(hidden)]
    Ldn = 3538,
    #[doc(hidden)]
    Ldo = 3539,
    #[doc(hidden)]
    Ldp = 3540,
    #[doc(hidden)]
    Ldq = 3541,
    #[doc(hidden)]
    Lea = 3542,
    #[doc(hidden)]
    Leb = 3543,
    #[doc(hidden)]
    Lec = 3544,
    #[doc(hidden)]
    Led = 3545,
    #[doc(hidden)]
    Lee = 3546,
    #[doc(hidden)]
    Lef = 3547,
    #[doc(hidden)]
    Leh = 3548,
    #[doc(hidden)]
    Lei = 3549,
    #[doc(hidden)]
    Lej = 3550,
    #[doc(hidden)]
    Lek = 3551,
    #[doc(hidden)]
    Lel = 3552,
    #[doc(hidden)]
    Lem = 3553,
    #[doc(hidden)]
    Len = 3554,
    #[doc(hidden)]
    Leo = 3555,
    #[doc(hidden)]
    Lep = 3556,
    #[doc(hidden)]
    Leq = 3557,
    #[doc(hidden)]
    Ler = 3558,
    #[doc(hidden)]
    Les = 3559,
    #[doc(hidden)]
    Let = 3560,
    #[doc(hidden)]
    Leu = 3561,
    #[doc(hidden)]
    Lev = 3562,
    #[doc(hidden)]
    Lew = 3563,
    #[doc(hidden)]
    Lex = 3564,
    #[doc(hidden)]
    Ley = 3565,
    #[doc(hidden)]
    Lez = 3566,
    #[doc(hidden)]
    Lfa = 3567,
    #[doc(hidden)]
    Lfn = 3568,
    #[doc(hidden)]
    Lga = 3569,
    #[doc(hidden)]
    Lgb = 3570,
    #[doc(hidden)]
    Lgg = 3571,
    #[doc(hidden)]
    Lgh = 3572,
    #[doc(hidden)]
    Lgi = 3573,
    #[doc(hidden)]
    Lgk = 3574,
    #[doc(hidden)]
    Lgl = 3575,
    #[doc(hidden)]
    Lgm = 3576,
    #[doc(hidden)]
    Lgn = 3577,
    #[doc(hidden)]
    Lgo = 3578,
    #[doc(hidden)]
    Lgq = 3579,
    #[doc(hidden)]
    Lgr = 3580,
    #[doc(hidden)]
    Lgt = 3581,
    #[doc(hidden)]
    Lgu = 3582,
    #[doc(hidden)]
    Lgz = 3583,
    #[doc(hidden)]
    Lha = 3584,
    #[doc(hidden)]
    Lhh = 3585,
    #[doc(hidden)]
    Lhi = 3586,
    #[doc(hidden)]
    Lhl = 3587,
    #[doc(hidden)]
    Lhm = 3588,
    #[doc(hidden)]
    Lhn = 3589,
    #[doc(hidden)]
    Lhp = 3590,
    #[doc(hidden)]
    Lhs = 3591,
    #[doc(hidden)]
    Lht = 3592,
    #[doc(hidden)]
    Lhu = 3593,
    #[doc(hidden)]
    Lia = 3594,
    #[doc(hidden)]
    Lib = 3595,
    #[doc(hidden)]
    Lic = 3596,
    #[doc(hidden)]
    Lid = 3597,
    #[doc(hidden)]
    Lie = 3598,
    #[doc(hidden)]
    Lif = 3599,
    #[doc(hidden)]
    Lig = 3600,
    #[doc(hidden)]
    Lih = 3601,
    #[doc(hidden)]
    Lij = 3602,
    #[doc(hidden)]
    Lik = 3603,
    #[doc(hidden)]
    Lil = 3604,
    #[doc(hidden)]
    Lim = 3605,
    #[doc(hidden)]
    Lin = 3606,
    #[doc(hidden)]
    Lio = 3607,
    #[doc(hidden)]
    Lip = 3608,
    #[doc(hidden)]
    Liq = 3609,
    #[doc(hidden)]
    Lir = 3610,
    #[doc(hidden)]
    Lis = 3611,
    #[doc(hidden)]
    Lit = 3612,
    #[doc(hidden)]
    Liu = 3613,
    #[doc(hidden)]
    Liv = 3614,
    #[doc(hidden)]
    Liw = 3615,
    #[doc(hidden)]
    Lix = 3616,
    #[doc(hidden)]
    Liy = 3617,
    #[doc(hidden)]
    Liz = 3618,
    #[doc(hidden)]
    Lja = 3619,
    #[doc(hidden)]
    Lje = 3620,
    #[doc(hidden)]
    Lji = 3621,
    #[doc(hidden)]
    Ljl = 3622,
    #[doc(hidden)]
    Ljp = 3623,
    #[doc(hidden)]
    Ljw = 3624,
    #[doc(hidden)]
    Ljx = 3625,
    #[doc(hidden)]
    Lka = 3626,
    #[doc(hidden)]
    Lkb = 3627,
    #[doc(hidden)]
    Lkc = 3628,
    #[doc(hidden)]
    Lkd = 3629,
    #[doc(hidden)]
    Lke = 3630,
    #[doc(hidden)]
    Lkh = 3631,
    #[doc(hidden)]
    Lki = 3632,
    #[doc(hidden)]
    Lkj = 3633,
    #[doc(hidden)]
    Lkl = 3634,
    #[doc(hidden)]
    Lkm = 3635,
    #[doc(hidden)]
    Lkn = 3636,
    #[doc(hidden)]
    Lko = 3637,
    #[doc(hidden)]
    Lkr = 3638,
    #[doc(hidden)]
    Lks = 3639,
    #[doc(hidden)]
    Lkt = 3640,
    #[doc(hidden)]
    Lku = 3641,
    #[doc(hidden)]
    Lky = 3642,
    #[doc(hidden)]
    Lla = 3643,
    #[doc(hidden)]
    Llb = 3644,
    #[doc(hidden)]
    Llc = 3645,
    #[doc(hidden)]
    Lld = 3646,
    #[doc(hidden)]
    Lle = 3647,
    #[doc(hidden)]
    Llf = 3648,
    #[doc(hidden)]
    Llg = 3649,
    #[doc(hidden)]
    Llh = 3650,
    #[doc(hidden)]
    Lli = 3651,
    #[doc(hidden)]
    Llj = 3652,
    #[doc(hidden)]
    Llk = 3653,
    #[doc(hidden)]
    Lll = 3654,
    #[doc(hidden)]
    Llm = 3655,
    #[doc(hidden)]
    Lln = 3656,
    #[doc(hidden)]
    Llp = 3657,
    #[doc(hidden)]
    Llq = 3658,
    #[doc(hidden)]
    Lls = 3659,
    #[doc(hidden)]
    Llu = 3660,
    #[doc(hidden)]
    Llx = 3661,
    #[doc(hidden)]
    Lma = 3662,
    #[doc(hidden)]
    Lmb = 3663,
    #[doc(hidden)]
    Lmc = 3664,
    #[doc(hidden)]
    Lmd = 3665,
    #[doc(hidden)]
    Lme = 3666,
    #[doc(hidden)]
    Lmf = 3667,
    #[doc(hidden)]
    Lmg = 3668,
    #[doc(hidden)]
    Lmh = 3669,
    #[doc(hidden)]
    Lmi = 3670,
    #[doc(hidden)]
    Lmj = 3671,
    #[doc(hidden)]
    Lmk = 3672,
    #[doc(hidden)]
    Lml = 3673,
    #[doc(hidden)]
    Lmn = 3674,
    #[doc(hidden)]
    Lmo = 3675,
    #[doc(hidden)]
    Lmp = 3676,
    #[doc(hidden)]
    Lmq = 3677,
    #[doc(hidden)]
    Lmr = 3678,
    #[doc(hidden)]
    Lmu = 3679,
    #[doc(hidden)]
    Lmv = 3680,
    #[doc(hidden)]
    Lmw = 3681,
    #[doc(hidden)]
    Lmx = 3682,
    #[doc(hidden)]
    Lmy = 3683,
    #[doc(hidden)]
    Lna = 3684,
    #[doc(hidden)]
    Lnb = 3685,
    #[doc(hidden)]
    Lnd = 3686,
    #[doc(hidden)]
    Lng = 3687,
    #[doc(hidden)]
    Lnh = 3688,
    #[doc(hidden)]
    Lni = 3689,
    #[doc(hidden)]
    Lnj = 3690,
    #[doc(hidden)]
    Lnl = 3691,
    #[doc(hidden)]
    Lnm = 3692,
    #[doc(hidden)]
    Lnn = 3693,
    #[doc(hidden)]
    Lns = 3694,
    #[doc(hidden)]
    Lnu = 3695,
    #[doc(hidden)]
    Lnw = 3696,
    #[doc(hidden)]
    Lnz = 3697,
    #[doc(hidden)]
    Loa = 3698,
    #[doc(hidden)]
    Lob = 3699,
    #[doc(hidden)]
    Loc = 3700,
    #[doc(hidden)]
    Loe = 3701,
    #[doc(hidden)]
    Lof = 3702,
    #[doc(hidden)]
    Log = 3703,
    #[doc(hidden)]
    Loh = 3704,
    #[doc(hidden)]
    Loi = 3705,
    #[doc(hidden)]
    Loj = 3706,
    #[doc(hidden)]
    Lok = 3707,
    #[doc(hidden)]
    Lol = 3708,
    #[doc(hidden)]
    Lom = 3709,
    #[doc(hidden)]
    Lon = 3710,
    #[doc(hidden)]
    Loo = 3711,
    #[doc(hidden)]
    Lop = 3712,
    #[doc(hidden)]
    Loq = 3713,
    #[doc(hidden)]
    Lor = 3714,
    #[doc(hidden)]
    Los = 3715,
    #[doc(hidden)]
    Lot = 3716,
    #[doc(hidden)]
    Lou = 3717,
    #[doc(hidden)]
    Lov = 3718,
    #[doc(hidden)]
    Low = 3719,
    #[doc(hidden)]
    Lox = 3720,
    #[doc(hidden)]
    Loy = 3721,
    #[doc(hidden)]
    Loz = 3722,
    #[doc(hidden)]
    Lpa = 3723,
    #[doc(hidden)]
    Lpe = 3724,
    #[doc(hidden)]
    Lpn = 3725,
    #[doc(hidden)]
    Lpo = 3726,
    #[doc(hidden)]
    Lpx = 3727,
    #[doc(hidden)]
    Lqr = 3728,
    #[doc(hidden)]
    Lra = 3729,
    #[doc(hidden)]
    Lrc = 3730,
    #[doc(hidden)]
    Lre = 3731,
    #[doc(hidden)]
    Lrg = 3732,
    #[doc(hidden)]
    Lri = 3733,
    #[doc(hidden)]
    Lrk = 3734,
    #[doc(hidden)]
    Lrl = 3735,
    #[doc(hidden)]
    Lrm = 3736,
    #[doc(hidden)]
    Lrn = 3737,
    #[doc(hidden)]
    Lro = 3738,
    #[doc(hidden)]
    Lrr = 3739,
    #[doc(hidden)]
    Lrt = 3740,
    #[doc(hidden)]
    Lrv = 3741,
    #[doc(hidden)]
    Lrz = 3742,
    #[doc(hidden)]
    Lsa = 3743,
    #[doc(hidden)]
    Lsb = 3744,
    #[doc(hidden)]
    Lsc = 3745,
    #[doc(hidden)]
    Lsd = 3746,
    #[doc(hidden)]
    Lse = 3747,
    #[doc(hidden)]
    Lsh = 3748,
    #[doc(hidden)]
    Lsi = 3749,
    #[doc(hidden)]
    Lsl = 3750,
    #[doc(hidden)]
    Lsm = 3751,
    #[doc(hidden)]
    Lsn = 3752,
    #[doc(hidden)]
    Lso = 3753,
    #[doc(hidden)]
    Lsp = 3754,
    #[doc(hidden)]
    Lsr = 3755,
    #[doc(hidden)]
    Lss = 3756,
    #[doc(hidden)]
    Lst = 3757,
    #[doc(hidden)]
    Lsv = 3758,
    #[doc(hidden)]
    Lsw = 3759,
    #[doc(hidden)]
    Lsy = 3760,
    #[doc(hidden)]
    Ltc = 3761,
    #[doc(hidden)]
    Ltg = 3762,
    #[doc(hidden)]
    Lth = 3763,
    #[doc(hidden)]
    Lti = 3764,
    #[doc(hidden)]
    Ltn = 3765,
    #[doc(hidden)]
    Lto = 3766,
    #[doc(hidden)]
    Lts = 3767,
    #[doc(hidden)]
    Ltu = 3768,
    #[doc(hidden)]
    Ltz = 3769,
    #[doc(hidden)]
    Lua = 3770,
    #[doc(hidden)]
    Lub = 3771,
    #[doc(hidden)]
    Luc = 3772,
    #[doc(hidden)]
    Lud = 3773,
    #[doc(hidden)]
    Lue = 3774,
    #[doc(hidden)]
    Luf = 3775,
    #[doc(hidden)]
    Lug = 3776,
    #[doc(hidden)]
    Lui = 3777,
    #[doc(hidden)]
    Luj = 3778,
    #[doc(hidden)]
    Luk = 3779,
    #[doc(hidden)]
    Lul = 3780,
    #[doc(hidden)]
    Lum = 3781,
    #[doc(hidden)]
    Lun = 3782,
    #[doc(hidden)]
    Luo = 3783,
    #[doc(hidden)]
    Lup = 3784,
    #[doc(hidden)]
    Luq = 3785,
    #[doc(hidden)]
    Lur = 3786,
    #[doc(hidden)]
    Lus = 3787,
    #[doc(hidden)]
    Lut = 3788,
    #[doc(hidden)]
    Luu = 3789,
    #[doc(hidden)]
    Luv = 3790,
    #[doc(hidden)]
    Luw = 3791,
    #[doc(hidden)]
    Luy = 3792,
    #[doc(hidden)]
    Luz = 3793,
    #[doc(hidden)]
    Lva = 3794,
    #[doc(hidden)]
    Lvi = 3795,
    #[doc(hidden)]
    Lvk = 3796,
    #[doc(hidden)]
    Lvs = 3797,
    #[doc(hidden)]
    Lvu = 3798,
    #[doc(hidden)]
    Lwa = 3799,
    #[doc(hidden)]
    Lwe = 3800,
    #[doc(hidden)]
    Lwg = 3801,
    #[doc(hidden)]
    Lwh = 3802,
    #[doc(hidden)]
    Lwl = 3803,
    #[doc(hidden)]
    Lwm = 3804,
    #[doc(hidden)]
    Lwo = 3805,
    #[doc(hidden)]
    Lws = 3806,
    #[doc(hidden)]
    Lwt = 3807,
    #[doc(hidden)]
    Lwu = 3808,
    #[doc(hidden)]
    Lww = 3809,
    #[doc(hidden)]
    Lxm = 3810,
    #[doc(hidden)]
    Lya = 3811,
    #[doc(hidden)]
    Lyg = 3812,
    #[doc(hidden)]
    Lyn = 3813,
    #[doc(hidden)]
    Lzh = 3814,
    #[doc(hidden)]
    Lzl = 3815,
    #[doc(hidden)]
    Lzn = 3816,
    #[doc(hidden)]
    Lzz = 3817,
    #[doc(hidden)]
    Maa = 3818,
    #[doc(hidden)]
    Mab = 3819,
    #[doc(hidden)]
    Mad = 3820,
    #[doc(hidden)]
    Mae = 3821,
    #[doc(hidden)]
    Maf = 3822,
    #[doc(hidden)]
    Mag = 3823,
    #[doc(hidden)]
    Mah = 3824,
    #[doc(hidden)]
    Mai = 3825,
    #[doc(hidden)]
    Maj = 3826,
    #[doc(hidden)]
    Mak = 3827,
    #[doc(hidden)]
    Mal = 3828,
    #[doc(hidden)]
    Mam = 3829,
    #[doc(hidden)]
    Man = 3830,
    #[doc(hidden)]
    Maq = 3831,
    #[doc(hidden)]
    Mar = 3832,
    #[doc(hidden)]
    Mas = 3833,
    #[doc(hidden)]
    Mat = 3834,
    #[doc(hidden)]
    Mau = 3835,
    #[doc(hidden)]
    Mav = 3836,
    #[doc(hidden)]
    Maw = 3837,
    #[doc(hidden)]
    Max = 3838,
    #[doc(hidden)]
    Maz = 3839,
    #[doc(hidden)]
    Mba = 3840,
    #[doc(hidden)]
    Mbb = 3841,
    #[doc(hidden)]
    Mbc = 3842,
    #[doc(hidden)]
    Mbd = 3843,
    #[doc(hidden)]
    Mbe = 3844,
    #[doc(hidden)]
    Mbf = 3845,
    #[doc(hidden)]
    Mbh = 3846,
    #[doc(hidden)]
    Mbi = 3847,
    #[doc(hidden)]
    Mbj = 3848,
    #[doc(hidden)]
    Mbk = 3849,
    #[doc(hidden)]
    Mbl = 3850,
    #[doc(hidden)]
    Mbm = 3851,
    #[doc(hidden)]
    Mbn = 3852,
    #[doc(hidden)]
    Mbo = 3853,
    #[doc(hidden)]
    Mbp = 3854,
    #[doc(hidden)]
    Mbq = 3855,
    #[doc(hidden)]
    Mbr = 3856,
    #[doc(hidden)]
    Mbs = 3857,
    #[doc(hidden)]
    Mbt = 3858,
    #[doc(hidden)]
    Mbu = 3859,
    #[doc(hidden)]
    Mbv = 3860,
    #[doc(hidden)]
    Mbw = 3861,
    #[doc(hidden)]
    Mbx = 3862,
    #[doc(hidden)]
    Mby = 3863,
    #[doc(hidden)]
    Mbz = 3864,
    #[doc(hidden)]
    Mca = 3865,
    #[doc(hidden)]
    Mcb = 3866,
    #[doc(hidden)]
    Mcc = 3867,
    #[doc(hidden)]
    Mcd = 3868,
    #[doc(hidden)]
    Mce = 3869,
    #[doc(hidden)]
    Mcf = 3870,
    #[doc(hidden)]
    Mcg = 3871,
    #[doc(hidden)]
    Mch = 3872,
    #[doc(hidden)]
    Mci = 3873,
    #[doc(hidden)]
    Mcj = 3874,
    #[doc(hidden)]
    Mck = 3875,
    #[doc(hidden)]
    Mcl = 3876,
    #[doc(hidden)]
    Mcm = 3877,
    #[doc(hidden)]
    Mcn = 3878,
    #[doc(hidden)]
    Mco = 3879,
    #[doc(hidden)]
    Mcp = 3880,
    #[doc(hidden)]
    Mcq = 3881,
    #[doc(hidden)]
    Mcr = 3882,
    #[doc(hidden)]
    Mcs = 3883,
    #[doc(hidden)]
    Mct = 3884,
    #[doc(hidden)]
    Mcu = 3885,
    #[doc(hidden)]
    Mcv = 3886,
    #[doc(hidden)]
    Mcw = 3887,
    #[doc(hidden)]
    Mcx = 3888,
    #[doc(hidden)]
    Mcy = 3889,
    #[doc(hidden)]
    Mcz = 3890,
    #[doc(hidden)]
    Mda = 3891,
    #[doc(hidden)]
    Mdb = 3892,
    #[doc(hidden)]
    Mdc = 3893,
    #[doc(hidden)]
    Mdd = 3894,
    #[doc(hidden)]
    Mde = 3895,
    #[doc(hidden)]
    Mdf = 3896,
    #[doc(hidden)]
    Mdg = 3897,
    #[doc(hidden)]
    Mdh = 3898,
    #[doc(hidden)]
    Mdi = 3899,
    #[doc(hidden)]
    Mdj = 3900,
    #[doc(hidden)]
    Mdk = 3901,
    #[doc(hidden)]
    Mdl = 3902,
    #[doc(hidden)]
    Mdm = 3903,
    #[doc(hidden)]
    Mdn = 3904,
    #[doc(hidden)]
    Mdp = 3905,
    #[doc(hidden)]
    Mdq = 3906,
    #[doc(hidden)]
    Mdr = 3907,
    #[doc(hidden)]
    Mds = 3908,
    #[doc(hidden)]
    Mdt = 3909,
    #[doc(hidden)]
    Mdu = 3910,
    #[doc(hidden)]
    Mdv = 3911,
    #[doc(hidden)]
    Mdw = 3912,
    #[doc(hidden)]
    Mdx = 3913,
    #[doc(hidden)]
    Mdy = 3914,
    #[doc(hidden)]
    Mdz = 3915,
    #[doc(hidden)]
    Mea = 3916,
    #[doc(hidden)]
    Meb = 3917,
    #[doc(hidden)]
    Mec = 3918,
    #[doc(hidden)]
    Med = 3919,
    #[doc(hidden)]
    Mee = 3920,
    #[doc(hidden)]
    Mef = 3921,
    #[doc(hidden)]
    Meh = 3922,
    #[doc(hidden)]
    Mei = 3923,
    #[doc(hidden)]
    Mej = 3924,
    #[doc(hidden)]
    Mek = 3925,
    #[doc(hidden)]
    Mel = 3926,
    #[doc(hidden)]
    Mem = 3927,
    #[doc(hidden)]
    Men = 3928,
    #[doc(hidden)]
    Meo = 3929,
    #[doc(hidden)]
    Mep = 3930,
    #[doc(hidden)]
    Meq = 3931,
    #[doc(hidden)]
    Mer = 3932,
    #[doc(hidden)]
    Mes = 3933,
    #[doc(hidden)]
    Met = 3934,
    #[doc(hidden)]
    Meu = 3935,
    #[doc(hidden)]
    Mev = 3936,
    #[doc(hidden)]
    Mew = 3937,
    #[doc(hidden)]
    Mey = 3938,
    #[doc(hidden)]
    Mez = 3939,
    #[doc(hidden)]
    Mfa = 3940,
    #[doc(hidden)]
    Mfb = 3941,
    #[doc(hidden)]
    Mfc = 3942,
    #[doc(hidden)]
    Mfd = 3943,
    #[doc(hidden)]
    Mfe = 3944,
    #[doc(hidden)]
    Mff = 3945,
    #[doc(hidden)]
    Mfg = 3946,
    #[doc(hidden)]
    Mfh = 3947,
    #[doc(hidden)]
    Mfi = 3948,
    #[doc(hidden)]
    Mfj = 3949,
    #[doc(hidden)]
    Mfk = 3950,
    #[doc(hidden)]
    Mfl = 3951,
    #[doc(hidden)]
    Mfm = 3952,
    #[doc(hidden)]
    Mfn = 3953,
    #[doc(hidden)]
    Mfo = 3954,
    #[doc(hidden)]
    Mfp = 3955,
    #[doc(hidden)]
    Mfq = 3956,
    #[doc(hidden)]
    Mfr = 3957,
    #[doc(hidden)]
    Mfs = 3958,
    #[doc(hidden)]
    Mft = 3959,
    #[doc(hidden)]
    Mfu = 3960,
    #[doc(hidden)]
    Mfv = 3961,
    #[doc(hidden)]
    Mfw = 3962,
    #[doc(hidden)]
    Mfx = 3963,
    #[doc(hidden)]
    Mfy = 3964,
    #[doc(hidden)]
    Mfz = 3965,
    #[doc(hidden)]
    Mga = 3966,
    #[doc(hidden)]
    Mgb = 3967,
    #[doc(hidden)]
    Mgc = 3968,
    #[doc(hidden)]
    Mgd = 3969,
    #[doc(hidden)]
    Mge = 3970,
    #[doc(hidden)]
    Mgf = 3971,
    #[doc(hidden)]
    Mgg = 3972,
    #[doc(hidden)]
    Mgh = 3973,
    #[doc(hidden)]
    Mgi = 3974,
    #[doc(hidden)]
    Mgj = 3975,
    #[doc(hidden)]
    Mgk = 3976,
    #[doc(hidden)]
    Mgl = 3977,
    #[doc(hidden)]
    Mgm = 3978,
    #[doc(hidden)]
    Mgn = 3979,
    #[doc(hidden)]
    Mgo = 3980,
    #[doc(hidden)]
    Mgp = 3981,
    #[doc(hidden)]
    Mgq = 3982,
    #[doc(hidden)]
    Mgr = 3983,
    #[doc(hidden)]
    Mgs = 3984,
    #[doc(hidden)]
    Mgt = 3985,
    #[doc(hidden)]
    Mgu = 3986,
    #[doc(hidden)]
    Mgv = 3987,
    #[doc(hidden)]
    Mgw = 3988,
    #[doc(hidden)]
    Mgy = 3989,
    #[doc(hidden)]
    Mgz = 3990,
    #[doc(hidden)]
    Mha = 3991,
    #[doc(hidden)]
    Mhb = 3992,
    #[doc(hidden)]
    Mhc = 3993,
    #[doc(hidden)]
    Mhd = 3994,
    #[doc(hidden)]
    Mhe = 3995,
    #[doc(hidden)]
    Mhf = 3996,
    #[doc(hidden)]
    Mhg = 3997,
    #[doc(hidden)]
    Mhi = 3998,
    #[doc(hidden)]
    Mhj = 3999,
    #[doc(hidden)]
    Mhk = 4000,
    #[doc(hidden)]
    Mhl = 4001,
    #[doc(hidden)]
    Mhm = 4002,
    #[doc(hidden)]
    Mhn = 4003,
    #[doc(hidden)]
    Mho = 4004,
    #[doc(hidden)]
    Mhp = 4005,
    #[doc(hidden)]
    Mhq = 4006,
    #[doc(hidden)]
    Mhr = 4007,
    #[doc(hidden)]
    Mhs = 4008,
    #[doc(hidden)]
    Mht = 4009,
    #[doc(hidden)]
    Mhu = 4010,
    #[doc(hidden)]
    Mhw = 4011,
    #[doc(hidden)]
    Mhx = 4012,
    #[doc(hidden)]
    Mhy = 4013,
    #[doc(hidden)]
    Mhz = 4014,
    #[doc(hidden)]
    Mia = 4015,
    #[doc(hidden)]
    Mib = 4016,
    #[doc(hidden)]
    Mic = 4017,
    #[doc(hidden)]
    Mid = 4018,
    #[doc(hidden)]
    Mie = 4019,
    #[doc(hidden)]
    Mif = 4020,
    #[doc(hidden)]
    Mig = 4021,
    #[doc(hidden)]
    Mih = 4022,
    #[doc(hidden)]
    Mii = 4023,
    #[doc(hidden)]
    Mij = 4024,
    #[doc(hidden)]
    Mik = 4025,
    #[doc(hidden)]
    Mil = 4026,
    #[doc(hidden)]
    Mim = 4027,
    #[doc(hidden)]
    Min = 4028,
    #[doc(hidden)]
    Mio = 4029,
    #[doc(hidden)]
    Mip = 4030,
    #[doc(hidden)]
    Miq = 4031,
    #[doc(hidden)]
    Mir = 4032,
    #[doc(hidden)]
    Mis = 4033,
    #[doc(hidden)]
    Mit = 4034,
    #[doc(hidden)]
    Miu = 4035,
    #[doc(hidden)]
    Miw = 4036,
    #[doc(hidden)]
    Mix = 4037,
    #[doc(hidden)]
    Miy = 4038,
    #[doc(hidden)]
    Miz = 4039,
    #[doc(hidden)]
    Mjb = 4040,
    #[doc(hidden)]
    Mjc = 4041,
    #[doc(hidden)]
    Mjd = 4042,
    #[doc(hidden)]
    Mje = 4043,
    #[doc(hidden)]
    Mjg = 4044,
    #[doc(hidden)]
    Mjh = 4045,
    #[doc(hidden)]
    Mji = 4046,
    #[doc(hidden)]
    Mjj = 4047,
    #[doc(hidden)]
    Mjk = 4048,
    #[doc(hidden)]
    Mjl = 4049,
    #[doc(hidden)]
    Mjm = 4050,
    #[doc(hidden)]
    Mjn = 4051,
    #[doc(hidden)]
    Mjo = 4052,
    #[doc(hidden)]
    Mjp = 4053,
    #[doc(hidden)]
    Mjq = 4054,
    #[doc(hidden)]
    Mjr = 4055,
    #[doc(hidden)]
    Mjs = 4056,
    #[doc(hidden)]
    Mjt = 4057,
    #[doc(hidden)]
    Mju = 4058,
    #[doc(hidden)]
    Mjv = 4059,
    #[doc(hidden)]
    Mjw = 4060,
    #[doc(hidden)]
    Mjx = 4061,
    #[doc(hidden)]
    Mjy = 4062,
    #[doc(hidden)]
    Mjz = 4063,
    #[doc(hidden)]
    Mka = 4064,
    #[doc(hidden)]
    Mkb = 4065,
    #[doc(hidden)]
    Mkc = 4066,
    #[doc(hidden)]
    Mkd = 4067,
    #[doc(hidden)]
    Mke = 4068,
    #[doc(hidden)]
    Mkf = 4069,
    #[doc(hidden)]
    Mkg = 4070,
    #[doc(hidden)]
    Mki = 4071,
    #[doc(hidden)]
    Mkj = 4072,
    #[doc(hidden)]
    Mkk = 4073,
    #[doc(hidden)]
    Mkl = 4074,
    #[doc(hidden)]
    Mkm = 4075,
    #[doc(hidden)]
    Mkn = 4076,
    #[doc(hidden)]
    Mko = 4077,
    #[doc(hidden)]
    Mkp = 4078,
    #[doc(hidden)]
    Mkq = 4079,
    #[doc(hidden)]
    Mkr = 4080,
    #[doc(hidden)]
    Mks = 4081,
    #[doc(hidden)]
    Mkt = 4082,
    #[doc(hidden)]
    Mku = 4083,
    #[doc(hidden)]
    Mkv = 4084,
    #[doc(hidden)]
    Mkw = 4085,
    #[doc(hidden)]
    Mkx = 4086,
    #[doc(hidden)]
    Mky = 4087,
    #[doc(hidden)]
    Mkz = 4088,
    #[doc(hidden)]
    Mla = 4089,
    #[doc(hidden)]
    Mlb = 4090,
    #[doc(hidden)]
    Mlc = 4091,
    #[doc(hidden)]
    Mle = 4092,
    #[doc(hidden)]
    Mlf = 4093,
    #[doc(hidden)]
    Mlg = 4094,
    #[doc(hidden)]
    Mlh = 4095,
    #[doc(hidden)]
    Mli = 4096,
    #[doc(hidden)]
    Mlj = 4097,
    #[doc(hidden)]
    Mlk = 4098,
    #[doc(hidden)]
    Mll = 4099,
    #[doc(hidden)]
    Mlm = 4100,
    #[doc(hidden)]
    Mln = 4101,
    #[doc(hidden)]
    Mlo = 4102,
    #[doc(hidden)]
    Mlp = 4103,
    #[doc(hidden)]
    Mlq = 4104,
    #[doc(hidden)]
    Mlr = 4105,
    #[doc(hidden)]
    Mls = 4106,
    #[doc(hidden)]
    Mlt = 4107,
    #[doc(hidden)]
    Mlu = 4108,
    #[doc(hidden)]
    Mlv = 4109,
    #[doc(hidden)]
    Mlw = 4110,
    #[doc(hidden)]
    Mlx = 4111,
    #[doc(hidden)]
    Mlz = 4112,
    #[doc(hidden)]
    Mma = 4113,
    #[doc(hidden)]
    Mmb = 4114,
    #[doc(hidden)]
    Mmc = 4115,
    #[doc(hidden)]
    Mmd = 4116,
    #[doc(hidden)]
    Mme = 4117,
    #[doc(hidden)]
    Mmf = 4118,
    #[doc(hidden)]
    Mmg = 4119,
    #[doc(hidden)]
    Mmh = 4120,
    #[doc(hidden)]
    Mmi = 4121,
    #[doc(hidden)]
    Mmj = 4122,
    #[doc(hidden)]
    Mmk = 4123,
    #[doc(hidden)]
    Mml = 4124,
    #[doc(hidden)]
    Mmm = 4125,
    #[doc(hidden)]
    Mmn = 4126,
    #[doc(hidden)]
    Mmo = 4127,
    #[doc(hidden)]
    Mmp = 4128,
    #[doc(hidden)]
    Mmq = 4129,
    #[doc(hidden)]
    Mmr = 4130,
    #[doc(hidden)]
    Mmt = 4131,
    #[doc(hidden)]
    Mmu = 4132,
    #[doc(hidden)]
    Mmv = 4133,
    #[doc(hidden)]
    Mmw = 4134,
    #[doc(hidden)]
    Mmx = 4135,
    #[doc(hidden)]
    Mmy = 4136,
    #[doc(hidden)]
    Mmz = 4137,
    #[doc(hidden)]
    Mna = 4138,
    #[doc(hidden)]
    Mnb = 4139,
    #[doc(hidden)]
    Mnc = 4140,
    #[doc(hidden)]
    Mnd = 4141,
    #[doc(hidden)]
    Mne = 4142,
    #[doc(hidden)]
    Mnf = 4143,
    #[doc(hidden)]
    Mng = 4144,
    #[doc(hidden)]
    Mnh = 4145,
    #[doc(hidden)]
    Mni = 4146,
    #[doc(hidden)]
    Mnj = 4147,
    #[doc(hidden)]
    Mnk = 4148,
    #[doc(hidden)]
    Mnl = 4149,
    #[doc(hidden)]
    Mnm = 4150,
    #[doc(hidden)]
    Mnn = 4151,
    #[doc(hidden)]
    Mnp = 4152,
    #[doc(hidden)]
    Mnq = 4153,
    #[doc(hidden)]
    Mnr = 4154,
    #[doc(hidden)]
    Mns = 4155,
    #[doc(hidden)]
    Mnu = 4156,
    #[doc(hidden)]
    Mnv = 4157,
    #[doc(hidden)]
    Mnw = 4158,
    #[doc(hidden)]
    Mnx = 4159,
    #[doc(hidden)]
    Mny = 4160,
    #[doc(hidden)]
    Mnz = 4161,
    #[doc(hidden)]
    Moa = 4162,
    #[doc(hidden)]
    Moc = 4163,
    #[doc(hidden)]
    Mod = 4164,
    #[doc(hidden)]
    Moe = 4165,
    #[doc(hidden)]
    Mog = 4166,
    #[doc(hidden)]
    Moh = 4167,
    #[doc(hidden)]
    Moi = 4168,
    #[doc(hidden)]
    Moj = 4169,
    #[doc(hidden)]
    Mok = 4170,
    #[doc(hidden)]
    Mom = 4171,
    #[doc(hidden)]
    Mon = 4172,
    #[doc(hidden)]
    Moo = 4173,
    #[doc(hidden)]
    Mop = 4174,
    #[doc(hidden)]
    Moq = 4175,
    #[doc(hidden)]
    Mor = 4176,
    #[doc(hidden)]
    Mos = 4177,
    #[doc(hidden)]
    Mot = 4178,
    #[doc(hidden)]
    Mou = 4179,
    #[doc(hidden)]
    Mov = 4180,
    #[doc(hidden)]
    Mow = 4181,
    #[doc(hidden)]
    Mox = 4182,
    #[doc(hidden)]
    Moy = 4183,
    #[doc(hidden)]
    Moz = 4184,
    #[doc(hidden)]
    Mpa = 4185,
    #[doc(hidden)]
    Mpb = 4186,
    #[doc(hidden)]
    Mpc = 4187,
    #[doc(hidden)]
    Mpd = 4188,
    #[doc(hidden)]
    Mpe = 4189,
    #[doc(hidden)]
    Mpg = 4190,
    #[doc(hidden)]
    Mph = 4191,
    #[doc(hidden)]
    Mpi = 4192,
    #[doc(hidden)]
    Mpj = 4193,
    #[doc(hidden)]
    Mpk = 4194,
    #[doc(hidden)]
    Mpl = 4195,
    #[doc(hidden)]
    Mpm = 4196,
    #[doc(hidden)]
    Mpn = 4197,
    #[doc(hidden)]
    Mpo = 4198,
    #[doc(hidden)]
    Mpp = 4199,
    #[doc(hidden)]
    Mpq = 4200,
    #[doc(hidden)]
    Mpr = 4201,
    #[doc(hidden)]
    Mps = 4202,
    #[doc(hidden)]
    Mpt = 4203,
    #[doc(hidden)]
    Mpu = 4204,
    #[doc(hidden)]
    Mpv = 4205,
    #[doc(hidden)]
    Mpw = 4206,
    #[doc(hidden)]
    Mpx = 4207,
    #[doc(hidden)]
    Mpy = 4208,
    #[doc(hidden)]
    Mpz = 4209,
    #[doc(hidden)]
    Mqa = 4210,
    #[doc(hidden)]
    Mqb = 4211,
    #[doc(hidden)]
    Mqc = 4212,
    #[doc(hidden)]
    Mqe = 4213,
    #[doc(hidden)]
    Mqf = 4214,
    #[doc(hidden)]
    Mqg = 4215,
    #[doc(hidden)]
    Mqh = 4216,
    #[doc(hidden)]
    Mqi = 4217,
    #[doc(hidden)]
    Mqj = 4218,
    #[doc(hidden)]
    Mqk = 4219,
    #[doc(hidden)]
    Mql = 4220,
    #[doc(hidden)]
    Mqm = 4221,
    #[doc(hidden)]
    Mqn = 4222,
    #[doc(hidden)]
    Mqo = 4223,
    #[doc(hidden)]
    Mqp = 4224,
    #[doc(hidden)]
    Mqq = 4225,
    #[doc(hidden)]
    Mqr = 4226,
    #[doc(hidden)]
    Mqs = 4227,
    #[doc(hidden)]
    Mqt = 4228,
    #[doc(hidden)]
    Mqu = 4229,
    #[doc(hidden)]
    Mqv = 4230,
    #[doc(hidden)]
    Mqw = 4231,
    #[doc(hidden)]
    Mqx = 4232,
    #[doc(hidden)]
    Mqy = 4233,
    #[doc(hidden)]
    Mqz = 4234,
    #[doc(hidden)]
    Mra = 4235,
    #[doc(hidden)]
    Mrb = 4236,
    #[doc(hidden)]
    Mrc = 4237,
    #[doc(hidden)]
    Mrd = 4238,
    #[doc(hidden)]
    Mre = 4239,
    #[doc(hidden)]
    Mrf = 4240,
    #[doc(hidden)]
    Mrg = 4241,
    #[doc(hidden)]
    Mrh = 4242,
    #[doc(hidden)]
    Mri = 4243,
    #[doc(hidden)]
    Mrj = 4244,
    #[doc(hidden)]
    Mrk = 4245,
    #[doc(hidden)]
    Mrl = 4246,
    #[doc(hidden)]
    Mrm = 4247,
    #[doc(hidden)]
    Mrn = 4248,
    #[doc(hidden)]
    Mro = 4249,
    #[doc(hidden)]
    Mrp = 4250,
    #[doc(hidden)]
    Mrq = 4251,
    #[doc(hidden)]
    Mrr = 4252,
    #[doc(hidden)]
    Mrs = 4253,
    #[doc(hidden)]
    Mrt = 4254,
    #[doc(hidden)]
    Mru = 4255,
    #[doc(hidden)]
    Mrv = 4256,
    #[doc(hidden)]
    Mrw = 4257,
    #[doc(hidden)]
    Mrx = 4258,
    #[doc(hidden)]
    Mry = 4259,
    #[doc(hidden)]
    Mrz = 4260,
    #[doc(hidden)]
    Msa = 4261,
    #[doc(hidden)]
    Msb = 4262,
    #[doc(hidden)]
    Msc = 4263,
    #[doc(hidden)]
    Msd = 4264,
    #[doc(hidden)]
    Mse = 4265,
    #[doc(hidden)]
    Msf = 4266,
    #[doc(hidden)]
    Msg = 4267,
    #[doc(hidden)]
    Msh = 4268,
    #[doc(hidden)]
    Msi = 4269,
    #[doc(hidden)]
    Msj = 4270,
    #[doc(hidden)]
    Msk = 4271,
    #[doc(hidden)]
    Msl = 4272,
    #[doc(hidden)]
    Msm = 4273,
    #[doc(hidden)]
    Msn = 4274,
    #[doc(hidden)]
    Mso = 4275,
    #[doc(hidden)]
    Msp = 4276,
    #[doc(hidden)]
    Msq = 4277,
    #[doc(hidden)]
    Msr = 4278,
    #[doc(hidden)]
    Mss = 4279,
    #[doc(hidden)]
    Msu = 4280,
    #[doc(hidden)]
    Msv = 4281,
    #[doc(hidden)]
    Msw = 4282,
    #[doc(hidden)]
    Msx = 4283,
    #[doc(hidden)]
    Msy = 4284,
    #[doc(hidden)]
    Msz = 4285,
    #[doc(hidden)]
    Mta = 4286,
    #[doc(hidden)]
    Mtb = 4287,
    #[doc(hidden)]
    Mtc = 4288,
    #[doc(hidden)]
    Mtd = 4289,
    #[doc(hidden)]
    Mte = 4290,
    #[doc(hidden)]
    Mtf = 4291,
    #[doc(hidden)]
    Mtg = 4292,
    #[doc(hidden)]
    Mth = 4293,
    #[doc(hidden)]
    Mti = 4294,
    #[doc(hidden)]
    Mtj = 4295,
    #[doc(hidden)]
    Mtk = 4296,
    #[doc(hidden)]
    Mtl = 4297,
    #[doc(hidden)]
    Mtm = 4298,
    #[doc(hidden)]
    Mtn = 4299,
    #[doc(hidden)]
    Mto = 4300,
    #[doc(hidden)]
    Mtp = 4301,
    #[doc(hidden)]
    Mtq = 4302,
    #[doc(hidden)]
    Mtr = 4303,
    #[doc(hidden)]
    Mts = 4304,
    #[doc(hidden)]
    Mtt = 4305,
    #[doc(hidden)]
    Mtu = 4306,
    #[doc(hidden)]
    Mtv = 4307,
    #[doc(hidden)]
    Mtw = 4308,
    #[doc(hidden)]
    Mtx = 4309,
    #[doc(hidden)]
    Mty = 4310,
    #[doc(hidden)]
    Mua = 4311,
    #[doc(hidden)]
    Mub = 4312,
    #[doc(hidden)]
    Muc = 4313,
    #[doc(hidden)]
    Mud = 4314,
    #[doc(hidden)]
    Mue = 4315,
    #[doc(hidden)]
    Mug = 4316,
    #[doc(hidden)]
    Muh = 4317,
    #[doc(hidden)]
    Mui = 4318,
    #[doc(hidden)]
    Muj = 4319,
    #[doc(hidden)]
    Muk = 4320,
    #[doc(hidden)]
    Mul = 4321,
    #[doc(hidden)]
    Mum = 4322,
    #[doc(hidden)]
    Muo = 4323,
    #[doc(hidden)]
    Mup = 4324,
    #[doc(hidden)]
    Muq = 4325,
    #[doc(hidden)]
    Mur = 4326,
    #[doc(hidden)]
    Mus = 4327,
    #[doc(hidden)]
    Mut = 4328,
    #[doc(hidden)]
    Muu = 4329,
    #[doc(hidden)]
    Muv = 4330,
    #[doc(hidden)]
    Mux = 4331,
    #[doc(hidden)]
    Muy = 4332,
    #[doc(hidden)]
    Muz = 4333,
    #[doc(hidden)]
    Mva = 4334,
    #[doc(hidden)]
    Mvb = 4335,
    #[doc(hidden)]
    Mvd = 4336,
    #[doc(hidden)]
    Mve = 4337,
    #[doc(hidden)]
    Mvf = 4338,
    #[doc(hidden)]
    Mvg = 4339,
    #[doc(hidden)]
    Mvh = 4340,
    #[doc(hidden)]
    Mvi = 4341,
    #[doc(hidden)]
    Mvk = 4342,
    #[doc(hidden)]
    Mvl = 4343,
    #[doc(hidden)]
    Mvn = 4344,
    #[doc(hidden)]
    Mvo = 4345,
    #[doc(hidden)]
    Mvp = 4346,
    #[doc(hidden)]
    Mvq = 4347,
    #[doc(hidden)]
    Mvr = 4348,
    #[doc(hidden)]
    Mvs = 4349,
    #[doc(hidden)]
    Mvt = 4350,
    #[doc(hidden)]
    Mvu = 4351,
    #[doc(hidden)]
    Mvv = 4352,
    #[doc(hidden)]
    Mvw = 4353,
    #[doc(hidden)]
    Mvx = 4354,
    #[doc(hidden)]
    Mvy = 4355,
    #[doc(hidden)]
    Mvz = 4356,
    #[doc(hidden)]
    Mwa = 4357,
    #[doc(hidden)]
    Mwb = 4358,
    #[doc(hidden)]
    Mwc = 4359,
    #[doc(hidden)]
    Mwe = 4360,
    #[doc(hidden)]
    Mwf = 4361,
    #[doc(hidden)]
    Mwg = 4362,
    #[doc(hidden)]
    Mwh = 4363,
    #[doc(hidden)]
    Mwi = 4364,
    #[doc(hidden)]
    Mwk = 4365,
    #[doc(hidden)]
    Mwl = 4366,
    #[doc(hidden)]
    Mwm = 4367,
    #[doc(hidden)]
    Mwn = 4368,
    #[doc(hidden)]
    Mwo = 4369,
    #[doc(hidden)]
    Mwp = 4370,
    #[doc(hidden)]
    Mwq = 4371,
    #[doc(hidden)]
    Mwr = 4372,
    #[doc(hidden)]
    Mws = 4373,
    #[doc(hidden)]
    Mwt = 4374,
    #[doc(hidden)]
    Mwu = 4375,
    #[doc(hidden)]
    Mwv = 4376,
    #[doc(hidden)]
    Mww = 4377,
    #[doc(hidden)]
    Mwz = 4378,
    #[doc(hidden)]
    Mxa = 4379,
    #[doc(hidden)]
    Mxb = 4380,
    #[doc(hidden)]
    Mxc = 4381,
    #[doc(hidden)]
    Mxd = 4382,
    #[doc(hidden)]
    Mxe = 4383,
    #[doc(hidden)]
    Mxf = 4384,
    #[doc(hidden)]
    Mxg = 4385,
    #[doc(hidden)]
    Mxh = 4386,
    #[doc(hidden)]
    Mxi = 4387,
    #[doc(hidden)]
    Mxj = 4388,
    #[doc(hidden)]
    Mxk = 4389,
    #[doc(hidden)]
    Mxl = 4390,
    #[doc(hidden)]
    Mxm = 4391,
    #[doc(hidden)]
    Mxn = 4392,
    #[doc(hidden)]
    Mxo = 4393,
    #[doc(hidden)]
    Mxp = 4394,
    #[doc(hidden)]
    Mxq = 4395,
    #[doc(hidden)]
    Mxr = 4396,
    #[doc(hidden)]
    Mxs = 4397,
    #[doc(hidden)]
    Mxt = 4398,
    #[doc(hidden)]
    Mxu = 4399,
    #[doc(hidden)]
    Mxv = 4400,
    #[doc(hidden)]
    Mxw = 4401,
    #[doc(hidden)]
    Mxx = 4402,
    #[doc(hidden)]
    Mxy = 4403,
    #[doc(hidden)]
    Mxz = 4404,
    #[doc(hidden)]
    Mya = 4405,
    #[doc(hidden)]
    Myb = 4406,
    #[doc(hidden)]
    Myc = 4407,
    #[doc(hidden)]
    Mye = 4408,
    #[doc(hidden)]
    Myf = 4409,
    #[doc(hidden)]
    Myg = 4410,
    #[doc(hidden)]
    Myh = 4411,
    #[doc(hidden)]
    Myj = 4412,
    #[doc(hidden)]
    Myk = 4413,
    #[doc(hidden)]
    Myl = 4414,
    #[doc(hidden)]
    Mym = 4415,
    #[doc(hidden)]
    Myo = 4416,
    #[doc(hidden)]
    Myp = 4417,
    #[doc(hidden)]
    Myr = 4418,
    #[doc(hidden)]
    Mys = 4419,
    #[doc(hidden)]
    Myu = 4420,
    #[doc(hidden)]
    Myv = 4421,
    #[doc(hidden)]
    Myw = 4422,
    #[doc(hidden)]
    Myx = 4423,
    #[doc(hidden)]
    Myy = 4424,
    #[doc(hidden)]
    Myz = 4425,
    #[doc(hidden)]
    Mza = 4426,
    #[doc(hidden)]
    Mzb = 4427,
    #[doc(hidden)]
    Mzc = 4428,
    #[doc(hidden)]
    Mzd = 4429,
    #[doc(hidden)]
    Mze = 4430,
    #[doc(hidden)]
    Mzg = 4431,
    #[doc(hidden)]
    Mzh = 4432,
    #[doc(hidden)]
    Mzi = 4433,
    #[doc(hidden)]
    Mzj = 4434,
    #[doc(hidden)]
    Mzk = 4435,
    #[doc(hidden)]
    Mzl = 4436,
    #[doc(hidden)]
    Mzm = 4437,
    #[doc(hidden)]
    Mzn = 4438,
    #[doc(hidden)]
    Mzo = 4439,
    #[doc(hidden)]
    Mzp = 4440,
    #[doc(hidden)]
    Mzq = 4441,
    #[doc(hidden)]
    Mzr = 4442,
    #[doc(hidden)]
    Mzs = 4443,
    #[doc(hidden)]
    Mzt = 4444,
    #[doc(hidden)]
    Mzu = 4445,
    #[doc(hidden)]
    Mzv = 4446,
    #[doc(hidden)]
    Mzw = 4447,
    #[doc(hidden)]
    Mzx = 4448,
    #[doc(hidden)]
    Mzy = 4449,
    #[doc(hidden)]
    Mzz = 4450,
    #[doc(hidden)]
    Naa = 4451,
    #[doc(hidden)]
    Nab = 4452,
    #[doc(hidden)]
    Nac = 4453,
    #[doc(hidden)]
    Nae = 4454,
    #[doc(hidden)]
    Naf = 4455,
    #[doc(hidden)]
    Nag = 4456,
    #[doc(hidden)]
    Naj = 4457,
    #[doc(hidden)]
    Nak = 4458,
    #[doc(hidden)]
    Nal = 4459,
    #[doc(hidden)]
    Nam = 4460,
    #[doc(hidden)]
    Nan = 4461,
    #[doc(hidden)]
    Nao = 4462,
    #[doc(hidden)]
    Nap = 4463,
    #[doc(hidden)]
    Naq = 4464,
    #[doc(hidden)]
    Nar = 4465,
    #[doc(hidden)]
    Nas = 4466,
    #[doc(hidden)]
    Nat = 4467,
    #[doc(hidden)]
    Nau = 4468,
    #[doc(hidden)]
    Nav = 4469,
    #[doc(hidden)]
    Naw = 4470,
    #[doc(hidden)]
    Nax = 4471,
    #[doc(hidden)]
    Nay = 4472,
    #[doc(hidden)]
    Naz = 4473,
    #[doc(hidden)]
    Nba = 4474,
    #[doc(hidden)]
    Nbb = 4475,
    #[doc(hidden)]
    Nbc = 4476,
    #[doc(hidden)]
    Nbd = 4477,
    #[doc(hidden)]
    Nbe = 4478,
    #[doc(hidden)]
    Nbg = 4479,
    #[doc(hidden)]
    Nbh = 4480,
    #[doc(hidden)]
    Nbi = 4481,
    #[doc(hidden)]
    Nbj = 4482,
    #[doc(hidden)]
    Nbk = 4483,
    #[doc(hidden)]
    Nbl = 4484,
    #[doc(hidden)]
    Nbm = 4485,
    #[doc(hidden)]
    Nbn = 4486,
    #[doc(hidden)]
    Nbo = 4487,
    #[doc(hidden)]
    Nbp = 4488,
    #[doc(hidden)]
    Nbq = 4489,
    #[doc(hidden)]
    Nbr = 4490,
    #[doc(hidden)]
    Nbs = 4491,
    #[doc(hidden)]
    Nbt = 4492,
    #[doc(hidden)]
    Nbu = 4493,
    #[doc(hidden)]
    Nbv = 4494,
    #[doc(hidden)]
    Nbw = 4495,
    #[doc(hidden)]
    Nby = 4496,
    #[doc(hidden)]
    Nca = 4497,
    #[doc(hidden)]
    Ncb = 4498,
    #[doc(hidden)]
    Ncc = 4499,
    #[doc(hidden)]
    Ncd = 4500,
    #[doc(hidden)]
    Nce = 4501,
    #[doc(hidden)]
    Ncf = 4502,
    #[doc(hidden)]
    Ncg = 4503,
    #[doc(hidden)]
    Nch = 4504,
    #[doc(hidden)]
    Nci = 4505,
    #[doc(hidden)]
    Ncj = 4506,
    #[doc(hidden)]
    Nck = 4507,
    #[doc(hidden)]
    Ncl = 4508,
    #[doc(hidden)]
    Ncm = 4509,
    #[doc(hidden)]
    Ncn = 4510,
    #[doc(hidden)]
    Nco = 4511,
    #[doc(hidden)]
    Ncq = 4512,
    #[doc(hidden)]
    Ncr = 4513,
    #[doc(hidden)]
    Ncs = 4514,
    #[doc(hidden)]
    Nct = 4515,
    #[doc(hidden)]
    Ncu = 4516,
    #[doc(hidden)]
    Ncx = 4517,
    #[doc(hidden)]
    Ncz = 4518,
    #[doc(hidden)]
    Nda = 4519,
    #[doc(hidden)]
    Ndb = 4520,
    #[doc(hidden)]
    Ndc = 4521,
    #[doc(hidden)]
    Ndd = 4522,
    #[doc(hidden)]
    Nde = 4523,
    #[doc(hidden)]
    Ndf = 4524,
    #[doc(hidden)]
    Ndg = 4525,
    #[doc(hidden)]
    Ndh = 4526,
    #[doc(hidden)]
    Ndi = 4527,
    #[doc(hidden)]
    Ndj = 4528,
    #[doc(hidden)]
    Ndk = 4529,
    #[doc(hidden)]
    Ndl = 4530,
    #[doc(hidden)]
    Ndm = 4531,
    #[doc(hidden)]
    Ndn = 4532,
    #[doc(hidden)]
    Ndo = 4533,
    #[doc(hidden)]
    Ndp = 4534,
    #[doc(hidden)]
    Ndq = 4535,
    #[doc(hidden)]
    Ndr = 4536,
    #[doc(hidden)]
    Nds = 4537,
    #[doc(hidden)]
    Ndt = 4538,
    #[doc(hidden)]
    Ndu = 4539,
    #[doc(hidden)]
    Ndv = 4540,
    #[doc(hidden)]
    Ndw = 4541,
    #[doc(hidden)]
    Ndx = 4542,
    #[doc(hidden)]
    Ndy = 4543,
    #[doc(hidden)]
    Ndz = 4544,
    #[doc(hidden)]
    Nea = 4545,
    #[doc(hidden)]
    Neb = 4546,
    #[doc(hidden)]
    Nec = 4547,
    #[doc(hidden)]
    Ned = 4548,
    #[doc(hidden)]
    Nee = 4549,
    #[doc(hidden)]
    Nef = 4550,
    #[doc(hidden)]
    Neg = 4551,
    #[doc(hidden)]
    Neh = 4552,
    #[doc(hidden)]
    Nei = 4553,
    #[doc(hidden)]
    Nej = 4554,
    #[doc(hidden)]
    Nek = 4555,
    #[doc(hidden)]
    Nem = 4556,
    #[doc(hidden)]
    Nen = 4557,
    #[doc(hidden)]
    Neo = 4558,
    #[doc(hidden)]
    Nep = 4559,
    #[doc(hidden)]
    Neq = 4560,
    #[doc(hidden)]
    Ner = 4561,
    #[doc(hidden)]
    Nes = 4562,
    #[doc(hidden)]
    Net = 4563,
    #[doc(hidden)]
    Neu = 4564,
    #[doc(hidden)]
    Nev = 4565,
    #[doc(hidden)]
    New = 4566,
    #[doc(hidden)]
    Nex = 4567,
    #[doc(hidden)]
    Ney = 4568,
    #[doc(hidden)]
    Nez = 4569,
    #[doc(hidden)]
    Nfa = 4570,
    #[doc(hidden)]
    Nfd = 4571,
    #[doc(hidden)]
    Nfl = 4572,
    #[doc(hidden)]
    Nfr = 4573,
    #[doc(hidden)]
    Nfu = 4574,
    #[doc(hidden)]
    Nga = 4575,
    #[doc(hidden)]
    Ngb = 4576,
    #[doc(hidden)]
    Ngc = 4577,
    #[doc(hidden)]
    Ngd = 4578,
    #[doc(hidden)]
    Nge = 4579,
    #[doc(hidden)]
    Ngg = 4580,
    #[doc(hidden)]
    Ngh = 4581,
    #[doc(hidden)]
    Ngi = 4582,
    #[doc(hidden)]
    Ngj = 4583,
    #[doc(hidden)]
    Ngk = 4584,
    #[doc(hidden)]
    Ngl = 4585,
    #[doc(hidden)]
    Ngm = 4586,
    #[doc(hidden)]
    Ngn = 4587,
    #[doc(hidden)]
    Ngp = 4588,
    #[doc(hidden)]
    Ngq = 4589,
    #[doc(hidden)]
    Ngr = 4590,
    #[doc(hidden)]
    Ngs = 4591,
    #[doc(hidden)]
    Ngt = 4592,
    #[doc(hidden)]
    Ngu = 4593,
    #[doc(hidden)]
    Ngv = 4594,
    #[doc(hidden)]
    Ngw = 4595,
    #[doc(hidden)]
    Ngx = 4596,
    #[doc(hidden)]
    Ngy = 4597,
    #[doc(hidden)]
    Ngz = 4598,
    #[doc(hidden)]
    Nha = 4599,
    #[doc(hidden)]
    Nhb = 4600,
    #[doc(hidden)]
    Nhc = 4601,
    #[doc(hidden)]
    Nhd = 4602,
    #[doc(hidden)]
    Nhe = 4603,
    #[doc(hidden)]
    Nhf = 4604,
    #[doc(hidden)]
    Nhg = 4605,
    #[doc(hidden)]
    Nhh = 4606,
    #[doc(hidden)]
    Nhi = 4607,
    #[doc(hidden)]
    Nhk = 4608,
    #[doc(hidden)]
    Nhm = 4609,
    #[doc(hidden)]
    Nhn = 4610,
    #[doc(hidden)]
    Nho = 4611,
    #[doc(hidden)]
    Nhp = 4612,
    #[doc(hidden)]
    Nhq = 4613,
    #[doc(hidden)]
    Nhr = 4614,
    #[doc(hidden)]
    Nht = 4615,
    #[doc(hidden)]
    Nhu = 4616,
    #[doc(hidden)]
    Nhv = 4617,
    #[doc(hidden)]
    Nhw = 4618,
    #[doc(hidden)]
    Nhx = 4619,
    #[doc(hidden)]
    Nhy = 4620,
    #[doc(hidden)]
    Nhz = 4621,
    #[doc(hidden)]
    Nia = 4622,
    #[doc(hidden)]
    Nib = 4623,
    #[doc(hidden)]
    Nid = 4624,
    #[doc(hidden)]
    Nie = 4625,
    #[doc(hidden)]
    Nif = 4626,
    #[doc(hidden)]
    Nig = 4627,
    #[doc(hidden)]
    Nih = 4628,
    #[doc(hidden)]
    Nii = 4629,
    #[doc(hidden)]
    Nij = 4630,
    #[doc(hidden)]
    Nik = 4631,
    #[doc(hidden)]
    Nil = 4632,
    #[doc(hidden)]
    Nim = 4633,
    #[doc(hidden)]
    Nin = 4634,
    #[doc(hidden)]
    Nio = 4635,
    #[doc(hidden)]
    Niq = 4636,
    #[doc(hidden)]
    Nir = 4637,
    #[doc(hidden)]
    Nis = 4638,
    #[doc(hidden)]
    Nit = 4639,
    #[doc(hidden)]
    Niu = 4640,
    #[doc(hidden)]
    Niv = 4641,
    #[doc(hidden)]
    Niw = 4642,
    #[doc(hidden)]
    Nix = 4643,
    #[doc(hidden)]
    Niy = 4644,
    #[doc(hidden)]
    Niz = 4645,
    #[doc(hidden)]
    Nja = 4646,
    #[doc(hidden)]
    Njb = 4647,
    #[doc(hidden)]
    Njd = 4648,
    #[doc(hidden)]
    Njh = 4649,
    #[doc(hidden)]
    Nji = 4650,
    #[doc(hidden)]
    Njj = 4651,
    #[doc(hidden)]
    Njl = 4652,
    #[doc(hidden)]
    Njm = 4653,
    #[doc(hidden)]
    Njn = 4654,
    #[doc(hidden)]
    Njo = 4655,
    #[doc(hidden)]
    Njr = 4656,
    #[doc(hidden)]
    Njs = 4657,
    #[doc(hidden)]
    Njt = 4658,
    #[doc(hidden)]
    Nju = 4659,
    #[doc(hidden)]
    Njx = 4660,
    #[doc(hidden)]
    Njy = 4661,
    #[doc(hidden)]
    Njz = 4662,
    #[doc(hidden)]
    Nka = 4663,
    #[doc(hidden)]
    Nkb = 4664,
    #[doc(hidden)]
    Nkc = 4665,
    #[doc(hidden)]
    Nkd = 4666,
    #[doc(hidden)]
    Nke = 4667,
    #[doc(hidden)]
    Nkf = 4668,
    #[doc(hidden)]
    Nkg = 4669,
    #[doc(hidden)]
    Nkh = 4670,
    #[doc(hidden)]
    Nki = 4671,
    #[doc(hidden)]
    Nkj = 4672,
    #[doc(hidden)]
    Nkk = 4673,
    #[doc(hidden)]
    Nkm = 4674,
    #[doc(hidden)]
    Nkn = 4675,
    #[doc(hidden)]
    Nko = 4676,
    #[doc(hidden)]
    Nkp = 4677,
    #[doc(hidden)]
    Nkq = 4678,
    #[doc(hidden)]
    Nkr = 4679,
    #[doc(hidden)]
    Nks = 4680,
    #[doc(hidden)]
    Nkt = 4681,
    #[doc(hidden)]
    Nku = 4682,
    #[doc(hidden)]
    Nkv = 4683,
    #[doc(hidden)]
    Nkw = 4684,
    #[doc(hidden)]
    Nkx = 4685,
    #[doc(hidden)]
    Nkz = 4686,
    #[doc(hidden)]
    Nla = 4687,
    #[doc(hidden)]
    Nlc = 4688,
    #[doc(hidden)]
    Nld = 4689,
    #[doc(hidden)]
    Nle = 4690,
    #[doc(hidden)]
    Nlg = 4691,
    #[doc(hidden)]
    Nli = 4692,
    #[doc(hidden)]
    Nlj = 4693,
    #[doc(hidden)]
    Nlk = 4694,
    #[doc(hidden)]
    Nll = 4695,
    #[doc(hidden)]
    Nlm = 4696,
    #[doc(hidden)]
    Nlo = 4697,
    #[doc(hidden)]
    Nlq = 4698,
    #[doc(hidden)]
    Nlu = 4699,
    #[doc(hidden)]
    Nlv = 4700,
    #[doc(hidden)]
    Nlw = 4701,
    #[doc(hidden)]
    Nlx = 4702,
    #[doc(hidden)]
    Nly = 4703,
    #[doc(hidden)]
    Nlz = 4704,
    #[doc(hidden)]
    Nma = 4705,
    #[doc(hidden)]
    Nmb = 4706,
    #[doc(hidden)]
    Nmc = 4707,
    #[doc(hidden)]
    Nmd = 4708,
    #[doc(hidden)]
    Nme = 4709,
    #[doc(hidden)]
    Nmf = 4710,
    #[doc(hidden)]
    Nmg = 4711,
    #[doc(hidden)]
    Nmh = 4712,
    #[doc(hidden)]
    Nmi = 4713,
    #[doc(hidden)]
    Nmj = 4714,
    #[doc(hidden)]
    Nmk = 4715,
    #[doc(hidden)]
    Nml = 4716,
    #[doc(hidden)]
    Nmm = 4717,
    #[doc(hidden)]
    Nmn = 4718,
    #[doc(hidden)]
    Nmo = 4719,
    #[doc(hidden)]
    Nmp = 4720,
    #[doc(hidden)]
    Nmq = 4721,
    #[doc(hidden)]
    Nmr = 4722,
    #[doc(hidden)]
    Nms = 4723,
    #[doc(hidden)]
    Nmt = 4724,
    #[doc(hidden)]
    Nmu = 4725,
    #[doc(hidden)]
    Nmv = 4726,
    #[doc(hidden)]
    Nmw = 4727,
    #[doc(hidden)]
    Nmx = 4728,
    #[doc(hidden)]
    Nmy = 4729,
    #[doc(hidden)]
    Nmz = 4730,
    #[doc(hidden)]
    Nna = 4731,
    #[doc(hidden)]
    Nnb = 4732,
    #[doc(hidden)]
    Nnc = 4733,
    #[doc(hidden)]
    Nnd = 4734,
    #[doc(hidden)]
    Nne = 4735,
    #[doc(hidden)]
    Nnf = 4736,
    #[doc(hidden)]
    Nng = 4737,
    #[doc(hidden)]
    Nnh = 4738,
    #[doc(hidden)]
    Nni = 4739,
    #[doc(hidden)]
    Nnj = 4740,
    #[doc(hidden)]
    Nnk = 4741,
    #[doc(hidden)]
    Nnl = 4742,
    #[doc(hidden)]
    Nnm = 4743,
    #[doc(hidden)]
    Nnn = 4744,
    #[doc(hidden)]
    Nno = 4745,
    #[doc(hidden)]
    Nnp = 4746,
    #[doc(hidden)]
    Nnq = 4747,
    #[doc(hidden)]
    Nnr = 4748,
    #[doc(hidden)]
    Nnt = 4749,
    #[doc(hidden)]
    Nnu = 4750,
    #[doc(hidden)]
    Nnv = 4751,
    #[doc(hidden)]
    Nnw = 4752,
    #[doc(hidden)]
    Nny = 4753,
    #[doc(hidden)]
    Nnz = 4754,
    #[doc(hidden)]
    Noa = 4755,
    #[doc(hidden)]
    Nob = 4756,
    #[doc(hidden)]
    Noc = 4757,
    #[doc(hidden)]
    Nod = 4758,
    #[doc(hidden)]
    Noe = 4759,
    #[doc(hidden)]
    Nof = 4760,
    #[doc(hidden)]
    Nog = 4761,
    #[doc(hidden)]
    Noh = 4762,
    #[doc(hidden)]
    Noi = 4763,
    #[doc(hidden)]
    Noj = 4764,
    #[doc(hidden)]
    Nok = 4765,
    #[doc(hidden)]
    Nol = 4766,
    #[doc(hidden)]
    Nom = 4767,
    #[doc(hidden)]
    Non = 4768,
    #[doc(hidden)]
    Nop = 4769,
    #[doc(hidden)]
    Noq = 4770,
    #[doc(hidden)]
    Nor = 4771,
    #[doc(hidden)]
    Nos = 4772,
    #[doc(hidden)]
    Not = 4773,
    #[doc(hidden)]
    Nou = 4774,
    #[doc(hidden)]
    Nov = 4775,
    #[doc(hidden)]
    Now = 4776,
    #[doc(hidden)]
    Noy = 4777,
    #[doc(hidden)]
    Noz = 4778,
    #[doc(hidden)]
    Npa = 4779,
    #[doc(hidden)]
    Npb = 4780,
    #[doc(hidden)]
    Npg = 4781,
    #[doc(hidden)]
    Nph = 4782,
    #[doc(hidden)]
    Npi = 4783,
    #[doc(hidden)]
    Npl = 4784,
    #[doc(hidden)]
    Npn = 4785,
    #[doc(hidden)]
    Npo = 4786,
    #[doc(hidden)]
    Nps = 4787,
    #[doc(hidden)]
    Npu = 4788,
    #[doc(hidden)]
    Npx = 4789,
    #[doc(hidden)]
    Npy = 4790,
    #[doc(hidden)]
    Nqg = 4791,
    #[doc(hidden)]
    Nqk = 4792,
    #[doc(hidden)]
    Nql = 4793,
    #[doc(hidden)]
    Nqm = 4794,
    #[doc(hidden)]
    Nqn = 4795,
    #[doc(hidden)]
    Nqo = 4796,
    #[doc(hidden)]
    Nqq = 4797,
    #[doc(hidden)]
    Nqt = 4798,
    #[doc(hidden)]
    Nqy = 4799,
    #[doc(hidden)]
    Nra = 4800,
    #[doc(hidden)]
    Nrb = 4801,
    #[doc(hidden)]
    Nrc = 4802,
    #[doc(hidden)]
    Nre = 4803,
    #[doc(hidden)]
    Nrf = 4804,
    #[doc(hidden)]
    Nrg = 4805,
    #[doc(hidden)]
    Nri = 4806,
    #[doc(hidden)]
    Nrk = 4807,
    #[doc(hidden)]
    Nrl = 4808,
    #[doc(hidden)]
    Nrm = 4809,
    #[doc(hidden)]
    Nrn = 4810,
    #[doc(hidden)]
    Nrp = 4811,
    #[doc(hidden)]
    Nrr = 4812,
    #[doc(hidden)]
    Nrt = 4813,
    #[doc(hidden)]
    Nru = 4814,
    #[doc(hidden)]
    Nrx = 4815,
    #[doc(hidden)]
    Nrz = 4816,
    #[doc(hidden)]
    Nsa = 4817,
    #[doc(hidden)]
    Nsb = 4818,
    #[doc(hidden)]
    Nsc = 4819,
    #[doc(hidden)]
    Nsd = 4820,
    #[doc(hidden)]
    Nse = 4821,
    #[doc(hidden)]
    Nsf = 4822,
    #[doc(hidden)]
    Nsg = 4823,
    #[doc(hidden)]
    Nsh = 4824,
    #[doc(hidden)]
    Nsi = 4825,
    #[doc(hidden)]
    Nsk = 4826,
    #[doc(hidden)]
    Nsl = 4827,
    #[doc(hidden)]
    Nsm = 4828,
    #[doc(hidden)]
    Nsn = 4829,
    #[doc(hidden)]
    Nso = 4830,
    #[doc(hidden)]
    Nsp = 4831,
    #[doc(hidden)]
    Nsq = 4832,
    #[doc(hidden)]
    Nsr = 4833,
    #[doc(hidden)]
    Nss = 4834,
    #[doc(hidden)]
    Nst = 4835,
    #[doc(hidden)]
    Nsu = 4836,
    #[doc(hidden)]
    Nsv = 4837,
    #[doc(hidden)]
    Nsw = 4838,
    #[doc(hidden)]
    Nsx = 4839,
    #[doc(hidden)]
    Nsy = 4840,
    #[doc(hidden)]
    Nsz = 4841,
    #[doc(hidden)]
    Ntd = 4842,
    #[doc(hidden)]
    Nte = 4843,
    #[doc(hidden)]
    Ntg = 4844,
    #[doc(hidden)]
    Nti = 4845,
    #[doc(hidden)]
    Ntj = 4846,
    #[doc(hidden)]
    Ntk = 4847,
    #[doc(hidden)]
    Ntm = 4848,
    #[doc(hidden)]
    Nto = 4849,
    #[doc(hidden)]
    Ntp = 4850,
    #[doc(hidden)]
    Ntr = 4851,
    #[doc(hidden)]
    Ntu = 4852,
    #[doc(hidden)]
    Ntw = 4853,
    #[doc(hidden)]
    Ntx = 4854,
    #[doc(hidden)]
    Nty = 4855,
    #[doc(hidden)]
    Ntz = 4856,
    #[doc(hidden)]
    Nua = 4857,
    #[doc(hidden)]
    Nuc = 4858,
    #[doc(hidden)]
    Nud = 4859,
    #[doc(hidden)]
    Nue = 4860,
    #[doc(hidden)]
    Nuf = 4861,
    #[doc(hidden)]
    Nug = 4862,
    #[doc(hidden)]
    Nuh = 4863,
    #[doc(hidden)]
    Nui = 4864,
    #[doc(hidden)]
    Nuj = 4865,
    #[doc(hidden)]
    Nuk = 4866,
    #[doc(hidden)]
    Nul = 4867,
    #[doc(hidden)]
    Num = 4868,
    #[doc(hidden)]
    Nun = 4869,
    #[doc(hidden)]
    Nuo = 4870,
    #[doc(hidden)]
    Nup = 4871,
    #[doc(hidden)]
    Nuq = 4872,
    #[doc(hidden)]
    Nur = 4873,
    #[doc(hidden)]
    Nus = 4874,
    #[doc(hidden)]
    Nut = 4875,
    #[doc(hidden)]
    Nuu = 4876,
    #[doc(hidden)]
    Nuv = 4877,
    #[doc(hidden)]
    Nuw = 4878,
    #[doc(hidden)]
    Nux = 4879,
    #[doc(hidden)]
    Nuy = 4880,
    #[doc(hidden)]
    Nuz = 4881,
    #[doc(hidden)]
    Nvh = 4882,
    #[doc(hidden)]
    Nvm = 4883,
    #[doc(hidden)]
    Nvo = 4884,
    #[doc(hidden)]
    Nwa = 4885,
    #[doc(hidden)]
    Nwb = 4886,
    #[doc(hidden)]
    Nwc = 4887,
    #[doc(hidden)]
    Nwe = 4888,
    #[doc(hidden)]
    Nwg = 4889,
    #[doc(hidden)]
    Nwi = 4890,
    #[doc(hidden)]
    Nwm = 4891,
    #[doc(hidden)]
    Nwo = 4892,
    #[doc(hidden)]
    Nwr = 4893,
    #[doc(hidden)]
    Nww = 4894,
    #[doc(hidden)]
    Nwx = 4895,
    #[doc(hidden)]
    Nwy = 4896,
    #[doc(hidden)]
    Nxa = 4897,
    #[doc(hidden)]
    Nxd = 4898,
    #[doc(hidden)]
    Nxe = 4899,
    #[doc(hidden)]
    Nxg = 4900,
    #[doc(hidden)]
    Nxi = 4901,
    #[doc(hidden)]
    Nxk = 4902,
    #[doc(hidden)]
    Nxl = 4903,
    #[doc(hidden)]
    Nxm = 4904,
    #[doc(hidden)]
    Nxn = 4905,
    #[doc(hidden)]
    Nxo = 4906,
    #[doc(hidden)]
    Nxq = 4907,
    #[doc(hidden)]
    Nxr = 4908,
    #[doc(hidden)]
    Nxx = 4909,
    #[doc(hidden)]
    Nya = 4910,
    #[doc(hidden)]
    Nyb = 4911,
    #[doc(hidden)]
    Nyc = 4912,
    #[doc(hidden)]
    Nyd = 4913,
    #[doc(hidden)]
    Nye = 4914,
    #[doc(hidden)]
    Nyf = 4915,
    #[doc(hidden)]
    Nyg = 4916,
    #[doc(hidden)]
    Nyh = 4917,
    #[doc(hidden)]
    Nyi = 4918,
    #[doc(hidden)]
    Nyj = 4919,
    #[doc(hidden)]
    Nyk = 4920,
    #[doc(hidden)]
    Nyl = 4921,
    #[doc(hidden)]
    Nym = 4922,
    #[doc(hidden)]
    Nyn = 4923,
    #[doc(hidden)]
    Nyo = 4924,
    #[doc(hidden)]
    Nyp = 4925,
    #[doc(hidden)]
    Nyq = 4926,
    #[doc(hidden)]
    Nyr = 4927,
    #[doc(hidden)]
    Nys = 4928,
    #[doc(hidden)]
    Nyt = 4929,
    #[doc(hidden)]
    Nyu = 4930,
    #[doc(hidden)]
    Nyv = 4931,
    #[doc(hidden)]
    Nyw = 4932,
    #[doc(hidden)]
    Nyx = 4933,
    #[doc(hidden)]
    Nyy = 4934,
    #[doc(hidden)]
    Nza = 4935,
    #[doc(hidden)]
    Nzb = 4936,
    #[doc(hidden)]
    Nzd = 4937,
    #[doc(hidden)]
    Nzi = 4938,
    #[doc(hidden)]
    Nzk = 4939,
    #[doc(hidden)]
    Nzm = 4940,
    #[doc(hidden)]
    Nzs = 4941,
    #[doc(hidden)]
    Nzu = 4942,
    #[doc(hidden)]
    Nzy = 4943,
    #[doc(hidden)]
    Nzz = 4944,
    #[doc(hidden)]
    Oaa = 4945,
    #[doc(hidden)]
    Oac = 4946,
    #[doc(hidden)]
    Oar = 4947,
    #[doc(hidden)]
    Oav = 4948,
    #[doc(hidden)]
    Obi = 4949,
    #[doc(hidden)]
    Obk = 4950,
    #[doc(hidden)]
    Obl = 4951,
    #[doc(hidden)]
    Obm = 4952,
    #[doc(hidden)]
    Obo = 4953,
    #[doc(hidden)]
    Obr = 4954,
    #[doc(hidden)]
    Obt = 4955,
    #[doc(hidden)]
    Obu = 4956,
    #[doc(hidden)]
    Oca = 4957,
    #[doc(hidden)]
    Och = 4958,
    #[doc(hidden)]
    Oci = 4959,
    #[doc(hidden)]
    Ocm = 4960,
    #[doc(hidden)]
    Oco = 4961,
    #[doc(hidden)]
    Ocu = 4962,
    #[doc(hidden)]
    Oda = 4963,
    #[doc(hidden)]
    Odk = 4964,
    #[doc(hidden)]
    Odt = 4965,
    #[doc(hidden)]
    Odu = 4966,
    #[doc(hidden)]
    Ofo = 4967,
    #[doc(hidden)]
    Ofs = 4968,
    #[doc(hidden)]
    Ofu = 4969,
    #[doc(hidden)]
    Ogb = 4970,
    #[doc(hidden)]
    Ogc = 4971,
    #[doc(hidden)]
    Oge = 4972,
    #[doc(hidden)]
    Ogg = 4973,
    #[doc(hidden)]
    Ogo = 4974,
    #[doc(hidden)]
    Ogu = 4975,
    #[doc(hidden)]
    Oht = 4976,
    #[doc(hidden)]
    Ohu = 4977,
    #[doc(hidden)]
    Oia = 4978,
    #[doc(hidden)]
    Oie = 4979,
    #[doc(hidden)]
    Oin = 4980,
    #[doc(hidden)]
    Ojb = 4981,
    #[doc(hidden)]
    Ojc = 4982,
    #[doc(hidden)]
    Ojg = 4983,
    #[doc(hidden)]
    Oji = 4984,
    #[doc(hidden)]
    Ojp = 4985,
    #[doc(hidden)]
    Ojs = 4986,
    #[doc(hidden)]
    Ojv = 4987,
    #[doc(hidden)]
    Ojw = 4988,
    #[doc(hidden)]
    Oka = 4989,
    #[doc(hidden)]
    Okb = 4990,
    #[doc(hidden)]
    Okc = 4991,
    #[doc(hidden)]
    Okd = 4992,
    #[doc(hidden)]
    Oke = 4993,
    #[doc(hidden)]
    Okg = 4994,
    #[doc(hidden)]
    Okh = 4995,
    #[doc(hidden)]
    Oki = 4996,
    #[doc(hidden)]
    Okj = 4997,
    #[doc(hidden)]
    Okk = 4998,
    #[doc(hidden)]
    Okl = 4999,
    #[doc(hidden)]
    Okm = 5000,
    #[doc(hidden)]
    Okn = 5001,
    #[doc(hidden)]
    Oko = 5002,
    #[doc(hidden)]
    Okr = 5003,
    #[doc(hidden)]
    Oks = 5004,
    #[doc(hidden)]
    Oku = 5005,
    #[doc(hidden)]
    Okv = 5006,
    #[doc(hidden)]
    Okx = 5007,
    #[doc(hidden)]
    Okz = 5008,
    #[doc(hidden)]
    Ola = 5009,
    #[doc(hidden)]
    Old = 5010,
    #[doc(hidden)]
    Ole = 5011,
    #[doc(hidden)]
    Olk = 5012,
    #[doc(hidden)]
    Olm = 5013,
    #[doc(hidden)]
    Olo = 5014,
    #[doc(hidden)]
    Olr = 5015,
    #[doc(hidden)]
    Olt = 5016,
    #[doc(hidden)]
    Olu = 5017,
    #[doc(hidden)]
    Oma = 5018,
    #[doc(hidden)]
    Omb = 5019,
    #[doc(hidden)]
    Omc = 5020,
    #[doc(hidden)]
    Omg = 5021,
    #[doc(hidden)]
    Omi = 5022,
    #[doc(hidden)]
    Omk = 5023,
    #[doc(hidden)]
    Oml = 5024,
    #[doc(hidden)]
    Omn = 5025,
    #[doc(hidden)]
    Omo = 5026,
    #[doc(hidden)]
    Omp = 5027,
    #[doc(hidden)]
    Omr = 5028,
    #[doc(hidden)]
    Omt = 5029,
    #[doc(hidden)]
    Omu = 5030,
    #[doc(hidden)]
    Omw = 5031,
    #[doc(hidden)]
    Omx = 5032,
    #[doc(hidden)]
    Omy = 5033,
    #[doc(hidden)]
    Ona = 5034,
    #[doc(hidden)]
    Onb = 5035,
    #[doc(hidden)]
    One = 5036,
    #[doc(hidden)]
    Ong = 5037,
    #[doc(hidden)]
    Oni = 5038,
    #[doc(hidden)]
    Onj = 5039,
    #[doc(hidden)]
    Onk = 5040,
    #[doc(hidden)]
    Onn = 5041,
    #[doc(hidden)]
    Ono = 5042,
    #[doc(hidden)]
    Onp = 5043,
    #[doc(hidden)]
    Onr = 5044,
    #[doc(hidden)]
    Ons = 5045,
    #[doc(hidden)]
    Ont = 5046,
    #[doc(hidden)]
    Onu = 5047,
    #[doc(hidden)]
    Onw = 5048,
    #[doc(hidden)]
    Onx = 5049,
    #[doc(hidden)]
    Ood = 5050,
    #[doc(hidden)]
    Oog = 5051,
    #[doc(hidden)]
    Oon = 5052,
    #[doc(hidden)]
    Oor = 5053,
    #[doc(hidden)]
    Oos = 5054,
    #[doc(hidden)]
    Opa = 5055,
    #[doc(hidden)]
    Opk = 5056,
    #[doc(hidden)]
    Opm = 5057,
    #[doc(hidden)]
    Opo = 5058,
    #[doc(hidden)]
    Opt = 5059,
    #[doc(hidden)]
    Opy = 5060,
    #[doc(hidden)]
    Ora = 5061,
    #[doc(hidden)]
    Orc = 5062,
    #[doc(hidden)]
    Ore = 5063,
    #[doc(hidden)]
    Org = 5064,
    #[doc(hidden)]
    Orh = 5065,
    #[doc(hidden)]
    Ori = 5066,
    #[doc(hidden)]
    Orm = 5067,
    #[doc(hidden)]
    Orn = 5068,
    #[doc(hidden)]
    Oro = 5069,
    #[doc(hidden)]
    Orr = 5070,
    #[doc(hidden)]
    Ors = 5071,
    #[doc(hidden)]
    Ort = 5072,
    #[doc(hidden)]
    Oru = 5073,
    #[doc(hidden)]
    Orv = 5074,
    #[doc(hidden)]
    Orw = 5075,
    #[doc(hidden)]
    Orx = 5076,
    #[doc(hidden)]
    Ory = 5077,
    #[doc(hidden)]
    Orz = 5078,
    #[doc(hidden)]
    Osa = 5079,
    #[doc(hidden)]
    Osc = 5080,
    #[doc(hidden)]
    Osi = 5081,
    #[doc(hidden)]
    Osn = 5082,
    #[doc(hidden)]
    Oso = 5083,
    #[doc(hidden)]
    Osp = 5084,
    #[doc(hidden)]
    Oss = 5085,
    #[doc(hidden)]
    Ost = 5086,
    #[doc(hidden)]
    Osu = 5087,
    #[doc(hidden)]
    Osx = 5088,
    #[doc(hidden)]
    Ota = 5089,
    #[doc(hidden)]
    Otb = 5090,
    #[doc(hidden)]
    Otd = 5091,
    #[doc(hidden)]
    Ote = 5092,
    #[doc(hidden)]
    Oti = 5093,
    #[doc(hidden)]
    Otk = 5094,
    #[doc(hidden)]
    Otl = 5095,
    #[doc(hidden)]
    Otm = 5096,
    #[doc(hidden)]
    Otn = 5097,
    #[doc(hidden)]
    Otq = 5098,
    #[doc(hidden)]
    Otr = 5099,
    #[doc(hidden)]
    Ots = 5100,
    #[doc(hidden)]
    Ott = 5101,
    #[doc(hidden)]
    Otu = 5102,
    #[doc(hidden)]
    Otw = 5103,
    #[doc(hidden)]
    Otx = 5104,
    #[doc(hidden)]
    Oty = 5105,
    #[doc(hidden)]
    Otz = 5106,
    #[doc(hidden)]
    Oua = 5107,
    #[doc(hidden)]
    Oub = 5108,
    #[doc(hidden)]
    Oue = 5109,
    #[doc(hidden)]
    Oui = 5110,
    #[doc(hidden)]
    Oum = 5111,
    #[doc(hidden)]
    Ovd = 5112,
    #[doc(hidden)]
    Owi = 5113,
    #[doc(hidden)]
    Owl = 5114,
    #[doc(hidden)]
    Oyb = 5115,
    #[doc(hidden)]
    Oyd = 5116,
    #[doc(hidden)]
    Oym = 5117,
    #[doc(hidden)]
    Oyy = 5118,
    #[doc(hidden)]
    Ozm = 5119,
    #[doc(hidden)]
    Pab = 5120,
    #[doc(hidden)]
    Pac = 5121,
    #[doc(hidden)]
    Pad = 5122,
    #[doc(hidden)]
    Pae = 5123,
    #[doc(hidden)]
    Paf = 5124,
    #[doc(hidden)]
    Pag = 5125,
    #[doc(hidden)]
    Pah = 5126,
    #[doc(hidden)]
    Pai = 5127,
    #[doc(hidden)]
    Pak = 5128,
    #[doc(hidden)]
    Pal = 5129,
    #[doc(hidden)]
    Pam = 5130,
    #[doc(hidden)]
    Pan = 5131,
    #[doc(hidden)]
    Pao = 5132,
    #[doc(hidden)]
    Pap = 5133,
    #[doc(hidden)]
    Paq = 5134,
    #[doc(hidden)]
    Par = 5135,
    #[doc(hidden)]
    Pas = 5136,
    #[doc(hidden)]
    Pau = 5137,
    #[doc(hidden)]
    Pav = 5138,
    #[doc(hidden)]
    Paw = 5139,
    #[doc(hidden)]
    Pax = 5140,
    #[doc(hidden)]
    Pay = 5141,
    #[doc(hidden)]
    Paz = 5142,
    #[doc(hidden)]
    Pbb = 5143,
    #[doc(hidden)]
    Pbc = 5144,
    #[doc(hidden)]
    Pbe = 5145,
    #[doc(hidden)]
    Pbf = 5146,
    #[doc(hidden)]
    Pbg = 5147,
    #[doc(hidden)]
    Pbh = 5148,
    #[doc(hidden)]
    Pbi = 5149,
    #[doc(hidden)]
    Pbl = 5150,
    #[doc(hidden)]
    Pbm = 5151,
    #[doc(hidden)]
    Pbn = 5152,
    #[doc(hidden)]
    Pbo = 5153,
    #[doc(hidden)]
    Pbp = 5154,
    #[doc(hidden)]
    Pbr = 5155,
    #[doc(hidden)]
    Pbs = 5156,
    #[doc(hidden)]
    Pbt = 5157,
    #[doc(hidden)]
    Pbu = 5158,
    #[doc(hidden)]
    Pbv = 5159,
    #[doc(hidden)]
    Pby = 5160,
    #[doc(hidden)]
    Pca = 5161,
    #[doc(hidden)]
    Pcb = 5162,
    #[doc(hidden)]
    Pcc = 5163,
    #[doc(hidden)]
    Pcd = 5164,
    #[doc(hidden)]
    Pce = 5165,
    #[doc(hidden)]
    Pcf = 5166,
    #[doc(hidden)]
    Pcg = 5167,
    #[doc(hidden)]
    Pch = 5168,
    #[doc(hidden)]
    Pci = 5169,
    #[doc(hidden)]
    Pcj = 5170,
    #[doc(hidden)]
    Pck = 5171,
    #[doc(hidden)]
    Pcl = 5172,
    #[doc(hidden)]
    Pcm = 5173,
    #[doc(hidden)]
    Pcn = 5174,
    #[doc(hidden)]
    Pcp = 5175,
    #[doc(hidden)]
    Pcw = 5176,
    #[doc(hidden)]
    Pda = 5177,
    #[doc(hidden)]
    Pdc = 5178,
    #[doc(hidden)]
    Pdi = 5179,
    #[doc(hidden)]
    Pdn = 5180,
    #[doc(hidden)]
    Pdo = 5181,
    #[doc(hidden)]
    Pdt = 5182,
    #[doc(hidden)]
    Pdu = 5183,
    #[doc(hidden)]
    Pea = 5184,
    #[doc(hidden)]
    Peb = 5185,
    #[doc(hidden)]
    Ped = 5186,
    #[doc(hidden)]
    Pee = 5187,
    #[doc(hidden)]
    Pef = 5188,
    #[doc(hidden)]
    Peg = 5189,
    #[doc(hidden)]
    Peh = 5190,
    #[doc(hidden)]
    Pei = 5191,
    #[doc(hidden)]
    Pej = 5192,
    #[doc(hidden)]
    Pek = 5193,
    #[doc(hidden)]
    Pel = 5194,
    #[doc(hidden)]
    Pem = 5195,
    #[doc(hidden)]
    Peo = 5196,
    #[doc(hidden)]
    Pep = 5197,
    #[doc(hidden)]
    Peq = 5198,
    #[doc(hidden)]
    Pes = 5199,
    #[doc(hidden)]
    Pev = 5200,
    #[doc(hidden)]
    Pex = 5201,
    #[doc(hidden)]
    Pey = 5202,
    #[doc(hidden)]
    Pez = 5203,
    #[doc(hidden)]
    Pfa = 5204,
    #[doc(hidden)]
    Pfe = 5205,
    #[doc(hidden)]
    Pfl = 5206,
    #[doc(hidden)]
    Pga = 5207,
    #[doc(hidden)]
    Pgd = 5208,
    #[doc(hidden)]
    Pgg = 5209,
    #[doc(hidden)]
    Pgi = 5210,
    #[doc(hidden)]
    Pgk = 5211,
    #[doc(hidden)]
    Pgl = 5212,
    #[doc(hidden)]
    Pgn = 5213,
    #[doc(hidden)]
    Pgs = 5214,
    #[doc(hidden)]
    Pgu = 5215,
    #[doc(hidden)]
    Pgz = 5216,
    #[doc(hidden)]
    Pha = 5217,
    #[doc(hidden)]
    Phd = 5218,
    #[doc(hidden)]
    Phg = 5219,
    #[doc(hidden)]
    Phh = 5220,
    #[doc(hidden)]
    Phj = 5221,
    #[doc(hidden)]
    Phk = 5222,
    #[doc(hidden)]
    Phl = 5223,
    #[doc(hidden)]
    Phm = 5224,
    #[doc(hidden)]
    Phn = 5225,
    #[doc(hidden)]
    Pho = 5226,
    #[doc(hidden)]
    Phq = 5227,
    #[doc(hidden)]
    Phr = 5228,
    #[doc(hidden)]
    Pht = 5229,
    #[doc(hidden)]
    Phu = 5230,
    #[doc(hidden)]
    Phv = 5231,
    #[doc(hidden)]
    Phw = 5232,
    #[doc(hidden)]
    Pia = 5233,
    #[doc(hidden)]
    Pib = 5234,
    #[doc(hidden)]
    Pic = 5235,
    #[doc(hidden)]
    Pid = 5236,
    #[doc(hidden)]
    Pie = 5237,
    #[doc(hidden)]
    Pif = 5238,
    #[doc(hidden)]
    Pig = 5239,
    #[doc(hidden)]
    Pih = 5240,
    #[doc(hidden)]
    Pij = 5241,
    #[doc(hidden)]
    Pil = 5242,
    #[doc(hidden)]
    Pim = 5243,
    #[doc(hidden)]
    Pin = 5244,
    #[doc(hidden)]
    Pio = 5245,
    #[doc(hidden)]
    Pip = 5246,
    #[doc(hidden)]
    Pir = 5247,
    #[doc(hidden)]
    Pis = 5248,
    #[doc(hidden)]
    Pit = 5249,
    #[doc(hidden)]
    Piu = 5250,
    #[doc(hidden)]
    Piv = 5251,
    #[doc(hidden)]
    Piw = 5252,
    #[doc(hidden)]
    Pix = 5253,
    #[doc(hidden)]
    Piy = 5254,
    #[doc(hidden)]
    Piz = 5255,
    #[doc(hidden)]
    Pjt = 5256,
    #[doc(hidden)]
    Pka = 5257,
    #[doc(hidden)]
    Pkb = 5258,
    #[doc(hidden)]
    Pkc = 5259,
    #[doc(hidden)]
    Pkg = 5260,
    #[doc(hidden)]
    Pkh = 5261,
    #[doc(hidden)]
    Pkn = 5262,
    #[doc(hidden)]
    Pko = 5263,
    #[doc(hidden)]
    Pkp = 5264,
    #[doc(hidden)]
    Pkr = 5265,
    #[doc(hidden)]
    Pks = 5266,
    #[doc(hidden)]
    Pkt = 5267,
    #[doc(hidden)]
    Pku = 5268,
    #[doc(hidden)]
    Pla = 5269,
    #[doc(hidden)]
    Plb = 5270,
    #[doc(hidden)]
    Plc = 5271,
    #[doc(hidden)]
    Pld = 5272,
    #[doc(hidden)]
    Ple = 5273,
    #[doc(hidden)]
    Plg = 5274,
    #[doc(hidden)]
    Plh = 5275,
    #[doc(hidden)]
    Pli = 5276,
    #[doc(hidden)]
    Plj = 5277,
    #[doc(hidden)]
    Plk = 5278,
    #[doc(hidden)]
    Pll = 5279,
    #[doc(hidden)]
    Pln = 5280,
    #[doc(hidden)]
    Plo = 5281,
    #[doc(hidden)]
    Plq = 5282,
    #[doc(hidden)]
    Plr = 5283,
    #[doc(hidden)]
    Pls = 5284,
    #[doc(hidden)]
    Plt = 5285,
    #[doc(hidden)]
    Plu = 5286,
    #[doc(hidden)]
    Plv = 5287,
    #[doc(hidden)]
    Plw = 5288,
    #[doc(hidden)]
    Ply = 5289,
    #[doc(hidden)]
    Plz = 5290,
    #[doc(hidden)]
    Pma = 5291,
    #[doc(hidden)]
    Pmb = 5292,
    #[doc(hidden)]
    Pmd = 5293,
    #[doc(hidden)]
    Pme = 5294,
    #[doc(hidden)]
    Pmf = 5295,
    #[doc(hidden)]
    Pmh = 5296,
    #[doc(hidden)]
    Pmi = 5297,
    #[doc(hidden)]
    Pmj = 5298,
    #[doc(hidden)]
    Pmk = 5299,
    #[doc(hidden)]
    Pml = 5300,
    #[doc(hidden)]
    Pmm = 5301,
    #[doc(hidden)]
    Pmn = 5302,
    #[doc(hidden)]
    Pmo = 5303,
    #[doc(hidden)]
    Pmq = 5304,
    #[doc(hidden)]
    Pmr = 5305,
    #[doc(hidden)]
    Pms = 5306,
    #[doc(hidden)]
    Pmt = 5307,
    #[doc(hidden)]
    Pmw = 5308,
    #[doc(hidden)]
    Pmx = 5309,
    #[doc(hidden)]
    Pmy = 5310,
    #[doc(hidden)]
    Pmz = 5311,
    #[doc(hidden)]
    Pna = 5312,
    #[doc(hidden)]
    Pnb = 5313,
    #[doc(hidden)]
    Pnc = 5314,
    #[doc(hidden)]
    Pnd = 5315,
    #[doc(hidden)]
    Pne = 5316,
    #[doc(hidden)]
    Png = 5317,
    #[doc(hidden)]
    Pnh = 5318,
    #[doc(hidden)]
    Pni = 5319,
    #[doc(hidden)]
    Pnj = 5320,
    #[doc(hidden)]
    Pnk = 5321,
    #[doc(hidden)]
    Pnl = 5322,
    #[doc(hidden)]
    Pnm = 5323,
    #[doc(hidden)]
    Pnn = 5324,
    #[doc(hidden)]
    Pno = 5325,
    #[doc(hidden)]
    Pnp = 5326,
    #[doc(hidden)]
    Pnq = 5327,
    #[doc(hidden)]
    Pnr = 5328,
    #[doc(hidden)]
    Pns = 5329,
    #[doc(hidden)]
    Pnt = 5330,
    #[doc(hidden)]
    Pnu = 5331,
    #[doc(hidden)]
    Pnv = 5332,
    #[doc(hidden)]
    Pnw = 5333,
    #[doc(hidden)]
    Pnx = 5334,
    #[doc(hidden)]
    Pny = 5335,
    #[doc(hidden)]
    Pnz = 5336,
    #[doc(hidden)]
    Poc = 5337,
    #[doc(hidden)]
    Poe = 5338,
    #[doc(hidden)]
    Pof = 5339,
    #[doc(hidden)]
    Pog = 5340,
    #[doc(hidden)]
    Poh = 5341,
    #[doc(hidden)]
    Poi = 5342,
    #[doc(hidden)]
    Pok = 5343,
    #[doc(hidden)]
    Pol = 5344,
    #[doc(hidden)]
    Pom = 5345,
    #[doc(hidden)]
    Pon = 5346,
    #[doc(hidden)]
    Poo = 5347,
    #[doc(hidden)]
    Pop = 5348,
    #[doc(hidden)]
    Poq = 5349,
    #[doc(hidden)]
    Por = 5350,
    #[doc(hidden)]
    Pos = 5351,
    #[doc(hidden)]
    Pot = 5352,
    #[doc(hidden)]
    Pov = 5353,
    #[doc(hidden)]
    Pow = 5354,
    #[doc(hidden)]
    Pox = 5355,
    #[doc(hidden)]
    Poy = 5356,
    #[doc(hidden)]
    Ppe = 5357,
    #[doc(hidden)]
    Ppi = 5358,
    #[doc(hidden)]
    Ppk = 5359,
    #[doc(hidden)]
    Ppl = 5360,
    #[doc(hidden)]
    Ppm = 5361,
    #[doc(hidden)]
    Ppn = 5362,
    #[doc(hidden)]
    Ppo = 5363,
    #[doc(hidden)]
    Ppp = 5364,
    #[doc(hidden)]
    Ppq = 5365,
    #[doc(hidden)]
    Pps = 5366,
    #[doc(hidden)]
    Ppt = 5367,
    #[doc(hidden)]
    Ppu = 5368,
    #[doc(hidden)]
    Pqa = 5369,
    #[doc(hidden)]
    Pqm = 5370,
    #[doc(hidden)]
    Prc = 5371,
    #[doc(hidden)]
    Prd = 5372,
    #[doc(hidden)]
    Pre = 5373,
    #[doc(hidden)]
    Prf = 5374,
    #[doc(hidden)]
    Prg = 5375,
    #[doc(hidden)]
    Prh = 5376,
    #[doc(hidden)]
    Pri = 5377,
    #[doc(hidden)]
    Prk = 5378,
    #[doc(hidden)]
    Prl = 5379,
    #[doc(hidden)]
    Prm = 5380,
    #[doc(hidden)]
    Prn = 5381,
    #[doc(hidden)]
    Pro = 5382,
    #[doc(hidden)]
    Prp = 5383,
    #[doc(hidden)]
    Prq = 5384,
    #[doc(hidden)]
    Prr = 5385,
    #[doc(hidden)]
    Prs = 5386,
    #[doc(hidden)]
    Prt = 5387,
    #[doc(hidden)]
    Pru = 5388,
    #[doc(hidden)]
    Prw = 5389,
    #[doc(hidden)]
    Prx = 5390,
    #[doc(hidden)]
    Prz = 5391,
    #[doc(hidden)]
    Psa = 5392,
    #[doc(hidden)]
    Psc = 5393,
    #[doc(hidden)]
    Psd = 5394,
    #[doc(hidden)]
    Pse = 5395,
    #[doc(hidden)]
    Psg = 5396,
    #[doc(hidden)]
    Psh = 5397,
    #[doc(hidden)]
    Psi = 5398,
    #[doc(hidden)]
    Psl = 5399,
    #[doc(hidden)]
    Psm = 5400,
    #[doc(hidden)]
    Psn = 5401,
    #[doc(hidden)]
    Pso = 5402,
    #[doc(hidden)]
    Psp = 5403,
    #[doc(hidden)]
    Psq = 5404,
    #[doc(hidden)]
    Psr = 5405,
    #[doc(hidden)]
    Pss = 5406,
    #[doc(hidden)]
    Pst = 5407,
    #[doc(hidden)]
    Psu = 5408,
    #[doc(hidden)]
    Psw = 5409,
    #[doc(hidden)]
    Psy = 5410,
    #[doc(hidden)]
    Pta = 5411,
    #[doc(hidden)]
    Pth = 5412,
    #[doc(hidden)]
    Pti = 5413,
    #[doc(hidden)]
    Ptn = 5414,
    #[doc(hidden)]
    Pto = 5415,
    #[doc(hidden)]
    Ptp = 5416,
    #[doc(hidden)]
    Ptq = 5417,
    #[doc(hidden)]
    Ptr = 5418,
    #[doc(hidden)]
    Ptt = 5419,
    #[doc(hidden)]
    Ptu = 5420,
    #[doc(hidden)]
    Ptv = 5421,
    #[doc(hidden)]
    Ptw = 5422,
    #[doc(hidden)]
    Pty = 5423,
    #[doc(hidden)]
    Pua = 5424,
    #[doc(hidden)]
    Pub = 5425,
    #[doc(hidden)]
    Puc = 5426,
    #[doc(hidden)]
    Pud = 5427,
    #[doc(hidden)]
    Pue = 5428,
    #[doc(hidden)]
    Puf = 5429,
    #[doc(hidden)]
    Pug = 5430,
    #[doc(hidden)]
    Pui = 5431,
    #[doc(hidden)]
    Puj = 5432,
    #[doc(hidden)]
    Pum = 5433,
    #[doc(hidden)]
    Puo = 5434,
    #[doc(hidden)]
    Pup = 5435,
    #[doc(hidden)]
    Puq = 5436,
    #[doc(hidden)]
    Pur = 5437,
    #[doc(hidden)]
    Pus = 5438,
    #[doc(hidden)]
    Put = 5439,
    #[doc(hidden)]
    Puu = 5440,
    #[doc(hidden)]
    Puw = 5441,
    #[doc(hidden)]
    Pux = 5442,
    #[doc(hidden)]
    Puy = 5443,
    #[doc(hidden)]
    Pwa = 5444,
    #[doc(hidden)]
    Pwb = 5445,
    #[doc(hidden)]
    Pwg = 5446,
    #[doc(hidden)]
    Pwi = 5447,
    #[doc(hidden)]
    Pwm = 5448,
    #[doc(hidden)]
    Pwn = 5449,
    #[doc(hidden)]
    Pwo = 5450,
    #[doc(hidden)]
    Pwr = 5451,
    #[doc(hidden)]
    Pww = 5452,
    #[doc(hidden)]
    Pxm = 5453,
    #[doc(hidden)]
    Pye = 5454,
    #[doc(hidden)]
    Pym = 5455,
    #[doc(hidden)]
    Pyn = 5456,
    #[doc(hidden)]
    Pys = 5457,
    #[doc(hidden)]
    Pyu = 5458,
    #[doc(hidden)]
    Pyx = 5459,
    #[doc(hidden)]
    Pyy = 5460,
    #[doc(hidden)]
    Pzh = 5461,
    #[doc(hidden)]
    Pzn = 5462,
    #[doc(hidden)]
    Qua = 5463,
    #[doc(hidden)]
    Qub = 5464,
    #[doc(hidden)]
    Quc = 5465,
    #[doc(hidden)]
    Qud = 5466,
    #[doc(hidden)]
    Que = 5467,
    #[doc(hidden)]
    Quf = 5468,
    #[doc(hidden)]
    Qug = 5469,
    #[doc(hidden)]
    Quh = 5470,
    #[doc(hidden)]
    Qui = 5471,
    #[doc(hidden)]
    Quk = 5472,
    #[doc(hidden)]
    Qul = 5473,
    #[doc(hidden)]
    Qum = 5474,
    #[doc(hidden)]
    Qun = 5475,
    #[doc(hidden)]
    Qup = 5476,
    #[doc(hidden)]
    Quq = 5477,
    #[doc(hidden)]
    Qur = 5478,
    #[doc(hidden)]
    Qus = 5479,
    #[doc(hidden)]
    Quv = 5480,
    #[doc(hidden)]
    Quw = 5481,
    #[doc(hidden)]
    Qux = 5482,
    #[doc(hidden)]
    Quy = 5483,
    #[doc(hidden)]
    Quz = 5484,
    #[doc(hidden)]
    Qva = 5485,
    #[doc(hidden)]
    Qvc = 5486,
    #[doc(hidden)]
    Qve = 5487,
    #[doc(hidden)]
    Qvh = 5488,
    #[doc(hidden)]
    Qvi = 5489,
    #[doc(hidden)]
    Qvj = 5490,
    #[doc(hidden)]
    Qvl = 5491,
    #[doc(hidden)]
    Qvm = 5492,
    #[doc(hidden)]
    Qvn = 5493,
    #[doc(hidden)]
    Qvo = 5494,
    #[doc(hidden)]
    Qvp = 5495,
    #[doc(hidden)]
    Qvs = 5496,
    #[doc(hidden)]
    Qvw = 5497,
    #[doc(hidden)]
    Qvy = 5498,
    #[doc(hidden)]
    Qvz = 5499,
    #[doc(hidden)]
    Qwa = 5500,
    #[doc(hidden)]
    Qwc = 5501,
    #[doc(hidden)]
    Qwh = 5502,
    #[doc(hidden)]
    Qwm = 5503,
    #[doc(hidden)]
    Qws = 5504,
    #[doc(hidden)]
    Qwt = 5505,
    #[doc(hidden)]
    Qxa = 5506,
    #[doc(hidden)]
    Qxc = 5507,
    #[doc(hidden)]
    Qxh = 5508,
    #[doc(hidden)]
    Qxl = 5509,
    #[doc(hidden)]
    Qxn = 5510,
    #[doc(hidden)]
    Qxo = 5511,
    #[doc(hidden)]
    Qxp = 5512,
    #[doc(hidden)]
    Qxq = 5513,
    #[doc(hidden)]
    Qxr = 5514,
    #[doc(hidden)]
    Qxs = 5515,
    #[doc(hidden)]
    Qxt = 5516,
    #[doc(hidden)]
    Qxu = 5517,
    #[doc(hidden)]
    Qxw = 5518,
    #[doc(hidden)]
    Qya = 5519,
    #[doc(hidden)]
    Qyp = 5520,
    #[doc(hidden)]
    Raa = 5521,
    #[doc(hidden)]
    Rab = 5522,
    #[doc(hidden)]
    Rac = 5523,
    #[doc(hidden)]
    Rad = 5524,
    #[doc(hidden)]
    Raf = 5525,
    #[doc(hidden)]
    Rag = 5526,
    #[doc(hidden)]
    Rah = 5527,
    #[doc(hidden)]
    Rai = 5528,
    #[doc(hidden)]
    Raj = 5529,
    #[doc(hidden)]
    Rak = 5530,
    #[doc(hidden)]
    Ral = 5531,
    #[doc(hidden)]
    Ram = 5532,
    #[doc(hidden)]
    Ran = 5533,
    #[doc(hidden)]
    Rao = 5534,
    #[doc(hidden)]
    Rap = 5535,
    #[doc(hidden)]
    Raq = 5536,
    #[doc(hidden)]
    Rar = 5537,
    #[doc(hidden)]
    Ras = 5538,
    #[doc(hidden)]
    Rat = 5539,
    #[doc(hidden)]
    Rau = 5540,
    #[doc(hidden)]
    Rav = 5541,
    #[doc(hidden)]
    Raw = 5542,
    #[doc(hidden)]
    Rax = 5543,
    #[doc(hidden)]
    Ray = 5544,
    #[doc(hidden)]
    Raz = 5545,
    #[doc(hidden)]
    Rbb = 5546,
    #[doc(hidden)]
    Rbk = 5547,
    #[doc(hidden)]
    Rbl = 5548,
    #[doc(hidden)]
    Rbp = 5549,
    #[doc(hidden)]
    Rcf = 5550,
    #[doc(hidden)]
    Rdb = 5551,
    #[doc(hidden)]
    Rea = 5552,
    #[doc(hidden)]
    Reb = 5553,
    #[doc(hidden)]
    Ree = 5554,
    #[doc(hidden)]
    Reg = 5555,
    #[doc(hidden)]
    Rei = 5556,
    #[doc(hidden)]
    Rej = 5557,
    #[doc(hidden)]
    Rel = 5558,
    #[doc(hidden)]
    Rem = 5559,
    #[doc(hidden)]
    Ren = 5560,
    #[doc(hidden)]
    Rer = 5561,
    #[doc(hidden)]
    Res = 5562,
    #[doc(hidden)]
    Ret = 5563,
    #[doc(hidden)]
    Rey = 5564,
    #[doc(hidden)]
    Rga = 5565,
    #[doc(hidden)]
    Rge = 5566,
    #[doc(hidden)]
    Rgk = 5567,
    #[doc(hidden)]
    Rgn = 5568,
    #[doc(hidden)]
    Rgr = 5569,
    #[doc(hidden)]
    Rgs = 5570,
    #[doc(hidden)]
    Rgu = 5571,
    #[doc(hidden)]
    Rhg = 5572,
    #[doc(hidden)]
    Rhp = 5573,
    #[doc(hidden)]
    Ria = 5574,
    #[doc(hidden)]
    Rib = 5575,
    #[doc(hidden)]
    Rif = 5576,
    #[doc(hidden)]
    Ril = 5577,
    #[doc(hidden)]
    Rim = 5578,
    #[doc(hidden)]
    Rin = 5579,
    #[doc(hidden)]
    Rir = 5580,
    #[doc(hidden)]
    Rit = 5581,
    #[doc(hidden)]
    Riu = 5582,
    #[doc(hidden)]
    Rjg = 5583,
    #[doc(hidden)]
    Rji = 5584,
    #[doc(hidden)]
    Rjs = 5585,
    #[doc(hidden)]
    Rka = 5586,
    #[doc(hidden)]
    Rkb = 5587,
    #[doc(hidden)]
    Rkh = 5588,
    #[doc(hidden)]
    Rki = 5589,
    #[doc(hidden)]
    Rkm = 5590,
    #[doc(hidden)]
    Rkt = 5591,
    #[doc(hidden)]
    Rkw = 5592,
    #[doc(hidden)]
    Rma = 5593,
    #[doc(hidden)]
    Rmb = 5594,
    #[doc(hidden)]
    Rmc = 5595,
    #[doc(hidden)]
    Rmd = 5596,
    #[doc(hidden)]
    Rme = 5597,
    #[doc(hidden)]
    Rmf = 5598,
    #[doc(hidden)]
    Rmg = 5599,
    #[doc(hidden)]
    Rmh = 5600,
    #[doc(hidden)]
    Rmi = 5601,
    #[doc(hidden)]
    Rmk = 5602,
    #[doc(hidden)]
    Rml = 5603,
    #[doc(hidden)]
    Rmm = 5604,
    #[doc(hidden)]
    Rmn = 5605,
    #[doc(hidden)]
    Rmo = 5606,
    #[doc(hidden)]
    Rmp = 5607,
    #[doc(hidden)]
    Rmq = 5608,
    #[doc(hidden)]
    Rms = 5609,
    #[doc(hidden)]
    Rmt = 5610,
    #[doc(hidden)]
    Rmu = 5611,
    #[doc(hidden)]
    Rmv = 5612,
    #[doc(hidden)]
    Rmw = 5613,
    #[doc(hidden)]
    Rmx = 5614,
    #[doc(hidden)]
    Rmy = 5615,
    #[doc(hidden)]
    Rmz = 5616,
    #[doc(hidden)]
    Rnb = 5617,
    #[doc(hidden)]
    Rnd = 5618,
    #[doc(hidden)]
    Rng = 5619,
    #[doc(hidden)]
    Rnl = 5620,
    #[doc(hidden)]
    Rnn = 5621,
    #[doc(hidden)]
    Rnp = 5622,
    #[doc(hidden)]
    Rnr = 5623,
    #[doc(hidden)]
    Rnw = 5624,
    #[doc(hidden)]
    Rob = 5625,
    #[doc(hidden)]
    Roc = 5626,
    #[doc(hidden)]
    Rod = 5627,
    #[doc(hidden)]
    Roe = 5628,
    #[doc(hidden)]
    Rof = 5629,
    #[doc(hidden)]
    Rog = 5630,
    #[doc(hidden)]
    Roh = 5631,
    #[doc(hidden)]
    Rol = 5632,
    #[doc(hidden)]
    Rom = 5633,
    #[doc(hidden)]
    Ron = 5634,
    #[doc(hidden)]
    Roo = 5635,
    #[doc(hidden)]
    Rop = 5636,
    #[doc(hidden)]
    Ror = 5637,
    #[doc(hidden)]
    Rou = 5638,
    #[doc(hidden)]
    Row = 5639,
    #[doc(hidden)]
    Rpn = 5640,
    #[doc(hidden)]
    Rpt = 5641,
    #[doc(hidden)]
    Rri = 5642,
    #[doc(hidden)]
    Rro = 5643,
    #[doc(hidden)]
    Rrt = 5644,
    #[doc(hidden)]
    Rsb = 5645,
    #[doc(hidden)]
    Rsk = 5646,
    #[doc(hidden)]
    Rsl = 5647,
    #[doc(hidden)]
    Rsm = 5648,
    #[doc(hidden)]
    Rsn = 5649,
    #[doc(hidden)]
    Rtc = 5650,
    #[doc(hidden)]
    Rth = 5651,
    #[doc(hidden)]
    Rtm = 5652,
    #[doc(hidden)]
    Rts = 5653,
    #[doc(hidden)]
    Rtw = 5654,
    #[doc(hidden)]
    Rub = 5655,
    #[doc(hidden)]
    Ruc = 5656,
    #[doc(hidden)]
    Rue = 5657,
    #[doc(hidden)]
    Ruf = 5658,
    #[doc(hidden)]
    Rug = 5659,
    #[doc(hidden)]
    Ruh = 5660,
    #[doc(hidden)]
    Rui = 5661,
    #[doc(hidden)]
    Ruk = 5662,
    #[doc(hidden)]
    Run = 5663,
    #[doc(hidden)]
    Ruo = 5664,
    #[doc(hidden)]
    Rup = 5665,
    #[doc(hidden)]
    Ruq = 5666,
    #[doc(hidden)]
    Rus = 5667,
    #[doc(hidden)]
    Rut = 5668,
    #[doc(hidden)]
    Ruu = 5669,
    #[doc(hidden)]
    Ruy = 5670,
    #[doc(hidden)]
    Ruz = 5671,
    #[doc(hidden)]
    Rwa = 5672,
    #[doc(hidden)]
    Rwk = 5673,
    #[doc(hidden)]
    Rwl = 5674,
    #[doc(hidden)]
    Rwm = 5675,
    #[doc(hidden)]
    Rwo = 5676,
    #[doc(hidden)]
    Rwr = 5677,
    #[doc(hidden)]
    Rxd = 5678,
    #[doc(hidden)]
    Rxw = 5679,
    #[doc(hidden)]
    Ryn = 5680,
    #[doc(hidden)]
    Rys = 5681,
    #[doc(hidden)]
    Ryu = 5682,
    #[doc(hidden)]
    Rzh = 5683,
    #[doc(hidden)]
    Saa = 5684,
    #[doc(hidden)]
    Sab = 5685,
    #[doc(hidden)]
    Sac = 5686,
    #[doc(hidden)]
    Sad = 5687,
    #[doc(hidden)]
    Sae = 5688,
    #[doc(hidden)]
    Saf = 5689,
    #[doc(hidden)]
    Sag = 5690,
    #[doc(hidden)]
    Sah = 5691,
    #[doc(hidden)]
    Saj = 5692,
    #[doc(hidden)]
    Sak = 5693,
    #[doc(hidden)]
    Sam = 5694,
    #[doc(hidden)]
    San = 5695,
    #[doc(hidden)]
    Sao = 5696,
    #[doc(hidden)]
    Saq = 5697,
    #[doc(hidden)]
    Sar = 5698,
    #[doc(hidden)]
    Sas = 5699,
    #[doc(hidden)]
    Sat = 5700,
    #[doc(hidden)]
    Sau = 5701,
    #[doc(hidden)]
    Sav = 5702,
    #[doc(hidden)]
    Saw = 5703,
    #[doc(hidden)]
    Sax = 5704,
    #[doc(hidden)]
    Say = 5705,
    #[doc(hidden)]
    Saz = 5706,
    #[doc(hidden)]
    Sba = 5707,
    #[doc(hidden)]
    Sbb = 5708,
    #[doc(hidden)]
    Sbc = 5709,
    #[doc(hidden)]
    Sbd = 5710,
    #[doc(hidden)]
    Sbe = 5711,
    #[doc(hidden)]
    Sbf = 5712,
    #[doc(hidden)]
    Sbg = 5713,
    #[doc(hidden)]
    Sbh = 5714,
    #[doc(hidden)]
    Sbi = 5715,
    #[doc(hidden)]
    Sbj = 5716,
    #[doc(hidden)]
    Sbk = 5717,
    #[doc(hidden)]
    Sbl = 5718,
    #[doc(hidden)]
    Sbm = 5719,
    #[doc(hidden)]
    Sbn = 5720,
    #[doc(hidden)]
    Sbo = 5721,
    #[doc(hidden)]
    Sbp = 5722,
    #[doc(hidden)]
    Sbq = 5723,
    #[doc(hidden)]
    Sbr = 5724,
    #[doc(hidden)]
    Sbs = 5725,
    #[doc(hidden)]
    Sbt = 5726,
    #[doc(hidden)]
    Sbu = 5727,
    #[doc(hidden)]
    Sbv = 5728,
    #[doc(hidden)]
    Sbw = 5729,
    #[doc(hidden)]
    Sbx = 5730,
    #[doc(hidden)]
    Sby = 5731,
    #[doc(hidden)]
    Sbz = 5732,
    #[doc(hidden)]
    Scb = 5733,
    #[doc(hidden)]
    Sce = 5734,
    #[doc(hidden)]
    Scf = 5735,
    #[doc(hidden)]
    Scg = 5736,
    #[doc(hidden)]
    Sch = 5737,
    #[doc(hidden)]
    Sci = 5738,
    #[doc(hidden)]
    Sck = 5739,
    #[doc(hidden)]
    Scl = 5740,
    #[doc(hidden)]
    Scn = 5741,
    #[doc(hidden)]
    Sco = 5742,
    #[doc(hidden)]
    Scp = 5743,
    #[doc(hidden)]
    Scq = 5744,
    #[doc(hidden)]
    Scs = 5745,
    #[doc(hidden)]
    Sct = 5746,
    #[doc(hidden)]
    Scu = 5747,
    #[doc(hidden)]
    Scv = 5748,
    #[doc(hidden)]
    Scw = 5749,
    #[doc(hidden)]
    Scx = 5750,
    #[doc(hidden)]
    Sda = 5751,
    #[doc(hidden)]
    Sdb = 5752,
    #[doc(hidden)]
    Sdc = 5753,
    #[doc(hidden)]
    Sde = 5754,
    #[doc(hidden)]
    Sdf = 5755,
    #[doc(hidden)]
    Sdg = 5756,
    #[doc(hidden)]
    Sdh = 5757,
    #[doc(hidden)]
    Sdj = 5758,
    #[doc(hidden)]
    Sdk = 5759,
    #[doc(hidden)]
    Sdl = 5760,
    #[doc(hidden)]
    Sdn = 5761,
    #[doc(hidden)]
    Sdo = 5762,
    #[doc(hidden)]
    Sdp = 5763,
    #[doc(hidden)]
    Sdq = 5764,
    #[doc(hidden)]
    Sdr = 5765,
    #[doc(hidden)]
    Sds = 5766,
    #[doc(hidden)]
    Sdt = 5767,
    #[doc(hidden)]
    Sdu = 5768,
    #[doc(hidden)]
    Sdx = 5769,
    #[doc(hidden)]
    Sdz = 5770,
    #[doc(hidden)]
    Sea = 5771,
    #[doc(hidden)]
    Seb = 5772,
    #[doc(hidden)]
    Sec = 5773,
    #[doc(hidden)]
    Sed = 5774,
    #[doc(hidden)]
    See = 5775,
    #[doc(hidden)]
    Sef = 5776,
    #[doc(hidden)]
    Seg = 5777,
    #[doc(hidden)]
    Seh = 5778,
    #[doc(hidden)]
    Sei = 5779,
    #[doc(hidden)]
    Sej = 5780,
    #[doc(hidden)]
    Sek = 5781,
    #[doc(hidden)]
    Sel = 5782,
    #[doc(hidden)]
    Sen = 5783,
    #[doc(hidden)]
    Seo = 5784,
    #[doc(hidden)]
    Sep = 5785,
    #[doc(hidden)]
    Seq = 5786,
    #[doc(hidden)]
    Ser = 5787,
    #[doc(hidden)]
    Ses = 5788,
    #[doc(hidden)]
    Set = 5789,
    #[doc(hidden)]
    Seu = 5790,
    #[doc(hidden)]
    Sev = 5791,
    #[doc(hidden)]
    Sew = 5792,
    #[doc(hidden)]
    Sey = 5793,
    #[doc(hidden)]
    Sez = 5794,
    #[doc(hidden)]
    Sfb = 5795,
    #[doc(hidden)]
    Sfe = 5796,
    #[doc(hidden)]
    Sfm = 5797,
    #[doc(hidden)]
    Sfs = 5798,
    #[doc(hidden)]
    Sfw = 5799,
    #[doc(hidden)]
    Sga = 5800,
    #[doc(hidden)]
    Sgb = 5801,
    #[doc(hidden)]
    Sgc = 5802,
    #[doc(hidden)]
    Sgd = 5803,
    #[doc(hidden)]
    Sge = 5804,
    #[doc(hidden)]
    Sgg = 5805,
    #[doc(hidden)]
    Sgh = 5806,
    #[doc(hidden)]
    Sgi = 5807,
    #[doc(hidden)]
    Sgj = 5808,
    #[doc(hidden)]
    Sgk = 5809,
    #[doc(hidden)]
    Sgm = 5810,
    #[doc(hidden)]
    Sgp = 5811,
    #[doc(hidden)]
    Sgr = 5812,
    #[doc(hidden)]
    Sgs = 5813,
    #[doc(hidden)]
    Sgt = 5814,
    #[doc(hidden)]
    Sgu = 5815,
    #[doc(hidden)]
    Sgw = 5816,
    #[doc(hidden)]
    Sgx = 5817,
    #[doc(hidden)]
    Sgy = 5818,
    #[doc(hidden)]
    Sgz = 5819,
    #[doc(hidden)]
    Sha = 5820,
    #[doc(hidden)]
    Shb = 5821,
    #[doc(hidden)]
    Shc = 5822,
    #[doc(hidden)]
    Shd = 5823,
    #[doc(hidden)]
    She = 5824,
    #[doc(hidden)]
    Shg = 5825,
    #[doc(hidden)]
    Shh = 5826,
    #[doc(hidden)]
    Shi = 5827,
    #[doc(hidden)]
    Shj = 5828,
    #[doc(hidden)]
    Shk = 5829,
    #[doc(hidden)]
    Shl = 5830,
    #[doc(hidden)]
    Shm = 5831,
    #[doc(hidden)]
    Shn = 5832,
    #[doc(hidden)]
    Sho = 5833,
    #[doc(hidden)]
    Shp = 5834,
    #[doc(hidden)]
    Shq = 5835,
    #[doc(hidden)]
    Shr = 5836,
    #[doc(hidden)]
    Shs = 5837,
    #[doc(hidden)]
    Sht = 5838,
    #[doc(hidden)]
    Shu = 5839,
    #[doc(hidden)]
    Shv = 5840,
    #[doc(hidden)]
    Shw = 5841,
    #[doc(hidden)]
    Shx = 5842,
    #[doc(hidden)]
    Shy = 5843,
    #[doc(hidden)]
    Shz = 5844,
    #[doc(hidden)]
    Sia = 5845,
    #[doc(hidden)]
    Sib = 5846,
    #[doc(hidden)]
    Sid = 5847,
    #[doc(hidden)]
    Sie = 5848,
    #[doc(hidden)]
    Sif = 5849,
    #[doc(hidden)]
    Sig = 5850,
    #[doc(hidden)]
    Sih = 5851,
    #[doc(hidden)]
    Sii = 5852,
    #[doc(hidden)]
    Sij = 5853,
    #[doc(hidden)]
    Sik = 5854,
    #[doc(hidden)]
    Sil = 5855,
    #[doc(hidden)]
    Sim = 5856,
    #[doc(hidden)]
    Sin = 5857,
    #[doc(hidden)]
    Sip = 5858,
    #[doc(hidden)]
    Siq = 5859,
    #[doc(hidden)]
    Sir = 5860,
    #[doc(hidden)]
    Sis = 5861,
    #[doc(hidden)]
    Siu = 5862,
    #[doc(hidden)]
    Siv = 5863,
    #[doc(hidden)]
    Siw = 5864,
    #[doc(hidden)]
    Six = 5865,
    #[doc(hidden)]
    Siy = 5866,
    #[doc(hidden)]
    Siz = 5867,
    #[doc(hidden)]
    Sja = 5868,
    #[doc(hidden)]
    Sjb = 5869,
    #[doc(hidden)]
    Sjd = 5870,
    #[doc(hidden)]
    Sje = 5871,
    #[doc(hidden)]
    Sjg = 5872,
    #[doc(hidden)]
    Sjk = 5873,
    #[doc(hidden)]
    Sjl = 5874,
    #[doc(hidden)]
    Sjm = 5875,
    #[doc(hidden)]
    Sjn = 5876,
    #[doc(hidden)]
    Sjo = 5877,
    #[doc(hidden)]
    Sjp = 5878,
    #[doc(hidden)]
    Sjr = 5879,
    #[doc(hidden)]
    Sjs = 5880,
    #[doc(hidden)]
    Sjt = 5881,
    #[doc(hidden)]
    Sju = 5882,
    #[doc(hidden)]
    Sjw = 5883,
    #[doc(hidden)]
    Ska = 5884,
    #[doc(hidden)]
    Skb = 5885,
    #[doc(hidden)]
    Skc = 5886,
    #[doc(hidden)]
    Skd = 5887,
    #[doc(hidden)]
    Ske = 5888,
    #[doc(hidden)]
    Skf = 5889,
    #[doc(hidden)]
    Skg = 5890,
    #[doc(hidden)]
    Skh = 5891,
    #[doc(hidden)]
    Ski = 5892,
    #[doc(hidden)]
    Skj = 5893,
    #[doc(hidden)]
    Skm = 5894,
    #[doc(hidden)]
    Skn = 5895,
    #[doc(hidden)]
    Sko = 5896,
    #[doc(hidden)]
    Skp = 5897,
    #[doc(hidden)]
    Skq = 5898,
    #[doc(hidden)]
    Skr = 5899,
    #[doc(hidden)]
    Sks = 5900,
    #[doc(hidden)]
    Skt = 5901,
    #[doc(hidden)]
    Sku = 5902,
    #[doc(hidden)]
    Skv = 5903,
    #[doc(hidden)]
    Skw = 5904,
    #[doc(hidden)]
    Skx = 5905,
    #[doc(hidden)]
    Sky = 5906,
    #[doc(hidden)]
    Skz = 5907,
    #[doc(hidden)]
    Slc = 5908,
    #[doc(hidden)]
    Sld = 5909,
    #[doc(hidden)]
    Sle = 5910,
    #[doc(hidden)]
    Slf = 5911,
    #[doc(hidden)]
    Slg = 5912,
    #[doc(hidden)]
    Slh = 5913,
    #[doc(hidden)]
    Sli = 5914,
    #[doc(hidden)]
    Slj = 5915,
    #[doc(hidden)]
    Slk = 5916,
    #[doc(hidden)]
    Sll = 5917,
    #[doc(hidden)]
    Slm = 5918,
    #[doc(hidden)]
    Sln = 5919,
    #[doc(hidden)]
    Slp = 5920,
    #[doc(hidden)]
    Slq = 5921,
    #[doc(hidden)]
    Slr = 5922,
    #[doc(hidden)]
    Sls = 5923,
    #[doc(hidden)]
    Slt = 5924,
    #[doc(hidden)]
    Slu = 5925,
    #[doc(hidden)]
    Slv = 5926,
    #[doc(hidden)]
    Slw = 5927,
    #[doc(hidden)]
    Slx = 5928,
    #[doc(hidden)]
    Sly = 5929,
    #[doc(hidden)]
    Slz = 5930,
    #[doc(hidden)]
    Sma = 5931,
    #[doc(hidden)]
    Smb = 5932,
    #[doc(hidden)]
    Smc = 5933,
    #[doc(hidden)]
    Sme = 5934,
    #[doc(hidden)]
    Smf = 5935,
    #[doc(hidden)]
    Smg = 5936,
    #[doc(hidden)]
    Smh = 5937,
    #[doc(hidden)]
    Smj = 5938,
    #[doc(hidden)]
    Smk = 5939,
    #[doc(hidden)]
    Sml = 5940,
    #[doc(hidden)]
    Smm = 5941,
    #[doc(hidden)]
    Smn = 5942,
    #[doc(hidden)]
    Smo = 5943,
    #[doc(hidden)]
    Smp = 5944,
    #[doc(hidden)]
    Smq = 5945,
    #[doc(hidden)]
    Smr = 5946,
    #[doc(hidden)]
    Sms = 5947,
    #[doc(hidden)]
    Smt = 5948,
    #[doc(hidden)]
    Smu = 5949,
    #[doc(hidden)]
    Smv = 5950,
    #[doc(hidden)]
    Smw = 5951,
    #[doc(hidden)]
    Smx = 5952,
    #[doc(hidden)]
    Smy = 5953,
    #[doc(hidden)]
    Smz = 5954,
    #[doc(hidden)]
    Sna = 5955,
    #[doc(hidden)]
    Snc = 5956,
    #[doc(hidden)]
    Snd = 5957,
    #[doc(hidden)]
    Sne = 5958,
    #[doc(hidden)]
    Snf = 5959,
    #[doc(hidden)]
    Sng = 5960,
    #[doc(hidden)]
    Sni = 5961,
    #[doc(hidden)]
    Snj = 5962,
    #[doc(hidden)]
    Snk = 5963,
    #[doc(hidden)]
    Snl = 5964,
    #[doc(hidden)]
    Snm = 5965,
    #[doc(hidden)]
    Snn = 5966,
    #[doc(hidden)]
    Sno = 5967,
    #[doc(hidden)]
    Snp = 5968,
    #[doc(hidden)]
    Snq = 5969,
    #[doc(hidden)]
    Snr = 5970,
    #[doc(hidden)]
    Sns = 5971,
    #[doc(hidden)]
    Snu = 5972,
    #[doc(hidden)]
    Snv = 5973,
    #[doc(hidden)]
    Snw = 5974,
    #[doc(hidden)]
    Snx = 5975,
    #[doc(hidden)]
    Sny = 5976,
    #[doc(hidden)]
    Snz = 5977,
    #[doc(hidden)]
    Soa = 5978,
    #[doc(hidden)]
    Sob = 5979,
    #[doc(hidden)]
    Soc = 5980,
    #[doc(hidden)]
    Sod = 5981,
    #[doc(hidden)]
    Soe = 5982,
    #[doc(hidden)]
    Sog = 5983,
    #[doc(hidden)]
    Soh = 5984,
    #[doc(hidden)]
    Soi = 5985,
    #[doc(hidden)]
    Soj = 5986,
    #[doc(hidden)]
    Sok = 5987,
    #[doc(hidden)]
    Sol = 5988,
    #[doc(hidden)]
    Som = 5989,
    #[doc(hidden)]
    Soo = 5990,
    #[doc(hidden)]
    Sop = 5991,
    #[doc(hidden)]
    Soq = 5992,
    #[doc(hidden)]
    Sor = 5993,
    #[doc(hidden)]
    Sos = 5994,
    #[doc(hidden)]
    Sot = 5995,
    #[doc(hidden)]
    Sou = 5996,
    #[doc(hidden)]
    Sov = 5997,
    #[doc(hidden)]
    Sow = 5998,
    #[doc(hidden)]
    Sox = 5999,
    #[doc(hidden)]
    Soy = 6000,
    #[doc(hidden)]
    Soz = 6001,
    #[doc(hidden)]
    Spa = 6002,
    #[doc(hidden)]
    Spb = 6003,
    #[doc(hidden)]
    Spc = 6004,
    #[doc(hidden)]
    Spd = 6005,
    #[doc(hidden)]
    Spe = 6006,
    #[doc(hidden)]
    Spg = 6007,
    #[doc(hidden)]
    Spi = 6008,
    #[doc(hidden)]
    Spk = 6009,
    #[doc(hidden)]
    Spl = 6010,
    #[doc(hidden)]
    Spm = 6011,
    #[doc(hidden)]
    Spn = 6012,
    #[doc(hidden)]
    Spo = 6013,
    #[doc(hidden)]
    Spp = 6014,
    #[doc(hidden)]
    Spq = 6015,
    #[doc(hidden)]
    Spr = 6016,
    #[doc(hidden)]
    Sps = 6017,
    #[doc(hidden)]
    Spt = 6018,
    #[doc(hidden)]
    Spu = 6019,
    #[doc(hidden)]
    Spv = 6020,
    #[doc(hidden)]
    Spx = 6021,
    #[doc(hidden)]
    Spy = 6022,
    #[doc(hidden)]
    Sqa = 6023,
    #[doc(hidden)]
    Sqh = 6024,
    #[doc(hidden)]
    Sqi = 6025,
    #[doc(hidden)]
    Sqk = 6026,
    #[doc(hidden)]
    Sqm = 6027,
    #[doc(hidden)]
    Sqn = 6028,
    #[doc(hidden)]
    Sqo = 6029,
    #[doc(hidden)]
    Sqq = 6030,
    #[doc(hidden)]
    Sqr = 6031,
    #[doc(hidden)]
    Sqs = 6032,
    #[doc(hidden)]
    Sqt = 6033,
    #[doc(hidden)]
    Squ = 6034,
    #[doc(hidden)]
    Sqx = 6035,
    #[doc(hidden)]
    Sra = 6036,
    #[doc(hidden)]
    Srb = 6037,
    #[doc(hidden)]
    Src = 6038,
    #[doc(hidden)]
    Srd = 6039,
    #[doc(hidden)]
    Sre = 6040,
    #[doc(hidden)]
    Srf = 6041,
    #[doc(hidden)]
    Srg = 6042,
    #[doc(hidden)]
    Srh = 6043,
    #[doc(hidden)]
    Sri = 6044,
    #[doc(hidden)]
    Srk = 6045,
    #[doc(hidden)]
    Srl = 6046,
    #[doc(hidden)]
    Srm = 6047,
    #[doc(hidden)]
    Srn = 6048,
    #[doc(hidden)]
    Sro = 6049,
    #[doc(hidden)]
    Srp = 6050,
    #[doc(hidden)]
    Srq = 6051,
    #[doc(hidden)]
    Srr = 6052,
    #[doc(hidden)]
    Srs = 6053,
    #[doc(hidden)]
    Srt = 6054,
    #[doc(hidden)]
    Sru = 6055,
    #[doc(hidden)]
    Srv = 6056,
    #[doc(hidden)]
    Srw = 6057,
    #[doc(hidden)]
    Srx = 6058,
    #[doc(hidden)]
    Sry = 6059,
    #[doc(hidden)]
    Srz = 6060,
    #[doc(hidden)]
    Ssb = 6061,
    #[doc(hidden)]
    Ssc = 6062,
    #[doc(hidden)]
    Ssd = 6063,
    #[doc(hidden)]
    Sse = 6064,
    #[doc(hidden)]
    Ssf = 6065,
    #[doc(hidden)]
    Ssg = 6066,
    #[doc(hidden)]
    Ssh = 6067,
    #[doc(hidden)]
    Ssi = 6068,
    #[doc(hidden)]
    Ssj = 6069,
    #[doc(hidden)]
    Ssk = 6070,
    #[doc(hidden)]
    Ssl = 6071,
    #[doc(hidden)]
    Ssm = 6072,
    #[doc(hidden)]
    Ssn = 6073,
    #[doc(hidden)]
    Sso = 6074,
    #[doc(hidden)]
    Ssp = 6075,
    #[doc(hidden)]
    Ssq = 6076,
    #[doc(hidden)]
    Ssr = 6077,
    #[doc(hidden)]
    Sss = 6078,
    #[doc(hidden)]
    Sst = 6079,
    #[doc(hidden)]
    Ssu = 6080,
    #[doc(hidden)]
    Ssv = 6081,
    #[doc(hidden)]
    Ssw = 6082,
    #[doc(hidden)]
    Ssx = 6083,
    #[doc(hidden)]
    Ssy = 6084,
    #[doc(hidden)]
    Ssz = 6085,
    #[doc(hidden)]
    Sta = 6086,
    #[doc(hidden)]
    Stb = 6087,
    #[doc(hidden)]
    Std = 6088,
    #[doc(hidden)]
    Ste = 6089,
    #[doc(hidden)]
    Stf = 6090,
    #[doc(hidden)]
    Stg = 6091,
    #[doc(hidden)]
    Sth = 6092,
    #[doc(hidden)]
    Sti = 6093,
    #[doc(hidden)]
    Stj = 6094,
    #[doc(hidden)]
    Stk = 6095,
    #[doc(hidden)]
    Stl = 6096,
    #[doc(hidden)]
    Stm = 6097,
    #[doc(hidden)]
    Stn = 6098,
    #[doc(hidden)]
    Sto = 6099,
    #[doc(hidden)]
    Stp = 6100,
    #[doc(hidden)]
    Stq = 6101,
    #[doc(hidden)]
    Str = 6102,
    #[doc(hidden)]
    Sts = 6103,
    #[doc(hidden)]
    Stt = 6104,
    #[doc(hidden)]
    Stu = 6105,
    #[doc(hidden)]
    Stv = 6106,
    #[doc(hidden)]
    Stw = 6107,
    #[doc(hidden)]
    Sty = 6108,
    #[doc(hidden)]
    Sua = 6109,
    #[doc(hidden)]
    Sub = 6110,
    #[doc(hidden)]
    Suc = 6111,
    #[doc(hidden)]
    Sue = 6112,
    #[doc(hidden)]
    Sug = 6113,
    #[doc(hidden)]
    Sui = 6114,
    #[doc(hidden)]
    Suj = 6115,
    #[doc(hidden)]
    Suk = 6116,
    #[doc(hidden)]
    Sun = 6117,
    #[doc(hidden)]
    Suo = 6118,
    #[doc(hidden)]
    Suq = 6119,
    #[doc(hidden)]
    Sur = 6120,
    #[doc(hidden)]
    Sus = 6121,
    #[doc(hidden)]
    Sut = 6122,
    #[doc(hidden)]
    Suv = 6123,
    #[doc(hidden)]
    Suw = 6124,
    #[doc(hidden)]
    Sux = 6125,
    #[doc(hidden)]
    Suy = 6126,
    #[doc(hidden)]
    Suz = 6127,
    #[doc(hidden)]
    Sva = 6128,
    #[doc(hidden)]
    Svb = 6129,
    #[doc(hidden)]
    Svc = 6130,
    #[doc(hidden)]
    Sve = 6131,
    #[doc(hidden)]
    Svk = 6132,
    #[doc(hidden)]
    Svm = 6133,
    #[doc(hidden)]
    Svs = 6134,
    #[doc(hidden)]
    Svx = 6135,
    #[doc(hidden)]
    Swa = 6136,
    #[doc(hidden)]
    Swb = 6137,
    #[doc(hidden)]
    Swc = 6138,
    #[doc(hidden)]
    Swe = 6139,
    #[doc(hidden)]
    Swf = 6140,
    #[doc(hidden)]
    Swg = 6141,
    #[doc(hidden)]
    Swh = 6142,
    #[doc(hidden)]
    Swi = 6143,
    #[doc(hidden)]
    Swj = 6144,
    #[doc(hidden)]
    Swk = 6145,
    #[doc(hidden)]
    Swl = 6146,
    #[doc(hidden)]
    Swm = 6147,
    #[doc(hidden)]
    Swn = 6148,
    #[doc(hidden)]
    Swo = 6149,
    #[doc(hidden)]
    Swp = 6150,
    #[doc(hidden)]
    Swq = 6151,
    #[doc(hidden)]
    Swr = 6152,
    #[doc(hidden)]
    Sws = 6153,
    #[doc(hidden)]
    Swt = 6154,
    #[doc(hidden)]
    Swu = 6155,
    #[doc(hidden)]
    Swv = 6156,
    #[doc(hidden)]
    Sww = 6157,
    #[doc(hidden)]
    Swx = 6158,
    #[doc(hidden)]
    Swy = 6159,
    #[doc(hidden)]
    Sxb = 6160,
    #[doc(hidden)]
    Sxc = 6161,
    #[doc(hidden)]
    Sxe = 6162,
    #[doc(hidden)]
    Sxg = 6163,
    #[doc(hidden)]
    Sxk = 6164,
    #[doc(hidden)]
    Sxl = 6165,
    #[doc(hidden)]
    Sxm = 6166,
    #[doc(hidden)]
    Sxn = 6167,
    #[doc(hidden)]
    Sxo = 6168,
    #[doc(hidden)]
    Sxr = 6169,
    #[doc(hidden)]
    Sxs = 6170,
    #[doc(hidden)]
    Sxu = 6171,
    #[doc(hidden)]
    Sxw = 6172,
    #[doc(hidden)]
    Sya = 6173,
    #[doc(hidden)]
    Syb = 6174,
    #[doc(hidden)]
    Syc = 6175,
    #[doc(hidden)]
    Syi = 6176,
    #[doc(hidden)]
    Syk = 6177,
    #[doc(hidden)]
    Syl = 6178,
    #[doc(hidden)]
    Sym = 6179,
    #[doc(hidden)]
    Syn = 6180,
    #[doc(hidden)]
    Syo = 6181,
    #[doc(hidden)]
    Syr = 6182,
    #[doc(hidden)]
    Sys = 6183,
    #[doc(hidden)]
    Syw = 6184,
    #[doc(hidden)]
    Syx = 6185,
    #[doc(hidden)]
    Syy = 6186,
    #[doc(hidden)]
    Sza = 6187,
    #[doc(hidden)]
    Szb = 6188,
    #[doc(hidden)]
    Szc = 6189,
    #[doc(hidden)]
    Szd = 6190,
    #[doc(hidden)]
    Sze = 6191,
    #[doc(hidden)]
    Szg = 6192,
    #[doc(hidden)]
    Szl = 6193,
    #[doc(hidden)]
    Szn = 6194,
    #[doc(hidden)]
    Szp = 6195,
    #[doc(hidden)]
    Szs = 6196,
    #[doc(hidden)]
    Szv = 6197,
    #[doc(hidden)]
    Szw = 6198,
    #[doc(hidden)]
    Szy = 6199,
    #[doc(hidden)]
    Taa = 6200,
    #[doc(hidden)]
    Tab = 6201,
    #[doc(hidden)]
    Tac = 6202,
    #[doc(hidden)]
    Tad = 6203,
    #[doc(hidden)]
    Tae = 6204,
    #[doc(hidden)]
    Taf = 6205,
    #[doc(hidden)]
    Tag = 6206,
    #[doc(hidden)]
    Tah = 6207,
    #[doc(hidden)]
    Taj = 6208,
    #[doc(hidden)]
    Tak = 6209,
    #[doc(hidden)]
    Tal = 6210,
    #[doc(hidden)]
    Tam = 6211,
    #[doc(hidden)]
    Tan = 6212,
    #[doc(hidden)]
    Tao = 6213,
    #[doc(hidden)]
    Tap = 6214,
    #[doc(hidden)]
    Taq = 6215,
    #[doc(hidden)]
    Tar = 6216,
    #[doc(hidden)]
    Tas = 6217,
    #[doc(hidden)]
    Tat = 6218,
    #[doc(hidden)]
    Tau = 6219,
    #[doc(hidden)]
    Tav = 6220,
    #[doc(hidden)]
    Taw = 6221,
    #[doc(hidden)]
    Tax = 6222,
    #[doc(hidden)]
    Tay = 6223,
    #[doc(hidden)]
    Taz = 6224,
    #[doc(hidden)]
    Tba = 6225,
    #[doc(hidden)]
    Tbc = 6226,
    #[doc(hidden)]
    Tbd = 6227,
    #[doc(hidden)]
    Tbe = 6228,
    #[doc(hidden)]
    Tbf = 6229,
    #[doc(hidden)]
    Tbg = 6230,
    #[doc(hidden)]
    Tbh = 6231,
    #[doc(hidden)]
    Tbi = 6232,
    #[doc(hidden)]
    Tbj = 6233,
    #[doc(hidden)]
    Tbk = 6234,
    #[doc(hidden)]
    Tbl = 6235,
    #[doc(hidden)]
    Tbm = 6236,
    #[doc(hidden)]
    Tbn = 6237,
    #[doc(hidden)]
    Tbo = 6238,
    #[doc(hidden)]
    Tbp = 6239,
    #[doc(hidden)]
    Tbr = 6240,
    #[doc(hidden)]
    Tbs = 6241,
    #[doc(hidden)]
    Tbt = 6242,
    #[doc(hidden)]
    Tbu = 6243,
    #[doc(hidden)]
    Tbv = 6244,
    #[doc(hidden)]
    Tbw = 6245,
    #[doc(hidden)]
    Tbx = 6246,
    #[doc(hidden)]
    Tby = 6247,
    #[doc(hidden)]
    Tbz = 6248,
    #[doc(hidden)]
    Tca = 6249,
    #[doc(hidden)]
    Tcb = 6250,
    #[doc(hidden)]
    Tcc = 6251,
    #[doc(hidden)]
    Tcd = 6252,
    #[doc(hidden)]
    Tce = 6253,
    #[doc(hidden)]
    Tcf = 6254,
    #[doc(hidden)]
    Tcg = 6255,
    #[doc(hidden)]
    Tch = 6256,
    #[doc(hidden)]
    Tci = 6257,
    #[doc(hidden)]
    Tck = 6258,
    #[doc(hidden)]
    Tcl = 6259,
    #[doc(hidden)]
    Tcm = 6260,
    #[doc(hidden)]
    Tcn = 6261,
    #[doc(hidden)]
    Tco = 6262,
    #[doc(hidden)]
    Tcp = 6263,
    #[doc(hidden)]
    Tcq = 6264,
    #[doc(hidden)]
    Tcs = 6265,
    #[doc(hidden)]
    Tct = 6266,
    #[doc(hidden)]
    Tcu = 6267,
    #[doc(hidden)]
    Tcw = 6268,
    #[doc(hidden)]
    Tcx = 6269,
    #[doc(hidden)]
    Tcy = 6270,
    #[doc(hidden)]
    Tcz = 6271,
    #[doc(hidden)]
    Tda = 6272,
    #[doc(hidden)]
    Tdb = 6273,
    #[doc(hidden)]
    Tdc = 6274,
    #[doc(hidden)]
    Tdd = 6275,
    #[doc(hidden)]
    Tde = 6276,
    #[doc(hidden)]
    Tdf = 6277,
    #[doc(hidden)]
    Tdg = 6278,
    #[doc(hidden)]
    Tdh = 6279,
    #[doc(hidden)]
    Tdi = 6280,
    #[doc(hidden)]
    Tdj = 6281,
    #[doc(hidden)]
    Tdk = 6282,
    #[doc(hidden)]
    Tdl = 6283,
    #[doc(hidden)]
    Tdm = 6284,
    #[doc(hidden)]
    Tdn = 6285,
    #[doc(hidden)]
    Tdo = 6286,
    #[doc(hidden)]
    Tdq = 6287,
    #[doc(hidden)]
    Tdr = 6288,
    #[doc(hidden)]
    Tds = 6289,
    #[doc(hidden)]
    Tdt = 6290,
    #[doc(hidden)]
    Tdv = 6291,
    #[doc(hidden)]
    Tdx = 6292,
    #[doc(hidden)]
    Tdy = 6293,
    #[doc(hidden)]
    Tea = 6294,
    #[doc(hidden)]
    Teb = 6295,
    #[doc(hidden)]
    Tec = 6296,
    #[doc(hidden)]
    Ted = 6297,
    #[doc(hidden)]
    Tee = 6298,
    #[doc(hidden)]
    Tef = 6299,
    #[doc(hidden)]
    Teg = 6300,
    #[doc(hidden)]
    Teh = 6301,
    #[doc(hidden)]
    Tei = 6302,
    #[doc(hidden)]
    Tek = 6303,
    #[doc(hidden)]
    Tel = 6304,
    #[doc(hidden)]
    Tem = 6305,
    #[doc(hidden)]
    Ten = 6306,
    #[doc(hidden)]
    Teo = 6307,
    #[doc(hidden)]
    Tep = 6308,
    #[doc(hidden)]
    Teq = 6309,
    #[doc(hidden)]
    Ter = 6310,
    #[doc(hidden)]
    Tes = 6311,
    #[doc(hidden)]
    Tet = 6312,
    #[doc(hidden)]
    Teu = 6313,
    #[doc(hidden)]
    Tev = 6314,
    #[doc(hidden)]
    Tew = 6315,
    #[doc(hidden)]
    Tex = 6316,
    #[doc(hidden)]
    Tey = 6317,
    #[doc(hidden)]
    Tez = 6318,
    #[doc(hidden)]
    Tfi = 6319,
    #[doc(hidden)]
    Tfn = 6320,
    #[doc(hidden)]
    Tfo = 6321,
    #[doc(hidden)]
    Tfr = 6322,
    #[doc(hidden)]
    Tft = 6323,
    #[doc(hidden)]
    Tga = 6324,
    #[doc(hidden)]
    Tgb = 6325,
    #[doc(hidden)]
    Tgc = 6326,
    #[doc(hidden)]
    Tgd = 6327,
    #[doc(hidden)]
    Tge = 6328,
    #[doc(hidden)]
    Tgf = 6329,
    #[doc(hidden)]
    Tgh = 6330,
    #[doc(hidden)]
    Tgi = 6331,
    #[doc(hidden)]
    Tgj = 6332,
    #[doc(hidden)]
    Tgk = 6333,
    #[doc(hidden)]
    Tgl = 6334,
    #[doc(hidden)]
    Tgn = 6335,
    #[doc(hidden)]
    Tgo = 6336,
    #[doc(hidden)]
    Tgp = 6337,
    #[doc(hidden)]
    Tgq = 6338,
    #[doc(hidden)]
    Tgr = 6339,
    #[doc(hidden)]
    Tgs = 6340,
    #[doc(hidden)]
    Tgt = 6341,
    #[doc(hidden)]
    Tgu = 6342,
    #[doc(hidden)]
    Tgv = 6343,
    #[doc(hidden)]
    Tgw = 6344,
    #[doc(hidden)]
    Tgx = 6345,
    #[doc(hidden)]
    Tgy = 6346,
    #[doc(hidden)]
    Tgz = 6347,
    #[doc(hidden)]
    Tha = 6348,
    #[doc(hidden)]
    Thd = 6349,
    #[doc(hidden)]
    The = 6350,
    #[doc(hidden)]
    Thf = 6351,
    #[doc(hidden)]
    Thh = 6352,
    #[doc(hidden)]
    Thi = 6353,
    #[doc(hidden)]
    Thk = 6354,
    #[doc(hidden)]
    Thl = 6355,
    #[doc(hidden)]
    Thm = 6356,
    #[doc(hidden)]
    Thn = 6357,
    #[doc(hidden)]
    Thp = 6358,
    #[doc(hidden)]
    Thq = 6359,
    #[doc(hidden)]
    Thr = 6360,
    #[doc(hidden)]
    Ths = 6361,
    #[doc(hidden)]
    Tht = 6362,
    #[doc(hidden)]
    Thu = 6363,
    #[doc(hidden)]
    Thv = 6364,
    #[doc(hidden)]
    Thy = 6365,
    #[doc(hidden)]
    Thz = 6366,
    #[doc(hidden)]
    Tia = 6367,
    #[doc(hidden)]
    Tic = 6368,
    #[doc(hidden)]
    Tif = 6369,
    #[doc(hidden)]
    Tig = 6370,
    #[doc(hidden)]
    Tih = 6371,
    #[doc(hidden)]
    Tii = 6372,
    #[doc(hidden)]
    Tij = 6373,
    #[doc(hidden)]
    Tik = 6374,
    #[doc(hidden)]
    Til = 6375,
    #[doc(hidden)]
    Tim = 6376,
    #[doc(hidden)]
    Tin = 6377,
    #[doc(hidden)]
    Tio = 6378,
    #[doc(hidden)]
    Tip = 6379,
    #[doc(hidden)]
    Tiq = 6380,
    #[doc(hidden)]
    Tir = 6381,
    #[doc(hidden)]
    Tis = 6382,
    #[doc(hidden)]
    Tit = 6383,
    #[doc(hidden)]
    Tiu = 6384,
    #[doc(hidden)]
    Tiv = 6385,
    #[doc(hidden)]
    Tiw = 6386,
    #[doc(hidden)]
    Tix = 6387,
    #[doc(hidden)]
    Tiy = 6388,
    #[doc(hidden)]
    Tiz = 6389,
    #[doc(hidden)]
    Tja = 6390,
    #[doc(hidden)]
    Tjg = 6391,
    #[doc(hidden)]
    Tji = 6392,
    #[doc(hidden)]
    Tjj = 6393,
    #[doc(hidden)]
    Tjl = 6394,
    #[doc(hidden)]
    Tjm = 6395,
    #[doc(hidden)]
    Tjn = 6396,
    #[doc(hidden)]
    Tjo = 6397,
    #[doc(hidden)]
    Tjp = 6398,
    #[doc(hidden)]
    Tjs = 6399,
    #[doc(hidden)]
    Tju = 6400,
    #[doc(hidden)]
    Tjw = 6401,
    #[doc(hidden)]
    Tka = 6402,
    #[doc(hidden)]
    Tkb = 6403,
    #[doc(hidden)]
    Tkd = 6404,
    #[doc(hidden)]
    Tke = 6405,
    #[doc(hidden)]
    Tkf = 6406,
    #[doc(hidden)]
    Tkg = 6407,
    #[doc(hidden)]
    Tkl = 6408,
    #[doc(hidden)]
    Tkm = 6409,
    #[doc(hidden)]
    Tkn = 6410,
    #[doc(hidden)]
    Tkp = 6411,
    #[doc(hidden)]
    Tkq = 6412,
    #[doc(hidden)]
    Tkr = 6413,
    #[doc(hidden)]
    Tks = 6414,
    #[doc(hidden)]
    Tkt = 6415,
    #[doc(hidden)]
    Tku = 6416,
    #[doc(hidden)]
    Tkv = 6417,
    #[doc(hidden)]
    Tkw = 6418,
    #[doc(hidden)]
    Tkx = 6419,
    #[doc(hidden)]
    Tkz = 6420,
    #[doc(hidden)]
    Tla = 6421,
    #[doc(hidden)]
    Tlb = 6422,
    #[doc(hidden)]
    Tlc = 6423,
    #[doc(hidden)]
    Tld = 6424,
    #[doc(hidden)]
    Tlf = 6425,
    #[doc(hidden)]
    Tlg = 6426,
    #[doc(hidden)]
    Tlh = 6427,
    #[doc(hidden)]
    Tli = 6428,
    #[doc(hidden)]
    Tlj = 6429,
    #[doc(hidden)]
    Tlk = 6430,
    #[doc(hidden)]
    Tll = 6431,
    #[doc(hidden)]
    Tlm = 6432,
    #[doc(hidden)]
    Tln = 6433,
    #[doc(hidden)]
    Tlo = 6434,
    #[doc(hidden)]
    Tlp = 6435,
    #[doc(hidden)]
    Tlq = 6436,
    #[doc(hidden)]
    Tlr = 6437,
    #[doc(hidden)]
    Tls = 6438,
    #[doc(hidden)]
    Tlt = 6439,
    #[doc(hidden)]
    Tlu = 6440,
    #[doc(hidden)]
    Tlv = 6441,
    #[doc(hidden)]
    Tlx = 6442,
    #[doc(hidden)]
    Tly = 6443,
    #[doc(hidden)]
    Tma = 6444,
    #[doc(hidden)]
    Tmb = 6445,
    #[doc(hidden)]
    Tmc = 6446,
    #[doc(hidden)]
    Tmd = 6447,
    #[doc(hidden)]
    Tme = 6448,
    #[doc(hidden)]
    Tmf = 6449,
    #[doc(hidden)]
    Tmg = 6450,
    #[doc(hidden)]
    Tmh = 6451,
    #[doc(hidden)]
    Tmi = 6452,
    #[doc(hidden)]
    Tmj = 6453,
    #[doc(hidden)]
    Tmk = 6454,
    #[doc(hidden)]
    Tml = 6455,
    #[doc(hidden)]
    Tmm = 6456,
    #[doc(hidden)]
    Tmn = 6457,
    #[doc(hidden)]
    Tmo = 6458,
    #[doc(hidden)]
    Tmq = 6459,
    #[doc(hidden)]
    Tmr = 6460,
    #[doc(hidden)]
    Tms = 6461,
    #[doc(hidden)]
    Tmt = 6462,
    #[doc(hidden)]
    Tmu = 6463,
    #[doc(hidden)]
    Tmv = 6464,
    #[doc(hidden)]
    Tmw = 6465,
    #[doc(hidden)]
    Tmy = 6466,
    #[doc(hidden)]
    Tmz = 6467,
    #[doc(hidden)]
    Tna = 6468,
    #[doc(hidden)]
    Tnb = 6469,
    #[doc(hidden)]
    Tnc = 6470,
    #[doc(hidden)]
    Tnd = 6471,
    #[doc(hidden)]
    Tng = 6472,
    #[doc(hidden)]
    Tnh = 6473,
    #[doc(hidden)]
    Tni = 6474,
    #[doc(hidden)]
    Tnk = 6475,
    #[doc(hidden)]
    Tnl = 6476,
    #[doc(hidden)]
    Tnm = 6477,
    #[doc(hidden)]
    Tnn = 6478,
    #[doc(hidden)]
    Tno = 6479,
    #[doc(hidden)]
    Tnp = 6480,
    #[doc(hidden)]
    Tnq = 6481,
    #[doc(hidden)]
    Tnr = 6482,
    #[doc(hidden)]
    Tns = 6483,
    #[doc(hidden)]
    Tnt = 6484,
    #[doc(hidden)]
    Tnu = 6485,
    #[doc(hidden)]
    Tnv = 6486,
    #[doc(hidden)]
    Tnw = 6487,
    #[doc(hidden)]
    Tnx = 6488,
    #[doc(hidden)]
    Tny = 6489,
    #[doc(hidden)]
    Tnz = 6490,
    #[doc(hidden)]
    Tob = 6491,
    #[doc(hidden)]
    Toc = 6492,
    #[doc(hidden)]
    Tod = 6493,
    #[doc(hidden)]
    Tof = 6494,
    #[doc(hidden)]
    Tog = 6495,
    #[doc(hidden)]
    Toh = 6496,
    #[doc(hidden)]
    Toi = 6497,
    #[doc(hidden)]
    Toj = 6498,
    #[doc(hidden)]
    Tok = 6499,
    #[doc(hidden)]
    Tol = 6500,
    #[doc(hidden)]
    Tom = 6501,
    #[doc(hidden)]
    Ton = 6502,
    #[doc(hidden)]
    Too = 6503,
    #[doc(hidden)]
    Top = 6504,
    #[doc(hidden)]
    Toq = 6505,
    #[doc(hidden)]
    Tor = 6506,
    #[doc(hidden)]
    Tos = 6507,
    #[doc(hidden)]
    Tou = 6508,
    #[doc(hidden)]
    Tov = 6509,
    #[doc(hidden)]
    Tow = 6510,
    #[doc(hidden)]
    Tox = 6511,
    #[doc(hidden)]
    Toy = 6512,
    #[doc(hidden)]
    Toz = 6513,
    #[doc(hidden)]
    Tpa = 6514,
    #[doc(hidden)]
    Tpc = 6515,
    #[doc(hidden)]
    Tpe = 6516,
    #[doc(hidden)]
    Tpf = 6517,
    #[doc(hidden)]
    Tpg = 6518,
    #[doc(hidden)]
    Tpi = 6519,
    #[doc(hidden)]
    Tpj = 6520,
    #[doc(hidden)]
    Tpk = 6521,
    #[doc(hidden)]
    Tpl = 6522,
    #[doc(hidden)]
    Tpm = 6523,
    #[doc(hidden)]
    Tpn = 6524,
    #[doc(hidden)]
    Tpo = 6525,
    #[doc(hidden)]
    Tpp = 6526,
    #[doc(hidden)]
    Tpq = 6527,
    #[doc(hidden)]
    Tpr = 6528,
    #[doc(hidden)]
    Tpt = 6529,
    #[doc(hidden)]
    Tpu = 6530,
    #[doc(hidden)]
    Tpv = 6531,
    #[doc(hidden)]
    Tpw = 6532,
    #[doc(hidden)]
    Tpx = 6533,
    #[doc(hidden)]
    Tpy = 6534,
    #[doc(hidden)]
    Tpz = 6535,
    #[doc(hidden)]
    Tqb = 6536,
    #[doc(hidden)]
    Tql = 6537,
    #[doc(hidden)]
    Tqm = 6538,
    #[doc(hidden)]
    Tqn = 6539,
    #[doc(hidden)]
    Tqo = 6540,
    #[doc(hidden)]
    Tqp = 6541,
    #[doc(hidden)]
    Tqq = 6542,
    #[doc(hidden)]
    Tqr = 6543,
    #[doc(hidden)]
    Tqt = 6544,
    #[doc(hidden)]
    Tqu = 6545,
    #[doc(hidden)]
    Tqw = 6546,
    #[doc(hidden)]
    Tra = 6547,
    #[doc(hidden)]
    Trb = 6548,
    #[doc(hidden)]
    Trc = 6549,
    #[doc(hidden)]
    Trd = 6550,
    #[doc(hidden)]
    Tre = 6551,
    #[doc(hidden)]
    Trf = 6552,
    #[doc(hidden)]
    Trg = 6553,
    #[doc(hidden)]
    Trh = 6554,
    #[doc(hidden)]
    Tri = 6555,
    #[doc(hidden)]
    Trj = 6556,
    #[doc(hidden)]
    Trl = 6557,
    #[doc(hidden)]
    Trm = 6558,
    #[doc(hidden)]
    Trn = 6559,
    #[doc(hidden)]
    Tro = 6560,
    #[doc(hidden)]
    Trp = 6561,
    #[doc(hidden)]
    Trq = 6562,
    #[doc(hidden)]
    Trr = 6563,
    #[doc(hidden)]
    Trs = 6564,
    #[doc(hidden)]
    Trt = 6565,
    #[doc(hidden)]
    Tru = 6566,
    #[doc(hidden)]
    Trv = 6567,
    #[doc(hidden)]
    Trw = 6568,
    #[doc(hidden)]
    Trx = 6569,
    #[doc(hidden)]
    Try = 6570,
    #[doc(hidden)]
    Trz = 6571,
    #[doc(hidden)]
    Tsa = 6572,
    #[doc(hidden)]
    Tsb = 6573,
    #[doc(hidden)]
    Tsc = 6574,
    #[doc(hidden)]
    Tsd = 6575,
    #[doc(hidden)]
    Tse = 6576,
    #[doc(hidden)]
    Tsg = 6577,
    #[doc(hidden)]
    Tsh = 6578,
    #[doc(hidden)]
    Tsi = 6579,
    #[doc(hidden)]
    Tsj = 6580,
    #[doc(hidden)]
    Tsk = 6581,
    #[doc(hidden)]
    Tsl = 6582,
    #[doc(hidden)]
    Tsm = 6583,
    #[doc(hidden)]
    Tsn = 6584,
    #[doc(hidden)]
    Tso = 6585,
    #[doc(hidden)]
    Tsp = 6586,
    #[doc(hidden)]
    Tsq = 6587,
    #[doc(hidden)]
    Tsr = 6588,
    #[doc(hidden)]
    Tss = 6589,
    #[doc(hidden)]
    Tst = 6590,
    #[doc(hidden)]
    Tsu = 6591,
    #[doc(hidden)]
    Tsv = 6592,
    #[doc(hidden)]
    Tsw = 6593,
    #[doc(hidden)]
    Tsx = 6594,
    #[doc(hidden)]
    Tsy = 6595,
    #[doc(hidden)]
    Tsz = 6596,
    #[doc(hidden)]
    Tta = 6597,
    #[doc(hidden)]
    Ttb = 6598,
    #[doc(hidden)]
    Ttc = 6599,
    #[doc(hidden)]
    Ttd = 6600,
    #[doc(hidden)]
    Tte = 6601,
    #[doc(hidden)]
    Ttf = 6602,
    #[doc(hidden)]
    Ttg = 6603,
    #[doc(hidden)]
    Tth = 6604,
    #[doc(hidden)]
    Tti = 6605,
    #[doc(hidden)]
    Ttj = 6606,
    #[doc(hidden)]
    Ttk = 6607,
    #[doc(hidden)]
    Ttl = 6608,
    #[doc(hidden)]
    Ttm = 6609,
    #[doc(hidden)]
    Ttn = 6610,
    #[doc(hidden)]
    Tto = 6611,
    #[doc(hidden)]
    Ttp = 6612,
    #[doc(hidden)]
    Ttq = 6613,
    #[doc(hidden)]
    Ttr = 6614,
    #[doc(hidden)]
    Tts = 6615,
    #[doc(hidden)]
    Ttt = 6616,
    #[doc(hidden)]
    Ttu = 6617,
    #[doc(hidden)]
    Ttv = 6618,
    #[doc(hidden)]
    Ttw = 6619,
    #[doc(hidden)]
    Tty = 6620,
    #[doc(hidden)]
    Ttz = 6621,
    #[doc(hidden)]
    Tua = 6622,
    #[doc(hidden)]
    Tub = 6623,
    #[doc(hidden)]
    Tuc = 6624,
    #[doc(hidden)]
    Tud = 6625,
    #[doc(hidden)]
    Tue = 6626,
    #[doc(hidden)]
    Tuf = 6627,
    #[doc(hidden)]
    Tug = 6628,
    #[doc(hidden)]
    Tuh = 6629,
    #[doc(hidden)]
    Tui = 6630,
    #[doc(hidden)]
    Tuj = 6631,
    #[doc(hidden)]
    Tuk = 6632,
    #[doc(hidden)]
    Tul = 6633,
    #[doc(hidden)]
    Tum = 6634,
    #[doc(hidden)]
    Tun = 6635,
    #[doc(hidden)]
    Tuo = 6636,
    #[doc(hidden)]
    Tuq = 6637,
    #[doc(hidden)]
    Tur = 6638,
    #[doc(hidden)]
    Tus = 6639,
    #[doc(hidden)]
    Tuu = 6640,
    #[doc(hidden)]
    Tuv = 6641,
    #[doc(hidden)]
    Tux = 6642,
    #[doc(hidden)]
    Tuy = 6643,
    #[doc(hidden)]
    Tuz = 6644,
    #[doc(hidden)]
    Tva = 6645,
    #[doc(hidden)]
    Tvd = 6646,
    #[doc(hidden)]
    Tve = 6647,
    #[doc(hidden)]
    Tvk = 6648,
    #[doc(hidden)]
    Tvl = 6649,
    #[doc(hidden)]
    Tvm = 6650,
    #[doc(hidden)]
    Tvn = 6651,
    #[doc(hidden)]
    Tvo = 6652,
    #[doc(hidden)]
    Tvs = 6653,
    #[doc(hidden)]
    Tvt = 6654,
    #[doc(hidden)]
    Tvu = 6655,
    #[doc(hidden)]
    Tvw = 6656,
    #[doc(hidden)]
    Tvx = 6657,
    #[doc(hidden)]
    Tvy = 6658,
    #[doc(hidden)]
    Twa = 6659,
    #[doc(hidden)]
    Twb = 6660,
    #[doc(hidden)]
    Twc = 6661,
    #[doc(hidden)]
    Twd = 6662,
    #[doc(hidden)]
    Twe = 6663,
    #[doc(hidden)]
    Twf = 6664,
    #[doc(hidden)]
    Twg = 6665,
    #[doc(hidden)]
    Twh = 6666,
    #[doc(hidden)]
    Twi = 6667,
    #[doc(hidden)]
    Twl = 6668,
    #[doc(hidden)]
    Twm = 6669,
    #[doc(hidden)]
    Twn = 6670,
    #[doc(hidden)]
    Two = 6671,
    #[doc(hidden)]
    Twp = 6672,
    #[doc(hidden)]
    Twq = 6673,
    #[doc(hidden)]
    Twr = 6674,
    #[doc(hidden)]
    Twt = 6675,
    #[doc(hidden)]
    Twu = 6676,
    #[doc(hidden)]
    Tww = 6677,
    #[doc(hidden)]
    Twx = 6678,
    #[doc(hidden)]
    Twy = 6679,
    #[doc(hidden)]
    Txa = 6680,
    #[doc(hidden)]
    Txb = 6681,
    #[doc(hidden)]
    Txc = 6682,
    #[doc(hidden)]
    Txe = 6683,
    #[doc(hidden)]
    Txg = 6684,
    #[doc(hidden)]
    Txh = 6685,
    #[doc(hidden)]
    Txi = 6686,
    #[doc(hidden)]
    Txj = 6687,
    #[doc(hidden)]
    Txm = 6688,
    #[doc(hidden)]
    Txn = 6689,
    #[doc(hidden)]
    Txo = 6690,
    #[doc(hidden)]
    Txq = 6691,
    #[doc(hidden)]
    Txr = 6692,
    #[doc(hidden)]
    Txs = 6693,
    #[doc(hidden)]
    Txt = 6694,
    #[doc(hidden)]
    Txu = 6695,
    #[doc(hidden)]
    Txx = 6696,
    #[doc(hidden)]
    Txy = 6697,
    #[doc(hidden)]
    Tya = 6698,
    #[doc(hidden)]
    Tye = 6699,
    #[doc(hidden)]
    Tyh = 6700,
    #[doc(hidden)]
    Tyi = 6701,
    #[doc(hidden)]
    Tyj = 6702,
    #[doc(hidden)]
    Tyl = 6703,
    #[doc(hidden)]
    Tyn = 6704,
    #[doc(hidden)]
    Typ = 6705,
    #[doc(hidden)]
    Tyr = 6706,
    #[doc(hidden)]
    Tys = 6707,
    #[doc(hidden)]
    Tyt = 6708,
    #[doc(hidden)]
    Tyu = 6709,
    #[doc(hidden)]
    Tyv = 6710,
    #[doc(hidden)]
    Tyx = 6711,
    #[doc(hidden)]
    Tyy = 6712,
    #[doc(hidden)]
    Tyz = 6713,
    #[doc(hidden)]
    Tza = 6714,
    #[doc(hidden)]
    Tzh = 6715,
    #[doc(hidden)]
    Tzj = 6716,
    #[doc(hidden)]
    Tzl = 6717,
    #[doc(hidden)]
    Tzm = 6718,
    #[doc(hidden)]
    Tzn = 6719,
    #[doc(hidden)]
    Tzo = 6720,
    #[doc(hidden)]
    Tzx = 6721,
    #[doc(hidden)]
    Uam = 6722,
    #[doc(hidden)]
    Uan = 6723,
    #[doc(hidden)]
    Uar = 6724,
    #[doc(hidden)]
    Uba = 6725,
    #[doc(hidden)]
    Ubi = 6726,
    #[doc(hidden)]
    Ubl = 6727,
    #[doc(hidden)]
    Ubr = 6728,
    #[doc(hidden)]
    Ubu = 6729,
    #[doc(hidden)]
    Uby = 6730,
    #[doc(hidden)]
    Uda = 6731,
    #[doc(hidden)]
    Ude = 6732,
    #[doc(hidden)]
    Udg = 6733,
    #[doc(hidden)]
    Udi = 6734,
    #[doc(hidden)]
    Udj = 6735,
    #[doc(hidden)]
    Udl = 6736,
    #[doc(hidden)]
    Udm = 6737,
    #[doc(hidden)]
    Udu = 6738,
    #[doc(hidden)]
    Ues = 6739,
    #[doc(hidden)]
    Ufi = 6740,
    #[doc(hidden)]
    Uga = 6741,
    #[doc(hidden)]
    Ugb = 6742,
    #[doc(hidden)]
    Uge = 6743,
    #[doc(hidden)]
    Ugh = 6744,
    #[doc(hidden)]
    Ugn = 6745,
    #[doc(hidden)]
    Ugo = 6746,
    #[doc(hidden)]
    Ugy = 6747,
    #[doc(hidden)]
    Uha = 6748,
    #[doc(hidden)]
    Uhn = 6749,
    #[doc(hidden)]
    Uig = 6750,
    #[doc(hidden)]
    Uis = 6751,
    #[doc(hidden)]
    Uiv = 6752,
    #[doc(hidden)]
    Uji = 6753,
    #[doc(hidden)]
    Uka = 6754,
    #[doc(hidden)]
    Ukg = 6755,
    #[doc(hidden)]
    Ukh = 6756,
    #[doc(hidden)]
    Uki = 6757,
    #[doc(hidden)]
    Ukk = 6758,
    #[doc(hidden)]
    Ukl = 6759,
    #[doc(hidden)]
    Ukp = 6760,
    #[doc(hidden)]
    Ukq = 6761,
    #[doc(hidden)]
    Ukr = 6762,
    #[doc(hidden)]
    Uks = 6763,
    #[doc(hidden)]
    Uku = 6764,
    #[doc(hidden)]
    Ukv = 6765,
    #[doc(hidden)]
    Ukw = 6766,
    #[doc(hidden)]
    Uky = 6767,
    #[doc(hidden)]
    Ula = 6768,
    #[doc(hidden)]
    Ulb = 6769,
    #[doc(hidden)]
    Ulc = 6770,
    #[doc(hidden)]
    Ule = 6771,
    #[doc(hidden)]
    Ulf = 6772,
    #[doc(hidden)]
    Uli = 6773,
    #[doc(hidden)]
    Ulk = 6774,
    #[doc(hidden)]
    Ull = 6775,
    #[doc(hidden)]
    Ulm = 6776,
    #[doc(hidden)]
    Uln = 6777,
    #[doc(hidden)]
    Ulu = 6778,
    #[doc(hidden)]
    Ulw = 6779,
    #[doc(hidden)]
    Uma = 6780,
    #[doc(hidden)]
    Umb = 6781,
    #[doc(hidden)]
    Umc = 6782,
    #[doc(hidden)]
    Umd = 6783,
    #[doc(hidden)]
    Umg = 6784,
    #[doc(hidden)]
    Umi = 6785,
    #[doc(hidden)]
    Umm = 6786,
    #[doc(hidden)]
    Umn = 6787,
    #[doc(hidden)]
    Umo = 6788,
    #[doc(hidden)]
    Ump = 6789,
    #[doc(hidden)]
    Umr = 6790,
    #[doc(hidden)]
    Ums = 6791,
    #[doc(hidden)]
    Umu = 6792,
    #[doc(hidden)]
    Una = 6793,
    #[doc(hidden)]
    Und = 6794,
    #[doc(hidden)]
    Une = 6795,
    #[doc(hidden)]
    Ung = 6796,
    #[doc(hidden)]
    Uni = 6797,
    #[doc(hidden)]
    Unk = 6798,
    #[doc(hidden)]
    Unm = 6799,
    #[doc(hidden)]
    Unn = 6800,
    #[doc(hidden)]
    Unr = 6801,
    #[doc(hidden)]
    Unu = 6802,
    #[doc(hidden)]
    Unx = 6803,
    #[doc(hidden)]
    Unz = 6804,
    #[doc(hidden)]
    Uon = 6805,
    #[doc(hidden)]
    Upi = 6806,
    #[doc(hidden)]
    Upv = 6807,
    #[doc(hidden)]
    Ura = 6808,
    #[doc(hidden)]
    Urb = 6809,
    #[doc(hidden)]
    Urc = 6810,
    #[doc(hidden)]
    Urd = 6811,
    #[doc(hidden)]
    Ure = 6812,
    #[doc(hidden)]
    Urf = 6813,
    #[doc(hidden)]
    Urg = 6814,
    #[doc(hidden)]
    Urh = 6815,
    #[doc(hidden)]
    Uri = 6816,
    #[doc(hidden)]
    Urk = 6817,
    #[doc(hidden)]
    Url = 6818,
    #[doc(hidden)]
    Urm = 6819,
    #[doc(hidden)]
    Urn = 6820,
    #[doc(hidden)]
    Uro = 6821,
    #[doc(hidden)]
    Urp = 6822,
    #[doc(hidden)]
    Urr = 6823,
    #[doc(hidden)]
    Urt = 6824,
    #[doc(hidden)]
    Uru = 6825,
    #[doc(hidden)]
    Urv = 6826,
    #[doc(hidden)]
    Urw = 6827,
    #[doc(hidden)]
    Urx = 6828,
    #[doc(hidden)]
    Ury = 6829,
    #[doc(hidden)]
    Urz = 6830,
    #[doc(hidden)]
    Usa = 6831,
    #[doc(hidden)]
    Ush = 6832,
    #[doc(hidden)]
    Usi = 6833,
    #[doc(hidden)]
    Usk = 6834,
    #[doc(hidden)]
    Usp = 6835,
    #[doc(hidden)]
    Uss = 6836,
    #[doc(hidden)]
    Usu = 6837,
    #[doc(hidden)]
    Uta = 6838,
    #[doc(hidden)]
    Ute = 6839,
    #[doc(hidden)]
    Uth = 6840,
    #[doc(hidden)]
    Utp = 6841,
    #[doc(hidden)]
    Utr = 6842,
    #[doc(hidden)]
    Utu = 6843,
    #[doc(hidden)]
    Uum = 6844,
    #[doc(hidden)]
    Uur = 6845,
    #[doc(hidden)]
    Uuu = 6846,
    #[doc(hidden)]
    Uve = 6847,
    #[doc(hidden)]
    Uvh = 6848,
    #[doc(hidden)]
    Uvl = 6849,
    #[doc(hidden)]
    Uwa = 6850,
    #[doc(hidden)]
    Uya = 6851,
    #[doc(hidden)]
    Uzb = 6852,
    #[doc(hidden)]
    Uzn = 6853,
    #[doc(hidden)]
    Uzs = 6854,
    #[doc(hidden)]
    Vaa = 6855,
    #[doc(hidden)]
    Vae = 6856,
    #[doc(hidden)]
    Vaf = 6857,
    #[doc(hidden)]
    Vag = 6858,
    #[doc(hidden)]
    Vah = 6859,
    #[doc(hidden)]
    Vai = 6860,
    #[doc(hidden)]
    Vaj = 6861,
    #[doc(hidden)]
    Val = 6862,
    #[doc(hidden)]
    Vam = 6863,
    #[doc(hidden)]
    Van = 6864,
    #[doc(hidden)]
    Vao = 6865,
    #[doc(hidden)]
    Vap = 6866,
    #[doc(hidden)]
    Var = 6867,
    #[doc(hidden)]
    Vas = 6868,
    #[doc(hidden)]
    Vau = 6869,
    #[doc(hidden)]
    Vav = 6870,
    #[doc(hidden)]
    Vay = 6871,
    #[doc(hidden)]
    Vbb = 6872,
    #[doc(hidden)]
    Vbk = 6873,
    #[doc(hidden)]
    Vec = 6874,
    #[doc(hidden)]
    Ved = 6875,
    #[doc(hidden)]
    Vel = 6876,
    #[doc(hidden)]
    Vem = 6877,
    #[doc(hidden)]
    Ven = 6878,
    #[doc(hidden)]
    Veo = 6879,
    #[doc(hidden)]
    Vep = 6880,
    #[doc(hidden)]
    Ver = 6881,
    #[doc(hidden)]
    Vgr = 6882,
    #[doc(hidden)]
    Vgt = 6883,
    #[doc(hidden)]
    Vic = 6884,
    #[doc(hidden)]
    Vid = 6885,
    #[doc(hidden)]
    Vie = 6886,
    #[doc(hidden)]
    Vif = 6887,
    #[doc(hidden)]
    Vig = 6888,
    #[doc(hidden)]
    Vil = 6889,
    #[doc(hidden)]
    Vin = 6890,
    #[doc(hidden)]
    Vis = 6891,
    #[doc(hidden)]
    Vit = 6892,
    #[doc(hidden)]
    Viv = 6893,
    #[doc(hidden)]
    Vka = 6894,
    #[doc(hidden)]
    Vkj = 6895,
    #[doc(hidden)]
    Vkk = 6896,
    #[doc(hidden)]
    Vkl = 6897,
    #[doc(hidden)]
    Vkm = 6898,
    #[doc(hidden)]
    Vkn = 6899,
    #[doc(hidden)]
    Vko = 6900,
    #[doc(hidden)]
    Vkp = 6901,
    #[doc(hidden)]
    Vkt = 6902,
    #[doc(hidden)]
    Vku = 6903,
    #[doc(hidden)]
    Vkz = 6904,
    #[doc(hidden)]
    Vlp = 6905,
    #[doc(hidden)]
    Vls = 6906,
    #[doc(hidden)]
    Vma = 6907,
    #[doc(hidden)]
    Vmb = 6908,
    #[doc(hidden)]
    Vmc = 6909,
    #[doc(hidden)]
    Vmd = 6910,
    #[doc(hidden)]
    Vme = 6911,
    #[doc(hidden)]
    Vmf = 6912,
    #[doc(hidden)]
    Vmg = 6913,
    #[doc(hidden)]
    Vmh = 6914,
    #[doc(hidden)]
    Vmi = 6915,
    #[doc(hidden)]
    Vmj = 6916,
    #[doc(hidden)]
    Vmk = 6917,
    #[doc(hidden)]
    Vml = 6918,
    #[doc(hidden)]
    Vmm = 6919,
    #[doc(hidden)]
    Vmp = 6920,
    #[doc(hidden)]
    Vmq = 6921,
    #[doc(hidden)]
    Vmr = 6922,
    #[doc(hidden)]
    Vms = 6923,
    #[doc(hidden)]
    Vmu = 6924,
    #[doc(hidden)]
    Vmv = 6925,
    #[doc(hidden)]
    Vmw = 6926,
    #[doc(hidden)]
    Vmx = 6927,
    #[doc(hidden)]
    Vmy = 6928,
    #[doc(hidden)]
    Vmz = 6929,
    #[doc(hidden)]
    Vnk = 6930,
    #[doc(hidden)]
    Vnm = 6931,
    #[doc(hidden)]
    Vnp = 6932,
    #[doc(hidden)]
    Vol = 6933,
    #[doc(hidden)]
    Vor = 6934,
    #[doc(hidden)]
    Vot = 6935,
    #[doc(hidden)]
    Vra = 6936,
    #[doc(hidden)]
    Vro = 6937,
    #[doc(hidden)]
    Vrs = 6938,
    #[doc(hidden)]
    Vrt = 6939,
    #[doc(hidden)]
    Vsi = 6940,
    #[doc(hidden)]
    Vsl = 6941,
    #[doc(hidden)]
    Vsv = 6942,
    #[doc(hidden)]
    Vto = 6943,
    #[doc(hidden)]
    Vum = 6944,
    #[doc(hidden)]
    Vun = 6945,
    #[doc(hidden)]
    Vut = 6946,
    #[doc(hidden)]
    Vwa = 6947,
    #[doc(hidden)]
    Waa = 6948,
    #[doc(hidden)]
    Wab = 6949,
    #[doc(hidden)]
    Wac = 6950,
    #[doc(hidden)]
    Wad = 6951,
    #[doc(hidden)]
    Wae = 6952,
    #[doc(hidden)]
    Waf = 6953,
    #[doc(hidden)]
    Wag = 6954,
    #[doc(hidden)]
    Wah = 6955,
    #[doc(hidden)]
    Wai = 6956,
    #[doc(hidden)]
    Waj = 6957,
    #[doc(hidden)]
    Wal = 6958,
    #[doc(hidden)]
    Wam = 6959,
    #[doc(hidden)]
    Wan = 6960,
    #[doc(hidden)]
    Wao = 6961,
    #[doc(hidden)]
    Wap = 6962,
    #[doc(hidden)]
    Waq = 6963,
    #[doc(hidden)]
    War = 6964,
    #[doc(hidden)]
    Was = 6965,
    #[doc(hidden)]
    Wat = 6966,
    #[doc(hidden)]
    Wau = 6967,
    #[doc(hidden)]
    Wav = 6968,
    #[doc(hidden)]
    Waw = 6969,
    #[doc(hidden)]
    Wax = 6970,
    #[doc(hidden)]
    Way = 6971,
    #[doc(hidden)]
    Waz = 6972,
    #[doc(hidden)]
    Wba = 6973,
    #[doc(hidden)]
    Wbb = 6974,
    #[doc(hidden)]
    Wbe = 6975,
    #[doc(hidden)]
    Wbf = 6976,
    #[doc(hidden)]
    Wbh = 6977,
    #[doc(hidden)]
    Wbi = 6978,
    #[doc(hidden)]
    Wbj = 6979,
    #[doc(hidden)]
    Wbk = 6980,
    #[doc(hidden)]
    Wbl = 6981,
    #[doc(hidden)]
    Wbm = 6982,
    #[doc(hidden)]
    Wbp = 6983,
    #[doc(hidden)]
    Wbq = 6984,
    #[doc(hidden)]
    Wbr = 6985,
    #[doc(hidden)]
    Wbs = 6986,
    #[doc(hidden)]
    Wbt = 6987,
    #[doc(hidden)]
    Wbv = 6988,
    #[doc(hidden)]
    Wbw = 6989,
    #[doc(hidden)]
    Wca = 6990,
    #[doc(hidden)]
    Wci = 6991,
    #[doc(hidden)]
    Wdd = 6992,
    #[doc(hidden)]
    Wdg = 6993,
    #[doc(hidden)]
    Wdj = 6994,
    #[doc(hidden)]
    Wdk = 6995,
    #[doc(hidden)]
    Wdt = 6996,
    #[doc(hidden)]
    Wdu = 6997,
    #[doc(hidden)]
    Wdy = 6998,
    #[doc(hidden)]
    Wea = 6999,
    #[doc(hidden)]
    Wec = 7000,
    #[doc(hidden)]
    Wed = 7001,
    #[doc(hidden)]
    Weg = 7002,
    #[doc(hidden)]
    Weh = 7003,
    #[doc(hidden)]
    Wei = 7004,
    #[doc(hidden)]
    Wem = 7005,
    #[doc(hidden)]
    Weo = 7006,
    #[doc(hidden)]
    Wep = 7007,
    #[doc(hidden)]
    Wer = 7008,
    #[doc(hidden)]
    Wes = 7009,
    #[doc(hidden)]
    Wet = 7010,
    #[doc(hidden)]
    Weu = 7011,
    #[doc(hidden)]
    Wew = 7012,
    #[doc(hidden)]
    Wfg = 7013,
    #[doc(hidden)]
    Wga = 7014,
    #[doc(hidden)]
    Wgb = 7015,
    #[doc(hidden)]
    Wgg = 7016,
    #[doc(hidden)]
    Wgi = 7017,
    #[doc(hidden)]
    Wgo = 7018,
    #[doc(hidden)]
    Wgu = 7019,
    #[doc(hidden)]
    Wgy = 7020,
    #[doc(hidden)]
    Wha = 7021,
    #[doc(hidden)]
    Whg = 7022,
    #[doc(hidden)]
    Whk = 7023,
    #[doc(hidden)]
    Whu = 7024,
    #[doc(hidden)]
    Wib = 7025,
    #[doc(hidden)]
    Wic = 7026,
    #[doc(hidden)]
    Wie = 7027,
    #[doc(hidden)]
    Wif = 7028,
    #[doc(hidden)]
    Wig = 7029,
    #[doc(hidden)]
    Wih = 7030,
    #[doc(hidden)]
    Wii = 7031,
    #[doc(hidden)]
    Wij = 7032,
    #[doc(hidden)]
    Wik = 7033,
    #[doc(hidden)]
    Wil = 7034,
    #[doc(hidden)]
    Wim = 7035,
    #[doc(hidden)]
    Win = 7036,
    #[doc(hidden)]
    Wir = 7037,
    #[doc(hidden)]
    Wiu = 7038,
    #[doc(hidden)]
    Wiv = 7039,
    #[doc(hidden)]
    Wiy = 7040,
    #[doc(hidden)]
    Wja = 7041,
    #[doc(hidden)]
    Wji = 7042,
    #[doc(hidden)]
    Wka = 7043,
    #[doc(hidden)]
    Wkb = 7044,
    #[doc(hidden)]
    Wkd = 7045,
    #[doc(hidden)]
    Wkl = 7046,
    #[doc(hidden)]
    Wkr = 7047,
    #[doc(hidden)]
    Wku = 7048,
    #[doc(hidden)]
    Wkw = 7049,
    #[doc(hidden)]
    Wky = 7050,
    #[doc(hidden)]
    Wla = 7051,
    #[doc(hidden)]
    Wlc = 7052,
    #[doc(hidden)]
    Wle = 7053,
    #[doc(hidden)]
    Wlg = 7054,
    #[doc(hidden)]
    Wlh = 7055,
    #[doc(hidden)]
    Wli = 7056,
    #[doc(hidden)]
    Wlk = 7057,
    #[doc(hidden)]
    Wll = 7058,
    #[doc(hidden)]
    Wlm = 7059,
    #[doc(hidden)]
    Wln = 7060,
    #[doc(hidden)]
    Wlo = 7061,
    #[doc(hidden)]
    Wlr = 7062,
    #[doc(hidden)]
    Wls = 7063,
    #[doc(hidden)]
    Wlu = 7064,
    #[doc(hidden)]
    Wlv = 7065,
    #[doc(hidden)]
    Wlw = 7066,
    #[doc(hidden)]
    Wlx = 7067,
    #[doc(hidden)]
    Wly = 7068,
    #[doc(hidden)]
    Wma = 7069,
    #[doc(hidden)]
    Wmb = 7070,
    #[doc(hidden)]
    Wmc = 7071,
    #[doc(hidden)]
    Wmd = 7072,
    #[doc(hidden)]
    Wme = 7073,
    #[doc(hidden)]
    Wmg = 7074,
    #[doc(hidden)]
    Wmh = 7075,
    #[doc(hidden)]
    Wmi = 7076,
    #[doc(hidden)]
    Wmm = 7077,
    #[doc(hidden)]
    Wmn = 7078,
    #[doc(hidden)]
    Wmo = 7079,
    #[doc(hidden)]
    Wms = 7080,
    #[doc(hidden)]
    Wmt = 7081,
    #[doc(hidden)]
    Wmw = 7082,
    #[doc(hidden)]
    Wmx = 7083,
    #[doc(hidden)]
    Wnb = 7084,
    #[doc(hidden)]
    Wnc = 7085,
    #[doc(hidden)]
    Wnd = 7086,
    #[doc(hidden)]
    Wne = 7087,
    #[doc(hidden)]
    Wng = 7088,
    #[doc(hidden)]
    Wni = 7089,
    #[doc(hidden)]
    Wnk = 7090,
    #[doc(hidden)]
    Wnm = 7091,
    #[doc(hidden)]
    Wnn = 7092,
    #[doc(hidden)]
    Wno = 7093,
    #[doc(hidden)]
    Wnp = 7094,
    #[doc(hidden)]
    Wnu = 7095,
    #[doc(hidden)]
    Wnw = 7096,
    #[doc(hidden)]
    Wny = 7097,
    #[doc(hidden)]
    Woa = 7098,
    #[doc(hidden)]
    Wob = 7099,
    #[doc(hidden)]
    Woc = 7100,
    #[doc(hidden)]
    Wod = 7101,
    #[doc(hidden)]
    Woe = 7102,
    #[doc(hidden)]
    Wof = 7103,
    #[doc(hidden)]
    Wog = 7104,
    #[doc(hidden)]
    Woi = 7105,
    #[doc(hidden)]
    Wok = 7106,
    #[doc(hidden)]
    Wol = 7107,
    #[doc(hidden)]
    Wom = 7108,
    #[doc(hidden)]
    Won = 7109,
    #[doc(hidden)]
    Woo = 7110,
    #[doc(hidden)]
    Wor = 7111,
    #[doc(hidden)]
    Wos = 7112,
    #[doc(hidden)]
    Wow = 7113,
    #[doc(hidden)]
    Woy = 7114,
    #[doc(hidden)]
    Wpc = 7115,
    #[doc(hidden)]
    Wrb = 7116,
    #[doc(hidden)]
    Wrg = 7117,
    #[doc(hidden)]
    Wrh = 7118,
    #[doc(hidden)]
    Wri = 7119,
    #[doc(hidden)]
    Wrk = 7120,
    #[doc(hidden)]
    Wrl = 7121,
    #[doc(hidden)]
    Wrm = 7122,
    #[doc(hidden)]
    Wrn = 7123,
    #[doc(hidden)]
    Wro = 7124,
    #[doc(hidden)]
    Wrp = 7125,
    #[doc(hidden)]
    Wrr = 7126,
    #[doc(hidden)]
    Wrs = 7127,
    #[doc(hidden)]
    Wru = 7128,
    #[doc(hidden)]
    Wrv = 7129,
    #[doc(hidden)]
    Wrw = 7130,
    #[doc(hidden)]
    Wrx = 7131,
    #[doc(hidden)]
    Wry = 7132,
    #[doc(hidden)]
    Wrz = 7133,
    #[doc(hidden)]
    Wsa = 7134,
    #[doc(hidden)]
    Wsg = 7135,
    #[doc(hidden)]
    Wsi = 7136,
    #[doc(hidden)]
    Wsk = 7137,
    #[doc(hidden)]
    Wsr = 7138,
    #[doc(hidden)]
    Wss = 7139,
    #[doc(hidden)]
    Wsu = 7140,
    #[doc(hidden)]
    Wsv = 7141,
    #[doc(hidden)]
    Wtf = 7142,
    #[doc(hidden)]
    Wth = 7143,
    #[doc(hidden)]
    Wti = 7144,
    #[doc(hidden)]
    Wtk = 7145,
    #[doc(hidden)]
    Wtm = 7146,
    #[doc(hidden)]
    Wtw = 7147,
    #[doc(hidden)]
    Wua = 7148,
    #[doc(hidden)]
    Wub = 7149,
    #[doc(hidden)]
    Wud = 7150,
    #[doc(hidden)]
    Wuh = 7151,
    #[doc(hidden)]
    Wul = 7152,
    #[doc(hidden)]
    Wum = 7153,
    #[doc(hidden)]
    Wun = 7154,
    #[doc(hidden)]
    Wur = 7155,
    #[doc(hidden)]
    Wut = 7156,
    #[doc(hidden)]
    Wuu = 7157,
    #[doc(hidden)]
    Wuv = 7158,
    #[doc(hidden)]
    Wux = 7159,
    #[doc(hidden)]
    Wuy = 7160,
    #[doc(hidden)]
    Wwa = 7161,
    #[doc(hidden)]
    Wwb = 7162,
    #[doc(hidden)]
    Wwo = 7163,
    #[doc(hidden)]
    Wwr = 7164,
    #[doc(hidden)]
    Www = 7165,
    #[doc(hidden)]
    Wxa = 7166,
    #[doc(hidden)]
    Wxw = 7167,
    #[doc(hidden)]
    Wyb = 7168,
    #[doc(hidden)]
    Wyi = 7169,
    #[doc(hidden)]
    Wym = 7170,
    #[doc(hidden)]
    Wyn = 7171,
    #[doc(hidden)]
    Wyr = 7172,
    #[doc(hidden)]
    Wyy = 7173,
    #[doc(hidden)]
    Xaa = 7174,
    #[doc(hidden)]
    Xab = 7175,
    #[doc(hidden)]
    Xac = 7176,
    #[doc(hidden)]
    Xad = 7177,
    #[doc(hidden)]
    Xae = 7178,
    #[doc(hidden)]
    Xag = 7179,
    #[doc(hidden)]
    Xai = 7180,
    #[doc(hidden)]
    Xaj = 7181,
    #[doc(hidden)]
    Xak = 7182,
    #[doc(hidden)]
    Xal = 7183,
    #[doc(hidden)]
    Xam = 7184,
    #[doc(hidden)]
    Xan = 7185,
    #[doc(hidden)]
    Xao = 7186,
    #[doc(hidden)]
    Xap = 7187,
    #[doc(hidden)]
    Xaq = 7188,
    #[doc(hidden)]
    Xar = 7189,
    #[doc(hidden)]
    Xas = 7190,
    #[doc(hidden)]
    Xat = 7191,
    #[doc(hidden)]
    Xau = 7192,
    #[doc(hidden)]
    Xav = 7193,
    #[doc(hidden)]
    Xaw = 7194,
    #[doc(hidden)]
    Xay = 7195,
    #[doc(hidden)]
    Xbb = 7196,
    #[doc(hidden)]
    Xbc = 7197,
    #[doc(hidden)]
    Xbd = 7198,
    #[doc(hidden)]
    Xbe = 7199,
    #[doc(hidden)]
    Xbg = 7200,
    #[doc(hidden)]
    Xbi = 7201,
    #[doc(hidden)]
    Xbj = 7202,
    #[doc(hidden)]
    Xbm = 7203,
    #[doc(hidden)]
    Xbn = 7204,
    #[doc(hidden)]
    Xbo = 7205,
    #[doc(hidden)]
    Xbp = 7206,
    #[doc(hidden)]
    Xbr = 7207,
    #[doc(hidden)]
    Xbw = 7208,
    #[doc(hidden)]
    Xby = 7209,
    #[doc(hidden)]
    Xcb = 7210,
    #[doc(hidden)]
    Xcc = 7211,
    #[doc(hidden)]
    Xce = 7212,
    #[doc(hidden)]
    Xcg = 7213,
    #[doc(hidden)]
    Xch = 7214,
    #[doc(hidden)]
    Xcl = 7215,
    #[doc(hidden)]
    Xcm = 7216,
    #[doc(hidden)]
    Xcn = 7217,
    #[doc(hidden)]
    Xco = 7218,
    #[doc(hidden)]
    Xcr = 7219,
    #[doc(hidden)]
    Xct = 7220,
    #[doc(hidden)]
    Xcu = 7221,
    #[doc(hidden)]
    Xcv = 7222,
    #[doc(hidden)]
    Xcw = 7223,
    #[doc(hidden)]
    Xcy = 7224,
    #[doc(hidden)]
    Xda = 7225,
    #[doc(hidden)]
    Xdc = 7226,
    #[doc(hidden)]
    Xdk = 7227,
    #[doc(hidden)]
    Xdm = 7228,
    #[doc(hidden)]
    Xdo = 7229,
    #[doc(hidden)]
    Xdq = 7230,
    #[doc(hidden)]
    Xdy = 7231,
    #[doc(hidden)]
    Xeb = 7232,
    #[doc(hidden)]
    Xed = 7233,
    #[doc(hidden)]
    Xeg = 7234,
    #[doc(hidden)]
    Xel = 7235,
    #[doc(hidden)]
    Xem = 7236,
    #[doc(hidden)]
    Xep = 7237,
    #[doc(hidden)]
    Xer = 7238,
    #[doc(hidden)]
    Xes = 7239,
    #[doc(hidden)]
    Xet = 7240,
    #[doc(hidden)]
    Xeu = 7241,
    #[doc(hidden)]
    Xfa = 7242,
    #[doc(hidden)]
    Xga = 7243,
    #[doc(hidden)]
    Xgb = 7244,
    #[doc(hidden)]
    Xgd = 7245,
    #[doc(hidden)]
    Xgf = 7246,
    #[doc(hidden)]
    Xgg = 7247,
    #[doc(hidden)]
    Xgi = 7248,
    #[doc(hidden)]
    Xgl = 7249,
    #[doc(hidden)]
    Xgm = 7250,
    #[doc(hidden)]
    Xgr = 7251,
    #[doc(hidden)]
    Xgu = 7252,
    #[doc(hidden)]
    Xgw = 7253,
    #[doc(hidden)]
    Xha = 7254,
    #[doc(hidden)]
    Xhc = 7255,
    #[doc(hidden)]
    Xhd = 7256,
    #[doc(hidden)]
    Xhe = 7257,
    #[doc(hidden)]
    Xhm = 7258,
    #[doc(hidden)]
    Xho = 7259,
    #[doc(hidden)]
    Xhr = 7260,
    #[doc(hidden)]
    Xht = 7261,
    #[doc(hidden)]
    Xhu = 7262,
    #[doc(hidden)]
    Xhv = 7263,
    #[doc(hidden)]
    Xib = 7264,
    #[doc(hidden)]
    Xii = 7265,
    #[doc(hidden)]
    Xil = 7266,
    #[doc(hidden)]
    Xin = 7267,
    #[doc(hidden)]
    Xir = 7268,
    #[doc(hidden)]
    Xis = 7269,
    #[doc(hidden)]
    Xiv = 7270,
    #[doc(hidden)]
    Xiy = 7271,
    #[doc(hidden)]
    Xjb = 7272,
    #[doc(hidden)]
    Xjt = 7273,
    #[doc(hidden)]
    Xka = 7274,
    #[doc(hidden)]
    Xkb = 7275,
    #[doc(hidden)]
    Xkc = 7276,
    #[doc(hidden)]
    Xkd = 7277,
    #[doc(hidden)]
    Xke = 7278,
    #[doc(hidden)]
    Xkf = 7279,
    #[doc(hidden)]
    Xkg = 7280,
    #[doc(hidden)]
    Xki = 7281,
    #[doc(hidden)]
    Xkj = 7282,
    #[doc(hidden)]
    Xkk = 7283,
    #[doc(hidden)]
    Xkl = 7284,
    #[doc(hidden)]
    Xkn = 7285,
    #[doc(hidden)]
    Xko = 7286,
    #[doc(hidden)]
    Xkp = 7287,
    #[doc(hidden)]
    Xkq = 7288,
    #[doc(hidden)]
    Xkr = 7289,
    #[doc(hidden)]
    Xks = 7290,
    #[doc(hidden)]
    Xkt = 7291,
    #[doc(hidden)]
    Xku = 7292,
    #[doc(hidden)]
    Xkv = 7293,
    #[doc(hidden)]
    Xkw = 7294,
    #[doc(hidden)]
    Xkx = 7295,
    #[doc(hidden)]
    Xky = 7296,
    #[doc(hidden)]
    Xkz = 7297,
    #[doc(hidden)]
    Xla = 7298,
    #[doc(hidden)]
    Xlb = 7299,
    #[doc(hidden)]
    Xlc = 7300,
    #[doc(hidden)]
    Xld = 7301,
    #[doc(hidden)]
    Xle = 7302,
    #[doc(hidden)]
    Xlg = 7303,
    #[doc(hidden)]
    Xli = 7304,
    #[doc(hidden)]
    Xln = 7305,
    #[doc(hidden)]
    Xlo = 7306,
    #[doc(hidden)]
    Xlp = 7307,
    #[doc(hidden)]
    Xls = 7308,
    #[doc(hidden)]
    Xlu = 7309,
    #[doc(hidden)]
    Xly = 7310,
    #[doc(hidden)]
    Xma = 7311,
    #[doc(hidden)]
    Xmb = 7312,
    #[doc(hidden)]
    Xmc = 7313,
    #[doc(hidden)]
    Xmd = 7314,
    #[doc(hidden)]
    Xme = 7315,
    #[doc(hidden)]
    Xmf = 7316,
    #[doc(hidden)]
    Xmg = 7317,
    #[doc(hidden)]
    Xmh = 7318,
    #[doc(hidden)]
    Xmj = 7319,
    #[doc(hidden)]
    Xmk = 7320,
    #[doc(hidden)]
    Xml = 7321,
    #[doc(hidden)]
    Xmm = 7322,
    #[doc(hidden)]
    Xmn = 7323,
    #[doc(hidden)]
    Xmo = 7324,
    #[doc(hidden)]
    Xmp = 7325,
    #[doc(hidden)]
    Xmq = 7326,
    #[doc(hidden)]
    Xmr = 7327,
    #[doc(hidden)]
    Xms = 7328,
    #[doc(hidden)]
    Xmt = 7329,
    #[doc(hidden)]
    Xmu = 7330,
    #[doc(hidden)]
    Xmv = 7331,
    #[doc(hidden)]
    Xmw = 7332,
    #[doc(hidden)]
    Xmx = 7333,
    #[doc(hidden)]
    Xmy = 7334,
    #[doc(hidden)]
    Xmz = 7335,
    #[doc(hidden)]
    Xna = 7336,
    #[doc(hidden)]
    Xnb = 7337,
    #[doc(hidden)]
    Xng = 7338,
    #[doc(hidden)]
    Xnh = 7339,
    #[doc(hidden)]
    Xni = 7340,
    #[doc(hidden)]
    Xnj = 7341,
    #[doc(hidden)]
    Xnk = 7342,
    #[doc(hidden)]
    Xnm = 7343,
    #[doc(hidden)]
    Xnn = 7344,
    #[doc(hidden)]
    Xno = 7345,
    #[doc(hidden)]
    Xnq = 7346,
    #[doc(hidden)]
    Xnr = 7347,
    #[doc(hidden)]
    Xns = 7348,
    #[doc(hidden)]
    Xnt = 7349,
    #[doc(hidden)]
    Xnu = 7350,
    #[doc(hidden)]
    Xny = 7351,
    #[doc(hidden)]
    Xnz = 7352,
    #[doc(hidden)]
    Xoc = 7353,
    #[doc(hidden)]
    Xod = 7354,
    #[doc(hidden)]
    Xog = 7355,
    #[doc(hidden)]
    Xoi = 7356,
    #[doc(hidden)]
    Xok = 7357,
    #[doc(hidden)]
    Xom = 7358,
    #[doc(hidden)]
    Xon = 7359,
    #[doc(hidden)]
    Xoo = 7360,
    #[doc(hidden)]
    Xop = 7361,
    #[doc(hidden)]
    Xor = 7362,
    #[doc(hidden)]
    Xow = 7363,
    #[doc(hidden)]
    Xpa = 7364,
    #[doc(hidden)]
    Xpb = 7365,
    #[doc(hidden)]
    Xpc = 7366,
    #[doc(hidden)]
    Xpd = 7367,
    #[doc(hidden)]
    Xpe = 7368,
    #[doc(hidden)]
    Xpf = 7369,
    #[doc(hidden)]
    Xpg = 7370,
    #[doc(hidden)]
    Xph = 7371,
    #[doc(hidden)]
    Xpi = 7372,
    #[doc(hidden)]
    Xpj = 7373,
    #[doc(hidden)]
    Xpk = 7374,
    #[doc(hidden)]
    Xpl = 7375,
    #[doc(hidden)]
    Xpm = 7376,
    #[doc(hidden)]
    Xpn = 7377,
    #[doc(hidden)]
    Xpo = 7378,
    #[doc(hidden)]
    Xpp = 7379,
    #[doc(hidden)]
    Xpq = 7380,
    #[doc(hidden)]
    Xpr = 7381,
    #[doc(hidden)]
    Xps = 7382,
    #[doc(hidden)]
    Xpt = 7383,
    #[doc(hidden)]
    Xpu = 7384,
    #[doc(hidden)]
    Xpv = 7385,
    #[doc(hidden)]
    Xpw = 7386,
    #[doc(hidden)]
    Xpx = 7387,
    #[doc(hidden)]
    Xpy = 7388,
    #[doc(hidden)]
    Xpz = 7389,
    #[doc(hidden)]
    Xqa = 7390,
    #[doc(hidden)]
    Xqt = 7391,
    #[doc(hidden)]
    Xra = 7392,
    #[doc(hidden)]
    Xrb = 7393,
    #[doc(hidden)]
    Xrd = 7394,
    #[doc(hidden)]
    Xre = 7395,
    #[doc(hidden)]
    Xrg = 7396,
    #[doc(hidden)]
    Xri = 7397,
    #[doc(hidden)]
    Xrm = 7398,
    #[doc(hidden)]
    Xrn = 7399,
    #[doc(hidden)]
    Xrr = 7400,
    #[doc(hidden)]
    Xrt = 7401,
    #[doc(hidden)]
    Xru = 7402,
    #[doc(hidden)]
    Xrw = 7403,
    #[doc(hidden)]
    Xsa = 7404,
    #[doc(hidden)]
    Xsb = 7405,
    #[doc(hidden)]
    Xsc = 7406,
    #[doc(hidden)]
    Xsd = 7407,
    #[doc(hidden)]
    Xse = 7408,
    #[doc(hidden)]
    Xsh = 7409,
    #[doc(hidden)]
    Xsi = 7410,
    #[doc(hidden)]
    Xsj = 7411,
    #[doc(hidden)]
    Xsl = 7412,
    #[doc(hidden)]
    Xsm = 7413,
    #[doc(hidden)]
    Xsn = 7414,
    #[doc(hidden)]
    Xso = 7415,
    #[doc(hidden)]
    Xsp = 7416,
    #[doc(hidden)]
    Xsq = 7417,
    #[doc(hidden)]
    Xsr = 7418,
    #[doc(hidden)]
    Xss = 7419,
    #[doc(hidden)]
    Xsu = 7420,
    #[doc(hidden)]
    Xsv = 7421,
    #[doc(hidden)]
    Xsy = 7422,
    #[doc(hidden)]
    Xta = 7423,
    #[doc(hidden)]
    Xtb = 7424,
    #[doc(hidden)]
    Xtc = 7425,
    #[doc(hidden)]
    Xtd = 7426,
    #[doc(hidden)]
    Xte = 7427,
    #[doc(hidden)]
    Xtg = 7428,
    #[doc(hidden)]
    Xth = 7429,
    #[doc(hidden)]
    Xti = 7430,
    #[doc(hidden)]
    Xtj = 7431,
    #[doc(hidden)]
    Xtl = 7432,
    #[doc(hidden)]
    Xtm = 7433,
    #[doc(hidden)]
    Xtn = 7434,
    #[doc(hidden)]
    Xto = 7435,
    #[doc(hidden)]
    Xtp = 7436,
    #[doc(hidden)]
    Xtq = 7437,
    #[doc(hidden)]
    Xtr = 7438,
    #[doc(hidden)]
    Xts = 7439,
    #[doc(hidden)]
    Xtt = 7440,
    #[doc(hidden)]
    Xtu = 7441,
    #[doc(hidden)]
    Xtv = 7442,
    #[doc(hidden)]
    Xtw = 7443,
    #[doc(hidden)]
    Xty = 7444,
    #[doc(hidden)]
    Xua = 7445,
    #[doc(hidden)]
    Xub = 7446,
    #[doc(hidden)]
    Xud = 7447,
    #[doc(hidden)]
    Xug = 7448,
    #[doc(hidden)]
    Xuj = 7449,
    #[doc(hidden)]
    Xul = 7450,
    #[doc(hidden)]
    Xum = 7451,
    #[doc(hidden)]
    Xun = 7452,
    #[doc(hidden)]
    Xuo = 7453,
    #[doc(hidden)]
    Xup = 7454,
    #[doc(hidden)]
    Xur = 7455,
    #[doc(hidden)]
    Xut = 7456,
    #[doc(hidden)]
    Xuu = 7457,
    #[doc(hidden)]
    Xve = 7458,
    #[doc(hidden)]
    Xvi = 7459,
    #[doc(hidden)]
    Xvn = 7460,
    #[doc(hidden)]
    Xvo = 7461,
    #[doc(hidden)]
    Xvs = 7462,
    #[doc(hidden)]
    Xwa = 7463,
    #[doc(hidden)]
    Xwc = 7464,
    #[doc(hidden)]
    Xwd = 7465,
    #[doc(hidden)]
    Xwe = 7466,
    #[doc(hidden)]
    Xwg = 7467,
    #[doc(hidden)]
    Xwj = 7468,
    #[doc(hidden)]
    Xwk = 7469,
    #[doc(hidden)]
    Xwl = 7470,
    #[doc(hidden)]
    Xwo = 7471,
    #[doc(hidden)]
    Xwr = 7472,
    #[doc(hidden)]
    Xwt = 7473,
    #[doc(hidden)]
    Xww = 7474,
    #[doc(hidden)]
    Xxb = 7475,
    #[doc(hidden)]
    Xxk = 7476,
    #[doc(hidden)]
    Xxm = 7477,
    #[doc(hidden)]
    Xxr = 7478,
    #[doc(hidden)]
    Xxt = 7479,
    #[doc(hidden)]
    Xya = 7480,
    #[doc(hidden)]
    Xyb = 7481,
    #[doc(hidden)]
    Xyj = 7482,
    #[doc(hidden)]
    Xyk = 7483,
    #[doc(hidden)]
    Xyl = 7484,
    #[doc(hidden)]
    Xyt = 7485,
    #[doc(hidden)]
    Xyy = 7486,
    #[doc(hidden)]
    Xzh = 7487,
    #[doc(hidden)]
    Xzm = 7488,
    #[doc(hidden)]
    Xzp = 7489,
    #[doc(hidden)]
    Yaa = 7490,
    #[doc(hidden)]
    Yab = 7491,
    #[doc(hidden)]
    Yac = 7492,
    #[doc(hidden)]
    Yad = 7493,
    #[doc(hidden)]
    Yae = 7494,
    #[doc(hidden)]
    Yaf = 7495,
    #[doc(hidden)]
    Yag = 7496,
    #[doc(hidden)]
    Yah = 7497,
    #[doc(hidden)]
    Yai = 7498,
    #[doc(hidden)]
    Yaj = 7499,
    #[doc(hidden)]
    Yak = 7500,
    #[doc(hidden)]
    Yal = 7501,
    #[doc(hidden)]
    Yam = 7502,
    #[doc(hidden)]
    Yan = 7503,
    #[doc(hidden)]
    Yao = 7504,
    #[doc(hidden)]
    Yap = 7505,
    #[doc(hidden)]
    Yaq = 7506,
    #[doc(hidden)]
    Yar = 7507,
    #[doc(hidden)]
    Yas = 7508,
    #[doc(hidden)]
    Yat = 7509,
    #[doc(hidden)]
    Yau = 7510,
    #[doc(hidden)]
    Yav = 7511,
    #[doc(hidden)]
    Yaw = 7512,
    #[doc(hidden)]
    Yax = 7513,
    #[doc(hidden)]
    Yay = 7514,
    #[doc(hidden)]
    Yaz = 7515,
    #[doc(hidden)]
    Yba = 7516,
    #[doc(hidden)]
    Ybb = 7517,
    #[doc(hidden)]
    Ybe = 7518,
    #[doc(hidden)]
    Ybh = 7519,
    #[doc(hidden)]
    Ybi = 7520,
    #[doc(hidden)]
    Ybj = 7521,
    #[doc(hidden)]
    Ybk = 7522,
    #[doc(hidden)]
    Ybl = 7523,
    #[doc(hidden)]
    Ybm = 7524,
    #[doc(hidden)]
    Ybn = 7525,
    #[doc(hidden)]
    Ybo = 7526,
    #[doc(hidden)]
    Ybx = 7527,
    #[doc(hidden)]
    Yby = 7528,
    #[doc(hidden)]
    Ych = 7529,
    #[doc(hidden)]
    Ycl = 7530,
    #[doc(hidden)]
    Ycn = 7531,
    #[doc(hidden)]
    Ycp = 7532,
    #[doc(hidden)]
    Yda = 7533,
    #[doc(hidden)]
    Ydd = 7534,
    #[doc(hidden)]
    Yde = 7535,
    #[doc(hidden)]
    Ydg = 7536,
    #[doc(hidden)]
    Ydk = 7537,
    #[doc(hidden)]
    Yea = 7538,
    #[doc(hidden)]
    Yec = 7539,
    #[doc(hidden)]
    Yee = 7540,
    #[doc(hidden)]
    Yei = 7541,
    #[doc(hidden)]
    Yej = 7542,
    #[doc(hidden)]
    Yel = 7543,
    #[doc(hidden)]
    Yer = 7544,
    #[doc(hidden)]
    Yes = 7545,
    #[doc(hidden)]
    Yet = 7546,
    #[doc(hidden)]
    Yeu = 7547,
    #[doc(hidden)]
    Yev = 7548,
    #[doc(hidden)]
    Yey = 7549,
    #[doc(hidden)]
    Yga = 7550,
    #[doc(hidden)]
    Ygi = 7551,
    #[doc(hidden)]
    Ygl = 7552,
    #[doc(hidden)]
    Ygm = 7553,
    #[doc(hidden)]
    Ygp = 7554,
    #[doc(hidden)]
    Ygr = 7555,
    #[doc(hidden)]
    Ygs = 7556,
    #[doc(hidden)]
    Ygu = 7557,
    #[doc(hidden)]
    Ygw = 7558,
    #[doc(hidden)]
    Yha = 7559,
    #[doc(hidden)]
    Yhd = 7560,
    #[doc(hidden)]
    Yhl = 7561,
    #[doc(hidden)]
    Yhs = 7562,
    #[doc(hidden)]
    Yia = 7563,
    #[doc(hidden)]
    Yid = 7564,
    #[doc(hidden)]
    Yif = 7565,
    #[doc(hidden)]
    Yig = 7566,
    #[doc(hidden)]
    Yih = 7567,
    #[doc(hidden)]
    Yii = 7568,
    #[doc(hidden)]
    Yij = 7569,
    #[doc(hidden)]
    Yik = 7570,
    #[doc(hidden)]
    Yil = 7571,
    #[doc(hidden)]
    Yim = 7572,
    #[doc(hidden)]
    Yin = 7573,
    #[doc(hidden)]
    Yip = 7574,
    #[doc(hidden)]
    Yiq = 7575,
    #[doc(hidden)]
    Yir = 7576,
    #[doc(hidden)]
    Yis = 7577,
    #[doc(hidden)]
    Yit = 7578,
    #[doc(hidden)]
    Yiu = 7579,
    #[doc(hidden)]
    Yiv = 7580,
    #[doc(hidden)]
    Yix = 7581,
    #[doc(hidden)]
    Yiz = 7582,
    #[doc(hidden)]
    Yka = 7583,
    #[doc(hidden)]
    Ykg = 7584,
    #[doc(hidden)]
    Yki = 7585,
    #[doc(hidden)]
    Ykk = 7586,
    #[doc(hidden)]
    Ykl = 7587,
    #[doc(hidden)]
    Ykm = 7588,
    #[doc(hidden)]
    Ykn = 7589,
    #[doc(hidden)]
    Yko = 7590,
    #[doc(hidden)]
    Ykr = 7591,
    #[doc(hidden)]
    Ykt = 7592,
    #[doc(hidden)]
    Yku = 7593,
    #[doc(hidden)]
    Yky = 7594,
    #[doc(hidden)]
    Yla = 7595,
    #[doc(hidden)]
    Ylb = 7596,
    #[doc(hidden)]
    Yle = 7597,
    #[doc(hidden)]
    Ylg = 7598,
    #[doc(hidden)]
    Yli = 7599,
    #[doc(hidden)]
    Yll = 7600,
    #[doc(hidden)]
    Ylm = 7601,
    #[doc(hidden)]
    Yln = 7602,
    #[doc(hidden)]
    Ylo = 7603,
    #[doc(hidden)]
    Ylr = 7604,
    #[doc(hidden)]
    Ylu = 7605,
    #[doc(hidden)]
    Yly = 7606,
    #[doc(hidden)]
    Ymb = 7607,
    #[doc(hidden)]
    Ymc = 7608,
    #[doc(hidden)]
    Ymd = 7609,
    #[doc(hidden)]
    Yme = 7610,
    #[doc(hidden)]
    Ymg = 7611,
    #[doc(hidden)]
    Ymh = 7612,
    #[doc(hidden)]
    Ymi = 7613,
    #[doc(hidden)]
    Ymk = 7614,
    #[doc(hidden)]
    Yml = 7615,
    #[doc(hidden)]
    Ymm = 7616,
    #[doc(hidden)]
    Ymn = 7617,
    #[doc(hidden)]
    Ymo = 7618,
    #[doc(hidden)]
    Ymp = 7619,
    #[doc(hidden)]
    Ymq = 7620,
    #[doc(hidden)]
    Ymr = 7621,
    #[doc(hidden)]
    Yms = 7622,
    #[doc(hidden)]
    Ymx = 7623,
    #[doc(hidden)]
    Ymz = 7624,
    #[doc(hidden)]
    Yna = 7625,
    #[doc(hidden)]
    Ynd = 7626,
    #[doc(hidden)]
    Yne = 7627,
    #[doc(hidden)]
    Yng = 7628,
    #[doc(hidden)]
    Ynk = 7629,
    #[doc(hidden)]
    Ynl = 7630,
    #[doc(hidden)]
    Ynn = 7631,
    #[doc(hidden)]
    Yno = 7632,
    #[doc(hidden)]
    Ynq = 7633,
    #[doc(hidden)]
    Yns = 7634,
    #[doc(hidden)]
    Ynu = 7635,
    #[doc(hidden)]
    Yob = 7636,
    #[doc(hidden)]
    Yog = 7637,
    #[doc(hidden)]
    Yoi = 7638,
    #[doc(hidden)]
    Yok = 7639,
    #[doc(hidden)]
    Yol = 7640,
    #[doc(hidden)]
    Yom = 7641,
    #[doc(hidden)]
    Yon = 7642,
    #[doc(hidden)]
    Yor = 7643,
    #[doc(hidden)]
    Yot = 7644,
    #[doc(hidden)]
    Yox = 7645,
    #[doc(hidden)]
    Yoy = 7646,
    #[doc(hidden)]
    Ypa = 7647,
    #[doc(hidden)]
    Ypb = 7648,
    #[doc(hidden)]
    Ypg = 7649,
    #[doc(hidden)]
    Yph = 7650,
    #[doc(hidden)]
    Ypm = 7651,
    #[doc(hidden)]
    Ypn = 7652,
    #[doc(hidden)]
    Ypo = 7653,
    #[doc(hidden)]
    Ypp = 7654,
    #[doc(hidden)]
    Ypz = 7655,
    #[doc(hidden)]
    Yra = 7656,
    #[doc(hidden)]
    Yrb = 7657,
    #[doc(hidden)]
    Yre = 7658,
    #[doc(hidden)]
    Yrk = 7659,
    #[doc(hidden)]
    Yrl = 7660,
    #[doc(hidden)]
    Yrm = 7661,
    #[doc(hidden)]
    Yrn = 7662,
    #[doc(hidden)]
    Yro = 7663,
    #[doc(hidden)]
    Yrs = 7664,
    #[doc(hidden)]
    Yrw = 7665,
    #[doc(hidden)]
    Yry = 7666,
    #[doc(hidden)]
    Ysc = 7667,
    #[doc(hidden)]
    Ysd = 7668,
    #[doc(hidden)]
    Ysg = 7669,
    #[doc(hidden)]
    Ysl = 7670,
    #[doc(hidden)]
    Ysm = 7671,
    #[doc(hidden)]
    Ysn = 7672,
    #[doc(hidden)]
    Yso = 7673,
    #[doc(hidden)]
    Ysp = 7674,
    #[doc(hidden)]
    Ysr = 7675,
    #[doc(hidden)]
    Yss = 7676,
    #[doc(hidden)]
    Ysy = 7677,
    #[doc(hidden)]
    Yta = 7678,
    #[doc(hidden)]
    Ytl = 7679,
    #[doc(hidden)]
    Ytp = 7680,
    #[doc(hidden)]
    Ytw = 7681,
    #[doc(hidden)]
    Yty = 7682,
    #[doc(hidden)]
    Yua = 7683,
    #[doc(hidden)]
    Yub = 7684,
    #[doc(hidden)]
    Yuc = 7685,
    #[doc(hidden)]
    Yud = 7686,
    #[doc(hidden)]
    Yue = 7687,
    #[doc(hidden)]
    Yuf = 7688,
    #[doc(hidden)]
    Yug = 7689,
    #[doc(hidden)]
    Yui = 7690,
    #[doc(hidden)]
    Yuj = 7691,
    #[doc(hidden)]
    Yuk = 7692,
    #[doc(hidden)]
    Yul = 7693,
    #[doc(hidden)]
    Yum = 7694,
    #[doc(hidden)]
    Yun = 7695,
    #[doc(hidden)]
    Yup = 7696,
    #[doc(hidden)]
    Yuq = 7697,
    #[doc(hidden)]
    Yur = 7698,
    #[doc(hidden)]
    Yut = 7699,
    #[doc(hidden)]
    Yuw = 7700,
    #[doc(hidden)]
    Yux = 7701,
    #[doc(hidden)]
    Yuy = 7702,
    #[doc(hidden)]
    Yuz = 7703,
    #[doc(hidden)]
    Yva = 7704,
    #[doc(hidden)]
    Yvt = 7705,
    #[doc(hidden)]
    Ywa = 7706,
    #[doc(hidden)]
    Ywg = 7707,
    #[doc(hidden)]
    Ywl = 7708,
    #[doc(hidden)]
    Ywn = 7709,
    #[doc(hidden)]
    Ywq = 7710,
    #[doc(hidden)]
    Ywr = 7711,
    #[doc(hidden)]
    Ywt = 7712,
    #[doc(hidden)]
    Ywu = 7713,
    #[doc(hidden)]
    Yww = 7714,
    #[doc(hidden)]
    Yxa = 7715,
    #[doc(hidden)]
    Yxg = 7716,
    #[doc(hidden)]
    Yxl = 7717,
    #[doc(hidden)]
    Yxm = 7718,
    #[doc(hidden)]
    Yxu = 7719,
    #[doc(hidden)]
    Yxy = 7720,
    #[doc(hidden)]
    Yyr = 7721,
    #[doc(hidden)]
    Yyu = 7722,
    #[doc(hidden)]
    Yyz = 7723,
    #[doc(hidden)]
    Yzg = 7724,
    #[doc(hidden)]
    Yzk = 7725,
    #[doc(hidden)]
    Zaa = 7726,
    #[doc(hidden)]
    Zab = 7727,
    #[doc(hidden)]
    Zac = 7728,
    #[doc(hidden)]
    Zad = 7729,
    #[doc(hidden)]
    Zae = 7730,
    #[doc(hidden)]
    Zaf = 7731,
    #[doc(hidden)]
    Zag = 7732,
    #[doc(hidden)]
    Zah = 7733,
    #[doc(hidden)]
    Zai = 7734,
    #[doc(hidden)]
    Zaj = 7735,
    #[doc(hidden)]
    Zak = 7736,
    #[doc(hidden)]
    Zal = 7737,
    #[doc(hidden)]
    Zam = 7738,
    #[doc(hidden)]
    Zao = 7739,
    #[doc(hidden)]
    Zap = 7740,
    #[doc(hidden)]
    Zaq = 7741,
    #[doc(hidden)]
    Zar = 7742,
    #[doc(hidden)]
    Zas = 7743,
    #[doc(hidden)]
    Zat = 7744,
    #[doc(hidden)]
    Zau = 7745,
    #[doc(hidden)]
    Zav = 7746,
    #[doc(hidden)]
    Zaw = 7747,
    #[doc(hidden)]
    Zax = 7748,
    #[doc(hidden)]
    Zay = 7749,
    #[doc(hidden)]
    Zaz = 7750,
    #[doc(hidden)]
    Zba = 7751,
    #[doc(hidden)]
    Zbc = 7752,
    #[doc(hidden)]
    Zbe = 7753,
    #[doc(hidden)]
    Zbl = 7754,
    #[doc(hidden)]
    Zbt = 7755,
    #[doc(hidden)]
    Zbu = 7756,
    #[doc(hidden)]
    Zbw = 7757,
    #[doc(hidden)]
    Zca = 7758,
    #[doc(hidden)]
    Zcd = 7759,
    #[doc(hidden)]
    Zch = 7760,
    #[doc(hidden)]
    Zdj = 7761,
    #[doc(hidden)]
    Zea = 7762,
    #[doc(hidden)]
    Zeg = 7763,
    #[doc(hidden)]
    Zeh = 7764,
    #[doc(hidden)]
    Zen = 7765,
    #[doc(hidden)]
    Zga = 7766,
    #[doc(hidden)]
    Zgb = 7767,
    #[doc(hidden)]
    Zgh = 7768,
    #[doc(hidden)]
    Zgm = 7769,
    #[doc(hidden)]
    Zgn = 7770,
    #[doc(hidden)]
    Zgr = 7771,
    #[doc(hidden)]
    Zha = 7772,
    #[doc(hidden)]
    Zhb = 7773,
    #[doc(hidden)]
    Zhd = 7774,
    #[doc(hidden)]
    Zhi = 7775,
    #[doc(hidden)]
    Zhn = 7776,
    #[doc(hidden)]
    Zho = 7777,
    #[doc(hidden)]
    Zhw = 7778,
    #[doc(hidden)]
    Zia = 7779,
    #[doc(hidden)]
    Zib = 7780,
    #[doc(hidden)]
    Zik = 7781,
    #[doc(hidden)]
    Zil = 7782,
    #[doc(hidden)]
    Zim = 7783,
    #[doc(hidden)]
    Zin = 7784,
    #[doc(hidden)]
    Ziw = 7785,
    #[doc(hidden)]
    Ziz = 7786,
    #[doc(hidden)]
    Zka = 7787,
    #[doc(hidden)]
    Zkb = 7788,
    #[doc(hidden)]
    Zkd = 7789,
    #[doc(hidden)]
    Zkg = 7790,
    #[doc(hidden)]
    Zkh = 7791,
    #[doc(hidden)]
    Zkk = 7792,
    #[doc(hidden)]
    Zkn = 7793,
    #[doc(hidden)]
    Zko = 7794,
    #[doc(hidden)]
    Zkp = 7795,
    #[doc(hidden)]
    Zkr = 7796,
    #[doc(hidden)]
    Zkt = 7797,
    #[doc(hidden)]
    Zku = 7798,
    #[doc(hidden)]
    Zkv = 7799,
    #[doc(hidden)]
    Zkz = 7800,
    #[doc(hidden)]
    Zla = 7801,
    #[doc(hidden)]
    Zlj = 7802,
    #[doc(hidden)]
    Zlm = 7803,
    #[doc(hidden)]
    Zln = 7804,
    #[doc(hidden)]
    Zlq = 7805,
    #[doc(hidden)]
    Zma = 7806,
    #[doc(hidden)]
    Zmb = 7807,
    #[doc(hidden)]
    Zmc = 7808,
    #[doc(hidden)]
    Zmd = 7809,
    #[doc(hidden)]
    Zme = 7810,
    #[doc(hidden)]
    Zmf = 7811,
    #[doc(hidden)]
    Zmg = 7812,
    #[doc(hidden)]
    Zmh = 7813,
    #[doc(hidden)]
    Zmi = 7814,
    #[doc(hidden)]
    Zmj = 7815,
    #[doc(hidden)]
    Zmk = 7816,
    #[doc(hidden)]
    Zml = 7817,
    #[doc(hidden)]
    Zmm = 7818,
    #[doc(hidden)]
    Zmn = 7819,
    #[doc(hidden)]
    Zmo = 7820,
    #[doc(hidden)]
    Zmp = 7821,
    #[doc(hidden)]
    Zmq = 7822,
    #[doc(hidden)]
    Zmr = 7823,
    #[doc(hidden)]
    Zms = 7824,
    #[doc(hidden)]
    Zmt = 7825,
    #[doc(hidden)]
    Zmu = 7826,
    #[doc(hidden)]
    Zmv = 7827,
    #[doc(hidden)]
    Zmw = 7828,
    #[doc(hidden)]
    Zmx = 7829,
    #[doc(hidden)]
    Zmy = 7830,
    #[doc(hidden)]
    Zmz = 7831,
    #[doc(hidden)]
    Zna = 7832,
    #[doc(hidden)]
    Zne = 7833,
    #[doc(hidden)]
    Zng = 7834,
    #[doc(hidden)]
    Znk = 7835,
    #[doc(hidden)]
    Zns = 7836,
    #[doc(hidden)]
    Zoc = 7837,
    #[doc(hidden)]
    Zoh = 7838,
    #[doc(hidden)]
    Zom = 7839,
    #[doc(hidden)]
    Zoo = 7840,
    #[doc(hidden)]
    Zoq = 7841,
    #[doc(hidden)]
    Zor = 7842,
    #[doc(hidden)]
    Zos = 7843,
    #[doc(hidden)]
    Zpa = 7844,
    #[doc(hidden)]
    Zpb = 7845,
    #[doc(hidden)]
    Zpc = 7846,
    #[doc(hidden)]
    Zpd = 7847,
    #[doc(hidden)]
    Zpe = 7848,
    #[doc(hidden)]
    Zpf = 7849,
    #[doc(hidden)]
    Zpg = 7850,
    #[doc(hidden)]
    Zph = 7851,
    #[doc(hidden)]
    Zpi = 7852,
    #[doc(hidden)]
    Zpj = 7853,
    #[doc(hidden)]
    Zpk = 7854,
    #[doc(hidden)]
    Zpl = 7855,
    #[doc(hidden)]
    Zpm = 7856,
    #[doc(hidden)]
    Zpn = 7857,
    #[doc(hidden)]
    Zpo = 7858,
    #[doc(hidden)]
    Zpp = 7859,
    #[doc(hidden)]
    Zpq = 7860,
    #[doc(hidden)]
    Zpr = 7861,
    #[doc(hidden)]
    Zps = 7862,
    #[doc(hidden)]
    Zpt = 7863,
    #[doc(hidden)]
    Zpu = 7864,
    #[doc(hidden)]
    Zpv = 7865,
    #[doc(hidden)]
    Zpw = 7866,
    #[doc(hidden)]
    Zpx = 7867,
    #[doc(hidden)]
    Zpy = 7868,
    #[doc(hidden)]
    Zpz = 7869,
    #[doc(hidden)]
    Zqe = 7870,
    #[doc(hidden)]
    Zra = 7871,
    #[doc(hidden)]
    Zrg = 7872,
    #[doc(hidden)]
    Zrn = 7873,
    #[doc(hidden)]
    Zro = 7874,
    #[doc(hidden)]
    Zrp = 7875,
    #[doc(hidden)]
    Zrs = 7876,
    #[doc(hidden)]
    Zsa = 7877,
    #[doc(hidden)]
    Zsk = 7878,
    #[doc(hidden)]
    Zsl = 7879,
    #[doc(hidden)]
    Zsm = 7880,
    #[doc(hidden)]
    Zsr = 7881,
    #[doc(hidden)]
    Zsu = 7882,
    #[doc(hidden)]
    Zte = 7883,
    #[doc(hidden)]
    Ztg = 7884,
    #[doc(hidden)]
    Ztl = 7885,
    #[doc(hidden)]
    Ztm = 7886,
    #[doc(hidden)]
    Ztn = 7887,
    #[doc(hidden)]
    Ztp = 7888,
    #[doc(hidden)]
    Ztq = 7889,
    #[doc(hidden)]
    Zts = 7890,
    #[doc(hidden)]
    Ztt = 7891,
    #[doc(hidden)]
    Ztu = 7892,
    #[doc(hidden)]
    Ztx = 7893,
    #[doc(hidden)]
    Zty = 7894,
    #[doc(hidden)]
    Zua = 7895,
    #[doc(hidden)]
    Zuh = 7896,
    #[doc(hidden)]
    Zul = 7897,
    #[doc(hidden)]
    Zum = 7898,
    #[doc(hidden)]
    Zun = 7899,
    #[doc(hidden)]
    Zuy = 7900,
    #[doc(hidden)]
    Zwa = 7901,
    #[doc(hidden)]
    Zxx = 7902,
    #[doc(hidden)]
    Zyb = 7903,
    #[doc(hidden)]
    Zyg = 7904,
    #[doc(hidden)]
    Zyj = 7905,
    #[doc(hidden)]
    Zyn = 7906,
    #[doc(hidden)]
    Zyp = 7907,
    #[doc(hidden)]
    Zza = 7908,
    #[doc(hidden)]
    Zzj = 7909,
}

impl Language {
    pub fn from_usize(u: usize) -> Option<Self> {
        match u {
            0 => Some(Language::Aaa),
            1 => Some(Language::Aab),
            2 => Some(Language::Aac),
            3 => Some(Language::Aad),
            4 => Some(Language::Aae),
            5 => Some(Language::Aaf),
            6 => Some(Language::Aag),
            7 => Some(Language::Aah),
            8 => Some(Language::Aai),
            9 => Some(Language::Aak),
            10 => Some(Language::Aal),
            11 => Some(Language::Aan),
            12 => Some(Language::Aao),
            13 => Some(Language::Aap),
            14 => Some(Language::Aaq),
            15 => Some(Language::Aar),
            16 => Some(Language::Aas),
            17 => Some(Language::Aat),
            18 => Some(Language::Aau),
            19 => Some(Language::Aaw),
            20 => Some(Language::Aax),
            21 => Some(Language::Aaz),
            22 => Some(Language::Aba),
            23 => Some(Language::Abb),
            24 => Some(Language::Abc),
            25 => Some(Language::Abd),
            26 => Some(Language::Abe),
            27 => Some(Language::Abf),
            28 => Some(Language::Abg),
            29 => Some(Language::Abh),
            30 => Some(Language::Abi),
            31 => Some(Language::Abj),
            32 => Some(Language::Abk),
            33 => Some(Language::Abl),
            34 => Some(Language::Abm),
            35 => Some(Language::Abn),
            36 => Some(Language::Abo),
            37 => Some(Language::Abp),
            38 => Some(Language::Abq),
            39 => Some(Language::Abr),
            40 => Some(Language::Abs),
            41 => Some(Language::Abt),
            42 => Some(Language::Abu),
            43 => Some(Language::Abv),
            44 => Some(Language::Abw),
            45 => Some(Language::Abx),
            46 => Some(Language::Aby),
            47 => Some(Language::Abz),
            48 => Some(Language::Aca),
            49 => Some(Language::Acb),
            50 => Some(Language::Acd),
            51 => Some(Language::Ace),
            52 => Some(Language::Acf),
            53 => Some(Language::Ach),
            54 => Some(Language::Aci),
            55 => Some(Language::Ack),
            56 => Some(Language::Acl),
            57 => Some(Language::Acm),
            58 => Some(Language::Acn),
            59 => Some(Language::Acp),
            60 => Some(Language::Acq),
            61 => Some(Language::Acr),
            62 => Some(Language::Acs),
            63 => Some(Language::Act),
            64 => Some(Language::Acu),
            65 => Some(Language::Acv),
            66 => Some(Language::Acw),
            67 => Some(Language::Acx),
            68 => Some(Language::Acy),
            69 => Some(Language::Acz),
            70 => Some(Language::Ada),
            71 => Some(Language::Adb),
            72 => Some(Language::Add),
            73 => Some(Language::Ade),
            74 => Some(Language::Adf),
            75 => Some(Language::Adg),
            76 => Some(Language::Adh),
            77 => Some(Language::Adi),
            78 => Some(Language::Adj),
            79 => Some(Language::Adl),
            80 => Some(Language::Adn),
            81 => Some(Language::Ado),
            82 => Some(Language::Adq),
            83 => Some(Language::Adr),
            84 => Some(Language::Ads),
            85 => Some(Language::Adt),
            86 => Some(Language::Adu),
            87 => Some(Language::Adw),
            88 => Some(Language::Adx),
            89 => Some(Language::Ady),
            90 => Some(Language::Adz),
            91 => Some(Language::Aea),
            92 => Some(Language::Aeb),
            93 => Some(Language::Aec),
            94 => Some(Language::Aed),
            95 => Some(Language::Aee),
            96 => Some(Language::Aek),
            97 => Some(Language::Ael),
            98 => Some(Language::Aem),
            99 => Some(Language::Aen),
            100 => Some(Language::Aeq),
            101 => Some(Language::Aer),
            102 => Some(Language::Aes),
            103 => Some(Language::Aeu),
            104 => Some(Language::Aew),
            105 => Some(Language::Aey),
            106 => Some(Language::Aez),
            107 => Some(Language::Afb),
            108 => Some(Language::Afd),
            109 => Some(Language::Afe),
            110 => Some(Language::Afg),
            111 => Some(Language::Afh),
            112 => Some(Language::Afi),
            113 => Some(Language::Afk),
            114 => Some(Language::Afn),
            115 => Some(Language::Afo),
            116 => Some(Language::Afp),
            117 => Some(Language::Afr),
            118 => Some(Language::Afs),
            119 => Some(Language::Aft),
            120 => Some(Language::Afu),
            121 => Some(Language::Afz),
            122 => Some(Language::Aga),
            123 => Some(Language::Agb),
            124 => Some(Language::Agc),
            125 => Some(Language::Agd),
            126 => Some(Language::Age),
            127 => Some(Language::Agf),
            128 => Some(Language::Agg),
            129 => Some(Language::Agh),
            130 => Some(Language::Agi),
            131 => Some(Language::Agj),
            132 => Some(Language::Agk),
            133 => Some(Language::Agl),
            134 => Some(Language::Agm),
            135 => Some(Language::Agn),
            136 => Some(Language::Ago),
            137 => Some(Language::Agq),
            138 => Some(Language::Agr),
            139 => Some(Language::Ags),
            140 => Some(Language::Agt),
            141 => Some(Language::Agu),
            142 => Some(Language::Agv),
            143 => Some(Language::Agw),
            144 => Some(Language::Agx),
            145 => Some(Language::Agy),
            146 => Some(Language::Agz),
            147 => Some(Language::Aha),
            148 => Some(Language::Ahb),
            149 => Some(Language::Ahg),
            150 => Some(Language::Ahh),
            151 => Some(Language::Ahi),
            152 => Some(Language::Ahk),
            153 => Some(Language::Ahl),
            154 => Some(Language::Ahm),
            155 => Some(Language::Ahn),
            156 => Some(Language::Aho),
            157 => Some(Language::Ahp),
            158 => Some(Language::Ahr),
            159 => Some(Language::Ahs),
            160 => Some(Language::Aht),
            161 => Some(Language::Aia),
            162 => Some(Language::Aib),
            163 => Some(Language::Aic),
            164 => Some(Language::Aid),
            165 => Some(Language::Aie),
            166 => Some(Language::Aif),
            167 => Some(Language::Aig),
            168 => Some(Language::Aih),
            169 => Some(Language::Aii),
            170 => Some(Language::Aij),
            171 => Some(Language::Aik),
            172 => Some(Language::Ail),
            173 => Some(Language::Aim),
            174 => Some(Language::Ain),
            175 => Some(Language::Aio),
            176 => Some(Language::Aip),
            177 => Some(Language::Aiq),
            178 => Some(Language::Air),
            179 => Some(Language::Ait),
            180 => Some(Language::Aiw),
            181 => Some(Language::Aix),
            182 => Some(Language::Aiy),
            183 => Some(Language::Aja),
            184 => Some(Language::Ajg),
            185 => Some(Language::Aji),
            186 => Some(Language::Ajn),
            187 => Some(Language::Ajp),
            188 => Some(Language::Ajs),
            189 => Some(Language::Aju),
            190 => Some(Language::Ajw),
            191 => Some(Language::Ajz),
            192 => Some(Language::Aka),
            193 => Some(Language::Akb),
            194 => Some(Language::Akc),
            195 => Some(Language::Akd),
            196 => Some(Language::Ake),
            197 => Some(Language::Akf),
            198 => Some(Language::Akg),
            199 => Some(Language::Akh),
            200 => Some(Language::Aki),
            201 => Some(Language::Akj),
            202 => Some(Language::Akk),
            203 => Some(Language::Akl),
            204 => Some(Language::Akm),
            205 => Some(Language::Ako),
            206 => Some(Language::Akp),
            207 => Some(Language::Akq),
            208 => Some(Language::Akr),
            209 => Some(Language::Aks),
            210 => Some(Language::Akt),
            211 => Some(Language::Aku),
            212 => Some(Language::Akv),
            213 => Some(Language::Akw),
            214 => Some(Language::Akx),
            215 => Some(Language::Aky),
            216 => Some(Language::Akz),
            217 => Some(Language::Ala),
            218 => Some(Language::Alc),
            219 => Some(Language::Ald),
            220 => Some(Language::Ale),
            221 => Some(Language::Alf),
            222 => Some(Language::Alh),
            223 => Some(Language::Ali),
            224 => Some(Language::Alj),
            225 => Some(Language::Alk),
            226 => Some(Language::All),
            227 => Some(Language::Alm),
            228 => Some(Language::Aln),
            229 => Some(Language::Alo),
            230 => Some(Language::Alp),
            231 => Some(Language::Alq),
            232 => Some(Language::Alr),
            233 => Some(Language::Als),
            234 => Some(Language::Alt),
            235 => Some(Language::Alu),
            236 => Some(Language::Alw),
            237 => Some(Language::Alx),
            238 => Some(Language::Aly),
            239 => Some(Language::Alz),
            240 => Some(Language::Ama),
            241 => Some(Language::Amb),
            242 => Some(Language::Amc),
            243 => Some(Language::Ame),
            244 => Some(Language::Amf),
            245 => Some(Language::Amg),
            246 => Some(Language::Amh),
            247 => Some(Language::Ami),
            248 => Some(Language::Amj),
            249 => Some(Language::Amk),
            250 => Some(Language::Aml),
            251 => Some(Language::Amm),
            252 => Some(Language::Amn),
            253 => Some(Language::Amo),
            254 => Some(Language::Amp),
            255 => Some(Language::Amq),
            256 => Some(Language::Amr),
            257 => Some(Language::Ams),
            258 => Some(Language::Amt),
            259 => Some(Language::Amu),
            260 => Some(Language::Amv),
            261 => Some(Language::Amw),
            262 => Some(Language::Amx),
            263 => Some(Language::Amy),
            264 => Some(Language::Amz),
            265 => Some(Language::Ana),
            266 => Some(Language::Anb),
            267 => Some(Language::Anc),
            268 => Some(Language::And),
            269 => Some(Language::Ane),
            270 => Some(Language::Anf),
            271 => Some(Language::Ang),
            272 => Some(Language::Anh),
            273 => Some(Language::Ani),
            274 => Some(Language::Anj),
            275 => Some(Language::Ank),
            276 => Some(Language::Anl),
            277 => Some(Language::Anm),
            278 => Some(Language::Ann),
            279 => Some(Language::Ano),
            280 => Some(Language::Anp),
            281 => Some(Language::Anq),
            282 => Some(Language::Anr),
            283 => Some(Language::Ans),
            284 => Some(Language::Ant),
            285 => Some(Language::Anu),
            286 => Some(Language::Anv),
            287 => Some(Language::Anw),
            288 => Some(Language::Anx),
            289 => Some(Language::Any),
            290 => Some(Language::Anz),
            291 => Some(Language::Aoa),
            292 => Some(Language::Aob),
            293 => Some(Language::Aoc),
            294 => Some(Language::Aod),
            295 => Some(Language::Aoe),
            296 => Some(Language::Aof),
            297 => Some(Language::Aog),
            298 => Some(Language::Aoi),
            299 => Some(Language::Aoj),
            300 => Some(Language::Aok),
            301 => Some(Language::Aol),
            302 => Some(Language::Aom),
            303 => Some(Language::Aon),
            304 => Some(Language::Aor),
            305 => Some(Language::Aos),
            306 => Some(Language::Aot),
            307 => Some(Language::Aou),
            308 => Some(Language::Aox),
            309 => Some(Language::Aoz),
            310 => Some(Language::Apb),
            311 => Some(Language::Apc),
            312 => Some(Language::Apd),
            313 => Some(Language::Ape),
            314 => Some(Language::Apf),
            315 => Some(Language::Apg),
            316 => Some(Language::Aph),
            317 => Some(Language::Api),
            318 => Some(Language::Apj),
            319 => Some(Language::Apk),
            320 => Some(Language::Apl),
            321 => Some(Language::Apm),
            322 => Some(Language::Apn),
            323 => Some(Language::Apo),
            324 => Some(Language::App),
            325 => Some(Language::Apq),
            326 => Some(Language::Apr),
            327 => Some(Language::Aps),
            328 => Some(Language::Apt),
            329 => Some(Language::Apu),
            330 => Some(Language::Apv),
            331 => Some(Language::Apw),
            332 => Some(Language::Apx),
            333 => Some(Language::Apy),
            334 => Some(Language::Apz),
            335 => Some(Language::Aqc),
            336 => Some(Language::Aqd),
            337 => Some(Language::Aqg),
            338 => Some(Language::Aqk),
            339 => Some(Language::Aqm),
            340 => Some(Language::Aqn),
            341 => Some(Language::Aqp),
            342 => Some(Language::Aqr),
            343 => Some(Language::Aqt),
            344 => Some(Language::Aqz),
            345 => Some(Language::Ara),
            346 => Some(Language::Arb),
            347 => Some(Language::Arc),
            348 => Some(Language::Ard),
            349 => Some(Language::Are),
            350 => Some(Language::Arg),
            351 => Some(Language::Arh),
            352 => Some(Language::Ari),
            353 => Some(Language::Arj),
            354 => Some(Language::Ark),
            355 => Some(Language::Arl),
            356 => Some(Language::Arn),
            357 => Some(Language::Aro),
            358 => Some(Language::Arp),
            359 => Some(Language::Arq),
            360 => Some(Language::Arr),
            361 => Some(Language::Ars),
            362 => Some(Language::Aru),
            363 => Some(Language::Arv),
            364 => Some(Language::Arw),
            365 => Some(Language::Arx),
            366 => Some(Language::Ary),
            367 => Some(Language::Arz),
            368 => Some(Language::Asa),
            369 => Some(Language::Asb),
            370 => Some(Language::Asc),
            371 => Some(Language::Ase),
            372 => Some(Language::Asf),
            373 => Some(Language::Asg),
            374 => Some(Language::Ash),
            375 => Some(Language::Asi),
            376 => Some(Language::Asj),
            377 => Some(Language::Ask),
            378 => Some(Language::Asl),
            379 => Some(Language::Asm),
            380 => Some(Language::Asn),
            381 => Some(Language::Aso),
            382 => Some(Language::Asp),
            383 => Some(Language::Asq),
            384 => Some(Language::Asr),
            385 => Some(Language::Ass),
            386 => Some(Language::Ast),
            387 => Some(Language::Asu),
            388 => Some(Language::Asv),
            389 => Some(Language::Asw),
            390 => Some(Language::Asx),
            391 => Some(Language::Asy),
            392 => Some(Language::Asz),
            393 => Some(Language::Ata),
            394 => Some(Language::Atb),
            395 => Some(Language::Atc),
            396 => Some(Language::Atd),
            397 => Some(Language::Ate),
            398 => Some(Language::Atg),
            399 => Some(Language::Ati),
            400 => Some(Language::Atj),
            401 => Some(Language::Atk),
            402 => Some(Language::Atl),
            403 => Some(Language::Atm),
            404 => Some(Language::Atn),
            405 => Some(Language::Ato),
            406 => Some(Language::Atp),
            407 => Some(Language::Atq),
            408 => Some(Language::Atr),
            409 => Some(Language::Ats),
            410 => Some(Language::Att),
            411 => Some(Language::Atu),
            412 => Some(Language::Atv),
            413 => Some(Language::Atw),
            414 => Some(Language::Atx),
            415 => Some(Language::Aty),
            416 => Some(Language::Atz),
            417 => Some(Language::Aua),
            418 => Some(Language::Aub),
            419 => Some(Language::Auc),
            420 => Some(Language::Aud),
            421 => Some(Language::Aug),
            422 => Some(Language::Auh),
            423 => Some(Language::Aui),
            424 => Some(Language::Auj),
            425 => Some(Language::Auk),
            426 => Some(Language::Aul),
            427 => Some(Language::Aum),
            428 => Some(Language::Aun),
            429 => Some(Language::Auo),
            430 => Some(Language::Aup),
            431 => Some(Language::Auq),
            432 => Some(Language::Aur),
            433 => Some(Language::Aut),
            434 => Some(Language::Auu),
            435 => Some(Language::Auw),
            436 => Some(Language::Aux),
            437 => Some(Language::Auy),
            438 => Some(Language::Auz),
            439 => Some(Language::Ava),
            440 => Some(Language::Avb),
            441 => Some(Language::Avd),
            442 => Some(Language::Ave),
            443 => Some(Language::Avi),
            444 => Some(Language::Avk),
            445 => Some(Language::Avl),
            446 => Some(Language::Avm),
            447 => Some(Language::Avn),
            448 => Some(Language::Avo),
            449 => Some(Language::Avs),
            450 => Some(Language::Avt),
            451 => Some(Language::Avu),
            452 => Some(Language::Avv),
            453 => Some(Language::Awa),
            454 => Some(Language::Awb),
            455 => Some(Language::Awc),
            456 => Some(Language::Awe),
            457 => Some(Language::Awg),
            458 => Some(Language::Awh),
            459 => Some(Language::Awi),
            460 => Some(Language::Awk),
            461 => Some(Language::Awm),
            462 => Some(Language::Awn),
            463 => Some(Language::Awo),
            464 => Some(Language::Awr),
            465 => Some(Language::Aws),
            466 => Some(Language::Awt),
            467 => Some(Language::Awu),
            468 => Some(Language::Awv),
            469 => Some(Language::Aww),
            470 => Some(Language::Awx),
            471 => Some(Language::Awy),
            472 => Some(Language::Axb),
            473 => Some(Language::Axe),
            474 => Some(Language::Axg),
            475 => Some(Language::Axk),
            476 => Some(Language::Axl),
            477 => Some(Language::Axm),
            478 => Some(Language::Axx),
            479 => Some(Language::Aya),
            480 => Some(Language::Ayb),
            481 => Some(Language::Ayc),
            482 => Some(Language::Ayd),
            483 => Some(Language::Aye),
            484 => Some(Language::Ayg),
            485 => Some(Language::Ayh),
            486 => Some(Language::Ayi),
            487 => Some(Language::Ayk),
            488 => Some(Language::Ayl),
            489 => Some(Language::Aym),
            490 => Some(Language::Ayn),
            491 => Some(Language::Ayo),
            492 => Some(Language::Ayp),
            493 => Some(Language::Ayq),
            494 => Some(Language::Ayr),
            495 => Some(Language::Ays),
            496 => Some(Language::Ayt),
            497 => Some(Language::Ayu),
            498 => Some(Language::Ayz),
            499 => Some(Language::Aza),
            500 => Some(Language::Azb),
            501 => Some(Language::Azd),
            502 => Some(Language::Aze),
            503 => Some(Language::Azg),
            504 => Some(Language::Azj),
            505 => Some(Language::Azm),
            506 => Some(Language::Azn),
            507 => Some(Language::Azo),
            508 => Some(Language::Azt),
            509 => Some(Language::Azz),
            510 => Some(Language::Baa),
            511 => Some(Language::Bab),
            512 => Some(Language::Bac),
            513 => Some(Language::Bae),
            514 => Some(Language::Baf),
            515 => Some(Language::Bag),
            516 => Some(Language::Bah),
            517 => Some(Language::Baj),
            518 => Some(Language::Bak),
            519 => Some(Language::Bal),
            520 => Some(Language::Bam),
            521 => Some(Language::Ban),
            522 => Some(Language::Bao),
            523 => Some(Language::Bap),
            524 => Some(Language::Bar),
            525 => Some(Language::Bas),
            526 => Some(Language::Bau),
            527 => Some(Language::Bav),
            528 => Some(Language::Baw),
            529 => Some(Language::Bax),
            530 => Some(Language::Bay),
            531 => Some(Language::Bba),
            532 => Some(Language::Bbb),
            533 => Some(Language::Bbc),
            534 => Some(Language::Bbd),
            535 => Some(Language::Bbe),
            536 => Some(Language::Bbf),
            537 => Some(Language::Bbg),
            538 => Some(Language::Bbh),
            539 => Some(Language::Bbi),
            540 => Some(Language::Bbj),
            541 => Some(Language::Bbk),
            542 => Some(Language::Bbl),
            543 => Some(Language::Bbm),
            544 => Some(Language::Bbn),
            545 => Some(Language::Bbo),
            546 => Some(Language::Bbp),
            547 => Some(Language::Bbq),
            548 => Some(Language::Bbr),
            549 => Some(Language::Bbs),
            550 => Some(Language::Bbt),
            551 => Some(Language::Bbu),
            552 => Some(Language::Bbv),
            553 => Some(Language::Bbw),
            554 => Some(Language::Bbx),
            555 => Some(Language::Bby),
            556 => Some(Language::Bca),
            557 => Some(Language::Bcb),
            558 => Some(Language::Bcc),
            559 => Some(Language::Bcd),
            560 => Some(Language::Bce),
            561 => Some(Language::Bcf),
            562 => Some(Language::Bcg),
            563 => Some(Language::Bch),
            564 => Some(Language::Bci),
            565 => Some(Language::Bcj),
            566 => Some(Language::Bck),
            567 => Some(Language::Bcl),
            568 => Some(Language::Bcm),
            569 => Some(Language::Bcn),
            570 => Some(Language::Bco),
            571 => Some(Language::Bcp),
            572 => Some(Language::Bcq),
            573 => Some(Language::Bcr),
            574 => Some(Language::Bcs),
            575 => Some(Language::Bct),
            576 => Some(Language::Bcu),
            577 => Some(Language::Bcv),
            578 => Some(Language::Bcw),
            579 => Some(Language::Bcy),
            580 => Some(Language::Bcz),
            581 => Some(Language::Bda),
            582 => Some(Language::Bdb),
            583 => Some(Language::Bdc),
            584 => Some(Language::Bdd),
            585 => Some(Language::Bde),
            586 => Some(Language::Bdf),
            587 => Some(Language::Bdg),
            588 => Some(Language::Bdh),
            589 => Some(Language::Bdi),
            590 => Some(Language::Bdj),
            591 => Some(Language::Bdk),
            592 => Some(Language::Bdl),
            593 => Some(Language::Bdm),
            594 => Some(Language::Bdn),
            595 => Some(Language::Bdo),
            596 => Some(Language::Bdp),
            597 => Some(Language::Bdq),
            598 => Some(Language::Bdr),
            599 => Some(Language::Bds),
            600 => Some(Language::Bdt),
            601 => Some(Language::Bdu),
            602 => Some(Language::Bdv),
            603 => Some(Language::Bdw),
            604 => Some(Language::Bdx),
            605 => Some(Language::Bdy),
            606 => Some(Language::Bdz),
            607 => Some(Language::Bea),
            608 => Some(Language::Beb),
            609 => Some(Language::Bec),
            610 => Some(Language::Bed),
            611 => Some(Language::Bee),
            612 => Some(Language::Bef),
            613 => Some(Language::Beg),
            614 => Some(Language::Beh),
            615 => Some(Language::Bei),
            616 => Some(Language::Bej),
            617 => Some(Language::Bek),
            618 => Some(Language::Bel),
            619 => Some(Language::Bem),
            620 => Some(Language::Ben),
            621 => Some(Language::Beo),
            622 => Some(Language::Bep),
            623 => Some(Language::Beq),
            624 => Some(Language::Bes),
            625 => Some(Language::Bet),
            626 => Some(Language::Beu),
            627 => Some(Language::Bev),
            628 => Some(Language::Bew),
            629 => Some(Language::Bex),
            630 => Some(Language::Bey),
            631 => Some(Language::Bez),
            632 => Some(Language::Bfa),
            633 => Some(Language::Bfb),
            634 => Some(Language::Bfc),
            635 => Some(Language::Bfd),
            636 => Some(Language::Bfe),
            637 => Some(Language::Bff),
            638 => Some(Language::Bfg),
            639 => Some(Language::Bfh),
            640 => Some(Language::Bfi),
            641 => Some(Language::Bfj),
            642 => Some(Language::Bfk),
            643 => Some(Language::Bfl),
            644 => Some(Language::Bfm),
            645 => Some(Language::Bfn),
            646 => Some(Language::Bfo),
            647 => Some(Language::Bfp),
            648 => Some(Language::Bfq),
            649 => Some(Language::Bfr),
            650 => Some(Language::Bfs),
            651 => Some(Language::Bft),
            652 => Some(Language::Bfu),
            653 => Some(Language::Bfw),
            654 => Some(Language::Bfx),
            655 => Some(Language::Bfy),
            656 => Some(Language::Bfz),
            657 => Some(Language::Bga),
            658 => Some(Language::Bgb),
            659 => Some(Language::Bgc),
            660 => Some(Language::Bgd),
            661 => Some(Language::Bge),
            662 => Some(Language::Bgf),
            663 => Some(Language::Bgg),
            664 => Some(Language::Bgi),
            665 => Some(Language::Bgj),
            666 => Some(Language::Bgk),
            667 => Some(Language::Bgl),
            668 => Some(Language::Bgn),
            669 => Some(Language::Bgo),
            670 => Some(Language::Bgp),
            671 => Some(Language::Bgq),
            672 => Some(Language::Bgr),
            673 => Some(Language::Bgs),
            674 => Some(Language::Bgt),
            675 => Some(Language::Bgu),
            676 => Some(Language::Bgv),
            677 => Some(Language::Bgw),
            678 => Some(Language::Bgx),
            679 => Some(Language::Bgy),
            680 => Some(Language::Bgz),
            681 => Some(Language::Bha),
            682 => Some(Language::Bhb),
            683 => Some(Language::Bhc),
            684 => Some(Language::Bhd),
            685 => Some(Language::Bhe),
            686 => Some(Language::Bhf),
            687 => Some(Language::Bhg),
            688 => Some(Language::Bhh),
            689 => Some(Language::Bhi),
            690 => Some(Language::Bhj),
            691 => Some(Language::Bhl),
            692 => Some(Language::Bhm),
            693 => Some(Language::Bhn),
            694 => Some(Language::Bho),
            695 => Some(Language::Bhp),
            696 => Some(Language::Bhq),
            697 => Some(Language::Bhr),
            698 => Some(Language::Bhs),
            699 => Some(Language::Bht),
            700 => Some(Language::Bhu),
            701 => Some(Language::Bhv),
            702 => Some(Language::Bhw),
            703 => Some(Language::Bhx),
            704 => Some(Language::Bhy),
            705 => Some(Language::Bhz),
            706 => Some(Language::Bia),
            707 => Some(Language::Bib),
            708 => Some(Language::Bid),
            709 => Some(Language::Bie),
            710 => Some(Language::Bif),
            711 => Some(Language::Big),
            712 => Some(Language::Bik),
            713 => Some(Language::Bil),
            714 => Some(Language::Bim),
            715 => Some(Language::Bin),
            716 => Some(Language::Bio),
            717 => Some(Language::Bip),
            718 => Some(Language::Biq),
            719 => Some(Language::Bir),
            720 => Some(Language::Bis),
            721 => Some(Language::Bit),
            722 => Some(Language::Biu),
            723 => Some(Language::Biv),
            724 => Some(Language::Biw),
            725 => Some(Language::Bix),
            726 => Some(Language::Biy),
            727 => Some(Language::Biz),
            728 => Some(Language::Bja),
            729 => Some(Language::Bjb),
            730 => Some(Language::Bjc),
            731 => Some(Language::Bje),
            732 => Some(Language::Bjf),
            733 => Some(Language::Bjg),
            734 => Some(Language::Bjh),
            735 => Some(Language::Bji),
            736 => Some(Language::Bjj),
            737 => Some(Language::Bjk),
            738 => Some(Language::Bjl),
            739 => Some(Language::Bjm),
            740 => Some(Language::Bjn),
            741 => Some(Language::Bjo),
            742 => Some(Language::Bjp),
            743 => Some(Language::Bjr),
            744 => Some(Language::Bjs),
            745 => Some(Language::Bjt),
            746 => Some(Language::Bju),
            747 => Some(Language::Bjv),
            748 => Some(Language::Bjw),
            749 => Some(Language::Bjx),
            750 => Some(Language::Bjy),
            751 => Some(Language::Bjz),
            752 => Some(Language::Bka),
            753 => Some(Language::Bkc),
            754 => Some(Language::Bkd),
            755 => Some(Language::Bkf),
            756 => Some(Language::Bkg),
            757 => Some(Language::Bkh),
            758 => Some(Language::Bki),
            759 => Some(Language::Bkj),
            760 => Some(Language::Bkk),
            761 => Some(Language::Bkl),
            762 => Some(Language::Bkm),
            763 => Some(Language::Bkn),
            764 => Some(Language::Bko),
            765 => Some(Language::Bkp),
            766 => Some(Language::Bkq),
            767 => Some(Language::Bkr),
            768 => Some(Language::Bks),
            769 => Some(Language::Bkt),
            770 => Some(Language::Bku),
            771 => Some(Language::Bkv),
            772 => Some(Language::Bkw),
            773 => Some(Language::Bkx),
            774 => Some(Language::Bky),
            775 => Some(Language::Bkz),
            776 => Some(Language::Bla),
            777 => Some(Language::Blb),
            778 => Some(Language::Blc),
            779 => Some(Language::Bld),
            780 => Some(Language::Ble),
            781 => Some(Language::Blf),
            782 => Some(Language::Blh),
            783 => Some(Language::Bli),
            784 => Some(Language::Blj),
            785 => Some(Language::Blk),
            786 => Some(Language::Bll),
            787 => Some(Language::Blm),
            788 => Some(Language::Bln),
            789 => Some(Language::Blo),
            790 => Some(Language::Blp),
            791 => Some(Language::Blq),
            792 => Some(Language::Blr),
            793 => Some(Language::Bls),
            794 => Some(Language::Blt),
            795 => Some(Language::Blv),
            796 => Some(Language::Blw),
            797 => Some(Language::Blx),
            798 => Some(Language::Bly),
            799 => Some(Language::Blz),
            800 => Some(Language::Bma),
            801 => Some(Language::Bmb),
            802 => Some(Language::Bmc),
            803 => Some(Language::Bmd),
            804 => Some(Language::Bme),
            805 => Some(Language::Bmf),
            806 => Some(Language::Bmg),
            807 => Some(Language::Bmh),
            808 => Some(Language::Bmi),
            809 => Some(Language::Bmj),
            810 => Some(Language::Bmk),
            811 => Some(Language::Bml),
            812 => Some(Language::Bmm),
            813 => Some(Language::Bmn),
            814 => Some(Language::Bmo),
            815 => Some(Language::Bmp),
            816 => Some(Language::Bmq),
            817 => Some(Language::Bmr),
            818 => Some(Language::Bms),
            819 => Some(Language::Bmt),
            820 => Some(Language::Bmu),
            821 => Some(Language::Bmv),
            822 => Some(Language::Bmw),
            823 => Some(Language::Bmx),
            824 => Some(Language::Bmz),
            825 => Some(Language::Bna),
            826 => Some(Language::Bnb),
            827 => Some(Language::Bnc),
            828 => Some(Language::Bnd),
            829 => Some(Language::Bne),
            830 => Some(Language::Bnf),
            831 => Some(Language::Bng),
            832 => Some(Language::Bni),
            833 => Some(Language::Bnj),
            834 => Some(Language::Bnk),
            835 => Some(Language::Bnl),
            836 => Some(Language::Bnm),
            837 => Some(Language::Bnn),
            838 => Some(Language::Bno),
            839 => Some(Language::Bnp),
            840 => Some(Language::Bnq),
            841 => Some(Language::Bnr),
            842 => Some(Language::Bns),
            843 => Some(Language::Bnu),
            844 => Some(Language::Bnv),
            845 => Some(Language::Bnw),
            846 => Some(Language::Bnx),
            847 => Some(Language::Bny),
            848 => Some(Language::Bnz),
            849 => Some(Language::Boa),
            850 => Some(Language::Bob),
            851 => Some(Language::Bod),
            852 => Some(Language::Boe),
            853 => Some(Language::Bof),
            854 => Some(Language::Bog),
            855 => Some(Language::Boh),
            856 => Some(Language::Boi),
            857 => Some(Language::Boj),
            858 => Some(Language::Bok),
            859 => Some(Language::Bol),
            860 => Some(Language::Bom),
            861 => Some(Language::Bon),
            862 => Some(Language::Boo),
            863 => Some(Language::Bop),
            864 => Some(Language::Boq),
            865 => Some(Language::Bor),
            866 => Some(Language::Bos),
            867 => Some(Language::Bot),
            868 => Some(Language::Bou),
            869 => Some(Language::Bov),
            870 => Some(Language::Bow),
            871 => Some(Language::Box),
            872 => Some(Language::Boy),
            873 => Some(Language::Boz),
            874 => Some(Language::Bpa),
            875 => Some(Language::Bpc),
            876 => Some(Language::Bpd),
            877 => Some(Language::Bpe),
            878 => Some(Language::Bpg),
            879 => Some(Language::Bph),
            880 => Some(Language::Bpi),
            881 => Some(Language::Bpj),
            882 => Some(Language::Bpk),
            883 => Some(Language::Bpl),
            884 => Some(Language::Bpm),
            885 => Some(Language::Bpn),
            886 => Some(Language::Bpo),
            887 => Some(Language::Bpp),
            888 => Some(Language::Bpq),
            889 => Some(Language::Bpr),
            890 => Some(Language::Bps),
            891 => Some(Language::Bpt),
            892 => Some(Language::Bpu),
            893 => Some(Language::Bpv),
            894 => Some(Language::Bpw),
            895 => Some(Language::Bpx),
            896 => Some(Language::Bpy),
            897 => Some(Language::Bpz),
            898 => Some(Language::Bqa),
            899 => Some(Language::Bqb),
            900 => Some(Language::Bqc),
            901 => Some(Language::Bqd),
            902 => Some(Language::Bqf),
            903 => Some(Language::Bqg),
            904 => Some(Language::Bqh),
            905 => Some(Language::Bqi),
            906 => Some(Language::Bqj),
            907 => Some(Language::Bqk),
            908 => Some(Language::Bql),
            909 => Some(Language::Bqm),
            910 => Some(Language::Bqn),
            911 => Some(Language::Bqo),
            912 => Some(Language::Bqp),
            913 => Some(Language::Bqq),
            914 => Some(Language::Bqr),
            915 => Some(Language::Bqs),
            916 => Some(Language::Bqt),
            917 => Some(Language::Bqu),
            918 => Some(Language::Bqv),
            919 => Some(Language::Bqw),
            920 => Some(Language::Bqx),
            921 => Some(Language::Bqy),
            922 => Some(Language::Bqz),
            923 => Some(Language::Bra),
            924 => Some(Language::Brb),
            925 => Some(Language::Brc),
            926 => Some(Language::Brd),
            927 => Some(Language::Bre),
            928 => Some(Language::Brf),
            929 => Some(Language::Brg),
            930 => Some(Language::Brh),
            931 => Some(Language::Bri),
            932 => Some(Language::Brj),
            933 => Some(Language::Brk),
            934 => Some(Language::Brl),
            935 => Some(Language::Brm),
            936 => Some(Language::Brn),
            937 => Some(Language::Bro),
            938 => Some(Language::Brp),
            939 => Some(Language::Brq),
            940 => Some(Language::Brr),
            941 => Some(Language::Brs),
            942 => Some(Language::Brt),
            943 => Some(Language::Bru),
            944 => Some(Language::Brv),
            945 => Some(Language::Brw),
            946 => Some(Language::Brx),
            947 => Some(Language::Bry),
            948 => Some(Language::Brz),
            949 => Some(Language::Bsa),
            950 => Some(Language::Bsb),
            951 => Some(Language::Bsc),
            952 => Some(Language::Bse),
            953 => Some(Language::Bsf),
            954 => Some(Language::Bsg),
            955 => Some(Language::Bsh),
            956 => Some(Language::Bsi),
            957 => Some(Language::Bsj),
            958 => Some(Language::Bsk),
            959 => Some(Language::Bsl),
            960 => Some(Language::Bsm),
            961 => Some(Language::Bsn),
            962 => Some(Language::Bso),
            963 => Some(Language::Bsp),
            964 => Some(Language::Bsq),
            965 => Some(Language::Bsr),
            966 => Some(Language::Bss),
            967 => Some(Language::Bst),
            968 => Some(Language::Bsu),
            969 => Some(Language::Bsv),
            970 => Some(Language::Bsw),
            971 => Some(Language::Bsx),
            972 => Some(Language::Bsy),
            973 => Some(Language::Bta),
            974 => Some(Language::Btc),
            975 => Some(Language::Btd),
            976 => Some(Language::Bte),
            977 => Some(Language::Btf),
            978 => Some(Language::Btg),
            979 => Some(Language::Bth),
            980 => Some(Language::Bti),
            981 => Some(Language::Btj),
            982 => Some(Language::Btm),
            983 => Some(Language::Btn),
            984 => Some(Language::Bto),
            985 => Some(Language::Btp),
            986 => Some(Language::Btq),
            987 => Some(Language::Btr),
            988 => Some(Language::Bts),
            989 => Some(Language::Btt),
            990 => Some(Language::Btu),
            991 => Some(Language::Btv),
            992 => Some(Language::Btw),
            993 => Some(Language::Btx),
            994 => Some(Language::Bty),
            995 => Some(Language::Btz),
            996 => Some(Language::Bua),
            997 => Some(Language::Bub),
            998 => Some(Language::Buc),
            999 => Some(Language::Bud),
            1000 => Some(Language::Bue),
            1001 => Some(Language::Buf),
            1002 => Some(Language::Bug),
            1003 => Some(Language::Buh),
            1004 => Some(Language::Bui),
            1005 => Some(Language::Buj),
            1006 => Some(Language::Buk),
            1007 => Some(Language::Bul),
            1008 => Some(Language::Bum),
            1009 => Some(Language::Bun),
            1010 => Some(Language::Buo),
            1011 => Some(Language::Bup),
            1012 => Some(Language::Buq),
            1013 => Some(Language::Bus),
            1014 => Some(Language::But),
            1015 => Some(Language::Buu),
            1016 => Some(Language::Buv),
            1017 => Some(Language::Buw),
            1018 => Some(Language::Bux),
            1019 => Some(Language::Buy),
            1020 => Some(Language::Buz),
            1021 => Some(Language::Bva),
            1022 => Some(Language::Bvb),
            1023 => Some(Language::Bvc),
            1024 => Some(Language::Bvd),
            1025 => Some(Language::Bve),
            1026 => Some(Language::Bvf),
            1027 => Some(Language::Bvg),
            1028 => Some(Language::Bvh),
            1029 => Some(Language::Bvi),
            1030 => Some(Language::Bvj),
            1031 => Some(Language::Bvk),
            1032 => Some(Language::Bvl),
            1033 => Some(Language::Bvm),
            1034 => Some(Language::Bvn),
            1035 => Some(Language::Bvo),
            1036 => Some(Language::Bvp),
            1037 => Some(Language::Bvq),
            1038 => Some(Language::Bvr),
            1039 => Some(Language::Bvt),
            1040 => Some(Language::Bvu),
            1041 => Some(Language::Bvv),
            1042 => Some(Language::Bvw),
            1043 => Some(Language::Bvx),
            1044 => Some(Language::Bvy),
            1045 => Some(Language::Bvz),
            1046 => Some(Language::Bwa),
            1047 => Some(Language::Bwb),
            1048 => Some(Language::Bwc),
            1049 => Some(Language::Bwd),
            1050 => Some(Language::Bwe),
            1051 => Some(Language::Bwf),
            1052 => Some(Language::Bwg),
            1053 => Some(Language::Bwh),
            1054 => Some(Language::Bwi),
            1055 => Some(Language::Bwj),
            1056 => Some(Language::Bwk),
            1057 => Some(Language::Bwl),
            1058 => Some(Language::Bwm),
            1059 => Some(Language::Bwn),
            1060 => Some(Language::Bwo),
            1061 => Some(Language::Bwp),
            1062 => Some(Language::Bwq),
            1063 => Some(Language::Bwr),
            1064 => Some(Language::Bws),
            1065 => Some(Language::Bwt),
            1066 => Some(Language::Bwu),
            1067 => Some(Language::Bww),
            1068 => Some(Language::Bwx),
            1069 => Some(Language::Bwy),
            1070 => Some(Language::Bwz),
            1071 => Some(Language::Bxa),
            1072 => Some(Language::Bxb),
            1073 => Some(Language::Bxc),
            1074 => Some(Language::Bxd),
            1075 => Some(Language::Bxe),
            1076 => Some(Language::Bxf),
            1077 => Some(Language::Bxg),
            1078 => Some(Language::Bxh),
            1079 => Some(Language::Bxi),
            1080 => Some(Language::Bxj),
            1081 => Some(Language::Bxk),
            1082 => Some(Language::Bxl),
            1083 => Some(Language::Bxm),
            1084 => Some(Language::Bxn),
            1085 => Some(Language::Bxo),
            1086 => Some(Language::Bxp),
            1087 => Some(Language::Bxq),
            1088 => Some(Language::Bxr),
            1089 => Some(Language::Bxs),
            1090 => Some(Language::Bxu),
            1091 => Some(Language::Bxv),
            1092 => Some(Language::Bxw),
            1093 => Some(Language::Bxz),
            1094 => Some(Language::Bya),
            1095 => Some(Language::Byb),
            1096 => Some(Language::Byc),
            1097 => Some(Language::Byd),
            1098 => Some(Language::Bye),
            1099 => Some(Language::Byf),
            1100 => Some(Language::Byg),
            1101 => Some(Language::Byh),
            1102 => Some(Language::Byi),
            1103 => Some(Language::Byj),
            1104 => Some(Language::Byk),
            1105 => Some(Language::Byl),
            1106 => Some(Language::Bym),
            1107 => Some(Language::Byn),
            1108 => Some(Language::Byo),
            1109 => Some(Language::Byp),
            1110 => Some(Language::Byq),
            1111 => Some(Language::Byr),
            1112 => Some(Language::Bys),
            1113 => Some(Language::Byt),
            1114 => Some(Language::Byv),
            1115 => Some(Language::Byw),
            1116 => Some(Language::Byx),
            1117 => Some(Language::Byz),
            1118 => Some(Language::Bza),
            1119 => Some(Language::Bzb),
            1120 => Some(Language::Bzc),
            1121 => Some(Language::Bzd),
            1122 => Some(Language::Bze),
            1123 => Some(Language::Bzf),
            1124 => Some(Language::Bzg),
            1125 => Some(Language::Bzh),
            1126 => Some(Language::Bzi),
            1127 => Some(Language::Bzj),
            1128 => Some(Language::Bzk),
            1129 => Some(Language::Bzl),
            1130 => Some(Language::Bzm),
            1131 => Some(Language::Bzn),
            1132 => Some(Language::Bzo),
            1133 => Some(Language::Bzp),
            1134 => Some(Language::Bzq),
            1135 => Some(Language::Bzr),
            1136 => Some(Language::Bzs),
            1137 => Some(Language::Bzt),
            1138 => Some(Language::Bzu),
            1139 => Some(Language::Bzv),
            1140 => Some(Language::Bzw),
            1141 => Some(Language::Bzx),
            1142 => Some(Language::Bzy),
            1143 => Some(Language::Bzz),
            1144 => Some(Language::Caa),
            1145 => Some(Language::Cab),
            1146 => Some(Language::Cac),
            1147 => Some(Language::Cad),
            1148 => Some(Language::Cae),
            1149 => Some(Language::Caf),
            1150 => Some(Language::Cag),
            1151 => Some(Language::Cah),
            1152 => Some(Language::Caj),
            1153 => Some(Language::Cak),
            1154 => Some(Language::Cal),
            1155 => Some(Language::Cam),
            1156 => Some(Language::Can),
            1157 => Some(Language::Cao),
            1158 => Some(Language::Cap),
            1159 => Some(Language::Caq),
            1160 => Some(Language::Car),
            1161 => Some(Language::Cas),
            1162 => Some(Language::Cat),
            1163 => Some(Language::Cav),
            1164 => Some(Language::Caw),
            1165 => Some(Language::Cax),
            1166 => Some(Language::Cay),
            1167 => Some(Language::Caz),
            1168 => Some(Language::Cbb),
            1169 => Some(Language::Cbc),
            1170 => Some(Language::Cbd),
            1171 => Some(Language::Cbg),
            1172 => Some(Language::Cbi),
            1173 => Some(Language::Cbj),
            1174 => Some(Language::Cbk),
            1175 => Some(Language::Cbl),
            1176 => Some(Language::Cbn),
            1177 => Some(Language::Cbo),
            1178 => Some(Language::Cbq),
            1179 => Some(Language::Cbr),
            1180 => Some(Language::Cbs),
            1181 => Some(Language::Cbt),
            1182 => Some(Language::Cbu),
            1183 => Some(Language::Cbv),
            1184 => Some(Language::Cbw),
            1185 => Some(Language::Cby),
            1186 => Some(Language::Ccc),
            1187 => Some(Language::Ccd),
            1188 => Some(Language::Cce),
            1189 => Some(Language::Ccg),
            1190 => Some(Language::Cch),
            1191 => Some(Language::Ccj),
            1192 => Some(Language::Ccl),
            1193 => Some(Language::Ccm),
            1194 => Some(Language::Cco),
            1195 => Some(Language::Ccp),
            1196 => Some(Language::Ccr),
            1197 => Some(Language::Cda),
            1198 => Some(Language::Cde),
            1199 => Some(Language::Cdf),
            1200 => Some(Language::Cdh),
            1201 => Some(Language::Cdi),
            1202 => Some(Language::Cdj),
            1203 => Some(Language::Cdm),
            1204 => Some(Language::Cdn),
            1205 => Some(Language::Cdo),
            1206 => Some(Language::Cdr),
            1207 => Some(Language::Cds),
            1208 => Some(Language::Cdy),
            1209 => Some(Language::Cdz),
            1210 => Some(Language::Cea),
            1211 => Some(Language::Ceb),
            1212 => Some(Language::Ceg),
            1213 => Some(Language::Cek),
            1214 => Some(Language::Cen),
            1215 => Some(Language::Ces),
            1216 => Some(Language::Cet),
            1217 => Some(Language::Cey),
            1218 => Some(Language::Cfa),
            1219 => Some(Language::Cfd),
            1220 => Some(Language::Cfg),
            1221 => Some(Language::Cfm),
            1222 => Some(Language::Cga),
            1223 => Some(Language::Cgc),
            1224 => Some(Language::Cgg),
            1225 => Some(Language::Cgk),
            1226 => Some(Language::Cha),
            1227 => Some(Language::Chb),
            1228 => Some(Language::Chc),
            1229 => Some(Language::Chd),
            1230 => Some(Language::Che),
            1231 => Some(Language::Chf),
            1232 => Some(Language::Chg),
            1233 => Some(Language::Chh),
            1234 => Some(Language::Chj),
            1235 => Some(Language::Chk),
            1236 => Some(Language::Chl),
            1237 => Some(Language::Chm),
            1238 => Some(Language::Chn),
            1239 => Some(Language::Cho),
            1240 => Some(Language::Chp),
            1241 => Some(Language::Chq),
            1242 => Some(Language::Chr),
            1243 => Some(Language::Cht),
            1244 => Some(Language::Chu),
            1245 => Some(Language::Chv),
            1246 => Some(Language::Chw),
            1247 => Some(Language::Chx),
            1248 => Some(Language::Chy),
            1249 => Some(Language::Chz),
            1250 => Some(Language::Cia),
            1251 => Some(Language::Cib),
            1252 => Some(Language::Cic),
            1253 => Some(Language::Cid),
            1254 => Some(Language::Cie),
            1255 => Some(Language::Cih),
            1256 => Some(Language::Cik),
            1257 => Some(Language::Cim),
            1258 => Some(Language::Cin),
            1259 => Some(Language::Cip),
            1260 => Some(Language::Cir),
            1261 => Some(Language::Ciw),
            1262 => Some(Language::Ciy),
            1263 => Some(Language::Cja),
            1264 => Some(Language::Cje),
            1265 => Some(Language::Cjh),
            1266 => Some(Language::Cji),
            1267 => Some(Language::Cjk),
            1268 => Some(Language::Cjm),
            1269 => Some(Language::Cjn),
            1270 => Some(Language::Cjo),
            1271 => Some(Language::Cjp),
            1272 => Some(Language::Cjs),
            1273 => Some(Language::Cjv),
            1274 => Some(Language::Cjy),
            1275 => Some(Language::Ckb),
            1276 => Some(Language::Ckh),
            1277 => Some(Language::Ckl),
            1278 => Some(Language::Ckm),
            1279 => Some(Language::Ckn),
            1280 => Some(Language::Cko),
            1281 => Some(Language::Ckq),
            1282 => Some(Language::Ckr),
            1283 => Some(Language::Cks),
            1284 => Some(Language::Ckt),
            1285 => Some(Language::Cku),
            1286 => Some(Language::Ckv),
            1287 => Some(Language::Ckx),
            1288 => Some(Language::Cky),
            1289 => Some(Language::Ckz),
            1290 => Some(Language::Cla),
            1291 => Some(Language::Clc),
            1292 => Some(Language::Cld),
            1293 => Some(Language::Cle),
            1294 => Some(Language::Clh),
            1295 => Some(Language::Cli),
            1296 => Some(Language::Clj),
            1297 => Some(Language::Clk),
            1298 => Some(Language::Cll),
            1299 => Some(Language::Clm),
            1300 => Some(Language::Clo),
            1301 => Some(Language::Clt),
            1302 => Some(Language::Clu),
            1303 => Some(Language::Clw),
            1304 => Some(Language::Cly),
            1305 => Some(Language::Cma),
            1306 => Some(Language::Cme),
            1307 => Some(Language::Cmg),
            1308 => Some(Language::Cmi),
            1309 => Some(Language::Cml),
            1310 => Some(Language::Cmm),
            1311 => Some(Language::Cmn),
            1312 => Some(Language::Cmo),
            1313 => Some(Language::Cmr),
            1314 => Some(Language::Cms),
            1315 => Some(Language::Cmt),
            1316 => Some(Language::Cna),
            1317 => Some(Language::Cnb),
            1318 => Some(Language::Cnc),
            1319 => Some(Language::Cng),
            1320 => Some(Language::Cnh),
            1321 => Some(Language::Cni),
            1322 => Some(Language::Cnk),
            1323 => Some(Language::Cnl),
            1324 => Some(Language::Cno),
            1325 => Some(Language::Cnp),
            1326 => Some(Language::Cnq),
            1327 => Some(Language::Cnr),
            1328 => Some(Language::Cns),
            1329 => Some(Language::Cnt),
            1330 => Some(Language::Cnu),
            1331 => Some(Language::Cnw),
            1332 => Some(Language::Cnx),
            1333 => Some(Language::Coa),
            1334 => Some(Language::Cob),
            1335 => Some(Language::Coc),
            1336 => Some(Language::Cod),
            1337 => Some(Language::Coe),
            1338 => Some(Language::Cof),
            1339 => Some(Language::Cog),
            1340 => Some(Language::Coh),
            1341 => Some(Language::Coj),
            1342 => Some(Language::Cok),
            1343 => Some(Language::Col),
            1344 => Some(Language::Com),
            1345 => Some(Language::Con),
            1346 => Some(Language::Coo),
            1347 => Some(Language::Cop),
            1348 => Some(Language::Coq),
            1349 => Some(Language::Cor),
            1350 => Some(Language::Cos),
            1351 => Some(Language::Cot),
            1352 => Some(Language::Cou),
            1353 => Some(Language::Cov),
            1354 => Some(Language::Cow),
            1355 => Some(Language::Cox),
            1356 => Some(Language::Coz),
            1357 => Some(Language::Cpa),
            1358 => Some(Language::Cpb),
            1359 => Some(Language::Cpc),
            1360 => Some(Language::Cpg),
            1361 => Some(Language::Cpi),
            1362 => Some(Language::Cpn),
            1363 => Some(Language::Cpo),
            1364 => Some(Language::Cps),
            1365 => Some(Language::Cpu),
            1366 => Some(Language::Cpx),
            1367 => Some(Language::Cpy),
            1368 => Some(Language::Cqd),
            1369 => Some(Language::Cra),
            1370 => Some(Language::Crb),
            1371 => Some(Language::Crc),
            1372 => Some(Language::Crd),
            1373 => Some(Language::Cre),
            1374 => Some(Language::Crf),
            1375 => Some(Language::Crg),
            1376 => Some(Language::Crh),
            1377 => Some(Language::Cri),
            1378 => Some(Language::Crj),
            1379 => Some(Language::Crk),
            1380 => Some(Language::Crl),
            1381 => Some(Language::Crm),
            1382 => Some(Language::Crn),
            1383 => Some(Language::Cro),
            1384 => Some(Language::Crq),
            1385 => Some(Language::Crr),
            1386 => Some(Language::Crs),
            1387 => Some(Language::Crt),
            1388 => Some(Language::Crv),
            1389 => Some(Language::Crw),
            1390 => Some(Language::Crx),
            1391 => Some(Language::Cry),
            1392 => Some(Language::Crz),
            1393 => Some(Language::Csa),
            1394 => Some(Language::Csb),
            1395 => Some(Language::Csc),
            1396 => Some(Language::Csd),
            1397 => Some(Language::Cse),
            1398 => Some(Language::Csf),
            1399 => Some(Language::Csg),
            1400 => Some(Language::Csh),
            1401 => Some(Language::Csi),
            1402 => Some(Language::Csj),
            1403 => Some(Language::Csk),
            1404 => Some(Language::Csl),
            1405 => Some(Language::Csm),
            1406 => Some(Language::Csn),
            1407 => Some(Language::Cso),
            1408 => Some(Language::Csp),
            1409 => Some(Language::Csq),
            1410 => Some(Language::Csr),
            1411 => Some(Language::Css),
            1412 => Some(Language::Cst),
            1413 => Some(Language::Csv),
            1414 => Some(Language::Csw),
            1415 => Some(Language::Csx),
            1416 => Some(Language::Csy),
            1417 => Some(Language::Csz),
            1418 => Some(Language::Cta),
            1419 => Some(Language::Ctc),
            1420 => Some(Language::Ctd),
            1421 => Some(Language::Cte),
            1422 => Some(Language::Ctg),
            1423 => Some(Language::Cth),
            1424 => Some(Language::Ctl),
            1425 => Some(Language::Ctm),
            1426 => Some(Language::Ctn),
            1427 => Some(Language::Cto),
            1428 => Some(Language::Ctp),
            1429 => Some(Language::Cts),
            1430 => Some(Language::Ctt),
            1431 => Some(Language::Ctu),
            1432 => Some(Language::Cty),
            1433 => Some(Language::Ctz),
            1434 => Some(Language::Cua),
            1435 => Some(Language::Cub),
            1436 => Some(Language::Cuc),
            1437 => Some(Language::Cuh),
            1438 => Some(Language::Cui),
            1439 => Some(Language::Cuj),
            1440 => Some(Language::Cuk),
            1441 => Some(Language::Cul),
            1442 => Some(Language::Cuo),
            1443 => Some(Language::Cup),
            1444 => Some(Language::Cuq),
            1445 => Some(Language::Cur),
            1446 => Some(Language::Cut),
            1447 => Some(Language::Cuu),
            1448 => Some(Language::Cuv),
            1449 => Some(Language::Cuw),
            1450 => Some(Language::Cux),
            1451 => Some(Language::Cuy),
            1452 => Some(Language::Cvg),
            1453 => Some(Language::Cvn),
            1454 => Some(Language::Cwa),
            1455 => Some(Language::Cwb),
            1456 => Some(Language::Cwd),
            1457 => Some(Language::Cwe),
            1458 => Some(Language::Cwg),
            1459 => Some(Language::Cwt),
            1460 => Some(Language::Cya),
            1461 => Some(Language::Cyb),
            1462 => Some(Language::Cym),
            1463 => Some(Language::Cyo),
            1464 => Some(Language::Czh),
            1465 => Some(Language::Czk),
            1466 => Some(Language::Czn),
            1467 => Some(Language::Czo),
            1468 => Some(Language::Czt),
            1469 => Some(Language::Daa),
            1470 => Some(Language::Dac),
            1471 => Some(Language::Dad),
            1472 => Some(Language::Dae),
            1473 => Some(Language::Dag),
            1474 => Some(Language::Dah),
            1475 => Some(Language::Dai),
            1476 => Some(Language::Daj),
            1477 => Some(Language::Dak),
            1478 => Some(Language::Dal),
            1479 => Some(Language::Dam),
            1480 => Some(Language::Dan),
            1481 => Some(Language::Dao),
            1482 => Some(Language::Daq),
            1483 => Some(Language::Dar),
            1484 => Some(Language::Das),
            1485 => Some(Language::Dau),
            1486 => Some(Language::Dav),
            1487 => Some(Language::Daw),
            1488 => Some(Language::Dax),
            1489 => Some(Language::Daz),
            1490 => Some(Language::Dba),
            1491 => Some(Language::Dbb),
            1492 => Some(Language::Dbd),
            1493 => Some(Language::Dbe),
            1494 => Some(Language::Dbf),
            1495 => Some(Language::Dbg),
            1496 => Some(Language::Dbi),
            1497 => Some(Language::Dbj),
            1498 => Some(Language::Dbl),
            1499 => Some(Language::Dbm),
            1500 => Some(Language::Dbn),
            1501 => Some(Language::Dbo),
            1502 => Some(Language::Dbp),
            1503 => Some(Language::Dbq),
            1504 => Some(Language::Dbr),
            1505 => Some(Language::Dbt),
            1506 => Some(Language::Dbu),
            1507 => Some(Language::Dbv),
            1508 => Some(Language::Dbw),
            1509 => Some(Language::Dby),
            1510 => Some(Language::Dcc),
            1511 => Some(Language::Dcr),
            1512 => Some(Language::Dda),
            1513 => Some(Language::Ddd),
            1514 => Some(Language::Dde),
            1515 => Some(Language::Ddg),
            1516 => Some(Language::Ddi),
            1517 => Some(Language::Ddj),
            1518 => Some(Language::Ddn),
            1519 => Some(Language::Ddo),
            1520 => Some(Language::Ddr),
            1521 => Some(Language::Dds),
            1522 => Some(Language::Ddw),
            1523 => Some(Language::Dec),
            1524 => Some(Language::Ded),
            1525 => Some(Language::Dee),
            1526 => Some(Language::Def),
            1527 => Some(Language::Deg),
            1528 => Some(Language::Deh),
            1529 => Some(Language::Dei),
            1530 => Some(Language::Dek),
            1531 => Some(Language::Del),
            1532 => Some(Language::Dem),
            1533 => Some(Language::Den),
            1534 => Some(Language::Dep),
            1535 => Some(Language::Deq),
            1536 => Some(Language::Der),
            1537 => Some(Language::Des),
            1538 => Some(Language::Deu),
            1539 => Some(Language::Dev),
            1540 => Some(Language::Dez),
            1541 => Some(Language::Dga),
            1542 => Some(Language::Dgb),
            1543 => Some(Language::Dgc),
            1544 => Some(Language::Dgd),
            1545 => Some(Language::Dge),
            1546 => Some(Language::Dgg),
            1547 => Some(Language::Dgh),
            1548 => Some(Language::Dgi),
            1549 => Some(Language::Dgk),
            1550 => Some(Language::Dgl),
            1551 => Some(Language::Dgn),
            1552 => Some(Language::Dgo),
            1553 => Some(Language::Dgr),
            1554 => Some(Language::Dgs),
            1555 => Some(Language::Dgt),
            1556 => Some(Language::Dgw),
            1557 => Some(Language::Dgx),
            1558 => Some(Language::Dgz),
            1559 => Some(Language::Dhd),
            1560 => Some(Language::Dhg),
            1561 => Some(Language::Dhi),
            1562 => Some(Language::Dhl),
            1563 => Some(Language::Dhm),
            1564 => Some(Language::Dhn),
            1565 => Some(Language::Dho),
            1566 => Some(Language::Dhr),
            1567 => Some(Language::Dhs),
            1568 => Some(Language::Dhu),
            1569 => Some(Language::Dhv),
            1570 => Some(Language::Dhw),
            1571 => Some(Language::Dhx),
            1572 => Some(Language::Dia),
            1573 => Some(Language::Dib),
            1574 => Some(Language::Dic),
            1575 => Some(Language::Did),
            1576 => Some(Language::Dif),
            1577 => Some(Language::Dig),
            1578 => Some(Language::Dih),
            1579 => Some(Language::Dii),
            1580 => Some(Language::Dij),
            1581 => Some(Language::Dik),
            1582 => Some(Language::Dil),
            1583 => Some(Language::Dim),
            1584 => Some(Language::Din),
            1585 => Some(Language::Dio),
            1586 => Some(Language::Dip),
            1587 => Some(Language::Diq),
            1588 => Some(Language::Dir),
            1589 => Some(Language::Dis),
            1590 => Some(Language::Diu),
            1591 => Some(Language::Div),
            1592 => Some(Language::Diw),
            1593 => Some(Language::Dix),
            1594 => Some(Language::Diy),
            1595 => Some(Language::Diz),
            1596 => Some(Language::Dja),
            1597 => Some(Language::Djb),
            1598 => Some(Language::Djc),
            1599 => Some(Language::Djd),
            1600 => Some(Language::Dje),
            1601 => Some(Language::Djf),
            1602 => Some(Language::Dji),
            1603 => Some(Language::Djj),
            1604 => Some(Language::Djk),
            1605 => Some(Language::Djm),
            1606 => Some(Language::Djn),
            1607 => Some(Language::Djo),
            1608 => Some(Language::Djr),
            1609 => Some(Language::Dju),
            1610 => Some(Language::Djw),
            1611 => Some(Language::Dka),
            1612 => Some(Language::Dkg),
            1613 => Some(Language::Dkk),
            1614 => Some(Language::Dkr),
            1615 => Some(Language::Dks),
            1616 => Some(Language::Dkx),
            1617 => Some(Language::Dlg),
            1618 => Some(Language::Dlk),
            1619 => Some(Language::Dlm),
            1620 => Some(Language::Dln),
            1621 => Some(Language::Dma),
            1622 => Some(Language::Dmb),
            1623 => Some(Language::Dmc),
            1624 => Some(Language::Dmd),
            1625 => Some(Language::Dme),
            1626 => Some(Language::Dmf),
            1627 => Some(Language::Dmg),
            1628 => Some(Language::Dmk),
            1629 => Some(Language::Dml),
            1630 => Some(Language::Dmm),
            1631 => Some(Language::Dmo),
            1632 => Some(Language::Dmr),
            1633 => Some(Language::Dms),
            1634 => Some(Language::Dmu),
            1635 => Some(Language::Dmv),
            1636 => Some(Language::Dmw),
            1637 => Some(Language::Dmx),
            1638 => Some(Language::Dmy),
            1639 => Some(Language::Dna),
            1640 => Some(Language::Dnd),
            1641 => Some(Language::Dne),
            1642 => Some(Language::Dng),
            1643 => Some(Language::Dni),
            1644 => Some(Language::Dnj),
            1645 => Some(Language::Dnk),
            1646 => Some(Language::Dnn),
            1647 => Some(Language::Dno),
            1648 => Some(Language::Dnr),
            1649 => Some(Language::Dnt),
            1650 => Some(Language::Dnu),
            1651 => Some(Language::Dnv),
            1652 => Some(Language::Dnw),
            1653 => Some(Language::Dny),
            1654 => Some(Language::Doa),
            1655 => Some(Language::Dob),
            1656 => Some(Language::Doc),
            1657 => Some(Language::Doe),
            1658 => Some(Language::Dof),
            1659 => Some(Language::Doh),
            1660 => Some(Language::Doi),
            1661 => Some(Language::Dok),
            1662 => Some(Language::Dol),
            1663 => Some(Language::Don),
            1664 => Some(Language::Doo),
            1665 => Some(Language::Dop),
            1666 => Some(Language::Doq),
            1667 => Some(Language::Dor),
            1668 => Some(Language::Dos),
            1669 => Some(Language::Dot),
            1670 => Some(Language::Dov),
            1671 => Some(Language::Dow),
            1672 => Some(Language::Dox),
            1673 => Some(Language::Doy),
            1674 => Some(Language::Doz),
            1675 => Some(Language::Dpp),
            1676 => Some(Language::Drb),
            1677 => Some(Language::Drc),
            1678 => Some(Language::Drd),
            1679 => Some(Language::Dre),
            1680 => Some(Language::Drg),
            1681 => Some(Language::Dri),
            1682 => Some(Language::Drl),
            1683 => Some(Language::Drn),
            1684 => Some(Language::Dro),
            1685 => Some(Language::Drq),
            1686 => Some(Language::Drs),
            1687 => Some(Language::Drt),
            1688 => Some(Language::Dru),
            1689 => Some(Language::Dry),
            1690 => Some(Language::Dsb),
            1691 => Some(Language::Dse),
            1692 => Some(Language::Dsh),
            1693 => Some(Language::Dsi),
            1694 => Some(Language::Dsl),
            1695 => Some(Language::Dsn),
            1696 => Some(Language::Dso),
            1697 => Some(Language::Dsq),
            1698 => Some(Language::Dsz),
            1699 => Some(Language::Dta),
            1700 => Some(Language::Dtb),
            1701 => Some(Language::Dtd),
            1702 => Some(Language::Dth),
            1703 => Some(Language::Dti),
            1704 => Some(Language::Dtk),
            1705 => Some(Language::Dtm),
            1706 => Some(Language::Dtn),
            1707 => Some(Language::Dto),
            1708 => Some(Language::Dtp),
            1709 => Some(Language::Dtr),
            1710 => Some(Language::Dts),
            1711 => Some(Language::Dtt),
            1712 => Some(Language::Dtu),
            1713 => Some(Language::Dty),
            1714 => Some(Language::Dua),
            1715 => Some(Language::Dub),
            1716 => Some(Language::Duc),
            1717 => Some(Language::Due),
            1718 => Some(Language::Duf),
            1719 => Some(Language::Dug),
            1720 => Some(Language::Duh),
            1721 => Some(Language::Dui),
            1722 => Some(Language::Duk),
            1723 => Some(Language::Dul),
            1724 => Some(Language::Dum),
            1725 => Some(Language::Dun),
            1726 => Some(Language::Duo),
            1727 => Some(Language::Dup),
            1728 => Some(Language::Duq),
            1729 => Some(Language::Dur),
            1730 => Some(Language::Dus),
            1731 => Some(Language::Duu),
            1732 => Some(Language::Duv),
            1733 => Some(Language::Duw),
            1734 => Some(Language::Dux),
            1735 => Some(Language::Duy),
            1736 => Some(Language::Duz),
            1737 => Some(Language::Dva),
            1738 => Some(Language::Dwa),
            1739 => Some(Language::Dwk),
            1740 => Some(Language::Dwr),
            1741 => Some(Language::Dws),
            1742 => Some(Language::Dwu),
            1743 => Some(Language::Dww),
            1744 => Some(Language::Dwy),
            1745 => Some(Language::Dwz),
            1746 => Some(Language::Dya),
            1747 => Some(Language::Dyb),
            1748 => Some(Language::Dyd),
            1749 => Some(Language::Dyg),
            1750 => Some(Language::Dyi),
            1751 => Some(Language::Dym),
            1752 => Some(Language::Dyn),
            1753 => Some(Language::Dyo),
            1754 => Some(Language::Dyu),
            1755 => Some(Language::Dyy),
            1756 => Some(Language::Dza),
            1757 => Some(Language::Dze),
            1758 => Some(Language::Dzg),
            1759 => Some(Language::Dzl),
            1760 => Some(Language::Dzn),
            1761 => Some(Language::Dzo),
            1762 => Some(Language::Eaa),
            1763 => Some(Language::Ebc),
            1764 => Some(Language::Ebg),
            1765 => Some(Language::Ebk),
            1766 => Some(Language::Ebo),
            1767 => Some(Language::Ebr),
            1768 => Some(Language::Ebu),
            1769 => Some(Language::Ecr),
            1770 => Some(Language::Ecs),
            1771 => Some(Language::Ecy),
            1772 => Some(Language::Eee),
            1773 => Some(Language::Efa),
            1774 => Some(Language::Efe),
            1775 => Some(Language::Efi),
            1776 => Some(Language::Ega),
            1777 => Some(Language::Egl),
            1778 => Some(Language::Egm),
            1779 => Some(Language::Ego),
            1780 => Some(Language::Egy),
            1781 => Some(Language::Ehs),
            1782 => Some(Language::Ehu),
            1783 => Some(Language::Eip),
            1784 => Some(Language::Eit),
            1785 => Some(Language::Eiv),
            1786 => Some(Language::Eja),
            1787 => Some(Language::Eka),
            1788 => Some(Language::Eke),
            1789 => Some(Language::Ekg),
            1790 => Some(Language::Eki),
            1791 => Some(Language::Ekk),
            1792 => Some(Language::Ekl),
            1793 => Some(Language::Ekm),
            1794 => Some(Language::Eko),
            1795 => Some(Language::Ekp),
            1796 => Some(Language::Ekr),
            1797 => Some(Language::Eky),
            1798 => Some(Language::Ele),
            1799 => Some(Language::Elh),
            1800 => Some(Language::Eli),
            1801 => Some(Language::Elk),
            1802 => Some(Language::Ell),
            1803 => Some(Language::Elm),
            1804 => Some(Language::Elo),
            1805 => Some(Language::Elu),
            1806 => Some(Language::Elx),
            1807 => Some(Language::Ema),
            1808 => Some(Language::Emb),
            1809 => Some(Language::Eme),
            1810 => Some(Language::Emg),
            1811 => Some(Language::Emi),
            1812 => Some(Language::Emk),
            1813 => Some(Language::Emm),
            1814 => Some(Language::Emn),
            1815 => Some(Language::Emp),
            1816 => Some(Language::Emq),
            1817 => Some(Language::Ems),
            1818 => Some(Language::Emu),
            1819 => Some(Language::Emw),
            1820 => Some(Language::Emx),
            1821 => Some(Language::Emy),
            1822 => Some(Language::Emz),
            1823 => Some(Language::Ena),
            1824 => Some(Language::Enb),
            1825 => Some(Language::Enc),
            1826 => Some(Language::End),
            1827 => Some(Language::Enf),
            1828 => Some(Language::Eng),
            1829 => Some(Language::Enh),
            1830 => Some(Language::Enl),
            1831 => Some(Language::Enm),
            1832 => Some(Language::Enn),
            1833 => Some(Language::Eno),
            1834 => Some(Language::Enq),
            1835 => Some(Language::Enr),
            1836 => Some(Language::Enu),
            1837 => Some(Language::Env),
            1838 => Some(Language::Enw),
            1839 => Some(Language::Enx),
            1840 => Some(Language::Eot),
            1841 => Some(Language::Epi),
            1842 => Some(Language::Epo),
            1843 => Some(Language::Era),
            1844 => Some(Language::Erg),
            1845 => Some(Language::Erh),
            1846 => Some(Language::Eri),
            1847 => Some(Language::Erk),
            1848 => Some(Language::Ero),
            1849 => Some(Language::Err),
            1850 => Some(Language::Ers),
            1851 => Some(Language::Ert),
            1852 => Some(Language::Erw),
            1853 => Some(Language::Ese),
            1854 => Some(Language::Esg),
            1855 => Some(Language::Esh),
            1856 => Some(Language::Esi),
            1857 => Some(Language::Esk),
            1858 => Some(Language::Esl),
            1859 => Some(Language::Esm),
            1860 => Some(Language::Esn),
            1861 => Some(Language::Eso),
            1862 => Some(Language::Esq),
            1863 => Some(Language::Ess),
            1864 => Some(Language::Est),
            1865 => Some(Language::Esu),
            1866 => Some(Language::Esy),
            1867 => Some(Language::Etb),
            1868 => Some(Language::Etc),
            1869 => Some(Language::Eth),
            1870 => Some(Language::Etn),
            1871 => Some(Language::Eto),
            1872 => Some(Language::Etr),
            1873 => Some(Language::Ets),
            1874 => Some(Language::Ett),
            1875 => Some(Language::Etu),
            1876 => Some(Language::Etx),
            1877 => Some(Language::Etz),
            1878 => Some(Language::Eus),
            1879 => Some(Language::Eve),
            1880 => Some(Language::Evh),
            1881 => Some(Language::Evn),
            1882 => Some(Language::Ewe),
            1883 => Some(Language::Ewo),
            1884 => Some(Language::Ext),
            1885 => Some(Language::Eya),
            1886 => Some(Language::Eyo),
            1887 => Some(Language::Eza),
            1888 => Some(Language::Eze),
            1889 => Some(Language::Faa),
            1890 => Some(Language::Fab),
            1891 => Some(Language::Fad),
            1892 => Some(Language::Faf),
            1893 => Some(Language::Fag),
            1894 => Some(Language::Fah),
            1895 => Some(Language::Fai),
            1896 => Some(Language::Faj),
            1897 => Some(Language::Fak),
            1898 => Some(Language::Fal),
            1899 => Some(Language::Fam),
            1900 => Some(Language::Fan),
            1901 => Some(Language::Fao),
            1902 => Some(Language::Fap),
            1903 => Some(Language::Far),
            1904 => Some(Language::Fas),
            1905 => Some(Language::Fat),
            1906 => Some(Language::Fau),
            1907 => Some(Language::Fax),
            1908 => Some(Language::Fay),
            1909 => Some(Language::Faz),
            1910 => Some(Language::Fbl),
            1911 => Some(Language::Fcs),
            1912 => Some(Language::Fer),
            1913 => Some(Language::Ffi),
            1914 => Some(Language::Ffm),
            1915 => Some(Language::Fgr),
            1916 => Some(Language::Fia),
            1917 => Some(Language::Fie),
            1918 => Some(Language::Fif),
            1919 => Some(Language::Fij),
            1920 => Some(Language::Fil),
            1921 => Some(Language::Fin),
            1922 => Some(Language::Fip),
            1923 => Some(Language::Fir),
            1924 => Some(Language::Fit),
            1925 => Some(Language::Fiw),
            1926 => Some(Language::Fkk),
            1927 => Some(Language::Fkv),
            1928 => Some(Language::Fla),
            1929 => Some(Language::Flh),
            1930 => Some(Language::Fli),
            1931 => Some(Language::Fll),
            1932 => Some(Language::Fln),
            1933 => Some(Language::Flr),
            1934 => Some(Language::Fly),
            1935 => Some(Language::Fmp),
            1936 => Some(Language::Fmu),
            1937 => Some(Language::Fnb),
            1938 => Some(Language::Fng),
            1939 => Some(Language::Fni),
            1940 => Some(Language::Fod),
            1941 => Some(Language::Foi),
            1942 => Some(Language::Fom),
            1943 => Some(Language::Fon),
            1944 => Some(Language::For),
            1945 => Some(Language::Fos),
            1946 => Some(Language::Fpe),
            1947 => Some(Language::Fqs),
            1948 => Some(Language::Fra),
            1949 => Some(Language::Frc),
            1950 => Some(Language::Frd),
            1951 => Some(Language::Frk),
            1952 => Some(Language::Frm),
            1953 => Some(Language::Fro),
            1954 => Some(Language::Frp),
            1955 => Some(Language::Frq),
            1956 => Some(Language::Frr),
            1957 => Some(Language::Frs),
            1958 => Some(Language::Frt),
            1959 => Some(Language::Fry),
            1960 => Some(Language::Fse),
            1961 => Some(Language::Fsl),
            1962 => Some(Language::Fss),
            1963 => Some(Language::Fub),
            1964 => Some(Language::Fuc),
            1965 => Some(Language::Fud),
            1966 => Some(Language::Fue),
            1967 => Some(Language::Fuf),
            1968 => Some(Language::Fuh),
            1969 => Some(Language::Fui),
            1970 => Some(Language::Fuj),
            1971 => Some(Language::Ful),
            1972 => Some(Language::Fum),
            1973 => Some(Language::Fun),
            1974 => Some(Language::Fuq),
            1975 => Some(Language::Fur),
            1976 => Some(Language::Fut),
            1977 => Some(Language::Fuu),
            1978 => Some(Language::Fuv),
            1979 => Some(Language::Fuy),
            1980 => Some(Language::Fvr),
            1981 => Some(Language::Fwa),
            1982 => Some(Language::Fwe),
            1983 => Some(Language::Gaa),
            1984 => Some(Language::Gab),
            1985 => Some(Language::Gac),
            1986 => Some(Language::Gad),
            1987 => Some(Language::Gae),
            1988 => Some(Language::Gaf),
            1989 => Some(Language::Gag),
            1990 => Some(Language::Gah),
            1991 => Some(Language::Gai),
            1992 => Some(Language::Gaj),
            1993 => Some(Language::Gak),
            1994 => Some(Language::Gal),
            1995 => Some(Language::Gam),
            1996 => Some(Language::Gan),
            1997 => Some(Language::Gao),
            1998 => Some(Language::Gap),
            1999 => Some(Language::Gaq),
            2000 => Some(Language::Gar),
            2001 => Some(Language::Gas),
            2002 => Some(Language::Gat),
            2003 => Some(Language::Gau),
            2004 => Some(Language::Gaw),
            2005 => Some(Language::Gax),
            2006 => Some(Language::Gay),
            2007 => Some(Language::Gaz),
            2008 => Some(Language::Gba),
            2009 => Some(Language::Gbb),
            2010 => Some(Language::Gbd),
            2011 => Some(Language::Gbe),
            2012 => Some(Language::Gbf),
            2013 => Some(Language::Gbg),
            2014 => Some(Language::Gbh),
            2015 => Some(Language::Gbi),
            2016 => Some(Language::Gbj),
            2017 => Some(Language::Gbk),
            2018 => Some(Language::Gbl),
            2019 => Some(Language::Gbm),
            2020 => Some(Language::Gbn),
            2021 => Some(Language::Gbo),
            2022 => Some(Language::Gbp),
            2023 => Some(Language::Gbq),
            2024 => Some(Language::Gbr),
            2025 => Some(Language::Gbs),
            2026 => Some(Language::Gbu),
            2027 => Some(Language::Gbv),
            2028 => Some(Language::Gbw),
            2029 => Some(Language::Gbx),
            2030 => Some(Language::Gby),
            2031 => Some(Language::Gbz),
            2032 => Some(Language::Gcc),
            2033 => Some(Language::Gcd),
            2034 => Some(Language::Gce),
            2035 => Some(Language::Gcf),
            2036 => Some(Language::Gcl),
            2037 => Some(Language::Gcn),
            2038 => Some(Language::Gcr),
            2039 => Some(Language::Gct),
            2040 => Some(Language::Gda),
            2041 => Some(Language::Gdb),
            2042 => Some(Language::Gdc),
            2043 => Some(Language::Gdd),
            2044 => Some(Language::Gde),
            2045 => Some(Language::Gdf),
            2046 => Some(Language::Gdg),
            2047 => Some(Language::Gdh),
            2048 => Some(Language::Gdi),
            2049 => Some(Language::Gdj),
            2050 => Some(Language::Gdk),
            2051 => Some(Language::Gdl),
            2052 => Some(Language::Gdm),
            2053 => Some(Language::Gdn),
            2054 => Some(Language::Gdo),
            2055 => Some(Language::Gdq),
            2056 => Some(Language::Gdr),
            2057 => Some(Language::Gds),
            2058 => Some(Language::Gdt),
            2059 => Some(Language::Gdu),
            2060 => Some(Language::Gdx),
            2061 => Some(Language::Gea),
            2062 => Some(Language::Geb),
            2063 => Some(Language::Gec),
            2064 => Some(Language::Ged),
            2065 => Some(Language::Gef),
            2066 => Some(Language::Geg),
            2067 => Some(Language::Geh),
            2068 => Some(Language::Gei),
            2069 => Some(Language::Gej),
            2070 => Some(Language::Gek),
            2071 => Some(Language::Gel),
            2072 => Some(Language::Geq),
            2073 => Some(Language::Ges),
            2074 => Some(Language::Gev),
            2075 => Some(Language::Gew),
            2076 => Some(Language::Gex),
            2077 => Some(Language::Gey),
            2078 => Some(Language::Gez),
            2079 => Some(Language::Gfk),
            2080 => Some(Language::Gft),
            2081 => Some(Language::Gga),
            2082 => Some(Language::Ggb),
            2083 => Some(Language::Ggd),
            2084 => Some(Language::Gge),
            2085 => Some(Language::Ggg),
            2086 => Some(Language::Ggk),
            2087 => Some(Language::Ggl),
            2088 => Some(Language::Ggt),
            2089 => Some(Language::Ggu),
            2090 => Some(Language::Ggw),
            2091 => Some(Language::Gha),
            2092 => Some(Language::Ghc),
            2093 => Some(Language::Ghe),
            2094 => Some(Language::Ghh),
            2095 => Some(Language::Ghk),
            2096 => Some(Language::Ghl),
            2097 => Some(Language::Ghn),
            2098 => Some(Language::Gho),
            2099 => Some(Language::Ghr),
            2100 => Some(Language::Ghs),
            2101 => Some(Language::Ght),
            2102 => Some(Language::Gia),
            2103 => Some(Language::Gib),
            2104 => Some(Language::Gic),
            2105 => Some(Language::Gid),
            2106 => Some(Language::Gie),
            2107 => Some(Language::Gig),
            2108 => Some(Language::Gih),
            2109 => Some(Language::Gii),
            2110 => Some(Language::Gil),
            2111 => Some(Language::Gim),
            2112 => Some(Language::Gin),
            2113 => Some(Language::Gip),
            2114 => Some(Language::Giq),
            2115 => Some(Language::Gir),
            2116 => Some(Language::Gis),
            2117 => Some(Language::Git),
            2118 => Some(Language::Giu),
            2119 => Some(Language::Giw),
            2120 => Some(Language::Gix),
            2121 => Some(Language::Giy),
            2122 => Some(Language::Giz),
            2123 => Some(Language::Gjk),
            2124 => Some(Language::Gjm),
            2125 => Some(Language::Gjn),
            2126 => Some(Language::Gjr),
            2127 => Some(Language::Gju),
            2128 => Some(Language::Gka),
            2129 => Some(Language::Gkd),
            2130 => Some(Language::Gke),
            2131 => Some(Language::Gkn),
            2132 => Some(Language::Gko),
            2133 => Some(Language::Gkp),
            2134 => Some(Language::Gku),
            2135 => Some(Language::Gla),
            2136 => Some(Language::Glb),
            2137 => Some(Language::Glc),
            2138 => Some(Language::Gld),
            2139 => Some(Language::Gle),
            2140 => Some(Language::Glg),
            2141 => Some(Language::Glh),
            2142 => Some(Language::Glj),
            2143 => Some(Language::Glk),
            2144 => Some(Language::Gll),
            2145 => Some(Language::Glo),
            2146 => Some(Language::Glr),
            2147 => Some(Language::Glu),
            2148 => Some(Language::Glv),
            2149 => Some(Language::Glw),
            2150 => Some(Language::Gly),
            2151 => Some(Language::Gma),
            2152 => Some(Language::Gmb),
            2153 => Some(Language::Gmd),
            2154 => Some(Language::Gmg),
            2155 => Some(Language::Gmh),
            2156 => Some(Language::Gml),
            2157 => Some(Language::Gmm),
            2158 => Some(Language::Gmn),
            2159 => Some(Language::Gmr),
            2160 => Some(Language::Gmu),
            2161 => Some(Language::Gmv),
            2162 => Some(Language::Gmx),
            2163 => Some(Language::Gmy),
            2164 => Some(Language::Gmz),
            2165 => Some(Language::Gna),
            2166 => Some(Language::Gnb),
            2167 => Some(Language::Gnc),
            2168 => Some(Language::Gnd),
            2169 => Some(Language::Gne),
            2170 => Some(Language::Gng),
            2171 => Some(Language::Gnh),
            2172 => Some(Language::Gni),
            2173 => Some(Language::Gnj),
            2174 => Some(Language::Gnk),
            2175 => Some(Language::Gnl),
            2176 => Some(Language::Gnm),
            2177 => Some(Language::Gnn),
            2178 => Some(Language::Gno),
            2179 => Some(Language::Gnq),
            2180 => Some(Language::Gnr),
            2181 => Some(Language::Gnt),
            2182 => Some(Language::Gnu),
            2183 => Some(Language::Gnw),
            2184 => Some(Language::Gnz),
            2185 => Some(Language::Goa),
            2186 => Some(Language::Gob),
            2187 => Some(Language::Goc),
            2188 => Some(Language::God),
            2189 => Some(Language::Goe),
            2190 => Some(Language::Gof),
            2191 => Some(Language::Gog),
            2192 => Some(Language::Goh),
            2193 => Some(Language::Goi),
            2194 => Some(Language::Goj),
            2195 => Some(Language::Gok),
            2196 => Some(Language::Gol),
            2197 => Some(Language::Gom),
            2198 => Some(Language::Gon),
            2199 => Some(Language::Goo),
            2200 => Some(Language::Gop),
            2201 => Some(Language::Goq),
            2202 => Some(Language::Gor),
            2203 => Some(Language::Gos),
            2204 => Some(Language::Got),
            2205 => Some(Language::Gou),
            2206 => Some(Language::Gov),
            2207 => Some(Language::Gow),
            2208 => Some(Language::Gox),
            2209 => Some(Language::Goy),
            2210 => Some(Language::Goz),
            2211 => Some(Language::Gpa),
            2212 => Some(Language::Gpe),
            2213 => Some(Language::Gpn),
            2214 => Some(Language::Gqa),
            2215 => Some(Language::Gqi),
            2216 => Some(Language::Gqn),
            2217 => Some(Language::Gqr),
            2218 => Some(Language::Gqu),
            2219 => Some(Language::Gra),
            2220 => Some(Language::Grb),
            2221 => Some(Language::Grc),
            2222 => Some(Language::Grd),
            2223 => Some(Language::Grg),
            2224 => Some(Language::Grh),
            2225 => Some(Language::Gri),
            2226 => Some(Language::Grj),
            2227 => Some(Language::Grm),
            2228 => Some(Language::Grn),
            2229 => Some(Language::Gro),
            2230 => Some(Language::Grq),
            2231 => Some(Language::Grr),
            2232 => Some(Language::Grs),
            2233 => Some(Language::Grt),
            2234 => Some(Language::Gru),
            2235 => Some(Language::Grv),
            2236 => Some(Language::Grw),
            2237 => Some(Language::Grx),
            2238 => Some(Language::Gry),
            2239 => Some(Language::Grz),
            2240 => Some(Language::Gse),
            2241 => Some(Language::Gsg),
            2242 => Some(Language::Gsl),
            2243 => Some(Language::Gsm),
            2244 => Some(Language::Gsn),
            2245 => Some(Language::Gso),
            2246 => Some(Language::Gsp),
            2247 => Some(Language::Gss),
            2248 => Some(Language::Gsw),
            2249 => Some(Language::Gta),
            2250 => Some(Language::Gtu),
            2251 => Some(Language::Gua),
            2252 => Some(Language::Gub),
            2253 => Some(Language::Guc),
            2254 => Some(Language::Gud),
            2255 => Some(Language::Gue),
            2256 => Some(Language::Guf),
            2257 => Some(Language::Gug),
            2258 => Some(Language::Guh),
            2259 => Some(Language::Gui),
            2260 => Some(Language::Guj),
            2261 => Some(Language::Guk),
            2262 => Some(Language::Gul),
            2263 => Some(Language::Gum),
            2264 => Some(Language::Gun),
            2265 => Some(Language::Guo),
            2266 => Some(Language::Gup),
            2267 => Some(Language::Guq),
            2268 => Some(Language::Gur),
            2269 => Some(Language::Gus),
            2270 => Some(Language::Gut),
            2271 => Some(Language::Guu),
            2272 => Some(Language::Guw),
            2273 => Some(Language::Gux),
            2274 => Some(Language::Guz),
            2275 => Some(Language::Gva),
            2276 => Some(Language::Gvc),
            2277 => Some(Language::Gve),
            2278 => Some(Language::Gvf),
            2279 => Some(Language::Gvj),
            2280 => Some(Language::Gvl),
            2281 => Some(Language::Gvm),
            2282 => Some(Language::Gvn),
            2283 => Some(Language::Gvo),
            2284 => Some(Language::Gvp),
            2285 => Some(Language::Gvr),
            2286 => Some(Language::Gvs),
            2287 => Some(Language::Gvy),
            2288 => Some(Language::Gwa),
            2289 => Some(Language::Gwb),
            2290 => Some(Language::Gwc),
            2291 => Some(Language::Gwd),
            2292 => Some(Language::Gwe),
            2293 => Some(Language::Gwf),
            2294 => Some(Language::Gwg),
            2295 => Some(Language::Gwi),
            2296 => Some(Language::Gwj),
            2297 => Some(Language::Gwm),
            2298 => Some(Language::Gwn),
            2299 => Some(Language::Gwr),
            2300 => Some(Language::Gwt),
            2301 => Some(Language::Gwu),
            2302 => Some(Language::Gww),
            2303 => Some(Language::Gwx),
            2304 => Some(Language::Gxx),
            2305 => Some(Language::Gya),
            2306 => Some(Language::Gyb),
            2307 => Some(Language::Gyd),
            2308 => Some(Language::Gye),
            2309 => Some(Language::Gyf),
            2310 => Some(Language::Gyg),
            2311 => Some(Language::Gyi),
            2312 => Some(Language::Gyl),
            2313 => Some(Language::Gym),
            2314 => Some(Language::Gyn),
            2315 => Some(Language::Gyo),
            2316 => Some(Language::Gyr),
            2317 => Some(Language::Gyy),
            2318 => Some(Language::Gyz),
            2319 => Some(Language::Gza),
            2320 => Some(Language::Gzi),
            2321 => Some(Language::Gzn),
            2322 => Some(Language::Haa),
            2323 => Some(Language::Hab),
            2324 => Some(Language::Hac),
            2325 => Some(Language::Had),
            2326 => Some(Language::Hae),
            2327 => Some(Language::Haf),
            2328 => Some(Language::Hag),
            2329 => Some(Language::Hah),
            2330 => Some(Language::Hai),
            2331 => Some(Language::Haj),
            2332 => Some(Language::Hak),
            2333 => Some(Language::Hal),
            2334 => Some(Language::Ham),
            2335 => Some(Language::Han),
            2336 => Some(Language::Hao),
            2337 => Some(Language::Hap),
            2338 => Some(Language::Haq),
            2339 => Some(Language::Har),
            2340 => Some(Language::Has),
            2341 => Some(Language::Hat),
            2342 => Some(Language::Hau),
            2343 => Some(Language::Hav),
            2344 => Some(Language::Haw),
            2345 => Some(Language::Hax),
            2346 => Some(Language::Hay),
            2347 => Some(Language::Haz),
            2348 => Some(Language::Hba),
            2349 => Some(Language::Hbb),
            2350 => Some(Language::Hbn),
            2351 => Some(Language::Hbo),
            2352 => Some(Language::Hbs),
            2353 => Some(Language::Hbu),
            2354 => Some(Language::Hca),
            2355 => Some(Language::Hch),
            2356 => Some(Language::Hdn),
            2357 => Some(Language::Hds),
            2358 => Some(Language::Hdy),
            2359 => Some(Language::Hea),
            2360 => Some(Language::Heb),
            2361 => Some(Language::Hed),
            2362 => Some(Language::Heg),
            2363 => Some(Language::Heh),
            2364 => Some(Language::Hei),
            2365 => Some(Language::Hem),
            2366 => Some(Language::Her),
            2367 => Some(Language::Hgm),
            2368 => Some(Language::Hgw),
            2369 => Some(Language::Hhi),
            2370 => Some(Language::Hhr),
            2371 => Some(Language::Hhy),
            2372 => Some(Language::Hia),
            2373 => Some(Language::Hib),
            2374 => Some(Language::Hid),
            2375 => Some(Language::Hif),
            2376 => Some(Language::Hig),
            2377 => Some(Language::Hih),
            2378 => Some(Language::Hii),
            2379 => Some(Language::Hij),
            2380 => Some(Language::Hik),
            2381 => Some(Language::Hil),
            2382 => Some(Language::Hin),
            2383 => Some(Language::Hio),
            2384 => Some(Language::Hir),
            2385 => Some(Language::Hit),
            2386 => Some(Language::Hiw),
            2387 => Some(Language::Hix),
            2388 => Some(Language::Hji),
            2389 => Some(Language::Hka),
            2390 => Some(Language::Hke),
            2391 => Some(Language::Hkh),
            2392 => Some(Language::Hkk),
            2393 => Some(Language::Hkn),
            2394 => Some(Language::Hks),
            2395 => Some(Language::Hla),
            2396 => Some(Language::Hlb),
            2397 => Some(Language::Hld),
            2398 => Some(Language::Hle),
            2399 => Some(Language::Hlt),
            2400 => Some(Language::Hlu),
            2401 => Some(Language::Hma),
            2402 => Some(Language::Hmb),
            2403 => Some(Language::Hmc),
            2404 => Some(Language::Hmd),
            2405 => Some(Language::Hme),
            2406 => Some(Language::Hmf),
            2407 => Some(Language::Hmg),
            2408 => Some(Language::Hmh),
            2409 => Some(Language::Hmi),
            2410 => Some(Language::Hmj),
            2411 => Some(Language::Hmk),
            2412 => Some(Language::Hml),
            2413 => Some(Language::Hmm),
            2414 => Some(Language::Hmn),
            2415 => Some(Language::Hmo),
            2416 => Some(Language::Hmp),
            2417 => Some(Language::Hmq),
            2418 => Some(Language::Hmr),
            2419 => Some(Language::Hms),
            2420 => Some(Language::Hmt),
            2421 => Some(Language::Hmu),
            2422 => Some(Language::Hmv),
            2423 => Some(Language::Hmw),
            2424 => Some(Language::Hmy),
            2425 => Some(Language::Hmz),
            2426 => Some(Language::Hna),
            2427 => Some(Language::Hnd),
            2428 => Some(Language::Hne),
            2429 => Some(Language::Hng),
            2430 => Some(Language::Hnh),
            2431 => Some(Language::Hni),
            2432 => Some(Language::Hnj),
            2433 => Some(Language::Hnn),
            2434 => Some(Language::Hno),
            2435 => Some(Language::Hns),
            2436 => Some(Language::Hnu),
            2437 => Some(Language::Hoa),
            2438 => Some(Language::Hob),
            2439 => Some(Language::Hoc),
            2440 => Some(Language::Hod),
            2441 => Some(Language::Hoe),
            2442 => Some(Language::Hoh),
            2443 => Some(Language::Hoi),
            2444 => Some(Language::Hoj),
            2445 => Some(Language::Hol),
            2446 => Some(Language::Hom),
            2447 => Some(Language::Hoo),
            2448 => Some(Language::Hop),
            2449 => Some(Language::Hor),
            2450 => Some(Language::Hos),
            2451 => Some(Language::Hot),
            2452 => Some(Language::Hov),
            2453 => Some(Language::How),
            2454 => Some(Language::Hoy),
            2455 => Some(Language::Hoz),
            2456 => Some(Language::Hpo),
            2457 => Some(Language::Hps),
            2458 => Some(Language::Hra),
            2459 => Some(Language::Hrc),
            2460 => Some(Language::Hre),
            2461 => Some(Language::Hrk),
            2462 => Some(Language::Hrm),
            2463 => Some(Language::Hro),
            2464 => Some(Language::Hrp),
            2465 => Some(Language::Hrt),
            2466 => Some(Language::Hru),
            2467 => Some(Language::Hrv),
            2468 => Some(Language::Hrw),
            2469 => Some(Language::Hrx),
            2470 => Some(Language::Hrz),
            2471 => Some(Language::Hsb),
            2472 => Some(Language::Hsh),
            2473 => Some(Language::Hsl),
            2474 => Some(Language::Hsn),
            2475 => Some(Language::Hss),
            2476 => Some(Language::Hti),
            2477 => Some(Language::Hto),
            2478 => Some(Language::Hts),
            2479 => Some(Language::Htu),
            2480 => Some(Language::Htx),
            2481 => Some(Language::Hub),
            2482 => Some(Language::Huc),
            2483 => Some(Language::Hud),
            2484 => Some(Language::Hue),
            2485 => Some(Language::Huf),
            2486 => Some(Language::Hug),
            2487 => Some(Language::Huh),
            2488 => Some(Language::Hui),
            2489 => Some(Language::Huj),
            2490 => Some(Language::Huk),
            2491 => Some(Language::Hul),
            2492 => Some(Language::Hum),
            2493 => Some(Language::Hun),
            2494 => Some(Language::Huo),
            2495 => Some(Language::Hup),
            2496 => Some(Language::Huq),
            2497 => Some(Language::Hur),
            2498 => Some(Language::Hus),
            2499 => Some(Language::Hut),
            2500 => Some(Language::Huu),
            2501 => Some(Language::Huv),
            2502 => Some(Language::Huw),
            2503 => Some(Language::Hux),
            2504 => Some(Language::Huy),
            2505 => Some(Language::Huz),
            2506 => Some(Language::Hvc),
            2507 => Some(Language::Hve),
            2508 => Some(Language::Hvk),
            2509 => Some(Language::Hvn),
            2510 => Some(Language::Hvv),
            2511 => Some(Language::Hwa),
            2512 => Some(Language::Hwc),
            2513 => Some(Language::Hwo),
            2514 => Some(Language::Hya),
            2515 => Some(Language::Hye),
            2516 => Some(Language::Hyw),
            2517 => Some(Language::Iai),
            2518 => Some(Language::Ian),
            2519 => Some(Language::Iar),
            2520 => Some(Language::Iba),
            2521 => Some(Language::Ibb),
            2522 => Some(Language::Ibd),
            2523 => Some(Language::Ibe),
            2524 => Some(Language::Ibg),
            2525 => Some(Language::Ibh),
            2526 => Some(Language::Ibl),
            2527 => Some(Language::Ibm),
            2528 => Some(Language::Ibn),
            2529 => Some(Language::Ibo),
            2530 => Some(Language::Ibr),
            2531 => Some(Language::Ibu),
            2532 => Some(Language::Iby),
            2533 => Some(Language::Ica),
            2534 => Some(Language::Ich),
            2535 => Some(Language::Icl),
            2536 => Some(Language::Icr),
            2537 => Some(Language::Ida),
            2538 => Some(Language::Idb),
            2539 => Some(Language::Idc),
            2540 => Some(Language::Idd),
            2541 => Some(Language::Ide),
            2542 => Some(Language::Idi),
            2543 => Some(Language::Ido),
            2544 => Some(Language::Idr),
            2545 => Some(Language::Ids),
            2546 => Some(Language::Idt),
            2547 => Some(Language::Idu),
            2548 => Some(Language::Ifa),
            2549 => Some(Language::Ifb),
            2550 => Some(Language::Ife),
            2551 => Some(Language::Iff),
            2552 => Some(Language::Ifk),
            2553 => Some(Language::Ifm),
            2554 => Some(Language::Ifu),
            2555 => Some(Language::Ify),
            2556 => Some(Language::Igb),
            2557 => Some(Language::Ige),
            2558 => Some(Language::Igg),
            2559 => Some(Language::Igl),
            2560 => Some(Language::Igm),
            2561 => Some(Language::Ign),
            2562 => Some(Language::Igo),
            2563 => Some(Language::Igs),
            2564 => Some(Language::Igw),
            2565 => Some(Language::Ihb),
            2566 => Some(Language::Ihi),
            2567 => Some(Language::Ihp),
            2568 => Some(Language::Ihw),
            2569 => Some(Language::Iii),
            2570 => Some(Language::Iin),
            2571 => Some(Language::Ijc),
            2572 => Some(Language::Ije),
            2573 => Some(Language::Ijj),
            2574 => Some(Language::Ijn),
            2575 => Some(Language::Ijs),
            2576 => Some(Language::Ike),
            2577 => Some(Language::Iki),
            2578 => Some(Language::Ikk),
            2579 => Some(Language::Ikl),
            2580 => Some(Language::Iko),
            2581 => Some(Language::Ikp),
            2582 => Some(Language::Ikr),
            2583 => Some(Language::Iks),
            2584 => Some(Language::Ikt),
            2585 => Some(Language::Iku),
            2586 => Some(Language::Ikv),
            2587 => Some(Language::Ikw),
            2588 => Some(Language::Ikx),
            2589 => Some(Language::Ikz),
            2590 => Some(Language::Ila),
            2591 => Some(Language::Ilb),
            2592 => Some(Language::Ile),
            2593 => Some(Language::Ilg),
            2594 => Some(Language::Ili),
            2595 => Some(Language::Ilk),
            2596 => Some(Language::Ilm),
            2597 => Some(Language::Ilo),
            2598 => Some(Language::Ilp),
            2599 => Some(Language::Ils),
            2600 => Some(Language::Ilu),
            2601 => Some(Language::Ilv),
            2602 => Some(Language::Ima),
            2603 => Some(Language::Imi),
            2604 => Some(Language::Iml),
            2605 => Some(Language::Imn),
            2606 => Some(Language::Imo),
            2607 => Some(Language::Imr),
            2608 => Some(Language::Ims),
            2609 => Some(Language::Imt),
            2610 => Some(Language::Imy),
            2611 => Some(Language::Ina),
            2612 => Some(Language::Inb),
            2613 => Some(Language::Ind),
            2614 => Some(Language::Ing),
            2615 => Some(Language::Inh),
            2616 => Some(Language::Inj),
            2617 => Some(Language::Inl),
            2618 => Some(Language::Inm),
            2619 => Some(Language::Inn),
            2620 => Some(Language::Ino),
            2621 => Some(Language::Inp),
            2622 => Some(Language::Ins),
            2623 => Some(Language::Int),
            2624 => Some(Language::Inz),
            2625 => Some(Language::Ior),
            2626 => Some(Language::Iou),
            2627 => Some(Language::Iow),
            2628 => Some(Language::Ipi),
            2629 => Some(Language::Ipk),
            2630 => Some(Language::Ipo),
            2631 => Some(Language::Iqu),
            2632 => Some(Language::Iqw),
            2633 => Some(Language::Ire),
            2634 => Some(Language::Irh),
            2635 => Some(Language::Iri),
            2636 => Some(Language::Irk),
            2637 => Some(Language::Irn),
            2638 => Some(Language::Irr),
            2639 => Some(Language::Iru),
            2640 => Some(Language::Irx),
            2641 => Some(Language::Iry),
            2642 => Some(Language::Isa),
            2643 => Some(Language::Isc),
            2644 => Some(Language::Isd),
            2645 => Some(Language::Ise),
            2646 => Some(Language::Isg),
            2647 => Some(Language::Ish),
            2648 => Some(Language::Isi),
            2649 => Some(Language::Isk),
            2650 => Some(Language::Isl),
            2651 => Some(Language::Ism),
            2652 => Some(Language::Isn),
            2653 => Some(Language::Iso),
            2654 => Some(Language::Isr),
            2655 => Some(Language::Ist),
            2656 => Some(Language::Isu),
            2657 => Some(Language::Ita),
            2658 => Some(Language::Itb),
            2659 => Some(Language::Itd),
            2660 => Some(Language::Ite),
            2661 => Some(Language::Iti),
            2662 => Some(Language::Itk),
            2663 => Some(Language::Itl),
            2664 => Some(Language::Itm),
            2665 => Some(Language::Ito),
            2666 => Some(Language::Itr),
            2667 => Some(Language::Its),
            2668 => Some(Language::Itt),
            2669 => Some(Language::Itv),
            2670 => Some(Language::Itw),
            2671 => Some(Language::Itx),
            2672 => Some(Language::Ity),
            2673 => Some(Language::Itz),
            2674 => Some(Language::Ium),
            2675 => Some(Language::Ivb),
            2676 => Some(Language::Ivv),
            2677 => Some(Language::Iwk),
            2678 => Some(Language::Iwm),
            2679 => Some(Language::Iwo),
            2680 => Some(Language::Iws),
            2681 => Some(Language::Ixc),
            2682 => Some(Language::Ixl),
            2683 => Some(Language::Iya),
            2684 => Some(Language::Iyo),
            2685 => Some(Language::Iyx),
            2686 => Some(Language::Izh),
            2687 => Some(Language::Izr),
            2688 => Some(Language::Izz),
            2689 => Some(Language::Jaa),
            2690 => Some(Language::Jab),
            2691 => Some(Language::Jac),
            2692 => Some(Language::Jad),
            2693 => Some(Language::Jae),
            2694 => Some(Language::Jaf),
            2695 => Some(Language::Jah),
            2696 => Some(Language::Jaj),
            2697 => Some(Language::Jak),
            2698 => Some(Language::Jal),
            2699 => Some(Language::Jam),
            2700 => Some(Language::Jan),
            2701 => Some(Language::Jao),
            2702 => Some(Language::Jaq),
            2703 => Some(Language::Jas),
            2704 => Some(Language::Jat),
            2705 => Some(Language::Jau),
            2706 => Some(Language::Jav),
            2707 => Some(Language::Jax),
            2708 => Some(Language::Jay),
            2709 => Some(Language::Jaz),
            2710 => Some(Language::Jbe),
            2711 => Some(Language::Jbi),
            2712 => Some(Language::Jbj),
            2713 => Some(Language::Jbk),
            2714 => Some(Language::Jbm),
            2715 => Some(Language::Jbn),
            2716 => Some(Language::Jbo),
            2717 => Some(Language::Jbr),
            2718 => Some(Language::Jbt),
            2719 => Some(Language::Jbu),
            2720 => Some(Language::Jbw),
            2721 => Some(Language::Jcs),
            2722 => Some(Language::Jct),
            2723 => Some(Language::Jda),
            2724 => Some(Language::Jdg),
            2725 => Some(Language::Jdt),
            2726 => Some(Language::Jeb),
            2727 => Some(Language::Jee),
            2728 => Some(Language::Jeh),
            2729 => Some(Language::Jei),
            2730 => Some(Language::Jek),
            2731 => Some(Language::Jel),
            2732 => Some(Language::Jen),
            2733 => Some(Language::Jer),
            2734 => Some(Language::Jet),
            2735 => Some(Language::Jeu),
            2736 => Some(Language::Jgb),
            2737 => Some(Language::Jge),
            2738 => Some(Language::Jgk),
            2739 => Some(Language::Jgo),
            2740 => Some(Language::Jhi),
            2741 => Some(Language::Jhs),
            2742 => Some(Language::Jia),
            2743 => Some(Language::Jib),
            2744 => Some(Language::Jic),
            2745 => Some(Language::Jid),
            2746 => Some(Language::Jie),
            2747 => Some(Language::Jig),
            2748 => Some(Language::Jih),
            2749 => Some(Language::Jii),
            2750 => Some(Language::Jil),
            2751 => Some(Language::Jim),
            2752 => Some(Language::Jio),
            2753 => Some(Language::Jiq),
            2754 => Some(Language::Jit),
            2755 => Some(Language::Jiu),
            2756 => Some(Language::Jiv),
            2757 => Some(Language::Jiy),
            2758 => Some(Language::Jje),
            2759 => Some(Language::Jjr),
            2760 => Some(Language::Jka),
            2761 => Some(Language::Jkm),
            2762 => Some(Language::Jko),
            2763 => Some(Language::Jkp),
            2764 => Some(Language::Jkr),
            2765 => Some(Language::Jks),
            2766 => Some(Language::Jku),
            2767 => Some(Language::Jle),
            2768 => Some(Language::Jls),
            2769 => Some(Language::Jma),
            2770 => Some(Language::Jmb),
            2771 => Some(Language::Jmc),
            2772 => Some(Language::Jmd),
            2773 => Some(Language::Jmi),
            2774 => Some(Language::Jml),
            2775 => Some(Language::Jmn),
            2776 => Some(Language::Jmr),
            2777 => Some(Language::Jms),
            2778 => Some(Language::Jmw),
            2779 => Some(Language::Jmx),
            2780 => Some(Language::Jna),
            2781 => Some(Language::Jnd),
            2782 => Some(Language::Jng),
            2783 => Some(Language::Jni),
            2784 => Some(Language::Jnj),
            2785 => Some(Language::Jnl),
            2786 => Some(Language::Jns),
            2787 => Some(Language::Job),
            2788 => Some(Language::Jod),
            2789 => Some(Language::Jog),
            2790 => Some(Language::Jor),
            2791 => Some(Language::Jos),
            2792 => Some(Language::Jow),
            2793 => Some(Language::Jpa),
            2794 => Some(Language::Jpn),
            2795 => Some(Language::Jpr),
            2796 => Some(Language::Jqr),
            2797 => Some(Language::Jra),
            2798 => Some(Language::Jrb),
            2799 => Some(Language::Jrr),
            2800 => Some(Language::Jrt),
            2801 => Some(Language::Jru),
            2802 => Some(Language::Jsl),
            2803 => Some(Language::Jua),
            2804 => Some(Language::Jub),
            2805 => Some(Language::Juc),
            2806 => Some(Language::Jud),
            2807 => Some(Language::Juh),
            2808 => Some(Language::Jui),
            2809 => Some(Language::Juk),
            2810 => Some(Language::Jul),
            2811 => Some(Language::Jum),
            2812 => Some(Language::Jun),
            2813 => Some(Language::Juo),
            2814 => Some(Language::Jup),
            2815 => Some(Language::Jur),
            2816 => Some(Language::Jus),
            2817 => Some(Language::Jut),
            2818 => Some(Language::Juu),
            2819 => Some(Language::Juw),
            2820 => Some(Language::Juy),
            2821 => Some(Language::Jvd),
            2822 => Some(Language::Jvn),
            2823 => Some(Language::Jwi),
            2824 => Some(Language::Jya),
            2825 => Some(Language::Jye),
            2826 => Some(Language::Jyy),
            2827 => Some(Language::Kaa),
            2828 => Some(Language::Kab),
            2829 => Some(Language::Kac),
            2830 => Some(Language::Kad),
            2831 => Some(Language::Kae),
            2832 => Some(Language::Kaf),
            2833 => Some(Language::Kag),
            2834 => Some(Language::Kah),
            2835 => Some(Language::Kai),
            2836 => Some(Language::Kaj),
            2837 => Some(Language::Kak),
            2838 => Some(Language::Kal),
            2839 => Some(Language::Kam),
            2840 => Some(Language::Kan),
            2841 => Some(Language::Kao),
            2842 => Some(Language::Kap),
            2843 => Some(Language::Kaq),
            2844 => Some(Language::Kas),
            2845 => Some(Language::Kat),
            2846 => Some(Language::Kau),
            2847 => Some(Language::Kav),
            2848 => Some(Language::Kaw),
            2849 => Some(Language::Kax),
            2850 => Some(Language::Kay),
            2851 => Some(Language::Kaz),
            2852 => Some(Language::Kba),
            2853 => Some(Language::Kbb),
            2854 => Some(Language::Kbc),
            2855 => Some(Language::Kbd),
            2856 => Some(Language::Kbe),
            2857 => Some(Language::Kbg),
            2858 => Some(Language::Kbh),
            2859 => Some(Language::Kbi),
            2860 => Some(Language::Kbj),
            2861 => Some(Language::Kbk),
            2862 => Some(Language::Kbl),
            2863 => Some(Language::Kbm),
            2864 => Some(Language::Kbn),
            2865 => Some(Language::Kbo),
            2866 => Some(Language::Kbp),
            2867 => Some(Language::Kbq),
            2868 => Some(Language::Kbr),
            2869 => Some(Language::Kbs),
            2870 => Some(Language::Kbt),
            2871 => Some(Language::Kbu),
            2872 => Some(Language::Kbv),
            2873 => Some(Language::Kbw),
            2874 => Some(Language::Kbx),
            2875 => Some(Language::Kby),
            2876 => Some(Language::Kbz),
            2877 => Some(Language::Kca),
            2878 => Some(Language::Kcb),
            2879 => Some(Language::Kcc),
            2880 => Some(Language::Kcd),
            2881 => Some(Language::Kce),
            2882 => Some(Language::Kcf),
            2883 => Some(Language::Kcg),
            2884 => Some(Language::Kch),
            2885 => Some(Language::Kci),
            2886 => Some(Language::Kcj),
            2887 => Some(Language::Kck),
            2888 => Some(Language::Kcl),
            2889 => Some(Language::Kcm),
            2890 => Some(Language::Kcn),
            2891 => Some(Language::Kco),
            2892 => Some(Language::Kcp),
            2893 => Some(Language::Kcq),
            2894 => Some(Language::Kcr),
            2895 => Some(Language::Kcs),
            2896 => Some(Language::Kct),
            2897 => Some(Language::Kcu),
            2898 => Some(Language::Kcv),
            2899 => Some(Language::Kcw),
            2900 => Some(Language::Kcx),
            2901 => Some(Language::Kcy),
            2902 => Some(Language::Kcz),
            2903 => Some(Language::Kda),
            2904 => Some(Language::Kdc),
            2905 => Some(Language::Kdd),
            2906 => Some(Language::Kde),
            2907 => Some(Language::Kdf),
            2908 => Some(Language::Kdg),
            2909 => Some(Language::Kdh),
            2910 => Some(Language::Kdi),
            2911 => Some(Language::Kdj),
            2912 => Some(Language::Kdk),
            2913 => Some(Language::Kdl),
            2914 => Some(Language::Kdm),
            2915 => Some(Language::Kdn),
            2916 => Some(Language::Kdp),
            2917 => Some(Language::Kdq),
            2918 => Some(Language::Kdr),
            2919 => Some(Language::Kdt),
            2920 => Some(Language::Kdu),
            2921 => Some(Language::Kdw),
            2922 => Some(Language::Kdx),
            2923 => Some(Language::Kdy),
            2924 => Some(Language::Kdz),
            2925 => Some(Language::Kea),
            2926 => Some(Language::Keb),
            2927 => Some(Language::Kec),
            2928 => Some(Language::Ked),
            2929 => Some(Language::Kee),
            2930 => Some(Language::Kef),
            2931 => Some(Language::Keg),
            2932 => Some(Language::Keh),
            2933 => Some(Language::Kei),
            2934 => Some(Language::Kej),
            2935 => Some(Language::Kek),
            2936 => Some(Language::Kel),
            2937 => Some(Language::Kem),
            2938 => Some(Language::Ken),
            2939 => Some(Language::Keo),
            2940 => Some(Language::Kep),
            2941 => Some(Language::Keq),
            2942 => Some(Language::Ker),
            2943 => Some(Language::Kes),
            2944 => Some(Language::Ket),
            2945 => Some(Language::Keu),
            2946 => Some(Language::Kev),
            2947 => Some(Language::Kew),
            2948 => Some(Language::Kex),
            2949 => Some(Language::Key),
            2950 => Some(Language::Kez),
            2951 => Some(Language::Kfa),
            2952 => Some(Language::Kfb),
            2953 => Some(Language::Kfc),
            2954 => Some(Language::Kfd),
            2955 => Some(Language::Kfe),
            2956 => Some(Language::Kff),
            2957 => Some(Language::Kfg),
            2958 => Some(Language::Kfh),
            2959 => Some(Language::Kfi),
            2960 => Some(Language::Kfj),
            2961 => Some(Language::Kfk),
            2962 => Some(Language::Kfl),
            2963 => Some(Language::Kfm),
            2964 => Some(Language::Kfn),
            2965 => Some(Language::Kfo),
            2966 => Some(Language::Kfp),
            2967 => Some(Language::Kfq),
            2968 => Some(Language::Kfr),
            2969 => Some(Language::Kfs),
            2970 => Some(Language::Kft),
            2971 => Some(Language::Kfu),
            2972 => Some(Language::Kfv),
            2973 => Some(Language::Kfw),
            2974 => Some(Language::Kfx),
            2975 => Some(Language::Kfy),
            2976 => Some(Language::Kfz),
            2977 => Some(Language::Kga),
            2978 => Some(Language::Kgb),
            2979 => Some(Language::Kge),
            2980 => Some(Language::Kgf),
            2981 => Some(Language::Kgg),
            2982 => Some(Language::Kgi),
            2983 => Some(Language::Kgj),
            2984 => Some(Language::Kgk),
            2985 => Some(Language::Kgl),
            2986 => Some(Language::Kgm),
            2987 => Some(Language::Kgn),
            2988 => Some(Language::Kgo),
            2989 => Some(Language::Kgp),
            2990 => Some(Language::Kgq),
            2991 => Some(Language::Kgr),
            2992 => Some(Language::Kgs),
            2993 => Some(Language::Kgt),
            2994 => Some(Language::Kgu),
            2995 => Some(Language::Kgv),
            2996 => Some(Language::Kgw),
            2997 => Some(Language::Kgx),
            2998 => Some(Language::Kgy),
            2999 => Some(Language::Kha),
            3000 => Some(Language::Khb),
            3001 => Some(Language::Khc),
            3002 => Some(Language::Khd),
            3003 => Some(Language::Khe),
            3004 => Some(Language::Khf),
            3005 => Some(Language::Khg),
            3006 => Some(Language::Khh),
            3007 => Some(Language::Khj),
            3008 => Some(Language::Khk),
            3009 => Some(Language::Khl),
            3010 => Some(Language::Khm),
            3011 => Some(Language::Khn),
            3012 => Some(Language::Kho),
            3013 => Some(Language::Khp),
            3014 => Some(Language::Khq),
            3015 => Some(Language::Khr),
            3016 => Some(Language::Khs),
            3017 => Some(Language::Kht),
            3018 => Some(Language::Khu),
            3019 => Some(Language::Khv),
            3020 => Some(Language::Khw),
            3021 => Some(Language::Khx),
            3022 => Some(Language::Khy),
            3023 => Some(Language::Khz),
            3024 => Some(Language::Kia),
            3025 => Some(Language::Kib),
            3026 => Some(Language::Kic),
            3027 => Some(Language::Kid),
            3028 => Some(Language::Kie),
            3029 => Some(Language::Kif),
            3030 => Some(Language::Kig),
            3031 => Some(Language::Kih),
            3032 => Some(Language::Kii),
            3033 => Some(Language::Kij),
            3034 => Some(Language::Kik),
            3035 => Some(Language::Kil),
            3036 => Some(Language::Kim),
            3037 => Some(Language::Kin),
            3038 => Some(Language::Kio),
            3039 => Some(Language::Kip),
            3040 => Some(Language::Kiq),
            3041 => Some(Language::Kir),
            3042 => Some(Language::Kis),
            3043 => Some(Language::Kit),
            3044 => Some(Language::Kiu),
            3045 => Some(Language::Kiv),
            3046 => Some(Language::Kiw),
            3047 => Some(Language::Kix),
            3048 => Some(Language::Kiy),
            3049 => Some(Language::Kiz),
            3050 => Some(Language::Kja),
            3051 => Some(Language::Kjb),
            3052 => Some(Language::Kjc),
            3053 => Some(Language::Kjd),
            3054 => Some(Language::Kje),
            3055 => Some(Language::Kjg),
            3056 => Some(Language::Kjh),
            3057 => Some(Language::Kji),
            3058 => Some(Language::Kjj),
            3059 => Some(Language::Kjk),
            3060 => Some(Language::Kjl),
            3061 => Some(Language::Kjm),
            3062 => Some(Language::Kjn),
            3063 => Some(Language::Kjo),
            3064 => Some(Language::Kjp),
            3065 => Some(Language::Kjq),
            3066 => Some(Language::Kjr),
            3067 => Some(Language::Kjs),
            3068 => Some(Language::Kjt),
            3069 => Some(Language::Kju),
            3070 => Some(Language::Kjv),
            3071 => Some(Language::Kjx),
            3072 => Some(Language::Kjy),
            3073 => Some(Language::Kjz),
            3074 => Some(Language::Kka),
            3075 => Some(Language::Kkb),
            3076 => Some(Language::Kkc),
            3077 => Some(Language::Kkd),
            3078 => Some(Language::Kke),
            3079 => Some(Language::Kkf),
            3080 => Some(Language::Kkg),
            3081 => Some(Language::Kkh),
            3082 => Some(Language::Kki),
            3083 => Some(Language::Kkj),
            3084 => Some(Language::Kkk),
            3085 => Some(Language::Kkl),
            3086 => Some(Language::Kkm),
            3087 => Some(Language::Kkn),
            3088 => Some(Language::Kko),
            3089 => Some(Language::Kkp),
            3090 => Some(Language::Kkq),
            3091 => Some(Language::Kkr),
            3092 => Some(Language::Kks),
            3093 => Some(Language::Kkt),
            3094 => Some(Language::Kku),
            3095 => Some(Language::Kkv),
            3096 => Some(Language::Kkw),
            3097 => Some(Language::Kkx),
            3098 => Some(Language::Kky),
            3099 => Some(Language::Kkz),
            3100 => Some(Language::Kla),
            3101 => Some(Language::Klb),
            3102 => Some(Language::Klc),
            3103 => Some(Language::Kld),
            3104 => Some(Language::Kle),
            3105 => Some(Language::Klf),
            3106 => Some(Language::Klg),
            3107 => Some(Language::Klh),
            3108 => Some(Language::Kli),
            3109 => Some(Language::Klj),
            3110 => Some(Language::Klk),
            3111 => Some(Language::Kll),
            3112 => Some(Language::Klm),
            3113 => Some(Language::Kln),
            3114 => Some(Language::Klo),
            3115 => Some(Language::Klp),
            3116 => Some(Language::Klq),
            3117 => Some(Language::Klr),
            3118 => Some(Language::Kls),
            3119 => Some(Language::Klt),
            3120 => Some(Language::Klu),
            3121 => Some(Language::Klv),
            3122 => Some(Language::Klw),
            3123 => Some(Language::Klx),
            3124 => Some(Language::Kly),
            3125 => Some(Language::Klz),
            3126 => Some(Language::Kma),
            3127 => Some(Language::Kmb),
            3128 => Some(Language::Kmc),
            3129 => Some(Language::Kmd),
            3130 => Some(Language::Kme),
            3131 => Some(Language::Kmf),
            3132 => Some(Language::Kmg),
            3133 => Some(Language::Kmh),
            3134 => Some(Language::Kmi),
            3135 => Some(Language::Kmj),
            3136 => Some(Language::Kmk),
            3137 => Some(Language::Kml),
            3138 => Some(Language::Kmm),
            3139 => Some(Language::Kmn),
            3140 => Some(Language::Kmo),
            3141 => Some(Language::Kmp),
            3142 => Some(Language::Kmq),
            3143 => Some(Language::Kmr),
            3144 => Some(Language::Kms),
            3145 => Some(Language::Kmt),
            3146 => Some(Language::Kmu),
            3147 => Some(Language::Kmv),
            3148 => Some(Language::Kmw),
            3149 => Some(Language::Kmx),
            3150 => Some(Language::Kmy),
            3151 => Some(Language::Kmz),
            3152 => Some(Language::Kna),
            3153 => Some(Language::Knb),
            3154 => Some(Language::Knc),
            3155 => Some(Language::Knd),
            3156 => Some(Language::Kne),
            3157 => Some(Language::Knf),
            3158 => Some(Language::Kng),
            3159 => Some(Language::Kni),
            3160 => Some(Language::Knj),
            3161 => Some(Language::Knk),
            3162 => Some(Language::Knl),
            3163 => Some(Language::Knm),
            3164 => Some(Language::Knn),
            3165 => Some(Language::Kno),
            3166 => Some(Language::Knp),
            3167 => Some(Language::Knq),
            3168 => Some(Language::Knr),
            3169 => Some(Language::Kns),
            3170 => Some(Language::Knt),
            3171 => Some(Language::Knu),
            3172 => Some(Language::Knv),
            3173 => Some(Language::Knw),
            3174 => Some(Language::Knx),
            3175 => Some(Language::Kny),
            3176 => Some(Language::Knz),
            3177 => Some(Language::Koa),
            3178 => Some(Language::Koc),
            3179 => Some(Language::Kod),
            3180 => Some(Language::Koe),
            3181 => Some(Language::Kof),
            3182 => Some(Language::Kog),
            3183 => Some(Language::Koh),
            3184 => Some(Language::Koi),
            3185 => Some(Language::Kok),
            3186 => Some(Language::Kol),
            3187 => Some(Language::Kom),
            3188 => Some(Language::Kon),
            3189 => Some(Language::Koo),
            3190 => Some(Language::Kop),
            3191 => Some(Language::Koq),
            3192 => Some(Language::Kor),
            3193 => Some(Language::Kos),
            3194 => Some(Language::Kot),
            3195 => Some(Language::Kou),
            3196 => Some(Language::Kov),
            3197 => Some(Language::Kow),
            3198 => Some(Language::Koy),
            3199 => Some(Language::Koz),
            3200 => Some(Language::Kpa),
            3201 => Some(Language::Kpb),
            3202 => Some(Language::Kpc),
            3203 => Some(Language::Kpd),
            3204 => Some(Language::Kpe),
            3205 => Some(Language::Kpf),
            3206 => Some(Language::Kpg),
            3207 => Some(Language::Kph),
            3208 => Some(Language::Kpi),
            3209 => Some(Language::Kpj),
            3210 => Some(Language::Kpk),
            3211 => Some(Language::Kpl),
            3212 => Some(Language::Kpm),
            3213 => Some(Language::Kpn),
            3214 => Some(Language::Kpo),
            3215 => Some(Language::Kpq),
            3216 => Some(Language::Kpr),
            3217 => Some(Language::Kps),
            3218 => Some(Language::Kpt),
            3219 => Some(Language::Kpu),
            3220 => Some(Language::Kpv),
            3221 => Some(Language::Kpw),
            3222 => Some(Language::Kpx),
            3223 => Some(Language::Kpy),
            3224 => Some(Language::Kpz),
            3225 => Some(Language::Kqa),
            3226 => Some(Language::Kqb),
            3227 => Some(Language::Kqc),
            3228 => Some(Language::Kqd),
            3229 => Some(Language::Kqe),
            3230 => Some(Language::Kqf),
            3231 => Some(Language::Kqg),
            3232 => Some(Language::Kqh),
            3233 => Some(Language::Kqi),
            3234 => Some(Language::Kqj),
            3235 => Some(Language::Kqk),
            3236 => Some(Language::Kql),
            3237 => Some(Language::Kqm),
            3238 => Some(Language::Kqn),
            3239 => Some(Language::Kqo),
            3240 => Some(Language::Kqp),
            3241 => Some(Language::Kqq),
            3242 => Some(Language::Kqr),
            3243 => Some(Language::Kqs),
            3244 => Some(Language::Kqt),
            3245 => Some(Language::Kqu),
            3246 => Some(Language::Kqv),
            3247 => Some(Language::Kqw),
            3248 => Some(Language::Kqx),
            3249 => Some(Language::Kqy),
            3250 => Some(Language::Kqz),
            3251 => Some(Language::Kra),
            3252 => Some(Language::Krb),
            3253 => Some(Language::Krc),
            3254 => Some(Language::Krd),
            3255 => Some(Language::Kre),
            3256 => Some(Language::Krf),
            3257 => Some(Language::Krh),
            3258 => Some(Language::Kri),
            3259 => Some(Language::Krj),
            3260 => Some(Language::Krk),
            3261 => Some(Language::Krl),
            3262 => Some(Language::Krn),
            3263 => Some(Language::Krp),
            3264 => Some(Language::Krr),
            3265 => Some(Language::Krs),
            3266 => Some(Language::Krt),
            3267 => Some(Language::Kru),
            3268 => Some(Language::Krv),
            3269 => Some(Language::Krw),
            3270 => Some(Language::Krx),
            3271 => Some(Language::Kry),
            3272 => Some(Language::Krz),
            3273 => Some(Language::Ksa),
            3274 => Some(Language::Ksb),
            3275 => Some(Language::Ksc),
            3276 => Some(Language::Ksd),
            3277 => Some(Language::Kse),
            3278 => Some(Language::Ksf),
            3279 => Some(Language::Ksg),
            3280 => Some(Language::Ksh),
            3281 => Some(Language::Ksi),
            3282 => Some(Language::Ksj),
            3283 => Some(Language::Ksk),
            3284 => Some(Language::Ksl),
            3285 => Some(Language::Ksm),
            3286 => Some(Language::Ksn),
            3287 => Some(Language::Kso),
            3288 => Some(Language::Ksp),
            3289 => Some(Language::Ksq),
            3290 => Some(Language::Ksr),
            3291 => Some(Language::Kss),
            3292 => Some(Language::Kst),
            3293 => Some(Language::Ksu),
            3294 => Some(Language::Ksv),
            3295 => Some(Language::Ksw),
            3296 => Some(Language::Ksx),
            3297 => Some(Language::Ksy),
            3298 => Some(Language::Ksz),
            3299 => Some(Language::Kta),
            3300 => Some(Language::Ktb),
            3301 => Some(Language::Ktc),
            3302 => Some(Language::Ktd),
            3303 => Some(Language::Kte),
            3304 => Some(Language::Ktf),
            3305 => Some(Language::Ktg),
            3306 => Some(Language::Kth),
            3307 => Some(Language::Kti),
            3308 => Some(Language::Ktj),
            3309 => Some(Language::Ktk),
            3310 => Some(Language::Ktl),
            3311 => Some(Language::Ktm),
            3312 => Some(Language::Ktn),
            3313 => Some(Language::Kto),
            3314 => Some(Language::Ktp),
            3315 => Some(Language::Ktq),
            3316 => Some(Language::Kts),
            3317 => Some(Language::Ktt),
            3318 => Some(Language::Ktu),
            3319 => Some(Language::Ktv),
            3320 => Some(Language::Ktw),
            3321 => Some(Language::Ktx),
            3322 => Some(Language::Kty),
            3323 => Some(Language::Ktz),
            3324 => Some(Language::Kua),
            3325 => Some(Language::Kub),
            3326 => Some(Language::Kuc),
            3327 => Some(Language::Kud),
            3328 => Some(Language::Kue),
            3329 => Some(Language::Kuf),
            3330 => Some(Language::Kug),
            3331 => Some(Language::Kuh),
            3332 => Some(Language::Kui),
            3333 => Some(Language::Kuj),
            3334 => Some(Language::Kuk),
            3335 => Some(Language::Kul),
            3336 => Some(Language::Kum),
            3337 => Some(Language::Kun),
            3338 => Some(Language::Kuo),
            3339 => Some(Language::Kup),
            3340 => Some(Language::Kuq),
            3341 => Some(Language::Kur),
            3342 => Some(Language::Kus),
            3343 => Some(Language::Kut),
            3344 => Some(Language::Kuu),
            3345 => Some(Language::Kuv),
            3346 => Some(Language::Kuw),
            3347 => Some(Language::Kux),
            3348 => Some(Language::Kuy),
            3349 => Some(Language::Kuz),
            3350 => Some(Language::Kva),
            3351 => Some(Language::Kvb),
            3352 => Some(Language::Kvc),
            3353 => Some(Language::Kvd),
            3354 => Some(Language::Kve),
            3355 => Some(Language::Kvf),
            3356 => Some(Language::Kvg),
            3357 => Some(Language::Kvh),
            3358 => Some(Language::Kvi),
            3359 => Some(Language::Kvj),
            3360 => Some(Language::Kvk),
            3361 => Some(Language::Kvl),
            3362 => Some(Language::Kvm),
            3363 => Some(Language::Kvn),
            3364 => Some(Language::Kvo),
            3365 => Some(Language::Kvp),
            3366 => Some(Language::Kvq),
            3367 => Some(Language::Kvr),
            3368 => Some(Language::Kvt),
            3369 => Some(Language::Kvu),
            3370 => Some(Language::Kvv),
            3371 => Some(Language::Kvw),
            3372 => Some(Language::Kvx),
            3373 => Some(Language::Kvy),
            3374 => Some(Language::Kvz),
            3375 => Some(Language::Kwa),
            3376 => Some(Language::Kwb),
            3377 => Some(Language::Kwc),
            3378 => Some(Language::Kwd),
            3379 => Some(Language::Kwe),
            3380 => Some(Language::Kwf),
            3381 => Some(Language::Kwg),
            3382 => Some(Language::Kwh),
            3383 => Some(Language::Kwi),
            3384 => Some(Language::Kwj),
            3385 => Some(Language::Kwk),
            3386 => Some(Language::Kwl),
            3387 => Some(Language::Kwm),
            3388 => Some(Language::Kwn),
            3389 => Some(Language::Kwo),
            3390 => Some(Language::Kwp),
            3391 => Some(Language::Kwr),
            3392 => Some(Language::Kws),
            3393 => Some(Language::Kwt),
            3394 => Some(Language::Kwu),
            3395 => Some(Language::Kwv),
            3396 => Some(Language::Kww),
            3397 => Some(Language::Kwx),
            3398 => Some(Language::Kwy),
            3399 => Some(Language::Kwz),
            3400 => Some(Language::Kxa),
            3401 => Some(Language::Kxb),
            3402 => Some(Language::Kxc),
            3403 => Some(Language::Kxd),
            3404 => Some(Language::Kxf),
            3405 => Some(Language::Kxh),
            3406 => Some(Language::Kxi),
            3407 => Some(Language::Kxj),
            3408 => Some(Language::Kxk),
            3409 => Some(Language::Kxm),
            3410 => Some(Language::Kxn),
            3411 => Some(Language::Kxo),
            3412 => Some(Language::Kxp),
            3413 => Some(Language::Kxq),
            3414 => Some(Language::Kxr),
            3415 => Some(Language::Kxs),
            3416 => Some(Language::Kxt),
            3417 => Some(Language::Kxv),
            3418 => Some(Language::Kxw),
            3419 => Some(Language::Kxx),
            3420 => Some(Language::Kxy),
            3421 => Some(Language::Kxz),
            3422 => Some(Language::Kya),
            3423 => Some(Language::Kyb),
            3424 => Some(Language::Kyc),
            3425 => Some(Language::Kyd),
            3426 => Some(Language::Kye),
            3427 => Some(Language::Kyf),
            3428 => Some(Language::Kyg),
            3429 => Some(Language::Kyh),
            3430 => Some(Language::Kyi),
            3431 => Some(Language::Kyj),
            3432 => Some(Language::Kyk),
            3433 => Some(Language::Kyl),
            3434 => Some(Language::Kym),
            3435 => Some(Language::Kyn),
            3436 => Some(Language::Kyo),
            3437 => Some(Language::Kyp),
            3438 => Some(Language::Kyq),
            3439 => Some(Language::Kyr),
            3440 => Some(Language::Kys),
            3441 => Some(Language::Kyt),
            3442 => Some(Language::Kyu),
            3443 => Some(Language::Kyv),
            3444 => Some(Language::Kyw),
            3445 => Some(Language::Kyx),
            3446 => Some(Language::Kyy),
            3447 => Some(Language::Kyz),
            3448 => Some(Language::Kza),
            3449 => Some(Language::Kzb),
            3450 => Some(Language::Kzc),
            3451 => Some(Language::Kzd),
            3452 => Some(Language::Kze),
            3453 => Some(Language::Kzf),
            3454 => Some(Language::Kzg),
            3455 => Some(Language::Kzi),
            3456 => Some(Language::Kzk),
            3457 => Some(Language::Kzl),
            3458 => Some(Language::Kzm),
            3459 => Some(Language::Kzn),
            3460 => Some(Language::Kzo),
            3461 => Some(Language::Kzp),
            3462 => Some(Language::Kzq),
            3463 => Some(Language::Kzr),
            3464 => Some(Language::Kzs),
            3465 => Some(Language::Kzu),
            3466 => Some(Language::Kzv),
            3467 => Some(Language::Kzw),
            3468 => Some(Language::Kzx),
            3469 => Some(Language::Kzy),
            3470 => Some(Language::Kzz),
            3471 => Some(Language::Laa),
            3472 => Some(Language::Lab),
            3473 => Some(Language::Lac),
            3474 => Some(Language::Lad),
            3475 => Some(Language::Lae),
            3476 => Some(Language::Laf),
            3477 => Some(Language::Lag),
            3478 => Some(Language::Lah),
            3479 => Some(Language::Lai),
            3480 => Some(Language::Laj),
            3481 => Some(Language::Lal),
            3482 => Some(Language::Lam),
            3483 => Some(Language::Lan),
            3484 => Some(Language::Lao),
            3485 => Some(Language::Lap),
            3486 => Some(Language::Laq),
            3487 => Some(Language::Lar),
            3488 => Some(Language::Las),
            3489 => Some(Language::Lat),
            3490 => Some(Language::Lau),
            3491 => Some(Language::Lav),
            3492 => Some(Language::Law),
            3493 => Some(Language::Lax),
            3494 => Some(Language::Lay),
            3495 => Some(Language::Laz),
            3496 => Some(Language::Lbb),
            3497 => Some(Language::Lbc),
            3498 => Some(Language::Lbe),
            3499 => Some(Language::Lbf),
            3500 => Some(Language::Lbg),
            3501 => Some(Language::Lbi),
            3502 => Some(Language::Lbj),
            3503 => Some(Language::Lbk),
            3504 => Some(Language::Lbl),
            3505 => Some(Language::Lbm),
            3506 => Some(Language::Lbn),
            3507 => Some(Language::Lbo),
            3508 => Some(Language::Lbq),
            3509 => Some(Language::Lbr),
            3510 => Some(Language::Lbs),
            3511 => Some(Language::Lbt),
            3512 => Some(Language::Lbu),
            3513 => Some(Language::Lbv),
            3514 => Some(Language::Lbw),
            3515 => Some(Language::Lbx),
            3516 => Some(Language::Lby),
            3517 => Some(Language::Lbz),
            3518 => Some(Language::Lcc),
            3519 => Some(Language::Lcd),
            3520 => Some(Language::Lce),
            3521 => Some(Language::Lcf),
            3522 => Some(Language::Lch),
            3523 => Some(Language::Lcl),
            3524 => Some(Language::Lcm),
            3525 => Some(Language::Lcp),
            3526 => Some(Language::Lcq),
            3527 => Some(Language::Lcs),
            3528 => Some(Language::Lda),
            3529 => Some(Language::Ldb),
            3530 => Some(Language::Ldd),
            3531 => Some(Language::Ldg),
            3532 => Some(Language::Ldh),
            3533 => Some(Language::Ldi),
            3534 => Some(Language::Ldj),
            3535 => Some(Language::Ldk),
            3536 => Some(Language::Ldl),
            3537 => Some(Language::Ldm),
            3538 => Some(Language::Ldn),
            3539 => Some(Language::Ldo),
            3540 => Some(Language::Ldp),
            3541 => Some(Language::Ldq),
            3542 => Some(Language::Lea),
            3543 => Some(Language::Leb),
            3544 => Some(Language::Lec),
            3545 => Some(Language::Led),
            3546 => Some(Language::Lee),
            3547 => Some(Language::Lef),
            3548 => Some(Language::Leh),
            3549 => Some(Language::Lei),
            3550 => Some(Language::Lej),
            3551 => Some(Language::Lek),
            3552 => Some(Language::Lel),
            3553 => Some(Language::Lem),
            3554 => Some(Language::Len),
            3555 => Some(Language::Leo),
            3556 => Some(Language::Lep),
            3557 => Some(Language::Leq),
            3558 => Some(Language::Ler),
            3559 => Some(Language::Les),
            3560 => Some(Language::Let),
            3561 => Some(Language::Leu),
            3562 => Some(Language::Lev),
            3563 => Some(Language::Lew),
            3564 => Some(Language::Lex),
            3565 => Some(Language::Ley),
            3566 => Some(Language::Lez),
            3567 => Some(Language::Lfa),
            3568 => Some(Language::Lfn),
            3569 => Some(Language::Lga),
            3570 => Some(Language::Lgb),
            3571 => Some(Language::Lgg),
            3572 => Some(Language::Lgh),
            3573 => Some(Language::Lgi),
            3574 => Some(Language::Lgk),
            3575 => Some(Language::Lgl),
            3576 => Some(Language::Lgm),
            3577 => Some(Language::Lgn),
            3578 => Some(Language::Lgo),
            3579 => Some(Language::Lgq),
            3580 => Some(Language::Lgr),
            3581 => Some(Language::Lgt),
            3582 => Some(Language::Lgu),
            3583 => Some(Language::Lgz),
            3584 => Some(Language::Lha),
            3585 => Some(Language::Lhh),
            3586 => Some(Language::Lhi),
            3587 => Some(Language::Lhl),
            3588 => Some(Language::Lhm),
            3589 => Some(Language::Lhn),
            3590 => Some(Language::Lhp),
            3591 => Some(Language::Lhs),
            3592 => Some(Language::Lht),
            3593 => Some(Language::Lhu),
            3594 => Some(Language::Lia),
            3595 => Some(Language::Lib),
            3596 => Some(Language::Lic),
            3597 => Some(Language::Lid),
            3598 => Some(Language::Lie),
            3599 => Some(Language::Lif),
            3600 => Some(Language::Lig),
            3601 => Some(Language::Lih),
            3602 => Some(Language::Lij),
            3603 => Some(Language::Lik),
            3604 => Some(Language::Lil),
            3605 => Some(Language::Lim),
            3606 => Some(Language::Lin),
            3607 => Some(Language::Lio),
            3608 => Some(Language::Lip),
            3609 => Some(Language::Liq),
            3610 => Some(Language::Lir),
            3611 => Some(Language::Lis),
            3612 => Some(Language::Lit),
            3613 => Some(Language::Liu),
            3614 => Some(Language::Liv),
            3615 => Some(Language::Liw),
            3616 => Some(Language::Lix),
            3617 => Some(Language::Liy),
            3618 => Some(Language::Liz),
            3619 => Some(Language::Lja),
            3620 => Some(Language::Lje),
            3621 => Some(Language::Lji),
            3622 => Some(Language::Ljl),
            3623 => Some(Language::Ljp),
            3624 => Some(Language::Ljw),
            3625 => Some(Language::Ljx),
            3626 => Some(Language::Lka),
            3627 => Some(Language::Lkb),
            3628 => Some(Language::Lkc),
            3629 => Some(Language::Lkd),
            3630 => Some(Language::Lke),
            3631 => Some(Language::Lkh),
            3632 => Some(Language::Lki),
            3633 => Some(Language::Lkj),
            3634 => Some(Language::Lkl),
            3635 => Some(Language::Lkm),
            3636 => Some(Language::Lkn),
            3637 => Some(Language::Lko),
            3638 => Some(Language::Lkr),
            3639 => Some(Language::Lks),
            3640 => Some(Language::Lkt),
            3641 => Some(Language::Lku),
            3642 => Some(Language::Lky),
            3643 => Some(Language::Lla),
            3644 => Some(Language::Llb),
            3645 => Some(Language::Llc),
            3646 => Some(Language::Lld),
            3647 => Some(Language::Lle),
            3648 => Some(Language::Llf),
            3649 => Some(Language::Llg),
            3650 => Some(Language::Llh),
            3651 => Some(Language::Lli),
            3652 => Some(Language::Llj),
            3653 => Some(Language::Llk),
            3654 => Some(Language::Lll),
            3655 => Some(Language::Llm),
            3656 => Some(Language::Lln),
            3657 => Some(Language::Llp),
            3658 => Some(Language::Llq),
            3659 => Some(Language::Lls),
            3660 => Some(Language::Llu),
            3661 => Some(Language::Llx),
            3662 => Some(Language::Lma),
            3663 => Some(Language::Lmb),
            3664 => Some(Language::Lmc),
            3665 => Some(Language::Lmd),
            3666 => Some(Language::Lme),
            3667 => Some(Language::Lmf),
            3668 => Some(Language::Lmg),
            3669 => Some(Language::Lmh),
            3670 => Some(Language::Lmi),
            3671 => Some(Language::Lmj),
            3672 => Some(Language::Lmk),
            3673 => Some(Language::Lml),
            3674 => Some(Language::Lmn),
            3675 => Some(Language::Lmo),
            3676 => Some(Language::Lmp),
            3677 => Some(Language::Lmq),
            3678 => Some(Language::Lmr),
            3679 => Some(Language::Lmu),
            3680 => Some(Language::Lmv),
            3681 => Some(Language::Lmw),
            3682 => Some(Language::Lmx),
            3683 => Some(Language::Lmy),
            3684 => Some(Language::Lna),
            3685 => Some(Language::Lnb),
            3686 => Some(Language::Lnd),
            3687 => Some(Language::Lng),
            3688 => Some(Language::Lnh),
            3689 => Some(Language::Lni),
            3690 => Some(Language::Lnj),
            3691 => Some(Language::Lnl),
            3692 => Some(Language::Lnm),
            3693 => Some(Language::Lnn),
            3694 => Some(Language::Lns),
            3695 => Some(Language::Lnu),
            3696 => Some(Language::Lnw),
            3697 => Some(Language::Lnz),
            3698 => Some(Language::Loa),
            3699 => Some(Language::Lob),
            3700 => Some(Language::Loc),
            3701 => Some(Language::Loe),
            3702 => Some(Language::Lof),
            3703 => Some(Language::Log),
            3704 => Some(Language::Loh),
            3705 => Some(Language::Loi),
            3706 => Some(Language::Loj),
            3707 => Some(Language::Lok),
            3708 => Some(Language::Lol),
            3709 => Some(Language::Lom),
            3710 => Some(Language::Lon),
            3711 => Some(Language::Loo),
            3712 => Some(Language::Lop),
            3713 => Some(Language::Loq),
            3714 => Some(Language::Lor),
            3715 => Some(Language::Los),
            3716 => Some(Language::Lot),
            3717 => Some(Language::Lou),
            3718 => Some(Language::Lov),
            3719 => Some(Language::Low),
            3720 => Some(Language::Lox),
            3721 => Some(Language::Loy),
            3722 => Some(Language::Loz),
            3723 => Some(Language::Lpa),
            3724 => Some(Language::Lpe),
            3725 => Some(Language::Lpn),
            3726 => Some(Language::Lpo),
            3727 => Some(Language::Lpx),
            3728 => Some(Language::Lqr),
            3729 => Some(Language::Lra),
            3730 => Some(Language::Lrc),
            3731 => Some(Language::Lre),
            3732 => Some(Language::Lrg),
            3733 => Some(Language::Lri),
            3734 => Some(Language::Lrk),
            3735 => Some(Language::Lrl),
            3736 => Some(Language::Lrm),
            3737 => Some(Language::Lrn),
            3738 => Some(Language::Lro),
            3739 => Some(Language::Lrr),
            3740 => Some(Language::Lrt),
            3741 => Some(Language::Lrv),
            3742 => Some(Language::Lrz),
            3743 => Some(Language::Lsa),
            3744 => Some(Language::Lsb),
            3745 => Some(Language::Lsc),
            3746 => Some(Language::Lsd),
            3747 => Some(Language::Lse),
            3748 => Some(Language::Lsh),
            3749 => Some(Language::Lsi),
            3750 => Some(Language::Lsl),
            3751 => Some(Language::Lsm),
            3752 => Some(Language::Lsn),
            3753 => Some(Language::Lso),
            3754 => Some(Language::Lsp),
            3755 => Some(Language::Lsr),
            3756 => Some(Language::Lss),
            3757 => Some(Language::Lst),
            3758 => Some(Language::Lsv),
            3759 => Some(Language::Lsw),
            3760 => Some(Language::Lsy),
            3761 => Some(Language::Ltc),
            3762 => Some(Language::Ltg),
            3763 => Some(Language::Lth),
            3764 => Some(Language::Lti),
            3765 => Some(Language::Ltn),
            3766 => Some(Language::Lto),
            3767 => Some(Language::Lts),
            3768 => Some(Language::Ltu),
            3769 => Some(Language::Ltz),
            3770 => Some(Language::Lua),
            3771 => Some(Language::Lub),
            3772 => Some(Language::Luc),
            3773 => Some(Language::Lud),
            3774 => Some(Language::Lue),
            3775 => Some(Language::Luf),
            3776 => Some(Language::Lug),
            3777 => Some(Language::Lui),
            3778 => Some(Language::Luj),
            3779 => Some(Language::Luk),
            3780 => Some(Language::Lul),
            3781 => Some(Language::Lum),
            3782 => Some(Language::Lun),
            3783 => Some(Language::Luo),
            3784 => Some(Language::Lup),
            3785 => Some(Language::Luq),
            3786 => Some(Language::Lur),
            3787 => Some(Language::Lus),
            3788 => Some(Language::Lut),
            3789 => Some(Language::Luu),
            3790 => Some(Language::Luv),
            3791 => Some(Language::Luw),
            3792 => Some(Language::Luy),
            3793 => Some(Language::Luz),
            3794 => Some(Language::Lva),
            3795 => Some(Language::Lvi),
            3796 => Some(Language::Lvk),
            3797 => Some(Language::Lvs),
            3798 => Some(Language::Lvu),
            3799 => Some(Language::Lwa),
            3800 => Some(Language::Lwe),
            3801 => Some(Language::Lwg),
            3802 => Some(Language::Lwh),
            3803 => Some(Language::Lwl),
            3804 => Some(Language::Lwm),
            3805 => Some(Language::Lwo),
            3806 => Some(Language::Lws),
            3807 => Some(Language::Lwt),
            3808 => Some(Language::Lwu),
            3809 => Some(Language::Lww),
            3810 => Some(Language::Lxm),
            3811 => Some(Language::Lya),
            3812 => Some(Language::Lyg),
            3813 => Some(Language::Lyn),
            3814 => Some(Language::Lzh),
            3815 => Some(Language::Lzl),
            3816 => Some(Language::Lzn),
            3817 => Some(Language::Lzz),
            3818 => Some(Language::Maa),
            3819 => Some(Language::Mab),
            3820 => Some(Language::Mad),
            3821 => Some(Language::Mae),
            3822 => Some(Language::Maf),
            3823 => Some(Language::Mag),
            3824 => Some(Language::Mah),
            3825 => Some(Language::Mai),
            3826 => Some(Language::Maj),
            3827 => Some(Language::Mak),
            3828 => Some(Language::Mal),
            3829 => Some(Language::Mam),
            3830 => Some(Language::Man),
            3831 => Some(Language::Maq),
            3832 => Some(Language::Mar),
            3833 => Some(Language::Mas),
            3834 => Some(Language::Mat),
            3835 => Some(Language::Mau),
            3836 => Some(Language::Mav),
            3837 => Some(Language::Maw),
            3838 => Some(Language::Max),
            3839 => Some(Language::Maz),
            3840 => Some(Language::Mba),
            3841 => Some(Language::Mbb),
            3842 => Some(Language::Mbc),
            3843 => Some(Language::Mbd),
            3844 => Some(Language::Mbe),
            3845 => Some(Language::Mbf),
            3846 => Some(Language::Mbh),
            3847 => Some(Language::Mbi),
            3848 => Some(Language::Mbj),
            3849 => Some(Language::Mbk),
            3850 => Some(Language::Mbl),
            3851 => Some(Language::Mbm),
            3852 => Some(Language::Mbn),
            3853 => Some(Language::Mbo),
            3854 => Some(Language::Mbp),
            3855 => Some(Language::Mbq),
            3856 => Some(Language::Mbr),
            3857 => Some(Language::Mbs),
            3858 => Some(Language::Mbt),
            3859 => Some(Language::Mbu),
            3860 => Some(Language::Mbv),
            3861 => Some(Language::Mbw),
            3862 => Some(Language::Mbx),
            3863 => Some(Language::Mby),
            3864 => Some(Language::Mbz),
            3865 => Some(Language::Mca),
            3866 => Some(Language::Mcb),
            3867 => Some(Language::Mcc),
            3868 => Some(Language::Mcd),
            3869 => Some(Language::Mce),
            3870 => Some(Language::Mcf),
            3871 => Some(Language::Mcg),
            3872 => Some(Language::Mch),
            3873 => Some(Language::Mci),
            3874 => Some(Language::Mcj),
            3875 => Some(Language::Mck),
            3876 => Some(Language::Mcl),
            3877 => Some(Language::Mcm),
            3878 => Some(Language::Mcn),
            3879 => Some(Language::Mco),
            3880 => Some(Language::Mcp),
            3881 => Some(Language::Mcq),
            3882 => Some(Language::Mcr),
            3883 => Some(Language::Mcs),
            3884 => Some(Language::Mct),
            3885 => Some(Language::Mcu),
            3886 => Some(Language::Mcv),
            3887 => Some(Language::Mcw),
            3888 => Some(Language::Mcx),
            3889 => Some(Language::Mcy),
            3890 => Some(Language::Mcz),
            3891 => Some(Language::Mda),
            3892 => Some(Language::Mdb),
            3893 => Some(Language::Mdc),
            3894 => Some(Language::Mdd),
            3895 => Some(Language::Mde),
            3896 => Some(Language::Mdf),
            3897 => Some(Language::Mdg),
            3898 => Some(Language::Mdh),
            3899 => Some(Language::Mdi),
            3900 => Some(Language::Mdj),
            3901 => Some(Language::Mdk),
            3902 => Some(Language::Mdl),
            3903 => Some(Language::Mdm),
            3904 => Some(Language::Mdn),
            3905 => Some(Language::Mdp),
            3906 => Some(Language::Mdq),
            3907 => Some(Language::Mdr),
            3908 => Some(Language::Mds),
            3909 => Some(Language::Mdt),
            3910 => Some(Language::Mdu),
            3911 => Some(Language::Mdv),
            3912 => Some(Language::Mdw),
            3913 => Some(Language::Mdx),
            3914 => Some(Language::Mdy),
            3915 => Some(Language::Mdz),
            3916 => Some(Language::Mea),
            3917 => Some(Language::Meb),
            3918 => Some(Language::Mec),
            3919 => Some(Language::Med),
            3920 => Some(Language::Mee),
            3921 => Some(Language::Mef),
            3922 => Some(Language::Meh),
            3923 => Some(Language::Mei),
            3924 => Some(Language::Mej),
            3925 => Some(Language::Mek),
            3926 => Some(Language::Mel),
            3927 => Some(Language::Mem),
            3928 => Some(Language::Men),
            3929 => Some(Language::Meo),
            3930 => Some(Language::Mep),
            3931 => Some(Language::Meq),
            3932 => Some(Language::Mer),
            3933 => Some(Language::Mes),
            3934 => Some(Language::Met),
            3935 => Some(Language::Meu),
            3936 => Some(Language::Mev),
            3937 => Some(Language::Mew),
            3938 => Some(Language::Mey),
            3939 => Some(Language::Mez),
            3940 => Some(Language::Mfa),
            3941 => Some(Language::Mfb),
            3942 => Some(Language::Mfc),
            3943 => Some(Language::Mfd),
            3944 => Some(Language::Mfe),
            3945 => Some(Language::Mff),
            3946 => Some(Language::Mfg),
            3947 => Some(Language::Mfh),
            3948 => Some(Language::Mfi),
            3949 => Some(Language::Mfj),
            3950 => Some(Language::Mfk),
            3951 => Some(Language::Mfl),
            3952 => Some(Language::Mfm),
            3953 => Some(Language::Mfn),
            3954 => Some(Language::Mfo),
            3955 => Some(Language::Mfp),
            3956 => Some(Language::Mfq),
            3957 => Some(Language::Mfr),
            3958 => Some(Language::Mfs),
            3959 => Some(Language::Mft),
            3960 => Some(Language::Mfu),
            3961 => Some(Language::Mfv),
            3962 => Some(Language::Mfw),
            3963 => Some(Language::Mfx),
            3964 => Some(Language::Mfy),
            3965 => Some(Language::Mfz),
            3966 => Some(Language::Mga),
            3967 => Some(Language::Mgb),
            3968 => Some(Language::Mgc),
            3969 => Some(Language::Mgd),
            3970 => Some(Language::Mge),
            3971 => Some(Language::Mgf),
            3972 => Some(Language::Mgg),
            3973 => Some(Language::Mgh),
            3974 => Some(Language::Mgi),
            3975 => Some(Language::Mgj),
            3976 => Some(Language::Mgk),
            3977 => Some(Language::Mgl),
            3978 => Some(Language::Mgm),
            3979 => Some(Language::Mgn),
            3980 => Some(Language::Mgo),
            3981 => Some(Language::Mgp),
            3982 => Some(Language::Mgq),
            3983 => Some(Language::Mgr),
            3984 => Some(Language::Mgs),
            3985 => Some(Language::Mgt),
            3986 => Some(Language::Mgu),
            3987 => Some(Language::Mgv),
            3988 => Some(Language::Mgw),
            3989 => Some(Language::Mgy),
            3990 => Some(Language::Mgz),
            3991 => Some(Language::Mha),
            3992 => Some(Language::Mhb),
            3993 => Some(Language::Mhc),
            3994 => Some(Language::Mhd),
            3995 => Some(Language::Mhe),
            3996 => Some(Language::Mhf),
            3997 => Some(Language::Mhg),
            3998 => Some(Language::Mhi),
            3999 => Some(Language::Mhj),
            4000 => Some(Language::Mhk),
            4001 => Some(Language::Mhl),
            4002 => Some(Language::Mhm),
            4003 => Some(Language::Mhn),
            4004 => Some(Language::Mho),
            4005 => Some(Language::Mhp),
            4006 => Some(Language::Mhq),
            4007 => Some(Language::Mhr),
            4008 => Some(Language::Mhs),
            4009 => Some(Language::Mht),
            4010 => Some(Language::Mhu),
            4011 => Some(Language::Mhw),
            4012 => Some(Language::Mhx),
            4013 => Some(Language::Mhy),
            4014 => Some(Language::Mhz),
            4015 => Some(Language::Mia),
            4016 => Some(Language::Mib),
            4017 => Some(Language::Mic),
            4018 => Some(Language::Mid),
            4019 => Some(Language::Mie),
            4020 => Some(Language::Mif),
            4021 => Some(Language::Mig),
            4022 => Some(Language::Mih),
            4023 => Some(Language::Mii),
            4024 => Some(Language::Mij),
            4025 => Some(Language::Mik),
            4026 => Some(Language::Mil),
            4027 => Some(Language::Mim),
            4028 => Some(Language::Min),
            4029 => Some(Language::Mio),
            4030 => Some(Language::Mip),
            4031 => Some(Language::Miq),
            4032 => Some(Language::Mir),
            4033 => Some(Language::Mis),
            4034 => Some(Language::Mit),
            4035 => Some(Language::Miu),
            4036 => Some(Language::Miw),
            4037 => Some(Language::Mix),
            4038 => Some(Language::Miy),
            4039 => Some(Language::Miz),
            4040 => Some(Language::Mjb),
            4041 => Some(Language::Mjc),
            4042 => Some(Language::Mjd),
            4043 => Some(Language::Mje),
            4044 => Some(Language::Mjg),
            4045 => Some(Language::Mjh),
            4046 => Some(Language::Mji),
            4047 => Some(Language::Mjj),
            4048 => Some(Language::Mjk),
            4049 => Some(Language::Mjl),
            4050 => Some(Language::Mjm),
            4051 => Some(Language::Mjn),
            4052 => Some(Language::Mjo),
            4053 => Some(Language::Mjp),
            4054 => Some(Language::Mjq),
            4055 => Some(Language::Mjr),
            4056 => Some(Language::Mjs),
            4057 => Some(Language::Mjt),
            4058 => Some(Language::Mju),
            4059 => Some(Language::Mjv),
            4060 => Some(Language::Mjw),
            4061 => Some(Language::Mjx),
            4062 => Some(Language::Mjy),
            4063 => Some(Language::Mjz),
            4064 => Some(Language::Mka),
            4065 => Some(Language::Mkb),
            4066 => Some(Language::Mkc),
            4067 => Some(Language::Mkd),
            4068 => Some(Language::Mke),
            4069 => Some(Language::Mkf),
            4070 => Some(Language::Mkg),
            4071 => Some(Language::Mki),
            4072 => Some(Language::Mkj),
            4073 => Some(Language::Mkk),
            4074 => Some(Language::Mkl),
            4075 => Some(Language::Mkm),
            4076 => Some(Language::Mkn),
            4077 => Some(Language::Mko),
            4078 => Some(Language::Mkp),
            4079 => Some(Language::Mkq),
            4080 => Some(Language::Mkr),
            4081 => Some(Language::Mks),
            4082 => Some(Language::Mkt),
            4083 => Some(Language::Mku),
            4084 => Some(Language::Mkv),
            4085 => Some(Language::Mkw),
            4086 => Some(Language::Mkx),
            4087 => Some(Language::Mky),
            4088 => Some(Language::Mkz),
            4089 => Some(Language::Mla),
            4090 => Some(Language::Mlb),
            4091 => Some(Language::Mlc),
            4092 => Some(Language::Mle),
            4093 => Some(Language::Mlf),
            4094 => Some(Language::Mlg),
            4095 => Some(Language::Mlh),
            4096 => Some(Language::Mli),
            4097 => Some(Language::Mlj),
            4098 => Some(Language::Mlk),
            4099 => Some(Language::Mll),
            4100 => Some(Language::Mlm),
            4101 => Some(Language::Mln),
            4102 => Some(Language::Mlo),
            4103 => Some(Language::Mlp),
            4104 => Some(Language::Mlq),
            4105 => Some(Language::Mlr),
            4106 => Some(Language::Mls),
            4107 => Some(Language::Mlt),
            4108 => Some(Language::Mlu),
            4109 => Some(Language::Mlv),
            4110 => Some(Language::Mlw),
            4111 => Some(Language::Mlx),
            4112 => Some(Language::Mlz),
            4113 => Some(Language::Mma),
            4114 => Some(Language::Mmb),
            4115 => Some(Language::Mmc),
            4116 => Some(Language::Mmd),
            4117 => Some(Language::Mme),
            4118 => Some(Language::Mmf),
            4119 => Some(Language::Mmg),
            4120 => Some(Language::Mmh),
            4121 => Some(Language::Mmi),
            4122 => Some(Language::Mmj),
            4123 => Some(Language::Mmk),
            4124 => Some(Language::Mml),
            4125 => Some(Language::Mmm),
            4126 => Some(Language::Mmn),
            4127 => Some(Language::Mmo),
            4128 => Some(Language::Mmp),
            4129 => Some(Language::Mmq),
            4130 => Some(Language::Mmr),
            4131 => Some(Language::Mmt),
            4132 => Some(Language::Mmu),
            4133 => Some(Language::Mmv),
            4134 => Some(Language::Mmw),
            4135 => Some(Language::Mmx),
            4136 => Some(Language::Mmy),
            4137 => Some(Language::Mmz),
            4138 => Some(Language::Mna),
            4139 => Some(Language::Mnb),
            4140 => Some(Language::Mnc),
            4141 => Some(Language::Mnd),
            4142 => Some(Language::Mne),
            4143 => Some(Language::Mnf),
            4144 => Some(Language::Mng),
            4145 => Some(Language::Mnh),
            4146 => Some(Language::Mni),
            4147 => Some(Language::Mnj),
            4148 => Some(Language::Mnk),
            4149 => Some(Language::Mnl),
            4150 => Some(Language::Mnm),
            4151 => Some(Language::Mnn),
            4152 => Some(Language::Mnp),
            4153 => Some(Language::Mnq),
            4154 => Some(Language::Mnr),
            4155 => Some(Language::Mns),
            4156 => Some(Language::Mnu),
            4157 => Some(Language::Mnv),
            4158 => Some(Language::Mnw),
            4159 => Some(Language::Mnx),
            4160 => Some(Language::Mny),
            4161 => Some(Language::Mnz),
            4162 => Some(Language::Moa),
            4163 => Some(Language::Moc),
            4164 => Some(Language::Mod),
            4165 => Some(Language::Moe),
            4166 => Some(Language::Mog),
            4167 => Some(Language::Moh),
            4168 => Some(Language::Moi),
            4169 => Some(Language::Moj),
            4170 => Some(Language::Mok),
            4171 => Some(Language::Mom),
            4172 => Some(Language::Mon),
            4173 => Some(Language::Moo),
            4174 => Some(Language::Mop),
            4175 => Some(Language::Moq),
            4176 => Some(Language::Mor),
            4177 => Some(Language::Mos),
            4178 => Some(Language::Mot),
            4179 => Some(Language::Mou),
            4180 => Some(Language::Mov),
            4181 => Some(Language::Mow),
            4182 => Some(Language::Mox),
            4183 => Some(Language::Moy),
            4184 => Some(Language::Moz),
            4185 => Some(Language::Mpa),
            4186 => Some(Language::Mpb),
            4187 => Some(Language::Mpc),
            4188 => Some(Language::Mpd),
            4189 => Some(Language::Mpe),
            4190 => Some(Language::Mpg),
            4191 => Some(Language::Mph),
            4192 => Some(Language::Mpi),
            4193 => Some(Language::Mpj),
            4194 => Some(Language::Mpk),
            4195 => Some(Language::Mpl),
            4196 => Some(Language::Mpm),
            4197 => Some(Language::Mpn),
            4198 => Some(Language::Mpo),
            4199 => Some(Language::Mpp),
            4200 => Some(Language::Mpq),
            4201 => Some(Language::Mpr),
            4202 => Some(Language::Mps),
            4203 => Some(Language::Mpt),
            4204 => Some(Language::Mpu),
            4205 => Some(Language::Mpv),
            4206 => Some(Language::Mpw),
            4207 => Some(Language::Mpx),
            4208 => Some(Language::Mpy),
            4209 => Some(Language::Mpz),
            4210 => Some(Language::Mqa),
            4211 => Some(Language::Mqb),
            4212 => Some(Language::Mqc),
            4213 => Some(Language::Mqe),
            4214 => Some(Language::Mqf),
            4215 => Some(Language::Mqg),
            4216 => Some(Language::Mqh),
            4217 => Some(Language::Mqi),
            4218 => Some(Language::Mqj),
            4219 => Some(Language::Mqk),
            4220 => Some(Language::Mql),
            4221 => Some(Language::Mqm),
            4222 => Some(Language::Mqn),
            4223 => Some(Language::Mqo),
            4224 => Some(Language::Mqp),
            4225 => Some(Language::Mqq),
            4226 => Some(Language::Mqr),
            4227 => Some(Language::Mqs),
            4228 => Some(Language::Mqt),
            4229 => Some(Language::Mqu),
            4230 => Some(Language::Mqv),
            4231 => Some(Language::Mqw),
            4232 => Some(Language::Mqx),
            4233 => Some(Language::Mqy),
            4234 => Some(Language::Mqz),
            4235 => Some(Language::Mra),
            4236 => Some(Language::Mrb),
            4237 => Some(Language::Mrc),
            4238 => Some(Language::Mrd),
            4239 => Some(Language::Mre),
            4240 => Some(Language::Mrf),
            4241 => Some(Language::Mrg),
            4242 => Some(Language::Mrh),
            4243 => Some(Language::Mri),
            4244 => Some(Language::Mrj),
            4245 => Some(Language::Mrk),
            4246 => Some(Language::Mrl),
            4247 => Some(Language::Mrm),
            4248 => Some(Language::Mrn),
            4249 => Some(Language::Mro),
            4250 => Some(Language::Mrp),
            4251 => Some(Language::Mrq),
            4252 => Some(Language::Mrr),
            4253 => Some(Language::Mrs),
            4254 => Some(Language::Mrt),
            4255 => Some(Language::Mru),
            4256 => Some(Language::Mrv),
            4257 => Some(Language::Mrw),
            4258 => Some(Language::Mrx),
            4259 => Some(Language::Mry),
            4260 => Some(Language::Mrz),
            4261 => Some(Language::Msa),
            4262 => Some(Language::Msb),
            4263 => Some(Language::Msc),
            4264 => Some(Language::Msd),
            4265 => Some(Language::Mse),
            4266 => Some(Language::Msf),
            4267 => Some(Language::Msg),
            4268 => Some(Language::Msh),
            4269 => Some(Language::Msi),
            4270 => Some(Language::Msj),
            4271 => Some(Language::Msk),
            4272 => Some(Language::Msl),
            4273 => Some(Language::Msm),
            4274 => Some(Language::Msn),
            4275 => Some(Language::Mso),
            4276 => Some(Language::Msp),
            4277 => Some(Language::Msq),
            4278 => Some(Language::Msr),
            4279 => Some(Language::Mss),
            4280 => Some(Language::Msu),
            4281 => Some(Language::Msv),
            4282 => Some(Language::Msw),
            4283 => Some(Language::Msx),
            4284 => Some(Language::Msy),
            4285 => Some(Language::Msz),
            4286 => Some(Language::Mta),
            4287 => Some(Language::Mtb),
            4288 => Some(Language::Mtc),
            4289 => Some(Language::Mtd),
            4290 => Some(Language::Mte),
            4291 => Some(Language::Mtf),
            4292 => Some(Language::Mtg),
            4293 => Some(Language::Mth),
            4294 => Some(Language::Mti),
            4295 => Some(Language::Mtj),
            4296 => Some(Language::Mtk),
            4297 => Some(Language::Mtl),
            4298 => Some(Language::Mtm),
            4299 => Some(Language::Mtn),
            4300 => Some(Language::Mto),
            4301 => Some(Language::Mtp),
            4302 => Some(Language::Mtq),
            4303 => Some(Language::Mtr),
            4304 => Some(Language::Mts),
            4305 => Some(Language::Mtt),
            4306 => Some(Language::Mtu),
            4307 => Some(Language::Mtv),
            4308 => Some(Language::Mtw),
            4309 => Some(Language::Mtx),
            4310 => Some(Language::Mty),
            4311 => Some(Language::Mua),
            4312 => Some(Language::Mub),
            4313 => Some(Language::Muc),
            4314 => Some(Language::Mud),
            4315 => Some(Language::Mue),
            4316 => Some(Language::Mug),
            4317 => Some(Language::Muh),
            4318 => Some(Language::Mui),
            4319 => Some(Language::Muj),
            4320 => Some(Language::Muk),
            4321 => Some(Language::Mul),
            4322 => Some(Language::Mum),
            4323 => Some(Language::Muo),
            4324 => Some(Language::Mup),
            4325 => Some(Language::Muq),
            4326 => Some(Language::Mur),
            4327 => Some(Language::Mus),
            4328 => Some(Language::Mut),
            4329 => Some(Language::Muu),
            4330 => Some(Language::Muv),
            4331 => Some(Language::Mux),
            4332 => Some(Language::Muy),
            4333 => Some(Language::Muz),
            4334 => Some(Language::Mva),
            4335 => Some(Language::Mvb),
            4336 => Some(Language::Mvd),
            4337 => Some(Language::Mve),
            4338 => Some(Language::Mvf),
            4339 => Some(Language::Mvg),
            4340 => Some(Language::Mvh),
            4341 => Some(Language::Mvi),
            4342 => Some(Language::Mvk),
            4343 => Some(Language::Mvl),
            4344 => Some(Language::Mvn),
            4345 => Some(Language::Mvo),
            4346 => Some(Language::Mvp),
            4347 => Some(Language::Mvq),
            4348 => Some(Language::Mvr),
            4349 => Some(Language::Mvs),
            4350 => Some(Language::Mvt),
            4351 => Some(Language::Mvu),
            4352 => Some(Language::Mvv),
            4353 => Some(Language::Mvw),
            4354 => Some(Language::Mvx),
            4355 => Some(Language::Mvy),
            4356 => Some(Language::Mvz),
            4357 => Some(Language::Mwa),
            4358 => Some(Language::Mwb),
            4359 => Some(Language::Mwc),
            4360 => Some(Language::Mwe),
            4361 => Some(Language::Mwf),
            4362 => Some(Language::Mwg),
            4363 => Some(Language::Mwh),
            4364 => Some(Language::Mwi),
            4365 => Some(Language::Mwk),
            4366 => Some(Language::Mwl),
            4367 => Some(Language::Mwm),
            4368 => Some(Language::Mwn),
            4369 => Some(Language::Mwo),
            4370 => Some(Language::Mwp),
            4371 => Some(Language::Mwq),
            4372 => Some(Language::Mwr),
            4373 => Some(Language::Mws),
            4374 => Some(Language::Mwt),
            4375 => Some(Language::Mwu),
            4376 => Some(Language::Mwv),
            4377 => Some(Language::Mww),
            4378 => Some(Language::Mwz),
            4379 => Some(Language::Mxa),
            4380 => Some(Language::Mxb),
            4381 => Some(Language::Mxc),
            4382 => Some(Language::Mxd),
            4383 => Some(Language::Mxe),
            4384 => Some(Language::Mxf),
            4385 => Some(Language::Mxg),
            4386 => Some(Language::Mxh),
            4387 => Some(Language::Mxi),
            4388 => Some(Language::Mxj),
            4389 => Some(Language::Mxk),
            4390 => Some(Language::Mxl),
            4391 => Some(Language::Mxm),
            4392 => Some(Language::Mxn),
            4393 => Some(Language::Mxo),
            4394 => Some(Language::Mxp),
            4395 => Some(Language::Mxq),
            4396 => Some(Language::Mxr),
            4397 => Some(Language::Mxs),
            4398 => Some(Language::Mxt),
            4399 => Some(Language::Mxu),
            4400 => Some(Language::Mxv),
            4401 => Some(Language::Mxw),
            4402 => Some(Language::Mxx),
            4403 => Some(Language::Mxy),
            4404 => Some(Language::Mxz),
            4405 => Some(Language::Mya),
            4406 => Some(Language::Myb),
            4407 => Some(Language::Myc),
            4408 => Some(Language::Mye),
            4409 => Some(Language::Myf),
            4410 => Some(Language::Myg),
            4411 => Some(Language::Myh),
            4412 => Some(Language::Myj),
            4413 => Some(Language::Myk),
            4414 => Some(Language::Myl),
            4415 => Some(Language::Mym),
            4416 => Some(Language::Myo),
            4417 => Some(Language::Myp),
            4418 => Some(Language::Myr),
            4419 => Some(Language::Mys),
            4420 => Some(Language::Myu),
            4421 => Some(Language::Myv),
            4422 => Some(Language::Myw),
            4423 => Some(Language::Myx),
            4424 => Some(Language::Myy),
            4425 => Some(Language::Myz),
            4426 => Some(Language::Mza),
            4427 => Some(Language::Mzb),
            4428 => Some(Language::Mzc),
            4429 => Some(Language::Mzd),
            4430 => Some(Language::Mze),
            4431 => Some(Language::Mzg),
            4432 => Some(Language::Mzh),
            4433 => Some(Language::Mzi),
            4434 => Some(Language::Mzj),
            4435 => Some(Language::Mzk),
            4436 => Some(Language::Mzl),
            4437 => Some(Language::Mzm),
            4438 => Some(Language::Mzn),
            4439 => Some(Language::Mzo),
            4440 => Some(Language::Mzp),
            4441 => Some(Language::Mzq),
            4442 => Some(Language::Mzr),
            4443 => Some(Language::Mzs),
            4444 => Some(Language::Mzt),
            4445 => Some(Language::Mzu),
            4446 => Some(Language::Mzv),
            4447 => Some(Language::Mzw),
            4448 => Some(Language::Mzx),
            4449 => Some(Language::Mzy),
            4450 => Some(Language::Mzz),
            4451 => Some(Language::Naa),
            4452 => Some(Language::Nab),
            4453 => Some(Language::Nac),
            4454 => Some(Language::Nae),
            4455 => Some(Language::Naf),
            4456 => Some(Language::Nag),
            4457 => Some(Language::Naj),
            4458 => Some(Language::Nak),
            4459 => Some(Language::Nal),
            4460 => Some(Language::Nam),
            4461 => Some(Language::Nan),
            4462 => Some(Language::Nao),
            4463 => Some(Language::Nap),
            4464 => Some(Language::Naq),
            4465 => Some(Language::Nar),
            4466 => Some(Language::Nas),
            4467 => Some(Language::Nat),
            4468 => Some(Language::Nau),
            4469 => Some(Language::Nav),
            4470 => Some(Language::Naw),
            4471 => Some(Language::Nax),
            4472 => Some(Language::Nay),
            4473 => Some(Language::Naz),
            4474 => Some(Language::Nba),
            4475 => Some(Language::Nbb),
            4476 => Some(Language::Nbc),
            4477 => Some(Language::Nbd),
            4478 => Some(Language::Nbe),
            4479 => Some(Language::Nbg),
            4480 => Some(Language::Nbh),
            4481 => Some(Language::Nbi),
            4482 => Some(Language::Nbj),
            4483 => Some(Language::Nbk),
            4484 => Some(Language::Nbl),
            4485 => Some(Language::Nbm),
            4486 => Some(Language::Nbn),
            4487 => Some(Language::Nbo),
            4488 => Some(Language::Nbp),
            4489 => Some(Language::Nbq),
            4490 => Some(Language::Nbr),
            4491 => Some(Language::Nbs),
            4492 => Some(Language::Nbt),
            4493 => Some(Language::Nbu),
            4494 => Some(Language::Nbv),
            4495 => Some(Language::Nbw),
            4496 => Some(Language::Nby),
            4497 => Some(Language::Nca),
            4498 => Some(Language::Ncb),
            4499 => Some(Language::Ncc),
            4500 => Some(Language::Ncd),
            4501 => Some(Language::Nce),
            4502 => Some(Language::Ncf),
            4503 => Some(Language::Ncg),
            4504 => Some(Language::Nch),
            4505 => Some(Language::Nci),
            4506 => Some(Language::Ncj),
            4507 => Some(Language::Nck),
            4508 => Some(Language::Ncl),
            4509 => Some(Language::Ncm),
            4510 => Some(Language::Ncn),
            4511 => Some(Language::Nco),
            4512 => Some(Language::Ncq),
            4513 => Some(Language::Ncr),
            4514 => Some(Language::Ncs),
            4515 => Some(Language::Nct),
            4516 => Some(Language::Ncu),
            4517 => Some(Language::Ncx),
            4518 => Some(Language::Ncz),
            4519 => Some(Language::Nda),
            4520 => Some(Language::Ndb),
            4521 => Some(Language::Ndc),
            4522 => Some(Language::Ndd),
            4523 => Some(Language::Nde),
            4524 => Some(Language::Ndf),
            4525 => Some(Language::Ndg),
            4526 => Some(Language::Ndh),
            4527 => Some(Language::Ndi),
            4528 => Some(Language::Ndj),
            4529 => Some(Language::Ndk),
            4530 => Some(Language::Ndl),
            4531 => Some(Language::Ndm),
            4532 => Some(Language::Ndn),
            4533 => Some(Language::Ndo),
            4534 => Some(Language::Ndp),
            4535 => Some(Language::Ndq),
            4536 => Some(Language::Ndr),
            4537 => Some(Language::Nds),
            4538 => Some(Language::Ndt),
            4539 => Some(Language::Ndu),
            4540 => Some(Language::Ndv),
            4541 => Some(Language::Ndw),
            4542 => Some(Language::Ndx),
            4543 => Some(Language::Ndy),
            4544 => Some(Language::Ndz),
            4545 => Some(Language::Nea),
            4546 => Some(Language::Neb),
            4547 => Some(Language::Nec),
            4548 => Some(Language::Ned),
            4549 => Some(Language::Nee),
            4550 => Some(Language::Nef),
            4551 => Some(Language::Neg),
            4552 => Some(Language::Neh),
            4553 => Some(Language::Nei),
            4554 => Some(Language::Nej),
            4555 => Some(Language::Nek),
            4556 => Some(Language::Nem),
            4557 => Some(Language::Nen),
            4558 => Some(Language::Neo),
            4559 => Some(Language::Nep),
            4560 => Some(Language::Neq),
            4561 => Some(Language::Ner),
            4562 => Some(Language::Nes),
            4563 => Some(Language::Net),
            4564 => Some(Language::Neu),
            4565 => Some(Language::Nev),
            4566 => Some(Language::New),
            4567 => Some(Language::Nex),
            4568 => Some(Language::Ney),
            4569 => Some(Language::Nez),
            4570 => Some(Language::Nfa),
            4571 => Some(Language::Nfd),
            4572 => Some(Language::Nfl),
            4573 => Some(Language::Nfr),
            4574 => Some(Language::Nfu),
            4575 => Some(Language::Nga),
            4576 => Some(Language::Ngb),
            4577 => Some(Language::Ngc),
            4578 => Some(Language::Ngd),
            4579 => Some(Language::Nge),
            4580 => Some(Language::Ngg),
            4581 => Some(Language::Ngh),
            4582 => Some(Language::Ngi),
            4583 => Some(Language::Ngj),
            4584 => Some(Language::Ngk),
            4585 => Some(Language::Ngl),
            4586 => Some(Language::Ngm),
            4587 => Some(Language::Ngn),
            4588 => Some(Language::Ngp),
            4589 => Some(Language::Ngq),
            4590 => Some(Language::Ngr),
            4591 => Some(Language::Ngs),
            4592 => Some(Language::Ngt),
            4593 => Some(Language::Ngu),
            4594 => Some(Language::Ngv),
            4595 => Some(Language::Ngw),
            4596 => Some(Language::Ngx),
            4597 => Some(Language::Ngy),
            4598 => Some(Language::Ngz),
            4599 => Some(Language::Nha),
            4600 => Some(Language::Nhb),
            4601 => Some(Language::Nhc),
            4602 => Some(Language::Nhd),
            4603 => Some(Language::Nhe),
            4604 => Some(Language::Nhf),
            4605 => Some(Language::Nhg),
            4606 => Some(Language::Nhh),
            4607 => Some(Language::Nhi),
            4608 => Some(Language::Nhk),
            4609 => Some(Language::Nhm),
            4610 => Some(Language::Nhn),
            4611 => Some(Language::Nho),
            4612 => Some(Language::Nhp),
            4613 => Some(Language::Nhq),
            4614 => Some(Language::Nhr),
            4615 => Some(Language::Nht),
            4616 => Some(Language::Nhu),
            4617 => Some(Language::Nhv),
            4618 => Some(Language::Nhw),
            4619 => Some(Language::Nhx),
            4620 => Some(Language::Nhy),
            4621 => Some(Language::Nhz),
            4622 => Some(Language::Nia),
            4623 => Some(Language::Nib),
            4624 => Some(Language::Nid),
            4625 => Some(Language::Nie),
            4626 => Some(Language::Nif),
            4627 => Some(Language::Nig),
            4628 => Some(Language::Nih),
            4629 => Some(Language::Nii),
            4630 => Some(Language::Nij),
            4631 => Some(Language::Nik),
            4632 => Some(Language::Nil),
            4633 => Some(Language::Nim),
            4634 => Some(Language::Nin),
            4635 => Some(Language::Nio),
            4636 => Some(Language::Niq),
            4637 => Some(Language::Nir),
            4638 => Some(Language::Nis),
            4639 => Some(Language::Nit),
            4640 => Some(Language::Niu),
            4641 => Some(Language::Niv),
            4642 => Some(Language::Niw),
            4643 => Some(Language::Nix),
            4644 => Some(Language::Niy),
            4645 => Some(Language::Niz),
            4646 => Some(Language::Nja),
            4647 => Some(Language::Njb),
            4648 => Some(Language::Njd),
            4649 => Some(Language::Njh),
            4650 => Some(Language::Nji),
            4651 => Some(Language::Njj),
            4652 => Some(Language::Njl),
            4653 => Some(Language::Njm),
            4654 => Some(Language::Njn),
            4655 => Some(Language::Njo),
            4656 => Some(Language::Njr),
            4657 => Some(Language::Njs),
            4658 => Some(Language::Njt),
            4659 => Some(Language::Nju),
            4660 => Some(Language::Njx),
            4661 => Some(Language::Njy),
            4662 => Some(Language::Njz),
            4663 => Some(Language::Nka),
            4664 => Some(Language::Nkb),
            4665 => Some(Language::Nkc),
            4666 => Some(Language::Nkd),
            4667 => Some(Language::Nke),
            4668 => Some(Language::Nkf),
            4669 => Some(Language::Nkg),
            4670 => Some(Language::Nkh),
            4671 => Some(Language::Nki),
            4672 => Some(Language::Nkj),
            4673 => Some(Language::Nkk),
            4674 => Some(Language::Nkm),
            4675 => Some(Language::Nkn),
            4676 => Some(Language::Nko),
            4677 => Some(Language::Nkp),
            4678 => Some(Language::Nkq),
            4679 => Some(Language::Nkr),
            4680 => Some(Language::Nks),
            4681 => Some(Language::Nkt),
            4682 => Some(Language::Nku),
            4683 => Some(Language::Nkv),
            4684 => Some(Language::Nkw),
            4685 => Some(Language::Nkx),
            4686 => Some(Language::Nkz),
            4687 => Some(Language::Nla),
            4688 => Some(Language::Nlc),
            4689 => Some(Language::Nld),
            4690 => Some(Language::Nle),
            4691 => Some(Language::Nlg),
            4692 => Some(Language::Nli),
            4693 => Some(Language::Nlj),
            4694 => Some(Language::Nlk),
            4695 => Some(Language::Nll),
            4696 => Some(Language::Nlm),
            4697 => Some(Language::Nlo),
            4698 => Some(Language::Nlq),
            4699 => Some(Language::Nlu),
            4700 => Some(Language::Nlv),
            4701 => Some(Language::Nlw),
            4702 => Some(Language::Nlx),
            4703 => Some(Language::Nly),
            4704 => Some(Language::Nlz),
            4705 => Some(Language::Nma),
            4706 => Some(Language::Nmb),
            4707 => Some(Language::Nmc),
            4708 => Some(Language::Nmd),
            4709 => Some(Language::Nme),
            4710 => Some(Language::Nmf),
            4711 => Some(Language::Nmg),
            4712 => Some(Language::Nmh),
            4713 => Some(Language::Nmi),
            4714 => Some(Language::Nmj),
            4715 => Some(Language::Nmk),
            4716 => Some(Language::Nml),
            4717 => Some(Language::Nmm),
            4718 => Some(Language::Nmn),
            4719 => Some(Language::Nmo),
            4720 => Some(Language::Nmp),
            4721 => Some(Language::Nmq),
            4722 => Some(Language::Nmr),
            4723 => Some(Language::Nms),
            4724 => Some(Language::Nmt),
            4725 => Some(Language::Nmu),
            4726 => Some(Language::Nmv),
            4727 => Some(Language::Nmw),
            4728 => Some(Language::Nmx),
            4729 => Some(Language::Nmy),
            4730 => Some(Language::Nmz),
            4731 => Some(Language::Nna),
            4732 => Some(Language::Nnb),
            4733 => Some(Language::Nnc),
            4734 => Some(Language::Nnd),
            4735 => Some(Language::Nne),
            4736 => Some(Language::Nnf),
            4737 => Some(Language::Nng),
            4738 => Some(Language::Nnh),
            4739 => Some(Language::Nni),
            4740 => Some(Language::Nnj),
            4741 => Some(Language::Nnk),
            4742 => Some(Language::Nnl),
            4743 => Some(Language::Nnm),
            4744 => Some(Language::Nnn),
            4745 => Some(Language::Nno),
            4746 => Some(Language::Nnp),
            4747 => Some(Language::Nnq),
            4748 => Some(Language::Nnr),
            4749 => Some(Language::Nnt),
            4750 => Some(Language::Nnu),
            4751 => Some(Language::Nnv),
            4752 => Some(Language::Nnw),
            4753 => Some(Language::Nny),
            4754 => Some(Language::Nnz),
            4755 => Some(Language::Noa),
            4756 => Some(Language::Nob),
            4757 => Some(Language::Noc),
            4758 => Some(Language::Nod),
            4759 => Some(Language::Noe),
            4760 => Some(Language::Nof),
            4761 => Some(Language::Nog),
            4762 => Some(Language::Noh),
            4763 => Some(Language::Noi),
            4764 => Some(Language::Noj),
            4765 => Some(Language::Nok),
            4766 => Some(Language::Nol),
            4767 => Some(Language::Nom),
            4768 => Some(Language::Non),
            4769 => Some(Language::Nop),
            4770 => Some(Language::Noq),
            4771 => Some(Language::Nor),
            4772 => Some(Language::Nos),
            4773 => Some(Language::Not),
            4774 => Some(Language::Nou),
            4775 => Some(Language::Nov),
            4776 => Some(Language::Now),
            4777 => Some(Language::Noy),
            4778 => Some(Language::Noz),
            4779 => Some(Language::Npa),
            4780 => Some(Language::Npb),
            4781 => Some(Language::Npg),
            4782 => Some(Language::Nph),
            4783 => Some(Language::Npi),
            4784 => Some(Language::Npl),
            4785 => Some(Language::Npn),
            4786 => Some(Language::Npo),
            4787 => Some(Language::Nps),
            4788 => Some(Language::Npu),
            4789 => Some(Language::Npx),
            4790 => Some(Language::Npy),
            4791 => Some(Language::Nqg),
            4792 => Some(Language::Nqk),
            4793 => Some(Language::Nql),
            4794 => Some(Language::Nqm),
            4795 => Some(Language::Nqn),
            4796 => Some(Language::Nqo),
            4797 => Some(Language::Nqq),
            4798 => Some(Language::Nqt),
            4799 => Some(Language::Nqy),
            4800 => Some(Language::Nra),
            4801 => Some(Language::Nrb),
            4802 => Some(Language::Nrc),
            4803 => Some(Language::Nre),
            4804 => Some(Language::Nrf),
            4805 => Some(Language::Nrg),
            4806 => Some(Language::Nri),
            4807 => Some(Language::Nrk),
            4808 => Some(Language::Nrl),
            4809 => Some(Language::Nrm),
            4810 => Some(Language::Nrn),
            4811 => Some(Language::Nrp),
            4812 => Some(Language::Nrr),
            4813 => Some(Language::Nrt),
            4814 => Some(Language::Nru),
            4815 => Some(Language::Nrx),
            4816 => Some(Language::Nrz),
            4817 => Some(Language::Nsa),
            4818 => Some(Language::Nsb),
            4819 => Some(Language::Nsc),
            4820 => Some(Language::Nsd),
            4821 => Some(Language::Nse),
            4822 => Some(Language::Nsf),
            4823 => Some(Language::Nsg),
            4824 => Some(Language::Nsh),
            4825 => Some(Language::Nsi),
            4826 => Some(Language::Nsk),
            4827 => Some(Language::Nsl),
            4828 => Some(Language::Nsm),
            4829 => Some(Language::Nsn),
            4830 => Some(Language::Nso),
            4831 => Some(Language::Nsp),
            4832 => Some(Language::Nsq),
            4833 => Some(Language::Nsr),
            4834 => Some(Language::Nss),
            4835 => Some(Language::Nst),
            4836 => Some(Language::Nsu),
            4837 => Some(Language::Nsv),
            4838 => Some(Language::Nsw),
            4839 => Some(Language::Nsx),
            4840 => Some(Language::Nsy),
            4841 => Some(Language::Nsz),
            4842 => Some(Language::Ntd),
            4843 => Some(Language::Nte),
            4844 => Some(Language::Ntg),
            4845 => Some(Language::Nti),
            4846 => Some(Language::Ntj),
            4847 => Some(Language::Ntk),
            4848 => Some(Language::Ntm),
            4849 => Some(Language::Nto),
            4850 => Some(Language::Ntp),
            4851 => Some(Language::Ntr),
            4852 => Some(Language::Ntu),
            4853 => Some(Language::Ntw),
            4854 => Some(Language::Ntx),
            4855 => Some(Language::Nty),
            4856 => Some(Language::Ntz),
            4857 => Some(Language::Nua),
            4858 => Some(Language::Nuc),
            4859 => Some(Language::Nud),
            4860 => Some(Language::Nue),
            4861 => Some(Language::Nuf),
            4862 => Some(Language::Nug),
            4863 => Some(Language::Nuh),
            4864 => Some(Language::Nui),
            4865 => Some(Language::Nuj),
            4866 => Some(Language::Nuk),
            4867 => Some(Language::Nul),
            4868 => Some(Language::Num),
            4869 => Some(Language::Nun),
            4870 => Some(Language::Nuo),
            4871 => Some(Language::Nup),
            4872 => Some(Language::Nuq),
            4873 => Some(Language::Nur),
            4874 => Some(Language::Nus),
            4875 => Some(Language::Nut),
            4876 => Some(Language::Nuu),
            4877 => Some(Language::Nuv),
            4878 => Some(Language::Nuw),
            4879 => Some(Language::Nux),
            4880 => Some(Language::Nuy),
            4881 => Some(Language::Nuz),
            4882 => Some(Language::Nvh),
            4883 => Some(Language::Nvm),
            4884 => Some(Language::Nvo),
            4885 => Some(Language::Nwa),
            4886 => Some(Language::Nwb),
            4887 => Some(Language::Nwc),
            4888 => Some(Language::Nwe),
            4889 => Some(Language::Nwg),
            4890 => Some(Language::Nwi),
            4891 => Some(Language::Nwm),
            4892 => Some(Language::Nwo),
            4893 => Some(Language::Nwr),
            4894 => Some(Language::Nww),
            4895 => Some(Language::Nwx),
            4896 => Some(Language::Nwy),
            4897 => Some(Language::Nxa),
            4898 => Some(Language::Nxd),
            4899 => Some(Language::Nxe),
            4900 => Some(Language::Nxg),
            4901 => Some(Language::Nxi),
            4902 => Some(Language::Nxk),
            4903 => Some(Language::Nxl),
            4904 => Some(Language::Nxm),
            4905 => Some(Language::Nxn),
            4906 => Some(Language::Nxo),
            4907 => Some(Language::Nxq),
            4908 => Some(Language::Nxr),
            4909 => Some(Language::Nxx),
            4910 => Some(Language::Nya),
            4911 => Some(Language::Nyb),
            4912 => Some(Language::Nyc),
            4913 => Some(Language::Nyd),
            4914 => Some(Language::Nye),
            4915 => Some(Language::Nyf),
            4916 => Some(Language::Nyg),
            4917 => Some(Language::Nyh),
            4918 => Some(Language::Nyi),
            4919 => Some(Language::Nyj),
            4920 => Some(Language::Nyk),
            4921 => Some(Language::Nyl),
            4922 => Some(Language::Nym),
            4923 => Some(Language::Nyn),
            4924 => Some(Language::Nyo),
            4925 => Some(Language::Nyp),
            4926 => Some(Language::Nyq),
            4927 => Some(Language::Nyr),
            4928 => Some(Language::Nys),
            4929 => Some(Language::Nyt),
            4930 => Some(Language::Nyu),
            4931 => Some(Language::Nyv),
            4932 => Some(Language::Nyw),
            4933 => Some(Language::Nyx),
            4934 => Some(Language::Nyy),
            4935 => Some(Language::Nza),
            4936 => Some(Language::Nzb),
            4937 => Some(Language::Nzd),
            4938 => Some(Language::Nzi),
            4939 => Some(Language::Nzk),
            4940 => Some(Language::Nzm),
            4941 => Some(Language::Nzs),
            4942 => Some(Language::Nzu),
            4943 => Some(Language::Nzy),
            4944 => Some(Language::Nzz),
            4945 => Some(Language::Oaa),
            4946 => Some(Language::Oac),
            4947 => Some(Language::Oar),
            4948 => Some(Language::Oav),
            4949 => Some(Language::Obi),
            4950 => Some(Language::Obk),
            4951 => Some(Language::Obl),
            4952 => Some(Language::Obm),
            4953 => Some(Language::Obo),
            4954 => Some(Language::Obr),
            4955 => Some(Language::Obt),
            4956 => Some(Language::Obu),
            4957 => Some(Language::Oca),
            4958 => Some(Language::Och),
            4959 => Some(Language::Oci),
            4960 => Some(Language::Ocm),
            4961 => Some(Language::Oco),
            4962 => Some(Language::Ocu),
            4963 => Some(Language::Oda),
            4964 => Some(Language::Odk),
            4965 => Some(Language::Odt),
            4966 => Some(Language::Odu),
            4967 => Some(Language::Ofo),
            4968 => Some(Language::Ofs),
            4969 => Some(Language::Ofu),
            4970 => Some(Language::Ogb),
            4971 => Some(Language::Ogc),
            4972 => Some(Language::Oge),
            4973 => Some(Language::Ogg),
            4974 => Some(Language::Ogo),
            4975 => Some(Language::Ogu),
            4976 => Some(Language::Oht),
            4977 => Some(Language::Ohu),
            4978 => Some(Language::Oia),
            4979 => Some(Language::Oie),
            4980 => Some(Language::Oin),
            4981 => Some(Language::Ojb),
            4982 => Some(Language::Ojc),
            4983 => Some(Language::Ojg),
            4984 => Some(Language::Oji),
            4985 => Some(Language::Ojp),
            4986 => Some(Language::Ojs),
            4987 => Some(Language::Ojv),
            4988 => Some(Language::Ojw),
            4989 => Some(Language::Oka),
            4990 => Some(Language::Okb),
            4991 => Some(Language::Okc),
            4992 => Some(Language::Okd),
            4993 => Some(Language::Oke),
            4994 => Some(Language::Okg),
            4995 => Some(Language::Okh),
            4996 => Some(Language::Oki),
            4997 => Some(Language::Okj),
            4998 => Some(Language::Okk),
            4999 => Some(Language::Okl),
            5000 => Some(Language::Okm),
            5001 => Some(Language::Okn),
            5002 => Some(Language::Oko),
            5003 => Some(Language::Okr),
            5004 => Some(Language::Oks),
            5005 => Some(Language::Oku),
            5006 => Some(Language::Okv),
            5007 => Some(Language::Okx),
            5008 => Some(Language::Okz),
            5009 => Some(Language::Ola),
            5010 => Some(Language::Old),
            5011 => Some(Language::Ole),
            5012 => Some(Language::Olk),
            5013 => Some(Language::Olm),
            5014 => Some(Language::Olo),
            5015 => Some(Language::Olr),
            5016 => Some(Language::Olt),
            5017 => Some(Language::Olu),
            5018 => Some(Language::Oma),
            5019 => Some(Language::Omb),
            5020 => Some(Language::Omc),
            5021 => Some(Language::Omg),
            5022 => Some(Language::Omi),
            5023 => Some(Language::Omk),
            5024 => Some(Language::Oml),
            5025 => Some(Language::Omn),
            5026 => Some(Language::Omo),
            5027 => Some(Language::Omp),
            5028 => Some(Language::Omr),
            5029 => Some(Language::Omt),
            5030 => Some(Language::Omu),
            5031 => Some(Language::Omw),
            5032 => Some(Language::Omx),
            5033 => Some(Language::Omy),
            5034 => Some(Language::Ona),
            5035 => Some(Language::Onb),
            5036 => Some(Language::One),
            5037 => Some(Language::Ong),
            5038 => Some(Language::Oni),
            5039 => Some(Language::Onj),
            5040 => Some(Language::Onk),
            5041 => Some(Language::Onn),
            5042 => Some(Language::Ono),
            5043 => Some(Language::Onp),
            5044 => Some(Language::Onr),
            5045 => Some(Language::Ons),
            5046 => Some(Language::Ont),
            5047 => Some(Language::Onu),
            5048 => Some(Language::Onw),
            5049 => Some(Language::Onx),
            5050 => Some(Language::Ood),
            5051 => Some(Language::Oog),
            5052 => Some(Language::Oon),
            5053 => Some(Language::Oor),
            5054 => Some(Language::Oos),
            5055 => Some(Language::Opa),
            5056 => Some(Language::Opk),
            5057 => Some(Language::Opm),
            5058 => Some(Language::Opo),
            5059 => Some(Language::Opt),
            5060 => Some(Language::Opy),
            5061 => Some(Language::Ora),
            5062 => Some(Language::Orc),
            5063 => Some(Language::Ore),
            5064 => Some(Language::Org),
            5065 => Some(Language::Orh),
            5066 => Some(Language::Ori),
            5067 => Some(Language::Orm),
            5068 => Some(Language::Orn),
            5069 => Some(Language::Oro),
            5070 => Some(Language::Orr),
            5071 => Some(Language::Ors),
            5072 => Some(Language::Ort),
            5073 => Some(Language::Oru),
            5074 => Some(Language::Orv),
            5075 => Some(Language::Orw),
            5076 => Some(Language::Orx),
            5077 => Some(Language::Ory),
            5078 => Some(Language::Orz),
            5079 => Some(Language::Osa),
            5080 => Some(Language::Osc),
            5081 => Some(Language::Osi),
            5082 => Some(Language::Osn),
            5083 => Some(Language::Oso),
            5084 => Some(Language::Osp),
            5085 => Some(Language::Oss),
            5086 => Some(Language::Ost),
            5087 => Some(Language::Osu),
            5088 => Some(Language::Osx),
            5089 => Some(Language::Ota),
            5090 => Some(Language::Otb),
            5091 => Some(Language::Otd),
            5092 => Some(Language::Ote),
            5093 => Some(Language::Oti),
            5094 => Some(Language::Otk),
            5095 => Some(Language::Otl),
            5096 => Some(Language::Otm),
            5097 => Some(Language::Otn),
            5098 => Some(Language::Otq),
            5099 => Some(Language::Otr),
            5100 => Some(Language::Ots),
            5101 => Some(Language::Ott),
            5102 => Some(Language::Otu),
            5103 => Some(Language::Otw),
            5104 => Some(Language::Otx),
            5105 => Some(Language::Oty),
            5106 => Some(Language::Otz),
            5107 => Some(Language::Oua),
            5108 => Some(Language::Oub),
            5109 => Some(Language::Oue),
            5110 => Some(Language::Oui),
            5111 => Some(Language::Oum),
            5112 => Some(Language::Ovd),
            5113 => Some(Language::Owi),
            5114 => Some(Language::Owl),
            5115 => Some(Language::Oyb),
            5116 => Some(Language::Oyd),
            5117 => Some(Language::Oym),
            5118 => Some(Language::Oyy),
            5119 => Some(Language::Ozm),
            5120 => Some(Language::Pab),
            5121 => Some(Language::Pac),
            5122 => Some(Language::Pad),
            5123 => Some(Language::Pae),
            5124 => Some(Language::Paf),
            5125 => Some(Language::Pag),
            5126 => Some(Language::Pah),
            5127 => Some(Language::Pai),
            5128 => Some(Language::Pak),
            5129 => Some(Language::Pal),
            5130 => Some(Language::Pam),
            5131 => Some(Language::Pan),
            5132 => Some(Language::Pao),
            5133 => Some(Language::Pap),
            5134 => Some(Language::Paq),
            5135 => Some(Language::Par),
            5136 => Some(Language::Pas),
            5137 => Some(Language::Pau),
            5138 => Some(Language::Pav),
            5139 => Some(Language::Paw),
            5140 => Some(Language::Pax),
            5141 => Some(Language::Pay),
            5142 => Some(Language::Paz),
            5143 => Some(Language::Pbb),
            5144 => Some(Language::Pbc),
            5145 => Some(Language::Pbe),
            5146 => Some(Language::Pbf),
            5147 => Some(Language::Pbg),
            5148 => Some(Language::Pbh),
            5149 => Some(Language::Pbi),
            5150 => Some(Language::Pbl),
            5151 => Some(Language::Pbm),
            5152 => Some(Language::Pbn),
            5153 => Some(Language::Pbo),
            5154 => Some(Language::Pbp),
            5155 => Some(Language::Pbr),
            5156 => Some(Language::Pbs),
            5157 => Some(Language::Pbt),
            5158 => Some(Language::Pbu),
            5159 => Some(Language::Pbv),
            5160 => Some(Language::Pby),
            5161 => Some(Language::Pca),
            5162 => Some(Language::Pcb),
            5163 => Some(Language::Pcc),
            5164 => Some(Language::Pcd),
            5165 => Some(Language::Pce),
            5166 => Some(Language::Pcf),
            5167 => Some(Language::Pcg),
            5168 => Some(Language::Pch),
            5169 => Some(Language::Pci),
            5170 => Some(Language::Pcj),
            5171 => Some(Language::Pck),
            5172 => Some(Language::Pcl),
            5173 => Some(Language::Pcm),
            5174 => Some(Language::Pcn),
            5175 => Some(Language::Pcp),
            5176 => Some(Language::Pcw),
            5177 => Some(Language::Pda),
            5178 => Some(Language::Pdc),
            5179 => Some(Language::Pdi),
            5180 => Some(Language::Pdn),
            5181 => Some(Language::Pdo),
            5182 => Some(Language::Pdt),
            5183 => Some(Language::Pdu),
            5184 => Some(Language::Pea),
            5185 => Some(Language::Peb),
            5186 => Some(Language::Ped),
            5187 => Some(Language::Pee),
            5188 => Some(Language::Pef),
            5189 => Some(Language::Peg),
            5190 => Some(Language::Peh),
            5191 => Some(Language::Pei),
            5192 => Some(Language::Pej),
            5193 => Some(Language::Pek),
            5194 => Some(Language::Pel),
            5195 => Some(Language::Pem),
            5196 => Some(Language::Peo),
            5197 => Some(Language::Pep),
            5198 => Some(Language::Peq),
            5199 => Some(Language::Pes),
            5200 => Some(Language::Pev),
            5201 => Some(Language::Pex),
            5202 => Some(Language::Pey),
            5203 => Some(Language::Pez),
            5204 => Some(Language::Pfa),
            5205 => Some(Language::Pfe),
            5206 => Some(Language::Pfl),
            5207 => Some(Language::Pga),
            5208 => Some(Language::Pgd),
            5209 => Some(Language::Pgg),
            5210 => Some(Language::Pgi),
            5211 => Some(Language::Pgk),
            5212 => Some(Language::Pgl),
            5213 => Some(Language::Pgn),
            5214 => Some(Language::Pgs),
            5215 => Some(Language::Pgu),
            5216 => Some(Language::Pgz),
            5217 => Some(Language::Pha),
            5218 => Some(Language::Phd),
            5219 => Some(Language::Phg),
            5220 => Some(Language::Phh),
            5221 => Some(Language::Phj),
            5222 => Some(Language::Phk),
            5223 => Some(Language::Phl),
            5224 => Some(Language::Phm),
            5225 => Some(Language::Phn),
            5226 => Some(Language::Pho),
            5227 => Some(Language::Phq),
            5228 => Some(Language::Phr),
            5229 => Some(Language::Pht),
            5230 => Some(Language::Phu),
            5231 => Some(Language::Phv),
            5232 => Some(Language::Phw),
            5233 => Some(Language::Pia),
            5234 => Some(Language::Pib),
            5235 => Some(Language::Pic),
            5236 => Some(Language::Pid),
            5237 => Some(Language::Pie),
            5238 => Some(Language::Pif),
            5239 => Some(Language::Pig),
            5240 => Some(Language::Pih),
            5241 => Some(Language::Pij),
            5242 => Some(Language::Pil),
            5243 => Some(Language::Pim),
            5244 => Some(Language::Pin),
            5245 => Some(Language::Pio),
            5246 => Some(Language::Pip),
            5247 => Some(Language::Pir),
            5248 => Some(Language::Pis),
            5249 => Some(Language::Pit),
            5250 => Some(Language::Piu),
            5251 => Some(Language::Piv),
            5252 => Some(Language::Piw),
            5253 => Some(Language::Pix),
            5254 => Some(Language::Piy),
            5255 => Some(Language::Piz),
            5256 => Some(Language::Pjt),
            5257 => Some(Language::Pka),
            5258 => Some(Language::Pkb),
            5259 => Some(Language::Pkc),
            5260 => Some(Language::Pkg),
            5261 => Some(Language::Pkh),
            5262 => Some(Language::Pkn),
            5263 => Some(Language::Pko),
            5264 => Some(Language::Pkp),
            5265 => Some(Language::Pkr),
            5266 => Some(Language::Pks),
            5267 => Some(Language::Pkt),
            5268 => Some(Language::Pku),
            5269 => Some(Language::Pla),
            5270 => Some(Language::Plb),
            5271 => Some(Language::Plc),
            5272 => Some(Language::Pld),
            5273 => Some(Language::Ple),
            5274 => Some(Language::Plg),
            5275 => Some(Language::Plh),
            5276 => Some(Language::Pli),
            5277 => Some(Language::Plj),
            5278 => Some(Language::Plk),
            5279 => Some(Language::Pll),
            5280 => Some(Language::Pln),
            5281 => Some(Language::Plo),
            5282 => Some(Language::Plq),
            5283 => Some(Language::Plr),
            5284 => Some(Language::Pls),
            5285 => Some(Language::Plt),
            5286 => Some(Language::Plu),
            5287 => Some(Language::Plv),
            5288 => Some(Language::Plw),
            5289 => Some(Language::Ply),
            5290 => Some(Language::Plz),
            5291 => Some(Language::Pma),
            5292 => Some(Language::Pmb),
            5293 => Some(Language::Pmd),
            5294 => Some(Language::Pme),
            5295 => Some(Language::Pmf),
            5296 => Some(Language::Pmh),
            5297 => Some(Language::Pmi),
            5298 => Some(Language::Pmj),
            5299 => Some(Language::Pmk),
            5300 => Some(Language::Pml),
            5301 => Some(Language::Pmm),
            5302 => Some(Language::Pmn),
            5303 => Some(Language::Pmo),
            5304 => Some(Language::Pmq),
            5305 => Some(Language::Pmr),
            5306 => Some(Language::Pms),
            5307 => Some(Language::Pmt),
            5308 => Some(Language::Pmw),
            5309 => Some(Language::Pmx),
            5310 => Some(Language::Pmy),
            5311 => Some(Language::Pmz),
            5312 => Some(Language::Pna),
            5313 => Some(Language::Pnb),
            5314 => Some(Language::Pnc),
            5315 => Some(Language::Pnd),
            5316 => Some(Language::Pne),
            5317 => Some(Language::Png),
            5318 => Some(Language::Pnh),
            5319 => Some(Language::Pni),
            5320 => Some(Language::Pnj),
            5321 => Some(Language::Pnk),
            5322 => Some(Language::Pnl),
            5323 => Some(Language::Pnm),
            5324 => Some(Language::Pnn),
            5325 => Some(Language::Pno),
            5326 => Some(Language::Pnp),
            5327 => Some(Language::Pnq),
            5328 => Some(Language::Pnr),
            5329 => Some(Language::Pns),
            5330 => Some(Language::Pnt),
            5331 => Some(Language::Pnu),
            5332 => Some(Language::Pnv),
            5333 => Some(Language::Pnw),
            5334 => Some(Language::Pnx),
            5335 => Some(Language::Pny),
            5336 => Some(Language::Pnz),
            5337 => Some(Language::Poc),
            5338 => Some(Language::Poe),
            5339 => Some(Language::Pof),
            5340 => Some(Language::Pog),
            5341 => Some(Language::Poh),
            5342 => Some(Language::Poi),
            5343 => Some(Language::Pok),
            5344 => Some(Language::Pol),
            5345 => Some(Language::Pom),
            5346 => Some(Language::Pon),
            5347 => Some(Language::Poo),
            5348 => Some(Language::Pop),
            5349 => Some(Language::Poq),
            5350 => Some(Language::Por),
            5351 => Some(Language::Pos),
            5352 => Some(Language::Pot),
            5353 => Some(Language::Pov),
            5354 => Some(Language::Pow),
            5355 => Some(Language::Pox),
            5356 => Some(Language::Poy),
            5357 => Some(Language::Ppe),
            5358 => Some(Language::Ppi),
            5359 => Some(Language::Ppk),
            5360 => Some(Language::Ppl),
            5361 => Some(Language::Ppm),
            5362 => Some(Language::Ppn),
            5363 => Some(Language::Ppo),
            5364 => Some(Language::Ppp),
            5365 => Some(Language::Ppq),
            5366 => Some(Language::Pps),
            5367 => Some(Language::Ppt),
            5368 => Some(Language::Ppu),
            5369 => Some(Language::Pqa),
            5370 => Some(Language::Pqm),
            5371 => Some(Language::Prc),
            5372 => Some(Language::Prd),
            5373 => Some(Language::Pre),
            5374 => Some(Language::Prf),
            5375 => Some(Language::Prg),
            5376 => Some(Language::Prh),
            5377 => Some(Language::Pri),
            5378 => Some(Language::Prk),
            5379 => Some(Language::Prl),
            5380 => Some(Language::Prm),
            5381 => Some(Language::Prn),
            5382 => Some(Language::Pro),
            5383 => Some(Language::Prp),
            5384 => Some(Language::Prq),
            5385 => Some(Language::Prr),
            5386 => Some(Language::Prs),
            5387 => Some(Language::Prt),
            5388 => Some(Language::Pru),
            5389 => Some(Language::Prw),
            5390 => Some(Language::Prx),
            5391 => Some(Language::Prz),
            5392 => Some(Language::Psa),
            5393 => Some(Language::Psc),
            5394 => Some(Language::Psd),
            5395 => Some(Language::Pse),
            5396 => Some(Language::Psg),
            5397 => Some(Language::Psh),
            5398 => Some(Language::Psi),
            5399 => Some(Language::Psl),
            5400 => Some(Language::Psm),
            5401 => Some(Language::Psn),
            5402 => Some(Language::Pso),
            5403 => Some(Language::Psp),
            5404 => Some(Language::Psq),
            5405 => Some(Language::Psr),
            5406 => Some(Language::Pss),
            5407 => Some(Language::Pst),
            5408 => Some(Language::Psu),
            5409 => Some(Language::Psw),
            5410 => Some(Language::Psy),
            5411 => Some(Language::Pta),
            5412 => Some(Language::Pth),
            5413 => Some(Language::Pti),
            5414 => Some(Language::Ptn),
            5415 => Some(Language::Pto),
            5416 => Some(Language::Ptp),
            5417 => Some(Language::Ptq),
            5418 => Some(Language::Ptr),
            5419 => Some(Language::Ptt),
            5420 => Some(Language::Ptu),
            5421 => Some(Language::Ptv),
            5422 => Some(Language::Ptw),
            5423 => Some(Language::Pty),
            5424 => Some(Language::Pua),
            5425 => Some(Language::Pub),
            5426 => Some(Language::Puc),
            5427 => Some(Language::Pud),
            5428 => Some(Language::Pue),
            5429 => Some(Language::Puf),
            5430 => Some(Language::Pug),
            5431 => Some(Language::Pui),
            5432 => Some(Language::Puj),
            5433 => Some(Language::Pum),
            5434 => Some(Language::Puo),
            5435 => Some(Language::Pup),
            5436 => Some(Language::Puq),
            5437 => Some(Language::Pur),
            5438 => Some(Language::Pus),
            5439 => Some(Language::Put),
            5440 => Some(Language::Puu),
            5441 => Some(Language::Puw),
            5442 => Some(Language::Pux),
            5443 => Some(Language::Puy),
            5444 => Some(Language::Pwa),
            5445 => Some(Language::Pwb),
            5446 => Some(Language::Pwg),
            5447 => Some(Language::Pwi),
            5448 => Some(Language::Pwm),
            5449 => Some(Language::Pwn),
            5450 => Some(Language::Pwo),
            5451 => Some(Language::Pwr),
            5452 => Some(Language::Pww),
            5453 => Some(Language::Pxm),
            5454 => Some(Language::Pye),
            5455 => Some(Language::Pym),
            5456 => Some(Language::Pyn),
            5457 => Some(Language::Pys),
            5458 => Some(Language::Pyu),
            5459 => Some(Language::Pyx),
            5460 => Some(Language::Pyy),
            5461 => Some(Language::Pzh),
            5462 => Some(Language::Pzn),
            5463 => Some(Language::Qua),
            5464 => Some(Language::Qub),
            5465 => Some(Language::Quc),
            5466 => Some(Language::Qud),
            5467 => Some(Language::Que),
            5468 => Some(Language::Quf),
            5469 => Some(Language::Qug),
            5470 => Some(Language::Quh),
            5471 => Some(Language::Qui),
            5472 => Some(Language::Quk),
            5473 => Some(Language::Qul),
            5474 => Some(Language::Qum),
            5475 => Some(Language::Qun),
            5476 => Some(Language::Qup),
            5477 => Some(Language::Quq),
            5478 => Some(Language::Qur),
            5479 => Some(Language::Qus),
            5480 => Some(Language::Quv),
            5481 => Some(Language::Quw),
            5482 => Some(Language::Qux),
            5483 => Some(Language::Quy),
            5484 => Some(Language::Quz),
            5485 => Some(Language::Qva),
            5486 => Some(Language::Qvc),
            5487 => Some(Language::Qve),
            5488 => Some(Language::Qvh),
            5489 => Some(Language::Qvi),
            5490 => Some(Language::Qvj),
            5491 => Some(Language::Qvl),
            5492 => Some(Language::Qvm),
            5493 => Some(Language::Qvn),
            5494 => Some(Language::Qvo),
            5495 => Some(Language::Qvp),
            5496 => Some(Language::Qvs),
            5497 => Some(Language::Qvw),
            5498 => Some(Language::Qvy),
            5499 => Some(Language::Qvz),
            5500 => Some(Language::Qwa),
            5501 => Some(Language::Qwc),
            5502 => Some(Language::Qwh),
            5503 => Some(Language::Qwm),
            5504 => Some(Language::Qws),
            5505 => Some(Language::Qwt),
            5506 => Some(Language::Qxa),
            5507 => Some(Language::Qxc),
            5508 => Some(Language::Qxh),
            5509 => Some(Language::Qxl),
            5510 => Some(Language::Qxn),
            5511 => Some(Language::Qxo),
            5512 => Some(Language::Qxp),
            5513 => Some(Language::Qxq),
            5514 => Some(Language::Qxr),
            5515 => Some(Language::Qxs),
            5516 => Some(Language::Qxt),
            5517 => Some(Language::Qxu),
            5518 => Some(Language::Qxw),
            5519 => Some(Language::Qya),
            5520 => Some(Language::Qyp),
            5521 => Some(Language::Raa),
            5522 => Some(Language::Rab),
            5523 => Some(Language::Rac),
            5524 => Some(Language::Rad),
            5525 => Some(Language::Raf),
            5526 => Some(Language::Rag),
            5527 => Some(Language::Rah),
            5528 => Some(Language::Rai),
            5529 => Some(Language::Raj),
            5530 => Some(Language::Rak),
            5531 => Some(Language::Ral),
            5532 => Some(Language::Ram),
            5533 => Some(Language::Ran),
            5534 => Some(Language::Rao),
            5535 => Some(Language::Rap),
            5536 => Some(Language::Raq),
            5537 => Some(Language::Rar),
            5538 => Some(Language::Ras),
            5539 => Some(Language::Rat),
            5540 => Some(Language::Rau),
            5541 => Some(Language::Rav),
            5542 => Some(Language::Raw),
            5543 => Some(Language::Rax),
            5544 => Some(Language::Ray),
            5545 => Some(Language::Raz),
            5546 => Some(Language::Rbb),
            5547 => Some(Language::Rbk),
            5548 => Some(Language::Rbl),
            5549 => Some(Language::Rbp),
            5550 => Some(Language::Rcf),
            5551 => Some(Language::Rdb),
            5552 => Some(Language::Rea),
            5553 => Some(Language::Reb),
            5554 => Some(Language::Ree),
            5555 => Some(Language::Reg),
            5556 => Some(Language::Rei),
            5557 => Some(Language::Rej),
            5558 => Some(Language::Rel),
            5559 => Some(Language::Rem),
            5560 => Some(Language::Ren),
            5561 => Some(Language::Rer),
            5562 => Some(Language::Res),
            5563 => Some(Language::Ret),
            5564 => Some(Language::Rey),
            5565 => Some(Language::Rga),
            5566 => Some(Language::Rge),
            5567 => Some(Language::Rgk),
            5568 => Some(Language::Rgn),
            5569 => Some(Language::Rgr),
            5570 => Some(Language::Rgs),
            5571 => Some(Language::Rgu),
            5572 => Some(Language::Rhg),
            5573 => Some(Language::Rhp),
            5574 => Some(Language::Ria),
            5575 => Some(Language::Rib),
            5576 => Some(Language::Rif),
            5577 => Some(Language::Ril),
            5578 => Some(Language::Rim),
            5579 => Some(Language::Rin),
            5580 => Some(Language::Rir),
            5581 => Some(Language::Rit),
            5582 => Some(Language::Riu),
            5583 => Some(Language::Rjg),
            5584 => Some(Language::Rji),
            5585 => Some(Language::Rjs),
            5586 => Some(Language::Rka),
            5587 => Some(Language::Rkb),
            5588 => Some(Language::Rkh),
            5589 => Some(Language::Rki),
            5590 => Some(Language::Rkm),
            5591 => Some(Language::Rkt),
            5592 => Some(Language::Rkw),
            5593 => Some(Language::Rma),
            5594 => Some(Language::Rmb),
            5595 => Some(Language::Rmc),
            5596 => Some(Language::Rmd),
            5597 => Some(Language::Rme),
            5598 => Some(Language::Rmf),
            5599 => Some(Language::Rmg),
            5600 => Some(Language::Rmh),
            5601 => Some(Language::Rmi),
            5602 => Some(Language::Rmk),
            5603 => Some(Language::Rml),
            5604 => Some(Language::Rmm),
            5605 => Some(Language::Rmn),
            5606 => Some(Language::Rmo),
            5607 => Some(Language::Rmp),
            5608 => Some(Language::Rmq),
            5609 => Some(Language::Rms),
            5610 => Some(Language::Rmt),
            5611 => Some(Language::Rmu),
            5612 => Some(Language::Rmv),
            5613 => Some(Language::Rmw),
            5614 => Some(Language::Rmx),
            5615 => Some(Language::Rmy),
            5616 => Some(Language::Rmz),
            5617 => Some(Language::Rnb),
            5618 => Some(Language::Rnd),
            5619 => Some(Language::Rng),
            5620 => Some(Language::Rnl),
            5621 => Some(Language::Rnn),
            5622 => Some(Language::Rnp),
            5623 => Some(Language::Rnr),
            5624 => Some(Language::Rnw),
            5625 => Some(Language::Rob),
            5626 => Some(Language::Roc),
            5627 => Some(Language::Rod),
            5628 => Some(Language::Roe),
            5629 => Some(Language::Rof),
            5630 => Some(Language::Rog),
            5631 => Some(Language::Roh),
            5632 => Some(Language::Rol),
            5633 => Some(Language::Rom),
            5634 => Some(Language::Ron),
            5635 => Some(Language::Roo),
            5636 => Some(Language::Rop),
            5637 => Some(Language::Ror),
            5638 => Some(Language::Rou),
            5639 => Some(Language::Row),
            5640 => Some(Language::Rpn),
            5641 => Some(Language::Rpt),
            5642 => Some(Language::Rri),
            5643 => Some(Language::Rro),
            5644 => Some(Language::Rrt),
            5645 => Some(Language::Rsb),
            5646 => Some(Language::Rsk),
            5647 => Some(Language::Rsl),
            5648 => Some(Language::Rsm),
            5649 => Some(Language::Rsn),
            5650 => Some(Language::Rtc),
            5651 => Some(Language::Rth),
            5652 => Some(Language::Rtm),
            5653 => Some(Language::Rts),
            5654 => Some(Language::Rtw),
            5655 => Some(Language::Rub),
            5656 => Some(Language::Ruc),
            5657 => Some(Language::Rue),
            5658 => Some(Language::Ruf),
            5659 => Some(Language::Rug),
            5660 => Some(Language::Ruh),
            5661 => Some(Language::Rui),
            5662 => Some(Language::Ruk),
            5663 => Some(Language::Run),
            5664 => Some(Language::Ruo),
            5665 => Some(Language::Rup),
            5666 => Some(Language::Ruq),
            5667 => Some(Language::Rus),
            5668 => Some(Language::Rut),
            5669 => Some(Language::Ruu),
            5670 => Some(Language::Ruy),
            5671 => Some(Language::Ruz),
            5672 => Some(Language::Rwa),
            5673 => Some(Language::Rwk),
            5674 => Some(Language::Rwl),
            5675 => Some(Language::Rwm),
            5676 => Some(Language::Rwo),
            5677 => Some(Language::Rwr),
            5678 => Some(Language::Rxd),
            5679 => Some(Language::Rxw),
            5680 => Some(Language::Ryn),
            5681 => Some(Language::Rys),
            5682 => Some(Language::Ryu),
            5683 => Some(Language::Rzh),
            5684 => Some(Language::Saa),
            5685 => Some(Language::Sab),
            5686 => Some(Language::Sac),
            5687 => Some(Language::Sad),
            5688 => Some(Language::Sae),
            5689 => Some(Language::Saf),
            5690 => Some(Language::Sag),
            5691 => Some(Language::Sah),
            5692 => Some(Language::Saj),
            5693 => Some(Language::Sak),
            5694 => Some(Language::Sam),
            5695 => Some(Language::San),
            5696 => Some(Language::Sao),
            5697 => Some(Language::Saq),
            5698 => Some(Language::Sar),
            5699 => Some(Language::Sas),
            5700 => Some(Language::Sat),
            5701 => Some(Language::Sau),
            5702 => Some(Language::Sav),
            5703 => Some(Language::Saw),
            5704 => Some(Language::Sax),
            5705 => Some(Language::Say),
            5706 => Some(Language::Saz),
            5707 => Some(Language::Sba),
            5708 => Some(Language::Sbb),
            5709 => Some(Language::Sbc),
            5710 => Some(Language::Sbd),
            5711 => Some(Language::Sbe),
            5712 => Some(Language::Sbf),
            5713 => Some(Language::Sbg),
            5714 => Some(Language::Sbh),
            5715 => Some(Language::Sbi),
            5716 => Some(Language::Sbj),
            5717 => Some(Language::Sbk),
            5718 => Some(Language::Sbl),
            5719 => Some(Language::Sbm),
            5720 => Some(Language::Sbn),
            5721 => Some(Language::Sbo),
            5722 => Some(Language::Sbp),
            5723 => Some(Language::Sbq),
            5724 => Some(Language::Sbr),
            5725 => Some(Language::Sbs),
            5726 => Some(Language::Sbt),
            5727 => Some(Language::Sbu),
            5728 => Some(Language::Sbv),
            5729 => Some(Language::Sbw),
            5730 => Some(Language::Sbx),
            5731 => Some(Language::Sby),
            5732 => Some(Language::Sbz),
            5733 => Some(Language::Scb),
            5734 => Some(Language::Sce),
            5735 => Some(Language::Scf),
            5736 => Some(Language::Scg),
            5737 => Some(Language::Sch),
            5738 => Some(Language::Sci),
            5739 => Some(Language::Sck),
            5740 => Some(Language::Scl),
            5741 => Some(Language::Scn),
            5742 => Some(Language::Sco),
            5743 => Some(Language::Scp),
            5744 => Some(Language::Scq),
            5745 => Some(Language::Scs),
            5746 => Some(Language::Sct),
            5747 => Some(Language::Scu),
            5748 => Some(Language::Scv),
            5749 => Some(Language::Scw),
            5750 => Some(Language::Scx),
            5751 => Some(Language::Sda),
            5752 => Some(Language::Sdb),
            5753 => Some(Language::Sdc),
            5754 => Some(Language::Sde),
            5755 => Some(Language::Sdf),
            5756 => Some(Language::Sdg),
            5757 => Some(Language::Sdh),
            5758 => Some(Language::Sdj),
            5759 => Some(Language::Sdk),
            5760 => Some(Language::Sdl),
            5761 => Some(Language::Sdn),
            5762 => Some(Language::Sdo),
            5763 => Some(Language::Sdp),
            5764 => Some(Language::Sdq),
            5765 => Some(Language::Sdr),
            5766 => Some(Language::Sds),
            5767 => Some(Language::Sdt),
            5768 => Some(Language::Sdu),
            5769 => Some(Language::Sdx),
            5770 => Some(Language::Sdz),
            5771 => Some(Language::Sea),
            5772 => Some(Language::Seb),
            5773 => Some(Language::Sec),
            5774 => Some(Language::Sed),
            5775 => Some(Language::See),
            5776 => Some(Language::Sef),
            5777 => Some(Language::Seg),
            5778 => Some(Language::Seh),
            5779 => Some(Language::Sei),
            5780 => Some(Language::Sej),
            5781 => Some(Language::Sek),
            5782 => Some(Language::Sel),
            5783 => Some(Language::Sen),
            5784 => Some(Language::Seo),
            5785 => Some(Language::Sep),
            5786 => Some(Language::Seq),
            5787 => Some(Language::Ser),
            5788 => Some(Language::Ses),
            5789 => Some(Language::Set),
            5790 => Some(Language::Seu),
            5791 => Some(Language::Sev),
            5792 => Some(Language::Sew),
            5793 => Some(Language::Sey),
            5794 => Some(Language::Sez),
            5795 => Some(Language::Sfb),
            5796 => Some(Language::Sfe),
            5797 => Some(Language::Sfm),
            5798 => Some(Language::Sfs),
            5799 => Some(Language::Sfw),
            5800 => Some(Language::Sga),
            5801 => Some(Language::Sgb),
            5802 => Some(Language::Sgc),
            5803 => Some(Language::Sgd),
            5804 => Some(Language::Sge),
            5805 => Some(Language::Sgg),
            5806 => Some(Language::Sgh),
            5807 => Some(Language::Sgi),
            5808 => Some(Language::Sgj),
            5809 => Some(Language::Sgk),
            5810 => Some(Language::Sgm),
            5811 => Some(Language::Sgp),
            5812 => Some(Language::Sgr),
            5813 => Some(Language::Sgs),
            5814 => Some(Language::Sgt),
            5815 => Some(Language::Sgu),
            5816 => Some(Language::Sgw),
            5817 => Some(Language::Sgx),
            5818 => Some(Language::Sgy),
            5819 => Some(Language::Sgz),
            5820 => Some(Language::Sha),
            5821 => Some(Language::Shb),
            5822 => Some(Language::Shc),
            5823 => Some(Language::Shd),
            5824 => Some(Language::She),
            5825 => Some(Language::Shg),
            5826 => Some(Language::Shh),
            5827 => Some(Language::Shi),
            5828 => Some(Language::Shj),
            5829 => Some(Language::Shk),
            5830 => Some(Language::Shl),
            5831 => Some(Language::Shm),
            5832 => Some(Language::Shn),
            5833 => Some(Language::Sho),
            5834 => Some(Language::Shp),
            5835 => Some(Language::Shq),
            5836 => Some(Language::Shr),
            5837 => Some(Language::Shs),
            5838 => Some(Language::Sht),
            5839 => Some(Language::Shu),
            5840 => Some(Language::Shv),
            5841 => Some(Language::Shw),
            5842 => Some(Language::Shx),
            5843 => Some(Language::Shy),
            5844 => Some(Language::Shz),
            5845 => Some(Language::Sia),
            5846 => Some(Language::Sib),
            5847 => Some(Language::Sid),
            5848 => Some(Language::Sie),
            5849 => Some(Language::Sif),
            5850 => Some(Language::Sig),
            5851 => Some(Language::Sih),
            5852 => Some(Language::Sii),
            5853 => Some(Language::Sij),
            5854 => Some(Language::Sik),
            5855 => Some(Language::Sil),
            5856 => Some(Language::Sim),
            5857 => Some(Language::Sin),
            5858 => Some(Language::Sip),
            5859 => Some(Language::Siq),
            5860 => Some(Language::Sir),
            5861 => Some(Language::Sis),
            5862 => Some(Language::Siu),
            5863 => Some(Language::Siv),
            5864 => Some(Language::Siw),
            5865 => Some(Language::Six),
            5866 => Some(Language::Siy),
            5867 => Some(Language::Siz),
            5868 => Some(Language::Sja),
            5869 => Some(Language::Sjb),
            5870 => Some(Language::Sjd),
            5871 => Some(Language::Sje),
            5872 => Some(Language::Sjg),
            5873 => Some(Language::Sjk),
            5874 => Some(Language::Sjl),
            5875 => Some(Language::Sjm),
            5876 => Some(Language::Sjn),
            5877 => Some(Language::Sjo),
            5878 => Some(Language::Sjp),
            5879 => Some(Language::Sjr),
            5880 => Some(Language::Sjs),
            5881 => Some(Language::Sjt),
            5882 => Some(Language::Sju),
            5883 => Some(Language::Sjw),
            5884 => Some(Language::Ska),
            5885 => Some(Language::Skb),
            5886 => Some(Language::Skc),
            5887 => Some(Language::Skd),
            5888 => Some(Language::Ske),
            5889 => Some(Language::Skf),
            5890 => Some(Language::Skg),
            5891 => Some(Language::Skh),
            5892 => Some(Language::Ski),
            5893 => Some(Language::Skj),
            5894 => Some(Language::Skm),
            5895 => Some(Language::Skn),
            5896 => Some(Language::Sko),
            5897 => Some(Language::Skp),
            5898 => Some(Language::Skq),
            5899 => Some(Language::Skr),
            5900 => Some(Language::Sks),
            5901 => Some(Language::Skt),
            5902 => Some(Language::Sku),
            5903 => Some(Language::Skv),
            5904 => Some(Language::Skw),
            5905 => Some(Language::Skx),
            5906 => Some(Language::Sky),
            5907 => Some(Language::Skz),
            5908 => Some(Language::Slc),
            5909 => Some(Language::Sld),
            5910 => Some(Language::Sle),
            5911 => Some(Language::Slf),
            5912 => Some(Language::Slg),
            5913 => Some(Language::Slh),
            5914 => Some(Language::Sli),
            5915 => Some(Language::Slj),
            5916 => Some(Language::Slk),
            5917 => Some(Language::Sll),
            5918 => Some(Language::Slm),
            5919 => Some(Language::Sln),
            5920 => Some(Language::Slp),
            5921 => Some(Language::Slq),
            5922 => Some(Language::Slr),
            5923 => Some(Language::Sls),
            5924 => Some(Language::Slt),
            5925 => Some(Language::Slu),
            5926 => Some(Language::Slv),
            5927 => Some(Language::Slw),
            5928 => Some(Language::Slx),
            5929 => Some(Language::Sly),
            5930 => Some(Language::Slz),
            5931 => Some(Language::Sma),
            5932 => Some(Language::Smb),
            5933 => Some(Language::Smc),
            5934 => Some(Language::Sme),
            5935 => Some(Language::Smf),
            5936 => Some(Language::Smg),
            5937 => Some(Language::Smh),
            5938 => Some(Language::Smj),
            5939 => Some(Language::Smk),
            5940 => Some(Language::Sml),
            5941 => Some(Language::Smm),
            5942 => Some(Language::Smn),
            5943 => Some(Language::Smo),
            5944 => Some(Language::Smp),
            5945 => Some(Language::Smq),
            5946 => Some(Language::Smr),
            5947 => Some(Language::Sms),
            5948 => Some(Language::Smt),
            5949 => Some(Language::Smu),
            5950 => Some(Language::Smv),
            5951 => Some(Language::Smw),
            5952 => Some(Language::Smx),
            5953 => Some(Language::Smy),
            5954 => Some(Language::Smz),
            5955 => Some(Language::Sna),
            5956 => Some(Language::Snc),
            5957 => Some(Language::Snd),
            5958 => Some(Language::Sne),
            5959 => Some(Language::Snf),
            5960 => Some(Language::Sng),
            5961 => Some(Language::Sni),
            5962 => Some(Language::Snj),
            5963 => Some(Language::Snk),
            5964 => Some(Language::Snl),
            5965 => Some(Language::Snm),
            5966 => Some(Language::Snn),
            5967 => Some(Language::Sno),
            5968 => Some(Language::Snp),
            5969 => Some(Language::Snq),
            5970 => Some(Language::Snr),
            5971 => Some(Language::Sns),
            5972 => Some(Language::Snu),
            5973 => Some(Language::Snv),
            5974 => Some(Language::Snw),
            5975 => Some(Language::Snx),
            5976 => Some(Language::Sny),
            5977 => Some(Language::Snz),
            5978 => Some(Language::Soa),
            5979 => Some(Language::Sob),
            5980 => Some(Language::Soc),
            5981 => Some(Language::Sod),
            5982 => Some(Language::Soe),
            5983 => Some(Language::Sog),
            5984 => Some(Language::Soh),
            5985 => Some(Language::Soi),
            5986 => Some(Language::Soj),
            5987 => Some(Language::Sok),
            5988 => Some(Language::Sol),
            5989 => Some(Language::Som),
            5990 => Some(Language::Soo),
            5991 => Some(Language::Sop),
            5992 => Some(Language::Soq),
            5993 => Some(Language::Sor),
            5994 => Some(Language::Sos),
            5995 => Some(Language::Sot),
            5996 => Some(Language::Sou),
            5997 => Some(Language::Sov),
            5998 => Some(Language::Sow),
            5999 => Some(Language::Sox),
            6000 => Some(Language::Soy),
            6001 => Some(Language::Soz),
            6002 => Some(Language::Spa),
            6003 => Some(Language::Spb),
            6004 => Some(Language::Spc),
            6005 => Some(Language::Spd),
            6006 => Some(Language::Spe),
            6007 => Some(Language::Spg),
            6008 => Some(Language::Spi),
            6009 => Some(Language::Spk),
            6010 => Some(Language::Spl),
            6011 => Some(Language::Spm),
            6012 => Some(Language::Spn),
            6013 => Some(Language::Spo),
            6014 => Some(Language::Spp),
            6015 => Some(Language::Spq),
            6016 => Some(Language::Spr),
            6017 => Some(Language::Sps),
            6018 => Some(Language::Spt),
            6019 => Some(Language::Spu),
            6020 => Some(Language::Spv),
            6021 => Some(Language::Spx),
            6022 => Some(Language::Spy),
            6023 => Some(Language::Sqa),
            6024 => Some(Language::Sqh),
            6025 => Some(Language::Sqi),
            6026 => Some(Language::Sqk),
            6027 => Some(Language::Sqm),
            6028 => Some(Language::Sqn),
            6029 => Some(Language::Sqo),
            6030 => Some(Language::Sqq),
            6031 => Some(Language::Sqr),
            6032 => Some(Language::Sqs),
            6033 => Some(Language::Sqt),
            6034 => Some(Language::Squ),
            6035 => Some(Language::Sqx),
            6036 => Some(Language::Sra),
            6037 => Some(Language::Srb),
            6038 => Some(Language::Src),
            6039 => Some(Language::Srd),
            6040 => Some(Language::Sre),
            6041 => Some(Language::Srf),
            6042 => Some(Language::Srg),
            6043 => Some(Language::Srh),
            6044 => Some(Language::Sri),
            6045 => Some(Language::Srk),
            6046 => Some(Language::Srl),
            6047 => Some(Language::Srm),
            6048 => Some(Language::Srn),
            6049 => Some(Language::Sro),
            6050 => Some(Language::Srp),
            6051 => Some(Language::Srq),
            6052 => Some(Language::Srr),
            6053 => Some(Language::Srs),
            6054 => Some(Language::Srt),
            6055 => Some(Language::Sru),
            6056 => Some(Language::Srv),
            6057 => Some(Language::Srw),
            6058 => Some(Language::Srx),
            6059 => Some(Language::Sry),
            6060 => Some(Language::Srz),
            6061 => Some(Language::Ssb),
            6062 => Some(Language::Ssc),
            6063 => Some(Language::Ssd),
            6064 => Some(Language::Sse),
            6065 => Some(Language::Ssf),
            6066 => Some(Language::Ssg),
            6067 => Some(Language::Ssh),
            6068 => Some(Language::Ssi),
            6069 => Some(Language::Ssj),
            6070 => Some(Language::Ssk),
            6071 => Some(Language::Ssl),
            6072 => Some(Language::Ssm),
            6073 => Some(Language::Ssn),
            6074 => Some(Language::Sso),
            6075 => Some(Language::Ssp),
            6076 => Some(Language::Ssq),
            6077 => Some(Language::Ssr),
            6078 => Some(Language::Sss),
            6079 => Some(Language::Sst),
            6080 => Some(Language::Ssu),
            6081 => Some(Language::Ssv),
            6082 => Some(Language::Ssw),
            6083 => Some(Language::Ssx),
            6084 => Some(Language::Ssy),
            6085 => Some(Language::Ssz),
            6086 => Some(Language::Sta),
            6087 => Some(Language::Stb),
            6088 => Some(Language::Std),
            6089 => Some(Language::Ste),
            6090 => Some(Language::Stf),
            6091 => Some(Language::Stg),
            6092 => Some(Language::Sth),
            6093 => Some(Language::Sti),
            6094 => Some(Language::Stj),
            6095 => Some(Language::Stk),
            6096 => Some(Language::Stl),
            6097 => Some(Language::Stm),
            6098 => Some(Language::Stn),
            6099 => Some(Language::Sto),
            6100 => Some(Language::Stp),
            6101 => Some(Language::Stq),
            6102 => Some(Language::Str),
            6103 => Some(Language::Sts),
            6104 => Some(Language::Stt),
            6105 => Some(Language::Stu),
            6106 => Some(Language::Stv),
            6107 => Some(Language::Stw),
            6108 => Some(Language::Sty),
            6109 => Some(Language::Sua),
            6110 => Some(Language::Sub),
            6111 => Some(Language::Suc),
            6112 => Some(Language::Sue),
            6113 => Some(Language::Sug),
            6114 => Some(Language::Sui),
            6115 => Some(Language::Suj),
            6116 => Some(Language::Suk),
            6117 => Some(Language::Sun),
            6118 => Some(Language::Suo),
            6119 => Some(Language::Suq),
            6120 => Some(Language::Sur),
            6121 => Some(Language::Sus),
            6122 => Some(Language::Sut),
            6123 => Some(Language::Suv),
            6124 => Some(Language::Suw),
            6125 => Some(Language::Sux),
            6126 => Some(Language::Suy),
            6127 => Some(Language::Suz),
            6128 => Some(Language::Sva),
            6129 => Some(Language::Svb),
            6130 => Some(Language::Svc),
            6131 => Some(Language::Sve),
            6132 => Some(Language::Svk),
            6133 => Some(Language::Svm),
            6134 => Some(Language::Svs),
            6135 => Some(Language::Svx),
            6136 => Some(Language::Swa),
            6137 => Some(Language::Swb),
            6138 => Some(Language::Swc),
            6139 => Some(Language::Swe),
            6140 => Some(Language::Swf),
            6141 => Some(Language::Swg),
            6142 => Some(Language::Swh),
            6143 => Some(Language::Swi),
            6144 => Some(Language::Swj),
            6145 => Some(Language::Swk),
            6146 => Some(Language::Swl),
            6147 => Some(Language::Swm),
            6148 => Some(Language::Swn),
            6149 => Some(Language::Swo),
            6150 => Some(Language::Swp),
            6151 => Some(Language::Swq),
            6152 => Some(Language::Swr),
            6153 => Some(Language::Sws),
            6154 => Some(Language::Swt),
            6155 => Some(Language::Swu),
            6156 => Some(Language::Swv),
            6157 => Some(Language::Sww),
            6158 => Some(Language::Swx),
            6159 => Some(Language::Swy),
            6160 => Some(Language::Sxb),
            6161 => Some(Language::Sxc),
            6162 => Some(Language::Sxe),
            6163 => Some(Language::Sxg),
            6164 => Some(Language::Sxk),
            6165 => Some(Language::Sxl),
            6166 => Some(Language::Sxm),
            6167 => Some(Language::Sxn),
            6168 => Some(Language::Sxo),
            6169 => Some(Language::Sxr),
            6170 => Some(Language::Sxs),
            6171 => Some(Language::Sxu),
            6172 => Some(Language::Sxw),
            6173 => Some(Language::Sya),
            6174 => Some(Language::Syb),
            6175 => Some(Language::Syc),
            6176 => Some(Language::Syi),
            6177 => Some(Language::Syk),
            6178 => Some(Language::Syl),
            6179 => Some(Language::Sym),
            6180 => Some(Language::Syn),
            6181 => Some(Language::Syo),
            6182 => Some(Language::Syr),
            6183 => Some(Language::Sys),
            6184 => Some(Language::Syw),
            6185 => Some(Language::Syx),
            6186 => Some(Language::Syy),
            6187 => Some(Language::Sza),
            6188 => Some(Language::Szb),
            6189 => Some(Language::Szc),
            6190 => Some(Language::Szd),
            6191 => Some(Language::Sze),
            6192 => Some(Language::Szg),
            6193 => Some(Language::Szl),
            6194 => Some(Language::Szn),
            6195 => Some(Language::Szp),
            6196 => Some(Language::Szs),
            6197 => Some(Language::Szv),
            6198 => Some(Language::Szw),
            6199 => Some(Language::Szy),
            6200 => Some(Language::Taa),
            6201 => Some(Language::Tab),
            6202 => Some(Language::Tac),
            6203 => Some(Language::Tad),
            6204 => Some(Language::Tae),
            6205 => Some(Language::Taf),
            6206 => Some(Language::Tag),
            6207 => Some(Language::Tah),
            6208 => Some(Language::Taj),
            6209 => Some(Language::Tak),
            6210 => Some(Language::Tal),
            6211 => Some(Language::Tam),
            6212 => Some(Language::Tan),
            6213 => Some(Language::Tao),
            6214 => Some(Language::Tap),
            6215 => Some(Language::Taq),
            6216 => Some(Language::Tar),
            6217 => Some(Language::Tas),
            6218 => Some(Language::Tat),
            6219 => Some(Language::Tau),
            6220 => Some(Language::Tav),
            6221 => Some(Language::Taw),
            6222 => Some(Language::Tax),
            6223 => Some(Language::Tay),
            6224 => Some(Language::Taz),
            6225 => Some(Language::Tba),
            6226 => Some(Language::Tbc),
            6227 => Some(Language::Tbd),
            6228 => Some(Language::Tbe),
            6229 => Some(Language::Tbf),
            6230 => Some(Language::Tbg),
            6231 => Some(Language::Tbh),
            6232 => Some(Language::Tbi),
            6233 => Some(Language::Tbj),
            6234 => Some(Language::Tbk),
            6235 => Some(Language::Tbl),
            6236 => Some(Language::Tbm),
            6237 => Some(Language::Tbn),
            6238 => Some(Language::Tbo),
            6239 => Some(Language::Tbp),
            6240 => Some(Language::Tbr),
            6241 => Some(Language::Tbs),
            6242 => Some(Language::Tbt),
            6243 => Some(Language::Tbu),
            6244 => Some(Language::Tbv),
            6245 => Some(Language::Tbw),
            6246 => Some(Language::Tbx),
            6247 => Some(Language::Tby),
            6248 => Some(Language::Tbz),
            6249 => Some(Language::Tca),
            6250 => Some(Language::Tcb),
            6251 => Some(Language::Tcc),
            6252 => Some(Language::Tcd),
            6253 => Some(Language::Tce),
            6254 => Some(Language::Tcf),
            6255 => Some(Language::Tcg),
            6256 => Some(Language::Tch),
            6257 => Some(Language::Tci),
            6258 => Some(Language::Tck),
            6259 => Some(Language::Tcl),
            6260 => Some(Language::Tcm),
            6261 => Some(Language::Tcn),
            6262 => Some(Language::Tco),
            6263 => Some(Language::Tcp),
            6264 => Some(Language::Tcq),
            6265 => Some(Language::Tcs),
            6266 => Some(Language::Tct),
            6267 => Some(Language::Tcu),
            6268 => Some(Language::Tcw),
            6269 => Some(Language::Tcx),
            6270 => Some(Language::Tcy),
            6271 => Some(Language::Tcz),
            6272 => Some(Language::Tda),
            6273 => Some(Language::Tdb),
            6274 => Some(Language::Tdc),
            6275 => Some(Language::Tdd),
            6276 => Some(Language::Tde),
            6277 => Some(Language::Tdf),
            6278 => Some(Language::Tdg),
            6279 => Some(Language::Tdh),
            6280 => Some(Language::Tdi),
            6281 => Some(Language::Tdj),
            6282 => Some(Language::Tdk),
            6283 => Some(Language::Tdl),
            6284 => Some(Language::Tdm),
            6285 => Some(Language::Tdn),
            6286 => Some(Language::Tdo),
            6287 => Some(Language::Tdq),
            6288 => Some(Language::Tdr),
            6289 => Some(Language::Tds),
            6290 => Some(Language::Tdt),
            6291 => Some(Language::Tdv),
            6292 => Some(Language::Tdx),
            6293 => Some(Language::Tdy),
            6294 => Some(Language::Tea),
            6295 => Some(Language::Teb),
            6296 => Some(Language::Tec),
            6297 => Some(Language::Ted),
            6298 => Some(Language::Tee),
            6299 => Some(Language::Tef),
            6300 => Some(Language::Teg),
            6301 => Some(Language::Teh),
            6302 => Some(Language::Tei),
            6303 => Some(Language::Tek),
            6304 => Some(Language::Tel),
            6305 => Some(Language::Tem),
            6306 => Some(Language::Ten),
            6307 => Some(Language::Teo),
            6308 => Some(Language::Tep),
            6309 => Some(Language::Teq),
            6310 => Some(Language::Ter),
            6311 => Some(Language::Tes),
            6312 => Some(Language::Tet),
            6313 => Some(Language::Teu),
            6314 => Some(Language::Tev),
            6315 => Some(Language::Tew),
            6316 => Some(Language::Tex),
            6317 => Some(Language::Tey),
            6318 => Some(Language::Tez),
            6319 => Some(Language::Tfi),
            6320 => Some(Language::Tfn),
            6321 => Some(Language::Tfo),
            6322 => Some(Language::Tfr),
            6323 => Some(Language::Tft),
            6324 => Some(Language::Tga),
            6325 => Some(Language::Tgb),
            6326 => Some(Language::Tgc),
            6327 => Some(Language::Tgd),
            6328 => Some(Language::Tge),
            6329 => Some(Language::Tgf),
            6330 => Some(Language::Tgh),
            6331 => Some(Language::Tgi),
            6332 => Some(Language::Tgj),
            6333 => Some(Language::Tgk),
            6334 => Some(Language::Tgl),
            6335 => Some(Language::Tgn),
            6336 => Some(Language::Tgo),
            6337 => Some(Language::Tgp),
            6338 => Some(Language::Tgq),
            6339 => Some(Language::Tgr),
            6340 => Some(Language::Tgs),
            6341 => Some(Language::Tgt),
            6342 => Some(Language::Tgu),
            6343 => Some(Language::Tgv),
            6344 => Some(Language::Tgw),
            6345 => Some(Language::Tgx),
            6346 => Some(Language::Tgy),
            6347 => Some(Language::Tgz),
            6348 => Some(Language::Tha),
            6349 => Some(Language::Thd),
            6350 => Some(Language::The),
            6351 => Some(Language::Thf),
            6352 => Some(Language::Thh),
            6353 => Some(Language::Thi),
            6354 => Some(Language::Thk),
            6355 => Some(Language::Thl),
            6356 => Some(Language::Thm),
            6357 => Some(Language::Thn),
            6358 => Some(Language::Thp),
            6359 => Some(Language::Thq),
            6360 => Some(Language::Thr),
            6361 => Some(Language::Ths),
            6362 => Some(Language::Tht),
            6363 => Some(Language::Thu),
            6364 => Some(Language::Thv),
            6365 => Some(Language::Thy),
            6366 => Some(Language::Thz),
            6367 => Some(Language::Tia),
            6368 => Some(Language::Tic),
            6369 => Some(Language::Tif),
            6370 => Some(Language::Tig),
            6371 => Some(Language::Tih),
            6372 => Some(Language::Tii),
            6373 => Some(Language::Tij),
            6374 => Some(Language::Tik),
            6375 => Some(Language::Til),
            6376 => Some(Language::Tim),
            6377 => Some(Language::Tin),
            6378 => Some(Language::Tio),
            6379 => Some(Language::Tip),
            6380 => Some(Language::Tiq),
            6381 => Some(Language::Tir),
            6382 => Some(Language::Tis),
            6383 => Some(Language::Tit),
            6384 => Some(Language::Tiu),
            6385 => Some(Language::Tiv),
            6386 => Some(Language::Tiw),
            6387 => Some(Language::Tix),
            6388 => Some(Language::Tiy),
            6389 => Some(Language::Tiz),
            6390 => Some(Language::Tja),
            6391 => Some(Language::Tjg),
            6392 => Some(Language::Tji),
            6393 => Some(Language::Tjj),
            6394 => Some(Language::Tjl),
            6395 => Some(Language::Tjm),
            6396 => Some(Language::Tjn),
            6397 => Some(Language::Tjo),
            6398 => Some(Language::Tjp),
            6399 => Some(Language::Tjs),
            6400 => Some(Language::Tju),
            6401 => Some(Language::Tjw),
            6402 => Some(Language::Tka),
            6403 => Some(Language::Tkb),
            6404 => Some(Language::Tkd),
            6405 => Some(Language::Tke),
            6406 => Some(Language::Tkf),
            6407 => Some(Language::Tkg),
            6408 => Some(Language::Tkl),
            6409 => Some(Language::Tkm),
            6410 => Some(Language::Tkn),
            6411 => Some(Language::Tkp),
            6412 => Some(Language::Tkq),
            6413 => Some(Language::Tkr),
            6414 => Some(Language::Tks),
            6415 => Some(Language::Tkt),
            6416 => Some(Language::Tku),
            6417 => Some(Language::Tkv),
            6418 => Some(Language::Tkw),
            6419 => Some(Language::Tkx),
            6420 => Some(Language::Tkz),
            6421 => Some(Language::Tla),
            6422 => Some(Language::Tlb),
            6423 => Some(Language::Tlc),
            6424 => Some(Language::Tld),
            6425 => Some(Language::Tlf),
            6426 => Some(Language::Tlg),
            6427 => Some(Language::Tlh),
            6428 => Some(Language::Tli),
            6429 => Some(Language::Tlj),
            6430 => Some(Language::Tlk),
            6431 => Some(Language::Tll),
            6432 => Some(Language::Tlm),
            6433 => Some(Language::Tln),
            6434 => Some(Language::Tlo),
            6435 => Some(Language::Tlp),
            6436 => Some(Language::Tlq),
            6437 => Some(Language::Tlr),
            6438 => Some(Language::Tls),
            6439 => Some(Language::Tlt),
            6440 => Some(Language::Tlu),
            6441 => Some(Language::Tlv),
            6442 => Some(Language::Tlx),
            6443 => Some(Language::Tly),
            6444 => Some(Language::Tma),
            6445 => Some(Language::Tmb),
            6446 => Some(Language::Tmc),
            6447 => Some(Language::Tmd),
            6448 => Some(Language::Tme),
            6449 => Some(Language::Tmf),
            6450 => Some(Language::Tmg),
            6451 => Some(Language::Tmh),
            6452 => Some(Language::Tmi),
            6453 => Some(Language::Tmj),
            6454 => Some(Language::Tmk),
            6455 => Some(Language::Tml),
            6456 => Some(Language::Tmm),
            6457 => Some(Language::Tmn),
            6458 => Some(Language::Tmo),
            6459 => Some(Language::Tmq),
            6460 => Some(Language::Tmr),
            6461 => Some(Language::Tms),
            6462 => Some(Language::Tmt),
            6463 => Some(Language::Tmu),
            6464 => Some(Language::Tmv),
            6465 => Some(Language::Tmw),
            6466 => Some(Language::Tmy),
            6467 => Some(Language::Tmz),
            6468 => Some(Language::Tna),
            6469 => Some(Language::Tnb),
            6470 => Some(Language::Tnc),
            6471 => Some(Language::Tnd),
            6472 => Some(Language::Tng),
            6473 => Some(Language::Tnh),
            6474 => Some(Language::Tni),
            6475 => Some(Language::Tnk),
            6476 => Some(Language::Tnl),
            6477 => Some(Language::Tnm),
            6478 => Some(Language::Tnn),
            6479 => Some(Language::Tno),
            6480 => Some(Language::Tnp),
            6481 => Some(Language::Tnq),
            6482 => Some(Language::Tnr),
            6483 => Some(Language::Tns),
            6484 => Some(Language::Tnt),
            6485 => Some(Language::Tnu),
            6486 => Some(Language::Tnv),
            6487 => Some(Language::Tnw),
            6488 => Some(Language::Tnx),
            6489 => Some(Language::Tny),
            6490 => Some(Language::Tnz),
            6491 => Some(Language::Tob),
            6492 => Some(Language::Toc),
            6493 => Some(Language::Tod),
            6494 => Some(Language::Tof),
            6495 => Some(Language::Tog),
            6496 => Some(Language::Toh),
            6497 => Some(Language::Toi),
            6498 => Some(Language::Toj),
            6499 => Some(Language::Tok),
            6500 => Some(Language::Tol),
            6501 => Some(Language::Tom),
            6502 => Some(Language::Ton),
            6503 => Some(Language::Too),
            6504 => Some(Language::Top),
            6505 => Some(Language::Toq),
            6506 => Some(Language::Tor),
            6507 => Some(Language::Tos),
            6508 => Some(Language::Tou),
            6509 => Some(Language::Tov),
            6510 => Some(Language::Tow),
            6511 => Some(Language::Tox),
            6512 => Some(Language::Toy),
            6513 => Some(Language::Toz),
            6514 => Some(Language::Tpa),
            6515 => Some(Language::Tpc),
            6516 => Some(Language::Tpe),
            6517 => Some(Language::Tpf),
            6518 => Some(Language::Tpg),
            6519 => Some(Language::Tpi),
            6520 => Some(Language::Tpj),
            6521 => Some(Language::Tpk),
            6522 => Some(Language::Tpl),
            6523 => Some(Language::Tpm),
            6524 => Some(Language::Tpn),
            6525 => Some(Language::Tpo),
            6526 => Some(Language::Tpp),
            6527 => Some(Language::Tpq),
            6528 => Some(Language::Tpr),
            6529 => Some(Language::Tpt),
            6530 => Some(Language::Tpu),
            6531 => Some(Language::Tpv),
            6532 => Some(Language::Tpw),
            6533 => Some(Language::Tpx),
            6534 => Some(Language::Tpy),
            6535 => Some(Language::Tpz),
            6536 => Some(Language::Tqb),
            6537 => Some(Language::Tql),
            6538 => Some(Language::Tqm),
            6539 => Some(Language::Tqn),
            6540 => Some(Language::Tqo),
            6541 => Some(Language::Tqp),
            6542 => Some(Language::Tqq),
            6543 => Some(Language::Tqr),
            6544 => Some(Language::Tqt),
            6545 => Some(Language::Tqu),
            6546 => Some(Language::Tqw),
            6547 => Some(Language::Tra),
            6548 => Some(Language::Trb),
            6549 => Some(Language::Trc),
            6550 => Some(Language::Trd),
            6551 => Some(Language::Tre),
            6552 => Some(Language::Trf),
            6553 => Some(Language::Trg),
            6554 => Some(Language::Trh),
            6555 => Some(Language::Tri),
            6556 => Some(Language::Trj),
            6557 => Some(Language::Trl),
            6558 => Some(Language::Trm),
            6559 => Some(Language::Trn),
            6560 => Some(Language::Tro),
            6561 => Some(Language::Trp),
            6562 => Some(Language::Trq),
            6563 => Some(Language::Trr),
            6564 => Some(Language::Trs),
            6565 => Some(Language::Trt),
            6566 => Some(Language::Tru),
            6567 => Some(Language::Trv),
            6568 => Some(Language::Trw),
            6569 => Some(Language::Trx),
            6570 => Some(Language::Try),
            6571 => Some(Language::Trz),
            6572 => Some(Language::Tsa),
            6573 => Some(Language::Tsb),
            6574 => Some(Language::Tsc),
            6575 => Some(Language::Tsd),
            6576 => Some(Language::Tse),
            6577 => Some(Language::Tsg),
            6578 => Some(Language::Tsh),
            6579 => Some(Language::Tsi),
            6580 => Some(Language::Tsj),
            6581 => Some(Language::Tsk),
            6582 => Some(Language::Tsl),
            6583 => Some(Language::Tsm),
            6584 => Some(Language::Tsn),
            6585 => Some(Language::Tso),
            6586 => Some(Language::Tsp),
            6587 => Some(Language::Tsq),
            6588 => Some(Language::Tsr),
            6589 => Some(Language::Tss),
            6590 => Some(Language::Tst),
            6591 => Some(Language::Tsu),
            6592 => Some(Language::Tsv),
            6593 => Some(Language::Tsw),
            6594 => Some(Language::Tsx),
            6595 => Some(Language::Tsy),
            6596 => Some(Language::Tsz),
            6597 => Some(Language::Tta),
            6598 => Some(Language::Ttb),
            6599 => Some(Language::Ttc),
            6600 => Some(Language::Ttd),
            6601 => Some(Language::Tte),
            6602 => Some(Language::Ttf),
            6603 => Some(Language::Ttg),
            6604 => Some(Language::Tth),
            6605 => Some(Language::Tti),
            6606 => Some(Language::Ttj),
            6607 => Some(Language::Ttk),
            6608 => Some(Language::Ttl),
            6609 => Some(Language::Ttm),
            6610 => Some(Language::Ttn),
            6611 => Some(Language::Tto),
            6612 => Some(Language::Ttp),
            6613 => Some(Language::Ttq),
            6614 => Some(Language::Ttr),
            6615 => Some(Language::Tts),
            6616 => Some(Language::Ttt),
            6617 => Some(Language::Ttu),
            6618 => Some(Language::Ttv),
            6619 => Some(Language::Ttw),
            6620 => Some(Language::Tty),
            6621 => Some(Language::Ttz),
            6622 => Some(Language::Tua),
            6623 => Some(Language::Tub),
            6624 => Some(Language::Tuc),
            6625 => Some(Language::Tud),
            6626 => Some(Language::Tue),
            6627 => Some(Language::Tuf),
            6628 => Some(Language::Tug),
            6629 => Some(Language::Tuh),
            6630 => Some(Language::Tui),
            6631 => Some(Language::Tuj),
            6632 => Some(Language::Tuk),
            6633 => Some(Language::Tul),
            6634 => Some(Language::Tum),
            6635 => Some(Language::Tun),
            6636 => Some(Language::Tuo),
            6637 => Some(Language::Tuq),
            6638 => Some(Language::Tur),
            6639 => Some(Language::Tus),
            6640 => Some(Language::Tuu),
            6641 => Some(Language::Tuv),
            6642 => Some(Language::Tux),
            6643 => Some(Language::Tuy),
            6644 => Some(Language::Tuz),
            6645 => Some(Language::Tva),
            6646 => Some(Language::Tvd),
            6647 => Some(Language::Tve),
            6648 => Some(Language::Tvk),
            6649 => Some(Language::Tvl),
            6650 => Some(Language::Tvm),
            6651 => Some(Language::Tvn),
            6652 => Some(Language::Tvo),
            6653 => Some(Language::Tvs),
            6654 => Some(Language::Tvt),
            6655 => Some(Language::Tvu),
            6656 => Some(Language::Tvw),
            6657 => Some(Language::Tvx),
            6658 => Some(Language::Tvy),
            6659 => Some(Language::Twa),
            6660 => Some(Language::Twb),
            6661 => Some(Language::Twc),
            6662 => Some(Language::Twd),
            6663 => Some(Language::Twe),
            6664 => Some(Language::Twf),
            6665 => Some(Language::Twg),
            6666 => Some(Language::Twh),
            6667 => Some(Language::Twi),
            6668 => Some(Language::Twl),
            6669 => Some(Language::Twm),
            6670 => Some(Language::Twn),
            6671 => Some(Language::Two),
            6672 => Some(Language::Twp),
            6673 => Some(Language::Twq),
            6674 => Some(Language::Twr),
            6675 => Some(Language::Twt),
            6676 => Some(Language::Twu),
            6677 => Some(Language::Tww),
            6678 => Some(Language::Twx),
            6679 => Some(Language::Twy),
            6680 => Some(Language::Txa),
            6681 => Some(Language::Txb),
            6682 => Some(Language::Txc),
            6683 => Some(Language::Txe),
            6684 => Some(Language::Txg),
            6685 => Some(Language::Txh),
            6686 => Some(Language::Txi),
            6687 => Some(Language::Txj),
            6688 => Some(Language::Txm),
            6689 => Some(Language::Txn),
            6690 => Some(Language::Txo),
            6691 => Some(Language::Txq),
            6692 => Some(Language::Txr),
            6693 => Some(Language::Txs),
            6694 => Some(Language::Txt),
            6695 => Some(Language::Txu),
            6696 => Some(Language::Txx),
            6697 => Some(Language::Txy),
            6698 => Some(Language::Tya),
            6699 => Some(Language::Tye),
            6700 => Some(Language::Tyh),
            6701 => Some(Language::Tyi),
            6702 => Some(Language::Tyj),
            6703 => Some(Language::Tyl),
            6704 => Some(Language::Tyn),
            6705 => Some(Language::Typ),
            6706 => Some(Language::Tyr),
            6707 => Some(Language::Tys),
            6708 => Some(Language::Tyt),
            6709 => Some(Language::Tyu),
            6710 => Some(Language::Tyv),
            6711 => Some(Language::Tyx),
            6712 => Some(Language::Tyy),
            6713 => Some(Language::Tyz),
            6714 => Some(Language::Tza),
            6715 => Some(Language::Tzh),
            6716 => Some(Language::Tzj),
            6717 => Some(Language::Tzl),
            6718 => Some(Language::Tzm),
            6719 => Some(Language::Tzn),
            6720 => Some(Language::Tzo),
            6721 => Some(Language::Tzx),
            6722 => Some(Language::Uam),
            6723 => Some(Language::Uan),
            6724 => Some(Language::Uar),
            6725 => Some(Language::Uba),
            6726 => Some(Language::Ubi),
            6727 => Some(Language::Ubl),
            6728 => Some(Language::Ubr),
            6729 => Some(Language::Ubu),
            6730 => Some(Language::Uby),
            6731 => Some(Language::Uda),
            6732 => Some(Language::Ude),
            6733 => Some(Language::Udg),
            6734 => Some(Language::Udi),
            6735 => Some(Language::Udj),
            6736 => Some(Language::Udl),
            6737 => Some(Language::Udm),
            6738 => Some(Language::Udu),
            6739 => Some(Language::Ues),
            6740 => Some(Language::Ufi),
            6741 => Some(Language::Uga),
            6742 => Some(Language::Ugb),
            6743 => Some(Language::Uge),
            6744 => Some(Language::Ugh),
            6745 => Some(Language::Ugn),
            6746 => Some(Language::Ugo),
            6747 => Some(Language::Ugy),
            6748 => Some(Language::Uha),
            6749 => Some(Language::Uhn),
            6750 => Some(Language::Uig),
            6751 => Some(Language::Uis),
            6752 => Some(Language::Uiv),
            6753 => Some(Language::Uji),
            6754 => Some(Language::Uka),
            6755 => Some(Language::Ukg),
            6756 => Some(Language::Ukh),
            6757 => Some(Language::Uki),
            6758 => Some(Language::Ukk),
            6759 => Some(Language::Ukl),
            6760 => Some(Language::Ukp),
            6761 => Some(Language::Ukq),
            6762 => Some(Language::Ukr),
            6763 => Some(Language::Uks),
            6764 => Some(Language::Uku),
            6765 => Some(Language::Ukv),
            6766 => Some(Language::Ukw),
            6767 => Some(Language::Uky),
            6768 => Some(Language::Ula),
            6769 => Some(Language::Ulb),
            6770 => Some(Language::Ulc),
            6771 => Some(Language::Ule),
            6772 => Some(Language::Ulf),
            6773 => Some(Language::Uli),
            6774 => Some(Language::Ulk),
            6775 => Some(Language::Ull),
            6776 => Some(Language::Ulm),
            6777 => Some(Language::Uln),
            6778 => Some(Language::Ulu),
            6779 => Some(Language::Ulw),
            6780 => Some(Language::Uma),
            6781 => Some(Language::Umb),
            6782 => Some(Language::Umc),
            6783 => Some(Language::Umd),
            6784 => Some(Language::Umg),
            6785 => Some(Language::Umi),
            6786 => Some(Language::Umm),
            6787 => Some(Language::Umn),
            6788 => Some(Language::Umo),
            6789 => Some(Language::Ump),
            6790 => Some(Language::Umr),
            6791 => Some(Language::Ums),
            6792 => Some(Language::Umu),
            6793 => Some(Language::Una),
            6794 => Some(Language::Und),
            6795 => Some(Language::Une),
            6796 => Some(Language::Ung),
            6797 => Some(Language::Uni),
            6798 => Some(Language::Unk),
            6799 => Some(Language::Unm),
            6800 => Some(Language::Unn),
            6801 => Some(Language::Unr),
            6802 => Some(Language::Unu),
            6803 => Some(Language::Unx),
            6804 => Some(Language::Unz),
            6805 => Some(Language::Uon),
            6806 => Some(Language::Upi),
            6807 => Some(Language::Upv),
            6808 => Some(Language::Ura),
            6809 => Some(Language::Urb),
            6810 => Some(Language::Urc),
            6811 => Some(Language::Urd),
            6812 => Some(Language::Ure),
            6813 => Some(Language::Urf),
            6814 => Some(Language::Urg),
            6815 => Some(Language::Urh),
            6816 => Some(Language::Uri),
            6817 => Some(Language::Urk),
            6818 => Some(Language::Url),
            6819 => Some(Language::Urm),
            6820 => Some(Language::Urn),
            6821 => Some(Language::Uro),
            6822 => Some(Language::Urp),
            6823 => Some(Language::Urr),
            6824 => Some(Language::Urt),
            6825 => Some(Language::Uru),
            6826 => Some(Language::Urv),
            6827 => Some(Language::Urw),
            6828 => Some(Language::Urx),
            6829 => Some(Language::Ury),
            6830 => Some(Language::Urz),
            6831 => Some(Language::Usa),
            6832 => Some(Language::Ush),
            6833 => Some(Language::Usi),
            6834 => Some(Language::Usk),
            6835 => Some(Language::Usp),
            6836 => Some(Language::Uss),
            6837 => Some(Language::Usu),
            6838 => Some(Language::Uta),
            6839 => Some(Language::Ute),
            6840 => Some(Language::Uth),
            6841 => Some(Language::Utp),
            6842 => Some(Language::Utr),
            6843 => Some(Language::Utu),
            6844 => Some(Language::Uum),
            6845 => Some(Language::Uur),
            6846 => Some(Language::Uuu),
            6847 => Some(Language::Uve),
            6848 => Some(Language::Uvh),
            6849 => Some(Language::Uvl),
            6850 => Some(Language::Uwa),
            6851 => Some(Language::Uya),
            6852 => Some(Language::Uzb),
            6853 => Some(Language::Uzn),
            6854 => Some(Language::Uzs),
            6855 => Some(Language::Vaa),
            6856 => Some(Language::Vae),
            6857 => Some(Language::Vaf),
            6858 => Some(Language::Vag),
            6859 => Some(Language::Vah),
            6860 => Some(Language::Vai),
            6861 => Some(Language::Vaj),
            6862 => Some(Language::Val),
            6863 => Some(Language::Vam),
            6864 => Some(Language::Van),
            6865 => Some(Language::Vao),
            6866 => Some(Language::Vap),
            6867 => Some(Language::Var),
            6868 => Some(Language::Vas),
            6869 => Some(Language::Vau),
            6870 => Some(Language::Vav),
            6871 => Some(Language::Vay),
            6872 => Some(Language::Vbb),
            6873 => Some(Language::Vbk),
            6874 => Some(Language::Vec),
            6875 => Some(Language::Ved),
            6876 => Some(Language::Vel),
            6877 => Some(Language::Vem),
            6878 => Some(Language::Ven),
            6879 => Some(Language::Veo),
            6880 => Some(Language::Vep),
            6881 => Some(Language::Ver),
            6882 => Some(Language::Vgr),
            6883 => Some(Language::Vgt),
            6884 => Some(Language::Vic),
            6885 => Some(Language::Vid),
            6886 => Some(Language::Vie),
            6887 => Some(Language::Vif),
            6888 => Some(Language::Vig),
            6889 => Some(Language::Vil),
            6890 => Some(Language::Vin),
            6891 => Some(Language::Vis),
            6892 => Some(Language::Vit),
            6893 => Some(Language::Viv),
            6894 => Some(Language::Vka),
            6895 => Some(Language::Vkj),
            6896 => Some(Language::Vkk),
            6897 => Some(Language::Vkl),
            6898 => Some(Language::Vkm),
            6899 => Some(Language::Vkn),
            6900 => Some(Language::Vko),
            6901 => Some(Language::Vkp),
            6902 => Some(Language::Vkt),
            6903 => Some(Language::Vku),
            6904 => Some(Language::Vkz),
            6905 => Some(Language::Vlp),
            6906 => Some(Language::Vls),
            6907 => Some(Language::Vma),
            6908 => Some(Language::Vmb),
            6909 => Some(Language::Vmc),
            6910 => Some(Language::Vmd),
            6911 => Some(Language::Vme),
            6912 => Some(Language::Vmf),
            6913 => Some(Language::Vmg),
            6914 => Some(Language::Vmh),
            6915 => Some(Language::Vmi),
            6916 => Some(Language::Vmj),
            6917 => Some(Language::Vmk),
            6918 => Some(Language::Vml),
            6919 => Some(Language::Vmm),
            6920 => Some(Language::Vmp),
            6921 => Some(Language::Vmq),
            6922 => Some(Language::Vmr),
            6923 => Some(Language::Vms),
            6924 => Some(Language::Vmu),
            6925 => Some(Language::Vmv),
            6926 => Some(Language::Vmw),
            6927 => Some(Language::Vmx),
            6928 => Some(Language::Vmy),
            6929 => Some(Language::Vmz),
            6930 => Some(Language::Vnk),
            6931 => Some(Language::Vnm),
            6932 => Some(Language::Vnp),
            6933 => Some(Language::Vol),
            6934 => Some(Language::Vor),
            6935 => Some(Language::Vot),
            6936 => Some(Language::Vra),
            6937 => Some(Language::Vro),
            6938 => Some(Language::Vrs),
            6939 => Some(Language::Vrt),
            6940 => Some(Language::Vsi),
            6941 => Some(Language::Vsl),
            6942 => Some(Language::Vsv),
            6943 => Some(Language::Vto),
            6944 => Some(Language::Vum),
            6945 => Some(Language::Vun),
            6946 => Some(Language::Vut),
            6947 => Some(Language::Vwa),
            6948 => Some(Language::Waa),
            6949 => Some(Language::Wab),
            6950 => Some(Language::Wac),
            6951 => Some(Language::Wad),
            6952 => Some(Language::Wae),
            6953 => Some(Language::Waf),
            6954 => Some(Language::Wag),
            6955 => Some(Language::Wah),
            6956 => Some(Language::Wai),
            6957 => Some(Language::Waj),
            6958 => Some(Language::Wal),
            6959 => Some(Language::Wam),
            6960 => Some(Language::Wan),
            6961 => Some(Language::Wao),
            6962 => Some(Language::Wap),
            6963 => Some(Language::Waq),
            6964 => Some(Language::War),
            6965 => Some(Language::Was),
            6966 => Some(Language::Wat),
            6967 => Some(Language::Wau),
            6968 => Some(Language::Wav),
            6969 => Some(Language::Waw),
            6970 => Some(Language::Wax),
            6971 => Some(Language::Way),
            6972 => Some(Language::Waz),
            6973 => Some(Language::Wba),
            6974 => Some(Language::Wbb),
            6975 => Some(Language::Wbe),
            6976 => Some(Language::Wbf),
            6977 => Some(Language::Wbh),
            6978 => Some(Language::Wbi),
            6979 => Some(Language::Wbj),
            6980 => Some(Language::Wbk),
            6981 => Some(Language::Wbl),
            6982 => Some(Language::Wbm),
            6983 => Some(Language::Wbp),
            6984 => Some(Language::Wbq),
            6985 => Some(Language::Wbr),
            6986 => Some(Language::Wbs),
            6987 => Some(Language::Wbt),
            6988 => Some(Language::Wbv),
            6989 => Some(Language::Wbw),
            6990 => Some(Language::Wca),
            6991 => Some(Language::Wci),
            6992 => Some(Language::Wdd),
            6993 => Some(Language::Wdg),
            6994 => Some(Language::Wdj),
            6995 => Some(Language::Wdk),
            6996 => Some(Language::Wdt),
            6997 => Some(Language::Wdu),
            6998 => Some(Language::Wdy),
            6999 => Some(Language::Wea),
            7000 => Some(Language::Wec),
            7001 => Some(Language::Wed),
            7002 => Some(Language::Weg),
            7003 => Some(Language::Weh),
            7004 => Some(Language::Wei),
            7005 => Some(Language::Wem),
            7006 => Some(Language::Weo),
            7007 => Some(Language::Wep),
            7008 => Some(Language::Wer),
            7009 => Some(Language::Wes),
            7010 => Some(Language::Wet),
            7011 => Some(Language::Weu),
            7012 => Some(Language::Wew),
            7013 => Some(Language::Wfg),
            7014 => Some(Language::Wga),
            7015 => Some(Language::Wgb),
            7016 => Some(Language::Wgg),
            7017 => Some(Language::Wgi),
            7018 => Some(Language::Wgo),
            7019 => Some(Language::Wgu),
            7020 => Some(Language::Wgy),
            7021 => Some(Language::Wha),
            7022 => Some(Language::Whg),
            7023 => Some(Language::Whk),
            7024 => Some(Language::Whu),
            7025 => Some(Language::Wib),
            7026 => Some(Language::Wic),
            7027 => Some(Language::Wie),
            7028 => Some(Language::Wif),
            7029 => Some(Language::Wig),
            7030 => Some(Language::Wih),
            7031 => Some(Language::Wii),
            7032 => Some(Language::Wij),
            7033 => Some(Language::Wik),
            7034 => Some(Language::Wil),
            7035 => Some(Language::Wim),
            7036 => Some(Language::Win),
            7037 => Some(Language::Wir),
            7038 => Some(Language::Wiu),
            7039 => Some(Language::Wiv),
            7040 => Some(Language::Wiy),
            7041 => Some(Language::Wja),
            7042 => Some(Language::Wji),
            7043 => Some(Language::Wka),
            7044 => Some(Language::Wkb),
            7045 => Some(Language::Wkd),
            7046 => Some(Language::Wkl),
            7047 => Some(Language::Wkr),
            7048 => Some(Language::Wku),
            7049 => Some(Language::Wkw),
            7050 => Some(Language::Wky),
            7051 => Some(Language::Wla),
            7052 => Some(Language::Wlc),
            7053 => Some(Language::Wle),
            7054 => Some(Language::Wlg),
            7055 => Some(Language::Wlh),
            7056 => Some(Language::Wli),
            7057 => Some(Language::Wlk),
            7058 => Some(Language::Wll),
            7059 => Some(Language::Wlm),
            7060 => Some(Language::Wln),
            7061 => Some(Language::Wlo),
            7062 => Some(Language::Wlr),
            7063 => Some(Language::Wls),
            7064 => Some(Language::Wlu),
            7065 => Some(Language::Wlv),
            7066 => Some(Language::Wlw),
            7067 => Some(Language::Wlx),
            7068 => Some(Language::Wly),
            7069 => Some(Language::Wma),
            7070 => Some(Language::Wmb),
            7071 => Some(Language::Wmc),
            7072 => Some(Language::Wmd),
            7073 => Some(Language::Wme),
            7074 => Some(Language::Wmg),
            7075 => Some(Language::Wmh),
            7076 => Some(Language::Wmi),
            7077 => Some(Language::Wmm),
            7078 => Some(Language::Wmn),
            7079 => Some(Language::Wmo),
            7080 => Some(Language::Wms),
            7081 => Some(Language::Wmt),
            7082 => Some(Language::Wmw),
            7083 => Some(Language::Wmx),
            7084 => Some(Language::Wnb),
            7085 => Some(Language::Wnc),
            7086 => Some(Language::Wnd),
            7087 => Some(Language::Wne),
            7088 => Some(Language::Wng),
            7089 => Some(Language::Wni),
            7090 => Some(Language::Wnk),
            7091 => Some(Language::Wnm),
            7092 => Some(Language::Wnn),
            7093 => Some(Language::Wno),
            7094 => Some(Language::Wnp),
            7095 => Some(Language::Wnu),
            7096 => Some(Language::Wnw),
            7097 => Some(Language::Wny),
            7098 => Some(Language::Woa),
            7099 => Some(Language::Wob),
            7100 => Some(Language::Woc),
            7101 => Some(Language::Wod),
            7102 => Some(Language::Woe),
            7103 => Some(Language::Wof),
            7104 => Some(Language::Wog),
            7105 => Some(Language::Woi),
            7106 => Some(Language::Wok),
            7107 => Some(Language::Wol),
            7108 => Some(Language::Wom),
            7109 => Some(Language::Won),
            7110 => Some(Language::Woo),
            7111 => Some(Language::Wor),
            7112 => Some(Language::Wos),
            7113 => Some(Language::Wow),
            7114 => Some(Language::Woy),
            7115 => Some(Language::Wpc),
            7116 => Some(Language::Wrb),
            7117 => Some(Language::Wrg),
            7118 => Some(Language::Wrh),
            7119 => Some(Language::Wri),
            7120 => Some(Language::Wrk),
            7121 => Some(Language::Wrl),
            7122 => Some(Language::Wrm),
            7123 => Some(Language::Wrn),
            7124 => Some(Language::Wro),
            7125 => Some(Language::Wrp),
            7126 => Some(Language::Wrr),
            7127 => Some(Language::Wrs),
            7128 => Some(Language::Wru),
            7129 => Some(Language::Wrv),
            7130 => Some(Language::Wrw),
            7131 => Some(Language::Wrx),
            7132 => Some(Language::Wry),
            7133 => Some(Language::Wrz),
            7134 => Some(Language::Wsa),
            7135 => Some(Language::Wsg),
            7136 => Some(Language::Wsi),
            7137 => Some(Language::Wsk),
            7138 => Some(Language::Wsr),
            7139 => Some(Language::Wss),
            7140 => Some(Language::Wsu),
            7141 => Some(Language::Wsv),
            7142 => Some(Language::Wtf),
            7143 => Some(Language::Wth),
            7144 => Some(Language::Wti),
            7145 => Some(Language::Wtk),
            7146 => Some(Language::Wtm),
            7147 => Some(Language::Wtw),
            7148 => Some(Language::Wua),
            7149 => Some(Language::Wub),
            7150 => Some(Language::Wud),
            7151 => Some(Language::Wuh),
            7152 => Some(Language::Wul),
            7153 => Some(Language::Wum),
            7154 => Some(Language::Wun),
            7155 => Some(Language::Wur),
            7156 => Some(Language::Wut),
            7157 => Some(Language::Wuu),
            7158 => Some(Language::Wuv),
            7159 => Some(Language::Wux),
            7160 => Some(Language::Wuy),
            7161 => Some(Language::Wwa),
            7162 => Some(Language::Wwb),
            7163 => Some(Language::Wwo),
            7164 => Some(Language::Wwr),
            7165 => Some(Language::Www),
            7166 => Some(Language::Wxa),
            7167 => Some(Language::Wxw),
            7168 => Some(Language::Wyb),
            7169 => Some(Language::Wyi),
            7170 => Some(Language::Wym),
            7171 => Some(Language::Wyn),
            7172 => Some(Language::Wyr),
            7173 => Some(Language::Wyy),
            7174 => Some(Language::Xaa),
            7175 => Some(Language::Xab),
            7176 => Some(Language::Xac),
            7177 => Some(Language::Xad),
            7178 => Some(Language::Xae),
            7179 => Some(Language::Xag),
            7180 => Some(Language::Xai),
            7181 => Some(Language::Xaj),
            7182 => Some(Language::Xak),
            7183 => Some(Language::Xal),
            7184 => Some(Language::Xam),
            7185 => Some(Language::Xan),
            7186 => Some(Language::Xao),
            7187 => Some(Language::Xap),
            7188 => Some(Language::Xaq),
            7189 => Some(Language::Xar),
            7190 => Some(Language::Xas),
            7191 => Some(Language::Xat),
            7192 => Some(Language::Xau),
            7193 => Some(Language::Xav),
            7194 => Some(Language::Xaw),
            7195 => Some(Language::Xay),
            7196 => Some(Language::Xbb),
            7197 => Some(Language::Xbc),
            7198 => Some(Language::Xbd),
            7199 => Some(Language::Xbe),
            7200 => Some(Language::Xbg),
            7201 => Some(Language::Xbi),
            7202 => Some(Language::Xbj),
            7203 => Some(Language::Xbm),
            7204 => Some(Language::Xbn),
            7205 => Some(Language::Xbo),
            7206 => Some(Language::Xbp),
            7207 => Some(Language::Xbr),
            7208 => Some(Language::Xbw),
            7209 => Some(Language::Xby),
            7210 => Some(Language::Xcb),
            7211 => Some(Language::Xcc),
            7212 => Some(Language::Xce),
            7213 => Some(Language::Xcg),
            7214 => Some(Language::Xch),
            7215 => Some(Language::Xcl),
            7216 => Some(Language::Xcm),
            7217 => Some(Language::Xcn),
            7218 => Some(Language::Xco),
            7219 => Some(Language::Xcr),
            7220 => Some(Language::Xct),
            7221 => Some(Language::Xcu),
            7222 => Some(Language::Xcv),
            7223 => Some(Language::Xcw),
            7224 => Some(Language::Xcy),
            7225 => Some(Language::Xda),
            7226 => Some(Language::Xdc),
            7227 => Some(Language::Xdk),
            7228 => Some(Language::Xdm),
            7229 => Some(Language::Xdo),
            7230 => Some(Language::Xdq),
            7231 => Some(Language::Xdy),
            7232 => Some(Language::Xeb),
            7233 => Some(Language::Xed),
            7234 => Some(Language::Xeg),
            7235 => Some(Language::Xel),
            7236 => Some(Language::Xem),
            7237 => Some(Language::Xep),
            7238 => Some(Language::Xer),
            7239 => Some(Language::Xes),
            7240 => Some(Language::Xet),
            7241 => Some(Language::Xeu),
            7242 => Some(Language::Xfa),
            7243 => Some(Language::Xga),
            7244 => Some(Language::Xgb),
            7245 => Some(Language::Xgd),
            7246 => Some(Language::Xgf),
            7247 => Some(Language::Xgg),
            7248 => Some(Language::Xgi),
            7249 => Some(Language::Xgl),
            7250 => Some(Language::Xgm),
            7251 => Some(Language::Xgr),
            7252 => Some(Language::Xgu),
            7253 => Some(Language::Xgw),
            7254 => Some(Language::Xha),
            7255 => Some(Language::Xhc),
            7256 => Some(Language::Xhd),
            7257 => Some(Language::Xhe),
            7258 => Some(Language::Xhm),
            7259 => Some(Language::Xho),
            7260 => Some(Language::Xhr),
            7261 => Some(Language::Xht),
            7262 => Some(Language::Xhu),
            7263 => Some(Language::Xhv),
            7264 => Some(Language::Xib),
            7265 => Some(Language::Xii),
            7266 => Some(Language::Xil),
            7267 => Some(Language::Xin),
            7268 => Some(Language::Xir),
            7269 => Some(Language::Xis),
            7270 => Some(Language::Xiv),
            7271 => Some(Language::Xiy),
            7272 => Some(Language::Xjb),
            7273 => Some(Language::Xjt),
            7274 => Some(Language::Xka),
            7275 => Some(Language::Xkb),
            7276 => Some(Language::Xkc),
            7277 => Some(Language::Xkd),
            7278 => Some(Language::Xke),
            7279 => Some(Language::Xkf),
            7280 => Some(Language::Xkg),
            7281 => Some(Language::Xki),
            7282 => Some(Language::Xkj),
            7283 => Some(Language::Xkk),
            7284 => Some(Language::Xkl),
            7285 => Some(Language::Xkn),
            7286 => Some(Language::Xko),
            7287 => Some(Language::Xkp),
            7288 => Some(Language::Xkq),
            7289 => Some(Language::Xkr),
            7290 => Some(Language::Xks),
            7291 => Some(Language::Xkt),
            7292 => Some(Language::Xku),
            7293 => Some(Language::Xkv),
            7294 => Some(Language::Xkw),
            7295 => Some(Language::Xkx),
            7296 => Some(Language::Xky),
            7297 => Some(Language::Xkz),
            7298 => Some(Language::Xla),
            7299 => Some(Language::Xlb),
            7300 => Some(Language::Xlc),
            7301 => Some(Language::Xld),
            7302 => Some(Language::Xle),
            7303 => Some(Language::Xlg),
            7304 => Some(Language::Xli),
            7305 => Some(Language::Xln),
            7306 => Some(Language::Xlo),
            7307 => Some(Language::Xlp),
            7308 => Some(Language::Xls),
            7309 => Some(Language::Xlu),
            7310 => Some(Language::Xly),
            7311 => Some(Language::Xma),
            7312 => Some(Language::Xmb),
            7313 => Some(Language::Xmc),
            7314 => Some(Language::Xmd),
            7315 => Some(Language::Xme),
            7316 => Some(Language::Xmf),
            7317 => Some(Language::Xmg),
            7318 => Some(Language::Xmh),
            7319 => Some(Language::Xmj),
            7320 => Some(Language::Xmk),
            7321 => Some(Language::Xml),
            7322 => Some(Language::Xmm),
            7323 => Some(Language::Xmn),
            7324 => Some(Language::Xmo),
            7325 => Some(Language::Xmp),
            7326 => Some(Language::Xmq),
            7327 => Some(Language::Xmr),
            7328 => Some(Language::Xms),
            7329 => Some(Language::Xmt),
            7330 => Some(Language::Xmu),
            7331 => Some(Language::Xmv),
            7332 => Some(Language::Xmw),
            7333 => Some(Language::Xmx),
            7334 => Some(Language::Xmy),
            7335 => Some(Language::Xmz),
            7336 => Some(Language::Xna),
            7337 => Some(Language::Xnb),
            7338 => Some(Language::Xng),
            7339 => Some(Language::Xnh),
            7340 => Some(Language::Xni),
            7341 => Some(Language::Xnj),
            7342 => Some(Language::Xnk),
            7343 => Some(Language::Xnm),
            7344 => Some(Language::Xnn),
            7345 => Some(Language::Xno),
            7346 => Some(Language::Xnq),
            7347 => Some(Language::Xnr),
            7348 => Some(Language::Xns),
            7349 => Some(Language::Xnt),
            7350 => Some(Language::Xnu),
            7351 => Some(Language::Xny),
            7352 => Some(Language::Xnz),
            7353 => Some(Language::Xoc),
            7354 => Some(Language::Xod),
            7355 => Some(Language::Xog),
            7356 => Some(Language::Xoi),
            7357 => Some(Language::Xok),
            7358 => Some(Language::Xom),
            7359 => Some(Language::Xon),
            7360 => Some(Language::Xoo),
            7361 => Some(Language::Xop),
            7362 => Some(Language::Xor),
            7363 => Some(Language::Xow),
            7364 => Some(Language::Xpa),
            7365 => Some(Language::Xpb),
            7366 => Some(Language::Xpc),
            7367 => Some(Language::Xpd),
            7368 => Some(Language::Xpe),
            7369 => Some(Language::Xpf),
            7370 => Some(Language::Xpg),
            7371 => Some(Language::Xph),
            7372 => Some(Language::Xpi),
            7373 => Some(Language::Xpj),
            7374 => Some(Language::Xpk),
            7375 => Some(Language::Xpl),
            7376 => Some(Language::Xpm),
            7377 => Some(Language::Xpn),
            7378 => Some(Language::Xpo),
            7379 => Some(Language::Xpp),
            7380 => Some(Language::Xpq),
            7381 => Some(Language::Xpr),
            7382 => Some(Language::Xps),
            7383 => Some(Language::Xpt),
            7384 => Some(Language::Xpu),
            7385 => Some(Language::Xpv),
            7386 => Some(Language::Xpw),
            7387 => Some(Language::Xpx),
            7388 => Some(Language::Xpy),
            7389 => Some(Language::Xpz),
            7390 => Some(Language::Xqa),
            7391 => Some(Language::Xqt),
            7392 => Some(Language::Xra),
            7393 => Some(Language::Xrb),
            7394 => Some(Language::Xrd),
            7395 => Some(Language::Xre),
            7396 => Some(Language::Xrg),
            7397 => Some(Language::Xri),
            7398 => Some(Language::Xrm),
            7399 => Some(Language::Xrn),
            7400 => Some(Language::Xrr),
            7401 => Some(Language::Xrt),
            7402 => Some(Language::Xru),
            7403 => Some(Language::Xrw),
            7404 => Some(Language::Xsa),
            7405 => Some(Language::Xsb),
            7406 => Some(Language::Xsc),
            7407 => Some(Language::Xsd),
            7408 => Some(Language::Xse),
            7409 => Some(Language::Xsh),
            7410 => Some(Language::Xsi),
            7411 => Some(Language::Xsj),
            7412 => Some(Language::Xsl),
            7413 => Some(Language::Xsm),
            7414 => Some(Language::Xsn),
            7415 => Some(Language::Xso),
            7416 => Some(Language::Xsp),
            7417 => Some(Language::Xsq),
            7418 => Some(Language::Xsr),
            7419 => Some(Language::Xss),
            7420 => Some(Language::Xsu),
            7421 => Some(Language::Xsv),
            7422 => Some(Language::Xsy),
            7423 => Some(Language::Xta),
            7424 => Some(Language::Xtb),
            7425 => Some(Language::Xtc),
            7426 => Some(Language::Xtd),
            7427 => Some(Language::Xte),
            7428 => Some(Language::Xtg),
            7429 => Some(Language::Xth),
            7430 => Some(Language::Xti),
            7431 => Some(Language::Xtj),
            7432 => Some(Language::Xtl),
            7433 => Some(Language::Xtm),
            7434 => Some(Language::Xtn),
            7435 => Some(Language::Xto),
            7436 => Some(Language::Xtp),
            7437 => Some(Language::Xtq),
            7438 => Some(Language::Xtr),
            7439 => Some(Language::Xts),
            7440 => Some(Language::Xtt),
            7441 => Some(Language::Xtu),
            7442 => Some(Language::Xtv),
            7443 => Some(Language::Xtw),
            7444 => Some(Language::Xty),
            7445 => Some(Language::Xua),
            7446 => Some(Language::Xub),
            7447 => Some(Language::Xud),
            7448 => Some(Language::Xug),
            7449 => Some(Language::Xuj),
            7450 => Some(Language::Xul),
            7451 => Some(Language::Xum),
            7452 => Some(Language::Xun),
            7453 => Some(Language::Xuo),
            7454 => Some(Language::Xup),
            7455 => Some(Language::Xur),
            7456 => Some(Language::Xut),
            7457 => Some(Language::Xuu),
            7458 => Some(Language::Xve),
            7459 => Some(Language::Xvi),
            7460 => Some(Language::Xvn),
            7461 => Some(Language::Xvo),
            7462 => Some(Language::Xvs),
            7463 => Some(Language::Xwa),
            7464 => Some(Language::Xwc),
            7465 => Some(Language::Xwd),
            7466 => Some(Language::Xwe),
            7467 => Some(Language::Xwg),
            7468 => Some(Language::Xwj),
            7469 => Some(Language::Xwk),
            7470 => Some(Language::Xwl),
            7471 => Some(Language::Xwo),
            7472 => Some(Language::Xwr),
            7473 => Some(Language::Xwt),
            7474 => Some(Language::Xww),
            7475 => Some(Language::Xxb),
            7476 => Some(Language::Xxk),
            7477 => Some(Language::Xxm),
            7478 => Some(Language::Xxr),
            7479 => Some(Language::Xxt),
            7480 => Some(Language::Xya),
            7481 => Some(Language::Xyb),
            7482 => Some(Language::Xyj),
            7483 => Some(Language::Xyk),
            7484 => Some(Language::Xyl),
            7485 => Some(Language::Xyt),
            7486 => Some(Language::Xyy),
            7487 => Some(Language::Xzh),
            7488 => Some(Language::Xzm),
            7489 => Some(Language::Xzp),
            7490 => Some(Language::Yaa),
            7491 => Some(Language::Yab),
            7492 => Some(Language::Yac),
            7493 => Some(Language::Yad),
            7494 => Some(Language::Yae),
            7495 => Some(Language::Yaf),
            7496 => Some(Language::Yag),
            7497 => Some(Language::Yah),
            7498 => Some(Language::Yai),
            7499 => Some(Language::Yaj),
            7500 => Some(Language::Yak),
            7501 => Some(Language::Yal),
            7502 => Some(Language::Yam),
            7503 => Some(Language::Yan),
            7504 => Some(Language::Yao),
            7505 => Some(Language::Yap),
            7506 => Some(Language::Yaq),
            7507 => Some(Language::Yar),
            7508 => Some(Language::Yas),
            7509 => Some(Language::Yat),
            7510 => Some(Language::Yau),
            7511 => Some(Language::Yav),
            7512 => Some(Language::Yaw),
            7513 => Some(Language::Yax),
            7514 => Some(Language::Yay),
            7515 => Some(Language::Yaz),
            7516 => Some(Language::Yba),
            7517 => Some(Language::Ybb),
            7518 => Some(Language::Ybe),
            7519 => Some(Language::Ybh),
            7520 => Some(Language::Ybi),
            7521 => Some(Language::Ybj),
            7522 => Some(Language::Ybk),
            7523 => Some(Language::Ybl),
            7524 => Some(Language::Ybm),
            7525 => Some(Language::Ybn),
            7526 => Some(Language::Ybo),
            7527 => Some(Language::Ybx),
            7528 => Some(Language::Yby),
            7529 => Some(Language::Ych),
            7530 => Some(Language::Ycl),
            7531 => Some(Language::Ycn),
            7532 => Some(Language::Ycp),
            7533 => Some(Language::Yda),
            7534 => Some(Language::Ydd),
            7535 => Some(Language::Yde),
            7536 => Some(Language::Ydg),
            7537 => Some(Language::Ydk),
            7538 => Some(Language::Yea),
            7539 => Some(Language::Yec),
            7540 => Some(Language::Yee),
            7541 => Some(Language::Yei),
            7542 => Some(Language::Yej),
            7543 => Some(Language::Yel),
            7544 => Some(Language::Yer),
            7545 => Some(Language::Yes),
            7546 => Some(Language::Yet),
            7547 => Some(Language::Yeu),
            7548 => Some(Language::Yev),
            7549 => Some(Language::Yey),
            7550 => Some(Language::Yga),
            7551 => Some(Language::Ygi),
            7552 => Some(Language::Ygl),
            7553 => Some(Language::Ygm),
            7554 => Some(Language::Ygp),
            7555 => Some(Language::Ygr),
            7556 => Some(Language::Ygs),
            7557 => Some(Language::Ygu),
            7558 => Some(Language::Ygw),
            7559 => Some(Language::Yha),
            7560 => Some(Language::Yhd),
            7561 => Some(Language::Yhl),
            7562 => Some(Language::Yhs),
            7563 => Some(Language::Yia),
            7564 => Some(Language::Yid),
            7565 => Some(Language::Yif),
            7566 => Some(Language::Yig),
            7567 => Some(Language::Yih),
            7568 => Some(Language::Yii),
            7569 => Some(Language::Yij),
            7570 => Some(Language::Yik),
            7571 => Some(Language::Yil),
            7572 => Some(Language::Yim),
            7573 => Some(Language::Yin),
            7574 => Some(Language::Yip),
            7575 => Some(Language::Yiq),
            7576 => Some(Language::Yir),
            7577 => Some(Language::Yis),
            7578 => Some(Language::Yit),
            7579 => Some(Language::Yiu),
            7580 => Some(Language::Yiv),
            7581 => Some(Language::Yix),
            7582 => Some(Language::Yiz),
            7583 => Some(Language::Yka),
            7584 => Some(Language::Ykg),
            7585 => Some(Language::Yki),
            7586 => Some(Language::Ykk),
            7587 => Some(Language::Ykl),
            7588 => Some(Language::Ykm),
            7589 => Some(Language::Ykn),
            7590 => Some(Language::Yko),
            7591 => Some(Language::Ykr),
            7592 => Some(Language::Ykt),
            7593 => Some(Language::Yku),
            7594 => Some(Language::Yky),
            7595 => Some(Language::Yla),
            7596 => Some(Language::Ylb),
            7597 => Some(Language::Yle),
            7598 => Some(Language::Ylg),
            7599 => Some(Language::Yli),
            7600 => Some(Language::Yll),
            7601 => Some(Language::Ylm),
            7602 => Some(Language::Yln),
            7603 => Some(Language::Ylo),
            7604 => Some(Language::Ylr),
            7605 => Some(Language::Ylu),
            7606 => Some(Language::Yly),
            7607 => Some(Language::Ymb),
            7608 => Some(Language::Ymc),
            7609 => Some(Language::Ymd),
            7610 => Some(Language::Yme),
            7611 => Some(Language::Ymg),
            7612 => Some(Language::Ymh),
            7613 => Some(Language::Ymi),
            7614 => Some(Language::Ymk),
            7615 => Some(Language::Yml),
            7616 => Some(Language::Ymm),
            7617 => Some(Language::Ymn),
            7618 => Some(Language::Ymo),
            7619 => Some(Language::Ymp),
            7620 => Some(Language::Ymq),
            7621 => Some(Language::Ymr),
            7622 => Some(Language::Yms),
            7623 => Some(Language::Ymx),
            7624 => Some(Language::Ymz),
            7625 => Some(Language::Yna),
            7626 => Some(Language::Ynd),
            7627 => Some(Language::Yne),
            7628 => Some(Language::Yng),
            7629 => Some(Language::Ynk),
            7630 => Some(Language::Ynl),
            7631 => Some(Language::Ynn),
            7632 => Some(Language::Yno),
            7633 => Some(Language::Ynq),
            7634 => Some(Language::Yns),
            7635 => Some(Language::Ynu),
            7636 => Some(Language::Yob),
            7637 => Some(Language::Yog),
            7638 => Some(Language::Yoi),
            7639 => Some(Language::Yok),
            7640 => Some(Language::Yol),
            7641 => Some(Language::Yom),
            7642 => Some(Language::Yon),
            7643 => Some(Language::Yor),
            7644 => Some(Language::Yot),
            7645 => Some(Language::Yox),
            7646 => Some(Language::Yoy),
            7647 => Some(Language::Ypa),
            7648 => Some(Language::Ypb),
            7649 => Some(Language::Ypg),
            7650 => Some(Language::Yph),
            7651 => Some(Language::Ypm),
            7652 => Some(Language::Ypn),
            7653 => Some(Language::Ypo),
            7654 => Some(Language::Ypp),
            7655 => Some(Language::Ypz),
            7656 => Some(Language::Yra),
            7657 => Some(Language::Yrb),
            7658 => Some(Language::Yre),
            7659 => Some(Language::Yrk),
            7660 => Some(Language::Yrl),
            7661 => Some(Language::Yrm),
            7662 => Some(Language::Yrn),
            7663 => Some(Language::Yro),
            7664 => Some(Language::Yrs),
            7665 => Some(Language::Yrw),
            7666 => Some(Language::Yry),
            7667 => Some(Language::Ysc),
            7668 => Some(Language::Ysd),
            7669 => Some(Language::Ysg),
            7670 => Some(Language::Ysl),
            7671 => Some(Language::Ysm),
            7672 => Some(Language::Ysn),
            7673 => Some(Language::Yso),
            7674 => Some(Language::Ysp),
            7675 => Some(Language::Ysr),
            7676 => Some(Language::Yss),
            7677 => Some(Language::Ysy),
            7678 => Some(Language::Yta),
            7679 => Some(Language::Ytl),
            7680 => Some(Language::Ytp),
            7681 => Some(Language::Ytw),
            7682 => Some(Language::Yty),
            7683 => Some(Language::Yua),
            7684 => Some(Language::Yub),
            7685 => Some(Language::Yuc),
            7686 => Some(Language::Yud),
            7687 => Some(Language::Yue),
            7688 => Some(Language::Yuf),
            7689 => Some(Language::Yug),
            7690 => Some(Language::Yui),
            7691 => Some(Language::Yuj),
            7692 => Some(Language::Yuk),
            7693 => Some(Language::Yul),
            7694 => Some(Language::Yum),
            7695 => Some(Language::Yun),
            7696 => Some(Language::Yup),
            7697 => Some(Language::Yuq),
            7698 => Some(Language::Yur),
            7699 => Some(Language::Yut),
            7700 => Some(Language::Yuw),
            7701 => Some(Language::Yux),
            7702 => Some(Language::Yuy),
            7703 => Some(Language::Yuz),
            7704 => Some(Language::Yva),
            7705 => Some(Language::Yvt),
            7706 => Some(Language::Ywa),
            7707 => Some(Language::Ywg),
            7708 => Some(Language::Ywl),
            7709 => Some(Language::Ywn),
            7710 => Some(Language::Ywq),
            7711 => Some(Language::Ywr),
            7712 => Some(Language::Ywt),
            7713 => Some(Language::Ywu),
            7714 => Some(Language::Yww),
            7715 => Some(Language::Yxa),
            7716 => Some(Language::Yxg),
            7717 => Some(Language::Yxl),
            7718 => Some(Language::Yxm),
            7719 => Some(Language::Yxu),
            7720 => Some(Language::Yxy),
            7721 => Some(Language::Yyr),
            7722 => Some(Language::Yyu),
            7723 => Some(Language::Yyz),
            7724 => Some(Language::Yzg),
            7725 => Some(Language::Yzk),
            7726 => Some(Language::Zaa),
            7727 => Some(Language::Zab),
            7728 => Some(Language::Zac),
            7729 => Some(Language::Zad),
            7730 => Some(Language::Zae),
            7731 => Some(Language::Zaf),
            7732 => Some(Language::Zag),
            7733 => Some(Language::Zah),
            7734 => Some(Language::Zai),
            7735 => Some(Language::Zaj),
            7736 => Some(Language::Zak),
            7737 => Some(Language::Zal),
            7738 => Some(Language::Zam),
            7739 => Some(Language::Zao),
            7740 => Some(Language::Zap),
            7741 => Some(Language::Zaq),
            7742 => Some(Language::Zar),
            7743 => Some(Language::Zas),
            7744 => Some(Language::Zat),
            7745 => Some(Language::Zau),
            7746 => Some(Language::Zav),
            7747 => Some(Language::Zaw),
            7748 => Some(Language::Zax),
            7749 => Some(Language::Zay),
            7750 => Some(Language::Zaz),
            7751 => Some(Language::Zba),
            7752 => Some(Language::Zbc),
            7753 => Some(Language::Zbe),
            7754 => Some(Language::Zbl),
            7755 => Some(Language::Zbt),
            7756 => Some(Language::Zbu),
            7757 => Some(Language::Zbw),
            7758 => Some(Language::Zca),
            7759 => Some(Language::Zcd),
            7760 => Some(Language::Zch),
            7761 => Some(Language::Zdj),
            7762 => Some(Language::Zea),
            7763 => Some(Language::Zeg),
            7764 => Some(Language::Zeh),
            7765 => Some(Language::Zen),
            7766 => Some(Language::Zga),
            7767 => Some(Language::Zgb),
            7768 => Some(Language::Zgh),
            7769 => Some(Language::Zgm),
            7770 => Some(Language::Zgn),
            7771 => Some(Language::Zgr),
            7772 => Some(Language::Zha),
            7773 => Some(Language::Zhb),
            7774 => Some(Language::Zhd),
            7775 => Some(Language::Zhi),
            7776 => Some(Language::Zhn),
            7777 => Some(Language::Zho),
            7778 => Some(Language::Zhw),
            7779 => Some(Language::Zia),
            7780 => Some(Language::Zib),
            7781 => Some(Language::Zik),
            7782 => Some(Language::Zil),
            7783 => Some(Language::Zim),
            7784 => Some(Language::Zin),
            7785 => Some(Language::Ziw),
            7786 => Some(Language::Ziz),
            7787 => Some(Language::Zka),
            7788 => Some(Language::Zkb),
            7789 => Some(Language::Zkd),
            7790 => Some(Language::Zkg),
            7791 => Some(Language::Zkh),
            7792 => Some(Language::Zkk),
            7793 => Some(Language::Zkn),
            7794 => Some(Language::Zko),
            7795 => Some(Language::Zkp),
            7796 => Some(Language::Zkr),
            7797 => Some(Language::Zkt),
            7798 => Some(Language::Zku),
            7799 => Some(Language::Zkv),
            7800 => Some(Language::Zkz),
            7801 => Some(Language::Zla),
            7802 => Some(Language::Zlj),
            7803 => Some(Language::Zlm),
            7804 => Some(Language::Zln),
            7805 => Some(Language::Zlq),
            7806 => Some(Language::Zma),
            7807 => Some(Language::Zmb),
            7808 => Some(Language::Zmc),
            7809 => Some(Language::Zmd),
            7810 => Some(Language::Zme),
            7811 => Some(Language::Zmf),
            7812 => Some(Language::Zmg),
            7813 => Some(Language::Zmh),
            7814 => Some(Language::Zmi),
            7815 => Some(Language::Zmj),
            7816 => Some(Language::Zmk),
            7817 => Some(Language::Zml),
            7818 => Some(Language::Zmm),
            7819 => Some(Language::Zmn),
            7820 => Some(Language::Zmo),
            7821 => Some(Language::Zmp),
            7822 => Some(Language::Zmq),
            7823 => Some(Language::Zmr),
            7824 => Some(Language::Zms),
            7825 => Some(Language::Zmt),
            7826 => Some(Language::Zmu),
            7827 => Some(Language::Zmv),
            7828 => Some(Language::Zmw),
            7829 => Some(Language::Zmx),
            7830 => Some(Language::Zmy),
            7831 => Some(Language::Zmz),
            7832 => Some(Language::Zna),
            7833 => Some(Language::Zne),
            7834 => Some(Language::Zng),
            7835 => Some(Language::Znk),
            7836 => Some(Language::Zns),
            7837 => Some(Language::Zoc),
            7838 => Some(Language::Zoh),
            7839 => Some(Language::Zom),
            7840 => Some(Language::Zoo),
            7841 => Some(Language::Zoq),
            7842 => Some(Language::Zor),
            7843 => Some(Language::Zos),
            7844 => Some(Language::Zpa),
            7845 => Some(Language::Zpb),
            7846 => Some(Language::Zpc),
            7847 => Some(Language::Zpd),
            7848 => Some(Language::Zpe),
            7849 => Some(Language::Zpf),
            7850 => Some(Language::Zpg),
            7851 => Some(Language::Zph),
            7852 => Some(Language::Zpi),
            7853 => Some(Language::Zpj),
            7854 => Some(Language::Zpk),
            7855 => Some(Language::Zpl),
            7856 => Some(Language::Zpm),
            7857 => Some(Language::Zpn),
            7858 => Some(Language::Zpo),
            7859 => Some(Language::Zpp),
            7860 => Some(Language::Zpq),
            7861 => Some(Language::Zpr),
            7862 => Some(Language::Zps),
            7863 => Some(Language::Zpt),
            7864 => Some(Language::Zpu),
            7865 => Some(Language::Zpv),
            7866 => Some(Language::Zpw),
            7867 => Some(Language::Zpx),
            7868 => Some(Language::Zpy),
            7869 => Some(Language::Zpz),
            7870 => Some(Language::Zqe),
            7871 => Some(Language::Zra),
            7872 => Some(Language::Zrg),
            7873 => Some(Language::Zrn),
            7874 => Some(Language::Zro),
            7875 => Some(Language::Zrp),
            7876 => Some(Language::Zrs),
            7877 => Some(Language::Zsa),
            7878 => Some(Language::Zsk),
            7879 => Some(Language::Zsl),
            7880 => Some(Language::Zsm),
            7881 => Some(Language::Zsr),
            7882 => Some(Language::Zsu),
            7883 => Some(Language::Zte),
            7884 => Some(Language::Ztg),
            7885 => Some(Language::Ztl),
            7886 => Some(Language::Ztm),
            7887 => Some(Language::Ztn),
            7888 => Some(Language::Ztp),
            7889 => Some(Language::Ztq),
            7890 => Some(Language::Zts),
            7891 => Some(Language::Ztt),
            7892 => Some(Language::Ztu),
            7893 => Some(Language::Ztx),
            7894 => Some(Language::Zty),
            7895 => Some(Language::Zua),
            7896 => Some(Language::Zuh),
            7897 => Some(Language::Zul),
            7898 => Some(Language::Zum),
            7899 => Some(Language::Zun),
            7900 => Some(Language::Zuy),
            7901 => Some(Language::Zwa),
            7902 => Some(Language::Zxx),
            7903 => Some(Language::Zyb),
            7904 => Some(Language::Zyg),
            7905 => Some(Language::Zyj),
            7906 => Some(Language::Zyn),
            7907 => Some(Language::Zyp),
            7908 => Some(Language::Zza),
            7909 => Some(Language::Zzj),
            _ => None,
        }
    }
}

pub(crate) const TWO_TO_THREE: phf::Map<&str, Language> = ::phf::Map {
    key: 12913932095322966823,
    disps: &[
        (1, 92),
        (0, 0),
        (0, 0),
        (1, 52),
        (0, 110),
        (1, 6),
        (0, 1),
        (0, 12),
        (3, 173),
        (1, 18),
        (0, 0),
        (0, 28),
        (1, 4),
        (0, 20),
        (0, 38),
        (0, 43),
        (3, 80),
        (0, 0),
        (1, 69),
        (0, 147),
        (3, 7),
        (0, 92),
        (4, 73),
        (19, 137),
        (1, 113),
        (0, 0),
        (0, 78),
        (0, 25),
        (5, 149),
        (0, 128),
        (0, 0),
        (6, 149),
        (2, 32),
        (2, 128),
        (6, 147),
        (3, 150),
        (0, 3),
    ],
    entries: &[
        ("br", Language::Bre),
        ("hu", Language::Hun),
        ("tr", Language::Tur),
        ("no", Language::Nor),
        ("kk", Language::Kaz),
        ("fo", Language::Fao),
        ("rm", Language::Roh),
        ("vo", Language::Vol),
        ("km", Language::Khm),
        ("ve", Language::Ven),
        ("si", Language::Sin),
        ("kw", Language::Cor),
        ("mr", Language::Mar),
        ("vi", Language::Vie),
        ("zu", Language::Zul),
        ("sr", Language::Srp),
        ("he", Language::Heb),
        ("ts", Language::Tso),
        ("ml", Language::Mal),
        ("is", Language::Isl),
        ("cy", Language::Cym),
        ("ln", Language::Lin),
        ("bo", Language::Bod),
        ("rw", Language::Kin),
        ("as", Language::Asm),
        ("uz", Language::Uzb),
        ("kl", Language::Kal),
        ("ht", Language::Hat),
        ("ka", Language::Kat),
        ("mi", Language::Mri),
        ("el", Language::Ell),
        ("pt", Language::Por),
        ("ar", Language::Ara),
        ("lu", Language::Lub),
        ("ab", Language::Abk),
        ("lg", Language::Lug),
        ("nv", Language::Nav),
        ("ss", Language::Ssw),
        ("az", Language::Aze),
        ("av", Language::Ava),
        ("fy", Language::Fry),
        ("an", Language::Arg),
        ("sw", Language::Swa),
        ("ik", Language::Ipk),
        ("rn", Language::Run),
        ("ja", Language::Jpn),
        ("zh", Language::Zho),
        ("fi", Language::Fin),
        ("gl", Language::Glg),
        ("ru", Language::Rus),
        ("ms", Language::Msa),
        ("ku", Language::Kur),
        ("kr", Language::Kau),
        ("sv", Language::Swe),
        ("nr", Language::Nbl),
        ("ng", Language::Ndo),
        ("hr", Language::Hrv),
        ("mk", Language::Mkd),
        ("nn", Language::Nno),
        ("eu", Language::Eus),
        ("ch", Language::Cha),
        ("ce", Language::Che),
        ("sl", Language::Slv),
        ("ne", Language::Nep),
        ("mg", Language::Mlg),
        ("tt", Language::Tat),
        ("sc", Language::Srd),
        ("kg", Language::Kon),
        ("sh", Language::Hbs),
        ("wa", Language::Wln),
        ("oj", Language::Oji),
        ("st", Language::Sot),
        ("li", Language::Lim),
        ("ha", Language::Hau),
        ("co", Language::Cos),
        ("my", Language::Mya),
        ("gn", Language::Grn),
        ("da", Language::Dan),
        ("bs", Language::Bos),
        ("ae", Language::Ave),
        ("oc", Language::Oci),
        ("ie", Language::Ile),
        ("hi", Language::Hin),
        ("lt", Language::Lit),
        ("sq", Language::Sqi),
        ("id", Language::Ind),
        ("te", Language::Tel),
        ("ps", Language::Pus),
        ("bg", Language::Bul),
        ("it", Language::Ita),
        ("tl", Language::Tgl),
        ("ta", Language::Tam),
        ("pi", Language::Pli),
        ("yo", Language::Yor),
        ("ho", Language::Hmo),
        ("to", Language::Ton),
        ("cv", Language::Chv),
        ("pl", Language::Pol),
        ("ba", Language::Bak),
        ("gv", Language::Glv),
        ("eo", Language::Epo),
        ("tw", Language::Twi),
        ("sn", Language::Sna),
        ("se", Language::Sme),
        ("fj", Language::Fij),
        ("cs", Language::Ces),
        ("hy", Language::Hye),
        ("mh", Language::Mah),
        ("ig", Language::Ibo),
        ("pa", Language::Pan),
        ("na", Language::Nau),
        ("nl", Language::Nld),
        ("es", Language::Spa),
        ("dv", Language::Div),
        ("om", Language::Orm),
        ("gu", Language::Guj),
        ("cu", Language::Chu),
        ("ca", Language::Cat),
        ("wo", Language::Wol),
        ("kv", Language::Kom),
        ("ay", Language::Aym),
        ("su", Language::Sun),
        ("sk", Language::Slk),
        ("lv", Language::Lav),
        ("de", Language::Deu),
        ("iu", Language::Iku),
        ("nd", Language::Nde),
        ("ky", Language::Kir),
        ("ii", Language::Iii),
        ("sd", Language::Snd),
        ("sm", Language::Smo),
        ("ty", Language::Tah),
        ("aa", Language::Aar),
        ("qu", Language::Que),
        ("lo", Language::Lao),
        ("fr", Language::Fra),
        ("sg", Language::Sag),
        ("af", Language::Afr),
        ("lb", Language::Ltz),
        ("so", Language::Som),
        ("am", Language::Amh),
        ("nb", Language::Nob),
        ("or", Language::Ori),
        ("ti", Language::Tir),
        ("ee", Language::Ewe),
        ("tg", Language::Tgk),
        ("yi", Language::Yid),
        ("xh", Language::Xho),
        ("fa", Language::Fas),
        ("et", Language::Est),
        ("tk", Language::Tuk),
        ("tn", Language::Tsn),
        ("sa", Language::San),
        ("be", Language::Bel),
        ("mn", Language::Mon),
        ("dz", Language::Dzo),
        ("th", Language::Tha),
        ("cr", Language::Cre),
        ("mt", Language::Mlt),
        ("ia", Language::Ina),
        ("bi", Language::Bis),
        ("ro", Language::Ron),
        ("ff", Language::Ful),
        ("uk", Language::Ukr),
        ("za", Language::Zha),
        ("ug", Language::Uig),
        ("ny", Language::Nya),
        ("ga", Language::Gle),
        ("ur", Language::Urd),
        ("io", Language::Ido),
        ("ko", Language::Kor),
        ("bn", Language::Ben),
        ("os", Language::Oss),
        ("la", Language::Lat),
        ("ks", Language::Kas),
        ("hz", Language::Her),
        ("kn", Language::Kan),
        ("ak", Language::Aka),
        ("kj", Language::Kua),
        ("ki", Language::Kik),
        ("jv", Language::Jav),
        ("bm", Language::Bam),
        ("gd", Language::Gla),
        ("en", Language::Eng),
    ],
};

pub(crate) const THREE_TO_THREE: phf::Map<&str, Language> = ::phf::Map {
    key: 12913932095322966823,
    disps: &[
        (0, 230),
        (0, 224),
        (0, 146),
        (0, 12),
        (0, 46),
        (0, 92),
        (0, 46),
        (0, 0),
        (0, 2),
        (0, 63),
        (0, 3),
        (0, 62),
        (0, 78),
        (0, 96),
        (0, 6),
        (0, 5),
        (0, 458),
        (0, 0),
        (0, 635),
        (0, 1205),
        (0, 410),
        (0, 967),
        (0, 748),
        (0, 110),
        (0, 3),
        (0, 857),
        (0, 479),
        (0, 1),
        (0, 25),
        (0, 29),
        (0, 89),
        (0, 3),
        (0, 203),
        (0, 4),
        (0, 4),
        (0, 96),
        (0, 553),
        (0, 0),
        (0, 56),
        (0, 84),
        (0, 48),
        (0, 2),
        (0, 2631),
        (0, 189),
        (0, 133),
        (0, 73),
        (0, 462),
        (0, 26),
        (0, 211),
        (0, 171),
        (0, 480),
        (0, 12),
        (0, 252),
        (0, 341),
        (0, 263),
        (0, 80),
        (0, 23),
        (0, 83),
        (0, 2),
        (0, 3),
        (0, 29),
        (0, 2),
        (0, 126),
        (0, 47),
        (0, 538),
        (0, 38),
        (0, 168),
        (0, 443),
        (0, 13),
        (0, 406),
        (0, 0),
        (0, 856),
        (0, 53),
        (0, 537),
        (0, 4),
        (0, 439),
        (0, 4),
        (0, 23),
        (0, 1428),
        (0, 0),
        (0, 7),
        (0, 7),
        (0, 5),
        (0, 26),
        (0, 36),
        (0, 1),
        (0, 48),
        (0, 229),
        (0, 725),
        (0, 34),
        (0, 70),
        (0, 590),
        (0, 0),
        (0, 2),
        (0, 50),
        (0, 89),
        (0, 290),
        (0, 7),
        (0, 79),
        (0, 449),
        (0, 217),
        (0, 42),
        (0, 130),
        (0, 1),
        (0, 0),
        (0, 0),
        (0, 5),
        (0, 64),
        (0, 23),
        (0, 107),
        (0, 53),
        (0, 37),
        (0, 3),
        (0, 0),
        (0, 36),
        (0, 0),
        (0, 0),
        (0, 84),
        (0, 118),
        (0, 16),
        (0, 0),
        (0, 1),
        (0, 304),
        (0, 720),
        (0, 137),
        (0, 13),
        (0, 327),
        (0, 0),
        (0, 1),
        (0, 0),
        (0, 25),
        (0, 397),
        (0, 0),
        (0, 837),
        (0, 14),
        (0, 6),
        (0, 299),
        (0, 1230),
        (0, 90),
        (0, 4),
        (0, 3),
        (0, 866),
        (0, 2),
        (0, 132),
        (0, 174),
        (0, 21),
        (0, 0),
        (0, 187),
        (0, 202),
        (0, 31),
        (0, 508),
        (0, 6),
        (0, 161),
        (0, 121),
        (0, 209),
        (0, 336),
        (0, 593),
        (0, 188),
        (0, 412),
        (0, 38),
        (0, 116),
        (0, 238),
        (0, 1),
        (0, 215),
        (0, 1),
        (0, 0),
        (0, 59),
        (0, 19),
        (0, 23),
        (0, 70),
        (0, 393),
        (0, 26),
        (0, 8),
        (0, 0),
        (0, 85),
        (0, 1602),
        (0, 935),
        (0, 39),
        (0, 7),
        (0, 190),
        (0, 0),
        (0, 157),
        (0, 39),
        (0, 7),
        (0, 213),
        (0, 33),
        (0, 13),
        (0, 72),
        (0, 9),
        (0, 309),
        (0, 205),
        (0, 489),
        (0, 22),
        (0, 225),
        (0, 62),
        (0, 22),
        (0, 22),
        (0, 6),
        (0, 209),
        (0, 1424),
        (0, 1),
        (0, 605),
        (0, 57),
        (0, 35),
        (0, 0),
        (0, 13),
        (0, 83),
        (0, 47),
        (0, 16),
        (0, 6),
        (0, 126),
        (0, 74),
        (0, 20),
        (0, 17),
        (0, 12),
        (0, 198),
        (0, 0),
        (0, 20),
        (0, 244),
        (0, 13),
        (0, 0),
        (0, 160),
        (0, 896),
        (0, 74),
        (0, 170),
        (0, 119),
        (0, 15),
        (0, 123),
        (0, 198),
        (0, 0),
        (0, 765),
        (0, 6),
        (0, 1427),
        (0, 946),
        (0, 234),
        (0, 1),
        (0, 186),
        (0, 0),
        (0, 74),
        (0, 16),
        (0, 211),
        (0, 1540),
        (0, 1930),
        (0, 1377),
        (0, 2),
        (0, 578),
        (0, 156),
        (0, 163),
        (0, 112),
        (0, 20),
        (0, 89),
        (0, 117),
        (0, 3521),
        (0, 1328),
        (0, 13),
        (0, 1652),
        (0, 9),
        (0, 1326),
        (0, 299),
        (0, 1),
        (0, 1),
        (0, 1265),
        (0, 3),
        (0, 31),
        (0, 126),
        (0, 75),
        (0, 0),
        (0, 235),
        (0, 2),
        (0, 1),
        (0, 2),
        (0, 0),
        (0, 121),
        (0, 22),
        (0, 172),
        (0, 1484),
        (0, 129),
        (0, 0),
        (0, 0),
        (0, 1363),
        (0, 747),
        (0, 78),
        (0, 104),
        (0, 1237),
        (0, 2386),
        (0, 2),
        (0, 29),
        (0, 45),
        (0, 14),
        (0, 47),
        (0, 90),
        (0, 23),
        (0, 13),
        (0, 2),
        (0, 364),
        (0, 88),
        (0, 225),
        (0, 212),
        (0, 97),
        (0, 29),
        (0, 0),
        (0, 109),
        (0, 356),
        (0, 1142),
        (0, 52),
        (0, 863),
        (0, 6),
        (0, 17),
        (0, 1353),
        (0, 310),
        (0, 1189),
        (0, 17),
        (0, 64),
        (0, 171),
        (0, 114),
        (0, 20),
        (0, 3),
        (0, 5282),
        (0, 44),
        (0, 0),
        (0, 2273),
        (0, 33),
        (0, 10),
        (0, 1317),
        (0, 911),
        (0, 61),
        (0, 5),
        (0, 47),
        (0, 125),
        (0, 22),
        (0, 279),
        (0, 1294),
        (0, 0),
        (0, 526),
        (0, 291),
        (0, 1),
        (0, 42),
        (0, 4),
        (0, 1),
        (0, 24),
        (0, 34),
        (0, 7),
        (0, 3),
        (0, 30),
        (0, 1547),
        (0, 6),
        (0, 1819),
        (0, 690),
        (0, 372),
        (0, 376),
        (0, 4),
        (0, 471),
        (0, 4),
        (0, 9),
        (0, 439),
        (0, 2861),
        (0, 14),
        (0, 108),
        (0, 389),
        (0, 0),
        (0, 52),
        (0, 63),
        (0, 2),
        (0, 12),
        (0, 29),
        (0, 92),
        (0, 64),
        (0, 157),
        (0, 1472),
        (0, 1),
        (0, 89),
        (0, 41),
        (0, 729),
        (0, 143),
        (0, 1435),
        (0, 385),
        (0, 22),
        (0, 849),
        (0, 73),
        (0, 0),
        (0, 7),
        (0, 1166),
        (0, 1298),
        (0, 163),
        (0, 9),
        (0, 12),
        (0, 15),
        (0, 916),
        (0, 0),
        (0, 270),
        (0, 96),
        (0, 219),
        (0, 86),
        (0, 0),
        (0, 4),
        (0, 144),
        (0, 2),
        (0, 114),
        (0, 1068),
        (0, 0),
        (0, 3762),
        (0, 318),
        (0, 18),
        (0, 295),
        (0, 227),
        (0, 75),
        (0, 1),
        (0, 2295),
        (0, 25),
        (0, 0),
        (0, 251),
        (0, 11),
        (0, 233),
        (0, 25),
        (0, 9),
        (0, 370),
        (0, 1),
        (0, 17),
        (0, 101),
        (0, 712),
        (0, 5),
        (0, 0),
        (0, 553),
        (0, 260),
        (0, 57),
        (0, 569),
        (0, 24),
        (0, 5287),
        (0, 511),
        (0, 416),
        (0, 19),
        (0, 1580),
        (0, 4703),
        (0, 605),
        (0, 21),
        (0, 299),
        (0, 1699),
        (0, 0),
        (0, 0),
        (0, 6),
        (0, 5),
        (0, 31),
        (0, 111),
        (0, 6),
        (0, 59),
        (0, 109),
        (0, 0),
        (0, 30),
        (0, 1),
        (0, 256),
        (0, 250),
        (0, 200),
        (0, 127),
        (0, 731),
        (0, 99),
        (0, 56),
        (0, 13),
        (0, 121),
        (0, 5),
        (0, 531),
        (0, 5257),
        (0, 239),
        (0, 4),
        (0, 10),
        (0, 1001),
        (0, 47),
        (0, 6),
        (0, 13),
        (0, 503),
        (0, 691),
        (0, 94),
        (0, 134),
        (0, 35),
        (0, 3430),
        (0, 8),
        (0, 575),
        (0, 20),
        (0, 340),
        (0, 3),
        (0, 179),
        (0, 186),
        (0, 1),
        (0, 287),
        (0, 79),
        (0, 263),
        (0, 14),
        (0, 0),
        (0, 6),
        (0, 136),
        (0, 22),
        (0, 113),
        (0, 9),
        (0, 410),
        (0, 0),
        (0, 64),
        (0, 502),
        (0, 122),
        (0, 0),
        (0, 817),
        (0, 59),
        (0, 142),
        (0, 18),
        (0, 0),
        (0, 53),
        (0, 0),
        (0, 5),
        (0, 6),
        (0, 56),
        (0, 47),
        (0, 210),
        (0, 834),
        (0, 533),
        (0, 931),
        (0, 4),
        (0, 95),
        (0, 336),
        (0, 471),
        (0, 478),
        (0, 3),
        (0, 11),
        (0, 89),
        (0, 92),
        (0, 3524),
        (0, 18),
        (0, 466),
        (0, 458),
        (0, 1630),
        (0, 13),
        (0, 5),
        (0, 61),
        (0, 0),
        (0, 200),
        (0, 308),
        (0, 201),
        (0, 487),
        (0, 12),
        (0, 6),
        (0, 1083),
        (0, 42),
        (0, 490),
        (0, 191),
        (0, 142),
        (0, 62),
        (0, 1372),
        (0, 22),
        (0, 249),
        (0, 1),
        (0, 503),
        (0, 377),
        (0, 1),
        (0, 1214),
        (0, 91),
        (0, 5),
        (0, 91),
        (0, 4),
        (0, 38),
        (0, 183),
        (0, 23),
        (0, 14),
        (0, 16),
        (0, 14),
        (0, 153),
        (0, 3926),
        (0, 98),
        (0, 3),
        (0, 0),
        (0, 72),
        (0, 38),
        (0, 16),
        (0, 9),
        (0, 32),
        (0, 76),
        (0, 823),
        (0, 9),
        (0, 39),
        (0, 10),
        (0, 1448),
        (0, 322),
        (0, 9),
        (0, 29),
        (0, 689),
        (0, 19),
        (0, 0),
        (0, 1133),
        (0, 159),
        (0, 0),
        (0, 0),
        (0, 294),
        (0, 328),
        (0, 5),
        (0, 132),
        (0, 182),
        (0, 1609),
        (0, 534),
        (0, 514),
        (0, 288),
        (0, 4704),
        (0, 0),
        (0, 5),
        (0, 40),
        (0, 10),
        (0, 503),
        (0, 164),
        (0, 37),
        (0, 1),
        (0, 297),
        (0, 128),
        (0, 31),
        (0, 27),
        (0, 437),
        (0, 1),
        (0, 236),
        (0, 0),
        (0, 0),
        (0, 27),
        (0, 37),
        (0, 1727),
        (0, 1),
        (0, 742),
        (0, 225),
        (0, 84),
        (0, 212),
        (0, 552),
        (0, 1344),
        (0, 460),
        (0, 108),
        (0, 5),
        (0, 106),
        (0, 238),
        (0, 607),
        (0, 124),
        (0, 115),
        (0, 385),
        (0, 0),
        (0, 107),
        (0, 15),
        (0, 1202),
        (0, 601),
        (0, 0),
        (0, 37),
        (0, 942),
        (0, 2300),
        (0, 2417),
        (0, 106),
        (0, 121),
        (0, 1696),
        (0, 19),
        (0, 830),
        (0, 1046),
        (0, 119),
        (0, 491),
        (0, 94),
        (0, 168),
        (0, 1),
        (0, 18),
        (0, 47),
        (0, 0),
        (0, 518),
        (0, 25),
        (0, 1543),
        (0, 130),
        (0, 160),
        (0, 0),
        (0, 1170),
        (0, 302),
        (0, 4756),
        (0, 3),
        (0, 0),
        (0, 1422),
        (0, 69),
        (0, 292),
        (0, 73),
        (0, 119),
        (0, 220),
        (0, 206),
        (0, 389),
        (0, 352),
        (0, 1496),
        (0, 1),
        (0, 0),
        (0, 0),
        (0, 1),
        (0, 722),
        (0, 4),
        (0, 99),
        (0, 141),
        (0, 1307),
        (0, 82),
        (0, 17),
        (0, 4),
        (0, 19),
        (0, 15),
        (0, 46),
        (0, 5),
        (0, 5),
        (0, 5),
        (0, 1121),
        (0, 8),
        (0, 7),
        (0, 0),
        (0, 2772),
        (0, 54),
        (0, 1225),
        (0, 66),
        (0, 841),
        (0, 258),
        (0, 2),
        (0, 0),
        (0, 533),
        (0, 1917),
        (0, 191),
        (0, 16),
        (0, 7),
        (0, 289),
        (0, 24),
        (0, 99),
        (0, 286),
        (0, 482),
        (0, 1754),
        (0, 3),
        (0, 2922),
        (0, 27),
        (0, 1229),
        (0, 2572),
        (0, 6),
        (0, 1352),
        (0, 0),
        (0, 91),
        (0, 85),
        (0, 69),
        (0, 333),
        (0, 1662),
        (0, 24),
        (0, 0),
        (0, 28),
        (0, 756),
        (0, 286),
        (0, 62),
        (0, 346),
        (0, 36),
        (0, 577),
        (0, 277),
        (0, 0),
        (0, 2),
        (0, 336),
        (0, 1028),
        (0, 33),
        (0, 5),
        (0, 7),
        (0, 1303),
        (0, 1309),
        (0, 50),
        (0, 1220),
        (0, 5),
        (0, 157),
        (0, 7),
        (0, 328),
        (0, 263),
        (0, 0),
        (0, 1),
        (0, 67),
        (0, 2335),
        (0, 87),
        (0, 50),
        (0, 262),
        (0, 680),
        (0, 3985),
        (0, 705),
        (0, 113),
        (0, 1635),
        (0, 69),
        (0, 6),
        (0, 478),
        (0, 19),
        (0, 4373),
        (0, 2),
        (0, 1232),
        (0, 3034),
        (0, 15),
        (0, 558),
        (0, 9),
        (0, 1306),
        (0, 7),
        (0, 1248),
        (0, 77),
        (0, 27),
        (0, 10),
        (0, 1),
        (0, 18),
        (0, 1357),
        (0, 0),
        (0, 6),
        (0, 1418),
        (0, 1),
        (0, 4046),
        (0, 171),
        (0, 44),
        (0, 68),
        (0, 426),
        (0, 1),
        (0, 1084),
        (0, 0),
        (0, 24),
        (0, 7657),
        (0, 14),
        (0, 30),
        (0, 120),
        (0, 479),
        (0, 75),
        (0, 1856),
        (0, 154),
        (0, 122),
        (0, 0),
        (0, 11),
        (0, 3),
        (0, 143),
        (0, 7),
        (0, 168),
        (0, 157),
        (0, 3),
        (0, 128),
        (0, 13),
        (0, 4719),
        (0, 41),
        (0, 19),
        (0, 298),
        (0, 581),
        (0, 2120),
        (0, 2035),
        (0, 140),
        (0, 0),
        (0, 0),
        (0, 17),
        (0, 199),
        (0, 1933),
        (0, 1060),
        (0, 501),
        (0, 133),
        (0, 898),
        (0, 1573),
        (0, 378),
        (0, 772),
        (0, 582),
        (0, 1602),
        (0, 7391),
        (1, 3),
        (0, 243),
        (0, 0),
        (0, 1100),
        (0, 92),
        (0, 28),
        (0, 31),
        (0, 127),
        (0, 2810),
        (0, 6),
        (0, 2971),
        (0, 1),
        (0, 22),
        (1, 537),
        (0, 1971),
        (0, 112),
        (0, 25),
        (0, 4),
        (0, 810),
        (0, 122),
        (0, 2),
        (0, 403),
        (0, 0),
        (0, 56),
        (0, 120),
        (0, 5),
        (0, 234),
        (0, 1345),
        (0, 107),
        (0, 509),
        (0, 3233),
        (0, 15),
        (0, 78),
        (0, 82),
        (0, 1),
        (0, 16),
        (0, 3362),
        (0, 1),
        (0, 161),
        (0, 81),
        (0, 504),
        (0, 87),
        (0, 8),
        (0, 74),
        (0, 208),
        (0, 825),
        (0, 262),
        (0, 4),
        (0, 63),
        (0, 3),
        (0, 1023),
        (0, 1423),
        (1, 1485),
        (0, 15),
        (0, 731),
        (0, 32),
        (0, 2626),
        (0, 14),
        (0, 0),
        (0, 785),
        (0, 848),
        (0, 219),
        (0, 7),
        (0, 172),
        (0, 127),
        (0, 4),
        (0, 392),
        (0, 3837),
        (0, 1349),
        (0, 0),
        (0, 213),
        (0, 2401),
        (0, 104),
        (0, 16),
        (0, 570),
        (0, 625),
        (0, 66),
        (0, 1225),
        (0, 0),
        (0, 219),
        (0, 0),
        (0, 6163),
        (0, 8),
        (0, 1898),
        (0, 159),
        (0, 1856),
        (0, 264),
        (0, 2589),
        (0, 0),
        (0, 4),
        (0, 23),
        (0, 1175),
        (0, 4690),
        (0, 991),
        (0, 0),
        (0, 237),
        (0, 6),
        (0, 58),
        (0, 624),
        (0, 431),
        (0, 1033),
        (0, 571),
        (0, 3555),
        (0, 22),
        (0, 7),
        (0, 113),
        (0, 6),
        (0, 1194),
        (0, 93),
        (0, 17),
        (0, 3),
        (0, 0),
        (0, 16),
        (0, 10),
        (0, 358),
        (0, 6),
        (0, 684),
        (0, 4),
        (0, 51),
        (0, 979),
        (0, 1839),
        (0, 4),
        (0, 102),
        (0, 74),
        (0, 0),
        (0, 2349),
        (0, 21),
        (0, 0),
        (1, 4630),
        (0, 269),
        (0, 60),
        (0, 5477),
        (0, 192),
        (0, 2503),
        (0, 578),
        (0, 1200),
        (0, 1),
        (0, 5),
        (0, 0),
        (0, 462),
        (0, 0),
        (0, 1684),
        (0, 3),
        (0, 79),
        (0, 2871),
        (0, 4341),
        (0, 164),
        (0, 10),
        (0, 2),
        (0, 0),
        (0, 4111),
        (0, 58),
        (0, 0),
        (0, 1185),
        (0, 0),
        (0, 198),
        (0, 1606),
        (0, 551),
        (0, 83),
        (0, 3),
        (0, 242),
        (0, 3),
        (0, 57),
        (0, 97),
        (0, 11),
        (0, 12),
        (0, 27),
        (0, 29),
        (0, 0),
        (0, 462),
        (0, 6488),
        (0, 14),
        (0, 383),
        (0, 33),
        (0, 2),
        (0, 1997),
        (0, 36),
        (0, 0),
        (0, 57),
        (0, 235),
        (0, 464),
        (0, 9),
        (0, 58),
        (0, 232),
        (0, 23),
        (0, 299),
        (0, 462),
        (0, 89),
        (0, 1287),
        (0, 279),
        (0, 0),
        (0, 363),
        (0, 56),
        (0, 0),
        (0, 128),
        (0, 16),
        (0, 0),
        (0, 19),
        (0, 22),
        (0, 346),
        (0, 34),
        (0, 1844),
        (0, 3),
        (0, 48),
        (0, 97),
        (0, 1752),
        (0, 1216),
        (0, 6),
        (0, 280),
        (0, 1),
        (0, 0),
        (0, 137),
        (0, 528),
        (0, 9),
        (0, 0),
        (0, 1930),
        (0, 19),
        (0, 0),
        (0, 12),
        (0, 1435),
        (0, 925),
        (0, 3286),
        (0, 306),
        (0, 233),
        (0, 14),
        (0, 30),
        (0, 7879),
        (0, 17),
        (0, 18),
        (0, 477),
        (0, 188),
        (0, 0),
        (0, 448),
        (0, 4),
        (0, 3857),
        (0, 2980),
        (0, 982),
        (1, 637),
        (0, 1),
        (0, 21),
        (0, 347),
        (0, 6),
        (0, 579),
        (0, 2382),
        (0, 3),
        (0, 387),
        (0, 75),
        (0, 1297),
        (0, 0),
        (0, 72),
        (0, 327),
        (0, 9),
        (0, 6),
        (1, 1499),
        (0, 249),
        (0, 392),
        (0, 46),
        (0, 746),
        (0, 2898),
        (0, 2421),
        (1, 6164),
        (0, 52),
        (0, 248),
        (0, 4855),
        (0, 22),
        (0, 112),
        (0, 0),
        (0, 1),
        (0, 45),
        (0, 920),
        (0, 389),
        (0, 4),
        (0, 22),
        (0, 0),
        (0, 103),
        (0, 56),
        (0, 1332),
        (0, 2639),
        (0, 0),
        (0, 1),
        (0, 128),
        (0, 0),
        (0, 248),
        (0, 3084),
        (1, 5479),
        (0, 217),
        (0, 5),
        (0, 7860),
        (0, 109),
        (0, 355),
        (0, 2),
        (0, 651),
        (0, 9),
        (0, 336),
        (0, 6838),
        (0, 0),
        (0, 1749),
        (0, 2258),
        (0, 6131),
        (0, 734),
        (0, 421),
        (0, 19),
        (0, 485),
        (0, 0),
        (0, 7),
        (0, 5),
        (0, 124),
        (1, 7567),
        (0, 3876),
        (0, 6),
        (0, 2583),
        (0, 6),
        (0, 1),
        (0, 1516),
        (0, 985),
        (0, 171),
        (0, 1385),
        (0, 2504),
        (0, 83),
        (0, 2),
        (0, 34),
        (0, 160),
        (0, 43),
        (0, 183),
        (0, 443),
        (0, 67),
        (0, 741),
        (0, 0),
        (0, 4680),
        (0, 2427),
        (0, 5507),
        (0, 17),
        (0, 416),
        (0, 267),
        (0, 79),
        (0, 12),
        (0, 88),
        (0, 146),
        (0, 5300),
        (0, 1),
        (0, 14),
        (0, 57),
        (0, 1),
        (0, 3028),
        (0, 3),
        (0, 1902),
        (0, 1005),
        (0, 0),
        (0, 0),
        (0, 701),
        (0, 1046),
        (0, 209),
        (0, 646),
        (0, 4897),
        (0, 1813),
        (0, 5411),
        (0, 0),
        (0, 1539),
        (0, 12),
        (0, 13),
        (0, 1),
        (0, 4705),
        (0, 66),
        (0, 2740),
        (0, 1947),
        (0, 2262),
        (0, 118),
        (0, 24),
        (0, 127),
        (0, 11),
        (0, 0),
        (0, 38),
        (0, 2478),
        (0, 12),
        (0, 4),
        (0, 2810),
        (0, 6),
        (0, 9),
        (0, 366),
        (0, 2397),
        (0, 463),
        (0, 547),
        (1, 92),
        (0, 613),
        (0, 10),
        (0, 4944),
        (0, 1472),
        (0, 176),
        (0, 1626),
        (0, 241),
        (0, 8),
        (0, 3852),
        (0, 143),
        (0, 3),
        (0, 3),
        (0, 3116),
        (0, 75),
        (0, 49),
        (0, 85),
        (0, 367),
        (0, 2),
        (0, 1066),
        (0, 6),
        (0, 7471),
        (0, 847),
        (0, 2246),
        (0, 2362),
        (0, 3097),
        (0, 0),
        (0, 81),
        (0, 163),
        (0, 33),
        (0, 5322),
        (0, 14),
        (0, 650),
        (0, 1443),
        (0, 15),
        (0, 14),
        (0, 645),
        (0, 60),
        (0, 0),
        (1, 4154),
        (0, 8),
        (0, 146),
        (0, 1039),
        (0, 696),
        (0, 4252),
        (0, 5825),
        (1, 7825),
        (0, 31),
        (1, 2402),
        (0, 3),
        (0, 6159),
        (0, 5),
        (0, 89),
        (1, 5115),
        (0, 73),
        (0, 806),
        (0, 263),
        (0, 5241),
        (0, 7772),
        (0, 124),
        (0, 354),
        (0, 5),
        (0, 2343),
        (0, 521),
        (0, 412),
        (0, 2807),
        (0, 0),
        (0, 168),
        (0, 1812),
        (0, 618),
        (0, 101),
        (0, 0),
        (0, 1594),
        (0, 35),
        (0, 13),
        (0, 754),
        (0, 5463),
        (0, 5),
        (0, 106),
        (0, 13),
        (0, 1776),
        (0, 26),
        (0, 1649),
        (0, 31),
        (0, 1047),
        (0, 1996),
        (0, 26),
        (1, 999),
        (0, 2667),
        (0, 4),
        (0, 2),
        (0, 149),
        (0, 38),
        (0, 1),
        (0, 1436),
        (1, 7292),
        (0, 2076),
        (0, 5),
        (0, 17),
        (0, 276),
        (0, 3263),
        (0, 130),
        (0, 0),
        (0, 7809),
        (0, 307),
        (0, 1773),
        (0, 1016),
        (0, 0),
        (0, 6755),
        (0, 4710),
        (0, 248),
        (0, 166),
        (0, 547),
        (0, 56),
        (0, 1119),
        (0, 0),
        (0, 522),
        (0, 10),
        (0, 1107),
        (0, 11),
        (0, 86),
        (0, 980),
        (2, 1310),
        (0, 2),
        (0, 879),
        (0, 10),
        (0, 80),
        (0, 13),
        (0, 1252),
        (0, 2056),
        (0, 206),
        (0, 300),
        (1, 5696),
        (0, 301),
        (0, 4),
        (0, 1542),
        (0, 2),
        (0, 6949),
        (0, 6254),
        (0, 3),
        (0, 76),
        (0, 1064),
        (0, 70),
        (0, 3120),
        (1, 4039),
        (0, 3786),
        (0, 216),
        (0, 7188),
        (0, 3996),
        (0, 1367),
        (0, 32),
        (0, 5),
        (0, 24),
        (0, 53),
        (1, 312),
        (0, 12),
        (0, 664),
        (0, 209),
        (0, 24),
        (1, 7780),
        (0, 2955),
        (0, 1),
        (0, 6377),
        (0, 98),
        (0, 83),
        (0, 1365),
        (0, 6),
        (0, 536),
        (0, 9),
        (0, 264),
        (2, 5426),
        (0, 31),
        (0, 92),
        (0, 2),
        (0, 44),
        (0, 529),
        (0, 0),
        (1, 6228),
        (0, 2),
        (0, 7526),
        (0, 7369),
        (0, 5223),
        (0, 834),
        (0, 1928),
        (0, 16),
        (0, 95),
        (0, 730),
        (0, 27),
        (0, 0),
        (0, 3386),
        (0, 3288),
        (0, 996),
        (0, 278),
        (0, 1230),
        (1, 1723),
        (0, 567),
        (0, 2708),
        (0, 530),
        (0, 1),
        (0, 17),
        (0, 4252),
        (0, 2424),
        (0, 290),
        (0, 0),
        (1, 3137),
        (0, 2103),
        (0, 15),
        (1, 15),
        (0, 1231),
        (0, 7),
        (0, 199),
        (0, 90),
        (0, 16),
        (1, 1304),
        (0, 268),
        (0, 7104),
        (0, 225),
        (0, 4),
        (0, 642),
        (0, 42),
        (0, 1),
        (0, 759),
        (0, 5),
        (0, 6910),
        (0, 8),
        (0, 5061),
        (0, 27),
        (0, 11),
        (0, 10),
        (0, 4387),
        (0, 7189),
        (0, 1151),
        (2, 3145),
        (0, 1076),
        (0, 0),
        (0, 0),
        (0, 36),
        (6, 502),
        (0, 882),
        (0, 28),
        (0, 886),
        (0, 2),
        (0, 1),
        (0, 66),
        (0, 25),
        (2, 3325),
        (0, 218),
        (0, 947),
        (0, 1),
        (0, 258),
        (0, 781),
        (0, 184),
        (0, 13),
        (0, 6355),
        (0, 52),
        (0, 28),
        (0, 4348),
        (1, 7254),
        (0, 3515),
        (0, 246),
        (0, 5),
        (0, 21),
        (4, 7554),
        (0, 6965),
        (0, 998),
        (2, 3436),
        (0, 6252),
        (0, 277),
        (1, 390),
        (0, 48),
        (0, 0),
        (0, 221),
        (0, 167),
        (0, 264),
        (2, 934),
        (0, 1),
        (0, 64),
        (4, 6508),
        (0, 1035),
        (0, 418),
        (0, 16),
        (1, 7827),
        (0, 583),
        (0, 290),
        (0, 897),
        (0, 478),
        (0, 74),
        (0, 7149),
        (0, 12),
        (0, 1),
        (0, 46),
        (0, 368),
        (0, 43),
        (0, 1),
        (0, 81),
        (0, 993),
        (0, 7),
        (0, 5),
        (7, 5940),
        (1, 850),
        (0, 7661),
        (0, 43),
        (0, 146),
        (0, 3),
        (0, 27),
        (0, 9),
        (0, 208),
        (0, 49),
        (0, 2),
        (0, 13),
        (0, 1695),
        (0, 0),
        (0, 7),
        (0, 449),
        (0, 1903),
        (0, 36),
        (0, 18),
        (0, 4071),
        (0, 1998),
        (0, 26),
        (0, 160),
        (0, 151),
        (6, 2821),
        (3, 1087),
        (0, 841),
        (0, 80),
        (0, 0),
        (0, 12),
        (0, 8),
        (0, 661),
        (0, 4840),
        (3, 5944),
        (0, 15),
        (0, 1859),
        (0, 1237),
        (1, 5201),
        (0, 126),
        (0, 2740),
        (0, 1567),
        (0, 98),
        (9, 6122),
        (0, 1331),
        (0, 3),
        (0, 38),
        (0, 22),
        (1, 3159),
        (0, 16),
        (0, 4673),
        (0, 200),
        (0, 297),
        (0, 29),
        (0, 25),
        (0, 371),
        (0, 75),
        (0, 4),
    ],
    entries: &[
        ("kku", Language::Kku),
        ("srs", Language::Srs),
        ("bds", Language::Bds),
        ("uwa", Language::Uwa),
        ("idi", Language::Idi),
        ("rea", Language::Rea),
        ("xtc", Language::Xtc),
        ("wrr", Language::Wrr),
        ("cdh", Language::Cdh),
        ("wwa", Language::Wwa),
        ("mzp", Language::Mzp),
        ("nmn", Language::Nmn),
        ("ama", Language::Ama),
        ("bsx", Language::Bsx),
        ("kol", Language::Kol),
        ("ltn", Language::Ltn),
        ("tyh", Language::Tyh),
        ("mez", Language::Mez),
        ("kdz", Language::Kdz),
        ("xpo", Language::Xpo),
        ("lke", Language::Lke),
        ("kan", Language::Kan),
        ("loi", Language::Loi),
        ("xti", Language::Xti),
        ("mmg", Language::Mmg),
        ("nsh", Language::Nsh),
        ("ztl", Language::Ztl),
        ("xtd", Language::Xtd),
        ("kxs", Language::Kxs),
        ("nle", Language::Nle),
        ("kpi", Language::Kpi),
        ("mvh", Language::Mvh),
        ("msv", Language::Msv),
        ("kno", Language::Kno),
        ("zul", Language::Zul),
        ("zbu", Language::Zbu),
        ("stv", Language::Stv),
        ("cyo", Language::Cyo),
        ("cnw", Language::Cnw),
        ("cmi", Language::Cmi),
        ("ura", Language::Ura),
        ("sqt", Language::Sqt),
        ("kcr", Language::Kcr),
        ("xhe", Language::Xhe),
        ("nol", Language::Nol),
        ("bnp", Language::Bnp),
        ("tzl", Language::Tzl),
        ("lcm", Language::Lcm),
        ("nsr", Language::Nsr),
        ("tbj", Language::Tbj),
        ("hue", Language::Hue),
        ("unm", Language::Unm),
        ("hux", Language::Hux),
        ("sww", Language::Sww),
        ("xmy", Language::Xmy),
        ("pad", Language::Pad),
        ("bua", Language::Bua),
        ("nyi", Language::Nyi),
        ("gez", Language::Gez),
        ("kiv", Language::Kiv),
        ("tkr", Language::Tkr),
        ("bka", Language::Bka),
        ("mbm", Language::Mbm),
        ("tco", Language::Tco),
        ("yip", Language::Yip),
        ("wji", Language::Wji),
        ("pfl", Language::Pfl),
        ("avn", Language::Avn),
        ("tdh", Language::Tdh),
        ("mmt", Language::Mmt),
        ("ury", Language::Ury),
        ("twm", Language::Twm),
        ("nmj", Language::Nmj),
        ("djb", Language::Djb),
        ("twt", Language::Twt),
        ("tyt", Language::Tyt),
        ("qxp", Language::Qxp),
        ("dre", Language::Dre),
        ("mzz", Language::Mzz),
        ("ymh", Language::Ymh),
        ("dme", Language::Dme),
        ("rup", Language::Rup),
        ("ppl", Language::Ppl),
        ("hir", Language::Hir),
        ("ycl", Language::Ycl),
        ("bys", Language::Bys),
        ("teg", Language::Teg),
        ("dmc", Language::Dmc),
        ("bne", Language::Bne),
        ("xke", Language::Xke),
        ("sia", Language::Sia),
        ("png", Language::Png),
        ("mnk", Language::Mnk),
        ("kuk", Language::Kuk),
        ("cop", Language::Cop),
        ("tuf", Language::Tuf),
        ("ake", Language::Ake),
        ("xki", Language::Xki),
        ("rmy", Language::Rmy),
        ("efi", Language::Efi),
        ("nmr", Language::Nmr),
        ("mgr", Language::Mgr),
        ("sqa", Language::Sqa),
        ("nrm", Language::Nrm),
        ("pom", Language::Pom),
        ("naw", Language::Naw),
        ("wss", Language::Wss),
        ("ker", Language::Ker),
        ("oku", Language::Oku),
        ("zyn", Language::Zyn),
        ("xhr", Language::Xhr),
        ("huz", Language::Huz),
        ("myp", Language::Myp),
        ("glg", Language::Glg),
        ("emz", Language::Emz),
        ("eiv", Language::Eiv),
        ("zxx", Language::Zxx),
        ("mqj", Language::Mqj),
        ("ubi", Language::Ubi),
        ("lro", Language::Lro),
        ("dgz", Language::Dgz),
        ("bgj", Language::Bgj),
        ("mqk", Language::Mqk),
        ("xas", Language::Xas),
        ("oro", Language::Oro),
        ("ayb", Language::Ayb),
        ("bll", Language::Bll),
        ("mci", Language::Mci),
        ("tyu", Language::Tyu),
        ("nhx", Language::Nhx),
        ("bmh", Language::Bmh),
        ("omo", Language::Omo),
        ("slj", Language::Slj),
        ("bvq", Language::Bvq),
        ("yii", Language::Yii),
        ("dna", Language::Dna),
        ("ycp", Language::Ycp),
        ("baa", Language::Baa),
        ("gle", Language::Gle),
        ("xhm", Language::Xhm),
        ("knl", Language::Knl),
        ("smf", Language::Smf),
        ("gsp", Language::Gsp),
        ("vap", Language::Vap),
        ("msg", Language::Msg),
        ("dho", Language::Dho),
        ("zku", Language::Zku),
        ("kbr", Language::Kbr),
        ("btp", Language::Btp),
        ("lbf", Language::Lbf),
        ("tsk", Language::Tsk),
        ("gke", Language::Gke),
        ("kww", Language::Kww),
        ("cja", Language::Cja),
        ("dgo", Language::Dgo),
        ("rkt", Language::Rkt),
        ("bsg", Language::Bsg),
        ("rir", Language::Rir),
        ("tik", Language::Tik),
        ("itk", Language::Itk),
        ("zpf", Language::Zpf),
        ("tif", Language::Tif),
        ("sqk", Language::Sqk),
        ("enl", Language::Enl),
        ("xwg", Language::Xwg),
        ("gdl", Language::Gdl),
        ("imn", Language::Imn),
        ("aed", Language::Aed),
        ("inb", Language::Inb),
        ("mkm", Language::Mkm),
        ("ktm", Language::Ktm),
        ("bao", Language::Bao),
        ("spg", Language::Spg),
        ("tgy", Language::Tgy),
        ("sen", Language::Sen),
        ("ott", Language::Ott),
        ("aha", Language::Aha),
        ("rxw", Language::Rxw),
        ("gdb", Language::Gdb),
        ("jeh", Language::Jeh),
        ("yut", Language::Yut),
        ("ego", Language::Ego),
        ("ipk", Language::Ipk),
        ("bno", Language::Bno),
        ("fas", Language::Fas),
        ("oym", Language::Oym),
        ("gse", Language::Gse),
        ("yga", Language::Yga),
        ("mow", Language::Mow),
        ("beo", Language::Beo),
        ("bkh", Language::Bkh),
        ("nee", Language::Nee),
        ("mbc", Language::Mbc),
        ("nlk", Language::Nlk),
        ("bxk", Language::Bxk),
        ("sbz", Language::Sbz),
        ("wdk", Language::Wdk),
        ("has", Language::Has),
        ("djo", Language::Djo),
        ("bjk", Language::Bjk),
        ("wxa", Language::Wxa),
        ("nut", Language::Nut),
        ("gbq", Language::Gbq),
        ("brn", Language::Brn),
        ("rnn", Language::Rnn),
        ("kyk", Language::Kyk),
        ("tyy", Language::Tyy),
        ("aze", Language::Aze),
        ("spm", Language::Spm),
        ("gbj", Language::Gbj),
        ("phq", Language::Phq),
        ("mrw", Language::Mrw),
        ("ert", Language::Ert),
        ("ben", Language::Ben),
        ("jiv", Language::Jiv),
        ("mhf", Language::Mhf),
        ("bnu", Language::Bnu),
        ("psq", Language::Psq),
        ("dto", Language::Dto),
        ("jdt", Language::Jdt),
        ("dmy", Language::Dmy),
        ("mgq", Language::Mgq),
        ("laj", Language::Laj),
        ("usi", Language::Usi),
        ("klp", Language::Klp),
        ("mzs", Language::Mzs),
        ("txi", Language::Txi),
        ("key", Language::Key),
        ("cal", Language::Cal),
        ("ule", Language::Ule),
        ("kux", Language::Kux),
        ("lbu", Language::Lbu),
        ("nyu", Language::Nyu),
        ("sgz", Language::Sgz),
        ("poh", Language::Poh),
        ("gwr", Language::Gwr),
        ("lda", Language::Lda),
        ("tic", Language::Tic),
        ("ngv", Language::Ngv),
        ("mso", Language::Mso),
        ("mia", Language::Mia),
        ("kyc", Language::Kyc),
        ("mnl", Language::Mnl),
        ("lov", Language::Lov),
        ("mix", Language::Mix),
        ("kaz", Language::Kaz),
        ("rmb", Language::Rmb),
        ("lmo", Language::Lmo),
        ("mrt", Language::Mrt),
        ("kkp", Language::Kkp),
        ("ram", Language::Ram),
        ("apl", Language::Apl),
        ("udm", Language::Udm),
        ("mfe", Language::Mfe),
        ("mue", Language::Mue),
        ("myo", Language::Myo),
        ("gkd", Language::Gkd),
        ("afr", Language::Afr),
        ("tbw", Language::Tbw),
        ("egl", Language::Egl),
        ("ysm", Language::Ysm),
        ("krp", Language::Krp),
        ("ebk", Language::Ebk),
        ("ari", Language::Ari),
        ("plg", Language::Plg),
        ("mab", Language::Mab),
        ("zmg", Language::Zmg),
        ("bov", Language::Bov),
        ("sve", Language::Sve),
        ("nwm", Language::Nwm),
        ("dii", Language::Dii),
        ("mgj", Language::Mgj),
        ("gbl", Language::Gbl),
        ("xpz", Language::Xpz),
        ("bjb", Language::Bjb),
        ("tmc", Language::Tmc),
        ("sry", Language::Sry),
        ("pok", Language::Pok),
        ("szn", Language::Szn),
        ("xir", Language::Xir),
        ("nnb", Language::Nnb),
        ("nyy", Language::Nyy),
        ("bzo", Language::Bzo),
        ("apg", Language::Apg),
        ("emx", Language::Emx),
        ("aub", Language::Aub),
        ("mrl", Language::Mrl),
        ("sii", Language::Sii),
        ("cfg", Language::Cfg),
        ("nyb", Language::Nyb),
        ("wrb", Language::Wrb),
        ("sbg", Language::Sbg),
        ("cuk", Language::Cuk),
        ("niv", Language::Niv),
        ("sun", Language::Sun),
        ("mxu", Language::Mxu),
        ("tsl", Language::Tsl),
        ("tbv", Language::Tbv),
        ("dyu", Language::Dyu),
        ("she", Language::She),
        ("mah", Language::Mah),
        ("bip", Language::Bip),
        ("ica", Language::Ica),
        ("mua", Language::Mua),
        ("gha", Language::Gha),
        ("lbq", Language::Lbq),
        ("xkr", Language::Xkr),
        ("mby", Language::Mby),
        ("sho", Language::Sho),
        ("abu", Language::Abu),
        ("rpn", Language::Rpn),
        ("tat", Language::Tat),
        ("lhi", Language::Lhi),
        ("nek", Language::Nek),
        ("axb", Language::Axb),
        ("cek", Language::Cek),
        ("hma", Language::Hma),
        ("ryn", Language::Ryn),
        ("sgt", Language::Sgt),
        ("zyg", Language::Zyg),
        ("ard", Language::Ard),
        ("gmh", Language::Gmh),
        ("mkb", Language::Mkb),
        ("aah", Language::Aah),
        ("car", Language::Car),
        ("bug", Language::Bug),
        ("tdn", Language::Tdn),
        ("wut", Language::Wut),
        ("str", Language::Str),
        ("chw", Language::Chw),
        ("zpj", Language::Zpj),
        ("gww", Language::Gww),
        ("tou", Language::Tou),
        ("mnh", Language::Mnh),
        ("ssh", Language::Ssh),
        ("bbp", Language::Bbp),
        ("jwi", Language::Jwi),
        ("blh", Language::Blh),
        ("grv", Language::Grv),
        ("aqp", Language::Aqp),
        ("tef", Language::Tef),
        ("msi", Language::Msi),
        ("yor", Language::Yor),
        ("apq", Language::Apq),
        ("paq", Language::Paq),
        ("sht", Language::Sht),
        ("ijn", Language::Ijn),
        ("ksw", Language::Ksw),
        ("brs", Language::Brs),
        ("ctg", Language::Ctg),
        ("grz", Language::Grz),
        ("jsl", Language::Jsl),
        ("lrr", Language::Lrr),
        ("jae", Language::Jae),
        ("kgg", Language::Kgg),
        ("cnp", Language::Cnp),
        ("lnb", Language::Lnb),
        ("tkv", Language::Tkv),
        ("bgv", Language::Bgv),
        ("mbp", Language::Mbp),
        ("sre", Language::Sre),
        ("bof", Language::Bof),
        ("tjw", Language::Tjw),
        ("ala", Language::Ala),
        ("wiu", Language::Wiu),
        ("abg", Language::Abg),
        ("sns", Language::Sns),
        ("zlm", Language::Zlm),
        ("erw", Language::Erw),
        ("adz", Language::Adz),
        ("xsd", Language::Xsd),
        ("mzr", Language::Mzr),
        ("mdh", Language::Mdh),
        ("udi", Language::Udi),
        ("kfc", Language::Kfc),
        ("uji", Language::Uji),
        ("mge", Language::Mge),
        ("bhl", Language::Bhl),
        ("msb", Language::Msb),
        ("dbp", Language::Dbp),
        ("svc", Language::Svc),
        ("sga", Language::Sga),
        ("kxb", Language::Kxb),
        ("kfj", Language::Kfj),
        ("jek", Language::Jek),
        ("ipi", Language::Ipi),
        ("mco", Language::Mco),
        ("xna", Language::Xna),
        ("pks", Language::Pks),
        ("gya", Language::Gya),
        ("ttn", Language::Ttn),
        ("kru", Language::Kru),
        ("ula", Language::Ula),
        ("kpb", Language::Kpb),
        ("agt", Language::Agt),
        ("auw", Language::Auw),
        ("smv", Language::Smv),
        ("anl", Language::Anl),
        ("fud", Language::Fud),
        ("tds", Language::Tds),
        ("kgm", Language::Kgm),
        ("plv", Language::Plv),
        ("mtl", Language::Mtl),
        ("kay", Language::Kay),
        ("app", Language::App),
        ("xya", Language::Xya),
        ("num", Language::Num),
        ("inn", Language::Inn),
        ("tbx", Language::Tbx),
        ("slf", Language::Slf),
        ("kzr", Language::Kzr),
        ("wbq", Language::Wbq),
        ("aia", Language::Aia),
        ("mxt", Language::Mxt),
        ("pez", Language::Pez),
        ("arw", Language::Arw),
        ("agm", Language::Agm),
        ("aai", Language::Aai),
        ("kie", Language::Kie),
        ("nat", Language::Nat),
        ("mzm", Language::Mzm),
        ("mzl", Language::Mzl),
        ("dip", Language::Dip),
        ("pih", Language::Pih),
        ("itl", Language::Itl),
        ("kjm", Language::Kjm),
        ("bez", Language::Bez),
        ("quz", Language::Quz),
        ("tey", Language::Tey),
        ("sbv", Language::Sbv),
        ("aik", Language::Aik),
        ("kuj", Language::Kuj),
        ("hvv", Language::Hvv),
        ("rhp", Language::Rhp),
        ("bbd", Language::Bbd),
        ("dne", Language::Dne),
        ("kmt", Language::Kmt),
        ("pwr", Language::Pwr),
        ("wpc", Language::Wpc),
        ("bqb", Language::Bqb),
        ("etu", Language::Etu),
        ("gow", Language::Gow),
        ("yue", Language::Yue),
        ("mls", Language::Mls),
        ("ctm", Language::Ctm),
        ("nuk", Language::Nuk),
        ("nes", Language::Nes),
        ("bcp", Language::Bcp),
        ("mrb", Language::Mrb),
        ("boj", Language::Boj),
        ("dkr", Language::Dkr),
        ("mki", Language::Mki),
        ("jud", Language::Jud),
        ("neg", Language::Neg),
        ("mye", Language::Mye),
        ("baf", Language::Baf),
        ("apj", Language::Apj),
        ("bmm", Language::Bmm),
        ("fuq", Language::Fuq),
        ("tsx", Language::Tsx),
        ("mln", Language::Mln),
        ("sef", Language::Sef),
        ("slm", Language::Slm),
        ("xly", Language::Xly),
        ("khu", Language::Khu),
        ("sbo", Language::Sbo),
        ("mvr", Language::Mvr),
        ("ldg", Language::Ldg),
        ("tmh", Language::Tmh),
        ("hnj", Language::Hnj),
        ("ldi", Language::Ldi),
        ("ksq", Language::Ksq),
        ("sad", Language::Sad),
        ("ndo", Language::Ndo),
        ("xnt", Language::Xnt),
        ("ziw", Language::Ziw),
        ("apy", Language::Apy),
        ("han", Language::Han),
        ("kgy", Language::Kgy),
        ("ytp", Language::Ytp),
        ("bzf", Language::Bzf),
        ("bjo", Language::Bjo),
        ("mim", Language::Mim),
        ("sao", Language::Sao),
        ("ggw", Language::Ggw),
        ("var", Language::Var),
        ("otb", Language::Otb),
        ("umd", Language::Umd),
        ("did", Language::Did),
        ("mmj", Language::Mmj),
        ("aza", Language::Aza),
        ("aft", Language::Aft),
        ("nnf", Language::Nnf),
        ("crf", Language::Crf),
        ("mqa", Language::Mqa),
        ("nhi", Language::Nhi),
        ("ylm", Language::Ylm),
        ("wms", Language::Wms),
        ("khj", Language::Khj),
        ("mlw", Language::Mlw),
        ("ail", Language::Ail),
        ("abf", Language::Abf),
        ("scs", Language::Scs),
        ("sbk", Language::Sbk),
        ("hmh", Language::Hmh),
        ("acs", Language::Acs),
        ("duy", Language::Duy),
        ("twi", Language::Twi),
        ("mkk", Language::Mkk),
        ("hax", Language::Hax),
        ("sux", Language::Sux),
        ("kna", Language::Kna),
        ("acy", Language::Acy),
        ("gvy", Language::Gvy),
        ("bes", Language::Bes),
        ("tmd", Language::Tmd),
        ("bje", Language::Bje),
        ("oub", Language::Oub),
        ("abm", Language::Abm),
        ("xcr", Language::Xcr),
        ("ure", Language::Ure),
        ("jmd", Language::Jmd),
        ("jii", Language::Jii),
        ("unz", Language::Unz),
        ("mpa", Language::Mpa),
        ("srf", Language::Srf),
        ("tew", Language::Tew),
        ("obo", Language::Obo),
        ("bel", Language::Bel),
        ("xmw", Language::Xmw),
        ("puy", Language::Puy),
        ("owi", Language::Owi),
        ("jah", Language::Jah),
        ("ces", Language::Ces),
        ("jiy", Language::Jiy),
        ("cpb", Language::Cpb),
        ("axg", Language::Axg),
        ("zns", Language::Zns),
        ("ank", Language::Ank),
        ("kqs", Language::Kqs),
        ("moh", Language::Moh),
        ("hsb", Language::Hsb),
        ("ixc", Language::Ixc),
        ("zgr", Language::Zgr),
        ("aap", Language::Aap),
        ("afd", Language::Afd),
        ("yar", Language::Yar),
        ("hus", Language::Hus),
        ("gmu", Language::Gmu),
        ("bob", Language::Bob),
        ("pgd", Language::Pgd),
        ("mbb", Language::Mbb),
        ("ila", Language::Ila),
        ("uzn", Language::Uzn),
        ("muu", Language::Muu),
        ("nha", Language::Nha),
        ("maw", Language::Maw),
        ("sil", Language::Sil),
        ("oca", Language::Oca),
        ("xin", Language::Xin),
        ("mnd", Language::Mnd),
        ("mqw", Language::Mqw),
        ("spr", Language::Spr),
        ("kqk", Language::Kqk),
        ("tcz", Language::Tcz),
        ("msy", Language::Msy),
        ("nig", Language::Nig),
        ("hji", Language::Hji),
        ("bqi", Language::Bqi),
        ("gae", Language::Gae),
        ("cuw", Language::Cuw),
        ("ams", Language::Ams),
        ("dbm", Language::Dbm),
        ("sci", Language::Sci),
        ("jog", Language::Jog),
        ("xpd", Language::Xpd),
        ("aga", Language::Aga),
        ("chy", Language::Chy),
        ("bdt", Language::Bdt),
        ("kak", Language::Kak),
        ("mnq", Language::Mnq),
        ("wom", Language::Wom),
        ("isk", Language::Isk),
        ("stl", Language::Stl),
        ("sds", Language::Sds),
        ("liy", Language::Liy),
        ("tdt", Language::Tdt),
        ("eve", Language::Eve),
        ("trf", Language::Trf),
        ("dba", Language::Dba),
        ("xsu", Language::Xsu),
        ("hia", Language::Hia),
        ("ako", Language::Ako),
        ("kha", Language::Kha),
        ("prm", Language::Prm),
        ("djd", Language::Djd),
        ("tua", Language::Tua),
        ("anp", Language::Anp),
        ("kir", Language::Kir),
        ("ply", Language::Ply),
        ("sbl", Language::Sbl),
        ("foi", Language::Foi),
        ("ykn", Language::Ykn),
        ("skr", Language::Skr),
        ("coa", Language::Coa),
        ("hkh", Language::Hkh),
        ("quk", Language::Quk),
        ("kcy", Language::Kcy),
        ("che", Language::Che),
        ("msu", Language::Msu),
        ("kdw", Language::Kdw),
        ("bve", Language::Bve),
        ("lej", Language::Lej),
        ("gpe", Language::Gpe),
        ("kmq", Language::Kmq),
        ("xpl", Language::Xpl),
        ("swe", Language::Swe),
        ("ute", Language::Ute),
        ("rac", Language::Rac),
        ("lum", Language::Lum),
        ("coc", Language::Coc),
        ("ikx", Language::Ikx),
        ("rmk", Language::Rmk),
        ("tcb", Language::Tcb),
        ("dmg", Language::Dmg),
        ("duk", Language::Duk),
        ("kbe", Language::Kbe),
        ("hij", Language::Hij),
        ("lrt", Language::Lrt),
        ("osu", Language::Osu),
        ("mpl", Language::Mpl),
        ("xwr", Language::Xwr),
        ("nkx", Language::Nkx),
        ("mro", Language::Mro),
        ("bkr", Language::Bkr),
        ("aak", Language::Aak),
        ("ndx", Language::Ndx),
        ("poy", Language::Poy),
        ("ken", Language::Ken),
        ("kfv", Language::Kfv),
        ("xco", Language::Xco),
        ("qvc", Language::Qvc),
        ("mxj", Language::Mxj),
        ("mos", Language::Mos),
        ("kqd", Language::Kqd),
        ("zsl", Language::Zsl),
        ("imo", Language::Imo),
        ("daj", Language::Daj),
        ("anb", Language::Anb),
        ("adx", Language::Adx),
        ("bqz", Language::Bqz),
        ("mjh", Language::Mjh),
        ("ych", Language::Ych),
        ("qus", Language::Qus),
        ("idb", Language::Idb),
        ("xtb", Language::Xtb),
        ("bwa", Language::Bwa),
        ("kjk", Language::Kjk),
        ("bfy", Language::Bfy),
        ("grj", Language::Grj),
        ("tja", Language::Tja),
        ("ukh", Language::Ukh),
        ("lob", Language::Lob),
        ("gbv", Language::Gbv),
        ("tdl", Language::Tdl),
        ("pmx", Language::Pmx),
        ("bls", Language::Bls),
        ("oml", Language::Oml),
        ("sot", Language::Sot),
        ("ikz", Language::Ikz),
        ("nbr", Language::Nbr),
        ("rmf", Language::Rmf),
        ("win", Language::Win),
        ("mfy", Language::Mfy),
        ("ztn", Language::Ztn),
        ("rej", Language::Rej),
        ("ipo", Language::Ipo),
        ("snc", Language::Snc),
        ("dgs", Language::Dgs),
        ("nnd", Language::Nnd),
        ("ktg", Language::Ktg),
        ("buw", Language::Buw),
        ("yxy", Language::Yxy),
        ("poe", Language::Poe),
        ("xcv", Language::Xcv),
        ("dbl", Language::Dbl),
        ("xpp", Language::Xpp),
        ("hoz", Language::Hoz),
        ("haa", Language::Haa),
        ("xls", Language::Xls),
        ("kvk", Language::Kvk),
        ("xes", Language::Xes),
        ("mrm", Language::Mrm),
        ("saj", Language::Saj),
        ("far", Language::Far),
        ("kmn", Language::Kmn),
        ("xty", Language::Xty),
        ("xel", Language::Xel),
        ("ifb", Language::Ifb),
        ("ulm", Language::Ulm),
        ("gey", Language::Gey),
        ("how", Language::How),
        ("tab", Language::Tab),
        ("dmo", Language::Dmo),
        ("vmy", Language::Vmy),
        ("daw", Language::Daw),
        ("rut", Language::Rut),
        ("pcp", Language::Pcp),
        ("kdf", Language::Kdf),
        ("mgl", Language::Mgl),
        ("yec", Language::Yec),
        ("lvu", Language::Lvu),
        ("wah", Language::Wah),
        ("nvm", Language::Nvm),
        ("aew", Language::Aew),
        ("emw", Language::Emw),
        ("kkt", Language::Kkt),
        ("wep", Language::Wep),
        ("las", Language::Las),
        ("ldp", Language::Ldp),
        ("ism", Language::Ism),
        ("clt", Language::Clt),
        ("kdj", Language::Kdj),
        ("lay", Language::Lay),
        ("lit", Language::Lit),
        ("kks", Language::Kks),
        ("kpz", Language::Kpz),
        ("mgd", Language::Mgd),
        ("ang", Language::Ang),
        ("tnh", Language::Tnh),
        ("zrn", Language::Zrn),
        ("tpg", Language::Tpg),
        ("smj", Language::Smj),
        ("mju", Language::Mju),
        ("air", Language::Air),
        ("lal", Language::Lal),
        ("etz", Language::Etz),
        ("ofs", Language::Ofs),
        ("ntd", Language::Ntd),
        ("eto", Language::Eto),
        ("bvh", Language::Bvh),
        ("erh", Language::Erh),
        ("nxl", Language::Nxl),
        ("nae", Language::Nae),
        ("avu", Language::Avu),
        ("pck", Language::Pck),
        ("pcn", Language::Pcn),
        ("jax", Language::Jax),
        ("kzn", Language::Kzn),
        ("itm", Language::Itm),
        ("nab", Language::Nab),
        ("kah", Language::Kah),
        ("brx", Language::Brx),
        ("kcs", Language::Kcs),
        ("kda", Language::Kda),
        ("mtg", Language::Mtg),
        ("nwy", Language::Nwy),
        ("moq", Language::Moq),
        ("cie", Language::Cie),
        ("yim", Language::Yim),
        ("pnx", Language::Pnx),
        ("xhc", Language::Xhc),
        ("ttc", Language::Ttc),
        ("kae", Language::Kae),
        ("plw", Language::Plw),
        ("bjy", Language::Bjy),
        ("tlk", Language::Tlk),
        ("vkl", Language::Vkl),
        ("tzo", Language::Tzo),
        ("ywa", Language::Ywa),
        ("esu", Language::Esu),
        ("msh", Language::Msh),
        ("ler", Language::Ler),
        ("sws", Language::Sws),
        ("tow", Language::Tow),
        ("nng", Language::Nng),
        ("nyh", Language::Nyh),
        ("ggt", Language::Ggt),
        ("ndl", Language::Ndl),
        ("jeb", Language::Jeb),
        ("cmg", Language::Cmg),
        ("bxp", Language::Bxp),
        ("weg", Language::Weg),
        ("kkn", Language::Kkn),
        ("tju", Language::Tju),
        ("rbk", Language::Rbk),
        ("vwa", Language::Vwa),
        ("ncz", Language::Ncz),
        ("sby", Language::Sby),
        ("jid", Language::Jid),
        ("fse", Language::Fse),
        ("gdq", Language::Gdq),
        ("xbe", Language::Xbe),
        ("gup", Language::Gup),
        ("paf", Language::Paf),
        ("zau", Language::Zau),
        ("wux", Language::Wux),
        ("xtj", Language::Xtj),
        ("dif", Language::Dif),
        ("bpl", Language::Bpl),
        ("kqm", Language::Kqm),
        ("chh", Language::Chh),
        ("oua", Language::Oua),
        ("mrj", Language::Mrj),
        ("snn", Language::Snn),
        ("mhu", Language::Mhu),
        ("kde", Language::Kde),
        ("rng", Language::Rng),
        ("bhm", Language::Bhm),
        ("hgw", Language::Hgw),
        ("ibr", Language::Ibr),
        ("ckq", Language::Ckq),
        ("ixl", Language::Ixl),
        ("bvt", Language::Bvt),
        ("bms", Language::Bms),
        ("szy", Language::Szy),
        ("nil", Language::Nil),
        ("ktt", Language::Ktt),
        ("tks", Language::Tks),
        ("dbf", Language::Dbf),
        ("cpi", Language::Cpi),
        ("zmd", Language::Zmd),
        ("mpk", Language::Mpk),
        ("sjp", Language::Sjp),
        ("dai", Language::Dai),
        ("xao", Language::Xao),
        ("yha", Language::Yha),
        ("jms", Language::Jms),
        ("bzb", Language::Bzb),
        ("skg", Language::Skg),
        ("nep", Language::Nep),
        ("inj", Language::Inj),
        ("ztu", Language::Ztu),
        ("kmb", Language::Kmb),
        ("txr", Language::Txr),
        ("aea", Language::Aea),
        ("xng", Language::Xng),
        ("bmz", Language::Bmz),
        ("yug", Language::Yug),
        ("fie", Language::Fie),
        ("tss", Language::Tss),
        ("sas", Language::Sas),
        ("lon", Language::Lon),
        ("aqd", Language::Aqd),
        ("lsa", Language::Lsa),
        ("dox", Language::Dox),
        ("bvy", Language::Bvy),
        ("tec", Language::Tec),
        ("lsh", Language::Lsh),
        ("wnw", Language::Wnw),
        ("caf", Language::Caf),
        ("gdf", Language::Gdf),
        ("nsz", Language::Nsz),
        ("ttt", Language::Ttt),
        ("lzl", Language::Lzl),
        ("kqb", Language::Kqb),
        ("maf", Language::Maf),
        ("trg", Language::Trg),
        ("env", Language::Env),
        ("lme", Language::Lme),
        ("icl", Language::Icl),
        ("kft", Language::Kft),
        ("arj", Language::Arj),
        ("irx", Language::Irx),
        ("ahb", Language::Ahb),
        ("azj", Language::Azj),
        ("yej", Language::Yej),
        ("fab", Language::Fab),
        ("cbs", Language::Cbs),
        ("hmo", Language::Hmo),
        ("rbb", Language::Rbb),
        ("ste", Language::Ste),
        ("kuz", Language::Kuz),
        ("lab", Language::Lab),
        ("bri", Language::Bri),
        ("dsh", Language::Dsh),
        ("aiw", Language::Aiw),
        ("nds", Language::Nds),
        ("nnh", Language::Nnh),
        ("emm", Language::Emm),
        ("xvo", Language::Xvo),
        ("ykr", Language::Ykr),
        ("bcg", Language::Bcg),
        ("mor", Language::Mor),
        ("mfd", Language::Mfd),
        ("tbc", Language::Tbc),
        ("cuy", Language::Cuy),
        ("fax", Language::Fax),
        ("ggd", Language::Ggd),
        ("ywg", Language::Ywg),
        ("miw", Language::Miw),
        ("xbc", Language::Xbc),
        ("kwe", Language::Kwe),
        ("dwa", Language::Dwa),
        ("nmi", Language::Nmi),
        ("krj", Language::Krj),
        ("tnp", Language::Tnp),
        ("itr", Language::Itr),
        ("onw", Language::Onw),
        ("gku", Language::Gku),
        ("iks", Language::Iks),
        ("doe", Language::Doe),
        ("squ", Language::Squ),
        ("col", Language::Col),
        ("hac", Language::Hac),
        ("mfj", Language::Mfj),
        ("gbz", Language::Gbz),
        ("lid", Language::Lid),
        ("njl", Language::Njl),
        ("duh", Language::Duh),
        ("bxl", Language::Bxl),
        ("gwj", Language::Gwj),
        ("aix", Language::Aix),
        ("baw", Language::Baw),
        ("nya", Language::Nya),
        ("lnh", Language::Lnh),
        ("xmm", Language::Xmm),
        ("kgn", Language::Kgn),
        ("enr", Language::Enr),
        ("wll", Language::Wll),
        ("kug", Language::Kug),
        ("rri", Language::Rri),
        ("pla", Language::Pla),
        ("par", Language::Par),
        ("hmb", Language::Hmb),
        ("bfb", Language::Bfb),
        ("ral", Language::Ral),
        ("sxb", Language::Sxb),
        ("ats", Language::Ats),
        ("lwt", Language::Lwt),
        ("nuz", Language::Nuz),
        ("frd", Language::Frd),
        ("wbe", Language::Wbe),
        ("kmg", Language::Kmg),
        ("gbg", Language::Gbg),
        ("hrw", Language::Hrw),
        ("awy", Language::Awy),
        ("mcl", Language::Mcl),
        ("nup", Language::Nup),
        ("lwh", Language::Lwh),
        ("cjv", Language::Cjv),
        ("pak", Language::Pak),
        ("lli", Language::Lli),
        ("bwt", Language::Bwt),
        ("xts", Language::Xts),
        ("bjr", Language::Bjr),
        ("dgb", Language::Dgb),
        ("yad", Language::Yad),
        ("yik", Language::Yik),
        ("bzm", Language::Bzm),
        ("doz", Language::Doz),
        ("meu", Language::Meu),
        ("drl", Language::Drl),
        ("bpk", Language::Bpk),
        ("cqd", Language::Cqd),
        ("ria", Language::Ria),
        ("pwn", Language::Pwn),
        ("bzs", Language::Bzs),
        ("mpg", Language::Mpg),
        ("poo", Language::Poo),
        ("nxk", Language::Nxk),
        ("ymc", Language::Ymc),
        ("cbw", Language::Cbw),
        ("iti", Language::Iti),
        ("kju", Language::Kju),
        ("xht", Language::Xht),
        ("bpz", Language::Bpz),
        ("yam", Language::Yam),
        ("kcl", Language::Kcl),
        ("mae", Language::Mae),
        ("lll", Language::Lll),
        ("pch", Language::Pch),
        ("ahh", Language::Ahh),
        ("lok", Language::Lok),
        ("eno", Language::Eno),
        ("jbj", Language::Jbj),
        ("kqo", Language::Kqo),
        ("lkb", Language::Lkb),
        ("anu", Language::Anu),
        ("lim", Language::Lim),
        ("ofo", Language::Ofo),
        ("cnt", Language::Cnt),
        ("vay", Language::Vay),
        ("ebu", Language::Ebu),
        ("acf", Language::Acf),
        ("tez", Language::Tez),
        ("bki", Language::Bki),
        ("lqr", Language::Lqr),
        ("kut", Language::Kut),
        ("qxl", Language::Qxl),
        ("kdx", Language::Kdx),
        ("esk", Language::Esk),
        ("aju", Language::Aju),
        ("poc", Language::Poc),
        ("bgg", Language::Bgg),
        ("hap", Language::Hap),
        ("jod", Language::Jod),
        ("pon", Language::Pon),
        ("tog", Language::Tog),
        ("cho", Language::Cho),
        ("qvz", Language::Qvz),
        ("bsc", Language::Bsc),
        ("zhi", Language::Zhi),
        ("ifk", Language::Ifk),
        ("bmt", Language::Bmt),
        ("ojs", Language::Ojs),
        ("maq", Language::Maq),
        ("kel", Language::Kel),
        ("ges", Language::Ges),
        ("frs", Language::Frs),
        ("abo", Language::Abo),
        ("anz", Language::Anz),
        ("mup", Language::Mup),
        ("dwz", Language::Dwz),
        ("hle", Language::Hle),
        ("dbd", Language::Dbd),
        ("bth", Language::Bth),
        ("jkm", Language::Jkm),
        ("hiw", Language::Hiw),
        ("krc", Language::Krc),
        ("tht", Language::Tht),
        ("bxg", Language::Bxg),
        ("tpz", Language::Tpz),
        ("zkh", Language::Zkh),
        ("ndd", Language::Ndd),
        ("nfa", Language::Nfa),
        ("aja", Language::Aja),
        ("wyb", Language::Wyb),
        ("sxe", Language::Sxe),
        ("nie", Language::Nie),
        ("xdk", Language::Xdk),
        ("zkv", Language::Zkv),
        ("lkd", Language::Lkd),
        ("zpg", Language::Zpg),
        ("nsp", Language::Nsp),
        ("boo", Language::Boo),
        ("tth", Language::Tth),
        ("vaa", Language::Vaa),
        ("npl", Language::Npl),
        ("cbg", Language::Cbg),
        ("lsr", Language::Lsr),
        ("pcb", Language::Pcb),
        ("tgl", Language::Tgl),
        ("kgl", Language::Kgl),
        ("lsy", Language::Lsy),
        ("bxa", Language::Bxa),
        ("nsy", Language::Nsy),
        ("bkg", Language::Bkg),
        ("isd", Language::Isd),
        ("nkf", Language::Nkf),
        ("lcf", Language::Lcf),
        ("mgu", Language::Mgu),
        ("mfp", Language::Mfp),
        ("byf", Language::Byf),
        ("tlr", Language::Tlr),
        ("sep", Language::Sep),
        ("iow", Language::Iow),
        ("noj", Language::Noj),
        ("biv", Language::Biv),
        ("gyd", Language::Gyd),
        ("sce", Language::Sce),
        ("thf", Language::Thf),
        ("kaw", Language::Kaw),
        ("bav", Language::Bav),
        ("txe", Language::Txe),
        ("bwz", Language::Bwz),
        ("duc", Language::Duc),
        ("neb", Language::Neb),
        ("lbi", Language::Lbi),
        ("snl", Language::Snl),
        ("kia", Language::Kia),
        ("nrz", Language::Nrz),
        ("blb", Language::Blb),
        ("sja", Language::Sja),
        ("sxg", Language::Sxg),
        ("ctp", Language::Ctp),
        ("cgc", Language::Cgc),
        ("bda", Language::Bda),
        ("ijj", Language::Ijj),
        ("hab", Language::Hab),
        ("arp", Language::Arp),
        ("nzz", Language::Nzz),
        ("uig", Language::Uig),
        ("bit", Language::Bit),
        ("cap", Language::Cap),
        ("diz", Language::Diz),
        ("abz", Language::Abz),
        ("dos", Language::Dos),
        ("anc", Language::Anc),
        ("tud", Language::Tud),
        ("lox", Language::Lox),
        ("guz", Language::Guz),
        ("pwb", Language::Pwb),
        ("wrv", Language::Wrv),
        ("bgr", Language::Bgr),
        ("hoo", Language::Hoo),
        ("phn", Language::Phn),
        ("mgv", Language::Mgv),
        ("swk", Language::Swk),
        ("blw", Language::Blw),
        ("gew", Language::Gew),
        ("ptn", Language::Ptn),
        ("std", Language::Std),
        ("nuv", Language::Nuv),
        ("nmp", Language::Nmp),
        ("bed", Language::Bed),
        ("ogc", Language::Ogc),
        ("mul", Language::Mul),
        ("lcp", Language::Lcp),
        ("lip", Language::Lip),
        ("hps", Language::Hps),
        ("nzm", Language::Nzm),
        ("aaz", Language::Aaz),
        ("gdi", Language::Gdi),
        ("rmp", Language::Rmp),
        ("bcl", Language::Bcl),
        ("bwh", Language::Bwh),
        ("nid", Language::Nid),
        ("tnt", Language::Tnt),
        ("atk", Language::Atk),
        ("htu", Language::Htu),
        ("ppt", Language::Ppt),
        ("pmk", Language::Pmk),
        ("gmd", Language::Gmd),
        ("sbc", Language::Sbc),
        ("kdq", Language::Kdq),
        ("wrx", Language::Wrx),
        ("bwr", Language::Bwr),
        ("yrw", Language::Yrw),
        ("sts", Language::Sts),
        ("fad", Language::Fad),
        ("jnj", Language::Jnj),
        ("rib", Language::Rib),
        ("ysc", Language::Ysc),
        ("yvt", Language::Yvt),
        ("kpd", Language::Kpd),
        ("wnb", Language::Wnb),
        ("mgy", Language::Mgy),
        ("maj", Language::Maj),
        ("teq", Language::Teq),
        ("tbe", Language::Tbe),
        ("ngm", Language::Ngm),
        ("gqi", Language::Gqi),
        ("xgi", Language::Xgi),
        ("auu", Language::Auu),
        ("bcm", Language::Bcm),
        ("lie", Language::Lie),
        ("nar", Language::Nar),
        ("ssl", Language::Ssl),
        ("bnj", Language::Bnj),
        ("hhr", Language::Hhr),
        ("bwi", Language::Bwi),
        ("nlz", Language::Nlz),
        ("bub", Language::Bub),
        ("yai", Language::Yai),
        ("fut", Language::Fut),
        ("dob", Language::Dob),
        ("lbn", Language::Lbn),
        ("kpe", Language::Kpe),
        ("beh", Language::Beh),
        ("bte", Language::Bte),
        ("ppu", Language::Ppu),
        ("jaa", Language::Jaa),
        ("mxg", Language::Mxg),
        ("yuf", Language::Yuf),
        ("bwq", Language::Bwq),
        ("llf", Language::Llf),
        ("twg", Language::Twg),
        ("scv", Language::Scv),
        ("amw", Language::Amw),
        ("djf", Language::Djf),
        ("lae", Language::Lae),
        ("ael", Language::Ael),
        ("knb", Language::Knb),
        ("aso", Language::Aso),
        ("mid", Language::Mid),
        ("wbl", Language::Wbl),
        ("kve", Language::Kve),
        ("bcb", Language::Bcb),
        ("dgi", Language::Dgi),
        ("bbg", Language::Bbg),
        ("kzk", Language::Kzk),
        ("mqt", Language::Mqt),
        ("pej", Language::Pej),
        ("gko", Language::Gko),
        ("siw", Language::Siw),
        ("umn", Language::Umn),
        ("srr", Language::Srr),
        ("xkt", Language::Xkt),
        ("tro", Language::Tro),
        ("etn", Language::Etn),
        ("xpf", Language::Xpf),
        ("sxk", Language::Sxk),
        ("tsq", Language::Tsq),
        ("smp", Language::Smp),
        ("nij", Language::Nij),
        ("atq", Language::Atq),
        ("msa", Language::Msa),
        ("nex", Language::Nex),
        ("mkn", Language::Mkn),
        ("pag", Language::Pag),
        ("gej", Language::Gej),
        ("kpa", Language::Kpa),
        ("lcq", Language::Lcq),
        ("gwd", Language::Gwd),
        ("tdd", Language::Tdd),
        ("xog", Language::Xog),
        ("pbp", Language::Pbp),
        ("blq", Language::Blq),
        ("orh", Language::Orh),
        ("llb", Language::Llb),
        ("ogb", Language::Ogb),
        ("als", Language::Als),
        ("ong", Language::Ong),
        ("pit", Language::Pit),
        ("csw", Language::Csw),
        ("lua", Language::Lua),
        ("txn", Language::Txn),
        ("zma", Language::Zma),
        ("qxo", Language::Qxo),
        ("lth", Language::Lth),
        ("ajw", Language::Ajw),
        ("zap", Language::Zap),
        ("ydd", Language::Ydd),
        ("tax", Language::Tax),
        ("wad", Language::Wad),
        ("mxa", Language::Mxa),
        ("asw", Language::Asw),
        ("nbv", Language::Nbv),
        ("dei", Language::Dei),
        ("jeu", Language::Jeu),
        ("bdv", Language::Bdv),
        ("kyl", Language::Kyl),
        ("kqc", Language::Kqc),
        ("sfb", Language::Sfb),
        ("tnb", Language::Tnb),
        ("nom", Language::Nom),
        ("bnc", Language::Bnc),
        ("wgg", Language::Wgg),
        ("tev", Language::Tev),
        ("har", Language::Har),
        ("lgi", Language::Lgi),
        ("jbk", Language::Jbk),
        ("bfp", Language::Bfp),
        ("tuz", Language::Tuz),
        ("lpa", Language::Lpa),
        ("mws", Language::Mws),
        ("pux", Language::Pux),
        ("dnn", Language::Dnn),
        ("izz", Language::Izz),
        ("tgc", Language::Tgc),
        ("seh", Language::Seh),
        ("olr", Language::Olr),
        ("mlo", Language::Mlo),
        ("law", Language::Law),
        ("fuu", Language::Fuu),
        ("ret", Language::Ret),
        ("nev", Language::Nev),
        ("vbb", Language::Vbb),
        ("ykt", Language::Ykt),
        ("zpe", Language::Zpe),
        ("xru", Language::Xru),
        ("guu", Language::Guu),
        ("nui", Language::Nui),
        ("sgi", Language::Sgi),
        ("frt", Language::Frt),
        ("mzv", Language::Mzv),
        ("ukw", Language::Ukw),
        ("oor", Language::Oor),
        ("udu", Language::Udu),
        ("gyy", Language::Gyy),
        ("jos", Language::Jos),
        ("xpm", Language::Xpm),
        ("giq", Language::Giq),
        ("pli", Language::Pli),
        ("ole", Language::Ole),
        ("tpl", Language::Tpl),
        ("nba", Language::Nba),
        ("van", Language::Van),
        ("vut", Language::Vut),
        ("piw", Language::Piw),
        ("sok", Language::Sok),
        ("mdw", Language::Mdw),
        ("liz", Language::Liz),
        ("gux", Language::Gux),
        ("bho", Language::Bho),
        ("gmy", Language::Gmy),
        ("yxl", Language::Yxl),
        ("lmf", Language::Lmf),
        ("skm", Language::Skm),
        ("mba", Language::Mba),
        ("vnk", Language::Vnk),
        ("ikt", Language::Ikt),
        ("ina", Language::Ina),
        ("ngj", Language::Ngj),
        ("yij", Language::Yij),
        ("lna", Language::Lna),
        ("pww", Language::Pww),
        ("ten", Language::Ten),
        ("wne", Language::Wne),
        ("shh", Language::Shh),
        ("azm", Language::Azm),
        ("sko", Language::Sko),
        ("bec", Language::Bec),
        ("juc", Language::Juc),
        ("kqu", Language::Kqu),
        ("ksr", Language::Ksr),
        ("llk", Language::Llk),
        ("sos", Language::Sos),
        ("kjh", Language::Kjh),
        ("huq", Language::Huq),
        ("ngg", Language::Ngg),
        ("kum", Language::Kum),
        ("yix", Language::Yix),
        ("uln", Language::Uln),
        ("kxn", Language::Kxn),
        ("git", Language::Git),
        ("xud", Language::Xud),
        ("poi", Language::Poi),
        ("dub", Language::Dub),
        ("kiq", Language::Kiq),
        ("mdm", Language::Mdm),
        ("nbw", Language::Nbw),
        ("soy", Language::Soy),
        ("yiq", Language::Yiq),
        ("ett", Language::Ett),
        ("ktd", Language::Ktd),
        ("cey", Language::Cey),
        ("agn", Language::Agn),
        ("lis", Language::Lis),
        ("ojp", Language::Ojp),
        ("ukk", Language::Ukk),
        ("mkv", Language::Mkv),
        ("iki", Language::Iki),
        ("kfu", Language::Kfu),
        ("ruc", Language::Ruc),
        ("nbn", Language::Nbn),
        ("bto", Language::Bto),
        ("mjw", Language::Mjw),
        ("mpb", Language::Mpb),
        ("kkm", Language::Kkm),
        ("zpn", Language::Zpn),
        ("coe", Language::Coe),
        ("xam", Language::Xam),
        ("eng", Language::Eng),
        ("odt", Language::Odt),
        ("koc", Language::Koc),
        ("mpv", Language::Mpv),
        ("tgx", Language::Tgx),
        ("bjt", Language::Bjt),
        ("yso", Language::Yso),
        ("pkh", Language::Pkh),
        ("tue", Language::Tue),
        ("swr", Language::Swr),
        ("kcz", Language::Kcz),
        ("kej", Language::Kej),
        ("ylb", Language::Ylb),
        ("uzs", Language::Uzs),
        ("mpd", Language::Mpd),
        ("bpd", Language::Bpd),
        ("ils", Language::Ils),
        ("vas", Language::Vas),
        ("dga", Language::Dga),
        ("rmv", Language::Rmv),
        ("bhf", Language::Bhf),
        ("jih", Language::Jih),
        ("rwr", Language::Rwr),
        ("dot", Language::Dot),
        ("nhe", Language::Nhe),
        ("hoa", Language::Hoa),
        ("kus", Language::Kus),
        ("tzj", Language::Tzj),
        ("sir", Language::Sir),
        ("iba", Language::Iba),
        ("atl", Language::Atl),
        ("dzn", Language::Dzn),
        ("tka", Language::Tka),
        ("yrs", Language::Yrs),
        ("xce", Language::Xce),
        ("ige", Language::Ige),
        ("bhg", Language::Bhg),
        ("gbd", Language::Gbd),
        ("mel", Language::Mel),
        ("afi", Language::Afi),
        ("hmj", Language::Hmj),
        ("enm", Language::Enm),
        ("shw", Language::Shw),
        ("pcm", Language::Pcm),
        ("roc", Language::Roc),
        ("tre", Language::Tre),
        ("gel", Language::Gel),
        ("yml", Language::Yml),
        ("axe", Language::Axe),
        ("swl", Language::Swl),
        ("tol", Language::Tol),
        ("sdn", Language::Sdn),
        ("psm", Language::Psm),
        ("adj", Language::Adj),
        ("lsv", Language::Lsv),
        ("nxd", Language::Nxd),
        ("mcv", Language::Mcv),
        ("aeb", Language::Aeb),
        ("hkn", Language::Hkn),
        ("grb", Language::Grb),
        ("pka", Language::Pka),
        ("mdc", Language::Mdc),
        ("syk", Language::Syk),
        ("hpo", Language::Hpo),
        ("zpy", Language::Zpy),
        ("atz", Language::Atz),
        ("hrc", Language::Hrc),
        ("mcf", Language::Mcf),
        ("psh", Language::Psh),
        ("elo", Language::Elo),
        ("kow", Language::Kow),
        ("mqx", Language::Mqx),
        ("agf", Language::Agf),
        ("big", Language::Big),
        ("tyn", Language::Tyn),
        ("had", Language::Had),
        ("rsm", Language::Rsm),
        ("ghe", Language::Ghe),
        ("muo", Language::Muo),
        ("mqg", Language::Mqg),
        ("pfe", Language::Pfe),
        ("snw", Language::Snw),
        ("qwa", Language::Qwa),
        ("zpb", Language::Zpb),
        ("erk", Language::Erk),
        ("qvi", Language::Qvi),
        ("bzy", Language::Bzy),
        ("pye", Language::Pye),
        ("kyi", Language::Kyi),
        ("bbc", Language::Bbc),
        ("tbn", Language::Tbn),
        ("aos", Language::Aos),
        ("yxa", Language::Yxa),
        ("bxo", Language::Bxo),
        ("clm", Language::Clm),
        ("koq", Language::Koq),
        ("mzy", Language::Mzy),
        ("zpd", Language::Zpd),
        ("byc", Language::Byc),
        ("nuc", Language::Nuc),
        ("blo", Language::Blo),
        ("tfr", Language::Tfr),
        ("nlw", Language::Nlw),
        ("zgm", Language::Zgm),
        ("xae", Language::Xae),
        ("mal", Language::Mal),
        ("tkm", Language::Tkm),
        ("yta", Language::Yta),
        ("tir", Language::Tir),
        ("tao", Language::Tao),
        ("bmk", Language::Bmk),
        ("lot", Language::Lot),
        ("ttl", Language::Ttl),
        ("tkl", Language::Tkl),
        ("jcs", Language::Jcs),
        ("buy", Language::Buy),
        ("tsc", Language::Tsc),
        ("ham", Language::Ham),
        ("ney", Language::Ney),
        ("pdt", Language::Pdt),
        ("dgx", Language::Dgx),
        ("tul", Language::Tul),
        ("axx", Language::Axx),
        ("bce", Language::Bce),
        ("lpe", Language::Lpe),
        ("gok", Language::Gok),
        ("ibm", Language::Ibm),
        ("kps", Language::Kps),
        ("mnf", Language::Mnf),
        ("tpi", Language::Tpi),
        ("mhc", Language::Mhc),
        ("xkv", Language::Xkv),
        ("klb", Language::Klb),
        ("ruz", Language::Ruz),
        ("sxs", Language::Sxs),
        ("crs", Language::Crs),
        ("ahg", Language::Ahg),
        ("idt", Language::Idt),
        ("min", Language::Min),
        ("dkk", Language::Dkk),
        ("noh", Language::Noh),
        ("nxe", Language::Nxe),
        ("ukl", Language::Ukl),
        ("auz", Language::Auz),
        ("wbi", Language::Wbi),
        ("suj", Language::Suj),
        ("dor", Language::Dor),
        ("wca", Language::Wca),
        ("cnu", Language::Cnu),
        ("tkp", Language::Tkp),
        ("mzk", Language::Mzk),
        ("gyo", Language::Gyo),
        ("kxm", Language::Kxm),
        ("bpu", Language::Bpu),
        ("mcp", Language::Mcp),
        ("ydk", Language::Ydk),
        ("sso", Language::Sso),
        ("txq", Language::Txq),
        ("ymd", Language::Ymd),
        ("mmn", Language::Mmn),
        ("dty", Language::Dty),
        ("ygm", Language::Ygm),
        ("bvn", Language::Bvn),
        ("ilg", Language::Ilg),
        ("aou", Language::Aou),
        ("wag", Language::Wag),
        ("sml", Language::Sml),
        ("ndt", Language::Ndt),
        ("auj", Language::Auj),
        ("lea", Language::Lea),
        ("bvp", Language::Bvp),
        ("kqf", Language::Kqf),
        ("xcn", Language::Xcn),
        ("mbu", Language::Mbu),
        ("dup", Language::Dup),
        ("lbx", Language::Lbx),
        ("kdl", Language::Kdl),
        ("mvq", Language::Mvq),
        ("ncb", Language::Ncb),
        ("oks", Language::Oks),
        ("emg", Language::Emg),
        ("cgg", Language::Cgg),
        ("jbt", Language::Jbt),
        ("gix", Language::Gix),
        ("tnm", Language::Tnm),
        ("lef", Language::Lef),
        ("tbk", Language::Tbk),
        ("omb", Language::Omb),
        ("mlm", Language::Mlm),
        ("xxt", Language::Xxt),
        ("mpu", Language::Mpu),
        ("cse", Language::Cse),
        ("nhm", Language::Nhm),
        ("drg", Language::Drg),
        ("ccc", Language::Ccc),
        ("esq", Language::Esq),
        ("xib", Language::Xib),
        ("nkw", Language::Nkw),
        ("sxu", Language::Sxu),
        ("rml", Language::Rml),
        ("gno", Language::Gno),
        ("mqe", Language::Mqe),
        ("ybn", Language::Ybn),
        ("scu", Language::Scu),
        ("xrb", Language::Xrb),
        ("api", Language::Api),
        ("mtu", Language::Mtu),
        ("lav", Language::Lav),
        ("aek", Language::Aek),
        ("mcw", Language::Mcw),
        ("otk", Language::Otk),
        ("rgu", Language::Rgu),
        ("zen", Language::Zen),
        ("rkm", Language::Rkm),
        ("bix", Language::Bix),
        ("krd", Language::Krd),
        ("xmc", Language::Xmc),
        ("alu", Language::Alu),
        ("beg", Language::Beg),
        ("ssb", Language::Ssb),
        ("ntw", Language::Ntw),
        ("ypa", Language::Ypa),
        ("tiv", Language::Tiv),
        ("dbi", Language::Dbi),
        ("thu", Language::Thu),
        ("bqg", Language::Bqg),
        ("hnn", Language::Hnn),
        ("cch", Language::Cch),
        ("utu", Language::Utu),
        ("jrr", Language::Jrr),
        ("nea", Language::Nea),
        ("xml", Language::Xml),
        ("zay", Language::Zay),
        ("mcz", Language::Mcz),
        ("jau", Language::Jau),
        ("fir", Language::Fir),
        ("nix", Language::Nix),
        ("iso", Language::Iso),
        ("yko", Language::Yko),
        ("nud", Language::Nud),
        ("acd", Language::Acd),
        ("ayt", Language::Ayt),
        ("ldn", Language::Ldn),
        ("tme", Language::Tme),
        ("sik", Language::Sik),
        ("crj", Language::Crj),
        ("dsb", Language::Dsb),
        ("gbb", Language::Gbb),
        ("roh", Language::Roh),
        ("usk", Language::Usk),
        ("kkl", Language::Kkl),
        ("ned", Language::Ned),
        ("ncg", Language::Ncg),
        ("biy", Language::Biy),
        ("kto", Language::Kto),
        ("ilm", Language::Ilm),
        ("mvz", Language::Mvz),
        ("izh", Language::Izh),
        ("xsh", Language::Xsh),
        ("wbt", Language::Wbt),
        ("xcl", Language::Xcl),
        ("vor", Language::Vor),
        ("qxu", Language::Qxu),
        ("tsj", Language::Tsj),
        ("gwb", Language::Gwb),
        ("vkn", Language::Vkn),
        ("dbu", Language::Dbu),
        ("sgd", Language::Sgd),
        ("csn", Language::Csn),
        ("mxs", Language::Mxs),
        ("xsj", Language::Xsj),
        ("nmo", Language::Nmo),
        ("ccg", Language::Ccg),
        ("uth", Language::Uth),
        ("lmb", Language::Lmb),
        ("tts", Language::Tts),
        ("dus", Language::Dus),
        ("xjb", Language::Xjb),
        ("cbv", Language::Cbv),
        ("pev", Language::Pev),
        ("wka", Language::Wka),
        ("mdf", Language::Mdf),
        ("bqh", Language::Bqh),
        ("buv", Language::Buv),
        ("kiz", Language::Kiz),
        ("lra", Language::Lra),
        ("wkw", Language::Wkw),
        ("neq", Language::Neq),
        ("fah", Language::Fah),
        ("lko", Language::Lko),
        ("koy", Language::Koy),
        ("mrg", Language::Mrg),
        ("nho", Language::Nho),
        ("krv", Language::Krv),
        ("hmu", Language::Hmu),
        ("mcu", Language::Mcu),
        ("orw", Language::Orw),
        ("nwr", Language::Nwr),
        ("mnw", Language::Mnw),
        ("hks", Language::Hks),
        ("hra", Language::Hra),
        ("abc", Language::Abc),
        ("jya", Language::Jya),
        ("bhz", Language::Bhz),
        ("ecr", Language::Ecr),
        ("suw", Language::Suw),
        ("yyr", Language::Yyr),
        ("awn", Language::Awn),
        ("hat", Language::Hat),
        ("dda", Language::Dda),
        ("bqo", Language::Bqo),
        ("kci", Language::Kci),
        ("loq", Language::Loq),
        ("mre", Language::Mre),
        ("mha", Language::Mha),
        ("kmo", Language::Kmo),
        ("tlv", Language::Tlv),
        ("bme", Language::Bme),
        ("usp", Language::Usp),
        ("ahp", Language::Ahp),
        ("jaz", Language::Jaz),
        ("kfk", Language::Kfk),
        ("mum", Language::Mum),
        ("kuo", Language::Kuo),
        ("ntr", Language::Ntr),
        ("mpc", Language::Mpc),
        ("mie", Language::Mie),
        ("bff", Language::Bff),
        ("gpn", Language::Gpn),
        ("wow", Language::Wow),
        ("kfi", Language::Kfi),
        ("xmv", Language::Xmv),
        ("mui", Language::Mui),
        ("isl", Language::Isl),
        ("miu", Language::Miu),
        ("hlu", Language::Hlu),
        ("int", Language::Int),
        ("atc", Language::Atc),
        ("nej", Language::Nej),
        ("bud", Language::Bud),
        ("nyt", Language::Nyt),
        ("gih", Language::Gih),
        ("ykm", Language::Ykm),
        ("qup", Language::Qup),
        ("ynl", Language::Ynl),
        ("kvm", Language::Kvm),
        ("mzn", Language::Mzn),
        ("wob", Language::Wob),
        ("ksx", Language::Ksx),
        ("seo", Language::Seo),
        ("kwi", Language::Kwi),
        ("nji", Language::Nji),
        ("avi", Language::Avi),
        ("nch", Language::Nch),
        ("nnc", Language::Nnc),
        ("ylr", Language::Ylr),
        ("pbn", Language::Pbn),
        ("bpn", Language::Bpn),
        ("bkx", Language::Bkx),
        ("ati", Language::Ati),
        ("onb", Language::Onb),
        ("sie", Language::Sie),
        ("zat", Language::Zat),
        ("kxy", Language::Kxy),
        ("uha", Language::Uha),
        ("zbw", Language::Zbw),
        ("kmu", Language::Kmu),
        ("kte", Language::Kte),
        ("bro", Language::Bro),
        ("sgg", Language::Sgg),
        ("bos", Language::Bos),
        ("pne", Language::Pne),
        ("btr", Language::Btr),
        ("scp", Language::Scp),
        ("khf", Language::Khf),
        ("mhm", Language::Mhm),
        ("awg", Language::Awg),
        ("der", Language::Der),
        ("gyr", Language::Gyr),
        ("umg", Language::Umg),
        ("mbf", Language::Mbf),
        ("lhm", Language::Lhm),
        ("asp", Language::Asp),
        ("gan", Language::Gan),
        ("dwk", Language::Dwk),
        ("nqk", Language::Nqk),
        ("cna", Language::Cna),
        ("kao", Language::Kao),
        ("dnv", Language::Dnv),
        ("ilo", Language::Ilo),
        ("akd", Language::Akd),
        ("kxr", Language::Kxr),
        ("ohu", Language::Ohu),
        ("sui", Language::Sui),
        ("ist", Language::Ist),
        ("lbg", Language::Lbg),
        ("mzd", Language::Mzd),
        ("lgn", Language::Lgn),
        ("thm", Language::Thm),
        ("mee", Language::Mee),
        ("lek", Language::Lek),
        ("wes", Language::Wes),
        ("mlx", Language::Mlx),
        ("wkl", Language::Wkl),
        ("nme", Language::Nme),
        ("azb", Language::Azb),
        ("sze", Language::Sze),
        ("rol", Language::Rol),
        ("sxl", Language::Sxl),
        ("srn", Language::Srn),
        ("tep", Language::Tep),
        ("byn", Language::Byn),
        ("cps", Language::Cps),
        ("lao", Language::Lao),
        ("noa", Language::Noa),
        ("tct", Language::Tct),
        ("tlf", Language::Tlf),
        ("bsf", Language::Bsf),
        ("cjy", Language::Cjy),
        ("skn", Language::Skn),
        ("tjn", Language::Tjn),
        ("wan", Language::Wan),
        ("wlu", Language::Wlu),
        ("kfn", Language::Kfn),
        ("njt", Language::Njt),
        ("tcp", Language::Tcp),
        ("gbn", Language::Gbn),
        ("tkb", Language::Tkb),
        ("jak", Language::Jak),
        ("mdg", Language::Mdg),
        ("bha", Language::Bha),
        ("tso", Language::Tso),
        ("nus", Language::Nus),
        ("ulc", Language::Ulc),
        ("uri", Language::Uri),
        ("nrb", Language::Nrb),
        ("llc", Language::Llc),
        ("urw", Language::Urw),
        ("xul", Language::Xul),
        ("mdb", Language::Mdb),
        ("mec", Language::Mec),
        ("kvl", Language::Kvl),
        ("pbg", Language::Pbg),
        ("gla", Language::Gla),
        ("slc", Language::Slc),
        ("lks", Language::Lks),
        ("rkw", Language::Rkw),
        ("kxj", Language::Kxj),
        ("mgm", Language::Mgm),
        ("kai", Language::Kai),
        ("tin", Language::Tin),
        ("nyo", Language::Nyo),
        ("mir", Language::Mir),
        ("fmu", Language::Fmu),
        ("vkz", Language::Vkz),
        ("ihi", Language::Ihi),
        ("kzm", Language::Kzm),
        ("stq", Language::Stq),
        ("lmc", Language::Lmc),
        ("huh", Language::Huh),
        ("ygs", Language::Ygs),
        ("kfb", Language::Kfb),
        ("hol", Language::Hol),
        ("xby", Language::Xby),
        ("gnm", Language::Gnm),
        ("tbs", Language::Tbs),
        ("tdk", Language::Tdk),
        ("anj", Language::Anj),
        ("xbr", Language::Xbr),
        ("ssr", Language::Ssr),
        ("eka", Language::Eka),
        ("twf", Language::Twf),
        ("ner", Language::Ner),
        ("nuh", Language::Nuh),
        ("gve", Language::Gve),
        ("sbm", Language::Sbm),
        ("zaz", Language::Zaz),
        ("puq", Language::Puq),
        ("thq", Language::Thq),
        ("aij", Language::Aij),
        ("nxr", Language::Nxr),
        ("dlm", Language::Dlm),
        ("rmg", Language::Rmg),
        ("spk", Language::Spk),
        ("cfa", Language::Cfa),
        ("dbr", Language::Dbr),
        ("atn", Language::Atn),
        ("sly", Language::Sly),
        ("gwx", Language::Gwx),
        ("pgl", Language::Pgl),
        ("yea", Language::Yea),
        ("tsm", Language::Tsm),
        ("ims", Language::Ims),
        ("wij", Language::Wij),
        ("bhn", Language::Bhn),
        ("hng", Language::Hng),
        ("mfn", Language::Mfn),
        ("tva", Language::Tva),
        ("kzy", Language::Kzy),
        ("kfx", Language::Kfx),
        ("zsk", Language::Zsk),
        ("cbt", Language::Cbt),
        ("net", Language::Net),
        ("xom", Language::Xom),
        ("eme", Language::Eme),
        ("dae", Language::Dae),
        ("abq", Language::Abq),
        ("abi", Language::Abi),
        ("iwo", Language::Iwo),
        ("ppp", Language::Ppp),
        ("gax", Language::Gax),
        ("rop", Language::Rop),
        ("sdf", Language::Sdf),
        ("lmv", Language::Lmv),
        ("xpe", Language::Xpe),
        ("jpr", Language::Jpr),
        ("sbi", Language::Sbi),
        ("psg", Language::Psg),
        ("gil", Language::Gil),
        ("gnq", Language::Gnq),
        ("lns", Language::Lns),
        ("pnl", Language::Pnl),
        ("hor", Language::Hor),
        ("soz", Language::Soz),
        ("ser", Language::Ser),
        ("tby", Language::Tby),
        ("bbu", Language::Bbu),
        ("xum", Language::Xum),
        ("zkb", Language::Zkb),
        ("ozm", Language::Ozm),
        ("guh", Language::Guh),
        ("tov", Language::Tov),
        ("ian", Language::Ian),
        ("xrw", Language::Xrw),
        ("ldd", Language::Ldd),
        ("tlm", Language::Tlm),
        ("dln", Language::Dln),
        ("mle", Language::Mle),
        ("ksc", Language::Ksc),
        ("rnr", Language::Rnr),
        ("bre", Language::Bre),
        ("kya", Language::Kya),
        ("siq", Language::Siq),
        ("xkp", Language::Xkp),
        ("obl", Language::Obl),
        ("idu", Language::Idu),
        ("yda", Language::Yda),
        ("knd", Language::Knd),
        ("plj", Language::Plj),
        ("kpk", Language::Kpk),
        ("gua", Language::Gua),
        ("goq", Language::Goq),
        ("cma", Language::Cma),
        ("jee", Language::Jee),
        ("sjt", Language::Sjt),
        ("mmk", Language::Mmk),
        ("scx", Language::Scx),
        ("jia", Language::Jia),
        ("wli", Language::Wli),
        ("tuj", Language::Tuj),
        ("ayq", Language::Ayq),
        ("rme", Language::Rme),
        ("rxd", Language::Rxd),
        ("zoh", Language::Zoh),
        ("slp", Language::Slp),
        ("bfe", Language::Bfe),
        ("gnj", Language::Gnj),
        ("bkp", Language::Bkp),
        ("mem", Language::Mem),
        ("ayc", Language::Ayc),
        ("bdf", Language::Bdf),
        ("lbw", Language::Lbw),
        ("pip", Language::Pip),
        ("avk", Language::Avk),
        ("yud", Language::Yud),
        ("thz", Language::Thz),
        ("duq", Language::Duq),
        ("adw", Language::Adw),
        ("cmr", Language::Cmr),
        ("thi", Language::Thi),
        ("dtk", Language::Dtk),
        ("pnk", Language::Pnk),
        ("gnu", Language::Gnu),
        ("nsm", Language::Nsm),
        ("lrm", Language::Lrm),
        ("gna", Language::Gna),
        ("kuf", Language::Kuf),
        ("weu", Language::Weu),
        ("ctd", Language::Ctd),
        ("gmm", Language::Gmm),
        ("nyf", Language::Nyf),
        ("mhp", Language::Mhp),
        ("waq", Language::Waq),
        ("sev", Language::Sev),
        ("mlg", Language::Mlg),
        ("nod", Language::Nod),
        ("enw", Language::Enw),
        ("oty", Language::Oty),
        ("dti", Language::Dti),
        ("lse", Language::Lse),
        ("ndf", Language::Ndf),
        ("jru", Language::Jru),
        ("urr", Language::Urr),
        ("huu", Language::Huu),
        ("gge", Language::Gge),
        ("pea", Language::Pea),
        ("moc", Language::Moc),
        ("kwt", Language::Kwt),
        ("cbk", Language::Cbk),
        ("zml", Language::Zml),
        ("sha", Language::Sha),
        ("aan", Language::Aan),
        ("okb", Language::Okb),
        ("tpf", Language::Tpf),
        ("bqp", Language::Bqp),
        ("awu", Language::Awu),
        ("tvm", Language::Tvm),
        ("sih", Language::Sih),
        ("kgb", Language::Kgb),
        ("bzq", Language::Bzq),
        ("npy", Language::Npy),
        ("mud", Language::Mud),
        ("mlt", Language::Mlt),
        ("rkh", Language::Rkh),
        ("azn", Language::Azn),
        ("ayd", Language::Ayd),
        ("gat", Language::Gat),
        ("pby", Language::Pby),
        ("omn", Language::Omn),
        ("apr", Language::Apr),
        ("mfb", Language::Mfb),
        ("ldm", Language::Ldm),
        ("xkf", Language::Xkf),
        ("mrd", Language::Mrd),
        ("nhp", Language::Nhp),
        ("tgb", Language::Tgb),
        ("rog", Language::Rog),
        ("xdo", Language::Xdo),
        ("wln", Language::Wln),
        ("lzh", Language::Lzh),
        ("kyu", Language::Kyu),
        ("nbm", Language::Nbm),
        ("uhn", Language::Uhn),
        ("mvo", Language::Mvo),
        ("oki", Language::Oki),
        ("yuj", Language::Yuj),
        ("sgr", Language::Sgr),
        ("itt", Language::Itt),
        ("crx", Language::Crx),
        ("yub", Language::Yub),
        ("kfz", Language::Kfz),
        ("gip", Language::Gip),
        ("jig", Language::Jig),
        ("bbo", Language::Bbo),
        ("lnd", Language::Lnd),
        ("shv", Language::Shv),
        ("kad", Language::Kad),
        ("bnq", Language::Bnq),
        ("ags", Language::Ags),
        ("dah", Language::Dah),
        ("olu", Language::Olu),
        ("zmq", Language::Zmq),
        ("pir", Language::Pir),
        ("krh", Language::Krh),
        ("bop", Language::Bop),
        ("bif", Language::Bif),
        ("xoc", Language::Xoc),
        ("ccj", Language::Ccj),
        ("kvr", Language::Kvr),
        ("kws", Language::Kws),
        ("sqh", Language::Sqh),
        ("bfz", Language::Bfz),
        ("loj", Language::Loj),
        ("tkt", Language::Tkt),
        ("yei", Language::Yei),
        ("dtr", Language::Dtr),
        ("imy", Language::Imy),
        ("isn", Language::Isn),
        ("cen", Language::Cen),
        ("ncu", Language::Ncu),
        ("wkd", Language::Wkd),
        ("zun", Language::Zun),
        ("mfl", Language::Mfl),
        ("iml", Language::Iml),
        ("cuu", Language::Cuu),
        ("cmm", Language::Cmm),
        ("brj", Language::Brj),
        ("wsu", Language::Wsu),
        ("rim", Language::Rim),
        ("eko", Language::Eko),
        ("ndc", Language::Ndc),
        ("kez", Language::Kez),
        ("ext", Language::Ext),
        ("don", Language::Don),
        ("anx", Language::Anx),
        ("gmv", Language::Gmv),
        ("err", Language::Err),
        ("thh", Language::Thh),
        ("kxh", Language::Kxh),
        ("abl", Language::Abl),
        ("spu", Language::Spu),
        ("stj", Language::Stj),
        ("tru", Language::Tru),
        ("nnr", Language::Nnr),
        ("xgm", Language::Xgm),
        ("jpn", Language::Jpn),
        ("mwq", Language::Mwq),
        ("amr", Language::Amr),
        ("sqs", Language::Sqs),
        ("nbo", Language::Nbo),
        ("utp", Language::Utp),
        ("swx", Language::Swx),
        ("xmu", Language::Xmu),
        ("swf", Language::Swf),
        ("bon", Language::Bon),
        ("hem", Language::Hem),
        ("mwp", Language::Mwp),
        ("gkp", Language::Gkp),
        ("hbo", Language::Hbo),
        ("goi", Language::Goi),
        ("ues", Language::Ues),
        ("qum", Language::Qum),
        ("kac", Language::Kac),
        ("mwo", Language::Mwo),
        ("aiq", Language::Aiq),
        ("ikw", Language::Ikw),
        ("amg", Language::Amg),
        ("bba", Language::Bba),
        ("myh", Language::Myh),
        ("hsh", Language::Hsh),
        ("xps", Language::Xps),
        ("biw", Language::Biw),
        ("nir", Language::Nir),
        ("ess", Language::Ess),
        ("lun", Language::Lun),
        ("xka", Language::Xka),
        ("kxw", Language::Kxw),
        ("nef", Language::Nef),
        ("svk", Language::Svk),
        ("bnx", Language::Bnx),
        ("nkp", Language::Nkp),
        ("lbm", Language::Lbm),
        ("fwa", Language::Fwa),
        ("hbn", Language::Hbn),
        ("ihb", Language::Ihb),
        ("xwc", Language::Xwc),
        ("wxw", Language::Wxw),
        ("ahl", Language::Ahl),
        ("mcj", Language::Mcj),
        ("kzx", Language::Kzx),
        ("yro", Language::Yro),
        ("dso", Language::Dso),
        ("dyo", Language::Dyo),
        ("vot", Language::Vot),
        ("knv", Language::Knv),
        ("nxa", Language::Nxa),
        ("bod", Language::Bod),
        ("trs", Language::Trs),
        ("ffi", Language::Ffi),
        ("sks", Language::Sks),
        ("phd", Language::Phd),
        ("auk", Language::Auk),
        ("pis", Language::Pis),
        ("yir", Language::Yir),
        ("alx", Language::Alx),
        ("kxi", Language::Kxi),
        ("klo", Language::Klo),
        ("mvn", Language::Mvn),
        ("mdp", Language::Mdp),
        ("txa", Language::Txa),
        ("yit", Language::Yit),
        ("atg", Language::Atg),
        ("bdi", Language::Bdi),
        ("xkz", Language::Xkz),
        ("add", Language::Add),
        ("ktp", Language::Ktp),
        ("ymo", Language::Ymo),
        ("deq", Language::Deq),
        ("mhg", Language::Mhg),
        ("les", Language::Les),
        ("bxd", Language::Bxd),
        ("gbs", Language::Gbs),
        ("wmm", Language::Wmm),
        ("gss", Language::Gss),
        ("fnb", Language::Fnb),
        ("xhu", Language::Xhu),
        ("kmx", Language::Kmx),
        ("zsm", Language::Zsm),
        ("sjm", Language::Sjm),
        ("ysl", Language::Ysl),
        ("bok", Language::Bok),
        ("whu", Language::Whu),
        ("nwg", Language::Nwg),
        ("khv", Language::Khv),
        ("tmi", Language::Tmi),
        ("wdt", Language::Wdt),
        ("aal", Language::Aal),
        ("mzh", Language::Mzh),
        ("twb", Language::Twb),
        ("bdx", Language::Bdx),
        ("bpq", Language::Bpq),
        ("qud", Language::Qud),
        ("cyb", Language::Cyb),
        ("rtc", Language::Rtc),
        ("psw", Language::Psw),
        ("svb", Language::Svb),
        ("hun", Language::Hun),
        ("sod", Language::Sod),
        ("ash", Language::Ash),
        ("pij", Language::Pij),
        ("gri", Language::Gri),
        ("nmx", Language::Nmx),
        ("mdt", Language::Mdt),
        ("elx", Language::Elx),
        ("kzp", Language::Kzp),
        ("vmk", Language::Vmk),
        ("sdt", Language::Sdt),
        ("dts", Language::Dts),
        ("aax", Language::Aax),
        ("lkj", Language::Lkj),
        ("rbl", Language::Rbl),
        ("niq", Language::Niq),
        ("etb", Language::Etb),
        ("asl", Language::Asl),
        ("dgw", Language::Dgw),
        ("kck", Language::Kck),
        ("gly", Language::Gly),
        ("glv", Language::Glv),
        ("hbb", Language::Hbb),
        ("sov", Language::Sov),
        ("xmo", Language::Xmo),
        ("ngl", Language::Ngl),
        ("twx", Language::Twx),
        ("kui", Language::Kui),
        ("bck", Language::Bck),
        ("mkw", Language::Mkw),
        ("ygr", Language::Ygr),
        ("dil", Language::Dil),
        ("swp", Language::Swp),
        ("kav", Language::Kav),
        ("yki", Language::Yki),
        ("gzi", Language::Gzi),
        ("bbi", Language::Bbi),
        ("kxz", Language::Kxz),
        ("gnc", Language::Gnc),
        ("eee", Language::Eee),
        ("ayi", Language::Ayi),
        ("wau", Language::Wau),
        ("men", Language::Men),
        ("hno", Language::Hno),
        ("cuh", Language::Cuh),
        ("tpn", Language::Tpn),
        ("fij", Language::Fij),
        ("okx", Language::Okx),
        ("teu", Language::Teu),
        ("itw", Language::Itw),
        ("pap", Language::Pap),
        ("oaa", Language::Oaa),
        ("okv", Language::Okv),
        ("kpg", Language::Kpg),
        ("ppi", Language::Ppi),
        ("slq", Language::Slq),
        ("bsy", Language::Bsy),
        ("ise", Language::Ise),
        ("tak", Language::Tak),
        ("ktl", Language::Ktl),
        ("nkg", Language::Nkg),
        ("ago", Language::Ago),
        ("jna", Language::Jna),
        ("gak", Language::Gak),
        ("dgd", Language::Dgd),
        ("sei", Language::Sei),
        ("wrh", Language::Wrh),
        ("try", Language::Try),
        ("lib", Language::Lib),
        ("xxm", Language::Xxm),
        ("tas", Language::Tas),
        ("wmx", Language::Wmx),
        ("vep", Language::Vep),
        ("hur", Language::Hur),
        ("wrw", Language::Wrw),
        ("knm", Language::Knm),
        ("tce", Language::Tce),
        ("mqp", Language::Mqp),
        ("kjj", Language::Kjj),
        ("uma", Language::Uma),
        ("zpm", Language::Zpm),
        ("kur", Language::Kur),
        ("wsa", Language::Wsa),
        ("kzd", Language::Kzd),
        ("sau", Language::Sau),
        ("shk", Language::Shk),
        ("twa", Language::Twa),
        ("luw", Language::Luw),
        ("kwj", Language::Kwj),
        ("rah", Language::Rah),
        ("gnb", Language::Gnb),
        ("akl", Language::Akl),
        ("bfc", Language::Bfc),
        ("ccd", Language::Ccd),
        ("aid", Language::Aid),
        ("gds", Language::Gds),
        ("daz", Language::Daz),
        ("akp", Language::Akp),
        ("zbl", Language::Zbl),
        ("ful", Language::Ful),
        ("kdc", Language::Kdc),
        ("bab", Language::Bab),
        ("apm", Language::Apm),
        ("akt", Language::Akt),
        ("buo", Language::Buo),
        ("kni", Language::Kni),
        ("mih", Language::Mih),
        ("gxx", Language::Gxx),
        ("gye", Language::Gye),
        ("sur", Language::Sur),
        ("tiq", Language::Tiq),
        ("ypn", Language::Ypn),
        ("xmd", Language::Xmd),
        ("aht", Language::Aht),
        ("enn", Language::Enn),
        ("abt", Language::Abt),
        ("mlu", Language::Mlu),
        ("mlz", Language::Mlz),
        ("lem", Language::Lem),
        ("aau", Language::Aau),
        ("ysn", Language::Ysn),
        ("twr", Language::Twr),
        ("yaw", Language::Yaw),
        ("tub", Language::Tub),
        ("bch", Language::Bch),
        ("gea", Language::Gea),
        ("apc", Language::Apc),
        ("rjg", Language::Rjg),
        ("xcw", Language::Xcw),
        ("bgd", Language::Bgd),
        ("mwa", Language::Mwa),
        ("sbb", Language::Sbb),
        ("oue", Language::Oue),
        ("buq", Language::Buq),
        ("aup", Language::Aup),
        ("tng", Language::Tng),
        ("jbm", Language::Jbm),
        ("tuu", Language::Tuu),
        ("lbl", Language::Lbl),
        ("ngz", Language::Ngz),
        ("xak", Language::Xak),
        ("bdc", Language::Bdc),
        ("grx", Language::Grx),
        ("zme", Language::Zme),
        ("acz", Language::Acz),
        ("tii", Language::Tii),
        ("mhx", Language::Mhx),
        ("abv", Language::Abv),
        ("vum", Language::Vum),
        ("gim", Language::Gim),
        ("zoo", Language::Zoo),
        ("gcc", Language::Gcc),
        ("cow", Language::Cow),
        ("ast", Language::Ast),
        ("mjc", Language::Mjc),
        ("lmx", Language::Lmx),
        ("ran", Language::Ran),
        ("sri", Language::Sri),
        ("mkg", Language::Mkg),
        ("czk", Language::Czk),
        ("mff", Language::Mff),
        ("gmn", Language::Gmn),
        ("tpe", Language::Tpe),
        ("rel", Language::Rel),
        ("das", Language::Das),
        ("lmj", Language::Lmj),
        ("gid", Language::Gid),
        ("src", Language::Src),
        ("yka", Language::Yka),
        ("avs", Language::Avs),
        ("yhd", Language::Yhd),
        ("ged", Language::Ged),
        ("lbj", Language::Lbj),
        ("sln", Language::Sln),
        ("tft", Language::Tft),
        ("acp", Language::Acp),
        ("brw", Language::Brw),
        ("cid", Language::Cid),
        ("dsq", Language::Dsq),
        ("leu", Language::Leu),
        ("apd", Language::Apd),
        ("ant", Language::Ant),
        ("tls", Language::Tls),
        ("njm", Language::Njm),
        ("lwm", Language::Lwm),
        ("wrz", Language::Wrz),
        ("unu", Language::Unu),
        ("mfr", Language::Mfr),
        ("afs", Language::Afs),
        ("aps", Language::Aps),
        ("lix", Language::Lix),
        ("acq", Language::Acq),
        ("kmk", Language::Kmk),
        ("csz", Language::Csz),
        ("mxh", Language::Mxh),
        ("lbt", Language::Lbt),
        ("rmz", Language::Rmz),
        ("zmk", Language::Zmk),
        ("ore", Language::Ore),
        ("tob", Language::Tob),
        ("kfl", Language::Kfl),
        ("aoj", Language::Aoj),
        ("alr", Language::Alr),
        ("ojc", Language::Ojc),
        ("uar", Language::Uar),
        ("ztm", Language::Ztm),
        ("fau", Language::Fau),
        ("pbt", Language::Pbt),
        ("skq", Language::Skq),
        ("hld", Language::Hld),
        ("xnn", Language::Xnn),
        ("snx", Language::Snx),
        ("ums", Language::Ums),
        ("wga", Language::Wga),
        ("myb", Language::Myb),
        ("bjp", Language::Bjp),
        ("kxx", Language::Kxx),
        ("lhh", Language::Lhh),
        ("adu", Language::Adu),
        ("agg", Language::Agg),
        ("rkb", Language::Rkb),
        ("ncl", Language::Ncl),
        ("any", Language::Any),
        ("esh", Language::Esh),
        ("srq", Language::Srq),
        ("ksy", Language::Ksy),
        ("kgt", Language::Kgt),
        ("kvw", Language::Kvw),
        ("czt", Language::Czt),
        ("mio", Language::Mio),
        ("kld", Language::Kld),
        ("qvw", Language::Qvw),
        ("csp", Language::Csp),
        ("cbi", Language::Cbi),
        ("tna", Language::Tna),
        ("ple", Language::Ple),
        ("wtm", Language::Wtm),
        ("sno", Language::Sno),
        ("nen", Language::Nen),
        ("pov", Language::Pov),
        ("sgh", Language::Sgh),
        ("axl", Language::Axl),
        ("ofu", Language::Ofu),
        ("xbg", Language::Xbg),
        ("waj", Language::Waj),
        ("mxv", Language::Mxv),
        ("mne", Language::Mne),
        ("yzk", Language::Yzk),
        ("sth", Language::Sth),
        ("arn", Language::Arn),
        ("bxr", Language::Bxr),
        ("kfe", Language::Kfe),
        ("mtv", Language::Mtv),
        ("jka", Language::Jka),
        ("pos", Language::Pos),
        ("rom", Language::Rom),
        ("sor", Language::Sor),
        ("cax", Language::Cax),
        ("tyj", Language::Tyj),
        ("iin", Language::Iin),
        ("kmj", Language::Kmj),
        ("dng", Language::Dng),
        ("war", Language::War),
        ("rub", Language::Rub),
        ("khd", Language::Khd),
        ("ssv", Language::Ssv),
        ("kbv", Language::Kbv),
        ("cce", Language::Cce),
        ("ksl", Language::Ksl),
        ("tya", Language::Tya),
        ("atd", Language::Atd),
        ("nua", Language::Nua),
        ("axm", Language::Axm),
        ("shi", Language::Shi),
        ("bdl", Language::Bdl),
        ("gyi", Language::Gyi),
        ("bap", Language::Bap),
        ("sfs", Language::Sfs),
        ("ung", Language::Ung),
        ("tms", Language::Tms),
        ("kor", Language::Kor),
        ("xra", Language::Xra),
        ("ccp", Language::Ccp),
        ("csh", Language::Csh),
        ("tlg", Language::Tlg),
        ("efe", Language::Efe),
        ("hob", Language::Hob),
        ("iff", Language::Iff),
        ("cwb", Language::Cwb),
        ("kbp", Language::Kbp),
        ("khs", Language::Khs),
        ("bkz", Language::Bkz),
        ("tbu", Language::Tbu),
        ("lnw", Language::Lnw),
        ("rag", Language::Rag),
        ("pca", Language::Pca),
        ("czh", Language::Czh),
        ("uon", Language::Uon),
        ("gev", Language::Gev),
        ("kkd", Language::Kkd),
        ("bml", Language::Bml),
        ("ttg", Language::Ttg),
        ("cos", Language::Cos),
        ("waz", Language::Waz),
        ("lwo", Language::Lwo),
        ("clh", Language::Clh),
        ("shs", Language::Shs),
        ("gag", Language::Gag),
        ("kjy", Language::Kjy),
        ("amb", Language::Amb),
        ("whk", Language::Whk),
        ("crl", Language::Crl),
        ("wul", Language::Wul),
        ("cod", Language::Cod),
        ("mew", Language::Mew),
        ("cco", Language::Cco),
        ("brp", Language::Brp),
        ("pgs", Language::Pgs),
        ("wwo", Language::Wwo),
        ("dik", Language::Dik),
        ("raf", Language::Raf),
        ("ort", Language::Ort),
        ("tpu", Language::Tpu),
        ("ydg", Language::Ydg),
        ("tdy", Language::Tdy),
        ("krx", Language::Krx),
        ("wuh", Language::Wuh),
        ("kwx", Language::Kwx),
        ("aaf", Language::Aaf),
        ("akm", Language::Akm),
        ("lmi", Language::Lmi),
        ("xrg", Language::Xrg),
        ("bcf", Language::Bcf),
        ("ncq", Language::Ncq),
        ("smx", Language::Smx),
        ("tmb", Language::Tmb),
        ("vol", Language::Vol),
        ("qwt", Language::Qwt),
        ("tor", Language::Tor),
        ("gdr", Language::Gdr),
        ("brl", Language::Brl),
        ("lld", Language::Lld),
        ("agq", Language::Agq),
        ("isi", Language::Isi),
        ("plu", Language::Plu),
        ("msp", Language::Msp),
        ("hod", Language::Hod),
        ("nca", Language::Nca),
        ("lbo", Language::Lbo),
        ("asc", Language::Asc),
        ("aku", Language::Aku),
        ("lax", Language::Lax),
        ("dij", Language::Dij),
        ("mav", Language::Mav),
        ("pcj", Language::Pcj),
        ("byj", Language::Byj),
        ("ddw", Language::Ddw),
        ("nhf", Language::Nhf),
        ("tzn", Language::Tzn),
        ("nwo", Language::Nwo),
        ("kyp", Language::Kyp),
        ("mdr", Language::Mdr),
        ("yhl", Language::Yhl),
        ("jra", Language::Jra),
        ("dmb", Language::Dmb),
        ("ttb", Language::Ttb),
        ("bcs", Language::Bcs),
        ("bmf", Language::Bmf),
        ("xyb", Language::Xyb),
        ("mmy", Language::Mmy),
        ("llm", Language::Llm),
        ("por", Language::Por),
        ("kbj", Language::Kbj),
        ("mot", Language::Mot),
        ("crz", Language::Crz),
        ("bwu", Language::Bwu),
        ("cjh", Language::Cjh),
        ("oog", Language::Oog),
        ("tpm", Language::Tpm),
        ("xss", Language::Xss),
        ("lif", Language::Lif),
        ("mep", Language::Mep),
        ("pmf", Language::Pmf),
        ("alk", Language::Alk),
        ("tdf", Language::Tdf),
        ("luu", Language::Luu),
        ("mvk", Language::Mvk),
        ("tsg", Language::Tsg),
        ("xbj", Language::Xbj),
        ("rms", Language::Rms),
        ("kvb", Language::Kvb),
        ("ema", Language::Ema),
        ("kfo", Language::Kfo),
        ("lee", Language::Lee),
        ("zrp", Language::Zrp),
        ("kbg", Language::Kbg),
        ("kjn", Language::Kjn),
        ("luz", Language::Luz),
        ("smh", Language::Smh),
        ("vag", Language::Vag),
        ("tra", Language::Tra),
        ("lmd", Language::Lmd),
        ("kiy", Language::Kiy),
        ("akx", Language::Akx),
        ("scn", Language::Scn),
        ("xsb", Language::Xsb),
        ("bhs", Language::Bhs),
        ("nlu", Language::Nlu),
        ("tcl", Language::Tcl),
        ("jay", Language::Jay),
        ("gbp", Language::Gbp),
        ("dau", Language::Dau),
        ("pab", Language::Pab),
        ("tkq", Language::Tkq),
        ("agw", Language::Agw),
        ("gay", Language::Gay),
        ("bdj", Language::Bdj),
        ("mjo", Language::Mjo),
        ("muv", Language::Muv),
        ("mty", Language::Mty),
        ("kei", Language::Kei),
        ("sqr", Language::Sqr),
        ("otz", Language::Otz),
        ("hvn", Language::Hvn),
        ("osn", Language::Osn),
        ("biu", Language::Biu),
        ("mtw", Language::Mtw),
        ("nor", Language::Nor),
        ("mdv", Language::Mdv),
        ("lti", Language::Lti),
        ("cts", Language::Cts),
        ("kuv", Language::Kuv),
        ("mok", Language::Mok),
        ("hav", Language::Hav),
        ("pls", Language::Pls),
        ("tse", Language::Tse),
        ("dlg", Language::Dlg),
        ("nao", Language::Nao),
        ("boq", Language::Boq),
        ("gjm", Language::Gjm),
        ("wsr", Language::Wsr),
        ("mts", Language::Mts),
        ("pmr", Language::Pmr),
        ("njh", Language::Njh),
        ("cns", Language::Cns),
        ("chj", Language::Chj),
        ("ida", Language::Ida),
        ("xkc", Language::Xkc),
        ("yba", Language::Yba),
        ("jgk", Language::Jgk),
        ("zpv", Language::Zpv),
        ("imr", Language::Imr),
        ("sqn", Language::Sqn),
        ("pkn", Language::Pkn),
        ("dbv", Language::Dbv),
        ("kco", Language::Kco),
        ("nly", Language::Nly),
        ("bra", Language::Bra),
        ("amp", Language::Amp),
        ("mxp", Language::Mxp),
        ("yae", Language::Yae),
        ("arv", Language::Arv),
        ("mox", Language::Mox),
        ("rth", Language::Rth),
        ("bjl", Language::Bjl),
        ("man", Language::Man),
        ("pnd", Language::Pnd),
        ("oia", Language::Oia),
        ("dtp", Language::Dtp),
        ("bko", Language::Bko),
        ("lzz", Language::Lzz),
        ("boz", Language::Boz),
        ("vis", Language::Vis),
        ("buu", Language::Buu),
        ("kxc", Language::Kxc),
        ("xpb", Language::Xpb),
        ("krk", Language::Krk),
        ("ans", Language::Ans),
        ("nnz", Language::Nnz),
        ("rat", Language::Rat),
        ("cde", Language::Cde),
        ("dlk", Language::Dlk),
        ("mzb", Language::Mzb),
        ("atj", Language::Atj),
        ("org", Language::Org),
        ("cme", Language::Cme),
        ("wun", Language::Wun),
        ("nkz", Language::Nkz),
        ("rmt", Language::Rmt),
        ("bnb", Language::Bnb),
        ("hme", Language::Hme),
        ("pcd", Language::Pcd),
        ("plq", Language::Plq),
        ("rsn", Language::Rsn),
        ("xtl", Language::Xtl),
        ("eza", Language::Eza),
        ("mgp", Language::Mgp),
        ("uzb", Language::Uzb),
        ("bwg", Language::Bwg),
        ("nkm", Language::Nkm),
        ("laa", Language::Laa),
        ("bsu", Language::Bsu),
        ("bgb", Language::Bgb),
        ("bvw", Language::Bvw),
        ("xdc", Language::Xdc),
        ("kyt", Language::Kyt),
        ("skb", Language::Skb),
        ("mdu", Language::Mdu),
        ("urk", Language::Urk),
        ("obu", Language::Obu),
        ("xgr", Language::Xgr),
        ("oon", Language::Oon),
        ("bqy", Language::Bqy),
        ("sty", Language::Sty),
        ("nza", Language::Nza),
        ("kzi", Language::Kzi),
        ("ndi", Language::Ndi),
        ("kvv", Language::Kvv),
        ("bfu", Language::Bfu),
        ("jas", Language::Jas),
        ("jks", Language::Jks),
        ("ocm", Language::Ocm),
        ("ccm", Language::Ccm),
        ("kla", Language::Kla),
        ("pse", Language::Pse),
        ("zga", Language::Zga),
        ("iqw", Language::Iqw),
        ("ptu", Language::Ptu),
        ("prt", Language::Prt),
        ("aar", Language::Aar),
        ("hae", Language::Hae),
        ("sjb", Language::Sjb),
        ("gnt", Language::Gnt),
        ("gwg", Language::Gwg),
        ("nmu", Language::Nmu),
        ("nnj", Language::Nnj),
        ("nww", Language::Nww),
        ("rnw", Language::Rnw),
        ("crk", Language::Crk),
        ("ciy", Language::Ciy),
        ("mgs", Language::Mgs),
        ("lap", Language::Lap),
        ("umo", Language::Umo),
        ("faj", Language::Faj),
        ("bdw", Language::Bdw),
        ("mau", Language::Mau),
        ("orc", Language::Orc),
        ("blt", Language::Blt),
        ("prp", Language::Prp),
        ("scg", Language::Scg),
        ("ynk", Language::Ynk),
        ("ggk", Language::Ggk),
        ("nmf", Language::Nmf),
        ("oht", Language::Oht),
        ("mzj", Language::Mzj),
        ("non", Language::Non),
        ("kil", Language::Kil),
        ("say", Language::Say),
        ("grt", Language::Grt),
        ("miy", Language::Miy),
        ("ktc", Language::Ktc),
        ("vsi", Language::Vsi),
        ("ilb", Language::Ilb),
        ("pnq", Language::Pnq),
        ("bgw", Language::Bgw),
        ("tcs", Language::Tcs),
        ("jaf", Language::Jaf),
        ("xtt", Language::Xtt),
        ("dgr", Language::Dgr),
        ("apt", Language::Apt),
        ("hmv", Language::Hmv),
        ("cry", Language::Cry),
        ("oti", Language::Oti),
        ("blp", Language::Blp),
        ("kle", Language::Kle),
        ("byl", Language::Byl),
        ("pko", Language::Pko),
        ("uta", Language::Uta),
        ("cov", Language::Cov),
        ("kdy", Language::Kdy),
        ("aon", Language::Aon),
        ("yuw", Language::Yuw),
        ("mdi", Language::Mdi),
        ("twl", Language::Twl),
        ("ask", Language::Ask),
        ("trm", Language::Trm),
        ("lep", Language::Lep),
        ("kjc", Language::Kjc),
        ("hos", Language::Hos),
        ("mai", Language::Mai),
        ("oyy", Language::Oyy),
        ("thd", Language::Thd),
        ("woi", Language::Woi),
        ("rjs", Language::Rjs),
        ("zkn", Language::Zkn),
        ("tom", Language::Tom),
        ("uda", Language::Uda),
        ("tns", Language::Tns),
        ("bsr", Language::Bsr),
        ("ndn", Language::Ndn),
        ("gob", Language::Gob),
        ("esm", Language::Esm),
        ("nla", Language::Nla),
        ("pss", Language::Pss),
        ("tnv", Language::Tnv),
        ("ykl", Language::Ykl),
        ("nin", Language::Nin),
        ("kwk", Language::Kwk),
        ("mkp", Language::Mkp),
        ("loe", Language::Loe),
        ("zzj", Language::Zzj),
        ("usu", Language::Usu),
        ("pnu", Language::Pnu),
        ("dcc", Language::Dcc),
        ("mfi", Language::Mfi),
        ("bbt", Language::Bbt),
        ("wla", Language::Wla),
        ("xpr", Language::Xpr),
        ("bbm", Language::Bbm),
        ("mhw", Language::Mhw),
        ("uku", Language::Uku),
        ("zmo", Language::Zmo),
        ("cof", Language::Cof),
        ("sys", Language::Sys),
        ("xcy", Language::Xcy),
        ("srd", Language::Srd),
        ("liv", Language::Liv),
        ("lij", Language::Lij),
        ("upv", Language::Upv),
        ("czo", Language::Czo),
        ("zza", Language::Zza),
        ("xmk", Language::Xmk),
        ("nzb", Language::Nzb),
        ("dak", Language::Dak),
        ("kjv", Language::Kjv),
        ("its", Language::Its),
        ("omr", Language::Omr),
        ("kqv", Language::Kqv),
        ("bjm", Language::Bjm),
        ("cvn", Language::Cvn),
        ("ssm", Language::Ssm),
        ("lhp", Language::Lhp),
        ("ire", Language::Ire),
        ("fod", Language::Fod),
        ("xop", Language::Xop),
        ("box", Language::Box),
        ("swa", Language::Swa),
        ("anm", Language::Anm),
        ("sjg", Language::Sjg),
        ("cst", Language::Cst),
        ("qwc", Language::Qwc),
        ("nbk", Language::Nbk),
        ("mtx", Language::Mtx),
        ("sub", Language::Sub),
        ("wap", Language::Wap),
        ("ack", Language::Ack),
        ("tmz", Language::Tmz),
        ("gjn", Language::Gjn),
        ("rhg", Language::Rhg),
        ("kos", Language::Kos),
        ("yil", Language::Yil),
        ("srl", Language::Srl),
        ("mtb", Language::Mtb),
        ("soa", Language::Soa),
        ("pll", Language::Pll),
        ("nur", Language::Nur),
        ("szs", Language::Szs),
        ("kmv", Language::Kmv),
        ("gaj", Language::Gaj),
        ("bvj", Language::Bvj),
        ("her", Language::Her),
        ("nos", Language::Nos),
        ("nwc", Language::Nwc),
        ("tof", Language::Tof),
        ("dbj", Language::Dbj),
        ("ptv", Language::Ptv),
        ("zpl", Language::Zpl),
        ("spb", Language::Spb),
        ("flh", Language::Flh),
        ("byt", Language::Byt),
        ("ndh", Language::Ndh),
        ("due", Language::Due),
        ("csa", Language::Csa),
        ("bso", Language::Bso),
        ("mga", Language::Mga),
        ("rif", Language::Rif),
        ("mph", Language::Mph),
        ("cji", Language::Cji),
        ("snq", Language::Snq),
        ("tly", Language::Tly),
        ("dgc", Language::Dgc),
        ("klu", Language::Klu),
        ("hrm", Language::Hrm),
        ("wei", Language::Wei),
        ("hke", Language::Hke),
        ("tll", Language::Tll),
        ("brh", Language::Brh),
        ("klx", Language::Klx),
        ("ulu", Language::Ulu),
        ("ntg", Language::Ntg),
        ("ndz", Language::Ndz),
        ("otr", Language::Otr),
        ("wii", Language::Wii),
        ("spq", Language::Spq),
        ("txh", Language::Txh),
        ("wdy", Language::Wdy),
        ("jrb", Language::Jrb),
        ("xap", Language::Xap),
        ("keh", Language::Keh),
        ("xpg", Language::Xpg),
        ("mit", Language::Mit),
        ("gad", Language::Gad),
        ("crr", Language::Crr),
        ("sgm", Language::Sgm),
        ("mch", Language::Mch),
        ("ayz", Language::Ayz),
        ("khz", Language::Khz),
        ("ivb", Language::Ivb),
        ("fkk", Language::Fkk),
        ("jpa", Language::Jpa),
        ("acb", Language::Acb),
        ("mmd", Language::Mmd),
        ("zmp", Language::Zmp),
        ("smt", Language::Smt),
        ("bzt", Language::Bzt),
        ("okr", Language::Okr),
        ("kms", Language::Kms),
        ("ikp", Language::Ikp),
        ("wrn", Language::Wrn),
        ("esl", Language::Esl),
        ("cbo", Language::Cbo),
        ("nlq", Language::Nlq),
        ("pms", Language::Pms),
        ("xqa", Language::Xqa),
        ("prx", Language::Prx),
        ("noz", Language::Noz),
        ("pfa", Language::Pfa),
        ("onp", Language::Onp),
        ("sam", Language::Sam),
        ("ddi", Language::Ddi),
        ("xaj", Language::Xaj),
        ("bmo", Language::Bmo),
        ("ivv", Language::Ivv),
        ("hdy", Language::Hdy),
        ("nuf", Language::Nuf),
        ("dnd", Language::Dnd),
        ("bwy", Language::Bwy),
        ("smu", Language::Smu),
        ("kuy", Language::Kuy),
        ("ptr", Language::Ptr),
        ("cni", Language::Cni),
        ("aii", Language::Aii),
        ("mkj", Language::Mkj),
        ("akq", Language::Akq),
        ("omp", Language::Omp),
        ("tdx", Language::Tdx),
        ("nkk", Language::Nkk),
        ("kgj", Language::Kgj),
        ("trr", Language::Trr),
        ("shu", Language::Shu),
        ("igm", Language::Igm),
        ("ntp", Language::Ntp),
        ("zaj", Language::Zaj),
        ("des", Language::Des),
        ("mgf", Language::Mgf),
        ("nra", Language::Nra),
        ("wem", Language::Wem),
        ("tye", Language::Tye),
        ("tpx", Language::Tpx),
        ("fmp", Language::Fmp),
        ("tuh", Language::Tuh),
        ("bmq", Language::Bmq),
        ("awk", Language::Awk),
        ("juu", Language::Juu),
        ("oco", Language::Oco),
        ("cje", Language::Cje),
        ("nwb", Language::Nwb),
        ("hhy", Language::Hhy),
        ("wec", Language::Wec),
        ("swh", Language::Swh),
        ("eki", Language::Eki),
        ("mhs", Language::Mhs),
        ("zeh", Language::Zeh),
        ("vem", Language::Vem),
        ("mtt", Language::Mtt),
        ("myc", Language::Myc),
        ("scl", Language::Scl),
        ("sbx", Language::Sbx),
        ("wim", Language::Wim),
        ("ekk", Language::Ekk),
        ("nga", Language::Nga),
        ("oge", Language::Oge),
        ("lec", Language::Lec),
        ("guj", Language::Guj),
        ("xlg", Language::Xlg),
        ("esg", Language::Esg),
        ("dbb", Language::Dbb),
        ("toi", Language::Toi),
        ("pre", Language::Pre),
        ("khx", Language::Khx),
        ("suk", Language::Suk),
        ("bvg", Language::Bvg),
        ("ksf", Language::Ksf),
        ("drd", Language::Drd),
        ("mdl", Language::Mdl),
        ("rrt", Language::Rrt),
        ("spi", Language::Spi),
        ("amj", Language::Amj),
        ("pnv", Language::Pnv),
        ("pty", Language::Pty),
        ("twh", Language::Twh),
        ("xai", Language::Xai),
        ("vmu", Language::Vmu),
        ("kdd", Language::Kdd),
        ("tjs", Language::Tjs),
        ("bib", Language::Bib),
        ("xnh", Language::Xnh),
        ("bus", Language::Bus),
        ("kee", Language::Kee),
        ("jen", Language::Jen),
        ("sto", Language::Sto),
        ("pqa", Language::Pqa),
        ("sjd", Language::Sjd),
        ("kxo", Language::Kxo),
        ("sis", Language::Sis),
        ("tji", Language::Tji),
        ("urt", Language::Urt),
        ("mcy", Language::Mcy),
        ("ayk", Language::Ayk),
        ("dgn", Language::Dgn),
        ("xnj", Language::Xnj),
        ("yoy", Language::Yoy),
        ("mma", Language::Mma),
        ("nqq", Language::Nqq),
        ("ltg", Language::Ltg),
        ("kje", Language::Kje),
        ("zos", Language::Zos),
        ("lop", Language::Lop),
        ("kod", Language::Kod),
        ("uss", Language::Uss),
        ("luc", Language::Luc),
        ("ted", Language::Ted),
        ("oyb", Language::Oyb),
        ("tni", Language::Tni),
        ("gab", Language::Gab),
        ("zpp", Language::Zpp),
        ("haq", Language::Haq),
        ("srk", Language::Srk),
        ("mtc", Language::Mtc),
        ("lmu", Language::Lmu),
        ("flr", Language::Flr),
        ("sdh", Language::Sdh),
        ("ots", Language::Ots),
        ("spo", Language::Spo),
        ("xcc", Language::Xcc),
        ("glo", Language::Glo),
        ("ahn", Language::Ahn),
        ("ota", Language::Ota),
        ("slh", Language::Slh),
        ("are", Language::Are),
        ("suc", Language::Suc),
        ("uli", Language::Uli),
        ("wci", Language::Wci),
        ("cko", Language::Cko),
        ("wyr", Language::Wyr),
        ("ceb", Language::Ceb),
        ("koi", Language::Koi),
        ("aur", Language::Aur),
        ("sug", Language::Sug),
        ("eth", Language::Eth),
        ("jum", Language::Jum),
        ("cuo", Language::Cuo),
        ("nmv", Language::Nmv),
        ("xpj", Language::Xpj),
        ("rzh", Language::Rzh),
        ("crq", Language::Crq),
        ("abk", Language::Abk),
        ("sip", Language::Sip),
        ("wfg", Language::Wfg),
        ("abe", Language::Abe),
        ("kko", Language::Kko),
        ("sol", Language::Sol),
        ("trt", Language::Trt),
        ("aqn", Language::Aqn),
        ("lsp", Language::Lsp),
        ("tam", Language::Tam),
        ("syb", Language::Syb),
        ("wbh", Language::Wbh),
        ("gut", Language::Gut),
        ("xnk", Language::Xnk),
        ("grq", Language::Grq),
        ("rus", Language::Rus),
        ("brk", Language::Brk),
        ("tcn", Language::Tcn),
        ("mkr", Language::Mkr),
        ("mce", Language::Mce),
        ("amo", Language::Amo),
        ("ybj", Language::Ybj),
        ("yav", Language::Yav),
        ("hlb", Language::Hlb),
        ("hla", Language::Hla),
        ("mli", Language::Mli),
        ("wsi", Language::Wsi),
        ("wlg", Language::Wlg),
        ("tba", Language::Tba),
        ("mgg", Language::Mgg),
        ("npx", Language::Npx),
        ("nov", Language::Nov),
        ("xiy", Language::Xiy),
        ("czn", Language::Czn),
        ("sbd", Language::Sbd),
        ("tbh", Language::Tbh),
        ("xbb", Language::Xbb),
        ("rtw", Language::Rtw),
        ("seg", Language::Seg),
        ("bow", Language::Bow),
        ("mbk", Language::Mbk),
        ("ngc", Language::Ngc),
        ("mis", Language::Mis),
        ("arx", Language::Arx),
        ("ghn", Language::Ghn),
        ("whg", Language::Whg),
        ("sgp", Language::Sgp),
        ("kjs", Language::Kjs),
        ("aym", Language::Aym),
        ("bfi", Language::Bfi),
        ("xlp", Language::Xlp),
        ("grh", Language::Grh),
        ("xla", Language::Xla),
        ("pgu", Language::Pgu),
        ("xeb", Language::Xeb),
        ("mkf", Language::Mkf),
        ("pru", Language::Pru),
        ("tei", Language::Tei),
        ("xyl", Language::Xyl),
        ("xnq", Language::Xnq),
        ("jbn", Language::Jbn),
        ("ras", Language::Ras),
        ("igo", Language::Igo),
        ("heb", Language::Heb),
        ("vid", Language::Vid),
        ("sab", Language::Sab),
        ("rgr", Language::Rgr),
        ("hif", Language::Hif),
        ("lad", Language::Lad),
        ("kjq", Language::Kjq),
        ("bny", Language::Bny),
        ("irk", Language::Irk),
        ("swi", Language::Swi),
        ("cml", Language::Cml),
        ("kca", Language::Kca),
        ("kly", Language::Kly),
        ("oac", Language::Oac),
        ("hre", Language::Hre),
        ("pdu", Language::Pdu),
        ("kfs", Language::Kfs),
        ("jul", Language::Jul),
        ("isr", Language::Isr),
        ("mpm", Language::Mpm),
        ("lud", Language::Lud),
        ("pny", Language::Pny),
        ("bzc", Language::Bzc),
        ("bft", Language::Bft),
        ("raj", Language::Raj),
        ("bhj", Language::Bhj),
        ("nuy", Language::Nuy),
        ("oci", Language::Oci),
        ("djr", Language::Djr),
        ("pst", Language::Pst),
        ("kzq", Language::Kzq),
        ("cab", Language::Cab),
        ("kim", Language::Kim),
        ("zkk", Language::Zkk),
        ("alh", Language::Alh),
        ("loy", Language::Loy),
        ("wnm", Language::Wnm),
        ("csb", Language::Csb),
        ("mys", Language::Mys),
        ("hop", Language::Hop),
        ("gbr", Language::Gbr),
        ("xsm", Language::Xsm),
        ("mnn", Language::Mnn),
        ("bsj", Language::Bsj),
        ("gda", Language::Gda),
        ("bzv", Language::Bzv),
        ("bkl", Language::Bkl),
        ("bau", Language::Bau),
        ("sdk", Language::Sdk),
        ("yuk", Language::Yuk),
        ("nrl", Language::Nrl),
        ("all", Language::All),
        ("bot", Language::Bot),
        ("lin", Language::Lin),
        ("lbb", Language::Lbb),
        ("trn", Language::Trn),
        ("apb", Language::Apb),
        ("ygw", Language::Ygw),
        ("csc", Language::Csc),
        ("kcb", Language::Kcb),
        ("quw", Language::Quw),
        ("bzu", Language::Bzu),
        ("nti", Language::Nti),
        ("btw", Language::Btw),
        ("bim", Language::Bim),
        ("ddd", Language::Ddd),
        ("ghr", Language::Ghr),
        ("hdn", Language::Hdn),
        ("len", Language::Len),
        ("dtm", Language::Dtm),
        ("saq", Language::Saq),
        ("tqp", Language::Tqp),
        ("nog", Language::Nog),
        ("wif", Language::Wif),
        ("bsb", Language::Bsb),
        ("xtn", Language::Xtn),
        ("tay", Language::Tay),
        ("eli", Language::Eli),
        ("bay", Language::Bay),
        ("tdb", Language::Tdb),
        ("xli", Language::Xli),
        ("dtd", Language::Dtd),
        ("bhx", Language::Bhx),
        ("bxf", Language::Bxf),
        ("leq", Language::Leq),
        ("tzx", Language::Tzx),
        ("yll", Language::Yll),
        ("kdp", Language::Kdp),
        ("bmv", Language::Bmv),
        ("goj", Language::Goj),
        ("ywl", Language::Ywl),
        ("bsh", Language::Bsh),
        ("ybb", Language::Ybb),
        ("tuy", Language::Tuy),
        ("ckx", Language::Ckx),
        ("bsi", Language::Bsi),
        ("meh", Language::Meh),
        ("twq", Language::Twq),
        ("hch", Language::Hch),
        ("mbw", Language::Mbw),
        ("kns", Language::Kns),
        ("rue", Language::Rue),
        ("qxq", Language::Qxq),
        ("zca", Language::Zca),
        ("iyo", Language::Iyo),
        ("xiv", Language::Xiv),
        ("kwm", Language::Kwm),
        ("nuo", Language::Nuo),
        ("nbc", Language::Nbc),
        ("ksp", Language::Ksp),
        ("tur", Language::Tur),
        ("gis", Language::Gis),
        ("bjg", Language::Bjg),
        ("mib", Language::Mib),
        ("gdt", Language::Gdt),
        ("btt", Language::Btt),
        ("dju", Language::Dju),
        ("zmf", Language::Zmf),
        ("dgl", Language::Dgl),
        ("adb", Language::Adb),
        ("lex", Language::Lex),
        ("nnt", Language::Nnt),
        ("pal", Language::Pal),
        ("aok", Language::Aok),
        ("pnp", Language::Pnp),
        ("xrr", Language::Xrr),
        ("pas", Language::Pas),
        ("zyj", Language::Zyj),
        ("ddo", Language::Ddo),
        ("tzh", Language::Tzh),
        ("bgs", Language::Bgs),
        ("ngq", Language::Ngq),
        ("gvp", Language::Gvp),
        ("shb", Language::Shb),
        ("apf", Language::Apf),
        ("gra", Language::Gra),
        ("qwh", Language::Qwh),
        ("bgo", Language::Bgo),
        ("anv", Language::Anv),
        ("bxb", Language::Bxb),
        ("mrp", Language::Mrp),
        ("gsw", Language::Gsw),
        ("oji", Language::Oji),
        ("ysr", Language::Ysr),
        ("dnr", Language::Dnr),
        ("swn", Language::Swn),
        ("muq", Language::Muq),
        ("oke", Language::Oke),
        ("hmg", Language::Hmg),
        ("cre", Language::Cre),
        ("bbn", Language::Bbn),
        ("prz", Language::Prz),
        ("nqy", Language::Nqy),
        ("emu", Language::Emu),
        ("coj", Language::Coj),
        ("nwe", Language::Nwe),
        ("rmw", Language::Rmw),
        ("mzu", Language::Mzu),
        ("tuk", Language::Tuk),
        ("rro", Language::Rro),
        ("omg", Language::Omg),
        ("uvl", Language::Uvl),
        ("ell", Language::Ell),
        ("sbe", Language::Sbe),
        ("wik", Language::Wik),
        ("jit", Language::Jit),
        ("psl", Language::Psl),
        ("xgd", Language::Xgd),
        ("vif", Language::Vif),
        ("abn", Language::Abn),
        ("xbd", Language::Xbd),
        ("jle", Language::Jle),
        ("tql", Language::Tql),
        ("bpx", Language::Bpx),
        ("awh", Language::Awh),
        ("chn", Language::Chn),
        ("ktx", Language::Ktx),
        ("jbr", Language::Jbr),
        ("yuy", Language::Yuy),
        ("pys", Language::Pys),
        ("ukr", Language::Ukr),
        ("pbb", Language::Pbb),
        ("shd", Language::Shd),
        ("ayn", Language::Ayn),
        ("tcu", Language::Tcu),
        ("gza", Language::Gza),
        ("jio", Language::Jio),
        ("btq", Language::Btq),
        ("rin", Language::Rin),
        ("ace", Language::Ace),
        ("dyn", Language::Dyn),
        ("twp", Language::Twp),
        ("cib", Language::Cib),
        ("lgm", Language::Lgm),
        ("quv", Language::Quv),
        ("cuj", Language::Cuj),
        ("jan", Language::Jan),
        ("way", Language::Way),
        ("hgm", Language::Hgm),
        ("dby", Language::Dby),
        ("xmh", Language::Xmh),
        ("lkh", Language::Lkh),
        ("dsz", Language::Dsz),
        ("kbq", Language::Kbq),
        ("xrt", Language::Xrt),
        ("mbz", Language::Mbz),
        ("lhl", Language::Lhl),
        ("jao", Language::Jao),
        ("cte", Language::Cte),
        ("lam", Language::Lam),
        ("zmy", Language::Zmy),
        ("dni", Language::Dni),
        ("skz", Language::Skz),
        ("dar", Language::Dar),
        ("nko", Language::Nko),
        ("ptq", Language::Ptq),
        ("sjl", Language::Sjl),
        ("bqd", Language::Bqd),
        ("sbu", Language::Sbu),
        ("crm", Language::Crm),
        ("zax", Language::Zax),
        ("kub", Language::Kub),
        ("phu", Language::Phu),
        ("qub", Language::Qub),
        ("arc", Language::Arc),
        ("ghh", Language::Ghh),
        ("otd", Language::Otd),
        ("mbo", Language::Mbo),
        ("weo", Language::Weo),
        ("con", Language::Con),
        ("mrr", Language::Mrr),
        ("kcg", Language::Kcg),
        ("bfj", Language::Bfj),
        ("xsv", Language::Xsv),
        ("prq", Language::Prq),
        ("mxw", Language::Mxw),
        ("tvo", Language::Tvo),
        ("zmw", Language::Zmw),
        ("kyz", Language::Kyz),
        ("tqm", Language::Tqm),
        ("wnc", Language::Wnc),
        ("nhq", Language::Nhq),
        ("aki", Language::Aki),
        ("nco", Language::Nco),
        ("onk", Language::Onk),
        ("bym", Language::Bym),
        ("tdr", Language::Tdr),
        ("hoe", Language::Hoe),
        ("elh", Language::Elh),
        ("lew", Language::Lew),
        ("jal", Language::Jal),
        ("gus", Language::Gus),
        ("mlf", Language::Mlf),
        ("xgu", Language::Xgu),
        ("bum", Language::Bum),
        ("asz", Language::Asz),
        ("kln", Language::Kln),
        ("mvb", Language::Mvb),
        ("doh", Language::Doh),
        ("dtn", Language::Dtn),
        ("caa", Language::Caa),
        ("rem", Language::Rem),
        ("ush", Language::Ush),
        ("mbl", Language::Mbl),
        ("zdj", Language::Zdj),
        ("vkt", Language::Vkt),
        ("ero", Language::Ero),
        ("qug", Language::Qug),
        ("lva", Language::Lva),
        ("dmf", Language::Dmf),
        ("dml", Language::Dml),
        ("bdo", Language::Bdo),
        ("asy", Language::Asy),
        ("job", Language::Job),
        ("jns", Language::Jns),
        ("ksn", Language::Ksn),
        ("gnd", Language::Gnd),
        ("anf", Language::Anf),
        ("pig", Language::Pig),
        ("dhd", Language::Dhd),
        ("itx", Language::Itx),
        ("pro", Language::Pro),
        ("aby", Language::Aby),
        ("tar", Language::Tar),
        ("drq", Language::Drq),
        ("pku", Language::Pku),
        ("ttz", Language::Ttz),
        ("oma", Language::Oma),
        ("bjz", Language::Bjz),
        ("yey", Language::Yey),
        ("unr", Language::Unr),
        ("zin", Language::Zin),
        ("ogu", Language::Ogu),
        ("twe", Language::Twe),
        ("hah", Language::Hah),
        ("cta", Language::Cta),
        ("sob", Language::Sob),
        ("sxn", Language::Sxn),
        ("rou", Language::Rou),
        ("jur", Language::Jur),
        ("gkn", Language::Gkn),
        ("mko", Language::Mko),
        ("lkn", Language::Lkn),
        ("wmb", Language::Wmb),
        ("uro", Language::Uro),
        ("kql", Language::Kql),
        ("cvg", Language::Cvg),
        ("vko", Language::Vko),
        ("soe", Language::Soe),
        ("mqh", Language::Mqh),
        ("ars", Language::Ars),
        ("ktj", Language::Ktj),
        ("kpu", Language::Kpu),
        ("mtm", Language::Mtm),
        ("lto", Language::Lto),
        ("kwd", Language::Kwd),
        ("kfh", Language::Kfh),
        ("tux", Language::Tux),
        ("kgk", Language::Kgk),
        ("ngs", Language::Ngs),
        ("vnm", Language::Vnm),
        ("tbd", Language::Tbd),
        ("myf", Language::Myf),
        ("kre", Language::Kre),
        ("nse", Language::Nse),
        ("knk", Language::Knk),
        ("nlx", Language::Nlx),
        ("fpe", Language::Fpe),
        ("dbq", Language::Dbq),
        ("kdm", Language::Kdm),
        ("mwt", Language::Mwt),
        ("aaw", Language::Aaw),
        ("ses", Language::Ses),
        ("dka", Language::Dka),
        ("jgo", Language::Jgo),
        ("kpw", Language::Kpw),
        ("lau", Language::Lau),
        ("ukv", Language::Ukv),
        ("xyy", Language::Xyy),
        ("bkv", Language::Bkv),
        ("zim", Language::Zim),
        ("era", Language::Era),
        ("phv", Language::Phv),
        ("hrv", Language::Hrv),
        ("uya", Language::Uya),
        ("ebr", Language::Ebr),
        ("pug", Language::Pug),
        ("kvo", Language::Kvo),
        ("pnn", Language::Pnn),
        ("byq", Language::Byq),
        ("awe", Language::Awe),
        ("piz", Language::Piz),
        ("ssf", Language::Ssf),
        ("ght", Language::Ght),
        ("kyn", Language::Kyn),
        ("tes", Language::Tes),
        ("yaq", Language::Yaq),
        ("kqp", Language::Kqp),
        ("gba", Language::Gba),
        ("caz", Language::Caz),
        ("kct", Language::Kct),
        ("kwb", Language::Kwb),
        ("mkt", Language::Mkt),
        ("dhm", Language::Dhm),
        ("gwe", Language::Gwe),
        ("ske", Language::Ske),
        ("mza", Language::Mza),
        ("gyb", Language::Gyb),
        ("zms", Language::Zms),
        ("muh", Language::Muh),
        ("ghs", Language::Ghs),
        ("dzl", Language::Dzl),
        ("noe", Language::Noe),
        ("nyv", Language::Nyv),
        ("ilk", Language::Ilk),
        ("ihp", Language::Ihp),
        ("spc", Language::Spc),
        ("mad", Language::Mad),
        ("ywu", Language::Ywu),
        ("tsv", Language::Tsv),
        ("lht", Language::Lht),
        ("cbl", Language::Cbl),
        ("aom", Language::Aom),
        ("nfr", Language::Nfr),
        ("yuz", Language::Yuz),
        ("khg", Language::Khg),
        ("tlh", Language::Tlh),
        ("kty", Language::Kty),
        ("nwa", Language::Nwa),
        ("anh", Language::Anh),
        ("kgp", Language::Kgp),
        ("hub", Language::Hub),
        ("bwk", Language::Bwk),
        ("aat", Language::Aat),
        ("krs", Language::Krs),
        ("bpt", Language::Bpt),
        ("bly", Language::Bly),
        ("ium", Language::Ium),
        ("kch", Language::Kch),
        ("tvs", Language::Tvs),
        ("prs", Language::Prs),
        ("loz", Language::Loz),
        ("cdn", Language::Cdn),
        ("mva", Language::Mva),
        ("pid", Language::Pid),
        ("stu", Language::Stu),
        ("crg", Language::Crg),
        ("kkh", Language::Kkh),
        ("let", Language::Let),
        ("ccr", Language::Ccr),
        ("pcl", Language::Pcl),
        ("oum", Language::Oum),
        ("rao", Language::Rao),
        ("jqr", Language::Jqr),
        ("kwl", Language::Kwl),
        ("wku", Language::Wku),
        ("bnw", Language::Bnw),
        ("hca", Language::Hca),
        ("tny", Language::Tny),
        ("nrn", Language::Nrn),
        ("bcz", Language::Bcz),
        ("pnz", Language::Pnz),
        ("mcc", Language::Mcc),
        ("nki", Language::Nki),
        ("ali", Language::Ali),
        ("yle", Language::Yle),
        ("chx", Language::Chx),
        ("nso", Language::Nso),
        ("jut", Language::Jut),
        ("meo", Language::Meo),
        ("cjs", Language::Cjs),
        ("bns", Language::Bns),
        ("cir", Language::Cir),
        ("xsy", Language::Xsy),
        ("rof", Language::Rof),
        ("mhl", Language::Mhl),
        ("mpq", Language::Mpq),
        ("mcd", Language::Mcd),
        ("aca", Language::Aca),
        ("gym", Language::Gym),
        ("kti", Language::Kti),
        ("blz", Language::Blz),
        ("krr", Language::Krr),
        ("ajp", Language::Ajp),
        ("tac", Language::Tac),
        ("xga", Language::Xga),
        ("aya", Language::Aya),
        ("ska", Language::Ska),
        ("abs", Language::Abs),
        ("kry", Language::Kry),
        ("mdy", Language::Mdy),
        ("tku", Language::Tku),
        ("jma", Language::Jma),
        ("rmd", Language::Rmd),
        ("umb", Language::Umb),
        ("ach", Language::Ach),
        ("opa", Language::Opa),
        ("bdb", Language::Bdb),
        ("bjx", Language::Bjx),
        ("mca", Language::Mca),
        ("xpq", Language::Xpq),
        ("boa", Language::Boa),
        ("knu", Language::Knu),
        ("mti", Language::Mti),
        ("yiz", Language::Yiz),
        ("gir", Language::Gir),
        ("eze", Language::Eze),
        ("sgw", Language::Sgw),
        ("ssw", Language::Ssw),
        ("bsl", Language::Bsl),
        ("inl", Language::Inl),
        ("fao", Language::Fao),
        ("gby", Language::Gby),
        ("iou", Language::Iou),
        ("plo", Language::Plo),
        ("siu", Language::Siu),
        ("vnp", Language::Vnp),
        ("kbz", Language::Kbz),
        ("tcg", Language::Tcg),
        ("ntk", Language::Ntk),
        ("ekg", Language::Ekg),
        ("kri", Language::Kri),
        ("kpj", Language::Kpj),
        ("ruh", Language::Ruh),
        ("unn", Language::Unn),
        ("tti", Language::Tti),
        ("krb", Language::Krb),
        ("lkl", Language::Lkl),
        ("gvo", Language::Gvo),
        ("jmw", Language::Jmw),
        ("emi", Language::Emi),
        ("szc", Language::Szc),
        ("ugn", Language::Ugn),
        ("ijc", Language::Ijc),
        ("dzg", Language::Dzg),
        ("nmb", Language::Nmb),
        ("qul", Language::Qul),
        ("amu", Language::Amu),
        ("reb", Language::Reb),
        ("pdc", Language::Pdc),
        ("bue", Language::Bue),
        ("jkr", Language::Jkr),
        ("zaq", Language::Zaq),
        ("nxm", Language::Nxm),
        ("bsq", Language::Bsq),
        ("agd", Language::Agd),
        ("cdf", Language::Cdf),
        ("pyy", Language::Pyy),
        ("kvf", Language::Kvf),
        ("nap", Language::Nap),
        ("kff", Language::Kff),
        ("atm", Language::Atm),
        ("gpa", Language::Gpa),
        ("nsf", Language::Nsf),
        ("kij", Language::Kij),
        ("vam", Language::Vam),
        ("cpc", Language::Cpc),
        ("moi", Language::Moi),
        ("wog", Language::Wog),
        ("kpo", Language::Kpo),
        ("zsa", Language::Zsa),
        ("yak", Language::Yak),
        ("mux", Language::Mux),
        ("ady", Language::Ady),
        ("bbj", Language::Bbj),
        ("ysg", Language::Ysg),
        ("bol", Language::Bol),
        ("kja", Language::Kja),
        ("mzg", Language::Mzg),
        ("sge", Language::Sge),
        ("gqn", Language::Gqn),
        ("sxc", Language::Sxc),
        ("bej", Language::Bej),
        ("lww", Language::Lww),
        ("gga", Language::Gga),
        ("wil", Language::Wil),
        ("nru", Language::Nru),
        ("ray", Language::Ray),
        ("ctc", Language::Ctc),
        ("mjn", Language::Mjn),
        ("myj", Language::Myj),
        ("kwv", Language::Kwv),
        ("lup", Language::Lup),
        ("onu", Language::Onu),
        ("njx", Language::Njx),
        ("mrv", Language::Mrv),
        ("ndw", Language::Ndw),
        ("was", Language::Was),
        ("nzi", Language::Nzi),
        ("pgz", Language::Pgz),
        ("tch", Language::Tch),
        ("ymg", Language::Ymg),
        ("zpo", Language::Zpo),
        ("xii", Language::Xii),
        ("yui", Language::Yui),
        ("ncn", Language::Ncn),
        ("esy", Language::Esy),
        ("tjm", Language::Tjm),
        ("zik", Language::Zik),
        ("zba", Language::Zba),
        ("cle", Language::Cle),
        ("slt", Language::Slt),
        ("ubr", Language::Ubr),
        ("gvc", Language::Gvc),
        ("wyi", Language::Wyi),
        ("ana", Language::Ana),
        ("kit", Language::Kit),
        ("boe", Language::Boe),
        ("shn", Language::Shn),
        ("vig", Language::Vig),
        ("pgi", Language::Pgi),
        ("xrd", Language::Xrd),
        ("gde", Language::Gde),
        ("otn", Language::Otn),
        ("kwp", Language::Kwp),
        ("ndr", Language::Ndr),
        ("olo", Language::Olo),
        ("grd", Language::Grd),
        ("enx", Language::Enx),
        ("mbt", Language::Mbt),
        ("fom", Language::Fom),
        ("kis", Language::Kis),
        ("ygp", Language::Ygp),
        ("aob", Language::Aob),
        ("lrn", Language::Lrn),
        ("zyp", Language::Zyp),
        ("fly", Language::Fly),
        ("cbc", Language::Cbc),
        ("zac", Language::Zac),
        ("akf", Language::Akf),
        ("xtq", Language::Xtq),
        ("gwa", Language::Gwa),
        ("btn", Language::Btn),
        ("fuc", Language::Fuc),
        ("txo", Language::Txo),
        ("omt", Language::Omt),
        ("ora", Language::Ora),
        ("umm", Language::Umm),
        ("nkb", Language::Nkb),
        ("etc", Language::Etc),
        ("mev", Language::Mev),
        ("ndq", Language::Ndq),
        ("koe", Language::Koe),
        ("ybm", Language::Ybm),
        ("sij", Language::Sij),
        ("taq", Language::Taq),
        ("psy", Language::Psy),
        ("yna", Language::Yna),
        ("xjt", Language::Xjt),
        ("bjc", Language::Bjc),
        ("acw", Language::Acw),
        ("xho", Language::Xho),
        ("psi", Language::Psi),
        ("jye", Language::Jye),
        ("cur", Language::Cur),
        ("lmy", Language::Lmy),
        ("byh", Language::Byh),
        ("nas", Language::Nas),
        ("acl", Language::Acl),
        ("gdn", Language::Gdn),
        ("shz", Language::Shz),
        ("gdx", Language::Gdx),
        ("tnz", Language::Tnz),
        ("viv", Language::Viv),
        ("gex", Language::Gex),
        ("tbi", Language::Tbi),
        ("xun", Language::Xun),
        ("sli", Language::Sli),
        ("yrl", Language::Yrl),
        ("npa", Language::Npa),
        ("tdi", Language::Tdi),
        ("ebg", Language::Ebg),
        ("qxh", Language::Qxh),
        ("opo", Language::Opo),
        ("ktv", Language::Ktv),
        ("hoh", Language::Hoh),
        ("mop", Language::Mop),
        ("msz", Language::Msz),
        ("wja", Language::Wja),
        ("akc", Language::Akc),
        ("ckt", Language::Ckt),
        ("snu", Language::Snu),
        ("rak", Language::Rak),
        ("kwn", Language::Kwn),
        ("bfk", Language::Bfk),
        ("yhs", Language::Yhs),
        ("bdn", Language::Bdn),
        ("auq", Language::Auq),
        ("cnb", Language::Cnb),
        ("ntj", Language::Ntj),
        ("kov", Language::Kov),
        ("nyp", Language::Nyp),
        ("xsn", Language::Xsn),
        ("bze", Language::Bze),
        ("enq", Language::Enq),
        ("keq", Language::Keq),
        ("pyx", Language::Pyx),
        ("mei", Language::Mei),
        ("shy", Language::Shy),
        ("kst", Language::Kst),
        ("paw", Language::Paw),
        ("bkn", Language::Bkn),
        ("bxh", Language::Bxh),
        ("cmn", Language::Cmn),
        ("pbv", Language::Pbv),
        ("gka", Language::Gka),
        ("nwi", Language::Nwi),
        ("onj", Language::Onj),
        ("nrg", Language::Nrg),
        ("ovd", Language::Ovd),
        ("sjw", Language::Sjw),
        ("bbq", Language::Bbq),
        ("mnj", Language::Mnj),
        ("xlo", Language::Xlo),
        ("mlq", Language::Mlq),
        ("klk", Language::Klk),
        ("mhd", Language::Mhd),
        ("tsa", Language::Tsa),
        ("mhj", Language::Mhj),
        ("akk", Language::Akk),
        ("qvs", Language::Qvs),
        ("aog", Language::Aog),
        ("lsn", Language::Lsn),
        ("bnv", Language::Bnv),
        ("bwd", Language::Bwd),
        ("urg", Language::Urg),
        ("wig", Language::Wig),
        ("ekp", Language::Ekp),
        ("kxq", Language::Kxq),
        ("wof", Language::Wof),
        ("omw", Language::Omw),
        ("xnm", Language::Xnm),
        ("sjr", Language::Sjr),
        ("tsr", Language::Tsr),
        ("zah", Language::Zah),
        ("npo", Language::Npo),
        ("tlu", Language::Tlu),
        ("frk", Language::Frk),
        ("yrk", Language::Yrk),
        ("zka", Language::Zka),
        ("blr", Language::Blr),
        ("tmk", Language::Tmk),
        ("grm", Language::Grm),
        ("eip", Language::Eip),
        ("mqq", Language::Mqq),
        ("dem", Language::Dem),
        ("mfh", Language::Mfh),
        ("kcq", Language::Kcq),
        ("cro", Language::Cro),
        ("lcd", Language::Lcd),
        ("hii", Language::Hii),
        ("syo", Language::Syo),
        ("kvj", Language::Kvj),
        ("suv", Language::Suv),
        ("csj", Language::Csj),
        ("xzm", Language::Xzm),
        ("srp", Language::Srp),
        ("aip", Language::Aip),
        ("mkd", Language::Mkd),
        ("mnc", Language::Mnc),
        ("orm", Language::Orm),
        ("sar", Language::Sar),
        ("wrl", Language::Wrl),
        ("lar", Language::Lar),
        ("oie", Language::Oie),
        ("xms", Language::Xms),
        ("skw", Language::Skw),
        ("nsu", Language::Nsu),
        ("nst", Language::Nst),
        ("gcn", Language::Gcn),
        ("pwa", Language::Pwa),
        ("chb", Language::Chb),
        ("tuv", Language::Tuv),
        ("url", Language::Url),
        ("ndm", Language::Ndm),
        ("geg", Language::Geg),
        ("nqg", Language::Nqg),
        ("guk", Language::Guk),
        ("nnu", Language::Nnu),
        ("ont", Language::Ont),
        ("peh", Language::Peh),
        ("nxi", Language::Nxi),
        ("kse", Language::Kse),
        ("nok", Language::Nok),
        ("uks", Language::Uks),
        ("luq", Language::Luq),
        ("coq", Language::Coq),
        ("zgb", Language::Zgb),
        ("nsw", Language::Nsw),
        ("krl", Language::Krl),
        ("aag", Language::Aag),
        ("lyn", Language::Lyn),
        ("kwh", Language::Kwh),
        ("vgt", Language::Vgt),
        ("imt", Language::Imt),
        ("zkg", Language::Zkg),
        ("mzi", Language::Mzi),
        ("syn", Language::Syn),
        ("toh", Language::Toh),
        ("tpc", Language::Tpc),
        ("bww", Language::Bww),
        ("ldk", Language::Ldk),
        ("orr", Language::Orr),
        ("mgn", Language::Mgn),
        ("frp", Language::Frp),
        ("xpv", Language::Xpv),
        ("neu", Language::Neu),
        ("ped", Language::Ped),
        ("fan", Language::Fan),
        ("tkg", Language::Tkg),
        ("pav", Language::Pav),
        ("zpx", Language::Zpx),
        ("ibe", Language::Ibe),
        ("pmq", Language::Pmq),
        ("zua", Language::Zua),
        ("mzt", Language::Mzt),
        ("huw", Language::Huw),
        ("aib", Language::Aib),
        ("huc", Language::Huc),
        ("xub", Language::Xub),
        ("xkk", Language::Xkk),
        ("djj", Language::Djj),
        ("zoq", Language::Zoq),
        ("bxz", Language::Bxz),
        ("cub", Language::Cub),
        ("tcq", Language::Tcq),
        ("cya", Language::Cya),
        ("mnv", Language::Mnv),
        ("kyd", Language::Kyd),
        ("ppm", Language::Ppm),
        ("xod", Language::Xod),
        ("mhk", Language::Mhk),
        ("mwi", Language::Mwi),
        ("uga", Language::Uga),
        ("vmh", Language::Vmh),
        ("fat", Language::Fat),
        ("sru", Language::Sru),
        ("byi", Language::Byi),
        ("igw", Language::Igw),
        ("bmn", Language::Bmn),
        ("lom", Language::Lom),
        ("sdx", Language::Sdx),
        ("nye", Language::Nye),
        ("jic", Language::Jic),
        ("rka", Language::Rka),
        ("ory", Language::Ory),
        ("udj", Language::Udj),
        ("kbb", Language::Kbb),
        ("ite", Language::Ite),
        ("kab", Language::Kab),
        ("ttk", Language::Ttk),
        ("mat", Language::Mat),
        ("nnp", Language::Nnp),
        ("juk", Language::Juk),
        ("kyo", Language::Kyo),
        ("kdh", Language::Kdh),
        ("piu", Language::Piu),
        ("acn", Language::Acn),
        ("lyg", Language::Lyg),
        ("bog", Language::Bog),
        ("hlt", Language::Hlt),
        ("wbk", Language::Wbk),
        ("aba", Language::Aba),
        ("pme", Language::Pme),
        ("lnm", Language::Lnm),
        ("aic", Language::Aic),
        ("yol", Language::Yol),
        ("cia", Language::Cia),
        ("sak", Language::Sak),
        ("gdh", Language::Gdh),
        ("lol", Language::Lol),
        ("bex", Language::Bex),
        ("moo", Language::Moo),
        ("sbf", Language::Sbf),
        ("sek", Language::Sek),
        ("mya", Language::Mya),
        ("asn", Language::Asn),
        ("khb", Language::Khb),
        ("amn", Language::Amn),
        ("ldb", Language::Ldb),
        ("wmg", Language::Wmg),
        ("noi", Language::Noi),
        ("aim", Language::Aim),
        ("aww", Language::Aww),
        ("coz", Language::Coz),
        ("bhb", Language::Bhb),
        ("huj", Language::Huj),
        ("zps", Language::Zps),
        ("ybi", Language::Ybi),
        ("ybl", Language::Ybl),
        ("tcy", Language::Tcy),
        ("slx", Language::Slx),
        ("mnz", Language::Mnz),
        ("sbs", Language::Sbs),
        ("smz", Language::Smz),
        ("ktk", Language::Ktk),
        ("bht", Language::Bht),
        ("mgk", Language::Mgk),
        ("klm", Language::Klm),
        ("aty", Language::Aty),
        ("glr", Language::Glr),
        ("kvn", Language::Kvn),
        ("mmm", Language::Mmm),
        ("goa", Language::Goa),
        ("tbf", Language::Tbf),
        ("dhr", Language::Dhr),
        ("bwj", Language::Bwj),
        ("gva", Language::Gva),
        ("srz", Language::Srz),
        ("tlx", Language::Tlx),
        ("wtk", Language::Wtk),
        ("yde", Language::Yde),
        ("hnu", Language::Hnu),
        ("wub", Language::Wub),
        ("ile", Language::Ile),
        ("kzu", Language::Kzu),
        ("mjz", Language::Mjz),
        ("pht", Language::Pht),
        ("lel", Language::Lel),
        ("ngu", Language::Ngu),
        ("kls", Language::Kls),
        ("omx", Language::Omx),
        ("yao", Language::Yao),
        ("lls", Language::Lls),
        ("apn", Language::Apn),
        ("xhv", Language::Xhv),
        ("wlr", Language::Wlr),
        ("yoi", Language::Yoi),
        ("kpq", Language::Kpq),
        ("yaz", Language::Yaz),
        ("dnw", Language::Dnw),
        ("ayh", Language::Ayh),
        ("oos", Language::Oos),
        ("ctt", Language::Ctt),
        ("oav", Language::Oav),
        ("agb", Language::Agb),
        ("ecy", Language::Ecy),
        ("txc", Language::Txc),
        ("wnp", Language::Wnp),
        ("tnc", Language::Tnc),
        ("awm", Language::Awm),
        ("tlj", Language::Tlj),
        ("hhi", Language::Hhi),
        ("avb", Language::Avb),
        ("kqr", Language::Kqr),
        ("nsc", Language::Nsc),
        ("mse", Language::Mse),
        ("cwg", Language::Cwg),
        ("smr", Language::Smr),
        ("xuj", Language::Xuj),
        ("kxp", Language::Kxp),
        ("nyg", Language::Nyg),
        ("nis", Language::Nis),
        ("chz", Language::Chz),
        ("rnp", Language::Rnp),
        ("aen", Language::Aen),
        ("ded", Language::Ded),
        ("nrk", Language::Nrk),
        ("dri", Language::Dri),
        ("ase", Language::Ase),
        ("mcs", Language::Mcs),
        ("bvr", Language::Bvr),
        ("uki", Language::Uki),
        ("ttq", Language::Ttq),
        ("pml", Language::Pml),
        ("prh", Language::Prh),
        ("koo", Language::Koo),
        ("gug", Language::Gug),
        ("xtr", Language::Xtr),
        ("ngh", Language::Ngh),
        ("nyq", Language::Nyq),
        ("skj", Language::Skj),
        ("nik", Language::Nik),
        ("lrk", Language::Lrk),
        ("nuq", Language::Nuq),
        ("liu", Language::Liu),
        ("llx", Language::Llx),
        ("tyz", Language::Tyz),
        ("luk", Language::Luk),
        ("coo", Language::Coo),
        ("ymp", Language::Ymp),
        ("pex", Language::Pex),
        ("myv", Language::Myv),
        ("pmm", Language::Pmm),
        ("led", Language::Led),
        ("bzl", Language::Bzl),
        ("xil", Language::Xil),
        ("chc", Language::Chc),
        ("aox", Language::Aox),
        ("giy", Language::Giy),
        ("hoj", Language::Hoj),
        ("kjz", Language::Kjz),
        ("kml", Language::Kml),
        ("und", Language::Und),
        ("xto", Language::Xto),
        ("lml", Language::Lml),
        ("azt", Language::Azt),
        ("yln", Language::Yln),
        ("kcv", Language::Kcv),
        ("sqo", Language::Sqo),
        ("kgx", Language::Kgx),
        ("sky", Language::Sky),
        ("asi", Language::Asi),
        ("gqa", Language::Gqa),
        ("tpp", Language::Tpp),
        ("gol", Language::Gol),
        ("agr", Language::Agr),
        ("ibd", Language::Ibd),
        ("wrs", Language::Wrs),
        ("sva", Language::Sva),
        ("gbw", Language::Gbw),
        ("psc", Language::Psc),
        ("tgu", Language::Tgu),
        ("otw", Language::Otw),
        ("nnm", Language::Nnm),
        ("nbu", Language::Nbu),
        ("awc", Language::Awc),
        ("bgx", Language::Bgx),
        ("tcd", Language::Tcd),
        ("blx", Language::Blx),
        ("xuo", Language::Xuo),
        ("gic", Language::Gic),
        ("sfw", Language::Sfw),
        ("nmq", Language::Nmq),
        ("fun", Language::Fun),
        ("tvk", Language::Tvk),
        ("jkp", Language::Jkp),
        ("iwk", Language::Iwk),
        ("bni", Language::Bni),
        ("cnl", Language::Cnl),
        ("xcb", Language::Xcb),
        ("hao", Language::Hao),
        ("mmv", Language::Mmv),
        ("dum", Language::Dum),
        ("yme", Language::Yme),
        ("sra", Language::Sra),
        ("pyu", Language::Pyu),
        ("ccl", Language::Ccl),
        ("lri", Language::Lri),
        ("nob", Language::Nob),
        ("lrz", Language::Lrz),
        ("mub", Language::Mub),
        ("yah", Language::Yah),
        ("nkc", Language::Nkc),
        ("slk", Language::Slk),
        ("nml", Language::Nml),
        ("cay", Language::Cay),
        ("vie", Language::Vie),
        ("gud", Language::Gud),
        ("qwm", Language::Qwm),
        ("ons", Language::Ons),
        ("koh", Language::Koh),
        ("tmm", Language::Tmm),
        ("kvg", Language::Kvg),
        ("bse", Language::Bse),
        ("etx", Language::Etx),
        ("tgw", Language::Tgw),
        ("lig", Language::Lig),
        ("udg", Language::Udg),
        ("pam", Language::Pam),
        ("ndp", Language::Ndp),
        ("gwt", Language::Gwt),
        ("nhw", Language::Nhw),
        ("hba", Language::Hba),
        ("sps", Language::Sps),
        ("xmj", Language::Xmj),
        ("ywn", Language::Ywn),
        ("aka", Language::Aka),
        ("omy", Language::Omy),
        ("tuc", Language::Tuc),
        ("sdb", Language::Sdb),
        ("gia", Language::Gia),
        ("tca", Language::Tca),
        ("sdu", Language::Sdu),
        ("eke", Language::Eke),
        ("kjb", Language::Kjb),
        ("tfo", Language::Tfo),
        ("ays", Language::Ays),
        ("xbm", Language::Xbm),
        ("vma", Language::Vma),
        ("urx", Language::Urx),
        ("ter", Language::Ter),
        ("dix", Language::Dix),
        ("bnz", Language::Bnz),
        ("nce", Language::Nce),
        ("bqj", Language::Bqj),
        ("kul", Language::Kul),
        ("soh", Language::Soh),
        ("kbs", Language::Kbs),
        ("khq", Language::Khq),
        ("vmb", Language::Vmb),
        ("opm", Language::Opm),
        ("huy", Language::Huy),
        ("fry", Language::Fry),
        ("ytl", Language::Ytl),
        ("jib", Language::Jib),
        ("rav", Language::Rav),
        ("waa", Language::Waa),
        ("sgs", Language::Sgs),
        ("nul", Language::Nul),
        ("gac", Language::Gac),
        ("upi", Language::Upi),
        ("tiz", Language::Tiz),
        ("woe", Language::Woe),
        ("ayl", Language::Ayl),
        ("anr", Language::Anr),
        ("dal", Language::Dal),
        ("niy", Language::Niy),
        ("zpc", Language::Zpc),
        ("lrl", Language::Lrl),
        ("jmr", Language::Jmr),
        ("xbw", Language::Xbw),
        ("zsr", Language::Zsr),
        ("evh", Language::Evh),
        ("tde", Language::Tde),
        ("yaa", Language::Yaa),
        ("oni", Language::Oni),
        ("mag", Language::Mag),
        ("lor", Language::Lor),
        ("plh", Language::Plh),
        ("trl", Language::Trl),
        ("kuq", Language::Kuq),
        ("ssy", Language::Ssy),
        ("thy", Language::Thy),
        ("wmo", Language::Wmo),
        ("rmh", Language::Rmh),
        ("okg", Language::Okg),
        ("rwm", Language::Rwm),
        ("lmn", Language::Lmn),
        ("tmt", Language::Tmt),
        ("mds", Language::Mds),
        ("gct", Language::Gct),
        ("iyx", Language::Iyx),
        ("crc", Language::Crc),
        ("vec", Language::Vec),
        ("aji", Language::Aji),
        ("klr", Language::Klr),
        ("mxd", Language::Mxd),
        ("bin", Language::Bin),
        ("ese", Language::Ese),
        ("akw", Language::Akw),
        ("tsh", Language::Tsh),
        ("grg", Language::Grg),
        ("liw", Language::Liw),
        ("mig", Language::Mig),
        ("fag", Language::Fag),
        ("nbq", Language::Nbq),
        ("ypm", Language::Ypm),
        ("lgr", Language::Lgr),
        ("tha", Language::Tha),
        ("gml", Language::Gml),
        ("ave", Language::Ave),
        ("yur", Language::Yur),
        ("afz", Language::Afz),
        ("ysp", Language::Ysp),
        ("hns", Language::Hns),
        ("xlc", Language::Xlc),
        ("smg", Language::Smg),
        ("tsb", Language::Tsb),
        ("grc", Language::Grc),
        ("rcf", Language::Rcf),
        ("xph", Language::Xph),
        ("bwo", Language::Bwo),
        ("bpm", Language::Bpm),
        ("thn", Language::Thn),
        ("mov", Language::Mov),
        ("mam", Language::Mam),
        ("aio", Language::Aio),
        ("shq", Language::Shq),
        ("nuu", Language::Nuu),
        ("agk", Language::Agk),
        ("xmz", Language::Xmz),
        ("heg", Language::Heg),
        ("mek", Language::Mek),
        ("gon", Language::Gon),
        ("gmz", Language::Gmz),
        ("sax", Language::Sax),
        ("xmf", Language::Xmf),
        ("rgs", Language::Rgs),
        ("jge", Language::Jge),
        ("kqa", Language::Kqa),
        ("ebc", Language::Ebc),
        ("leb", Language::Leb),
        ("xve", Language::Xve),
        ("fbl", Language::Fbl),
        ("ado", Language::Ado),
        ("xln", Language::Xln),
        ("bvv", Language::Bvv),
        ("bwl", Language::Bwl),
        ("muc", Language::Muc),
        ("zha", Language::Zha),
        ("wti", Language::Wti),
        ("kdn", Language::Kdn),
        ("gwm", Language::Gwm),
        ("kbn", Language::Kbn),
        ("kke", Language::Kke),
        ("yax", Language::Yax),
        ("vro", Language::Vro),
        ("sma", Language::Sma),
        ("mhn", Language::Mhn),
        ("mpx", Language::Mpx),
        ("amf", Language::Amf),
        ("pnh", Language::Pnh),
        ("bng", Language::Bng),
        ("mnb", Language::Mnb),
        ("mqm", Language::Mqm),
        ("nkj", Language::Nkj),
        ("mar", Language::Mar),
        ("unk", Language::Unk),
        ("otl", Language::Otl),
        ("xni", Language::Xni),
        ("kep", Language::Kep),
        ("zgn", Language::Zgn),
        ("ctz", Language::Ctz),
        ("mjy", Language::Mjy),
        ("zmj", Language::Zmj),
        ("mxy", Language::Mxy),
        ("ngt", Language::Ngt),
        ("bag", Language::Bag),
        ("blf", Language::Blf),
        ("mtq", Language::Mtq),
        ("dmr", Language::Dmr),
        ("kup", Language::Kup),
        ("rug", Language::Rug),
        ("kyy", Language::Kyy),
        ("rge", Language::Rge),
        ("puf", Language::Puf),
        ("hoc", Language::Hoc),
        ("kka", Language::Kka),
        ("rei", Language::Rei),
        ("cms", Language::Cms),
        ("mog", Language::Mog),
        ("mvp", Language::Mvp),
        ("ltc", Language::Ltc),
        ("itv", Language::Itv),
        ("ylg", Language::Ylg),
        ("prg", Language::Prg),
        ("zhn", Language::Zhn),
        ("fcs", Language::Fcs),
        ("zpu", Language::Zpu),
        ("aee", Language::Aee),
        ("bfr", Language::Bfr),
        ("hix", Language::Hix),
        ("gyn", Language::Gyn),
        ("xkw", Language::Xkw),
        ("tno", Language::Tno),
        ("mbe", Language::Mbe),
        ("zib", Language::Zib),
        ("auo", Language::Auo),
        ("xtg", Language::Xtg),
        ("yeu", Language::Yeu),
        ("enb", Language::Enb),
        ("uky", Language::Uky),
        ("gbu", Language::Gbu),
        ("kic", Language::Kic),
        ("ogg", Language::Ogg),
        ("yul", Language::Yul),
        ("yun", Language::Yun),
        ("dir", Language::Dir),
        ("bbe", Language::Bbe),
        ("kyw", Language::Kyw),
        ("tci", Language::Tci),
        ("cdi", Language::Cdi),
        ("mrc", Language::Mrc),
        ("wbb", Language::Wbb),
        ("wkr", Language::Wkr),
        ("jbe", Language::Jbe),
        ("uge", Language::Uge),
        ("lvk", Language::Lvk),
        ("gma", Language::Gma),
        ("pax", Language::Pax),
        ("amq", Language::Amq),
        ("hwa", Language::Hwa),
        ("usa", Language::Usa),
        ("uka", Language::Uka),
        ("jrt", Language::Jrt),
        ("skc", Language::Skc),
        ("mwu", Language::Mwu),
        ("alz", Language::Alz),
        ("mym", Language::Mym),
        ("tqr", Language::Tqr),
        ("kev", Language::Kev),
        ("knt", Language::Knt),
        ("auh", Language::Auh),
        ("kzv", Language::Kzv),
        ("stm", Language::Stm),
        ("ulw", Language::Ulw),
        ("xct", Language::Xct),
        ("ffm", Language::Ffm),
        ("pln", Language::Pln),
        ("dag", Language::Dag),
        ("awi", Language::Awi),
        ("dbt", Language::Dbt),
        ("igl", Language::Igl),
        ("lui", Language::Lui),
        ("zmu", Language::Zmu),
        ("ypg", Language::Ypg),
        ("tgk", Language::Tgk),
        ("fiw", Language::Fiw),
        ("juy", Language::Juy),
        ("wol", Language::Wol),
        ("ctl", Language::Ctl),
        ("pbl", Language::Pbl),
        ("bcq", Language::Bcq),
        ("dtu", Language::Dtu),
        ("mda", Language::Mda),
        ("bdd", Language::Bdd),
        ("duf", Language::Duf),
        ("scq", Language::Scq),
        ("giz", Language::Giz),
        ("bcn", Language::Bcn),
        ("nhv", Language::Nhv),
        ("dcr", Language::Dcr),
        ("aug", Language::Aug),
        ("lhu", Language::Lhu),
        ("nlv", Language::Nlv),
        ("kfa", Language::Kfa),
        ("agl", Language::Agl),
        ("hed", Language::Hed),
        ("pnc", Language::Pnc),
        ("vmg", Language::Vmg),
        ("loh", Language::Loh),
        ("emp", Language::Emp),
        ("yon", Language::Yon),
        ("pau", Language::Pau),
        ("gbx", Language::Gbx),
        ("kll", Language::Kll),
        ("yaj", Language::Yaj),
        ("jub", Language::Jub),
        ("hoy", Language::Hoy),
        ("lia", Language::Lia),
        ("aac", Language::Aac),
        ("hal", Language::Hal),
        ("orz", Language::Orz),
        ("hmr", Language::Hmr),
        ("bfl", Language::Bfl),
        ("kkc", Language::Kkc),
        ("dnk", Language::Dnk),
        ("pah", Language::Pah),
        ("sym", Language::Sym),
        ("pkb", Language::Pkb),
        ("pnj", Language::Pnj),
        ("pbo", Language::Pbo),
        ("ass", Language::Ass),
        ("dza", Language::Dza),
        ("jng", Language::Jng),
        ("sde", Language::Sde),
        ("mxl", Language::Mxl),
        ("cnr", Language::Cnr),
        ("pld", Language::Pld),
        ("ldo", Language::Ldo),
        ("kqz", Language::Kqz),
        ("lgb", Language::Lgb),
        ("oka", Language::Oka),
        ("dji", Language::Dji),
        ("gnn", Language::Gnn),
        ("pho", Language::Pho),
        ("kkk", Language::Kkk),
        ("hrx", Language::Hrx),
        ("suy", Language::Suy),
        ("fgr", Language::Fgr),
        ("zal", Language::Zal),
        ("guq", Language::Guq),
        ("mjj", Language::Mjj),
        ("ctu", Language::Ctu),
        ("zhb", Language::Zhb),
        ("nop", Language::Nop),
        ("pif", Language::Pif),
        ("brg", Language::Brg),
        ("yzg", Language::Yzg),
        ("arz", Language::Arz),
        ("bma", Language::Bma),
        ("boh", Language::Boh),
        ("ndy", Language::Ndy),
        ("dde", Language::Dde),
        ("mxz", Language::Mxz),
        ("bfd", Language::Bfd),
        ("asm", Language::Asm),
        ("tte", Language::Tte),
        ("szp", Language::Szp),
        ("dul", Language::Dul),
        ("ths", Language::Ths),
        ("nlm", Language::Nlm),
        ("siv", Language::Siv),
        ("lky", Language::Lky),
        ("bpr", Language::Bpr),
        ("deu", Language::Deu),
        ("wir", Language::Wir),
        ("bgf", Language::Bgf),
        ("wky", Language::Wky),
        ("pue", Language::Pue),
        ("otu", Language::Otu),
        ("wnn", Language::Wnn),
        ("msx", Language::Msx),
        ("sro", Language::Sro),
        ("kem", Language::Kem),
        ("rys", Language::Rys),
        ("aot", Language::Aot),
        ("seu", Language::Seu),
        ("teh", Language::Teh),
        ("tap", Language::Tap),
        ("wae", Language::Wae),
        ("njs", Language::Njs),
        ("nyx", Language::Nyx),
        ("xwo", Language::Xwo),
        ("swq", Language::Swq),
        ("xet", Language::Xet),
        ("doa", Language::Doa),
        ("wlv", Language::Wlv),
        ("gwf", Language::Gwf),
        ("tvd", Language::Tvd),
        ("phl", Language::Phl),
        ("mtr", Language::Mtr),
        ("sti", Language::Sti),
        ("kdu", Language::Kdu),
        ("hin", Language::Hin),
        ("apz", Language::Apz),
        ("tdm", Language::Tdm),
        ("soi", Language::Soi),
        ("bst", Language::Bst),
        ("dis", Language::Dis),
        ("ycn", Language::Ycn),
        ("lez", Language::Lez),
        ("bxi", Language::Bxi),
        ("llp", Language::Llp),
        ("jam", Language::Jam),
        ("qva", Language::Qva),
        ("kpy", Language::Kpy),
        ("ukq", Language::Ukq),
        ("alj", Language::Alj),
        ("ysd", Language::Ysd),
        ("lag", Language::Lag),
        ("qya", Language::Qya),
        ("ynn", Language::Ynn),
        ("bep", Language::Bep),
        ("yxg", Language::Yxg),
        ("iwm", Language::Iwm),
        ("opy", Language::Opy),
        ("gne", Language::Gne),
        ("nbe", Language::Nbe),
        ("xug", Language::Xug),
        ("bsp", Language::Bsp),
        ("xoo", Language::Xoo),
        ("guc", Language::Guc),
        ("pem", Language::Pem),
        ("vsv", Language::Vsv),
        ("trc", Language::Trc),
        ("csl", Language::Csl),
        ("bie", Language::Bie),
        ("aqg", Language::Aqg),
        ("kgr", Language::Kgr),
        ("csg", Language::Csg),
        ("akz", Language::Akz),
        ("jjr", Language::Jjr),
        ("chd", Language::Chd),
        ("mxm", Language::Mxm),
        ("azz", Language::Azz),
        ("qvo", Language::Qvo),
        ("vmx", Language::Vmx),
        ("pci", Language::Pci),
        ("wle", Language::Wle),
        ("yky", Language::Yky),
        ("gyz", Language::Gyz),
        ("kon", Language::Kon),
        ("afu", Language::Afu),
        ("fak", Language::Fak),
        ("sdp", Language::Sdp),
        ("kjr", Language::Kjr),
        ("ids", Language::Ids),
        ("laz", Language::Laz),
        ("ame", Language::Ame),
        ("koa", Language::Koa),
        ("mni", Language::Mni),
        ("pmn", Language::Pmn),
        ("acr", Language::Acr),
        ("lbs", Language::Lbs),
        ("mkl", Language::Mkl),
        ("onn", Language::Onn),
        ("gvj", Language::Gvj),
        ("btu", Language::Btu),
        ("phr", Language::Phr),
        ("myr", Language::Myr),
        ("ztq", Language::Ztq),
        ("mru", Language::Mru),
        ("bye", Language::Bye),
        ("ble", Language::Ble),
        ("syy", Language::Syy),
        ("gvl", Language::Gvl),
        ("myx", Language::Myx),
        ("trq", Language::Trq),
        ("kce", Language::Kce),
        ("dgt", Language::Dgt),
        ("gaq", Language::Gaq),
        ("mom", Language::Mom),
        ("kso", Language::Kso),
        ("aun", Language::Aun),
        ("top", Language::Top),
        ("pox", Language::Pox),
        ("mho", Language::Mho),
        ("nsl", Language::Nsl),
        ("bvb", Language::Bvb),
        ("xac", Language::Xac),
        ("lmh", Language::Lmh),
        ("eky", Language::Eky),
        ("veo", Language::Veo),
        ("gho", Language::Gho),
        ("lwa", Language::Lwa),
        ("mww", Language::Mww),
        ("yyz", Language::Yyz),
        ("zkp", Language::Zkp),
        ("chu", Language::Chu),
        ("tmr", Language::Tmr),
        ("hot", Language::Hot),
        ("kcp", Language::Kcp),
        ("kof", Language::Kof),
        ("stw", Language::Stw),
        ("maz", Language::Maz),
        ("lbk", Language::Lbk),
        ("gsm", Language::Gsm),
        ("bts", Language::Bts),
        ("swy", Language::Swy),
        ("rwa", Language::Rwa),
        ("rap", Language::Rap),
        ("gwi", Language::Gwi),
        ("tdo", Language::Tdo),
        ("msc", Language::Msc),
        ("kkw", Language::Kkw),
        ("mxo", Language::Mxo),
        ("nma", Language::Nma),
        ("otq", Language::Otq),
        ("mxi", Language::Mxi),
        ("bmp", Language::Bmp),
        ("hwc", Language::Hwc),
        ("ino", Language::Ino),
        ("bvk", Language::Bvk),
        ("jhi", Language::Jhi),
        ("zmi", Language::Zmi),
        ("pud", Language::Pud),
        ("oso", Language::Oso),
        ("bmg", Language::Bmg),
        ("mgo", Language::Mgo),
        ("bpv", Language::Bpv),
        ("doi", Language::Doi),
        ("bid", Language::Bid),
        ("klc", Language::Klc),
        ("pgk", Language::Pgk),
        ("ykg", Language::Ykg),
        ("arg", Language::Arg),
        ("frm", Language::Frm),
        ("bfq", Language::Bfq),
        ("lnj", Language::Lnj),
        ("blc", Language::Blc),
        ("sls", Language::Sls),
        ("hya", Language::Hya),
        ("tiw", Language::Tiw),
        ("urz", Language::Urz),
        ("swc", Language::Swc),
        ("cpn", Language::Cpn),
        ("lbv", Language::Lbv),
        ("kin", Language::Kin),
        ("six", Language::Six),
        ("mmb", Language::Mmb),
        ("sdo", Language::Sdo),
        ("ilv", Language::Ilv),
        ("coh", Language::Coh),
        ("msr", Language::Msr),
        ("cas", Language::Cas),
        ("ldl", Language::Ldl),
        ("fsl", Language::Fsl),
        ("ttw", Language::Ttw),
        ("toc", Language::Toc),
        ("sbj", Language::Sbj),
        ("xth", Language::Xth),
        ("kpc", Language::Kpc),
        ("old", Language::Old),
        ("lhs", Language::Lhs),
        ("brq", Language::Brq),
        ("ssi", Language::Ssi),
        ("iby", Language::Iby),
        ("srg", Language::Srg),
        ("svs", Language::Svs),
        ("cpy", Language::Cpy),
        ("lgt", Language::Lgt),
        ("cnx", Language::Cnx),
        ("obm", Language::Obm),
        ("xcg", Language::Xcg),
        ("cbd", Language::Cbd),
        ("ylu", Language::Ylu),
        ("lrc", Language::Lrc),
        ("ssx", Language::Ssx),
        ("bjj", Language::Bjj),
        ("noc", Language::Noc),
        ("mjd", Language::Mjd),
        ("zhd", Language::Zhd),
        ("spt", Language::Spt),
        ("ldh", Language::Ldh),
        ("mii", Language::Mii),
        ("bsa", Language::Bsa),
        ("rmq", Language::Rmq),
        ("gyf", Language::Gyf),
        ("sgb", Language::Sgb),
        ("adi", Language::Adi),
        ("ajs", Language::Ajs),
        ("ptp", Language::Ptp),
        ("ssg", Language::Ssg),
        ("cui", Language::Cui),
        ("jnd", Language::Jnd),
        ("mra", Language::Mra),
        ("cip", Language::Cip),
        ("rnb", Language::Rnb),
        ("tqb", Language::Tqb),
        ("zmt", Language::Zmt),
        ("ona", Language::Ona),
        ("duz", Language::Duz),
        ("twy", Language::Twy),
        ("cav", Language::Cav),
        ("fuj", Language::Fuj),
        ("mfq", Language::Mfq),
        ("adn", Language::Adn),
        ("bar", Language::Bar),
        ("lcc", Language::Lcc),
        ("glc", Language::Glc),
        ("bvf", Language::Bvf),
        ("ost", Language::Ost),
        ("zmv", Language::Zmv),
        ("omi", Language::Omi),
        ("kpx", Language::Kpx),
        ("shr", Language::Shr),
        ("asr", Language::Asr),
        ("gjk", Language::Gjk),
        ("wng", Language::Wng),
        ("bcu", Language::Bcu),
        ("ckv", Language::Ckv),
        ("trh", Language::Trh),
        ("som", Language::Som),
        ("kax", Language::Kax),
        ("byv", Language::Byv),
        ("gbi", Language::Gbi),
        ("dur", Language::Dur),
        ("mnu", Language::Mnu),
        ("lir", Language::Lir),
        ("mpn", Language::Mpn),
        ("alc", Language::Alc),
        ("dva", Language::Dva),
        ("eus", Language::Eus),
        ("hio", Language::Hio),
        ("pjt", Language::Pjt),
        ("muz", Language::Muz),
        ("xww", Language::Xww),
        ("wua", Language::Wua),
        ("ktz", Language::Ktz),
        ("sah", Language::Sah),
        ("yua", Language::Yua),
        ("zea", Language::Zea),
        ("gld", Language::Gld),
        ("nqt", Language::Nqt),
        ("szl", Language::Szl),
        ("bmr", Language::Bmr),
        ("pdi", Language::Pdi),
        ("sin", Language::Sin),
        ("wym", Language::Wym),
        ("mdd", Language::Mdd),
        ("los", Language::Los),
        ("ets", Language::Ets),
        ("sco", Language::Sco),
        ("nkr", Language::Nkr),
        ("wum", Language::Wum),
        ("agu", Language::Agu),
        ("bjh", Language::Bjh),
        ("swj", Language::Swj),
        ("tui", Language::Tui),
        ("pmd", Language::Pmd),
        ("pmh", Language::Pmh),
        ("nhr", Language::Nhr),
        ("ttp", Language::Ttp),
        ("nit", Language::Nit),
        ("crh", Language::Crh),
        ("aem", Language::Aem),
        ("nhk", Language::Nhk),
        ("sjs", Language::Sjs),
        ("kyx", Language::Kyx),
        ("ptt", Language::Ptt),
        ("dmx", Language::Dmx),
        ("ity", Language::Ity),
        ("afn", Language::Afn),
        ("kdr", Language::Kdr),
        ("kbh", Language::Kbh),
        ("ayr", Language::Ayr),
        ("khl", Language::Khl),
        ("bdp", Language::Bdp),
        ("xpk", Language::Xpk),
        ("pno", Language::Pno),
        ("mwl", Language::Mwl),
        ("spv", Language::Spv),
        ("ybo", Language::Ybo),
        ("kyb", Language::Kyb),
        ("myg", Language::Myg),
        ("kot", Language::Kot),
        ("bea", Language::Bea),
        ("zar", Language::Zar),
        ("bgi", Language::Bgi),
        ("hut", Language::Hut),
        ("uni", Language::Uni),
        ("chg", Language::Chg),
        ("jvn", Language::Jvn),
        ("dig", Language::Dig),
        ("dyi", Language::Dyi),
        ("tke", Language::Tke),
        ("hmm", Language::Hmm),
        ("aly", Language::Aly),
        ("uve", Language::Uve),
        ("pnb", Language::Pnb),
        ("mcg", Language::Mcg),
        ("kss", Language::Kss),
        ("dua", Language::Dua),
        ("mef", Language::Mef),
        ("caw", Language::Caw),
        ("ain", Language::Ain),
        ("tnr", Language::Tnr),
        ("atp", Language::Atp),
        ("sch", Language::Sch),
        ("kfq", Language::Kfq),
        ("cua", Language::Cua),
        ("lka", Language::Lka),
        ("gdk", Language::Gdk),
        ("xer", Language::Xer),
        ("avm", Language::Avm),
        ("tnu", Language::Tnu),
        ("onx", Language::Onx),
        ("rnd", Language::Rnd),
        ("mss", Language::Mss),
        ("stg", Language::Stg),
        ("xwa", Language::Xwa),
        ("aph", Language::Aph),
        ("yms", Language::Yms),
        ("csf", Language::Csf),
        ("quh", Language::Quh),
        ("tos", Language::Tos),
        ("psa", Language::Psa),
        ("cda", Language::Cda),
        ("shc", Language::Shc),
        ("qvh", Language::Qvh),
        ("muj", Language::Muj),
        ("qxt", Language::Qxt),
        ("thp", Language::Thp),
        ("lwu", Language::Lwu),
        ("vkm", Language::Vkm),
        ("kkr", Language::Kkr),
        ("wlm", Language::Wlm),
        ("juw", Language::Juw),
        ("yog", Language::Yog),
        ("btm", Language::Btm),
        ("slr", Language::Slr),
        ("pai", Language::Pai),
        ("kgo", Language::Kgo),
        ("mqc", Language::Mqc),
        ("njj", Language::Njj),
        ("xxb", Language::Xxb),
        ("ton", Language::Ton),
        ("buz", Language::Buz),
        ("won", Language::Won),
        ("tnn", Language::Tnn),
        ("csy", Language::Csy),
        ("dyd", Language::Dyd),
        ("tlb", Language::Tlb),
        ("xok", Language::Xok),
        ("lsi", Language::Lsi),
        ("xar", Language::Xar),
        ("jaj", Language::Jaj),
        ("jbi", Language::Jbi),
        ("xon", Language::Xon),
        ("dkg", Language::Dkg),
        ("kqh", Language::Kqh),
        ("que", Language::Que),
        ("yre", Language::Yre),
        ("eri", Language::Eri),
        ("tug", Language::Tug),
        ("kiw", Language::Kiw),
        ("kpn", Language::Kpn),
        ("ems", Language::Ems),
        ("iai", Language::Iai),
        ("xno", Language::Xno),
        ("bnn", Language::Bnn),
        ("aln", Language::Aln),
        ("afh", Language::Afh),
        ("xsa", Language::Xsa),
        ("drs", Language::Drs),
        ("iku", Language::Iku),
        ("aez", Language::Aez),
        ("okh", Language::Okh),
        ("uru", Language::Uru),
        ("www", Language::Www),
        ("nav", Language::Nav),
        ("slg", Language::Slg),
        ("ldj", Language::Ldj),
        ("typ", Language::Typ),
        ("ltz", Language::Ltz),
        ("nvo", Language::Nvo),
        ("yac", Language::Yac),
        ("uby", Language::Uby),
        ("smc", Language::Smc),
        ("lla", Language::Lla),
        ("bmw", Language::Bmw),
        ("yok", Language::Yok),
        ("asa", Language::Asa),
        ("lkt", Language::Lkt),
        ("swv", Language::Swv),
        ("tpr", Language::Tpr),
        ("gur", Language::Gur),
        ("btg", Language::Btg),
        ("xse", Language::Xse),
        ("igs", Language::Igs),
        ("nmg", Language::Nmg),
        ("blj", Language::Blj),
        ("toy", Language::Toy),
        ("klg", Language::Klg),
        ("dok", Language::Dok),
        ("nka", Language::Nka),
        ("abd", Language::Abd),
        ("epi", Language::Epi),
        ("clw", Language::Clw),
        ("ruu", Language::Ruu),
        ("lmr", Language::Lmr),
        ("ruo", Language::Ruo),
        ("pef", Language::Pef),
        ("bet", Language::Bet),
        ("zao", Language::Zao),
        ("aqc", Language::Aqc),
        ("kky", Language::Kky),
        ("mcb", Language::Mcb),
        ("udl", Language::Udl),
        ("nms", Language::Nms),
        ("cor", Language::Cor),
        ("swm", Language::Swm),
        ("ife", Language::Ife),
        ("sst", Language::Sst),
        ("xtv", Language::Xtv),
        ("rwk", Language::Rwk),
        ("cly", Language::Cly),
        ("awr", Language::Awr),
        ("ifa", Language::Ifa),
        ("ywr", Language::Ywr),
        ("mto", Language::Mto),
        ("phh", Language::Phh),
        ("apw", Language::Apw),
        ("wmd", Language::Wmd),
        ("xpy", Language::Xpy),
        ("kye", Language::Kye),
        ("mmz", Language::Mmz),
        ("nal", Language::Nal),
        ("nof", Language::Nof),
        ("tmj", Language::Tmj),
        ("klv", Language::Klv),
        ("bhd", Language::Bhd),
        ("lwg", Language::Lwg),
        ("lgk", Language::Lgk),
        ("nps", Language::Nps),
        ("mme", Language::Mme),
        ("heh", Language::Heh),
        ("cac", Language::Cac),
        ("kng", Language::Kng),
        ("uiv", Language::Uiv),
        ("nif", Language::Nif),
        ("dks", Language::Dks),
        ("nhd", Language::Nhd),
        ("tjl", Language::Tjl),
        ("xsi", Language::Xsi),
        ("thv", Language::Thv),
        ("mbq", Language::Mbq),
        ("yia", Language::Yia),
        ("bcy", Language::Bcy),
        ("lul", Language::Lul),
        ("cog", Language::Cog),
        ("ssu", Language::Ssu),
        ("wiv", Language::Wiv),
        ("yel", Language::Yel),
        ("hrt", Language::Hrt),
        ("xpx", Language::Xpx),
        ("wos", Language::Wos),
        ("luy", Language::Luy),
        ("enu", Language::Enu),
        ("bqr", Language::Bqr),
        ("agx", Language::Agx),
        ("mdn", Language::Mdn),
        ("ajg", Language::Ajg),
        ("tmn", Language::Tmn),
        ("gqr", Language::Gqr),
        ("lgz", Language::Lgz),
        ("taw", Language::Taw),
        ("rmx", Language::Rmx),
        ("gdd", Language::Gdd),
        ("mgc", Language::Mgc),
        ("max", Language::Max),
        ("enh", Language::Enh),
        ("nim", Language::Nim),
        ("nbs", Language::Nbs),
        ("tmu", Language::Tmu),
        ("clk", Language::Clk),
        ("xny", Language::Xny),
        ("tgz", Language::Tgz),
        ("kjg", Language::Kjg),
        ("qvp", Language::Qvp),
        ("knw", Language::Knw),
        ("krt", Language::Krt),
        ("mfu", Language::Mfu),
        ("mxr", Language::Mxr),
        ("vmq", Language::Vmq),
        ("xua", Language::Xua),
        ("kcf", Language::Kcf),
        ("ttf", Language::Ttf),
        ("bgt", Language::Bgt),
        ("xut", Language::Xut),
        ("tex", Language::Tex),
        ("sju", Language::Sju),
        ("knx", Language::Knx),
        ("doy", Language::Doy),
        ("svm", Language::Svm),
        ("ywt", Language::Ywt),
        ("phw", Language::Phw),
        ("agz", Language::Agz),
        ("gaw", Language::Gaw),
        ("lai", Language::Lai),
        ("aor", Language::Aor),
        ("baj", Language::Baj),
        ("muk", Language::Muk),
        ("lni", Language::Lni),
        ("snv", Language::Snv),
        ("adr", Language::Adr),
        ("tsu", Language::Tsu),
        ("csm", Language::Csm),
        ("btz", Language::Btz),
        ("rgn", Language::Rgn),
        ("awx", Language::Awx),
        ("wea", Language::Wea),
        ("lah", Language::Lah),
        ("myu", Language::Myu),
        ("cao", Language::Cao),
        ("gbo", Language::Gbo),
        ("ren", Language::Ren),
        ("wbp", Language::Wbp),
        ("lpn", Language::Lpn),
        ("rod", Language::Rod),
        ("kpm", Language::Kpm),
        ("snd", Language::Snd),
        ("zpr", Language::Zpr),
        ("vsl", Language::Vsl),
        ("wlc", Language::Wlc),
        ("gaf", Language::Gaf),
        ("mmh", Language::Mmh),
        ("liq", Language::Liq),
        ("ttv", Language::Ttv),
        ("bqn", Language::Bqn),
        ("zmn", Language::Zmn),
        ("nsn", Language::Nsn),
        ("gnk", Language::Gnk),
        ("wgi", Language::Wgi),
        ("bhu", Language::Bhu),
        ("nzd", Language::Nzd),
        ("kny", Language::Kny),
        ("kcn", Language::Kcn),
        ("cdo", Language::Cdo),
        ("dih", Language::Dih),
        ("ghk", Language::Ghk),
        ("bjf", Language::Bjf),
        ("yal", Language::Yal),
        ("ulb", Language::Ulb),
        ("pan", Language::Pan),
        ("bky", Language::Bky),
        ("cga", Language::Cga),
        ("loa", Language::Loa),
        ("nll", Language::Nll),
        ("dtb", Language::Dtb),
        ("tel", Language::Tel),
        ("eya", Language::Eya),
        ("ubl", Language::Ubl),
        ("ylo", Language::Ylo),
        ("vku", Language::Vku),
        ("lmq", Language::Lmq),
        ("cea", Language::Cea),
        ("jmi", Language::Jmi),
        ("bca", Language::Bca),
        ("gsn", Language::Gsn),
        ("ztx", Language::Ztx),
        ("sqx", Language::Sqx),
        ("pcw", Language::Pcw),
        ("cli", Language::Cli),
        ("wgy", Language::Wgy),
        ("but", Language::But),
        ("ypb", Language::Ypb),
        ("lac", Language::Lac),
        ("grn", Language::Grn),
        ("zrs", Language::Zrs),
        ("asb", Language::Asb),
        ("bjv", Language::Bjv),
        ("wax", Language::Wax),
        ("ctn", Language::Ctn),
        ("tmy", Language::Tmy),
        ("jgb", Language::Jgb),
        ("naf", Language::Naf),
        ("qvm", Language::Qvm),
        ("scw", Language::Scw),
        ("bik", Language::Bik),
        ("mng", Language::Mng),
        ("sng", Language::Sng),
        ("kfy", Language::Kfy),
        ("hne", Language::Hne),
        ("hoi", Language::Hoi),
        ("tnl", Language::Tnl),
        ("ngy", Language::Ngy),
        ("yig", Language::Yig),
        ("lvi", Language::Lvi),
        ("kdt", Language::Kdt),
        ("ote", Language::Ote),
        ("mrs", Language::Mrs),
        ("cll", Language::Cll),
        ("mta", Language::Mta),
        ("ril", Language::Ril),
        ("kva", Language::Kva),
        ("lat", Language::Lat),
        ("xbp", Language::Xbp),
        ("plk", Language::Plk),
        ("fal", Language::Fal),
        ("mjq", Language::Mjq),
        ("bdg", Language::Bdg),
        ("oda", Language::Oda),
        ("olm", Language::Olm),
        ("vmp", Language::Vmp),
        ("kht", Language::Kht),
        ("mvg", Language::Mvg),
        ("tyi", Language::Tyi),
        ("gso", Language::Gso),
        ("fln", Language::Fln),
        ("kzg", Language::Kzg),
        ("ibl", Language::Ibl),
        ("bct", Language::Bct),
        ("nlg", Language::Nlg),
        ("quc", Language::Quc),
        ("ywq", Language::Ywq),
        ("mcx", Language::Mcx),
        ("mfz", Language::Mfz),
        ("bmu", Language::Bmu),
        ("del", Language::Del),
        ("gvr", Language::Gvr),
        ("boy", Language::Boy),
        ("kzo", Language::Kzo),
        ("hml", Language::Hml),
        ("nno", Language::Nno),
        ("ypp", Language::Ypp),
        ("nia", Language::Nia),
        ("xeu", Language::Xeu),
        ("syc", Language::Syc),
        ("lso", Language::Lso),
        ("huk", Language::Huk),
        ("gap", Language::Gap),
        ("kxd", Language::Kxd),
        ("kvx", Language::Kvx),
        ("ngb", Language::Ngb),
        ("dge", Language::Dge),
        ("afb", Language::Afb),
        ("lnn", Language::Lnn),
        ("kdg", Language::Kdg),
        ("kgs", Language::Kgs),
        ("woo", Language::Woo),
        ("blk", Language::Blk),
        ("trw", Language::Trw),
        ("kxv", Language::Kxv),
        ("tfn", Language::Tfn),
        ("slw", Language::Slw),
        ("ntx", Language::Ntx),
        ("zbe", Language::Zbe),
        ("txj", Language::Txj),
        ("rmc", Language::Rmc),
        ("ssj", Language::Ssj),
        ("wur", Language::Wur),
        ("bbh", Language::Bbh),
        ("zhw", Language::Zhw),
        ("szg", Language::Szg),
        ("kkb", Language::Kkb),
        ("lbc", Language::Lbc),
        ("akj", Language::Akj),
        ("ekl", Language::Ekl),
        ("kat", Language::Kat),
        ("syx", Language::Syx),
        ("taj", Language::Taj),
        ("lsm", Language::Lsm),
        ("rbp", Language::Rbp),
        ("pmw", Language::Pmw),
        ("nri", Language::Nri),
        ("ors", Language::Ors),
        ("azo", Language::Azo),
        ("bhq", Language::Bhq),
        ("dim", Language::Dim),
        ("kgq", Language::Kgq),
        ("kvz", Language::Kvz),
        ("dgk", Language::Dgk),
        ("shp", Language::Shp),
        ("pao", Language::Pao),
        ("bkt", Language::Bkt),
        ("rsb", Language::Rsb),
        ("tae", Language::Tae),
        ("pha", Language::Pha),
        ("kze", Language::Kze),
        ("sgu", Language::Sgu),
        ("dhx", Language::Dhx),
        ("lfa", Language::Lfa),
        ("fng", Language::Fng),
        ("szb", Language::Szb),
        ("aky", Language::Aky),
        ("iru", Language::Iru),
        ("cto", Language::Cto),
        ("wud", Language::Wud),
        ("sms", Language::Sms),
        ("dmu", Language::Dmu),
        ("nnk", Language::Nnk),
        ("woy", Language::Woy),
        ("ikl", Language::Ikl),
        ("luf", Language::Luf),
        ("mwc", Language::Mwc),
        ("kmm", Language::Kmm),
        ("wtw", Language::Wtw),
        ("bwm", Language::Bwm),
        ("tio", Language::Tio),
        ("yup", Language::Yup),
        ("kaj", Language::Kaj),
        ("cuq", Language::Cuq),
        ("tig", Language::Tig),
        ("hmw", Language::Hmw),
        ("xad", Language::Xad),
        ("fro", Language::Fro),
        ("wry", Language::Wry),
        ("xpt", Language::Xpt),
        ("cul", Language::Cul),
        ("mmp", Language::Mmp),
        ("gox", Language::Gox),
        ("puc", Language::Puc),
        ("pur", Language::Pur),
        ("zaw", Language::Zaw),
        ("wbj", Language::Wbj),
        ("tdj", Language::Tdj),
        ("nqm", Language::Nqm),
        ("mdz", Language::Mdz),
        ("tkd", Language::Tkd),
        ("yng", Language::Yng),
        ("fit", Language::Fit),
        ("chp", Language::Chp),
        ("bfm", Language::Bfm),
        ("pus", Language::Pus),
        ("hit", Language::Hit),
        ("aye", Language::Aye),
        ("mqr", Language::Mqr),
        ("doo", Language::Doo),
        ("tvn", Language::Tvn),
        ("ttd", Language::Ttd),
        ("xqt", Language::Xqt),
        ("tbr", Language::Tbr),
        ("xaa", Language::Xaa),
        ("gue", Language::Gue),
        ("luv", Language::Luv),
        ("hig", Language::Hig),
        ("ocu", Language::Ocu),
        ("lub", Language::Lub),
        ("nqo", Language::Nqo),
        ("plr", Language::Plr),
        ("ghl", Language::Ghl),
        ("plz", Language::Plz),
        ("bxs", Language::Bxs),
        ("fam", Language::Fam),
        ("kun", Language::Kun),
        ("bxj", Language::Bxj),
        ("wlh", Language::Wlh),
        ("mka", Language::Mka),
        ("lbz", Language::Lbz),
        ("bxn", Language::Bxn),
        ("bxc", Language::Bxc),
        ("tgt", Language::Tgt),
        ("aml", Language::Aml),
        ("mml", Language::Mml),
        ("tjj", Language::Tjj),
        ("mer", Language::Mer),
        ("uam", Language::Uam),
        ("cgk", Language::Cgk),
        ("pmo", Language::Pmo),
        ("sey", Language::Sey),
        ("mas", Language::Mas),
        ("ldq", Language::Ldq),
        ("mpe", Language::Mpe),
        ("pie", Language::Pie),
        ("nyc", Language::Nyc),
        ("uuu", Language::Uuu),
        ("nxg", Language::Nxg),
        ("mfv", Language::Mfv),
        ("ckb", Language::Ckb),
        ("sxm", Language::Sxm),
        ("sxr", Language::Sxr),
        ("lkc", Language::Lkc),
        ("mlr", Language::Mlr),
        ("pui", Language::Pui),
        ("kgi", Language::Kgi),
        ("xrm", Language::Xrm),
        ("iws", Language::Iws),
        ("ega", Language::Ega),
        ("dop", Language::Dop),
        ("nyn", Language::Nyn),
        ("wmn", Language::Wmn),
        ("kph", Language::Kph),
        ("ibu", Language::Ibu),
        ("res", Language::Res),
        ("xkn", Language::Xkn),
        ("jmc", Language::Jmc),
        ("kqg", Language::Kqg),
        ("pbe", Language::Pbe),
        ("enf", Language::Enf),
        ("sje", Language::Sje),
        ("nke", Language::Nke),
        ("bdy", Language::Bdy),
        ("lus", Language::Lus),
        ("ssz", Language::Ssz),
        ("eja", Language::Eja),
        ("xzp", Language::Xzp),
        ("sse", Language::Sse),
        ("vao", Language::Vao),
        ("daa", Language::Daa),
        ("mjr", Language::Mjr),
        ("mvw", Language::Mvw),
        ("knp", Language::Knp),
        ("bui", Language::Bui),
        ("xtu", Language::Xtu),
        ("yih", Language::Yih),
        ("bpg", Language::Bpg),
        ("dsl", Language::Dsl),
        ("lma", Language::Lma),
        ("kgf", Language::Kgf),
        ("mbj", Language::Mbj),
        ("wav", Language::Wav),
        ("jmx", Language::Jmx),
        ("xha", Language::Xha),
        ("znk", Language::Znk),
        ("ngi", Language::Ngi),
        ("xkb", Language::Xkb),
        ("sct", Language::Sct),
        ("okz", Language::Okz),
        ("kvc", Language::Kvc),
        ("tvx", Language::Tvx),
        ("mxf", Language::Mxf),
        ("dry", Language::Dry),
        ("fqs", Language::Fqs),
        ("yox", Language::Yox),
        ("ind", Language::Ind),
        ("shg", Language::Shg),
        ("oss", Language::Oss),
        ("sku", Language::Sku),
        ("cux", Language::Cux),
        ("kqx", Language::Kqx),
        ("ide", Language::Ide),
        ("mjs", Language::Mjs),
        ("mku", Language::Mku),
        ("mur", Language::Mur),
        ("ena", Language::Ena),
        ("ajn", Language::Ajn),
        ("bhr", Language::Bhr),
        ("bgk", Language::Bgk),
        ("bah", Language::Bah),
        ("wgu", Language::Wgu),
        ("bhw", Language::Bhw),
        ("aua", Language::Aua),
        ("ndb", Language::Ndb),
        ("div", Language::Div),
        ("rdb", Language::Rdb),
        ("lof", Language::Lof),
        ("rwo", Language::Rwo),
        ("faz", Language::Faz),
        ("nfd", Language::Nfd),
        ("pkp", Language::Pkp),
        ("mhb", Language::Mhb),
        ("clu", Language::Clu),
        ("wro", Language::Wro),
        ("gau", Language::Gau),
        ("bln", Language::Bln),
        ("pgg", Language::Pgg),
        ("myl", Language::Myl),
        ("prd", Language::Prd),
        ("def", Language::Def),
        ("yid", Language::Yid),
        ("khy", Language::Khy),
        ("acu", Language::Acu),
        ("myy", Language::Myy),
        ("bkw", Language::Bkw),
        ("aof", Language::Aof),
        ("kcw", Language::Kcw),
        ("haw", Language::Haw),
        ("xtm", Language::Xtm),
        ("bde", Language::Bde),
        ("apk", Language::Apk),
        ("vrt", Language::Vrt),
        ("bps", Language::Bps),
        ("dbg", Language::Dbg),
        ("hni", Language::Hni),
        ("hul", Language::Hul),
        ("tpw", Language::Tpw),
        ("cjk", Language::Cjk),
        ("nby", Language::Nby),
        ("ndj", Language::Ndj),
        ("ita", Language::Ita),
        ("fil", Language::Fil),
        ("uvh", Language::Uvh),
        ("jun", Language::Jun),
        ("ymx", Language::Ymx),
        ("jei", Language::Jei),
        ("qvl", Language::Qvl),
        ("kbk", Language::Kbk),
        ("bxe", Language::Bxe),
        ("arh", Language::Arh),
        ("isg", Language::Isg),
        ("drn", Language::Drn),
        ("kiu", Language::Kiu),
        ("lan", Language::Lan),
        ("gyg", Language::Gyg),
        ("biz", Language::Biz),
        ("mcr", Language::Mcr),
        ("fum", Language::Fum),
        ("ani", Language::Ani),
        ("kwr", Language::Kwr),
        ("rar", Language::Rar),
        ("goz", Language::Goz),
        ("ybh", Language::Ybh),
        ("fif", Language::Fif),
        ("glh", Language::Glh),
        ("xwd", Language::Xwd),
        ("vrs", Language::Vrs),
        ("zam", Language::Zam),
        ("gog", Language::Gog),
        ("ale", Language::Ale),
        ("llq", Language::Llq),
        ("kag", Language::Kag),
        ("bwx", Language::Bwx),
        ("sdc", Language::Sdc),
        ("tla", Language::Tla),
        ("dol", Language::Dol),
        ("wic", Language::Wic),
        ("nhy", Language::Nhy),
        ("mpr", Language::Mpr),
        ("kpl", Language::Kpl),
        ("xkx", Language::Xkx),
        ("mmf", Language::Mmf),
        ("tkz", Language::Tkz),
        ("tbg", Language::Tbg),
        ("tim", Language::Tim),
        ("rey", Language::Rey),
        ("kdi", Language::Kdi),
        ("jbw", Language::Jbw),
        ("qxn", Language::Qxn),
        ("bzd", Language::Bzd),
        ("soc", Language::Soc),
        ("kwf", Language::Kwf),
        ("aqt", Language::Aqt),
        ("lya", Language::Lya),
        ("raa", Language::Raa),
        ("ayo", Language::Ayo),
        ("nih", Language::Nih),
        ("mpz", Language::Mpz),
        ("ann", Language::Ann),
        ("jav", Language::Jav),
        ("ike", Language::Ike),
        ("ano", Language::Ano),
        ("pkr", Language::Pkr),
        ("zsu", Language::Zsu),
        ("xkq", Language::Xkq),
        ("trv", Language::Trv),
        ("kcc", Language::Kcc),
        ("tdv", Language::Tdv),
        ("bmb", Language::Bmb),
        ("bcw", Language::Bcw),
        ("irn", Language::Irn),
        ("sdl", Language::Sdl),
        ("ije", Language::Ije),
        ("kop", Language::Kop),
        ("gvf", Language::Gvf),
        ("kjp", Language::Kjp),
        ("kip", Language::Kip),
        ("dhu", Language::Dhu),
        ("ley", Language::Ley),
        ("cad", Language::Cad),
        ("mri", Language::Mri),
        ("azg", Language::Azg),
        ("szd", Language::Szd),
        ("mej", Language::Mej),
        ("sjk", Language::Sjk),
        ("agy", Language::Agy),
        ("hro", Language::Hro),
        ("scb", Language::Scb),
        ("wuu", Language::Wuu),
        ("mtd", Language::Mtd),
        ("kge", Language::Kge),
        ("mip", Language::Mip),
        ("bqf", Language::Bqf),
        ("bpo", Language::Bpo),
        ("wmw", Language::Wmw),
        ("tbl", Language::Tbl),
        ("ato", Language::Ato),
        ("mpt", Language::Mpt),
        ("mkc", Language::Mkc),
        ("chf", Language::Chf),
        ("tcf", Language::Tcf),
        ("jor", Language::Jor),
        ("swo", Language::Swo),
        ("bjs", Language::Bjs),
        ("byg", Language::Byg),
        ("aks", Language::Aks),
        ("awb", Language::Awb),
        ("abh", Language::Abh),
        ("gai", Language::Gai),
        ("ktf", Language::Ktf),
        ("bbw", Language::Bbw),
        ("dmk", Language::Dmk),
        ("tge", Language::Tge),
        ("bgl", Language::Bgl),
        ("wib", Language::Wib),
        ("zak", Language::Zak),
        ("xsp", Language::Xsp),
        ("kap", Language::Kap),
        ("prn", Language::Prn),
        ("diw", Language::Diw),
        ("nsd", Language::Nsd),
        ("geq", Language::Geq),
        ("cds", Language::Cds),
        ("mvv", Language::Mvv),
        ("tyx", Language::Tyx),
        ("okl", Language::Okl),
        ("dnt", Language::Dnt),
        ("qxs", Language::Qxs),
        ("nyj", Language::Nyj),
        ("ysy", Language::Ysy),
        ("fuf", Language::Fuf),
        ("nyr", Language::Nyr),
        ("wbv", Language::Wbv),
        ("ndg", Language::Ndg),
        ("rax", Language::Rax),
        ("jmb", Language::Jmb),
        ("gal", Language::Gal),
        ("zmz", Language::Zmz),
        ("sop", Language::Sop),
        ("kec", Language::Kec),
        ("alt", Language::Alt),
        ("maa", Language::Maa),
        ("itb", Language::Itb),
        ("tri", Language::Tri),
        ("hmt", Language::Hmt),
        ("weh", Language::Weh),
        ("fip", Language::Fip),
        ("ckn", Language::Ckn),
        ("xnz", Language::Xnz),
        ("tia", Language::Tia),
        ("xks", Language::Xks),
        ("djw", Language::Djw),
        ("szv", Language::Szv),
        ("igg", Language::Igg),
        ("now", Language::Now),
        ("bas", Language::Bas),
        ("lik", Language::Lik),
        ("alo", Language::Alo),
        ("obt", Language::Obt),
        ("owl", Language::Owl),
        ("djm", Language::Djm),
        ("uum", Language::Uum),
        ("mbv", Language::Mbv),
        ("ljl", Language::Ljl),
        ("xgw", Language::Xgw),
        ("ymb", Language::Ymb),
        ("glk", Language::Glk),
        ("ksh", Language::Ksh),
        ("qvn", Language::Qvn),
        ("dic", Language::Dic),
        ("sed", Language::Sed),
        ("bru", Language::Bru),
        ("mth", Language::Mth),
        ("kas", Language::Kas),
        ("ypo", Language::Ypo),
        ("est", Language::Est),
        ("miq", Language::Miq),
        ("pin", Language::Pin),
        ("mxx", Language::Mxx),
        ("tis", Language::Tis),
        ("bbk", Language::Bbk),
        ("zwa", Language::Zwa),
        ("bqu", Language::Bqu),
        ("plc", Language::Plc),
        ("xgb", Language::Xgb),
        ("ybk", Language::Ybk),
        ("nkt", Language::Nkt),
        ("afk", Language::Afk),
        ("prf", Language::Prf),
        ("xab", Language::Xab),
        ("ngp", Language::Ngp),
        ("ksj", Language::Ksj),
        ("aul", Language::Aul),
        ("zav", Language::Zav),
        ("mzc", Language::Mzc),
        ("fuy", Language::Fuy),
        ("afp", Language::Afp),
        ("bsm", Language::Bsm),
        ("cam", Language::Cam),
        ("dje", Language::Dje),
        ("doq", Language::Doq),
        ("txy", Language::Txy),
        ("trz", Language::Trz),
        ("spl", Language::Spl),
        ("kma", Language::Kma),
        ("kbm", Language::Kbm),
        ("ddr", Language::Ddr),
        ("kuc", Language::Kuc),
        ("tiu", Language::Tiu),
        ("mqs", Language::Mqs),
        ("dez", Language::Dez),
        ("vkj", Language::Vkj),
        ("mwr", Language::Mwr),
        ("epo", Language::Epo),
        ("kyh", Language::Kyh),
        ("haf", Language::Haf),
        ("bwp", Language::Bwp),
        ("xwt", Language::Xwt),
        ("bzw", Language::Bzw),
        ("ski", Language::Ski),
        ("msm", Language::Msm),
        ("mwf", Language::Mwf),
        ("saf", Language::Saf),
        ("kwc", Language::Kwc),
        ("lut", Language::Lut),
        ("xbo", Language::Xbo),
        ("aer", Language::Aer),
        ("ibb", Language::Ibb),
        ("tag", Language::Tag),
        ("mxb", Language::Mxb),
        ("buf", Language::Buf),
        ("xch", Language::Xch),
        ("bal", Language::Bal),
        ("apx", Language::Apx),
        ("wmt", Language::Wmt),
        ("kcm", Language::Kcm),
        ("ljw", Language::Ljw),
        ("pbu", Language::Pbu),
        ("nmt", Language::Nmt),
        ("bwb", Language::Bwb),
        ("kuh", Language::Kuh),
        ("til", Language::Til),
        ("chk", Language::Chk),
        ("vme", Language::Vme),
        ("jko", Language::Jko),
        ("sgx", Language::Sgx),
        ("xup", Language::Xup),
        ("ssd", Language::Ssd),
        ("mvl", Language::Mvl),
        ("hsl", Language::Hsl),
        ("krf", Language::Krf),
        ("mik", Language::Mik),
        ("xrn", Language::Xrn),
        ("xvs", Language::Xvs),
        ("ish", Language::Ish),
        ("njr", Language::Njr),
        ("tgh", Language::Tgh),
        ("mrz", Language::Mrz),
        ("sni", Language::Sni),
        ("xld", Language::Xld),
        ("dtt", Language::Dtt),
        ("stt", Language::Stt),
        ("not", Language::Not),
        ("tmv", Language::Tmv),
        ("kyr", Language::Kyr),
        ("iry", Language::Iry),
        ("zne", Language::Zne),
        ("msf", Language::Msf),
        ("tqt", Language::Tqt),
        ("peg", Language::Peg),
        ("pcg", Language::Pcg),
        ("ktq", Language::Ktq),
        ("yry", Language::Yry),
        ("pae", Language::Pae),
        ("thl", Language::Thl),
        ("zlq", Language::Zlq),
        ("utr", Language::Utr),
        ("aqm", Language::Aqm),
        ("brc", Language::Brc),
        ("tnd", Language::Tnd),
        ("lgu", Language::Lgu),
        ("nlc", Language::Nlc),
        ("hna", Language::Hna),
        ("khk", Language::Khk),
        ("cri", Language::Cri),
        ("gzn", Language::Gzn),
        ("deh", Language::Deh),
        ("llg", Language::Llg),
        ("qvy", Language::Qvy),
        ("cbu", Language::Cbu),
        ("gmx", Language::Gmx),
        ("run", Language::Run),
        ("dds", Language::Dds),
        ("cbn", Language::Cbn),
        ("mny", Language::Mny),
        ("cup", Language::Cup),
        ("ygi", Language::Ygi),
        ("pkg", Language::Pkg),
        ("nag", Language::Nag),
        ("jab", Language::Jab),
        ("ihw", Language::Ihw),
        ("inz", Language::Inz),
        ("sow", Language::Sow),
        ("snj", Language::Snj),
        ("snr", Language::Snr),
        ("brv", Language::Brv),
        ("bph", Language::Bph),
        ("knn", Language::Knn),
        ("suq", Language::Suq),
        ("mmo", Language::Mmo),
        ("zuh", Language::Zuh),
        ("bil", Language::Bil),
        ("pmi", Language::Pmi),
        ("mvf", Language::Mvf),
        ("cbj", Language::Cbj),
        ("bfg", Language::Bfg),
        ("xsr", Language::Xsr),
        ("nzk", Language::Nzk),
        ("ait", Language::Ait),
        ("pmy", Language::Pmy),
        ("ojw", Language::Ojw),
        ("mmw", Language::Mmw),
        ("lvs", Language::Lvs),
        ("pyn", Language::Pyn),
        ("zmx", Language::Zmx),
        ("mtp", Language::Mtp),
        ("fub", Language::Fub),
        ("kfr", Language::Kfr),
        ("mqv", Language::Mqv),
        ("pbi", Language::Pbi),
        ("lgl", Language::Lgl),
        ("tcx", Language::Tcx),
        ("pgn", Language::Pgn),
        ("jls", Language::Jls),
        ("nge", Language::Nge),
        ("boi", Language::Boi),
        ("wam", Language::Wam),
        ("ksg", Language::Ksg),
        ("bkq", Language::Bkq),
        ("npg", Language::Npg),
        ("yra", Language::Yra),
        ("pri", Language::Pri),
        ("lxm", Language::Lxm),
        ("lwe", Language::Lwe),
        ("kmi", Language::Kmi),
        ("iko", Language::Iko),
        ("nni", Language::Nni),
        ("wai", Language::Wai),
        ("dhi", Language::Dhi),
        ("cpx", Language::Cpx),
        ("mug", Language::Mug),
        ("kmf", Language::Kmf),
        ("dsn", Language::Dsn),
        ("yet", Language::Yet),
        ("wsg", Language::Wsg),
        ("bbs", Language::Bbs),
        ("fer", Language::Fer),
        ("qxa", Language::Qxa),
        ("saz", Language::Saz),
        ("bey", Language::Bey),
        ("bge", Language::Bge),
        ("yva", Language::Yva),
        ("smb", Language::Smb),
        ("kvd", Language::Kvd),
        ("tox", Language::Tox),
        ("kcu", Language::Kcu),
        ("wbw", Language::Wbw),
        ("sox", Language::Sox),
        ("isc", Language::Isc),
        ("rer", Language::Rer),
        ("kjd", Language::Kjd),
        ("vaj", Language::Vaj),
        ("kbt", Language::Kbt),
        ("nxx", Language::Nxx),
        ("nyk", Language::Nyk),
        ("zil", Language::Zil),
        ("yin", Language::Yin),
        ("zab", Language::Zab),
        ("mjx", Language::Mjx),
        ("din", Language::Din),
        ("blv", Language::Blv),
        ("orv", Language::Orv),
        ("mut", Language::Mut),
        ("nys", Language::Nys),
        ("ngr", Language::Ngr),
        ("tys", Language::Tys),
        ("zum", Language::Zum),
        ("kza", Language::Kza),
        ("xnu", Language::Xnu),
        ("mwn", Language::Mwn),
        ("pee", Language::Pee),
        ("lch", Language::Lch),
        ("sog", Language::Sog),
        ("cnh", Language::Cnh),
        ("klt", Language::Klt),
        ("vit", Language::Vit),
        ("klw", Language::Klw),
        ("ori", Language::Ori),
        ("eyo", Language::Eyo),
        ("sej", Language::Sej),
        ("wyy", Language::Wyy),
        ("ksu", Language::Ksu),
        ("chl", Language::Chl),
        ("dmv", Language::Dmv),
        ("roo", Language::Roo),
        ("bwe", Language::Bwe),
        ("xcm", Language::Xcm),
        ("bqx", Language::Bqx),
        ("amt", Language::Amt),
        ("xor", Language::Xor),
        ("hwo", Language::Hwo),
        ("alp", Language::Alp),
        ("yot", Language::Yot),
        ("tsw", Language::Tsw),
        ("mlv", Language::Mlv),
        ("nbj", Language::Nbj),
        ("pda", Language::Pda),
        ("auy", Language::Auy),
        ("mwe", Language::Mwe),
        ("gru", Language::Gru),
        ("urm", Language::Urm),
        ("osx", Language::Osx),
        ("zpz", Language::Zpz),
        ("azd", Language::Azd),
        ("ojv", Language::Ojv),
        ("qux", Language::Qux),
        ("kys", Language::Kys),
        ("rma", Language::Rma),
        ("kqy", Language::Kqy),
        ("xvn", Language::Xvn),
        ("kqw", Language::Kqw),
        ("can", Language::Can),
        ("mbd", Language::Mbd),
        ("knr", Language::Knr),
        ("dma", Language::Dma),
        ("jua", Language::Jua),
        ("gun", Language::Gun),
        ("aeq", Language::Aeq),
        ("kbw", Language::Kbw),
        ("xpi", Language::Xpi),
        ("vms", Language::Vms),
        ("cah", Language::Cah),
        ("woc", Language::Woc),
        ("ald", Language::Ald),
        ("ttu", Language::Ttu),
        ("cbq", Language::Cbq),
        ("zmr", Language::Zmr),
        ("mns", Language::Mns),
        ("aig", Language::Aig),
        ("tzm", Language::Tzm),
        ("ade", Language::Ade),
        ("bya", Language::Bya),
        ("idd", Language::Idd),
        ("una", Language::Una),
        ("nzu", Language::Nzu),
        ("gdc", Language::Gdc),
        ("bax", Language::Bax),
        ("smy", Language::Smy),
        ("mwv", Language::Mwv),
        ("tuq", Language::Tuq),
        ("xem", Language::Xem),
        ("kli", Language::Kli),
        ("rgk", Language::Rgk),
        ("amh", Language::Amh),
        ("esn", Language::Esn),
        ("tml", Language::Tml),
        ("ppn", Language::Ppn),
        ("wrp", Language::Wrp),
        ("lio", Language::Lio),
        ("akg", Language::Akg),
        ("mte", Language::Mte),
        ("emb", Language::Emb),
        ("tsp", Language::Tsp),
        ("wbf", Language::Wbf),
        ("kym", Language::Kym),
        ("gyl", Language::Gyl),
        ("kaf", Language::Kaf),
        ("duu", Language::Duu),
        ("ssq", Language::Ssq),
        ("byd", Language::Byd),
        ("mjt", Language::Mjt),
        ("sue", Language::Sue),
        ("zeg", Language::Zeg),
        ("hmz", Language::Hmz),
        ("skp", Language::Skp),
        ("chr", Language::Chr),
        ("kes", Language::Kes),
        ("diu", Language::Diu),
        ("bli", Language::Bli),
        ("tqw", Language::Tqw),
        ("aiy", Language::Aiy),
        ("adf", Language::Adf),
        ("kgu", Language::Kgu),
        ("zgh", Language::Zgh),
        ("urc", Language::Urc),
        ("gwu", Language::Gwu),
        ("cjp", Language::Cjp),
        ("xda", Language::Xda),
        ("wmc", Language::Wmc),
        ("med", Language::Med),
        ("glb", Language::Glb),
        ("eso", Language::Eso),
        ("bae", Language::Bae),
        ("gjr", Language::Gjr),
        ("wls", Language::Wls),
        ("bqc", Language::Bqc),
        ("nbt", Language::Nbt),
        ("bgu", Language::Bgu),
        ("rmu", Language::Rmu),
        ("xku", Language::Xku),
        ("yns", Language::Yns),
        ("ckz", Language::Ckz),
        ("vka", Language::Vka),
        ("sua", Language::Sua),
        ("ppe", Language::Ppe),
        ("bfa", Language::Bfa),
        ("crt", Language::Crt),
        ("tlc", Language::Tlc),
        ("qui", Language::Qui),
        ("xoi", Language::Xoi),
        ("xmr", Language::Xmr),
        ("quq", Language::Quq),
        ("mgw", Language::Mgw),
        ("rmo", Language::Rmo),
        ("scf", Language::Scf),
        ("elu", Language::Elu),
        ("drc", Language::Drc),
        ("nwx", Language::Nwx),
        ("jie", Language::Jie),
        ("ukp", Language::Ukp),
        ("lwl", Language::Lwl),
        ("gin", Language::Gin),
        ("wed", Language::Wed),
        ("hui", Language::Hui),
        ("wma", Language::Wma),
        ("mhe", Language::Mhe),
        ("cae", Language::Cae),
        ("bam", Language::Bam),
        ("kio", Language::Kio),
        ("psd", Language::Psd),
        ("yum", Language::Yum),
        ("okd", Language::Okd),
        ("gll", Language::Gll),
        ("ilu", Language::Ilu),
        ("grw", Language::Grw),
        ("prk", Language::Prk),
        ("tmq", Language::Tmq),
        ("sta", Language::Sta),
        ("biq", Language::Biq),
        ("mkx", Language::Mkx),
        ("dax", Language::Dax),
        ("chm", Language::Chm),
        ("kdk", Language::Kdk),
        ("qxr", Language::Qxr),
        ("cob", Language::Cob),
        ("mod", Language::Mod),
        ("itd", Language::Itd),
        ("yla", Language::Yla),
        ("umr", Language::Umr),
        ("dhn", Language::Dhn),
        ("onr", Language::Onr),
        ("ror", Language::Ror),
        ("nmw", Language::Nmw),
        ("tea", Language::Tea),
        ("zpw", Language::Zpw),
        ("yly", Language::Yly),
        ("nnn", Language::Nnn),
        ("adt", Language::Adt),
        ("bcj", Language::Bcj),
        ("syw", Language::Syw),
        ("yay", Language::Yay),
        ("ahr", Language::Ahr),
        ("pnm", Language::Pnm),
        ("kjt", Language::Kjt),
        ("tjp", Language::Tjp),
        ("brz", Language::Brz),
        ("bld", Language::Bld),
        ("cat", Language::Cat),
        ("dgg", Language::Dgg),
        ("gni", Language::Gni),
        ("nuw", Language::Nuw),
        ("hnd", Language::Hnd),
        ("rji", Language::Rji),
        ("swg", Language::Swg),
        ("wly", Language::Wly),
        ("syi", Language::Syi),
        ("yrn", Language::Yrn),
        ("mcm", Language::Mcm),
        ("yab", Language::Yab),
        ("agc", Language::Agc),
        ("acv", Language::Acv),
        ("wdj", Language::Wdj),
        ("yee", Language::Yee),
        ("pna", Language::Pna),
        ("xal", Language::Xal),
        ("emn", Language::Emn),
        ("wru", Language::Wru),
        ("yuc", Language::Yuc),
        ("nxn", Language::Nxn),
        ("txt", Language::Txt),
        ("nug", Language::Nug),
        ("xis", Language::Xis),
        ("zpt", Language::Zpt),
        ("ymi", Language::Ymi),
        ("wha", Language::Wha),
        ("tjo", Language::Tjo),
        ("tvy", Language::Tvy),
        ("obi", Language::Obi),
        ("ngd", Language::Ngd),
        ("kyq", Language::Kyq),
        ("jad", Language::Jad),
        ("lmk", Language::Lmk),
        ("ilp", Language::Ilp),
        ("mft", Language::Mft),
        ("jhs", Language::Jhs),
        ("vun", Language::Vun),
        ("akr", Language::Akr),
        ("dee", Language::Dee),
        ("sbh", Language::Sbh),
        ("erg", Language::Erg),
        ("kbu", Language::Kbu),
        ("tbo", Language::Tbo),
        ("kmy", Language::Kmy),
        ("nak", Language::Nak),
        ("sim", Language::Sim),
        ("oru", Language::Oru),
        ("vls", Language::Vls),
        ("bac", Language::Bac),
        ("bmx", Language::Bmx),
        ("btc", Language::Btc),
        ("mmr", Language::Mmr),
        ("itz", Language::Itz),
        ("abj", Language::Abj),
        ("mjk", Language::Mjk),
        ("bup", Language::Bup),
        ("haj", Language::Haj),
        ("tum", Language::Tum),
        ("ola", Language::Ola),
        ("gta", Language::Gta),
        ("rga", Language::Rga),
        ("dia", Language::Dia),
        ("acx", Language::Acx),
        ("pto", Language::Pto),
        ("kcx", Language::Kcx),
        ("bpi", Language::Bpi),
        ("xta", Language::Xta),
        ("ono", Language::Ono),
        ("hms", Language::Hms),
        ("bei", Language::Bei),
        ("wnd", Language::Wnd),
        ("goo", Language::Goo),
        ("mck", Language::Mck),
        ("njn", Language::Njn),
        ("phj", Language::Phj),
        ("lkm", Language::Lkm),
        ("kpv", Language::Kpv),
        ("xed", Language::Xed),
        ("bcd", Language::Bcd),
        ("abp", Language::Abp),
        ("imi", Language::Imi),
        ("ulf", Language::Ulf),
        ("guo", Language::Guo),
        ("kmz", Language::Kmz),
        ("hti", Language::Hti),
        ("stk", Language::Stk),
        ("bnr", Language::Bnr),
        ("ket", Language::Ket),
        ("bqq", Language::Bqq),
        ("ksm", Language::Ksm),
        ("tvt", Language::Tvt),
        ("thr", Language::Thr),
        ("gdj", Language::Gdj),
        ("zom", Language::Zom),
        ("wab", Language::Wab),
        ("yxu", Language::Yxu),
        ("elm", Language::Elm),
        ("tnq", Language::Tnq),
        ("cok", Language::Cok),
        ("amk", Language::Amk),
        ("gef", Language::Gef),
        ("efa", Language::Efa),
        ("luo", Language::Luo),
        ("nmd", Language::Nmd),
        ("mif", Language::Mif),
        ("unx", Language::Unx),
        ("trb", Language::Trb),
        ("slz", Language::Slz),
        ("xep", Language::Xep),
        ("cnk", Language::Cnk),
        ("byo", Language::Byo),
        ("pib", Language::Pib),
        ("ign", Language::Ign),
        ("pti", Language::Pti),
        ("sel", Language::Sel),
        ("bvo", Language::Bvo),
        ("moz", Language::Moz),
        ("swt", Language::Swt),
        ("mqn", Language::Mqn),
        ("ewo", Language::Ewo),
        ("bgp", Language::Bgp),
        ("vil", Language::Vil),
        ("tmf", Language::Tmf),
        ("msk", Language::Msk),
        ("mbn", Language::Mbn),
        ("tgd", Language::Tgd),
        ("nbb", Language::Nbb),
        ("noq", Language::Noq),
        ("low", Language::Low),
        ("bzi", Language::Bzi),
        ("kvp", Language::Kvp),
        ("jer", Language::Jer),
        ("wkb", Language::Wkb),
        ("ktb", Language::Ktb),
        ("asg", Language::Asg),
        ("brm", Language::Brm),
        ("hsn", Language::Hsn),
        ("bfx", Language::Bfx),
        ("kkj", Language::Kkj),
        ("glj", Language::Glj),
        ("xow", Language::Xow),
        ("kfd", Language::Kfd),
        ("csr", Language::Csr),
        ("bmi", Language::Bmi),
        ("dhw", Language::Dhw),
        ("kmh", Language::Kmh),
        ("bfh", Language::Bfh),
        ("mrf", Language::Mrf),
        ("wme", Language::Wme),
        ("asu", Language::Asu),
        ("bvz", Language::Bvz),
        ("pth", Language::Pth),
        ("skd", Language::Skd),
        ("cby", Language::Cby),
        ("ima", Language::Ima),
        ("beu", Language::Beu),
        ("lic", Language::Lic),
        ("oui", Language::Oui),
        ("ncx", Language::Ncx),
        ("xyk", Language::Xyk),
        ("faa", Language::Faa),
        ("tza", Language::Tza),
        ("fui", Language::Fui),
        ("meq", Language::Meq),
        ("gui", Language::Gui),
        ("mke", Language::Mke),
        ("cfd", Language::Cfd),
        ("jow", Language::Jow),
        ("yxm", Language::Yxm),
        ("inp", Language::Inp),
        ("nss", Language::Nss),
        ("ssk", Language::Ssk),
        ("ing", Language::Ing),
        ("mlh", Language::Mlh),
        ("ygl", Language::Ygl),
        ("khp", Language::Khp),
        ("bzx", Language::Bzx),
        ("cng", Language::Cng),
        ("ada", Language::Ada),
        ("mpy", Language::Mpy),
        ("oin", Language::Oin),
        ("bvi", Language::Bvi),
        ("kpr", Language::Kpr),
        ("dbn", Language::Dbn),
        ("knq", Language::Knq),
        ("duw", Language::Duw),
        ("jil", Language::Jil),
        ("juh", Language::Juh),
        ("zkt", Language::Zkt),
        ("meb", Language::Meb),
        ("tad", Language::Tad),
        ("gfk", Language::Gfk),
        ("atb", Language::Atb),
        ("vmd", Language::Vmd),
        ("xkd", Language::Xkd),
        ("nvh", Language::Nvh),
        ("khn", Language::Khn),
        ("ciw", Language::Ciw),
        ("obr", Language::Obr),
        ("hmi", Language::Hmi),
        ("gmr", Language::Gmr),
        ("wao", Language::Wao),
        ("sid", Language::Sid),
        ("duv", Language::Duv),
        ("bcr", Language::Bcr),
        ("fuv", Language::Fuv),
        ("kib", Language::Kib),
        ("bnk", Language::Bnk),
        ("tkx", Language::Tkx),
        ("kcd", Language::Kcd),
        ("orn", Language::Orn),
        ("nte", Language::Nte),
        ("peq", Language::Peq),
        ("lja", Language::Lja),
        ("tee", Language::Tee),
        ("alf", Language::Alf),
        ("mvx", Language::Mvx),
        ("drt", Language::Drt),
        ("ytw", Language::Ytw),
        ("tnx", Language::Tnx),
        ("sea", Language::Sea),
        ("ymr", Language::Ymr),
        ("snp", Language::Snp),
        ("bnl", Language::Bnl),
        ("oyd", Language::Oyd),
        ("caq", Language::Caq),
        ("lts", Language::Lts),
        ("kki", Language::Kki),
        ("seq", Language::Seq),
        ("jku", Language::Jku),
        ("nux", Language::Nux),
        ("ava", Language::Ava),
        ("nbd", Language::Nbd),
        ("mbr", Language::Mbr),
        ("cih", Language::Cih),
        ("wwr", Language::Wwr),
        ("wbs", Language::Wbs),
        ("gnh", Language::Gnh),
        ("kzz", Language::Kzz),
        ("qua", Language::Qua),
        ("yes", Language::Yes),
        ("nsg", Language::Nsg),
        ("pps", Language::Pps),
        ("aes", Language::Aes),
        ("ssc", Language::Ssc),
        ("wny", Language::Wny),
        ("hih", Language::Hih),
        ("qun", Language::Qun),
        ("bzn", Language::Bzn),
        ("bvd", Language::Bvd),
        ("lgh", Language::Lgh),
        ("kfp", Language::Kfp),
        ("djc", Language::Djc),
        ("tkw", Language::Tkw),
        ("skx", Language::Skx),
        ("nrx", Language::Nrx),
        ("cuv", Language::Cuv),
        ("pbr", Language::Pbr),
        ("gah", Language::Gah),
        ("sbp", Language::Sbp),
        ("sqi", Language::Sqi),
        ("okn", Language::Okn),
        ("tpo", Language::Tpo),
        ("tld", Language::Tld),
        ("xmx", Language::Xmx),
        ("sjn", Language::Sjn),
        ("tgo", Language::Tgo),
        ("npu", Language::Npu),
        ("zia", Language::Zia),
        ("mmi", Language::Mmi),
        ("gov", Language::Gov),
        ("yas", Language::Yas),
        ("nsi", Language::Nsi),
        ("cdz", Language::Cdz),
        ("mfc", Language::Mfc),
        ("otm", Language::Otm),
        ("pta", Language::Pta),
        ("wrm", Language::Wrm),
        ("att", Language::Att),
        ("mmq", Language::Mmq),
        ("jiq", Language::Jiq),
        ("tck", Language::Tck),
        ("ili", Language::Ili),
        ("sbt", Language::Sbt),
        ("avv", Language::Avv),
        ("mou", Language::Mou),
        ("hbu", Language::Hbu),
        ("stb", Language::Stb),
        ("mcq", Language::Mcq),
        ("syr", Language::Syr),
        ("ane", Language::Ane),
        ("lpx", Language::Lpx),
        ("gec", Language::Gec),
        ("phm", Language::Phm),
        ("ude", Language::Ude),
        ("sut", Language::Sut),
        ("sza", Language::Sza),
        ("gju", Language::Gju),
        ("abb", Language::Abb),
        ("orx", Language::Orx),
        ("aeu", Language::Aeu),
        ("new", Language::New),
        ("bsv", Language::Bsv),
        ("wet", Language::Wet),
        ("wbm", Language::Wbm),
        ("pwi", Language::Pwi),
        ("hmk", Language::Hmk),
        ("tga", Language::Tga),
        ("bnd", Language::Bnd),
        ("and", Language::And),
        ("sif", Language::Sif),
        ("cty", Language::Cty),
        ("sna", Language::Sna),
        ("srv", Language::Srv),
        ("zla", Language::Zla),
        ("kud", Language::Kud),
        ("xgf", Language::Xgf),
        ("piv", Language::Piv),
        ("mdq", Language::Mdq),
        ("smk", Language::Smk),
        ("mxk", Language::Mxk),
        ("wat", Language::Wat),
        ("wlw", Language::Wlw),
        ("klf", Language::Klf),
        ("dhv", Language::Dhv),
        ("cra", Language::Cra),
        ("puu", Language::Puu),
        ("pwo", Language::Pwo),
        ("ibh", Language::Ibh),
        ("pbm", Language::Pbm),
        ("ebo", Language::Ebo),
        ("fur", Language::Fur),
        ("bvl", Language::Bvl),
        ("sew", Language::Sew),
        ("agh", Language::Agh),
        ("zoc", Language::Zoc),
        ("pof", Language::Pof),
        ("mjp", Language::Mjp),
        ("tpy", Language::Tpy),
        ("kmp", Language::Kmp),
        ("msl", Language::Msl),
        ("nei", Language::Nei),
        ("ojg", Language::Ojg),
        ("ttr", Language::Ttr),
        ("gig", Language::Gig),
        ("ckh", Language::Ckh),
        ("fue", Language::Fue),
        ("ulk", Language::Ulk),
        ("ppk", Language::Ppk),
        ("pcc", Language::Pcc),
        ("rnl", Language::Rnl),
        ("zty", Language::Zty),
        ("tgn", Language::Tgn),
        ("jdg", Language::Jdg),
        ("tkn", Language::Tkn),
        ("kcj", Language::Kcj),
        ("mbi", Language::Mbi),
        ("dwr", Language::Dwr),
        ("diq", Language::Diq),
        ("tgj", Language::Tgj),
        ("kta", Language::Kta),
        ("aih", Language::Aih),
        ("btv", Language::Btv),
        ("urd", Language::Urd),
        ("duo", Language::Duo),
        ("kig", Language::Kig),
        ("gbe", Language::Gbe),
        ("cut", Language::Cut),
        ("ykk", Language::Ykk),
        ("uur", Language::Uur),
        ("nhb", Language::Nhb),
        ("mrh", Language::Mrh),
        ("bia", Language::Bia),
        ("och", Language::Och),
        ("kjo", Language::Kjo),
        ("bew", Language::Bew),
        ("bak", Language::Bak),
        ("yif", Language::Yif),
        ("tqu", Language::Tqu),
        ("aho", Language::Aho),
        ("ahk", Language::Ahk),
        ("ypz", Language::Ypz),
        ("pol", Language::Pol),
        ("toj", Language::Toj),
        ("ybx", Language::Ybx),
        ("buh", Language::Buh),
        ("crw", Language::Crw),
        ("mbx", Language::Mbx),
        ("leo", Language::Leo),
        ("kii", Language::Kii),
        ("nbl", Language::Nbl),
        ("wrk", Language::Wrk),
        ("jel", Language::Jel),
        ("yis", Language::Yis),
        ("ikr", Language::Ikr),
        ("mdj", Language::Mdj),
        ("bpj", Language::Bpj),
        ("rad", Language::Rad),
        ("zpk", Language::Zpk),
        ("ntu", Language::Ntu),
        ("dnu", Language::Dnu),
        ("tpk", Language::Tpk),
        ("ktw", Language::Ktw),
        ("lgo", Language::Lgo),
        ("bux", Language::Bux),
        ("evn", Language::Evn),
        ("lih", Language::Lih),
        ("dux", Language::Dux),
        ("nct", Language::Nct),
        ("kam", Language::Kam),
        ("qxw", Language::Qxw),
        ("tiy", Language::Tiy),
        ("bzk", Language::Bzk),
        ("mnr", Language::Mnr),
        ("grs", Language::Grs),
        ("dpp", Language::Dpp),
        ("got", Language::Got),
        ("met", Language::Met),
        ("mwg", Language::Mwg),
        ("ate", Language::Ate),
        ("mht", Language::Mht),
        ("zaf", Language::Zaf),
        ("dev", Language::Dev),
        ("nun", Language::Nun),
        ("omu", Language::Omu),
        ("aqk", Language::Aqk),
        ("yiu", Language::Yiu),
        ("zcd", Language::Zcd),
        ("dan", Language::Dan),
        ("gar", Language::Gar),
        ("cik", Language::Cik),
        ("izr", Language::Izr),
        ("gcf", Language::Gcf),
        ("ksz", Language::Ksz),
        ("dth", Language::Dth),
        ("fwe", Language::Fwe),
        ("ved", Language::Ved),
        ("cdr", Language::Cdr),
        ("xma", Language::Xma),
        ("nsb", Language::Nsb),
        ("mzx", Language::Mzx),
        ("mqf", Language::Mqf),
        ("laq", Language::Laq),
        ("vmf", Language::Vmf),
        ("ceg", Language::Ceg),
        ("kho", Language::Kho),
        ("twn", Language::Twn),
        ("nec", Language::Nec),
        ("pek", Language::Pek),
        ("hov", Language::Hov),
        ("cuc", Language::Cuc),
        ("gbm", Language::Gbm),
        ("sbw", Language::Sbw),
        ("kpf", Language::Kpf),
        ("npn", Language::Npn),
        ("srt", Language::Srt),
        ("xyj", Language::Xyj),
        ("ikk", Language::Ikk),
        ("tlo", Language::Tlo),
        ("tuo", Language::Tuo),
        ("tqn", Language::Tqn),
        ("twu", Language::Twu),
        ("awv", Language::Awv),
        ("iqu", Language::Iqu),
        ("mfs", Language::Mfs),
        ("nuj", Language::Nuj),
        ("vkp", Language::Vkp),
        ("kzb", Language::Kzb),
        ("swu", Language::Swu),
        ("rmn", Language::Rmn),
        ("lmp", Language::Lmp),
        ("bou", Language::Bou),
        ("ttm", Language::Ttm),
        ("mqu", Language::Mqu),
        ("aey", Language::Aey),
        ("pes", Language::Pes),
        ("igb", Language::Igb),
        ("lsd", Language::Lsd),
        ("chv", Language::Chv),
        ("sgy", Language::Sgy),
        ("pmt", Language::Pmt),
        ("nrr", Language::Nrr),
        ("kyj", Language::Kyj),
        ("kvq", Language::Kvq),
        ("slv", Language::Slv),
        ("zuy", Language::Zuy),
        ("mjv", Language::Mjv),
        ("nue", Language::Nue),
        ("skh", Language::Skh),
        ("yby", Language::Yby),
        ("stp", Language::Stp),
        ("hts", Language::Hts),
        ("syl", Language::Syl),
        ("nhn", Language::Nhn),
        ("sqq", Language::Sqq),
        ("gub", Language::Gub),
        ("myz", Language::Myz),
        ("koz", Language::Koz),
        ("bwf", Language::Bwf),
        ("frr", Language::Frr),
        ("mjm", Language::Mjm),
        ("bby", Language::Bby),
        ("byr", Language::Byr),
        ("kfg", Language::Kfg),
        ("mwz", Language::Mwz),
        ("amz", Language::Amz),
        ("akb", Language::Akb),
        ("zna", Language::Zna),
        ("zor", Language::Zor),
        ("mvd", Language::Mvd),
        ("pnt", Language::Pnt),
        ("mxn", Language::Mxn),
        ("lil", Language::Lil),
        ("jat", Language::Jat),
        ("ryu", Language::Ryu),
        ("asf", Language::Asf),
        ("zkz", Language::Zkz),
        ("yli", Language::Yli),
        ("kek", Language::Kek),
        ("mrq", Language::Mrq),
        ("bss", Language::Bss),
        ("xkj", Language::Xkj),
        ("mvt", Language::Mvt),
        ("cin", Language::Cin),
        ("mkq", Language::Mkq),
        ("bvm", Language::Bvm),
        ("cox", Language::Cox),
        ("sgj", Language::Sgj),
        ("klh", Language::Klh),
        ("nan", Language::Nan),
        ("mjl", Language::Mjl),
        ("poq", Language::Poq),
        ("srx", Language::Srx),
        ("lss", Language::Lss),
        ("mjb", Language::Mjb),
        ("zmh", Language::Zmh),
        ("nlo", Language::Nlo),
        ("mrn", Language::Mrn),
        ("kik", Language::Kik),
        ("dws", Language::Dws),
        ("god", Language::God),
        ("ayp", Language::Ayp),
        ("wnu", Language::Wnu),
        ("ynd", Language::Ynd),
        ("mvu", Language::Mvu),
        ("tsn", Language::Tsn),
        ("xle", Language::Xle),
        ("mtn", Language::Mtn),
        ("bkf", Language::Bkf),
        ("hmc", Language::Hmc),
        ("bbx", Language::Bbx),
        ("fuh", Language::Fuh),
        ("smm", Language::Smm),
        ("kts", Language::Kts),
        ("spe", Language::Spe),
        ("sib", Language::Sib),
        ("ewe", Language::Ewe),
        ("apo", Language::Apo),
        ("pmz", Language::Pmz),
        ("lsb", Language::Lsb),
        ("mtk", Language::Mtk),
        ("gor", Language::Gor),
        ("bfn", Language::Bfn),
        ("gwn", Language::Gwn),
        ("pce", Language::Pce),
        ("tsy", Language::Tsy),
        ("dnj", Language::Dnj),
        ("cot", Language::Cot),
        ("arq", Language::Arq),
        ("osc", Language::Osc),
        ("dad", Language::Dad),
        ("sdq", Language::Sdq),
        ("nhg", Language::Nhg),
        ("yno", Language::Yno),
        ("cdy", Language::Cdy),
        ("kth", Language::Kth),
        ("tve", Language::Tve),
        ("bhy", Language::Bhy),
        ("aws", Language::Aws),
        ("nyw", Language::Nyw),
        ("kmr", Language::Kmr),
        ("ksd", Language::Ksd),
        ("vmc", Language::Vmc),
        ("lnu", Language::Lnu),
        ("cku", Language::Cku),
        ("btx", Language::Btx),
        ("zad", Language::Zad),
        ("wmi", Language::Wmi),
        ("dro", Language::Dro),
        ("act", Language::Act),
        ("mwk", Language::Mwk),
        ("wwb", Language::Wwb),
        ("idr", Language::Idr),
        ("bdr", Language::Bdr),
        ("kby", Language::Kby),
        ("dsi", Language::Dsi),
        ("byz", Language::Byz),
        ("dyg", Language::Dyg),
        ("xpn", Language::Xpn),
        ("wbr", Language::Wbr),
        ("kjx", Language::Kjx),
        ("dmw", Language::Dmw),
        ("gdu", Language::Gdu),
        ("shm", Language::Shm),
        ("ckl", Language::Ckl),
        ("icr", Language::Icr),
        ("mmc", Language::Mmc),
        ("nib", Language::Nib),
        ("isu", Language::Isu),
        ("goh", Language::Goh),
        ("kqj", Language::Kqj),
        ("bpe", Language::Bpe),
        ("xme", Language::Xme),
        ("rit", Language::Rit),
        ("abw", Language::Abw),
        ("apu", Language::Apu),
        ("bhh", Language::Bhh),
        ("kbd", Language::Kbd),
        ("xmp", Language::Xmp),
        ("agv", Language::Agv),
        ("ddg", Language::Ddg),
        ("stn", Language::Stn),
        ("vml", Language::Vml),
        ("mfm", Language::Mfm),
        ("xri", Language::Xri),
        ("bhp", Language::Bhp),
        ("bzr", Language::Bzr),
        ("urv", Language::Urv),
        ("tjg", Language::Tjg),
        ("tgr", Language::Tgr),
        ("mwh", Language::Mwh),
        ("asv", Language::Asv),
        ("kau", Language::Kau),
        ("smn", Language::Smn),
        ("rob", Language::Rob),
        ("wlo", Language::Wlo),
        ("ido", Language::Ido),
        ("hik", Language::Hik),
        ("aad", Language::Aad),
        ("gnl", Language::Gnl),
        ("pmb", Language::Pmb),
        ("kwy", Language::Kwy),
        ("mwm", Language::Mwm),
        ("fra", Language::Fra),
        ("taz", Language::Taz),
        ("gos", Language::Gos),
        ("crv", Language::Crv),
        ("bpc", Language::Bpc),
        ("dww", Language::Dww),
        ("nzs", Language::Nzs),
        ("ugh", Language::Ugh),
        ("aci", Language::Aci),
        ("npb", Language::Npb),
        ("jyy", Language::Jyy),
        ("wsk", Language::Wsk),
        ("gbk", Language::Gbk),
        ("nku", Language::Nku),
        ("ssn", Language::Ssn),
        ("cla", Language::Cla),
        ("csv", Language::Csv),
        ("tcm", Language::Tcm),
        ("sll", Language::Sll),
        ("wnk", Language::Wnk),
        ("kbc", Language::Kbc),
        ("puw", Language::Puw),
        ("sgk", Language::Sgk),
        ("mij", Language::Mij),
        ("spa", Language::Spa),
        ("dug", Language::Dug),
        ("peb", Language::Peb),
        ("jda", Language::Jda),
        ("kfm", Language::Kfm),
        ("mje", Language::Mje),
        ("puo", Language::Puo),
        ("vel", Language::Vel),
        ("kvt", Language::Kvt),
        ("hud", Language::Hud),
        ("tpv", Language::Tpv),
        ("tcw", Language::Tcw),
        ("olk", Language::Olk),
        ("adl", Language::Adl),
        ("kyv", Language::Kyv),
        ("aie", Language::Aie),
        ("cim", Language::Cim),
        ("mey", Language::Mey),
        ("pga", Language::Pga),
        ("bkk", Language::Bkk),
        ("tnk", Language::Tnk),
        ("hye", Language::Hye),
        ("trx", Language::Trx),
        ("gul", Language::Gul),
        ("noy", Language::Noy),
        ("vmm", Language::Vmm),
        ("dac", Language::Dac),
        ("tpa", Language::Tpa),
        ("sae", Language::Sae),
        ("shx", Language::Shx),
        ("tmw", Language::Tmw),
        ("eit", Language::Eit),
        ("deg", Language::Deg),
        ("tit", Language::Tit),
        ("dmm", Language::Dmm),
        ("nem", Language::Nem),
        ("ymn", Language::Ymn),
        ("dyy", Language::Dyy),
        ("mzo", Language::Mzo),
        ("ifm", Language::Ifm),
        ("zbt", Language::Zbt),
        ("sck", Language::Sck),
        ("bci", Language::Bci),
        ("kxk", Language::Kxk),
        ("cpu", Language::Cpu),
        ("nny", Language::Nny),
        ("etr", Language::Etr),
        ("mjg", Language::Mjg),
        ("tgv", Language::Tgv),
        ("hkk", Language::Hkk),
        ("kqq", Language::Kqq),
        ("xan", Language::Xan),
        ("kea", Language::Kea),
        ("knf", Language::Knf),
        ("xte", Language::Xte),
        ("thk", Language::Thk),
        ("cjo", Language::Cjo),
        ("moa", Language::Moa),
        ("gof", Language::Gof),
        ("urb", Language::Urb),
        ("gsg", Language::Gsg),
        ("dav", Language::Dav),
        ("xky", Language::Xky),
        ("pdn", Language::Pdn),
        ("psr", Language::Psr),
        ("pkc", Language::Pkc),
        ("pnw", Language::Pnw),
        ("bbl", Language::Bbl),
        ("tah", Language::Tah),
        ("ibn", Language::Ibn),
        ("wih", Language::Wih),
        ("zkd", Language::Zkd),
        ("bxw", Language::Bxw),
        ("osp", Language::Osp),
        ("kqe", Language::Kqe),
        ("xmt", Language::Xmt),
        ("yku", Language::Yku),
        ("pxm", Language::Pxm),
        ("peo", Language::Peo),
        ("amc", Language::Amc),
        ("inh", Language::Inh),
        ("tnw", Language::Tnw),
        ("zqe", Language::Zqe),
        ("ull", Language::Ull),
        ("aaq", Language::Aaq),
        ("bun", Language::Bun),
        ("obk", Language::Obk),
        ("luj", Language::Luj),
        ("ele", Language::Ele),
        ("bwc", Language::Bwc),
        ("emk", Language::Emk),
        ("atv", Language::Atv),
        ("quy", Language::Quy),
        ("goe", Language::Goe),
        ("yww", Language::Yww),
        ("umi", Language::Umi),
        ("nmy", Language::Nmy),
        ("kkv", Language::Kkv),
        ("yyu", Language::Yyu),
        ("wno", Language::Wno),
        ("lby", Language::Lby),
        ("bdh", Language::Bdh),
        ("tli", Language::Tli),
        ("tek", Language::Tek),
        ("mfg", Language::Mfg),
        ("mfa", Language::Mfa),
        ("cfm", Language::Cfm),
        ("ver", Language::Ver),
        ("nju", Language::Nju),
        ("sda", Language::Sda),
        ("skt", Language::Skt),
        ("taf", Language::Taf),
        ("kwo", Language::Kwo),
        ("mhy", Language::Mhy),
        ("smw", Language::Smw),
        ("okc", Language::Okc),
        ("nhh", Language::Nhh),
        ("bhc", Language::Bhc),
        ("cbb", Language::Cbb),
        ("bnf", Language::Bnf),
        ("atr", Language::Atr),
        ("nsk", Language::Nsk),
        ("bqv", Language::Bqv),
        ("xkg", Language::Xkg),
        ("mla", Language::Mla),
        ("dbw", Language::Dbw),
        ("nne", Language::Nne),
        ("guf", Language::Guf),
        ("jus", Language::Jus),
        ("ify", Language::Ify),
        ("bql", Language::Bql),
        ("wie", Language::Wie),
        ("bmd", Language::Bmd),
        ("yty", Language::Yty),
        ("gam", Language::Gam),
        ("bir", Language::Bir),
        ("bom", Language::Bom),
        ("bxm", Language::Bxm),
        ("ymz", Language::Ymz),
        ("tlt", Language::Tlt),
        ("cwt", Language::Cwt),
        ("arb", Language::Arb),
        ("pac", Language::Pac),
        ("kzs", Language::Kzs),
        ("yat", Language::Yat),
        ("shl", Language::Shl),
        ("yne", Language::Yne),
        ("bzh", Language::Bzh),
        ("nqn", Language::Nqn),
        ("ziz", Language::Ziz),
        ("dao", Language::Dao),
        ("dep", Language::Dep),
        ("kbo", Language::Kbo),
        ("ddj", Language::Ddj),
        ("zlj", Language::Zlj),
        ("hrp", Language::Hrp),
        ("wer", Language::Wer),
        ("tda", Language::Tda),
        ("ijs", Language::Ijs),
        ("sjo", Language::Sjo),
        ("kxa", Language::Kxa),
        ("spy", Language::Spy),
        ("mes", Language::Mes),
        ("snf", Language::Snf),
        ("twc", Language::Twc),
        ("snm", Language::Snm),
        ("sld", Language::Sld),
        ("bvu", Language::Bvu),
        ("yss", Language::Yss),
        ("zra", Language::Zra),
        ("gbf", Language::Gbf),
        ("nio", Language::Nio),
        ("nlj", Language::Nlj),
        ("ajz", Language::Ajz),
        ("nmh", Language::Nmh),
        ("ynu", Language::Ynu),
        ("dwy", Language::Dwy),
        ("crb", Language::Crb),
        ("lug", Language::Lug),
        ("mpj", Language::Mpj),
        ("jml", Language::Jml),
        ("gtu", Language::Gtu),
        ("ktu", Language::Ktu),
        ("tpq", Language::Tpq),
        ("ggg", Language::Ggg),
        ("bzp", Language::Bzp),
        ("nfl", Language::Nfl),
        ("zai", Language::Zai),
        ("sbn", Language::Sbn),
        ("wal", Language::Wal),
        ("cwd", Language::Cwd),
        ("aoc", Language::Aoc),
        ("txu", Language::Txu),
        ("gao", Language::Gao),
        ("dui", Language::Dui),
        ("san", Language::San),
        ("end", Language::End),
        ("bbr", Language::Bbr),
        ("zaa", Language::Zaa),
        ("hup", Language::Hup),
        ("klq", Language::Klq),
        ("lpo", Language::Lpo),
        ("one", Language::One),
        ("ukg", Language::Ukg),
        ("bpy", Language::Bpy),
        ("ngw", Language::Ngw),
        ("xwk", Language::Xwk),
        ("bkc", Language::Bkc),
        ("yux", Language::Yux),
        ("jvd", Language::Jvd),
        ("nmk", Language::Nmk),
        ("srb", Language::Srb),
        ("wuy", Language::Wuy),
        ("ngx", Language::Ngx),
        ("pzh", Language::Pzh),
        ("tvu", Language::Tvu),
        ("lcl", Language::Lcl),
        ("dgh", Language::Dgh),
        ("mze", Language::Mze),
        ("gaa", Language::Gaa),
        ("iar", Language::Iar),
        ("aae", Language::Aae),
        ("plt", Language::Plt),
        ("cht", Language::Cht),
        ("kga", Language::Kga),
        ("mfk", Language::Mfk),
        ("oar", Language::Oar),
        ("ehu", Language::Ehu),
        ("khr", Language::Khr),
        ("knc", Language::Knc),
        ("ksi", Language::Ksi),
        ("urn", Language::Urn),
        ("bzj", Language::Bzj),
        ("yrb", Language::Yrb),
        ("nmz", Language::Nmz),
        ("zro", Language::Zro),
        ("yap", Language::Yap),
        ("bfw", Language::Bfw),
        ("btj", Language::Btj),
        ("srm", Language::Srm),
        ("kbx", Language::Kbx),
        ("wtf", Language::Wtf),
        ("keu", Language::Keu),
        ("tta", Language::Tta),
        ("soj", Language::Soj),
        ("leh", Language::Leh),
        ("bgn", Language::Bgn),
        ("tgq", Language::Tgq),
        ("vae", Language::Vae),
        ("mtj", Language::Mtj),
        ("krz", Language::Krz),
        ("gib", Language::Gib),
        ("ood", Language::Ood),
        ("gro", Language::Gro),
        ("bkm", Language::Bkm),
        ("ata", Language::Ata),
        ("mzw", Language::Mzw),
        ("nsx", Language::Nsx),
        ("zmb", Language::Zmb),
        ("sne", Language::Sne),
        ("ara", Language::Ara),
        ("irr", Language::Irr),
        ("aoz", Language::Aoz),
        ("nkd", Language::Nkd),
        ("xmg", Language::Xmg),
        ("wsv", Language::Wsv),
        ("mxc", Language::Mxc),
        ("sfm", Language::Sfm),
        ("lou", Language::Lou),
        ("msd", Language::Msd),
        ("kwa", Language::Kwa),
        ("prc", Language::Prc),
        ("eaa", Language::Eaa),
        ("bjn", Language::Bjn),
        ("apv", Language::Apv),
        ("pup", Language::Pup),
        ("kxt", Language::Kxt),
        ("phg", Language::Phg),
        ("gei", Language::Gei),
        ("sya", Language::Sya),
        ("asj", Language::Asj),
        ("brd", Language::Brd),
        ("egm", Language::Egm),
        ("nrf", Language::Nrf),
        ("bee", Language::Bee),
        ("waw", Language::Waw),
        ("crd", Language::Crd),
        ("sqm", Language::Sqm),
        ("yuq", Language::Yuq),
        ("phk", Language::Phk),
        ("nja", Language::Nja),
        ("stf", Language::Stf),
        ("aoa", Language::Aoa),
        ("ggu", Language::Ggu),
        ("pbc", Language::Pbc),
        ("sus", Language::Sus),
        ("yer", Language::Yer),
        ("mgb", Language::Mgb),
        ("uan", Language::Uan),
        ("miz", Language::Miz),
        ("tdc", Language::Tdc),
        ("mll", Language::Mll),
        ("lbr", Language::Lbr),
        ("nac", Language::Nac),
        ("dof", Language::Dof),
        ("fai", Language::Fai),
        ("plb", Language::Plb),
        ("tem", Language::Tem),
        ("alm", Language::Alm),
        ("ztt", Language::Ztt),
        ("ban", Language::Ban),
        ("pso", Language::Pso),
        ("for", Language::For),
        ("mlk", Language::Mlk),
        ("yev", Language::Yev),
        ("tok", Language::Tok),
        ("csq", Language::Csq),
        ("byk", Language::Byk),
        ("hau", Language::Hau),
        ("sxo", Language::Sxo),
        ("pkt", Language::Pkt),
        ("jac", Language::Jac),
        ("mwb", Language::Mwb),
        ("xwj", Language::Xwj),
        ("amx", Language::Amx),
        ("zbc", Language::Zbc),
        ("amm", Language::Amm),
        ("aro", Language::Aro),
        ("kne", Language::Kne),
        ("ers", Language::Ers),
        ("ump", Language::Ump),
        ("geb", Language::Geb),
        ("lle", Language::Lle),
        ("byx", Language::Byx),
        ("yph", Language::Yph),
        ("njo", Language::Njo),
        ("amy", Language::Amy),
        ("lha", Language::Lha),
        ("tsd", Language::Tsd),
        ("spn", Language::Spn),
        ("nty", Language::Nty),
        ("saa", Language::Saa),
        ("roe", Language::Roe),
        ("tma", Language::Tma),
        ("rpt", Language::Rpt),
        ("mlc", Language::Mlc),
        ("szw", Language::Szw),
        ("mgh", Language::Mgh),
        ("zho", Language::Zho),
        ("yag", Language::Yag),
        ("gom", Language::Gom),
        ("aum", Language::Aum),
        ("ntm", Language::Ntm),
        ("cbr", Language::Cbr),
        ("uis", Language::Uis),
        ("lmg", Language::Lmg),
        ("bxv", Language::Bxv),
        ("csi", Language::Csi),
        ("ahm", Language::Ahm),
        ("ncf", Language::Ncf),
        ("gry", Language::Gry),
        ("vmw", Language::Vmw),
        ("seb", Language::Seb),
        ("nam", Language::Nam),
        ("fin", Language::Fin),
        ("bfs", Language::Bfs),
        ("buj", Language::Buj),
        ("wba", Language::Wba),
        ("akv", Language::Akv),
        ("kue", Language::Kue),
        ("mqi", Language::Mqi),
        ("aec", Language::Aec),
        ("kme", Language::Kme),
        ("nsa", Language::Nsa),
        ("bwn", Language::Bwn),
        ("ksb", Language::Ksb),
        ("zpi", Language::Zpi),
        ("sss", Language::Sss),
        ("mbs", Language::Mbs),
        ("bgq", Language::Bgq),
        ("bef", Language::Bef),
        ("tvl", Language::Tvl),
        ("cnq", Language::Cnq),
        ("hib", Language::Hib),
        ("afe", Language::Afe),
        ("cky", Language::Cky),
        ("bpp", Language::Bpp),
        ("gvn", Language::Gvn),
        ("ugo", Language::Ugo),
        ("css", Language::Css),
        ("dhl", Language::Dhl),
        ("xlb", Language::Xlb),
        ("wuv", Language::Wuv),
        ("isa", Language::Isa),
        ("mcn", Language::Mcn),
        ("ncc", Language::Ncc),
        ("hug", Language::Hug),
        ("abx", Language::Abx),
        ("bem", Language::Bem),
        ("lnl", Language::Lnl),
        ("mvs", Language::Mvs),
        ("zyb", Language::Zyb),
        ("hmy", Language::Hmy),
        ("prr", Language::Prr),
        ("chq", Language::Chq),
        ("aud", Language::Aud),
        ("yrm", Language::Yrm),
        ("bxq", Language::Bxq),
        ("zmc", Language::Zmc),
        ("fli", Language::Fli),
        ("beq", Language::Beq),
        ("nnq", Language::Nnq),
        ("kif", Language::Kif),
        ("kpt", Language::Kpt),
        ("agj", Language::Agj),
        ("mgz", Language::Mgz),
        ("hbs", Language::Hbs),
        ("vmv", Language::Vmv),
        ("kji", Language::Kji),
        ("xwl", Language::Xwl),
        ("suz", Language::Suz),
        ("msn", Language::Msn),
        ("trj", Language::Trj),
        ("mve", Language::Mve),
        ("nde", Language::Nde),
        ("njd", Language::Njd),
        ("lhn", Language::Lhn),
        ("okk", Language::Okk),
        ("khw", Language::Khw),
        ("nxo", Language::Nxo),
        ("ygu", Language::Ygu),
        ("olt", Language::Olt),
        ("rwl", Language::Rwl),
        ("quf", Language::Quf),
        ("afg", Language::Afg),
        ("mmx", Language::Mmx),
        ("khm", Language::Khm),
        ("acm", Language::Acm),
        ("bul", Language::Bul),
        ("mpp", Language::Mpp),
        ("kyf", Language::Kyf),
        ("ggl", Language::Ggl),
        ("nrc", Language::Nrc),
        ("moy", Language::Moy),
        ("gie", Language::Gie),
        ("pay", Language::Pay),
        ("tod", Language::Tod),
        ("bpa", Language::Bpa),
        ("axk", Language::Axk),
        ("xmq", Language::Xmq),
        ("cdm", Language::Cdm),
        ("agi", Language::Agi),
        ("kkg", Language::Kkg),
        ("xav", Language::Xav),
        ("vra", Language::Vra),
        ("mqz", Language::Mqz),
        ("hds", Language::Hds),
        ("sme", Language::Sme),
        ("ggb", Language::Ggb),
        ("vin", Language::Vin),
        ("fss", Language::Fss),
        ("dow", Language::Dow),
        ("bjw", Language::Bjw),
        ("vmz", Language::Vmz),
        ("bgz", Language::Bgz),
        ("aaa", Language::Aaa),
        ("lrv", Language::Lrv),
        ("kew", Language::Kew),
        ("gvs", Language::Gvs),
        ("gdo", Language::Gdo),
        ("nsq", Language::Nsq),
        ("ncd", Language::Ncd),
        ("saw", Language::Saw),
        ("nnl", Language::Nnl),
        ("bhe", Language::Bhe),
        ("xsq", Language::Xsq),
        ("kou", Language::Kou),
        ("uba", Language::Uba),
        ("vau", Language::Vau),
        ("ltu", Language::Ltu),
        ("xmb", Language::Xmb),
        ("pum", Language::Pum),
        ("tbt", Language::Tbt),
        ("tmo", Language::Tmo),
        ("fni", Language::Fni),
        ("omc", Language::Omc),
        ("krw", Language::Krw),
        ("mhz", Language::Mhz),
        ("ztg", Language::Ztg),
        ("giu", Language::Giu),
        ("ghc", Language::Ghc),
        ("tix", Language::Tix),
        ("kvy", Language::Kvy),
        ("avt", Language::Avt),
        ("aut", Language::Aut),
        ("cso", Language::Cso),
        ("cym", Language::Cym),
        ("xpa", Language::Xpa),
        ("ahs", Language::Ahs),
        ("skv", Language::Skv),
        ("lku", Language::Lku),
        ("afo", Language::Afo),
        ("msw", Language::Msw),
        ("dze", Language::Dze),
        ("vkk", Language::Vkk),
        ("suo", Language::Suo),
        ("cth", Language::Cth),
        ("vaf", Language::Vaf),
        ("tan", Language::Tan),
        ("kex", Language::Kex),
        ("wmh", Language::Wmh),
        ("yau", Language::Yau),
        ("bta", Language::Bta),
        ("byp", Language::Byp),
        ("yaf", Language::Yaf),
        ("kgw", Language::Kgw),
        ("omk", Language::Omk),
        ("ndv", Language::Ndv),
        ("mak", Language::Mak),
        ("xbi", Language::Xbi),
        ("put", Language::Put),
        ("tln", Language::Tln),
        ("nhz", Language::Nhz),
        ("grr", Language::Grr),
        ("ruk", Language::Ruk),
        ("ybe", Language::Ybe),
        ("haz", Language::Haz),
        ("faf", Language::Faf),
        ("abr", Language::Abr),
        ("adq", Language::Adq),
        ("iri", Language::Iri),
        ("wrg", Language::Wrg),
        ("xat", Language::Xat),
        ("anq", Language::Anq),
        ("pni", Language::Pni),
        ("mil", Language::Mil),
        ("bji", Language::Bji),
        ("qws", Language::Qws),
        ("vlp", Language::Vlp),
        ("llh", Language::Llh),
        ("soo", Language::Soo),
        ("nyl", Language::Nyl),
        ("msq", Language::Msq),
        ("btf", Language::Btf),
        ("aqr", Language::Aqr),
        ("mct", Language::Mct),
        ("bja", Language::Bja),
        ("yob", Language::Yob),
        ("cou", Language::Cou),
        ("nmc", Language::Nmc),
        ("caj", Language::Caj),
        ("mkz", Language::Mkz),
        ("mgt", Language::Mgt),
        ("tbm", Language::Tbm),
        ("nxq", Language::Nxq),
        ("gng", Language::Gng),
        ("lbe", Language::Lbe),
        ("dse", Language::Dse),
        ("kra", Language::Kra),
        ("ufi", Language::Ufi),
        ("pix", Language::Pix),
        ("blm", Language::Blm),
        ("cpo", Language::Cpo),
        ("hum", Language::Hum),
        ("hve", Language::Hve),
        ("nii", Language::Nii),
        ("niu", Language::Niu),
        ("bhi", Language::Bhi),
        ("set", Language::Set),
        ("amv", Language::Amv),
        ("svx", Language::Svx),
        ("rsk", Language::Rsk),
        ("bhv", Language::Bhv),
        ("mry", Language::Mry),
        ("ekr", Language::Ekr),
        ("sag", Language::Sag),
        ("aui", Language::Aui),
        ("rts", Language::Rts),
        ("see", Language::See),
        ("zkr", Language::Zkr),
        ("vmi", Language::Vmi),
        ("tmg", Language::Tmg),
        ("lgg", Language::Lgg),
        ("xyt", Language::Xyt),
        ("ruy", Language::Ruy),
        ("naq", Language::Naq),
        ("nax", Language::Nax),
        ("wdd", Language::Wdd),
        ("mvy", Language::Mvy),
        ("zpq", Language::Zpq),
        ("ngk", Language::Ngk),
        ("toq", Language::Toq),
        ("ayu", Language::Ayu),
        ("xag", Language::Xag),
        ("giw", Language::Giw),
        ("bti", Language::Bti),
        ("cmt", Language::Cmt),
        ("tun", Language::Tun),
        ("xko", Language::Xko),
        ("bbv", Language::Bbv),
        ("xgl", Language::Xgl),
        ("mqy", Language::Mqy),
        ("piy", Language::Piy),
        ("dya", Language::Dya),
        ("gft", Language::Gft),
        ("okm", Language::Okm),
        ("txs", Language::Txs),
        ("xtw", Language::Xtw),
        ("nrp", Language::Nrp),
        ("bnm", Language::Bnm),
        ("bgy", Language::Bgy),
        ("gop", Language::Gop),
        ("bis", Language::Bis),
        ("bga", Language::Bga),
        ("ekm", Language::Ekm),
        ("kef", Language::Kef),
        ("ecs", Language::Ecs),
        ("lkr", Language::Lkr),
        ("umc", Language::Umc),
        ("prl", Language::Prl),
        ("mna", Language::Mna),
        ("hrz", Language::Hrz),
        ("lrg", Language::Lrg),
        ("zph", Language::Zph),
        ("pym", Language::Pym),
        ("raw", Language::Raw),
        ("bdk", Language::Bdk),
        ("bna", Language::Bna),
        ("bws", Language::Bws),
        ("asx", Language::Asx),
        ("kzw", Language::Kzw),
        ("mks", Language::Mks),
        ("byb", Language::Byb),
        ("bqk", Language::Bqk),
        ("glu", Language::Glu),
        ("nkn", Language::Nkn),
        ("mic", Language::Mic),
        ("keg", Language::Keg),
        ("kok", Language::Kok),
        ("awt", Language::Awt),
        ("aoe", Language::Aoe),
        ("nld", Language::Nld),
        ("aux", Language::Aux),
        ("loc", Language::Loc),
        ("ztp", Language::Ztp),
        ("pim", Language::Pim),
        ("kog", Language::Kog),
        ("sig", Language::Sig),
        ("brr", Language::Brr),
        ("inm", Language::Inm),
        ("atu", Language::Atu),
        ("khh", Language::Khh),
        ("goy", Language::Goy),
        ("sgc", Language::Sgc),
        ("daq", Language::Daq),
        ("hto", Language::Hto),
        ("rai", Language::Rai),
        ("llj", Language::Llj),
        ("cet", Language::Cet),
        ("rau", Language::Rau),
        ("row", Language::Row),
        ("snk", Language::Snk),
        ("nbg", Language::Nbg),
        ("gcd", Language::Gcd),
        ("xvi", Language::Xvi),
        ("tav", Language::Tav),
        ("cpg", Language::Cpg),
        ("txm", Language::Txm),
        ("nto", Language::Nto),
        ("naa", Language::Naa),
        ("rtm", Language::Rtm),
        ("psn", Language::Psn),
        ("xfa", Language::Xfa),
        ("mgi", Language::Mgi),
        ("une", Language::Une),
        ("zag", Language::Zag),
        ("kwg", Language::Kwg),
        ("hss", Language::Hss),
        ("hid", Language::Hid),
        ("cwe", Language::Cwe),
        ("teb", Language::Teb),
        ("ksa", Language::Ksa),
        ("zng", Language::Zng),
        ("raz", Language::Raz),
        ("mtf", Language::Mtf),
        ("wod", Language::Wod),
        ("fll", Language::Fll),
        ("sle", Language::Sle),
        ("zpa", Language::Zpa),
        ("siz", Language::Siz),
        ("awa", Language::Awa),
        ("xdq", Language::Xdq),
        ("hai", Language::Hai),
        ("txg", Language::Txg),
        ("xaq", Language::Xaq),
        ("xxr", Language::Xxr),
        ("bdq", Language::Bdq),
        ("tty", Language::Tty),
        ("lji", Language::Lji),
        ("llu", Language::Llu),
        ("zch", Language::Zch),
        ("mhq", Language::Mhq),
        ("aif", Language::Aif),
        ("pbs", Language::Pbs),
        ("yom", Language::Yom),
        ("xso", Language::Xso),
        ("sat", Language::Sat),
        ("den", Language::Den),
        ("nck", Language::Nck),
        ("adh", Language::Adh),
        ("kmc", Language::Kmc),
        ("txx", Language::Txx),
        ("cno", Language::Cno),
        ("rsl", Language::Rsl),
        ("lnz", Language::Lnz),
        ("hru", Language::Hru),
        ("iya", Language::Iya),
        ("lje", Language::Lje),
        ("dbo", Language::Dbo),
        ("ssp", Language::Ssp),
        ("gwc", Language::Gwc),
        ("xgg", Language::Xgg),
        ("ifu", Language::Ifu),
        ("sny", Language::Sny),
        ("kmw", Language::Kmw),
        ("jiu", Language::Jiu),
        ("avo", Language::Avo),
        ("diy", Language::Diy),
        ("irh", Language::Irh),
        ("myk", Language::Myk),
        ("pns", Language::Pns),
        ("bkd", Language::Bkd),
        ("sxw", Language::Sxw),
        ("sec", Language::Sec),
        ("gmg", Language::Gmg),
        ("ncr", Language::Ncr),
        ("laf", Language::Laf),
        ("qxc", Language::Qxc),
        ("bks", Language::Bks),
        ("tvw", Language::Tvw),
        ("kyg", Language::Kyg),
        ("wdg", Language::Wdg),
        ("khc", Language::Khc),
        ("bev", Language::Bev),
        ("mus", Language::Mus),
        ("waf", Language::Waf),
        ("btd", Language::Btd),
        ("bsn", Language::Bsn),
        ("mfx", Language::Mfx),
        ("ibo", Language::Ibo),
        ("bza", Language::Bza),
        ("tij", Language::Tij),
        ("nks", Language::Nks),
        ("ymm", Language::Ymm),
        ("loo", Language::Loo),
        ("gcr", Language::Gcr),
        ("xdm", Language::Xdm),
        ("dny", Language::Dny),
        ("gii", Language::Gii),
        ("xpu", Language::Xpu),
        ("mlb", Language::Mlb),
        ("toz", Language::Toz),
        ("eot", Language::Eot),
        ("njy", Language::Njy),
        ("yan", Language::Yan),
        ("dno", Language::Dno),
        ("lfn", Language::Lfn),
        ("awo", Language::Awo),
        ("fon", Language::Fon),
        ("iii", Language::Iii),
        ("rab", Language::Rab),
        ("ppq", Language::Ppq),
        ("lws", Language::Lws),
        ("ugb", Language::Ugb),
        ("srh", Language::Srh),
        ("pic", Language::Pic),
        ("wlx", Language::Wlx),
        ("ape", Language::Ape),
        ("huo", Language::Huo),
        ("cjm", Language::Cjm),
        ("ckm", Language::Ckm),
        ("bry", Language::Bry),
        ("mhi", Language::Mhi),
        ("pot", Language::Pot),
        ("kkq", Language::Kkq),
        ("swb", Language::Swb),
        ("trp", Language::Trp),
        ("pep", Language::Pep),
        ("lei", Language::Lei),
        ("lcs", Language::Lcs),
        ("psp", Language::Psp),
        ("dek", Language::Dek),
        ("pow", Language::Pow),
        ("rmm", Language::Rmm),
        ("pdo", Language::Pdo),
        ("snz", Language::Snz),
        ("wok", Language::Wok),
        ("aas", Language::Aas),
        ("arr", Language::Arr),
        ("xay", Language::Xay),
        ("smo", Language::Smo),
        ("pbf", Language::Pbf),
        ("bsw", Language::Bsw),
        ("gbh", Language::Gbh),
        ("wor", Language::Wor),
        ("hmp", Language::Hmp),
        ("sdz", Language::Sdz),
        ("mnx", Language::Mnx),
        ("ktn", Language::Ktn),
        ("lst", Language::Lst),
        ("xxk", Language::Xxk),
        ("cag", Language::Cag),
        ("htx", Language::Htx),
        ("frc", Language::Frc),
        ("bmj", Language::Bmj),
        ("tip", Language::Tip),
        ("tkf", Language::Tkf),
        ("dio", Language::Dio),
        ("ree", Language::Ree),
        ("mps", Language::Mps),
        ("bio", Language::Bio),
        ("kua", Language::Kua),
        ("lln", Language::Lln),
        ("kkz", Language::Kkz),
        ("ppo", Language::Ppo),
        ("tih", Language::Tih),
        ("bvx", Language::Bvx),
        ("mvi", Language::Mvi),
        ("prw", Language::Prw),
        ("gvm", Language::Gvm),
        ("gcl", Language::Gcl),
        ("tww", Language::Tww),
        ("mdx", Language::Mdx),
        ("fkv", Language::Fkv),
        ("jbu", Language::Jbu),
        ("kzf", Language::Kzf),
        ("wri", Language::Wri),
        ("mpw", Language::Mpw),
        ("kuu", Language::Kuu),
        ("bku", Language::Bku),
        ("odu", Language::Odu),
        ("atx", Language::Atx),
        ("moj", Language::Moj),
        ("lmw", Language::Lmw),
        ("lsc", Language::Lsc),
        ("bzg", Language::Bzg),
        ("okj", Language::Okj),
        ("xpc", Language::Xpc),
        ("fay", Language::Fay),
        ("nph", Language::Nph),
        ("keo", Language::Keo),
        ("brt", Language::Brt),
        ("myw", Language::Myw),
        ("ndu", Language::Ndu),
        ("nht", Language::Nht),
        ("lng", Language::Lng),
        ("sav", Language::Sav),
        ("xns", Language::Xns),
        ("kqi", Language::Kqi),
        ("xnb", Language::Xnb),
        ("bfo", Language::Bfo),
        ("wac", Language::Wac),
        ("tgf", Language::Tgf),
        ("skf", Language::Skf),
        ("djk", Language::Djk),
        ("kgv", Language::Kgv),
        ("clc", Language::Clc),
        ("bxu", Language::Bxu),
        ("nzy", Language::Nzy),
        ("kbi", Language::Kbi),
        ("avd", Language::Avd),
        ("gdm", Language::Gdm),
        ("aod", Language::Aod),
        ("sdg", Language::Sdg),
        ("hka", Language::Hka),
        ("cks", Language::Cks),
        ("sdj", Language::Sdj),
        ("xmn", Language::Xmn),
        ("xsc", Language::Xsc),
        ("bco", Language::Bco),
        ("atw", Language::Atw),
        ("vto", Language::Vto),
        ("ehs", Language::Ehs),
        ("xau", Language::Xau),
        ("pzn", Language::Pzn),
        ("mqb", Language::Mqb),
        ("ojb", Language::Ojb),
        ("lzn", Language::Lzn),
        ("esi", Language::Esi),
        ("tdq", Language::Tdq),
        ("idc", Language::Idc),
        ("nbp", Language::Nbp),
        ("aol", Language::Aol),
        ("mde", Language::Mde),
        ("wyn", Language::Wyn),
        ("odk", Language::Odk),
        ("srw", Language::Srw),
        ("vbk", Language::Vbk),
        ("pio", Language::Pio),
        ("mrk", Language::Mrk),
        ("ymq", Language::Ymq),
        ("nkq", Language::Nkq),
        ("mqo", Language::Mqo),
        ("hag", Language::Hag),
        ("ruq", Language::Ruq),
        ("ugy", Language::Ugy),
        ("arl", Language::Arl),
        ("nsv", Language::Nsv),
        ("kvh", Language::Kvh),
        ("xnr", Language::Xnr),
        ("tgi", Language::Tgi),
        ("beb", Language::Beb),
        ("qve", Language::Qve),
        ("tsz", Language::Tsz),
        ("mql", Language::Mql),
        ("jct", Language::Jct),
        ("guw", Language::Guw),
        ("bdu", Language::Bdu),
        ("zae", Language::Zae),
        ("hmq", Language::Hmq),
        ("dym", Language::Dym),
        ("osa", Language::Osa),
        ("zln", Language::Zln),
        ("xur", Language::Xur),
        ("brb", Language::Brb),
        ("dja", Language::Dja),
        ("dyb", Language::Dyb),
        ("avl", Language::Avl),
        ("kwz", Language::Kwz),
        ("knj", Language::Knj),
        ("gnz", Language::Gnz),
        ("lga", Language::Lga),
        ("teo", Language::Teo),
        ("rmi", Language::Rmi),
        ("sou", Language::Sou),
        ("fos", Language::Fos),
        ("lre", Language::Lre),
        ("jmn", Language::Jmn),
        ("cwa", Language::Cwa),
        ("tyv", Language::Tyv),
        ("ynq", Language::Ynq),
        ("nli", Language::Nli),
        ("cak", Language::Cak),
        ("csx", Language::Csx),
        ("kkf", Language::Kkf),
        ("huv", Language::Huv),
        ("pbh", Language::Pbh),
        ("pqm", Language::Pqm),
        ("tfi", Language::Tfi),
        ("kbl", Language::Kbl),
        ("hay", Language::Hay),
        ("kkx", Language::Kkx),
        ("naz", Language::Naz),
        ("pwm", Language::Pwm),
        ("hmf", Language::Hmf),
        ("qvj", Language::Qvj),
        ("ked", Language::Ked),
        ("wew", Language::Wew),
        ("mfw", Language::Mfw),
        ("juo", Language::Juo),
        ("zrg", Language::Zrg),
        ("dec", Language::Dec),
        ("klz", Language::Klz),
        ("crn", Language::Crn),
        ("bvc", Language::Bvc),
        ("clo", Language::Clo),
        ("niw", Language::Niw),
        ("xbn", Language::Xbn),
        ("rki", Language::Rki),
        ("pey", Language::Pey),
        ("gce", Language::Gce),
        ("tcc", Language::Tcc),
        ("mnp", Language::Mnp),
        ("too", Language::Too),
        ("lsl", Language::Lsl),
        ("kuw", Language::Kuw),
        ("reg", Language::Reg),
        ("emq", Language::Emq),
        ("dms", Language::Dms),
        ("ogo", Language::Ogo),
        ("nfu", Language::Nfu),
        ("taa", Language::Taa),
        ("kix", Language::Kix),
        ("cld", Language::Cld),
        ("lue", Language::Lue),
        ("xwe", Language::Xwe),
        ("dbe", Language::Dbe),
        ("lsw", Language::Lsw),
        ("vmj", Language::Vmj),
        ("gas", Language::Gas),
        ("mlp", Language::Mlp),
        ("kfw", Language::Kfw),
        ("ljp", Language::Ljp),
        ("nym", Language::Nym),
        ("spd", Language::Spd),
        ("bdm", Language::Bdm),
        ("bju", Language::Bju),
        ("doc", Language::Doc),
        ("ruf", Language::Ruf),
        ("ndk", Language::Ndk),
        ("aao", Language::Aao),
        ("bpw", Language::Bpw),
        ("urp", Language::Urp),
        ("csk", Language::Csk),
        ("cnc", Language::Cnc),
        ("kaa", Language::Kaa),
        ("npi", Language::Npi),
        ("slu", Language::Slu),
        ("nql", Language::Nql),
        ("fia", Language::Fia),
        ("puj", Language::Puj),
        ("cha", Language::Cha),
        ("adg", Language::Adg),
        ("nbi", Language::Nbi),
        ("anw", Language::Anw),
        ("fvr", Language::Fvr),
        ("aab", Language::Aab),
        ("xpw", Language::Xpw),
        ("sez", Language::Sez),
        ("nau", Language::Nau),
        ("klj", Language::Klj),
        ("tet", Language::Tet),
        ("byw", Language::Byw),
        ("pua", Language::Pua),
        ("ahi", Language::Ahi),
        ("kwu", Language::Kwu),
        ("dov", Language::Dov),
        ("njb", Language::Njb),
        ("mrx", Language::Mrx),
        ("tbp", Language::Tbp),
        ("twd", Language::Twd),
        ("qur", Language::Qur),
        ("nhu", Language::Nhu),
        ("xzh", Language::Xzh),
        ("jnl", Language::Jnl),
        ("jbo", Language::Jbo),
        ("ncm", Language::Ncm),
        ("tlq", Language::Tlq),
        ("jaq", Language::Jaq),
        ("smq", Language::Smq),
        ("tal", Language::Tal),
        ("ami", Language::Ami),
        ("spx", Language::Spx),
        ("woa", Language::Woa),
        ("pop", Language::Pop),
        ("gum", Language::Gum),
        ("wth", Language::Wth),
        ("urh", Language::Urh),
        ("xuu", Language::Xuu),
        ("mdk", Language::Mdk),
        ("akh", Language::Akh),
        ("bty", Language::Bty),
        ("naj", Language::Naj),
        ("cpa", Language::Cpa),
        ("dun", Language::Dun),
        ("opk", Language::Opk),
        ("bzz", Language::Bzz),
        ("xhd", Language::Xhd),
        ("asq", Language::Asq),
        ("dru", Language::Dru),
        ("xeg", Language::Xeg),
        ("age", Language::Age),
        ("xtp", Language::Xtp),
        ("hnh", Language::Hnh),
        ("pog", Language::Pog),
        ("tbz", Language::Tbz),
        ("cjn", Language::Cjn),
        ("hei", Language::Hei),
        ("nmm", Language::Nmm),
        ("kid", Language::Kid),
        ("gqu", Language::Gqu),
        ("hil", Language::Hil),
        ("pmj", Language::Pmj),
        ("wlk", Language::Wlk),
        ("two", Language::Two),
        ("bqs", Language::Bqs),
        ("tyr", Language::Tyr),
        ("gsl", Language::Gsl),
        ("bva", Language::Bva),
        ("mpi", Language::Mpi),
        ("pei", Language::Pei),
        ("kxf", Language::Kxf),
        ("xkl", Language::Xkl),
        ("ich", Language::Ich),
        ("cic", Language::Cic),
        ("kaq", Language::Kaq),
        ("nez", Language::Nez),
        ("nyd", Language::Nyd),
        ("pil", Language::Pil),
        ("psu", Language::Psu),
        ("tgp", Language::Tgp),
        ("ads", Language::Ads),
        ("ljx", Language::Ljx),
        ("ior", Language::Ior),
        ("ibg", Language::Ibg),
        ("dmd", Language::Dmd),
        ("paz", Language::Paz),
        ("osi", Language::Osi),
        ("sac", Language::Sac),
        ("bek", Language::Bek),
        ("the", Language::The),
        ("gou", Language::Gou),
        ("cdj", Language::Cdj),
        ("csd", Language::Csd),
        ("hvc", Language::Hvc),
        ("dhg", Language::Dhg),
        ("tdg", Language::Tdg),
        ("gnw", Language::Gnw),
        ("glw", Language::Glw),
        ("gaz", Language::Gaz),
        ("auc", Language::Auc),
        ("mhr", Language::Mhr),
        ("tau", Language::Tau),
        ("nda", Language::Nda),
        ("kih", Language::Kih),
        ("bqt", Language::Bqt),
        ("geh", Language::Geh),
        ("dwu", Language::Dwu),
        ("khe", Language::Khe),
        ("hak", Language::Hak),
        ("bla", Language::Bla),
        ("vgr", Language::Vgr),
        ("lur", Language::Lur),
        ("egy", Language::Egy),
        ("txb", Language::Txb),
        ("nbh", Language::Nbh),
        ("otx", Language::Otx),
        ("elk", Language::Elk),
        ("nhc", Language::Nhc),
        ("jui", Language::Jui),
        ("soq", Language::Soq),
        ("ayg", Language::Ayg),
        ("kzc", Language::Kzc),
        ("fap", Language::Fap),
        ("lce", Language::Lce),
        ("fla", Language::Fla),
        ("zmm", Language::Zmm),
        ("ins", Language::Ins),
        ("pwg", Language::Pwg),
        ("gnr", Language::Gnr),
        ("ito", Language::Ito),
        ("keb", Language::Keb),
        ("xcu", Language::Xcu),
        ("wdu", Language::Wdu),
        ("ttj", Language::Ttj),
        ("ron", Language::Ron),
        ("bkj", Language::Bkj),
        ("umu", Language::Umu),
        ("ddn", Language::Ddn),
        ("ksk", Language::Ksk),
        ("tyl", Language::Tyl),
        ("mzq", Language::Mzq),
        ("mji", Language::Mji),
        ("dib", Language::Dib),
        ("xaw", Language::Xaw),
        ("mlj", Language::Mlj),
        ("muy", Language::Muy),
        ("tlp", Language::Tlp),
        ("wni", Language::Wni),
        ("aru", Language::Aru),
        ("lev", Language::Lev),
        ("aqz", Language::Aqz),
        ("kjl", Language::Kjl),
        ("ymk", Language::Ymk),
        ("tsi", Language::Tsi),
        ("wgb", Language::Wgb),
        ("alw", Language::Alw),
        ("nkv", Language::Nkv),
        ("msj", Language::Msj),
        ("jet", Language::Jet),
        ("emy", Language::Emy),
        ("urf", Language::Urf),
        ("wiy", Language::Wiy),
        ("hmd", Language::Hmd),
        ("buk", Language::Buk),
        ("jje", Language::Jje),
        ("nna", Language::Nna),
        ("sba", Language::Sba),
        ("clj", Language::Clj),
        ("bcc", Language::Bcc),
        ("kal", Language::Kal),
        ("nnv", Language::Nnv),
        ("sdr", Language::Sdr),
        ("ary", Language::Ary),
        ("drb", Language::Drb),
        ("kqn", Language::Kqn),
        ("goc", Language::Goc),
        ("ksv", Language::Ksv),
        ("hvk", Language::Hvk),
        ("mnm", Language::Mnm),
        ("tpt", Language::Tpt),
        ("bcv", Language::Bcv),
        ("com", Language::Com),
        ("bbf", Language::Bbf),
        ("tgs", Language::Tgs),
        ("djn", Language::Djn),
        ("opt", Language::Opt),
        ("oko", Language::Oko),
        ("nnw", Language::Nnw),
        ("enc", Language::Enc),
        ("bbb", Language::Bbb),
        ("xlu", Language::Xlu),
        ("siy", Language::Siy),
        ("qyp", Language::Qyp),
        ("ubu", Language::Ubu),
        ("niz", Language::Niz),
        ("mxq", Language::Mxq),
        ("dhs", Language::Dhs),
        ("zko", Language::Zko),
        ("zts", Language::Zts),
        ("tpj", Language::Tpj),
        ("brf", Language::Brf),
        ("log", Language::Log),
        ("mfo", Language::Mfo),
        ("dam", Language::Dam),
        ("pel", Language::Pel),
        ("vai", Language::Vai),
        ("zas", Language::Zas),
        ("pcf", Language::Pcf),
        ("xdy", Language::Xdy),
        ("kmd", Language::Kmd),
        ("hrk", Language::Hrk),
        ("shj", Language::Shj),
        ("hom", Language::Hom),
        ("mbh", Language::Mbh),
        ("nci", Language::Nci),
        ("sbq", Language::Sbq),
        ("dzo", Language::Dzo),
        ("kqt", Language::Kqt),
        ("trd", Language::Trd),
        ("jni", Language::Jni),
        ("kvi", Language::Kvi),
        ("pma", Language::Pma),
        ("buc", Language::Buc),
        ("zte", Language::Zte),
        ("ark", Language::Ark),
        ("gmb", Language::Gmb),
        ("spp", Language::Spp),
        ("lki", Language::Lki),
        ("hmn", Language::Hmn),
        ("aoi", Language::Aoi),
        ("ngn", Language::Ngn),
        ("vic", Language::Vic),
        ("vav", Language::Vav),
        ("gek", Language::Gek),
        ("dta", Language::Dta),
        ("vah", Language::Vah),
        ("ncs", Language::Ncs),
        ("kba", Language::Kba),
        ("kzl", Language::Kzl),
        ("tqq", Language::Tqq),
        ("alq", Language::Alq),
        ("dkx", Language::Dkx),
        ("pia", Language::Pia),
        ("tus", Language::Tus),
        ("kvu", Language::Kvu),
        ("mpo", Language::Mpo),
        ("pub", Language::Pub),
        ("jup", Language::Jup),
        ("hyw", Language::Hyw),
        ("tst", Language::Tst),
        ("ven", Language::Ven),
        ("tqo", Language::Tqo),
        ("frq", Language::Frq),
        ("bor", Language::Bor),
        ("gdg", Language::Gdg),
        ("sbr", Language::Sbr),
        ("bgc", Language::Bgc),
        ("lgq", Language::Lgq),
        ("huf", Language::Huf),
        ("xsl", Language::Xsl),
        ("mky", Language::Mky),
        ("bmc", Language::Bmc),
        ("mxe", Language::Mxe),
        ("tto", Language::Tto),
        ("xre", Language::Xre),
        ("moe", Language::Moe),
        ("bqw", Language::Bqw),
        ("neh", Language::Neh),
        ("cmo", Language::Cmo),
        ("nre", Language::Nre),
        ("knz", Language::Knz),
        ("pnr", Language::Pnr),
        ("wgo", Language::Wgo),
        ("nkh", Language::Nkh),
        ("neo", Language::Neo),
        ("rui", Language::Rui),
        ("ckr", Language::Ckr),
        ("hea", Language::Hea),
        ("sfe", Language::Sfe),
        ("mon", Language::Mon),
        ("bsk", Language::Bsk),
        ("raq", Language::Raq),
        ("ncj", Language::Ncj),
        ("yiv", Language::Yiv),
        ("riu", Language::Riu),
        ("jim", Language::Jim),
        ("val", Language::Val),
        ("mmu", Language::Mmu),
        ("bqm", Language::Bqm),
        ("nou", Language::Nou),
        ("krn", Language::Krn),
        ("ikv", Language::Ikv),
        ("bqa", Language::Bqa),
        ("mea", Language::Mea),
        ("kom", Language::Kom),
        ("vmr", Language::Vmr),
        ("bdz", Language::Bdz),
        ("ntz", Language::Ntz),
        ("njz", Language::Njz),
        ("nay", Language::Nay),
        ("nrt", Language::Nrt),
        ("ptw", Language::Ptw),
    ],
};
